/*  createhtml - library to generate html files
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// HtmlAttributes.h: interface for the HtmlAttributes classes.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__HTML_ATTRIBUTES_H__INCLUDED_)
#define __HTML_ATTRIBUTES_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "HtmlAttribute.h" // dep_chk_unknown 
#include "UrlItem.h" // dep_chk_needed 

#include <string>

using namespace std;

#define DECLARE_HTML_ATTRIB_STORAGE(nametype,type,reftype) \
	class HtmlAttrib__##nametype : public HtmlAttribute \
	{ \
	public: \
		HtmlAttrib__##nametype(const reftype value) { m_value = value; } \
		virtual ~HtmlAttrib__##nametype() { } \
	protected: \
		type m_value;

#define DECLARE_HTML_ATTRIB_STORAGE2(nametype,type,reftype,nametype2,type2,reftype2) \
	class HtmlAttrib__##nametype##_##nametype2 : public HtmlAttribute \
	{ \
	public: \
		HtmlAttrib__##nametype##_##nametype2(const reftype value,const reftype2 value2) { m_value = value; m_value2 = value2; } \
		virtual ~HtmlAttrib__##nametype##_##nametype2() { } \
	protected: \
		type m_value; \
		type2 m_value2;

/*	protected: \
		HtmlAttrib__##nametype() { } \*/

#define RENDER_HTML_ATTRIB_STORAGE \
	public: virtual void Render (ostream* pOut,HtmlPage* pPage) { HtmlAttribute::Render(pOut,pPage); (*pOut) << '=' << m_value; }
#define END_HTML_ATTRIB_STORAGE \
	};

#define DECLARE_HTML_ATTRIB(name,nametype,reftype) \
	class HtmlAttrib__##nametype##_##name : public HtmlAttrib__##nametype \
	{ \
	public: \
	HtmlAttrib__##nametype##_##name(const reftype value) : HtmlAttrib__##nametype(value) { } \
		virtual ~HtmlAttrib__##nametype##_##name() { } \
	public: \
		virtual const char* GetName () { return #name; } \
	};
/*	protected: \
		HtmlAttrib__##nametype##_##name() { } \ */

#define DECLARE_HTML_ATTRIB2(name,nametype,reftype,nametype2,reftype2) \
	class HtmlAttrib__##nametype##_##nametype2##_##name : public HtmlAttrib__##nametype##_##nametype2 \
	{ \
	public: \
	HtmlAttrib__##nametype##_##nametype2##_##name(const reftype value,const reftype2 value2) : HtmlAttrib__##nametype##_##nametype2(value,value2) { } \
		virtual ~HtmlAttrib__##nametype##_##nametype2##_##name() { } \
	public: \
		virtual const char* GetName () { return #name; } \
	};

DECLARE_HTML_ATTRIB_STORAGE(Integer,int,int)
RENDER_HTML_ATTRIB_STORAGE
END_HTML_ATTRIB_STORAGE

DECLARE_HTML_ATTRIB_STORAGE(String,string,string&)
RENDER_HTML_ATTRIB_STORAGE
END_HTML_ATTRIB_STORAGE

DECLARE_HTML_ATTRIB_STORAGE(UrlItemPtr,const UrlItem*,UrlItem*)
	public: virtual void Render (ostream* pOut,HtmlPage* pPage);
END_HTML_ATTRIB_STORAGE

DECLARE_HTML_ATTRIB_STORAGE2(UrlItemPtr,const UrlItem*,UrlItem*,String,string,string&)
	public: virtual void Render (ostream* pOut,HtmlPage* pPage);
END_HTML_ATTRIB_STORAGE

DECLARE_HTML_ATTRIB(href   ,UrlItemPtr,UrlItem*)
DECLARE_HTML_ATTRIB(src    ,UrlItemPtr,UrlItem*)

DECLARE_HTML_ATTRIB2(href  ,UrlItemPtr,UrlItem*,String,string&)

DECLARE_HTML_ATTRIB(href   ,String,string&)

DECLARE_HTML_ATTRIB(src    ,String,string&)
DECLARE_HTML_ATTRIB(alt    ,String,string&)
DECLARE_HTML_ATTRIB(name   ,String,string&)

DECLARE_HTML_ATTRIB(width  ,Integer,int)
DECLARE_HTML_ATTRIB(height ,Integer,int)
DECLARE_HTML_ATTRIB(border ,Integer,int)
DECLARE_HTML_ATTRIB(colspan,Integer,int)
DECLARE_HTML_ATTRIB(rowspan,Integer,int)

	/*
class HtmlImage;

typedef HtmlImage image;

class HtmlImage : public HtmlTag
{
public:
	HtmlImage();
	virtual ~HtmlImage();

public:
	HtmlImage& src(HtmlURL& url) { AddAttribute (ATTR_SRC,url); return *this; }
};
*/
/*
#define DECLARE_HTML_UNARY_TAG(name) \
	class Html_##name;\
	typedef Html_##name name;\
	class Html_##name : public HtmlUnaryTag\
	{\
	public:\
		Html_##name() { }\
		virtual ~Html_##name() { } \
		virtual const char* GetName () { return #name; }
			

#define DECLARE_HTML_BINARY_TAG(name) \
	class Html_##name;\
	typedef Html_##name name;\
	class Html_##name : public HtmlBinaryTag\
	{\
	public:\
		Html_##name() { }\
		Html_##name(const string& strText) : HtmlBinaryTag(strText) { }\
		Html_##name(const char* pText) : HtmlBinaryTag(pText) { }\
		virtual ~Html_##name() { }\
		virtual const char* GetName () { return #name; }

#define END_HTML_TAG(name) \
	public: virtual HtmlObject* Copy () { return new Html_##name(*this); } \
	};
//	friend HtmlTag& operator<<(HtmlTag& tag1,Html_##name& tag2) { tag1 << new Html_##name(tag2); return tag1; } \

#define HTML_TAG_ATTRIBUTE(name,attrib,type) public: Html_##name& attrib(type& url) { AddAttribute (ATTR_SRC,url); return *this; }

#define HTML_TAG_ATTRIBUTE_URL(name,attrib) \
	HTML_TAG_ATTRIBUTE(name,attrib,HtmlURL)
#define HTML_TAG_ATTRIBUTE_TEXT(name,attrib) \
	HTML_TAG_ATTRIBUTE(name,attrib,const char*) \
	HTML_TAG_ATTRIBUTE(name,attrib,string)
#define HTML_TAG_ATTRIBUTE_HEX(name,attrib) \
	HTML_TAG_ATTRIBUTE(name,attrib,dword)

DECLARE_HTML_BINARY_TAG(html)
END_HTML_TAG(html)

DECLARE_HTML_BINARY_TAG(head)
END_HTML_TAG(head)

DECLARE_HTML_BINARY_TAG(title)
END_HTML_TAG(title)

DECLARE_HTML_BINARY_TAG(body)
HTML_TAG_ATTRIBUTE_HEX (body,bgcolor)
END_HTML_TAG(body)

DECLARE_HTML_UNARY_TAG (img)
HTML_TAG_ATTRIBUTE_URL (img,src)
HTML_TAG_ATTRIBUTE_TEXT(img,alt)
END_HTML_TAG(img)

DECLARE_HTML_BINARY_TAG(a)
HTML_TAG_ATTRIBUTE_URL (a,href)
END_HTML_TAG(a)

DECLARE_HTML_UNARY_TAG (br)
END_HTML_TAG(br)

DECLARE_HTML_BINARY_TAG(p)
END_HTML_TAG(p)

DECLARE_HTML_BINARY_TAG(table)
END_HTML_TAG(table)

DECLARE_HTML_BINARY_TAG(tr)
END_HTML_TAG(tr)

DECLARE_HTML_BINARY_TAG(td)
END_HTML_TAG(td)

DECLARE_HTML_BINARY_TAG(h1)
END_HTML_TAG(h1)

DECLARE_HTML_BINARY_TAG(h2)
END_HTML_TAG(h2)

DECLARE_HTML_BINARY_TAG(h3)
END_HTML_TAG(h3)

DECLARE_HTML_BINARY_TAG(h4)
END_HTML_TAG(h4)

DECLARE_HTML_BINARY_TAG(h5)
END_HTML_TAG(h5)

DECLARE_HTML_BINARY_TAG(h6)
END_HTML_TAG(h6)

*/






#endif // !defined(__HTML_ATTRIBUTES_H__INCLUDED_)

