/*  createhtml - library to generate html files
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// HtmlPage.h: interface for the HtmlPage class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__HTML_PAGE_H__INCLUDED_)
#define __HTML_PAGE_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Common.h"
#include "HtmlObject.h" // dep_chk_needed 
#include "HtmlTag.h" // dep_chk_needed 
#include "HtmlBinaryTag.h" // dep_chk_needed 
// #include "HtmlSimpleText.h" // dep_chk_unknown 

class UrlFolder; // from  #include "UrlFolder.h"  by dep_chk 
class UrlFile;

#include <vector>
#include <stack>
#include <string>

using namespace std;

class HtmlPage  
{
public:
	HtmlPage();
	HtmlPage(UrlFile* pFile);
	virtual ~HtmlPage();

protected:
	ostream* m_pOutStream;
	bool m_bDeleteOutStream;

	UrlFile* m_pFile;

protected:
	vector<HtmlObject*> m_vectObjects;
	stack<HtmlBinaryTag*> m_stackBegunTags;

public:
	void SetOutStream (ostream* pOutStream,bool bDeleteOutStream = false);

public:
	UrlFile* GetFile () { return m_pFile; }

public:
	void AddObject(HtmlObject* pObject);
	void AddObject(HtmlObject& object) { AddObject (object.Copy()); }
	void AddObject(const HtmlTagPtr& object) { AddObject (object.ptr->Copy()); }

	void AddText(const string& strText);
	void AddText(const char* pText);
	void AddText(char token);

	void Begin (HtmlBinaryTag* pTag);
	void Begin (const HtmlBinaryTagPtr& tag) { Begin ((HtmlBinaryTag*) tag.ptr->Copy()); }
	void End ();

	void Header (const string& strText,bool bAlsoHeading = true);
	void Header (const char* pText,bool bAlsoHeading = true) { string str(pText); Header(str,bAlsoHeading); }
	void Header (bool bAlsoHeading = true);
	void HeaderNoBody ();

	HtmlPage& operator<<(HtmlObject* pObject) { AddObject (pObject); return *this; }
	HtmlPage& operator<<(HtmlObject& object) { AddObject (object.Copy ()); return *this; }

	HtmlPage& operator<<(const HtmlTagPtr& object) { AddObject (object.ptr->Copy ()); return *this; }

	HtmlPage& operator<<(const string& strText) { AddText (strText); return *this; }
	HtmlPage& operator<<(const char* pText) { AddText (pText); return *this; }
	HtmlPage& operator<<(char token) { AddText (token); return *this; }

public:
	void LinksToFiles (UrlFolder* pFolder,bool bOnlyFolders = false);
//	friend HtmlPage& operator+(HtmlPage& page,HtmlTag& tag);
};


/*
HtmlPage& operator<<(HtmlPage& page,HtmlObject* pObject);
HtmlPage& operator<<(HtmlPage& page,HtmlObject& object);
*/
#endif // !defined(__HTML_PAGE_H__INCLUDED_)


/*

<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN" "http://www.w3.org/TR/REC-html40/strict.dtd">
<html lang="en" dir="ltr" >
<head>
<title>GNU cgicc v3.2</title>
</head>



  */

