/*  createhtml - library to generate html files
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// HtmlTag.cpp: implementation of the HtmlTag class.
//
//////////////////////////////////////////////////////////////////////

#include "Common.h"
#include "HtmlTag.h" // dep_chk_needed 
#include "HtmlAttribute.h"
#include "HtmlSimpleText.h"
#include <algorithm>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

HtmlTag::HtmlTag()
{

}

void DeleteHtmlAttrib (HtmlAttribute* pAttrib)
{
	assert (pAttrib);
//	delete pAttrib;
}

HtmlTag::~HtmlTag()
{
	for_each (m_vectAttributes.begin(),m_vectAttributes.end(),DeleteHtmlAttrib);
}

void RenderHtmlAttrib (HtmlAttribute* pAttrib,ostream* pOut,HtmlPage* pPage)
{
	assert (pAttrib);
	pAttrib->Render (pOut,pPage);
}

void HtmlTag::RenderStartTag (ostream* pOut,HtmlPage* pPage)
{
	(*pOut) << '<' << GetName ();

	for_each (m_vectAttributes.begin(),m_vectAttributes.end(),RenderHtmlAttrib,pOut,pPage);

	(*pOut) << '>';
}


void HtmlTag::Render (ostream* pOut,HtmlPage* pPage)
{
	RenderStartTag(pOut,pPage);
}

void HtmlTag::AddObject(HtmlObject* pObject)
{
    if (pObject) // NULL objects are simply nothing
        m_vectObjects.push_back(pObject);
}

void HtmlTag::AddText(const string& strText)
{
	AddObject (new HtmlSimpleText (strText));
}

void HtmlTag::AddText(const char* pText)
{
    assert (pText);

        AddObject (new HtmlSimpleText (pText));
}

void HtmlTag::AddText(char token)
{
	AddObject (new HtmlSimpleText (token));
}

void HtmlTag::AddAttribute(HtmlAttribute* pAttrib)
{
	m_vectAttributes.push_back(pAttrib);
}

void HtmlTag::DoCopy (HtmlTag* pDest) const
{
  pDest->m_vectObjects = m_vectObjects;
  pDest->m_vectAttributes = m_vectAttributes;
}

//////////////////////////////////////////////////////////////////////

#include <stdio.h>

const HtmlTagPtr& operator<<(const HtmlTagPtr& tag1,const HtmlTagPtr& tag2) 
{
  tag1.ptr->AddObject (tag2.ptr->Copy ());
  return tag1; 
}

const HtmlTagPtr& operator<<(const HtmlTagPtr& tag1,HtmlObject* obj2)
{
  tag1.ptr->AddObject (obj2);
  return tag1; 
}





