/*  createhtml - library to generate html files
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// HtmlTag.h: interface for the HtmlTag class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__HTML_TAG_H__INCLUDED_)
#define __HTML_TAG_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

// #include "Common.h" // dep_chk_unknown 
#include "HtmlObject.h" // dep_chk_needed 
class HtmlAttribute; // from  #include "HtmlAttribute.h"  by dep_chk 

#include <vector>
#include <string>

using namespace std;

class HtmlTag;

class HtmlTagPtr 
{ 
 public:
  HtmlTagPtr() { }
  HtmlTagPtr(HtmlTag* src) { ptr = src; }
  HtmlTagPtr(const HtmlTagPtr& src) { ptr = src.ptr; }
  ~HtmlTagPtr() { }

 public:
  HtmlTagPtr& operator=(const HtmlTagPtr& src) { ptr = src.ptr; return *this; }

  HtmlTag* ptr;

  operator HtmlTag*() const { return ptr; }
  operator HtmlTag&() const
  {
      return *ptr;
  }

};

class HtmlTag : public HtmlObject
{
protected:
	HtmlTag();

public:
	//HtmlTag(const HtmlTag& src) { }
	virtual ~HtmlTag();

	//HtmlTag& operator=(const HtmlTag& src) { return *this; }

	operator HtmlTagPtr() { return HtmlTagPtr(this); }

protected:
	vector<HtmlObject*> m_vectObjects;
	vector<HtmlAttribute*> m_vectAttributes;

public:
	void AddObject(HtmlObject* pObject);
	void AddObject(const HtmlObject& object) { AddObject (object.Copy()); }

	void AddText(const string& strText);
	void AddText(const char* pText);
	void AddText(char token);

	void AddAttribute(HtmlAttribute* pAttrib);
/*
	void AddAttribute(int m_nType,HtmlURL& url);
	void AddAttribute(int m_nType,const char*& str);
	void AddAttribute(int m_nType,string& str);
	void AddAttribute(int m_nType,dword hex);
*/

public:
	virtual void Render (ostream* pOut,HtmlPage* pPage);

	void RenderStartTag (ostream* pOut,HtmlPage* pPage);

public:
	void DoCopy (HtmlTag* pDest) const;

	virtual const char* GetName () const = 0; 

public:
	//HtmlTag& operator<<(HtmlObject* pObject) { AddObject (pObject); return *this; }
	//const HtmlTag& operator<<(const HtmlObject* pObject) const { return *this; }

};

const HtmlTagPtr& operator<<(const HtmlTagPtr& tag1,const HtmlTagPtr& tag2);
const HtmlTagPtr& operator<<(const HtmlTagPtr& tag1,HtmlObject* obj2);

inline const HtmlTagPtr& operator<<(const HtmlTagPtr& tag1,const string& strText) { tag1.ptr->AddText (strText); return tag1; }
inline const HtmlTagPtr& operator<<(const HtmlTagPtr& tag1,const char* pText) { tag1.ptr->AddText (pText); return tag1; }
inline const HtmlTagPtr& operator<<(const HtmlTagPtr& tag1,char token) { tag1.ptr->AddText (token); return tag1; }

/*
HtmlTag& operator<<(HtmlTag& tag1,HtmlObject* pObject);
HtmlTag& operator<<(HtmlTag& tag1,HtmlObject& object);

HtmlTag& operator<<(HtmlTag& tag1,HtmlObject* pObject)
{
	tag1.AddObject (pObject);	
	return tag1;
}

HtmlTag& operator<<(HtmlTag& tag1,HtmlObject& object)
{
	tag1.AddObject (object.Copy ());
	return tag1;
}
*/
#endif // !defined(__HTML_TAG_H__INCLUDED_)



