/*  createhtml - library to generate html files
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// HtmlTags.h: interface for the HtmlTags classes.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__HTML_TAGS_H__INCLUDED_)
#define __HTML_TAGS_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "HtmlBinaryTag.h" // dep_chk_unknown 
#include "HtmlUnaryTag.h" // dep_chk_unknown 

#include "HtmlAttributes.h" // dep_chk_unknown 

/*
class HtmlImage;

typedef HtmlImage image;

class HtmlImage : public HtmlTag
{
public:
	HtmlImage();
	virtual ~HtmlImage();

public:
	HtmlImage& src(HtmlURL& url) { AddAttribute (ATTR_SRC,url); return *this; }
};
*/

#define DECLARE_HTML_TAG(name,basename) \
	class Html_##name;\
	typedef Html_##name name;\
	class Html_##name : public basename\
	{\
	public:\
		Html_##name() { }\
		virtual ~Html_##name() { } \
		virtual const char* GetName () const { return #name; } \
	public:\
		Html_##name& attrib(const string& strName,const string& strData) { AddAttribute (new HtmlAttribGeneric(strName,strData)); return *this; }

#define DECLARE_HTML_UNARY_TAG(name) \
	DECLARE_HTML_TAG(name,HtmlUnaryTag)
			

#define DECLARE_HTML_BINARY_TAG(name) \
	DECLARE_HTML_TAG(name,HtmlBinaryTag) \
	public:\
		Html_##name(const string& strText) : HtmlBinaryTag(strText) { }\
		Html_##name(const char* pText) : HtmlBinaryTag(pText) { }\

#define END_HTML_TAG(name) \
	public: virtual HtmlObject* Copy () const { Html_##name* pNew = new Html_##name(*this); HtmlTag::DoCopy(pNew); return pNew; } \
	};

#define DECLARE_HTML_SIMPLE_UNARY_TAG(name) \
DECLARE_HTML_UNARY_TAG (name) \
END_HTML_TAG(name)

#define DECLARE_HTML_SIMPLE_BINARY_TAG(name) \
DECLARE_HTML_BINARY_TAG (name) \
END_HTML_TAG(name)

// HtmlTag& operator<<(HtmlTag& tag1,const Html_##name& tag2) { tag1.AddObject (tag2.Copy()); return tag1; }

#define HTML_TAG_ATTRIBUTE(name,attrib,nametype,type) public: Html_##name& attrib(type data) { AddAttribute (new HtmlAttrib__##nametype##_##attrib(data)); return *this; }
#define HTML_TAG_ATTRIBUTE2(name,attrib,nametype,type,nametype2,type2) public: Html_##name& attrib(type data,type2 data2) { AddAttribute (new HtmlAttrib__##nametype##_##nametype2##_##attrib(data,data2)); return *this; }

#define HTML_TAG_ATTRIBUTE_TEXT(name,attrib) \
	HTML_TAG_ATTRIBUTE(name,attrib,String,const char*) \
	HTML_TAG_ATTRIBUTE(name,attrib,String,const string&)
#define HTML_TAG_ATTRIBUTE_URL(name,attrib) \
	HTML_TAG_ATTRIBUTE(name,attrib,UrlItemPtr,UrlItem*) 
#define HTML_TAG_ATTRIBUTE_URL_TEXT(name,attrib) \
	HTML_TAG_ATTRIBUTE2(name,attrib,UrlItemPtr,UrlItem*,String,const char*) \
	HTML_TAG_ATTRIBUTE2(name,attrib,UrlItemPtr,UrlItem*,String,const string&) 
#define HTML_TAG_ATTRIBUTE_INT(name,attrib) \
	HTML_TAG_ATTRIBUTE(name,attrib,Integer,int) 

/*
#define HTML_TAG_ATTRIBUTE_URL(name,attrib) \
	HTML_TAG_ATTRIBUTE(name,attrib,HtmlURL)
#define HTML_TAG_ATTRIBUTE_HEX(name,attrib) \
	HTML_TAG_ATTRIBUTE(name,attrib,dword)
*/

DECLARE_HTML_BINARY_TAG(html)
END_HTML_TAG(html)

DECLARE_HTML_BINARY_TAG(head)
END_HTML_TAG(head)

DECLARE_HTML_BINARY_TAG(title)
END_HTML_TAG(title)

DECLARE_HTML_BINARY_TAG(body)
//HTML_TAG_ATTRIBUTE_HEX (body,bgcolor)
END_HTML_TAG(body)

DECLARE_HTML_UNARY_TAG (img)
//HTML_TAG_ATTRIBUTE     (img,src)
HTML_TAG_ATTRIBUTE_URL (img,src)
HTML_TAG_ATTRIBUTE_TEXT(img,src)
HTML_TAG_ATTRIBUTE_TEXT(img,alt)
HTML_TAG_ATTRIBUTE_INT (img,width)
HTML_TAG_ATTRIBUTE_INT (img,height)
END_HTML_TAG(img)

DECLARE_HTML_BINARY_TAG(a)
HTML_TAG_ATTRIBUTE_URL (a,href)
HTML_TAG_ATTRIBUTE_TEXT(a,href)
HTML_TAG_ATTRIBUTE_TEXT(a,name)
HTML_TAG_ATTRIBUTE_URL_TEXT (a,href)
END_HTML_TAG(a)

DECLARE_HTML_BINARY_TAG(table)
HTML_TAG_ATTRIBUTE_INT (table,border)
END_HTML_TAG(table)

DECLARE_HTML_BINARY_TAG(tr)
END_HTML_TAG(tr)

DECLARE_HTML_BINARY_TAG(td)
HTML_TAG_ATTRIBUTE_INT (td,colspan)
HTML_TAG_ATTRIBUTE_INT (td,rowspan)
END_HTML_TAG(td)

DECLARE_HTML_BINARY_TAG(th)
HTML_TAG_ATTRIBUTE_INT (th,colspan)
HTML_TAG_ATTRIBUTE_INT (th,rowspan)
END_HTML_TAG(th)

DECLARE_HTML_SIMPLE_UNARY_TAG (br)

DECLARE_HTML_SIMPLE_BINARY_TAG(p)

DECLARE_HTML_SIMPLE_BINARY_TAG(h1)
DECLARE_HTML_SIMPLE_BINARY_TAG(h2)
DECLARE_HTML_SIMPLE_BINARY_TAG(h3)
DECLARE_HTML_SIMPLE_BINARY_TAG(h4)
DECLARE_HTML_SIMPLE_BINARY_TAG(h5)
DECLARE_HTML_SIMPLE_BINARY_TAG(h6)

DECLARE_HTML_SIMPLE_BINARY_TAG(ul)
DECLARE_HTML_SIMPLE_BINARY_TAG(ol)
DECLARE_HTML_SIMPLE_BINARY_TAG(li)

DECLARE_HTML_SIMPLE_BINARY_TAG(b)
// DECLARE_HTML_SIMPLE_BINARY_TAG(i) // caused shadow warnings of local variables vs global (if you need this, fix namespace for this library)
DECLARE_HTML_SIMPLE_BINARY_TAG(u)
DECLARE_HTML_SIMPLE_BINARY_TAG(q)
DECLARE_HTML_SIMPLE_BINARY_TAG(tt)
DECLARE_HTML_SIMPLE_BINARY_TAG(kbd)
DECLARE_HTML_SIMPLE_BINARY_TAG(code)
DECLARE_HTML_SIMPLE_BINARY_TAG(cite)
DECLARE_HTML_SIMPLE_BINARY_TAG(pre)

DECLARE_HTML_SIMPLE_BINARY_TAG(applet)

HtmlObject* text(const string& strText);
HtmlObject* text(const char* pText);

#endif // !defined(__HTML_TAGS_H__INCLUDED_)

