/*  createhtml - library to generate html files
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// UrlFolder.cpp: implementation of the UrlFolder class.
//
//////////////////////////////////////////////////////////////////////

#include "UrlFolder.h" // dep_chk_needed 
#include "UrlIndexFile.h" // dep_chk_needed 

#include <algorithm>
#include <assert.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

UrlFolder::UrlFolder(const string& strName)
	: UrlItem(strName)
{

}

template<class __II> inline
void for_each_delete_ptr(__II __F, __II __L)
{for (;__F != __L; ++__F)
  delete *__F; }

UrlFolder::~UrlFolder()
{
	for_each_delete_ptr (m_vectFolders.begin(),m_vectFolders.end());
	for_each_delete_ptr (m_vectFiles.begin(),m_vectFiles.end());
}

void UrlFolder::Add (UrlFile* pFile)
{
	assert (pFile);
	assert (!pFile->m_pParent);	
	pFile->m_pParent = this;
	
	m_vectFiles.push_back(pFile);
}

UrlFile* UrlFolder::AddFile (const string& strName)
{
	UrlFile* pFile = new UrlFile(strName);

	Add(pFile);

	return pFile;
}

UrlFile* UrlFolder::AddFile (const string& strName,const string& strDescription)
{
	UrlFile* pFile = new UrlFile(strName);

	pFile->SetDescription (strDescription);
	Add(pFile);

	return pFile;
}

UrlFile* UrlFolder::AddIndexFile (const string& strDescription)
{
	UrlFile* pFile = new UrlIndexFile("index.html");

	pFile->SetDescription (strDescription);
	Add(pFile);

	return pFile;
}

void UrlFolder::Add (UrlFolder* pFolder)
{
	assert (pFolder);
	assert (!pFolder->m_pParent);	
	pFolder->m_pParent = this;
	
	m_vectFolders.push_back(pFolder);
}

UrlFolder* UrlFolder::AddFolder (const string& strName)
{
	UrlFolder* pFolder = new UrlFolder(strName);

	Add(pFolder);

	return pFolder;
}

UrlFile* UrlFolder::GetIndexFile ()
{
	for (unsigned int i = 0; i < m_vectFiles.size(); i++)
	{
		UrlFile* pFile = dynamic_cast<UrlIndexFile*>(m_vectFiles[i]);

		if (pFile)
			return pFile;
	}	

	// TODO: make sure it's an index file
	
	return NULL;
}







