/*  createhtml - library to generate html files
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// UrlItem.cpp: implementation of the UrlItem class.
//
//////////////////////////////////////////////////////////////////////

#include <sys/stat.h>
#ifdef WIN32
#include <direct.h>
#endif

#include "UrlItem.h" // dep_chk_needed 
#include <stack>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

UrlItem::UrlItem(const string& strName)
{
	m_pParent = NULL;
	m_strName = strName;
}

UrlItem::~UrlItem()
{

}

string UrlItem::GetPath () const
{
  string strName;
  
  if (m_pParent)
    strName = m_pParent->GetPath();	
  else
    strName = '.';
  
  return strName+PATH_DELIMITER+m_strName; // +'/'+
}

void UrlItem::CreateDir () const
{
	// make dure directory exists

	if (m_pParent)
		m_pParent->CreateDir();	

	string strDir = GetPath ();

#ifdef WIN32
	_mkdir (strDir.c_str());
#else
	mkdir (strDir.c_str(),S_IXUSR | S_IWUSR | S_IRUSR);
#endif
}

string UrlItem::GetRelativePathToItem (const UrlItem* pTo) const
{
	string strRelative;

	// so the html page being generated is reflected by pPageFile
	// and we are UrlItem* m_value
	// now, we'll try to find a common parent

	stack<const UrlItem*> stackPageParents;
	stack<const UrlItem*> stackTargetParents;

	const UrlItem* pParent = this;
	
	while (pParent)
	{
		stackPageParents.push(pParent);
		pParent = pParent->m_pParent;
	}
	
	pParent = pTo->m_pParent;
	
	while (pParent)
	{
		stackTargetParents.push(pParent);
		pParent = pParent->m_pParent;
	}

	while (!stackPageParents.empty() &&
		!stackTargetParents.empty() &&
		stackPageParents.top () == stackTargetParents.top ())
	{
		stackPageParents.pop ();
		stackTargetParents.pop ();
	}

	for (int i = stackPageParents.size(); i; i--)
		strRelative += "../";  

	while (!stackTargetParents.empty())
	{
		strRelative += stackTargetParents.top()->GetName ();
		strRelative += '/';
		stackTargetParents.pop ();
	}

	strRelative += pTo->GetName ();

	return strRelative;
}







