/*  disasmviewer, part of exmhcg - executable program tuner
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



//import FormattingException;

public class CisBetter
{
	// first some wrappers for common calls
	
	static String sprintf(String fmt,Boolean arg1)   { return sprintf(fmt,new Object[]{arg1}); }
	static String sprintf(String fmt,Character arg1) { return sprintf(fmt,new Object[]{arg1}); }
	static String sprintf(String fmt,Byte arg1)      { return sprintf(fmt,new Object[]{arg1}); }
	static String sprintf(String fmt,Short arg1)     { return sprintf(fmt,new Object[]{arg1}); }	
	static String sprintf(String fmt,Integer arg1)   { return sprintf(fmt,new Object[]{arg1}); }	
	static String sprintf(String fmt,Long arg1)      { return sprintf(fmt,new Object[]{arg1}); }
	static String sprintf(String fmt,Float arg1)     { return sprintf(fmt,new Object[]{arg1}); }
	static String sprintf(String fmt,Double arg1)    { return sprintf(fmt,new Object[]{arg1}); }
	static String sprintf(String fmt,String arg1)    { return sprintf(fmt,new Object[]{arg1}); }
	
	static String sprintf(String fmt,boolean   arg1) 
	{ return sprintf(fmt,        new Boolean  (arg1)); }	
	static String sprintf(String fmt,char      arg1) 
	{ return sprintf(fmt,        new Character(arg1)); }
	static String sprintf(String fmt,byte      arg1) 
	{ return sprintf(fmt,        new Byte     (arg1)); }	
	static String sprintf(String fmt,short     arg1) 
	{ return sprintf(fmt,        new Short    (arg1)); }
	static String sprintf(String fmt,int       arg1) 
	{ return sprintf(fmt,        new Integer  (arg1)); }	
	static String sprintf(String fmt,long      arg1) 
	{ return sprintf(fmt,        new Long     (arg1)); }
	static String sprintf(String fmt,float     arg1) 
	{ return sprintf(fmt,        new Float    (arg1)); }	
	static String sprintf(String fmt,double    arg1) 
	{ return sprintf(fmt,        new Double   (arg1)); }

	// then tho operations themselves
		
	
	
	
	static String spaces(int spaces)
		{
			String out = "";
			while (spaces-- > 0)
				out += ' ';
			return out;			
		}
	
	static String spaces(int spaces,char space)
		{
			String out = "";
			while (spaces-- > 0)
				out += space;
			return out;			
		}
	
	static String formatField (int fieldLength,boolean leftAlign,String str)
		{
			String out = "";

			int padding = fieldLength-str.length();

			if (!leftAlign)
				out += spaces(padding);
				
			out += str;

			if (leftAlign)
				out += spaces(padding);

			return out;
		}
	
	static String formatChar (int fieldLength,boolean leftAlign,Object obj)
		{
			if (obj instanceof Character)
				return formatField (fieldLength,leftAlign,obj.toString());
			else
				throw new FormattingException("Cannot format %c from type "+obj.getClass().getName());
		}
	
	static String formatString (int fieldLength,boolean leftAlign,Object obj)
		{
			if (obj instanceof String)
				return formatField (fieldLength,leftAlign,(String) obj);
			else
				throw new FormattingException("Cannot format %c from type "+obj.getClass().getName());
		}
	
	static String formatReal (int fieldLength,int realPrecision,boolean leftAlign,boolean plusSign,boolean zeroExtend,boolean capitalLetters,Object obj)
		{
			String strRawValue;
			
			if (obj instanceof Double ||
					obj instanceof Float)
				strRawValue = obj.toString();
			else
				throw new FormattingException("Cannot format %feEgG from type "+obj.getClass().getName());
/*
			String out = "";
			
			while (fieldLength-- > 1)
				out += ' ';
				
			out += obj;	
			
			return out;
*/

			// deal with: realPrecision
			
			if (strRawValue.equals("NaN") ||
					strRawValue.equals("Infinity") ||
					strRawValue.equals("-Infinity"))
				return formatField(fieldLength,leftAlign,strRawValue);
				
			// so it is a normal number, chop it up so we can reformat it
			
			throw new FormattingException("Formatting of real values not implemented! DO SO!");

			//return formatZeroExtendSign(fieldLength,leftAlign,zeroExtend,plusSign,strRawValue);
		}
	
	static String formatZeroExtend (int fieldLength,boolean leftAlign,boolean zeroExtend,String value)
		{
			String out = "";
			
			int padding = fieldLength-value.length();

			if (!leftAlign)
				out += spaces(padding,(zeroExtend ? '0' : ' '));
				
			out += value;

			if (leftAlign)
				out += spaces(padding);

			return out;
		}
	
	static String formatZeroExtendSign(int fieldLength,boolean leftAlign,boolean plusSign,boolean zeroExtend,String value)
		{
			String out = "";
			
			String sign = null;
			
			if (value.startsWith("-"))
				{
					sign = "-";
					value = value.substring(1);				
				}
			else if (plusSign)
				sign = "+";			

			int padding = fieldLength-value.length()-(sign != null ? sign.length() : 0);

			if (!leftAlign)
				{
					if (zeroExtend)
						out += sign + spaces(padding,'0');
					else
						out += spaces(padding) + sign;
					out += value;
				}
			else
				out += sign + value;

			if (leftAlign)
				out += spaces(padding);

			return out;
		}
	
	static long UnsignedLongValue(Number obj)
		{
			if (obj instanceof Integer)
				{
					long unsignedValue = ((Integer) obj).longValue();	
					if (unsignedValue < 0)
						unsignedValue += Integer.MAX_VALUE - Integer.MIN_VALUE + 1;
					return unsignedValue;
				}
			if (obj instanceof Short)
				{
					long unsignedValue = ((Short) obj).longValue();	
					if (unsignedValue < 0)
						unsignedValue += Short.MAX_VALUE - Short.MIN_VALUE + 1;
					return unsignedValue;
				}
			if (obj instanceof Byte)
				{
					long unsignedValue = ((Byte) obj).longValue();	
					if (unsignedValue < 0)
						unsignedValue += Byte.MAX_VALUE - Byte.MIN_VALUE + 1;
					return unsignedValue;
				}
			return 0;
		}

	static String formatHex (int fieldLength,boolean leftAlign,boolean zeroExtend,boolean capitalLetters,Object obj)
		{
			String strRawValue;

			if (obj instanceof Long)
				strRawValue = Long.toHexString(((Long) (obj)).intValue());
			else if (obj instanceof Integer ||
							 obj instanceof Short ||
							 obj instanceof Byte)
				strRawValue = Long.toHexString(UnsignedLongValue ((Number) obj));
			else
				throw new FormattingException("Cannot format %xX from type "+obj.getClass().getName());
				
			if (capitalLetters)
				strRawValue.toUpperCase();

			return formatZeroExtend(fieldLength,leftAlign,zeroExtend,strRawValue);			
		}
		
	static String formatOct (int fieldLength,boolean leftAlign,boolean zeroExtend,Object obj)
		{
			String strRawValue;

			if (obj instanceof Long)
				strRawValue = Long.toOctalString(((Long) (obj)).intValue());
			else if (obj instanceof Integer ||
							 obj instanceof Short ||
							 obj instanceof Byte)
				strRawValue = Long.toOctalString(UnsignedLongValue ((Number) obj));
			else
				throw new FormattingException("Cannot format %o from type "+obj.getClass().getName());

			return formatZeroExtend(fieldLength,leftAlign,zeroExtend,strRawValue);			
		}
		
	static String formatUnsigned (int fieldLength,boolean leftAlign,boolean zeroExtend,Object obj)
		{
			long unsignedValue;
			
			if (obj instanceof Integer ||
					obj instanceof Short ||
					obj instanceof Byte)
				unsignedValue = UnsignedLongValue ((Number) obj);
			// we cannot handle longs, as we would need one extra bit
			else
				throw new FormattingException("Cannot format %xX from type "+obj.getClass().getName());

			String strRawValue = Long.toString(unsignedValue);

			return formatZeroExtend(fieldLength,leftAlign,zeroExtend,strRawValue);			
		}
		
	static String formatInt (int fieldLength,boolean leftAlign,boolean plusSign,boolean zeroExtend,Object obj)
		{
			String strRawValue;

			if (obj instanceof Long ||
					obj instanceof Integer ||
					obj instanceof Short ||
					obj instanceof Byte)
				strRawValue = obj.toString();
			else
				throw new FormattingException("Cannot format %xX from type "+obj.getClass().getName());
				
			return formatZeroExtendSign(fieldLength,leftAlign,plusSign,zeroExtend,strRawValue);			
		}
		
	static String sprintf(String fmt,Object objs[])
		{
			String out = new String();
			
			// find us an '%'
			
			int obj_i = 0;
			
			int i;

			while ((i = fmt.indexOf('%')) != -1)
				{
					// get what is before the '%'
					
					out += fmt.substring(0,i);
					
					// the format string ends at any of %dxXuofeEgGcs
					
					int j = i + 1;
					
					char nextChar = fmt.charAt(j++);
					
					if (nextChar == '%')
						{
							out += '%';	
							fmt = fmt.substring(j+1);
							continue;
						}

					boolean leftAlign = false;
					boolean plusSign = false;
					boolean zeroExtend = false;
					String fieldLength = null;
					String realPrecision = null;

					if (nextChar == '-') { leftAlign = true; nextChar = fmt.charAt(j++); }
					if (nextChar == '+') { plusSign = true; nextChar = fmt.charAt(j++); }
					if (nextChar == '0') { zeroExtend = true; nextChar = fmt.charAt(j++); }
					if (nextChar >= '1' && nextChar <= '9')
						{
							fieldLength = new String();
							while (nextChar >= '0' && nextChar <= '9')
								{
									fieldLength += nextChar;
									nextChar = fmt.charAt(j++);
								}					
						}
					if (nextChar == '.')
						{
							realPrecision = new String();
							nextChar = fmt.charAt(j++);
							while (nextChar >= '0' && nextChar <= '9')
								{
									realPrecision += nextChar;
									nextChar = fmt.charAt(j++);
								}					
						}
					if (nextChar == 'l' || nextChar == 'L') { /* ignore */ nextChar = fmt.charAt(j++); }
					
					boolean intFormat = false;
					boolean realFormat = false;
					boolean stringFormat = false;
					boolean charFormat = false;
				
					boolean capitalLetters = false;

					boolean hexFormat = false;
					boolean octFormat = false;

					boolean unsignedFormat = false;

					if      (nextChar == 'd') { intFormat = true; }
					else if (nextChar == 'x') { intFormat = true; hexFormat = true; }
					else if (nextChar == 'X') { intFormat = true; hexFormat = true; capitalLetters = true; }
					else if (nextChar == 'u') { intFormat = true; unsignedFormat = true; }
					else if (nextChar == 'o') { intFormat = true; octFormat = true; }
					else if (nextChar == 'f') { realFormat = true; }
					else if (nextChar == 'e') { realFormat = true; }
					else if (nextChar == 'E') { realFormat = true; capitalLetters = true; }
					else if (nextChar == 'g') { realFormat = true; }
					else if (nextChar == 'G') { realFormat = true; capitalLetters = true; }
					else if (nextChar == 'c') { charFormat = true; }
					else if (nextChar == 's') { stringFormat = true; }
					else
						{
							// unrecognised format
							
							throw new FormattingException("Unrecognised format: '"+fmt.substring(i,j-i)+"'");							
						}
					
					if (realPrecision != null && !realFormat)
						throw new FormattingException("Cannot have precision (.) with non-real format specifier: '"+fmt.substring(i,j-i)+"'");
					if (plusSign && (stringFormat || charFormat || hexFormat || unsignedFormat || octFormat))
						throw new FormattingException("Cannot have plus sign string format specifier: '"+fmt.substring(i,j-i)+"'");
					if (zeroExtend && (!intFormat && !realFormat))
						throw new FormattingException("Cannot have zero extend (0) with not-integer format specifier: '"+fmt.substring(i,j-i)+"'");
					if (zeroExtend && leftAlign)
						throw new FormattingException("Cannot left align (-) and zero extend (0): '"+fmt.substring(i,j-i)+"'");
						
					// now format the mess
					
					Object obj = objs[obj_i++];
					
					int nFieldLength = (fieldLength == null ? 0 : Integer.parseInt(fieldLength));
					
					if (charFormat)
						{
							out += formatChar (nFieldLength,leftAlign,obj);									
						}
					if (stringFormat)
						{
							out += formatString (nFieldLength,leftAlign,obj);									
						}
					if (realFormat)
						{							
							int nRealPrecision = (realPrecision == null ? 0 : Integer.parseInt(realPrecision));
							out += formatReal (nFieldLength,nRealPrecision,leftAlign,plusSign,zeroExtend,capitalLetters,obj);									
						}
					if (intFormat)
						{							
							if (hexFormat)
								out += formatHex (nFieldLength,leftAlign,zeroExtend,capitalLetters,obj);
							else if (octFormat)
								out += formatOct (nFieldLength,leftAlign,zeroExtend,obj);
							else if (hexFormat)
								out += formatUnsigned (nFieldLength,leftAlign,zeroExtend,obj);
							else
								out += formatInt (nFieldLength,leftAlign,plusSign,zeroExtend,obj);
						}
					
					fmt = fmt.substring(j);
				}		
				
			// whatever is left
			
			return out + fmt;	
		}
}
