/*  disasmviewer, part of exmhcg - executable program tuner
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

import java.awt.*;
import java.awt.font.*;
import java.awt.event.*;
import java.util.*;

//import CisBetter;

public class DataPane extends Canvas implements ItemListener, ActionListener, KeyListener
{
	Dimension dimNeedSize = null;
	
	DisAsmData data;
	
	static final int SORT_OFFSET = 0;
	static final int SORT_FLOW   = 1;
	static final int NUM_SORT   = 2;
	
	int nSortOrder = SORT_OFFSET;

	static final int COL_ID        = 0; // label/offset
	static final int COL_CODE      = 1; 
	static final int COL_JUMP_UP   = 2; 
	static final int COL_INSTR     = 3; 
	static final int COL_JUMP_DOWN = 4; 
	static final int NUM_COLS     = 5;
	
	boolean	bHaveCol[] = new boolean[NUM_COLS];
	boolean	bUseCol[]  = new boolean[NUM_COLS];
	Color colColors[]  = new Color[NUM_COLS];
	
	PopupMenuHelper popupMenu;
	MenuItem menuItemSort[] = new MenuItem[NUM_SORT];
	CheckboxMenuItem menuItemCols[] = new CheckboxMenuItem[NUM_COLS];	
	
	public void init(DisAsmData d)
		{
			addKeyListener (this);				
			
			Font font = getFont();
			setFont (new Font("Monospaced",font.getStyle(),font.getSize()));

			data = d;
			
			popupMenu = new PopupMenuHelper(this);
			
			menuItemSort[SORT_OFFSET] = popupMenu.addItem (this,"Sort by offset");
			menuItemSort[SORT_FLOW  ] = popupMenu.addItem (this,"Sort by flow");
			popupMenu.addSeparator ();
			menuItemCols[COL_ID     ] = popupMenu.addCheckbox (this,"Id/Label");
			menuItemCols[COL_CODE   ] = popupMenu.addCheckbox (this,"Raw code");
			menuItemCols[COL_JUMP_UP] = popupMenu.addCheckbox (this,"Jumps");
			
			for (int i = 0; i < NUM_COLS; i++)
				{
					bUseCol[i] = true;
					if (bUseCol[i] && menuItemCols[i] != null)
						menuItemCols[i].setState (true);
				}
			
			colColors[COL_CODE]      = new Color(.75f,.75f,.75f);
			colColors[COL_JUMP_UP]   = new Color(.90f,.95f,.90f);
			colColors[COL_JUMP_DOWN] = new Color(.90f,.95f,.90f);
		
			layoutData ();
			
			System.out.println ("init finsihed...");
			
			
			
			/*
	
			System.out.println ("DataPane::Graphics: "+getGraphics());

			setLayout (new GridBagLayout());
			
			Font font = getFont();
			
			System.out.println ("Font: "+font);
			
			
			Font f;
			setFont (f = new Font("Monospaced",Font.PLAIN,12));
			
			FontMetrics fm = getGraphics().getFontMetrics (f);
			
			System.out.println ("Dim: "+fm.getStringBounds("Hello",getGraphics()));
			
			
			
			for (Iterator i = data.vectSnippets.iterator(); i.hasNext(); )
				{
					CodeSnippet snippet = (CodeSnippet) i.next();
					
					System.out.println ("ID: "+snippet.strId);
					
					if (snippet instanceof AsmCodeSnippet)
						add((AsmCodeSnippet) snippet);
					
				}
			
			GridBagConstraints c = new GridBagConstraints();
			
			c.gridx = 3;
			c.gridy = 0;
			c.gridheight = GridBagConstraints.REMAINDER;
			c.fill = GridBagConstraints.VERTICAL;
			c.anchor = GridBagConstraints.WEST;
			
			add(new JumpsCanvas(),c);			
			
			{	
			c = new GridBagConstraints();
			
			c.gridx = 10;
			c.gridy = GridBagConstraints.RELATIVE;
			c.anchor = GridBagConstraints.WEST;
			c.weightx = 1;
			c.weighty = 1;
			
			add(new Label("Lower-Right"),c);
			}
			
/*			
			Canvas c = new Canvas();
			
			add(c);
			
			Graphics gc = c.getGraphics();
			
			System.out.println ("Graphics:"+gc);
**
			*/
		}
	
	public void actionPerformed(ActionEvent e)
		{
			for (int i = 0; i < NUM_SORT; i++)
				if (e.getSource() == menuItemSort[i])
					if (i != nSortOrder)
						{
							nSortOrder = i;
							layoutData();							
						}
		}
	
	public void itemStateChanged(ItemEvent e)
		{
			for (int i = 0; i < NUM_COLS; i++)
				if (e.getSource() == menuItemCols[i])
				{
					bUseCol[i] = (e.getStateChange() == ItemEvent.SELECTED);
					fillColumns ();
				}
			
		}
	
	public Dimension getPreferredSize ()
		{
			if (dimNeedSize != null)
				return dimNeedSize;
			else
				return super.getPreferredSize();
		}
		
	static final int JUMP_UP = 0;
	static final int JUMP_DOWN = 1;
	
	Vector vectSnippets = null;
	Hashtable hashIdIndexes = null;
	
	Vector vectJumpArrows[] = new Vector[2];
	Hashtable hashSnippetAnchor = null;
		
	public void sortSnippets()
		{
			// forget old sorting, and reload according to new
			// sorting
			
			vectSnippets = new Vector();
			
			for (Enumeration e = data.vectSnippets.elements(); e.hasMoreElements(); )
				{
					CodeSnippet snippet = (CodeSnippet) e.nextElement();
					
					vectSnippets.add (snippet);
				}
			
			hashIdIndexes = new Hashtable();
			
			for (int i = 0; i < vectSnippets.size(); i++)
				{
					CodeSnippet snippet = (CodeSnippet) vectSnippets.elementAt(i);					
					hashIdIndexes.put (snippet,new Integer(i));
				}
		}	
		
	public void addJumpArrows(int fi,CodeSnippet from,AsmBranchInstr instr)
		{
			Vector v = ((AsmBranchInstr) instr).vectInternalTargets;
			
			if (v == null)				
				return;
				
			JumpArrow arrow = null;
			
			boolean bArrowDown = false;
			
			for (Enumeration i = v.elements(); i.hasMoreElements(); )
				{
					CodeSnippet to = (CodeSnippet) i.nextElement();
					
					int ti = ((Integer) hashIdIndexes.get(to)).intValue();
					
					if (arrow == null)
						arrow = new JumpArrow(from,to,fi,ti);
					else
						arrow.addTo (to,ti);
					
					if (ti > fi)
						bArrowDown = true;
				}
			if (bArrowDown)
				vectJumpArrows[JUMP_DOWN].add (arrow);
			else
				vectJumpArrows[JUMP_UP].add (arrow);				
		}
	
	public void collapseSameTargetArrows (Vector v)
		{
			// can only collapse if targets are unique
			
			Hashtable singleTarget = new Hashtable();
			
			for (Iterator i = v.iterator(); i.hasNext(); )
				{
					JumpArrow a = (JumpArrow) i.next();
					
					if (a.vectTo.size() == 1)
						{
							JumpArrow a_same = (JumpArrow) singleTarget.get (a.getTo ().snippet);
							
							if (a_same != null)
								{
									// collapse these two into one, a_same is already in hash
									// so collapse into a_same
									
									a_same.addFrom (a.getFrom ());					
									
									i.remove(); // remove a from v
								}
							else
								singleTarget.put (a.getTo ().snippet,a);							
						}					
				}			
		}
	
	public void assignArrowLevels(Vector v)
		{
			class CompareArrows implements Comparator
				{
					public int compare(Object o1,Object o2)
						{
							JumpArrow arrow1 = (JumpArrow) o1;
							JumpArrow arrow2 = (JumpArrow) o2;
							
							return arrow1.minIndex - arrow2.minIndex; 
						}                   				
				}
			
			Collections.sort(v,new CompareArrows());
			
			// find what arrows begin inside other ones
			
			for (int i = 0; i < v.size()-1; i++)
				{
					JumpArrow ai = (JumpArrow) v.elementAt(i);
					
					for (int j = i+1; j < v.size(); j++)
						{
							JumpArrow aj = (JumpArrow) v.elementAt(j);
							
							if (aj.minIndex > ai.maxIndex)
								break;
							
							if (aj.maxIndex < ai.maxIndex)
								ai.addInsideFully (aj);							
							else
								ai.addInsideBegin (aj);							
						}
				}
			
			// inside fully no contain all fully enclosed jump arrows (as before)
			// and inside begin contain all partially enclosed that begin lower thet we do
			
			int nMaxLevel = 0;

			for (int i = 0; i < v.size(); i++)
				nMaxLevel = Math.max(nMaxLevel,((JumpArrow) v.elementAt(i)).assignLevel ());
				
			// now compact if possible
			
			for (int i = 0; i < v.size(); i++)
				((JumpArrow) v.elementAt(i)).clearInside ();
				
			for (int i = 0; i < v.size()-1; i++)
				{
					JumpArrow ai = (JumpArrow) v.elementAt(i);
					
					for (int j = i+1; j < v.size(); j++)
						{
							JumpArrow aj = (JumpArrow) v.elementAt(j);
							
							if (ai.maxIndex < aj.minIndex ||
									ai.minIndex > aj.maxIndex)
								continue;
							
							ai.addInsideBegin (aj);							
							aj.addInsideBegin (ai);							
						}
				}

			// inside fully no contain all fully enclosed jump arrows (as before)
			// and inside begin contain all partially enclosed

			boolean used[] = new boolean[nMaxLevel+1];

			for ( ; ; )
				{
					boolean bAgain = false;
					
					for (int i = 0; i < v.size(); i++)
						bAgain |= ((JumpArrow) v.elementAt(i)).compactLevel (used);
						
					if (!bAgain)
						break;
				}

			for (int i = 0; i < v.size(); i++)
				((JumpArrow) v.elementAt(i)).clearInside ();
		}
	
	public void assignSnippetAnchors(int index,Vector v)
		{
			for (int i = 0; i < v.size(); i++)
				{
					JumpArrow arrow = (JumpArrow) v.elementAt(i);
					
					arrow.assignSnippetAnchors (index,hashSnippetAnchor);					
				}	
		}
		
	public void setupJumpArrows()
		{
			// find all internal jumps
			
			for (int i = 0; i < 2; i++)
				vectJumpArrows[i] = new Vector();
			
			CodeSnippet from;
			AsmInstr instr;
			
			for (int fi = 0; fi < vectSnippets.size(); fi++)
				if ((from = (CodeSnippet) vectSnippets.elementAt(fi)) instanceof AsmCodeSnippet)
					for (Iterator i = ((AsmCodeSnippet) from).vectInstrs.iterator(); i.hasNext(); )
						if ((instr = (AsmInstr) i.next()) instanceof AsmBranchInstr)
							addJumpArrows (fi,from,(AsmBranchInstr) instr);
					
//					hashIdIndexes.put (snippet,new Integer(i));

			hashSnippetAnchor = new Hashtable();

			for (int i = 0; i < 2; i++)
				{
					if (vectJumpArrows[i].isEmpty())
						{
							vectJumpArrows[i] = null;
							continue;
						}
					collapseSameTargetArrows (vectJumpArrows[i]);
					assignArrowLevels (vectJumpArrows[i]);
					
					assignSnippetAnchors (i,vectJumpArrows[i]);
				}

			// sort 
		}
	
	public void haveColumns()
		{
			for (int i = 0; i < NUM_COLS; i++)
				bHaveCol[i] = false;
				
			bHaveCol[COL_JUMP_UP  ] = (vectJumpArrows[JUMP_UP  ] != null);
			bHaveCol[COL_JUMP_DOWN] = (vectJumpArrows[JUMP_DOWN] != null);
				
			// walk through all code snippets and see what we have
			
			CodeSnippet from;
			AsmInstr instr;
			
			for (int fi = 0; fi < vectSnippets.size(); fi++)
				{
					from = (CodeSnippet) vectSnippets.elementAt(fi);
					
					if (from.getPrefName() != null)
						bHaveCol[COL_ID] = true;						

					if (from instanceof AsmCodeSnippet)
						{
							for (Iterator i = ((AsmCodeSnippet) from).vectInstrs.iterator(); i.hasNext(); )
								{
									instr = (AsmInstr) i.next();
									
									bHaveCol[COL_INSTR] = true;						

									if (instr.getCode() != null)
										bHaveCol[COL_CODE] = true;						
								}
						}
				}	
		}
	
	int nWidthCol[] = new int[NUM_COLS];
	int nStartXCol[] = new int[NUM_COLS];
	
	int nBaselineY;
	int nColSeparation;
	
	Vector vectDataRows;
	
	class DataFiller
		{
			FontMetrics metrics;
			DataRow row;
			
			int textHeight;
			
			boolean bRowHasOutgoing = false;
			
			DataFiller()
				{
					metrics = getGraphics().getFontMetrics();					
					textHeight = metrics.getHeight();
					
					nBaselineY = metrics.getAscent()+metrics.getLeading()/2;
					nColSeparation = metrics.charWidth (' ');
				}		
				
			public int calcRowYs()
				{
					int y = 0;
					
					Iterator i = vectDataRows.iterator();
					
					while (i.hasNext())
						{
							DataRow row = (DataRow) i.next();
							
							y += row.leading;							
							row.y = y;
							y += row.height + row.ending;							
						}					
					
					return y;
				}	
			
			public DataRow addRow()
				{		
					if (row != null &&
							row.height == 0 && row.leading == 0 && row.ending == 0 &&
							row.vectDataItem.isEmpty())
						return row; // do not add a new one if current is unused
					
					row = new DataRow();
					
					vectDataRows.add (row);	
					
					return row;
				}
			
			public void addRowLeading(float fraction)
				{	
					row.addLeading ((int) (textHeight*fraction));
				}	

			public void minRowLeading(float fraction)
				{	
					row.minLeading ((int) (textHeight*fraction));
				}	

			public void addRowHeight(float fraction)
				{	
					row.addHeight ((int) (textHeight*fraction));
				}	

			public void minRowHeight(float fraction)
				{	
					row.minHeight ((int) (textHeight*fraction));
				}	

/*
			public void addRowEnding(float fraction)
				{	
					row.addEnding ((int) (textHeight*fraction));
				}	
*/
			public DataItemLine addLine (int c)
				{
					DataItemLine l = new DataItemLine (c);
					row.add (l);
					return l;
				}
	
			public DataText addText(int c,String text)
				{
					DataText t = new DataText(c,text);
					
					row.add (t);
					row.minHeight (textHeight);
					nWidthCol[c] = Math.max(nWidthCol[c],t.calcWidth (metrics));
					
					return t;
				}
			
			public DataTextLink addTextLink(int c,String pre,String text,String post,int rowoffset)
				{
					//System.out.println (""+pre+text+post);
					
					int yoffset = rowoffset*textHeight;
					
					DataTextLink t = new DataTextLink(c,pre,text,post,0,yoffset);
					
					row.add (t);
					
					if (yoffset < 0)
						row.minLeading (-yoffset);
					if (yoffset+textHeight > 0)
						row.minHeight (yoffset+textHeight);
						
					nWidthCol[c] = Math.max(nWidthCol[c],t.calcWidth (metrics));
					
					return t;
				}
			
			public DataTextMultiLine addTextMultiLine(int c,String text)
				{
					DataTextMultiLine tml = new DataTextMultiLine(c,text);						
	
					row.add (tml);
					row.minHeight (textHeight*tml.breaks.length);
					
					nWidthCol[c] = Math.max(nWidthCol[c],tml.calcWidth (metrics));

					return tml;
				}
		}

	public int calcArrowsWidth (Vector v)
		{
			int nMaxWidth = 0;
			
			for (int i = 0; i < v.size(); i++)
				nMaxWidth = Math.max(nMaxWidth,((JumpArrow) v.elementAt(i)).getWidth ());
				
			return nMaxWidth;
		}
		
	public void fillColumns ()
		{
			DataFiller filler = new DataFiller();
			
			for (int i = 0; i < NUM_COLS; i++)
				nWidthCol[i] = 0;
				
			vectDataRows = new Vector();

			Vector vectDataTextRightAlign = new Vector();

/*				
			// adjust widths for text width
			
			FontMetrics metrics = getGraphics().getFontMetrics();
						
			for (int i = 0; i < NUM_COLS; i++)
				nWidthCol[i] = 0;
*/				
			CodeSnippet prev = null;
			CodeSnippet snippet;
			AsmInstr instr;
			
			for (int fi = 0; fi < vectSnippets.size(); fi++)
				{
//					System.out.println ("layouting snippet: "+fi);
					
					snippet = (CodeSnippet) vectSnippets.elementAt(fi);
					
					if (snippet.prev == null && prev != null)
						{
							filler.addRow ();
							filler.addRowHeight (1);
						}
					
					filler.addRow ();
					
					filler.addLine (COL_INSTR);
					
					///////////////////////////////////////////////////////////
					// fix incoming jump arrows
					
					if (bUseCol[COL_JUMP_UP])
						{
							JumpAnchor anchor = (JumpAnchor) hashSnippetAnchor.get(snippet);
							
							if (anchor != null) // there is something (but may be only outgoing)
								{
									anchor.setRow (filler.row,filler.textHeight);
/*									
									int nNumIncoming = anchor.getNum(JumpAnchor.INCOMING);
									int nNumOutgoing = anchor.getNum(JumpAnchor.OUTGOING);
								
									if (nNumIncoming + nNumOutgoing > 1)
										filler.addRowLeading(nNumIncoming+(nNumOutgoing>0?1:0)-1);
									if (nNumOutgoing > 1)
										filler.addRowEnding(nNumOutgoing-1);
									anchor.setRow (filler.row,filler.textHeight); //,nNumOutgoing>0);
*/
								}
						}
					
					///////////////////////////////////////////////////////////
					// add id/label

					if (bUseCol[COL_ID] && snippet.isTarget())
						{
							String strName = snippet.getPrefName();
					
							if (strName != null)
								filler.addText (COL_ID,strName);
							}

					///////////////////////////////////////////////////////////
					// show external sources
					
					if (snippet.vectExternalCalls != null)
						{
							Vector v = snippet.vectExternalCalls;
							
							int num = v.size ();
							
							for (int i = 0; i < num; i++)
								{
									External from = (External) v.elementAt(i);
									//System.out.println (""+from+","+from.getPrefName());
									filler.addTextLink(COL_INSTR,"[",from.getPrefName(),"]",-num+i);
								} 
						}
					
					
					///////////////////////////////////////////////////////////
					// add instructions

					if (snippet instanceof AsmCodeSnippet)
						{
							boolean bFirst = true;
							
							for (Iterator i = ((AsmCodeSnippet) snippet).vectInstrs.iterator(); i.hasNext(); )
								{
									instr = (AsmInstr) i.next();
									
									filler.addText (COL_INSTR,instr.getAsText());
									
									if (instr instanceof AsmBranchInstr)
										{
											Vector v = ((AsmBranchInstr) instr).vectExternalTargets;
											
											if (v != null)
												{
													int num = v.size ();
													
													for (int j = 0; j < num; j++)
														{
															External to = (External) v.elementAt(j);
															
															//System.out.println (""+to+","+to.getPrefName());

															DataText t = filler.addTextLink(COL_INSTR,"[",to.getPrefName(),"]",j+1);
															
															vectDataTextRightAlign.add(t);
														} 
												}
										}
									
									if (bUseCol[COL_CODE])
										{
											String strCode = instr.getCodeAsText();
											
											if (strCode != null)
												{
													if (strCode.indexOf('\n') != -1)
														filler.addTextMultiLine (COL_CODE,strCode);
													else
														filler.addText (COL_CODE,strCode);
												}
										}
									filler.addRow ();
								}
						}

					///////////////////////////////////////////////////////////
					
					prev = snippet;
				}	
				
			System.out.println ("layouting snippets finsihed");

			// calculate width of jump arrows
			
			if (bUseCol[COL_JUMP_UP])
				{
					if (bHaveCol[COL_JUMP_UP])
						nWidthCol[COL_JUMP_UP] = calcArrowsWidth (vectJumpArrows[JUMP_UP]);
					if (bHaveCol[COL_JUMP_DOWN])
						nWidthCol[COL_JUMP_DOWN] = calcArrowsWidth (vectJumpArrows[JUMP_DOWN]);
				}
					
			//

			nStartXCol[0] = nColSeparation/2;			
			for (int i = 1; i < NUM_COLS; i++)
				{
					nStartXCol[i] = nStartXCol[i-1] + nWidthCol[i-1] +
						(nWidthCol[i-1] > 0 ? nColSeparation: 0);
					
					System.out.println ("col "+i+" : "+nStartXCol[i]+"  "+nWidthCol[i]);
				}
			
			filler.addRow (); // last (unused, bottom) row
			
			int nWidth = nStartXCol[NUM_COLS-1] + nWidthCol[NUM_COLS-1]+nColSeparation-nColSeparation/2;
			int nHeight = filler.calcRowYs ();
			dimNeedSize = new Dimension (nWidth,nHeight);
			
			Iterator dti = vectDataTextRightAlign.iterator();
			while (dti.hasNext())
				{
					DataTextOffset dt = (DataTextOffset) dti.next();	
					
					dt.rightAlign (nWidthCol[dt.col],filler.metrics);
				}
			
			System.out.println ("layouting finsihed, size" + dimNeedSize);

			repaint ();
			getParent().doLayout ();
			System.out.println ("layouting called repaint & dolayout");
		}
		
	public void layoutData()
		{
			// calculate size of all data items, columns etc
			
			// first sort data snippets according to current sorting order
			
			sortSnippets();
			
			setupJumpArrows();			
			
			// figure out what columns are needed	
			
			haveColumns ();
			
			
			
			fillColumns ();
			System.out.println ("layoutData finsihed...");
		}

	
/*	
	/*
	public void update(Graphics gc)
		{
			gc.drawString ("Hello",10,10);	
			
			
		}
	*
*/	
/*	
	void add(AsmCodeSnippet snippet)
		{
			Integer nOffset = snippet.nOffset;
			
			for (Iterator i = snippet.vectInstrs.iterator(); i.hasNext(); )
				{
					AsmInstr instr = (AsmInstr) i.next();
					
					// str += instr.DebugContents ();					
					
					if (nOffset != null)
						{					
							GridBagConstraints c = new GridBagConstraints();
							
							c.gridx = 0;
							c.gridy = GridBagConstraints.RELATIVE;
							c.anchor = GridBagConstraints.EAST;
							
							add(new Label(CisBetter.sprintf(nOffset == snippet.nOffset ? "0x%08x" : "%x",nOffset.intValue()),Label.RIGHT),c);
							
							if (instr.bCode != null)
								nOffset = new Integer(nOffset.intValue()+instr.bCode.length);
							else
								nOffset = null;					
						}
	
					add (instr);
				}
			
			
			
			
		}
	
	void add(AsmInstr instr)
		{
			
			if (instr.bCode != null && instr.strInstr.startsWith("J"))
			{
				String str = CisBetter.sprintf("%02x",instr.bCode[0]);
				for (int i = 1; i < instr.bCode.length; i++)
					str += CisBetter.sprintf(" %02x",instr.bCode[i]);							
					
				GridBagConstraints c = new GridBagConstraints();
				
				c.gridx = 2;
				c.gridy = GridBagConstraints.RELATIVE;
				c.anchor = GridBagConstraints.WEST;
				
				add(new Label(str),c);
			}
		
			
			
			GridBagConstraints c = new GridBagConstraints();
			
			c.gridx = 4;
			c.gridy = GridBagConstraints.RELATIVE;
			c.anchor = GridBagConstraints.WEST;
			c.ipady = -2;
			
			Label l;
			
			add(l = new Label(instr.strInstr),c);
			
			l.setBackground (new Color(.5f,.5f,1.f));
			
			System.out.println ("minsize:  "+l.getMinimumSize ());
			System.out.println ("maxsize:  "+l.getMinimumSize ());
			System.out.println ("prefsize: "+l.getPreferredSize ());
			
			System.out.println (instr.strInstr);
			
			
			
		}
	*/
	/*
			str += CisBetter.sprintf("Offset: %0x08x\n",nOffset);
			
			for (Iterator i = vectInstrs.iterator(); i.hasNext(); )
				{
					AsmInstr instr = (AsmInstr) i.next();
					
					str += instr.DebugContents ();					
				}
*/

	public void keyTyped(KeyEvent e) { }
	public void keyReleased(KeyEvent e) { }
	
	public void keyPressed(KeyEvent e)
		{
			if (e.getKeyCode() == KeyEvent.VK_UP)
				{
					int r = getFirstVisibleRow();

					if (!vectDataRows.isEmpty())
					{
						Point ptPos = ((ScrollPane) getParent()).getScrollPosition ();
						DataRow row = (DataRow) vectDataRows.elementAt(r);
						if (row.y-row.leading == ptPos.y)
							r--;
					}
					scrollToRow (r);
				}			
			if (e.getKeyCode() == KeyEvent.VK_DOWN)
					scrollToRow (getFirstVisibleRow()+1);
			if (e.getKeyCode() == KeyEvent.VK_PAGE_DOWN)
				{
					int r = getLastVisibleRow();
					scrollToRow (r == getFirstVisibleRow () ? r+1 : r);
				}			
			if (e.getKeyCode() == KeyEvent.VK_PAGE_UP)
				{
					int r = getFirstVisibleRow();

					DataRow rowlast = (DataRow) vectDataRows.elementAt(r);
					
					// we wan't rowlast to be at least a little bit visible
					
					ScrollPane scroll = (ScrollPane) getParent();
					Dimension dimSize = scroll.getViewportSize ();
					
					// see how mush we can back up
					
					r--; // backup at least one
					
					while (r > 0)
						{
							DataRow row = (DataRow) vectDataRows.elementAt(r-1);
							
							if (row.y-row.leading + dimSize.height < rowlast.y-rowlast.leading)
								break;
							r--; // visible with row (=> r-1) at top, so accept								
						}							
					scrollToRow (r);
				}			
			if (e.getKeyCode() == KeyEvent.VK_RIGHT)
				scrollHorizontally ( 5*nColSeparation);
			if (e.getKeyCode() == KeyEvent.VK_LEFT)
				scrollHorizontally (-5*nColSeparation);
		}
	
	public int getLastVisibleRow ()
		{
			ScrollPane scroll = (ScrollPane) getParent();

			Point ptPos = scroll.getScrollPosition ();
			Dimension dimSize = scroll.getViewportSize ();

			DataRow row = null;
			int r = 0;
			
			while (r < vectDataRows.size())
				{
					row = (DataRow) vectDataRows.elementAt(r);
					if (row.y-row.leading >= ptPos.y+dimSize.height)
						break; // this one is not visible, stop
					r++;
				}
			if (r > 0)
				r--; // and backup to the last visible
			return r;
		}
	
	public int getFirstVisibleRow ()
		{
			Point ptPos = ((ScrollPane) getParent()).getScrollPosition ();

			DataRow row = null;
			int r = 0;
			
			while (r < vectDataRows.size())
				{
					row = (DataRow) vectDataRows.elementAt(r);
					if (row.y+row.height+row.ending > ptPos.y)
						break; // this one is visible, stop
					r++;
				}
			return r;
		}
	
	public void scrollHorizontally (int dx)
		{
			ScrollPane scroll = (ScrollPane) getParent();
			
			Point ptPos = scroll.getScrollPosition ();
			scroll.setScrollPosition (ptPos.x+dx,ptPos.y);							
		}

	public void scrollToRow (int r)
		{
			if (vectDataRows.isEmpty ())
				return;
			
			if (r < 0)
				r = 0;
			if (r >= vectDataRows.size())
				r = vectDataRows.size()-1;

			DataRow row = (DataRow) vectDataRows.elementAt(r);
					
			ScrollPane scroll = (ScrollPane) getParent();
			
			Point ptPos = scroll.getScrollPosition ();
			scroll.setScrollPosition (ptPos.x,row.y-row.leading);							
		}

	public void paint(Graphics g)
		{
			Rectangle bound = g.getClipBounds();
			
			// draw some background colors
			
			for (int c = 0; c < NUM_COLS; c++)
				if (colColors[c] != null && nWidthCol[c] > 0)
					{
						g.setColor (colColors[c]);
						g.fillRect (nStartXCol[c]-nColSeparation/2,0,nWidthCol[c]+nColSeparation,dimNeedSize.height);	
					}
			

			// only draw text that has to be updated
			
			g.setColor (new Color(0,0,0));

			// text is separated into rows, first skip the ones that are not visible
			
			DataRow row = null;
			int r = 0;
			
			while (r < vectDataRows.size())
				{
					row = (DataRow) vectDataRows.elementAt(r);
					if (row.y+row.height+row.ending > bound.y)
						break; // this one is visible, start drawing
					r++;
				}

			while (r < vectDataRows.size()-1)
				{
					row = (DataRow) vectDataRows.elementAt(r);
					
					if (row.y-row.leading > bound.y+bound.height)
						break; // this is not visible, stop drawing
						
					row.paint (g,this);
											
					r++;
				}
			
			////// draw jump arrows
			
			if (bUseCol[COL_JUMP_UP])
				{
					if (bHaveCol[COL_JUMP_UP])
						{
							Iterator i = vectJumpArrows[JUMP_UP].iterator();
							
							while (i.hasNext())
								((JumpArrow) i.next()).paint(g,this,nStartXCol[COL_JUMP_UP]+nWidthCol[COL_JUMP_UP],-1);
						}						
					if (bHaveCol[COL_JUMP_DOWN])
						{
							Iterator i = vectJumpArrows[JUMP_DOWN].iterator();
							
							while (i.hasNext())
								((JumpArrow) i.next()).paint(g,this,nStartXCol[COL_JUMP_DOWN],1);
						}
				}
/*
	Vector vectDataText;
	Vector vectRowsY;
*/



	/*		
			g.drawOval (2,2,10,10);	
			
			g.drawString("Hello",5,50);
*/
/*			
			Font font = getFont();
			
			System.out.println ("Font: "+font);
			
			FontMetrics fm = g.getFontMetrics ();
			
			System.out.println ("fm: "+fm);
			System.out.println ("Dim: "+fm.getStringBounds("Hello",g));
			
			System.out.println ("|getAscent: "+fm.getAscent()+
													"|getDescent: "+fm.getDescent()+
													"|getMaxAscent: "+fm.getMaxAscent()+
													"|getMaxDescent: "+fm.getMaxDescent()+
													"|getHeight: "+fm.getHeight()
													);
*/

		}	

}
