/*  disasmviewer, part of exmhcg - executable program tuner
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


import java.awt.*;
import java.util.*;

public class JumpArrow
{
	Vector vectFrom = new Vector();
	Vector vectTo = new Vector();
	
	int nLevel = 0;
	
	int minIndex, maxIndex;
	
	JumpArrow (CodeSnippet f,CodeSnippet t,int fi,int ti)
		{
			fi = fi*2+1;
			ti = ti*2;
			
			vectFrom.add(new JumpArrowAnchor(f,fi));
			vectTo.add(new JumpArrowAnchor(t,ti));
			
			minIndex = Math.min(fi,ti);
			maxIndex = Math.max(fi,ti);
		}
	
	void addFrom (CodeSnippet f,int fi)
		{
			fi = fi*2+1;
			vectFrom.add(new JumpArrowAnchor(f,fi));
			minIndex = Math.min(fi,minIndex);
			maxIndex = Math.max(fi,maxIndex);
		}
	
	void addTo (CodeSnippet t,int ti)
		{
			ti = ti*2;
			vectTo.add(new JumpArrowAnchor(t,ti));
			minIndex = Math.min(ti,minIndex);
			maxIndex = Math.max(ti,maxIndex);
		}
	
	void addFrom (JumpArrowAnchor anchor)
		{
			vectFrom.add(anchor);
			minIndex = Math.min(anchor.index,minIndex);
			maxIndex = Math.max(anchor.index,maxIndex);
		}
	
	void addTo (JumpArrowAnchor anchor)
		{
			vectTo.add(anchor);
			minIndex = Math.min(anchor.index,minIndex);
			maxIndex = Math.max(anchor.index,maxIndex);
		}
	
	JumpArrowAnchor getTo ()
		{
			return (JumpArrowAnchor) vectTo.firstElement ();			
		}
	
	JumpArrowAnchor getFrom ()
		{
			return (JumpArrowAnchor) vectFrom.firstElement ();			
		}
	
	Vector vectInsideFully = null;
	Vector vectInsideBegin = null;
	
	void addInsideFully(JumpArrow inside)
		{
			if (vectInsideFully == null)
				vectInsideFully = new Vector();
			vectInsideFully.add (inside);			
		}
	
	void addInsideBegin(JumpArrow inside)
		{
			if (vectInsideBegin == null)
				vectInsideBegin = new Vector();
			vectInsideBegin.add (inside);			
		}
	
	int assignLevel ()
		{
			if (nLevel > 0)
				return nLevel;
				
			nLevel = 1;
				
			if (vectInsideFully != null)
				for (Iterator i = vectInsideFully.iterator(); i.hasNext(); )
					{						
						JumpArrow insideFully = (JumpArrow) i.next();
						nLevel = Math.max(nLevel,insideFully.assignLevel()+1);					
					}		
			if (vectInsideBegin != null)
				for (Iterator i = vectInsideBegin.iterator(); i.hasNext(); )
					{						
						JumpArrow insideBegin = (JumpArrow) i.next();
						nLevel = Math.max(nLevel,insideBegin.assignLevel()+1);					
					}		

			return nLevel;			
		}
	
	boolean compactLevel (boolean used[])
		{
			boolean bAgain = false;
			
			if (vectInsideFully != null)
				for (Iterator i = vectInsideFully.iterator(); i.hasNext(); )
					bAgain |= ((JumpArrow) i.next()).compactLevel(used);
			
			for (int l = 1; l < nLevel; l++)
				used[l] = false;
			
			// now find out if there are any unused levels
/*
			if (vectInsideFully != null)
				for (Iterator i = vectInsideFully.iterator(); i.hasNext(); )
					used[((JumpArrow) i.next()).nLevel] = true;
*/
			if (vectInsideBegin != null)
				for (Iterator i = vectInsideBegin.iterator(); i.hasNext(); )
					used[((JumpArrow) i.next()).nLevel] = true;

			for (int l = 1; l < nLevel; l++)
				if (!used[l])
					{
						// level l is < nLevel, and free, so use it instead
						nLevel = l;
						return true;
					}
			return false;
		}
	
	void clearInside ()
		{
			vectInsideFully = null;
			vectInsideBegin = null;
		}
	
	JumpAnchor getAnchorFromHashtable (Hashtable hashSnippetAnchor,CodeSnippet snippet)
		{
			JumpAnchor anchor = (JumpAnchor) hashSnippetAnchor.get(snippet);
			if (anchor != null)
				return anchor;
			anchor = new JumpAnchor();
			hashSnippetAnchor.put (snippet,anchor);
			return anchor;
		}
	
	void assignSnippetAnchors (int index,Hashtable hashSnippetAnchor)
		{
			Iterator i = vectFrom.iterator();
			
			while (i.hasNext())
				{
					JumpArrowAnchor arrowAnchor = (JumpArrowAnchor) i.next();	
					JumpAnchor anchor = getAnchorFromHashtable(hashSnippetAnchor,arrowAnchor.snippet);
					anchor.addArrow(arrowAnchor,JumpAnchor.OUTGOING,index);
				}
			
			i = vectTo.iterator();
			
			while (i.hasNext())
				{
					JumpArrowAnchor arrowAnchor = (JumpArrowAnchor) i.next();	
					JumpAnchor anchor = getAnchorFromHashtable(hashSnippetAnchor,arrowAnchor.snippet);
					anchor.addArrow(arrowAnchor,JumpAnchor.INCOMING,index);
				}
		}
	
	int getWidth()
		{
			// 5  free on each side   => 10
			// 5  for each level (lowest level is 1)
			// 5  for arrow           => 5			
			
			return 10 + 5 * nLevel + 5;		
		}
	
	static final Color colors[] = { new Color(.4f,.4f,.4f),new Color(0,0,0) };
	
	public void paint(Graphics g,DataPane pane,int nBorderX,int nDirX)
		{
			int minY = Integer.MAX_VALUE; 
			int maxY = Integer.MIN_VALUE; 
			
			int nEndX  = nBorderX + nDirX * 5;
			int nArrowX  = nBorderX + nDirX * 10;
			int nLineX = nBorderX + nDirX * (5 + 5 * nLevel + 5);
			
			g.setColor (colors[nLevel % 2]);
			
			Iterator i = vectFrom.iterator();
			
			while (i.hasNext())
				{
					JumpArrowAnchor arrowAnchor = (JumpArrowAnchor) i.next();	
					
					int y = arrowAnchor.getY();
					
					g.drawLine(nEndX,y,nLineX,y);
					
					minY = Math.min(minY,y);
					maxY = Math.max(maxY,y);
				}
			
			i = vectTo.iterator();
			
			while (i.hasNext())
				{
					JumpArrowAnchor arrowAnchor = (JumpArrowAnchor) i.next();	
					
					int y = arrowAnchor.getY();
					
					g.drawLine(nEndX,y,nLineX,y);
					g.drawLine(nEndX,y,nArrowX,y-5);
					g.drawLine(nEndX,y,nArrowX,y+5);
					
					minY = Math.min(minY,y);
					maxY = Math.max(maxY,y);
				}
			
			g.drawLine(nLineX,minY,nLineX,maxY);
			
			//g.drawString(""+minIndex,nLineX,minY);
			//g.drawString(""+maxIndex,nLineX,maxY);
		}
}
