/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// BranchInstr.cpp: implementation of the BranchInstr class.
//
//////////////////////////////////////////////////////////////////////

#include "BranchInstr.h" 
#include "OperandImmediate.h" 

#include <algorithm>

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

BranchInstr::BranchInstr()
{
  m_pTargetOp = NULL;
  
  // m_bVectTargetsComplete = false;
}

BranchInstr::~BranchInstr()
{
  delete m_pTargetOp;

}

void BranchInstr::AddTarget (CtrlTransferTargetBase* pTarget/*,bool bComplete*/)
{
  assert (pTarget);
  
  assert (find (m_vectTargets.begin(),m_vectTargets.end(),pTarget) == 
	  m_vectTargets.end()); // only one, and therefore, not before
  
  m_vectTargets.push_back (pTarget);
  // m_bVectTargetsComplete |= bComplete;
}

void BranchInstr::RemoveTarget (CtrlTransferTargetBase* pTarget)
{
  VectTargets::iterator erase = find (m_vectTargets.begin(),
				      m_vectTargets.end(),
				      pTarget);
  
  assert (erase != m_vectTargets.end());
  
  m_vectTargets.erase(erase);
  
  assert (find (m_vectTargets.begin(),m_vectTargets.end(),pTarget) == 
	  m_vectTargets.end()); // only one
  // m_bVectTargetsComplete = false ????
}

char* BranchInstr::GenOpAsmString (char* pDest) const
{
  pDest = Instruction::GenOpAsmString (pDest);
  
  if (m_pTargetOp)
    {
      pDest = m_pTargetOp->GenAsmString (pDest);
    }
  return pDest;
}

bool BranchInstr::CanTraceContinue ()
{
  assert ((m_nInstruction & instrgMASK) == instrgCTRL_TRANSFER);
  
  word nType = m_nInstruction & instrmCTRL_TRANSFER_TYPE_MASK;
  
  return nType == instrmJMP_COND ||
    nType == instrmINT ||
    nType == instrmCALL;
}

bool BranchInstr::IsTargetRelative ()
{
  assert ((m_nInstruction & instrgMASK) == instrgCTRL_TRANSFER);
  
  word nType = m_nInstruction & instrmCTRL_TRANSFER_TYPE_MASK;
  
  return nType == instrmJMP_COND ||
    ((nType == instrmCALL || nType == instrmJMP) && 
     (m_nInstruction & instrmABS_REL_MASK) == instrmREL);
}

sdword BranchInstr::GetOffset ()
{
  assert (dynamic_cast<OperandImmediate*>(m_pTargetOp));
  
  OperandImmediate *pOpImm = static_cast<OperandImmediate*>(m_pTargetOp);
  
  if (pOpImm->m_bImmBytes == 1)
    return (sdword) (sbyte) pOpImm->m_dwImm;
  if (pOpImm->m_bImmBytes == 2)
    return (sdword) (sword) pOpImm->m_dwImm;
  return pOpImm->m_dwImm;
}
	




