/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// BranchInstr.h: interface for the BranchInstr class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__BRANCH_INSTR_H__INCLUDED_)
#define __BRANCH_INSTR_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

class Operand;
#include "Instruction.h"
#include "CtrlTransferTarget.h" 

#include <vector>

using namespace std ;

class CtrlTransferTarget;

typedef vector<CtrlTransferTargetBase*> VectTargets;

class BranchInstr : public Instruction  
{
 public:
  BranchInstr();
  virtual ~BranchInstr();
  
 public:
  Operand* m_pTargetOp;
  
 public:
  void SetTargetOp (Operand* pOp) 
    { 
      assert (m_pTargetOp == NULL); 
      m_pTargetOp = pOp; 
    }
  
 public:
  virtual const Operand* GetTargetOp () const { return m_pTargetOp; }
  virtual const Operand* GetDestOp () const { return m_pTargetOp; }
  
 public:
  VectTargets	m_vectTargets;
  // bool		m_bVectTargetsComplete;	
  /* the list may be incomplete if the instruction get the target from register or
   * memory and we have not been able to find all possible targets (which
   * disallows jump/call/return interface modification (simplification))
   */
  
 public:
  void AddTarget (CtrlTransferTargetBase* pTarget/*-,bool bComplete = false*/);
  void RemoveTarget (CtrlTransferTargetBase* pTarget);
  
 public:
  bool IsInterrupt () 
    { assert ((m_nInstruction & instrgMASK) == instrgCTRL_TRANSFER);
    return (m_nInstruction & instrmCTRL_TRANSFER_TYPE_MASK) == instrmINT; }
  bool IsCall ()
    { assert ((m_nInstruction & instrgMASK) == instrgCTRL_TRANSFER);
    return (m_nInstruction & instrmCTRL_TRANSFER_TYPE_MASK) == instrmCALL; }
  bool IsJump ()
    { assert ((m_nInstruction & instrgMASK) == instrgCTRL_TRANSFER);
    return (m_nInstruction & instrmCTRL_TRANSFER_TYPE_MASK) == instrmJMP || 
      (m_nInstruction & instrmCTRL_TRANSFER_TYPE_MASK) == instrmJMP_COND; }
  bool IsUnConditionalJump ()
    { assert ((m_nInstruction & instrgMASK) == instrgCTRL_TRANSFER);
    return (m_nInstruction & instrmCTRL_TRANSFER_TYPE_MASK) == instrmJMP; }
  bool IsConditionalJump () 
    { assert ((m_nInstruction & instrgMASK) == instrgCTRL_TRANSFER);
    return (m_nInstruction & instrmCTRL_TRANSFER_TYPE_MASK) == instrmJMP_COND;}
  
  bool CanTraceContinue ();
  bool IsTargetRelative ();
  
  sdword GetOffset ();
  
 protected:
  virtual char* GenOpAsmString (char* pDest) const;
};

#endif // !defined(__BRANCH_INSTR_H__INCLUDED_)
