/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// BranchTarget.cpp: implementation of the BranchTarget class.
//
//////////////////////////////////////////////////////////////////////

#include "BranchTarget.h" 
#include "CtrlTransferTarget.h" 
#include <stdio.h>
#include "JumpTarget.h" 
#include "CallTarget.h" 
#include "JavaDisAsmFile.h" 
#include "ReportDisAsmInfo.h" 
#include "Instruction.h"
#include "Operand.h"

#include "DataAnalyseFunction.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

BranchTarget::BranchTarget()
{
  m_pSource = NULL;
  
  m_pPrev = NULL;
  m_pNext = NULL;
  
  //	m_bListFBIComplete = false;
}

BranchTarget::~BranchTarget()
{

}

void BranchTarget::SetCallSource (CallTarget* pCallSrc)
{
  if (m_pSource)
    {
      assert (!m_pSource->m_pMoreSource);
      assert (dynamic_cast<JumpTarget*>(m_pSource));
      
      m_pSource->m_pMoreSource = pCallSrc;
    }
  else
    m_pSource = pCallSrc;
  
  pCallSrc->m_pTarget = this;
}

void BranchTarget::SetJumpSource (JumpTarget* pJumpSrc)
{
  if (m_pSource)
    {
      assert (!pJumpSrc->m_pMoreSource);
      assert (dynamic_cast<CallTarget*>(m_pSource));
      
      pJumpSrc->m_pMoreSource = m_pSource;
      m_pSource = pJumpSrc;
    }
  else
    m_pSource = pJumpSrc;
  
  pJumpSrc->m_pTarget = this;
}

CallTarget* BranchTarget::GetCallSource ()
{
  JumpTarget* pJumpSrc = dynamic_cast<JumpTarget*>(m_pSource);
  return dynamic_cast<CallTarget*>(pJumpSrc ? 
				   m_pSource->m_pMoreSource : m_pSource);
}

JumpTarget* BranchTarget::GetJumpSource ()
{
  return dynamic_cast<JumpTarget*>(m_pSource);	
}

void BranchTarget::SetNextTarget (BranchTarget* pNext)
{
  assert (pNext != this);   // no loops
  assert (!m_pNext);        // must not have any next yet (as we're about to set one)
  
  assert (pNext);           // must be something
  assert (!pNext->m_pPrev); // must not have any prev
  
  m_pNext = pNext;
  pNext->m_pPrev = this;
  
#ifndef NDEBUG
  // now make sure we have not got a loop
  BranchTarget* pTry = this;
  
  while (pTry->m_pNext)
    {
      assert (pTry->m_pNext != this);
      pTry = pTry->m_pNext;
    }
#endif//NDEBUG
}

void BranchTarget::ClearNextTarget ()
{
  if (m_pNext)
    {
      assert (m_pNext->m_pPrev == this);
      m_pNext->m_pPrev = NULL;
      m_pNext = NULL;
    }
}

void BranchTarget::SpliceInAsPrev (BranchTarget* pPrev)
{
  assert (!pPrev->m_pNext);
  assert (!pPrev->m_pPrev);
  
  if (m_pPrev)
    {
      assert (m_pPrev->m_pNext == this);
      
      m_pPrev->m_pNext = pPrev;
      pPrev->m_pPrev = m_pPrev;
    }
  
  pPrev->m_pNext = this;
  m_pPrev = pPrev;
}

void BranchTarget::SpliceInAsNext (BranchTarget* pNext)
{
  assert (!pNext->m_pNext);
  assert (!pNext->m_pPrev);
  
  if (m_pNext)
    {
      assert (m_pNext->m_pPrev == this);
      
      m_pNext->m_pPrev = pNext;
      pNext->m_pNext = m_pNext;
    }
  
  pNext->m_pPrev = this;
  m_pNext = pNext;
}
/*
BranchTarget* BranchTarget::GetNextCodeTarget ()
{
	BranchTarget* pNext = m_pNext;

	while (dynamic_cast<CtrlTransferTarget*>(pNext))
		pNext = pNext->m_pNext;
	return pNext;
}
*/
/*
BranchTarget* BranchTarget::GetThisCodeTarget ()
{
	BranchTarget* pCode = this;

	while (dynamic_cast<CtrlTransferTarget*>(pCode))
		pCode = pCode->m_pNext;
	return pCode;
}
*/
/*
BranchTarget* BranchTarget::GetPrevCodeTarget ()
{
	BranchTarget* pPrev = m_pPrev;

	while (dynamic_cast<CtrlTransferTarget*>(pPrev))
		pPrev = pPrev->m_pPrev;
	return pPrev;
}
*/
/*
CtrlTransferTarget* BranchTarget::GetTraceHead ()
{
	BranchTarget* pHead = this;

	while (pHead->m_pPrev)
		pHead = pHead->m_pPrev;

	// the head should always be a target

	assert (dynamic_cast<CtrlTransferTarget*>(pHead));

	return static_cast<CtrlTransferTarget*>(pHead);
}
*/
BranchTarget* BranchTarget::GetTraceHead ()
{
  BranchTarget* pHead = this;
  
  while (pHead->m_pPrev)
    pHead = pHead->m_pPrev;
  
  return pHead;
}

void BranchTarget::RedirectPointers (BranchTarget* pFirstReplace,
				     BranchTarget* pLastReplace)
{
  if (m_pPrev)
    {
      assert (m_pPrev->m_pNext == this);
      m_pPrev->m_pNext = pFirstReplace;
      pFirstReplace->m_pPrev = m_pPrev;
    }
  if (m_pNext)
    {
      assert (m_pNext->m_pPrev == this);
      m_pNext->m_pPrev = pLastReplace;
      pLastReplace->m_pNext = m_pNext;
    }
  
  CtrlTransferTarget* pSource = m_pSource;
  
  while (pSource)
    {
      assert (pSource->m_pTarget == this);
      pSource->m_pTarget = pFirstReplace;
      pSource = pSource->m_pMoreSource;
    }
  pFirstReplace->m_pSource = m_pSource;
}

#ifndef NDEBUG
void BranchTarget::ConsistencyCheck (CallFunction* pFunction)
{
  // make sure linking is ok (fall-through lists)
  
  assert (!m_pPrev || m_pPrev->m_pNext == this);
  assert (!m_pNext || m_pNext->m_pPrev == this);
  
  // make sure linking is ok (source objects lists)
  
  CtrlTransferTarget* pSource = m_pSource;
  
  if (pSource)
    {
      JumpTarget* pJumpSrc = dynamic_cast<JumpTarget*>(pSource);
      
      if (pJumpSrc)
	{
	  assert (pSource->m_pTarget == this);
	  pSource = pJumpSrc->m_pMoreSource;
	}
      
      CallTarget* pCallSrc = dynamic_cast<CallTarget*>(pSource);
      
      if (pCallSrc)
	{
	  assert (pSource->m_pTarget == this);
	  assert (pCallSrc->GetCallFunction() == pFunction ||
		  pFunction == NULL);
	  pSource = pCallSrc->m_pMoreSource;
	}
      assert (pSource == NULL); // we should now have had all
    }
}
#endif

/*
string BranchTarget::JdafGetId (const MapBranchTargetVA* pMapBT_VA)
{
	char fmt[32];

	sprintf (fmt,"0x%08x",pMapBT_VA->find(this)->second);

	return string(fmt);
}

string BranchTarget::JdafGetName (const MapBranchTargetVA* pMapBT_VA)
{
	char fmt[32];

	sprintf (fmt,"%08x",pMapBT_VA->find(this)->second);

	return string(fmt);
}
*/

#include "HtmlPage.h"
#include "HtmlTags.h" // dep_chk_unknown     

void BranchTarget::HtmlPlainDisAsm (HtmlPage& page,
				    const ReportDisAsmInfo& reportDisAsmInfo) const
{
  CtrlTransferTarget* pSource = m_pSource;
  
  while (pSource)
    {
      pSource->HtmlPlainDisAsm(page,reportDisAsmInfo);
      
      if (reportDisAsmInfo.m_pAnalyseFunction)
	{
	  MapCTTEntryState::const_iterator entry = 
	    reportDisAsmInfo.m_pAnalyseFunction->mapEntryState.find(pSource);
	  
	  if (entry != reportDisAsmInfo.m_pAnalyseFunction->mapEntryState.end())
	    {
	      page << "ENTRY:\n";
	      entry->second->HtmlPlainDisAsm(page,reportDisAsmInfo,false);
	      //page << "ENDENTRY\n";
	    }
	  
	  MapCTTIncomingState::const_iterator incoming = 
	    reportDisAsmInfo.m_pAnalyseFunction->mapIncomingState.find(pSource);
	  
	  if (incoming != reportDisAsmInfo.m_pAnalyseFunction->mapIncomingState.end())
	    {
	      page << "INCOMING:\n";
	      incoming->second->HtmlPlainDisAsm(page,reportDisAsmInfo,false);
	      //page << "ENDINCOMING\n";
	    }
	  
	}      
      pSource = pSource->m_pMoreSource;
    }
}

void BranchTarget::JdafDisAsm (JdafAsmCodeSnippet& snippet,
			       const ReportDisAsmInfo& reportDisAsmInfo) const
{
  snippet.SetId(reportDisAsmInfo.GetLabel(this));
  
  dword dwOffset;
  
  if (reportDisAsmInfo.GetOffset(this,dwOffset))
    snippet.SetOffset(dwOffset);
  
  CtrlTransferTarget* pSource = m_pSource;
  
  while (pSource)
    {
      pSource->JdafDisAsm(snippet,reportDisAsmInfo);
      pSource = pSource->m_pMoreSource;
    }
  
  if (m_pPrev)
    snippet.SetFalledThrough();
  if (m_pNext)
    snippet.SetFallsThrough();
}


void BranchTarget::HtmlPlainDisAsm (HtmlPage& page,
				    const Instruction* pInstruction,
				    const byte* pRawCode,byte bLength,
				    dword dwVA,bool bVA,bool bFullAddress,
				    const ReportDisAsmInfo& reportDisAsmInfo) const
{	
  char fmt[256];
  char fmt2[64];
  char fmt3[64];
  char fmt4[128];
  
  fmt2[0] = 0;
  char* ptr = fmt2;
  
  int l =  bLength;
  for (int k = 7; l && k; l--, k--)
    {
      byte bByte = *(pRawCode++);				
      ptr += sprintf (ptr,"%02x ",bByte);
    }
  
  pInstruction->InstrAsmString(fmt3);
  pInstruction->OpAsmString(fmt4);
  
  if (bFullAddress)
    page.Begin(a().name(reportDisAsmInfo.GetLabel(this)));
  
  if (bVA)
    sprintf (fmt,bFullAddress ? "0x%08x  %-21s %-9s %s\n" :
             "  %8x  %-21s %-9s %s\n",dwVA,fmt2,fmt3,fmt4);
  else
    sprintf (fmt,           "            %-21s %-9s %s\n",fmt2,fmt3,fmt4);
  
  page << fmt;
  
  if (bFullAddress)
    page.End();

  while (l)
    {
      ptr = fmt2;
      for (int k = 7; l && k; l--, k--)
        {
          byte bByte = *(pRawCode++);				
          ptr += sprintf (ptr,"%02x ",bByte);
        }
      page << "            " << fmt2 << '\n';
    }
}

JdafAsmInstr* BranchTarget::JdafDisAsm (const Instruction* pInstr,const byte* pRawCode,byte bLength) const
{
  JdafAsmInstr* pJdafInstr = new JdafAsmInstr();
  
  char fmt[64];
  
  pInstr->InstrAsmString(fmt);

  pJdafInstr->SetCode(bLength,pRawCode);
  pJdafInstr->SetInstr(fmt);
  
  const Operand* pOp;
  
  if ((pOp = pInstr->GetDestOp()))
    pJdafInstr->m_pOpDest = pOp->JdafOp();
  if ((pOp = pInstr->GetSrcOp()))
    pJdafInstr->m_pOpSrc  = pOp->JdafOp();
  if ((pOp = pInstr->GetSrc2Op()))
    pJdafInstr->m_pOpSrc2 = pOp->JdafOp();
  
  return pJdafInstr;
}

/*
BranchInstr* BranchTarget::GetBranchInstr ()
{
	if (m_vectInstructions.empty())
		return NULL;
	
	return dynamic_cast<BranchInstr*>(m_vectInstructions[0]);
}
*/


/*
// // #include "HtmlTags.h" // dep_chk_unknown     

void BranchTarget::HtmlTable (HtmlPage& page)
{
//	page.Begin(table().border(1));
	
	char fmt1[64];
	char fmt2[64];
	char fmt3[64];
	char fmt4[64];
	
	dword dwVA = m_dwVA;
	byte* pRawCode = m_pRawCode;
	
	for (int i = 0; i < m_vectInstructions.size(); i++)
	{
		Instruction* pInstruction = m_vectInstructions[i];
		byte bLength = m_vectRawInstrLengths[i];

		sprintf (fmt1,"0x%08lx",dwVA);

		fmt2[0] = 0;
		char* ptr = fmt2;

		for (byte l = bLength; l; l--)
		{
			byte bByte = *(pRawCode++);				
			ptr += sprintf (ptr,"%02x ",bByte);
		}

		pInstruction->InstrAsmString(fmt3);
		pInstruction->OpAsmString(fmt4);

		page << (tr() << td(fmt1) << td(fmt2) << td(fmt3) << td(fmt4));


		
		
		dwVA += bLength;
	}
	
 // for (vector<Instruction*>::iterator i = m_vectInstructions.begin();
 //      i != m_vectInstructions.end (); i++)
 //   {
 //     (*i)->AsmString (fmt);
 //     
 //     page << (tr() << td(fmt));
 //   }

  //vector<Instruction*>& GetInstructions () { return m_vectInstructions; }
  //vector<Instruction*> m_vectInstructions;
  

 // page.End ();
}
*/



