/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// BranchTarget.h: interface for the BranchTarget class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__BRANCH_TARGET_H__INCLUDED_)
#define __BRANCH_TARGET_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "GlobalDefs.h"
class HtmlPage; 
class Instruction; 
class ReportDisAsmInfo; 
class CallTarget; 
class JumpTarget; 

class JdafAsmCodeSnippet;
class JdafAsmInstr;
class CallFunction;

#include <vector>
#include <set>
#include <map>

using namespace std;

class BranchTarget;

typedef set<BranchTarget*> SetBranchTargets;

typedef map<const BranchTarget*,dword> MapBranchTargetVA;

class CtrlTransferTarget;

class BranchTarget
{
 public:
  BranchTarget();
  virtual ~BranchTarget();
  
 protected:
  // how can we get here ?
  
  // dword m_dwVA;
  
 public: // TODO: to be removed, as is only used by cheating code
  CtrlTransferTarget* m_pSource;
  
 protected:
  BranchTarget* m_pPrev; // fall through from previous
  BranchTarget* m_pNext; // fall through to next
  
  //	bool m_bListFromTraceComplete;	 // is this useful?

 public:
  void SetCallSource (CallTarget* pCallSrc);
  void SetJumpSource (JumpTarget* pJumpSrc);
  
  CallTarget* GetCallSource ();
  JumpTarget* GetJumpSource ();
  
  CtrlTransferTarget* GetCtrlTransferTarget () const { return m_pSource; }
  
 public:
  void SetNextTarget (BranchTarget* pNext);
  // void SetPrevTarget (BranchTarget* pPrev); // use SetNextTarget on pPrev

  void ClearNextTarget(); // TODO: to be removed, as is only used by cheating code
  
  void SpliceInAsPrev (BranchTarget* pPrev);
  void SpliceInAsNext (BranchTarget* pNext);
  
  BranchTarget* GetNextTarget () { return m_pNext; }
  BranchTarget* GetPrevTarget () { return m_pPrev; }
  const BranchTarget* GetNextTarget () const { return m_pNext; }
  const BranchTarget* GetPrevTarget () const { return m_pPrev; }
  //	BranchTarget* GetNextCodeTarget (); // get next, skipping any CtrlTransferTarget
  //	BranchTarget* GetPrevCodeTarget ();
  //	BranchTarget* GetThisCodeTarget (); // jump pase any CtrlTransferTarget
  BranchTarget* GetTraceHead ();
  
 public:
  void RedirectPointers (BranchTarget* pFirstReplace,
			 BranchTarget* pLastReplace);
  
 public:
  virtual void HtmlPlainDisAsm (HtmlPage& page,
				const ReportDisAsmInfo& reportDisAsmInfo) const;
  virtual void JdafDisAsm (JdafAsmCodeSnippet& snippet,
			   const ReportDisAsmInfo& reportDisAsmInfo) const;
  
  // void HtmlPlainDisAsm (HtmlPage& page,ReportDisAsmInfo& reportDisAsmInfo);
  void HtmlPlainDisAsm (HtmlPage& page,
			const Instruction* pInstruction,
			const byte* pRawCode,byte bLength,
			dword dwVA,bool bVA,bool bFullAddress,
			const ReportDisAsmInfo& reportDisAsmInfo) const;
  JdafAsmInstr* JdafDisAsm (const Instruction* pInstr,
			    const byte* pRawCode,byte bLength) const;
  
  /*
    string JdafGetId (const MapBranchTargetVA* pMapBT_VA);
	string JdafGetName (const MapBranchTargetVA* pMapBT_VA);
  */
  /*	
	public:
	void HtmlTable (HtmlPage& page);
  */
  
 public:
  virtual dword GetLengthOfRawCode () const { return 0; }
  
 public:
#ifndef NDEBUG
  virtual void ConsistencyCheck (CallFunction* pFunction = NULL);
#endif
  
};

#endif // !defined(__BRANCH_TARGET_H__INCLUDED_)
