/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// BranchTargetCtrlTransfer.cpp: implementation of the BranchTargetCtrlTransfer class.
//
//////////////////////////////////////////////////////////////////////

#include "BranchTargetCtrlTransfer.h" 
#include "CallTarget.h" 
#include "JumpTarget.h"
#include "BranchTarget.h"

#include "ReportDisAsmInfo.h"
#include "JavaDisAsmFile.h"
#include "UrlFile.h"

#include "DataAnalyseFunction.h"

#include "Operand.h"

class JumpTarget; 

#include <stdio.h>
#include <algorithm>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

BranchTargetCtrlTransfer::BranchTargetCtrlTransfer()
{

}

BranchTargetCtrlTransfer::~BranchTargetCtrlTransfer()
{

}

#ifndef NDEBUG
void BranchTargetCtrlTransfer::ConsistencyCheck (CallFunction* pFunction)
{
  BranchTarget::ConsistencyCheck(pFunction);
  
  BranchInstr* pInstr = GetBranchInstr();
  
  FOR_EACH_ITER(VectTargets,pInstr->m_vectTargets,target)
    {
      CtrlTransferTargetBase* pTarget = *target;
      
      assert (find(pTarget->m_vectFromTrace.begin(),
		   pTarget->m_vectFromTrace.end(),this) !=
	      pTarget->m_vectFromTrace.end());
    }
}
#endif

#include "HtmlPage.h"
#include "HtmlTags.h" 

void HtmlAddInternalTargetLink(HtmlPage& page,
			       const CtrlTransferTarget* pTarget,
			       const ReportDisAsmInfo& reportDisAsmInfo)
{
  page << " [";
  
  const CallTarget* pCallTarget = dynamic_cast<const CallTarget*>(pTarget);
  
  string strLabel = reportDisAsmInfo.GetLabel(pTarget->m_pTarget);
  
  assert (dynamic_cast<const JumpTarget*>(pTarget) ||
	  dynamic_cast<const CallTarget*>(pTarget));
  
  if (!pCallTarget)
    {
      // local target (jump)
      
      page << a(strLabel).href("#"+strLabel);
    }
  else
    {
      // external target (call)
      
      page << a(strLabel).href(reportDisAsmInfo.GetFunctionUrlFile(pCallTarget->GetCallFunction(),DIS_ASM_URL_HTML),
			       "#"+strLabel);
    }
  
  page << ']';	
}

/*
			JdafExternal* pExternal = new JdafExternal(,
				reportDisAsmInfo.GetLabel(pCallTarget),
				reportDisAsmInfo.GetLabel(pCallTarget));
*/

void BranchTargetCtrlTransfer::HtmlPlainDisAsm (HtmlPage& page,
						const ReportDisAsmInfo& reportDisAsmInfo) const
{
  BranchTarget::HtmlPlainDisAsm(page,reportDisAsmInfo);
  
  char fmt[256];
  char fmt2[64];
  char fmt3[64];
  char fmt4[128];
  
  assert (m_pSource || m_pPrev);
  
  bool bFullAddress = (m_pSource != NULL);
  byte* pRawCode = m_pRawCode;
  
  fmt2[0] = 0;
  char* ptr = fmt2;
  
  int l = m_bRawCodeLength;
  for (int k = 7; l && k; l--, k--)
    {
      byte bByte = *(pRawCode++);				
      ptr += sprintf (ptr,"%02x ",bByte);
    }
  
  const BranchInstr* pInstr = GetBranchInstr();
  
  pInstr->InstrAsmString(fmt3);
  pInstr->OpAsmString(fmt4);
  
  if (bFullAddress)
    page.Begin(a().name(reportDisAsmInfo.GetLabel(this)));
  
  dword dwVA;
  verify(/*bool bVA = */ reportDisAsmInfo.GetOffset(this,dwVA));
  
  sprintf (fmt,bFullAddress ? "0x%08x  %-21s %-9s %-12s" :
	   "  %8x  %-21s %-9s %-12s",dwVA,fmt2,fmt3,fmt4);
  
  page << fmt;
  
  //	sprintf (fmt,"(%d)",pInstr->m_vectTargets.size());
  //	page << fmt;
  
  for (VectTargets::const_iterator iter = pInstr->m_vectTargets.begin();
       iter != pInstr->m_vectTargets.end(); iter++)
    {
      CtrlTransferTarget* pCtrlTransferTarget = dynamic_cast<CtrlTransferTarget*>(*iter);
      if (pCtrlTransferTarget)
	HtmlAddInternalTargetLink(page,pCtrlTransferTarget,reportDisAsmInfo);
      else
	page << "[Link for imported fcn goes here.]";
      //		printf ("%x ",*iter);
    }
  
  /*
    for_each(pInstr->m_vectTargets.begin(),pInstr->m_vectTargets.end(),
    HtmlAddTargetLink,page);	
  */	
  page << '\n';
  
  if (bFullAddress)
    page.End();
  
  while (l)
    {
      ptr = fmt2;
      for (int k = 7; l && k; l--, k--)
	{
	  byte bByte = *(pRawCode++);				
	  ptr += sprintf (ptr,"%02x ",bByte);
	}
      page << "            " << fmt2 << '\n';
    }
  
  
  if (reportDisAsmInfo.m_pAnalyseFunction)
    {
      MapBTOutgoingState::const_iterator outgoing = 
	reportDisAsmInfo.m_pAnalyseFunction->mapOutgoingState.find(this);

      page << "OUTGOING:\n";
      if (outgoing != 
	  reportDisAsmInfo.m_pAnalyseFunction->mapOutgoingState.end())
	outgoing->second->HtmlPlainDisAsm(page,reportDisAsmInfo,false);
      //page << "ENDOUTGOING\n";
    }

  if (reportDisAsmInfo.m_pAnalyseFunction)
    {
      MapBTOutgoingState::const_iterator outgoingBranch = 
	reportDisAsmInfo.m_pAnalyseFunction->mapOutgoingStateBranch.find(this);

      page << "BRANCH:\n";
      if (outgoingBranch != 
	  reportDisAsmInfo.m_pAnalyseFunction->mapOutgoingStateBranch.end())
	outgoingBranch->second->HtmlPlainDisAsm(page,reportDisAsmInfo,true);
      //page << "ENDBRANCH\n";
    }

}

//dword dwADisplayLength = 0;
 
void BranchTargetCtrlTransfer::JdafDisAsm (JdafAsmCodeSnippet& snippet,
					   const ReportDisAsmInfo& reportDisAsmInfo) const
{
  BranchTarget::JdafDisAsm (snippet,reportDisAsmInfo);
  
  //	dwADisplayLength += m_bRawCodeLength;
  //	printf ("        %d\n",dwADisplayLength);
  
  JdafAsmBranchInstr* pJdafInstr = new JdafAsmBranchInstr();
  
  const BranchInstr* pInstr = GetBranchInstr();
  
  char fmt[64];
  
  pInstr->InstrAsmString(fmt);
  
  pJdafInstr->SetCode(m_bRawCodeLength,m_pRawCode);
  pJdafInstr->SetInstr(fmt);
  
  //	pJdafInstr->SetType (pInstr->CanTraceContinue () ? JDAF_ABI_CAN_CONTINUE : JDAF_ABI_ENDS_TRACE);
  
  const Operand* pOp;
  
  if ((pOp = pInstr->GetTargetOp()))
    pJdafInstr->m_pOpDest = pOp->JdafOp();
  
  snippet.AddInstruction(pJdafInstr);
  
  for (VectTargets::const_iterator i = pInstr->m_vectTargets.begin(); i != pInstr->m_vectTargets.end(); i++)
    {
      CtrlTransferTarget* pCtrlTransferTarget = dynamic_cast<CtrlTransferTarget*>(*i);
      
      if (pCtrlTransferTarget)
	{
	  assert (dynamic_cast<JumpTarget*>(pCtrlTransferTarget) ||
		  dynamic_cast<CallTarget*>(pCtrlTransferTarget));
	  
	  CallTarget* pCallTarget = dynamic_cast<CallTarget*>(pCtrlTransferTarget);
	  
	  string strLabel = reportDisAsmInfo.GetLabel(pCtrlTransferTarget->m_pTarget);
	  
	  if (!pCallTarget)
	    {
				// internal target
	      
	      pJdafInstr->AddInternalTarget(strLabel);
	    }
	  else
	    {
				// external target
	      
	      JdafExternal* pExternal = new JdafExternal(reportDisAsmInfo.GetFunctionUrlFile(pCallTarget->GetCallFunction(),DIS_ASM_URL_JAVA),
							 strLabel,
							 strLabel);
	      
	      pJdafInstr->AddExternalTarget(pExternal);
	    }
	}
      else
	{
	  // imported function target
	  
	}
    }
}






