/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// BranchTargetTrace.cpp: implementation of the BranchTargetTrace class.
//
//////////////////////////////////////////////////////////////////////

#include "BranchTargetTrace.h" 
#include "BranchTargetTraceSingle.h" 

#include "ReportDisAsmInfo.h"
#include "JavaDisAsmFile.h"

#include <stdio.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

BranchTargetTrace::BranchTargetTrace()
{

}

BranchTargetTrace::~BranchTargetTrace()
{

}

void BranchTargetTrace::Split (dword dwOffset,
			       BranchTargetTrace* pNew,
			       dword dwCodeTraceRVA)
{
  CodeTrace::Split (dwOffset,pNew,dwCodeTraceRVA);
  
  SpliceInAsNext(pNew);
}

void BranchTargetTrace::HtmlPlainDisAsm (HtmlPage& page,
					 const ReportDisAsmInfo& reportDisAsmInfo) const
{	
  BranchTarget::HtmlPlainDisAsm(page,reportDisAsmInfo);
  
  dword dwVA;
  bool bVA = reportDisAsmInfo.GetOffset(this,dwVA);
  
  byte* pRawCode = m_pRawCode;
  
  for (unsigned int i = 0; i < m_vectInstructions.size(); i++)
    {
      Instruction* pInstruction = m_vectInstructions[i];
      byte bLength = m_vectRawInstrLengths[i];
      
      assert (m_pSource || m_pPrev);
      bool bFullAddress = (i == 0 && m_pSource);
      
      //////////////////////////////////////////////////////////

      BranchTarget::HtmlPlainDisAsm (page,pInstruction,pRawCode,
				     bLength,dwVA,bVA,
				     bFullAddress,reportDisAsmInfo);
      /*
	fmt2[0] = 0;
	char* ptr = fmt2;
	
	int l =  bLength;
		for (int k = 7; l && k; l--, k--)
		{
			byte bByte = *(pRawCode++);				
			ptr += sprintf (ptr,"%02x ",bByte);
		}

		pInstruction->InstrAsmString(fmt3);
		pInstruction->OpAsmString(fmt4);

		if (bFullAddress)
			page.Begin(a().name(reportDisAsmInfo.GetLabel(this)));

		assert (bVA); // or cannot use dwVA
		sprintf (fmt,bFullAddress ? "0x%08lx  %-21s %-9s %s\n" :
									 "  %8lx  %-21s %-9s %s\n",dwVA,fmt2,fmt3,fmt4);

		page << fmt;

		if (bFullAddress)
			page.End();

		while (l)
		{
			char* ptr = fmt2;
			for (int k = 7; l && k; l--, k--)
			{
				byte bByte = *(pRawCode++);				
				ptr += sprintf (ptr,"%02x ",bByte);
			}
			page << "            " << fmt2 << '\n';
		}
*/
//////////////////////////////////////////////////////////
		
      dwVA += bLength;
    }
}

//dword dwDisplayLength = 0;

void BranchTargetTrace::JdafDisAsm (JdafAsmCodeSnippet& snippet,
				    const ReportDisAsmInfo& reportDisAsmInfo) const
{
  BranchTarget::JdafDisAsm (snippet,reportDisAsmInfo);
  
  //	dwDisplayLength += m_dwRawCodeLength;
  //	printf ("%d\n",dwDisplayLength);
  
  byte* pRawCode = m_pRawCode;
  
  for (unsigned int i = 0; i < m_vectInstructions.size(); i++)
    {
      Instruction* pInstr = m_vectInstructions[i];
      byte bLength = m_vectRawInstrLengths[i];
      
      JdafAsmInstr* pJdafInstr = BranchTarget::JdafDisAsm (pInstr,pRawCode,bLength);
      /*
	JdafAsmInstr* pJdafInstr = new JdafAsmInstr();
	
	char fmt[64];
	
		pInstr->InstrAsmString(fmt);

		pJdafInstr->SetCode(bLength,pRawCode);
		pJdafInstr->SetInstr(fmt);

		const Operand* pOp;

		if (pOp = pInstr->GetDestOp())
			pJdafInstr->m_pOpDest = pOp->JdafOp();
		if (pOp = pInstr->GetSrcOp())
			pJdafInstr->m_pOpSrc  = pOp->JdafOp();
		if (pOp = pInstr->GetSrc2Op())
			pJdafInstr->m_pOpSrc2 = pOp->JdafOp();

*/
      pRawCode += bLength;
      snippet.AddInstruction(pJdafInstr);
    }
}

int dwDeletedMultiple = 0;
int dwCreatedSingle = 0;
int dwEstimatedRawSpill = 0;

BranchTargetTraceSingle* BranchTargetTrace::ConvertToTraceSingle (BranchTargetTraceSingle** pLastSingle)
{
  BranchTargetTraceSingle* pFirst = NULL;
  BranchTargetTraceSingle* pLast = NULL;
  
  byte* pRawCode = m_pRawCode;
  
  assert (m_vectInstructions.size()); // must have at least one instruction
  for (unsigned int i = 0; i < m_vectInstructions.size(); i++)
    {
      BranchTargetTraceSingle* pSingle = new BranchTargetTraceSingle;
      if (!pFirst)
	pFirst = pSingle;
      
      pSingle->SetInstruction(m_vectInstructions[i]);
      
      byte bLength = m_vectRawInstrLengths[i];
      
      pSingle->SetRawCode(pRawCode,bLength);
      pRawCode += bLength;
      
      if (pLast)
	pLast->SetNextTarget(pSingle);
      
      pLast = pSingle;
      
      dwCreatedSingle++;
      dwEstimatedRawSpill += (bLength + 3) & ~3;
    }
  
  assert ((dword) (pRawCode - m_pRawCode) == m_dwRawCodeLength);
  
  if (pLastSingle)
    *pLastSingle = pLast;
  
  RedirectPointers (pFirst,pLast);
  
  dwDeletedMultiple++;
  dwEstimatedRawSpill -= (m_dwRawCodeLength + 3) & ~3;
  
  return pFirst;
}
/*
void PrintConversionInfo()
{
	int nSingleSize = sizeof (BranchTargetTraceSingle);
	int nMultSize = sizeof (BranchTargetTrace);

	printf ("deleted: %d  created: %d\n",dwDeletedMultiple,dwCreatedSingle);
	printf ("size: %d  size: %d\n",nMultSize,nSingleSize);
	printf ("extra: %d\n",dwCreatedSingle*nSingleSize-dwDeletedMultiple*nMultSize);
	printf ("raw spill: %d\n",dwEstimatedRawSpill);

}
*/







