/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// CallFunction.cpp: implementation of the CCallFunction class.
//
//////////////////////////////////////////////////////////////////////

#ifdef WIN32
#include "GlobalDefs.h" 
#endif
#include "CallFunction.h" 
#include "BranchTargetCtrlTransfer.h"
#include "BranchTargetTrace.h"
#include "BranchTargetTraceSingle.h"
#include "JumpTarget.h" 
#include "ImportedFunc.h" 

#include <list>
#include <map>
#include <stack>

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CallFunction::CallFunction()
{

}

CallFunction::~CallFunction()
{
  DeleteCallReturnState (); // as we're history

}

void CallFunction::AddEntryPoint (CallTarget* pTarget)
{
  //printf ("CallFunction(%x)::AddEntryPoint(%08x)\n",this,pTarget);
  
  m_vectEntryPoints.push_back (pTarget);
  pTarget->SetCallFunction(this);
}

/*
void CallFunction::AddEntryPoint (CallTarget* pTarget,ExportedEntry* pExported)
{
	//printf ("CallFunction(%x)::AddEntryPoint(%08x)\n",this,pTarget);
	
	m_vectEntryPoints.push_back (EntryPoint(pTarget,pExported));
	pTarget->SetCallFunction(this);
}

void CallFunction::AddEntryPoint (const EntryPoint& entryPoint)
{
	m_vectEntryPoints.push_back (entryPoint);
	entryPoint.m_pTarget->SetCallFunction(this);
}
*/
void CallFunction::AddCodeTrace (BranchTarget* pTarget)
{
  //printf ("CallFunction(%x)::AddCodeTrace(%08x)\n",this,pTarget);
  //printf ("prev size: %d\n",m_setCodeTraces.size());
  
  m_setCodeTraces.insert (pTarget);
}

void CallFunction::Merge (CallFunction* pInsertFunc)
{
  //printf ("CallFunction::Merge(%08x)\n",pInsertFunc);
  
  FOR_EACH_ITER(VectEntryPoints,pInsertFunc->m_vectEntryPoints,entry)
    AddEntryPoint(*entry);
  FOR_EACH_ITER(SetCodeTraces,pInsertFunc->m_setCodeTraces,trace)
    AddCodeTrace(*trace);
}

void CallFunction::RemoveCodeTrace (BranchTarget* pTarget)
{
  assert (m_setCodeTraces.find(pTarget) != m_setCodeTraces.end());
  
  m_setCodeTraces.erase(pTarget);
}
/*
void CallFunction::GetCtrlTransferTargets (SetStackCtrlTransferTargets& setstackCtrlTransferTargets)
{
  FOR_EACH_ITER(SetCodeTraces,m_setCodeTraces,i)
    {
      BranchTarget* pCheck = *i;
      
      assert (pCheck->GetCtrlTransferTarget());
      
      while (pCheck)
	{
	  CtrlTransferTarget* pTarget = pCheck->GetCtrlTransferTarget();
	  
	  if (pTarget)
	    {
	      assert (pTarget->m_pTarget = pCheck);
	      setstackCtrlTransferTargets.push(pTarget);
	    }
	  
	  pCheck = pCheck->GetNextTarget();
	}
    }
}
*/
void CallFunction::GetCodeTraces (VectCodeTraces& vectCodeTraces)
{
  FOR_EACH_ITER(SetCodeTraces,m_setCodeTraces,i)
    vectCodeTraces.push_back(*i);
}

typedef stack<BranchTarget*> StackTargets;
typedef list<BranchTarget*> ListTargets;
typedef map<BranchTarget*,ListTargets::iterator> MapTargetsListIterator;

void CallFunction::GetSortedCodeTraces (VectCodeTraces& vectCodeTraces)
{
  ListTargets listTargets;
  MapTargetsListIterator mapTargetsAdded;
  
  StackTargets stackInvestigate;
  
  // now find all CtrlTransfers targets that can be reached by jumps
  // within this function
  
  // big problem: in what order to add them?
  //
  // a) as we only deal with the beginnings af traces, here in the
  // meaning that a trace is the entire route from a target though all
  // fall-throughs to an unconditional jump (jmp/ret), we only have to
  // order these ones the plan i simple
  //
  // 1) add them to the end as default (we assume any not yet added
  // targets that the trace points to will be added after this one
  // (which is true by default))
  //
  // 2) If the trace to be added has all it's targets already added,
  // it should of course be before those targets Therefore, we in that
  // case add it after the last trace already added that point to it
  
  // first get all entry points
  
  for (VectEntryPoints::reverse_iterator i = m_vectEntryPoints.rbegin();
       i != m_vectEntryPoints.rend(); ++i)
    stackInvestigate.push((*i)->m_pTarget->GetTraceHead());
  
  // now deal with all entry points
  
  while (!stackInvestigate.empty())
    {
      BranchTarget* pAddTarget = stackInvestigate.top();
      stackInvestigate.pop();
      
      if (mapTargetsAdded.find(pAddTarget) != mapTargetsAdded.end())
	continue; // already added (will only happen for targets multiply added by init from m_vectEntryPoints)

      assert (pAddTarget);
      
      BranchTarget* pCheck = pAddTarget;
      
      //		BranchTarget* pCheck = pAddTarget->GetThisCodeTarget(); // jump past any CtrlTransferTarget
      //
      //		if (!pCheck)
      //			continue; // empty (only jump/call target, but no code)

      // this is a target,
      // now find all internal targets and follow those too
      
      ListTargets::iterator addBefore = listTargets.end();
      ListTargets::iterator addAfter = listTargets.end();
      bool bAfterIsAfterBefore = false;
      
      BranchTargetCtrlTransfer* pCtrlTransfer;
      JumpTarget* pJumpTarget;
      
      while (pCheck)
	{
	  assert (dynamic_cast<BranchTargetCtrlTransfer*>(pCheck) ||
		  dynamic_cast<BranchTargetTrace*>(pCheck) ||
		  dynamic_cast<BranchTargetTraceSingle*>(pCheck));
	  
	  if ((pJumpTarget = pCheck->GetJumpSource()))
	    {
	      // we're an local target, see if any of our sources have
	      // already been added. In that case, we must not be
	      // added before them
	      
	      for (VectFromTrace::iterator j = 
		     pJumpTarget->m_vectFromTrace.begin();
		   j != pJumpTarget->m_vectFromTrace.end(); j++)
		{
		  BranchTarget* pSrc = (*j)->GetTraceHead ();
		  
		  MapTargetsListIterator::iterator src = 
		    mapTargetsAdded.find (pSrc);
		  
		  if (src != mapTargetsAdded.end())
		    {
		      // source is known
		      
		      if (addAfter == listTargets.end() ||
			  distance(addAfter,src->second) > 0)
			{
			  // src->second is after addAfter
			  
			  addAfter = src->second;
			  if (distance(addAfter,addBefore))
			    bAfterIsAfterBefore = true;
			}
		    }
		}
	    }
	  
	  if ((pCtrlTransfer = 
	       dynamic_cast<BranchTargetCtrlTransfer*>(pCheck)))
	    {
	      // it's a branch, let's find out where it's going
	      
	      BranchInstr* pBranchInstr = pCtrlTransfer->GetBranchInstr();
	      
	      for (VectTargets::iterator i = pBranchInstr->m_vectTargets.begin();
		   i != pBranchInstr->m_vectTargets.end (); i++)
		{
		  CtrlTransferTargetBase* pTargetDest = *i;
		  
		  assert (dynamic_cast<ImportedFunc*>(pTargetDest) || // local
			  dynamic_cast<JumpTarget*>(pTargetDest) || // local
			  dynamic_cast<CallTarget*>(pTargetDest)); // globel
		  
		  JumpTarget* pJumpTargetDest;
		  
		  if ((pJumpTargetDest = 
		       dynamic_cast<JumpTarget*>(pTargetDest)))
		    {
		      // it's a local target, let's add it
		      
		      BranchTarget* pTarget = 
			pJumpTargetDest->m_pTarget->GetTraceHead ();
		      
		      if (pTarget != pAddTarget)
			{
			  // not looping back to ourselves
			  
			  assert(m_setCodeTraces.find(pTarget) != 
				 m_setCodeTraces.end());
			  
			  // make sure we're not added before
			  
			  MapTargetsListIterator::iterator dest = 
			    mapTargetsAdded.find (pTarget);
			  
			  if (dest != mapTargetsAdded.end())
			    {
			      // target added before
			      
			      if (!bAfterIsAfterBefore && // we won't be added before addAfter anyway
				  distance(dest->second,addBefore) > 0)
				{
				  // dest->second is before addBefore
				  
				  addBefore = dest->second;
				  if (distance(addAfter,addBefore))
				    bAfterIsAfterBefore = true;
				}
			    }
			  else
			    {
			      // target not added before
			      
			      stackInvestigate.push(pTarget);
			    }
			}
		    }
		}
	    }
	  pCheck = pCheck->GetNextTarget();
	}
      // now add ourselves
      
      if (addAfter != listTargets.end() /*bAfterIsAfterBefore*/)
	{
	  addBefore = addAfter;
	  ++addBefore;
	}
      
      ListTargets::iterator pos = 
	listTargets.insert(addBefore,pAddTarget);
      
      mapTargetsAdded.insert(MapTargetsListIterator::value_type(pAddTarget,
								pos));
    }
  
  vectCodeTraces.reserve(listTargets.size());
  
#ifndef NDEBUG
  // make sure all trace members of the function have been found
  SetCodeTraces setCodeTracesCopy(m_setCodeTraces);
#endif
  
  for (ListTargets::iterator j = listTargets.begin(); 
       j != listTargets.end (); ++j)
    {
      // make sure all targets found are members of the function
      
      assert (m_setCodeTraces.find(*j) != m_setCodeTraces.end());
      
#ifndef NDEBUG
      assert (setCodeTracesCopy.find(*j) != setCodeTracesCopy.end()); // multiply found (bad)
      setCodeTracesCopy.erase(*j);
#endif		
      
      vectCodeTraces.push_back(*j);
    }
  
#ifndef NDEBUG
  assert (setCodeTracesCopy.empty()); // otherwise, we have members that were not found
#endif
}

void CallFunction::InitCallReturnState ()
{
  DeleteCallReturnState ();
  
}

void CallFunction::DeleteCallReturnState ()
{
  
}





