/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// CallFunction.h: interface for the CCallFunction class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__CALL_FUNCTION_H__INCLUDED_)
#define __CALL_FUNCTION_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

/* A object CallFunction represent one piece of the code of the
 * executable It's defined (or bounded) by the rule that is should
 * only be accessed by calls from other parts of the code (or possibly
 * recursively), and execution should only leave by returns. That mean
 * that all jumps are to be internal within the function.  That any
 * function may have multiple return statements is easy, but the above
 * contruction implies that we may also have multiple entry points, if
 * two different call targets via jumps access the same code (A). This
 * can of course be fixed by duplication of the code (A), but is not
 * (at least not during disassembly).
 *
 * About data-analysis. When not doing the analysis of the function,
 * we don't keep all data, as this will use a lot of memory, so we
 * keep the least set necessary not too loose any information, while
 * still being able to recover full information by only having to
 * reanalyse this function.  This small set is the entry stack and
 * register state at each entry point (the common least known
 * information from every possible caller) And the same information at
 * each call instruction, is it is used by other function And the same
 * information (least common known) of all return statements in the
 * CallFuntion */

#include "CallTarget.h" 

#include <vector>
#include <set>

using namespace std;
/*
  class ExportedEntry;

struct EntryPoint
{
public:
	EntryPoint(CallTarget* pTarget,ExportedEntry* pExported)
	{
		m_pTarget = pTarget;
		m_pExported = pExported; 
	}

public:
	CallTarget*		m_pTarget;
	ExportedEntry*	m_pExported; // how we were exported (or modules entry point)
};

typedef vector<EntryPoint> VectEntryPoints; 
*/
typedef vector<CallTarget*> VectEntryPoints; 
typedef set<BranchTarget*> SetCodeTraces;
typedef vector<BranchTarget*> VectCodeTraces;

//typedef setstack<CtrlTransferTarget*> SetStackCtrlTransferTargets;

class CallFunction  
{
 public:
  CallFunction();
  virtual ~CallFunction();
  
 public:
  VectEntryPoints m_vectEntryPoints; // all used entrypoints to this function. (hopefully only one)
  SetCodeTraces m_setCodeTraces;
  
 public:
  void AddEntryPoint (CallTarget* pTarget);
  //	void AddEntryPoint (CallTarget* pTarget,ExportedEntry* pExported);
  //	void AddEntryPoint (const EntryPoint& entryPoint);
  // void RemoveEntryPoint (CallTarget* pTarget);
  void AddCodeTrace (BranchTarget* pTarget);
  void RemoveCodeTrace (BranchTarget* pTarget);
  
 public:
  void Merge (CallFunction* pInsertFunc);
  
 public:
  //	EntryPoint& GetFirstEntryPoint () { return m_vectEntryPoints.front(); }
  CallTarget* GetFirstEntryPoint () { return m_vectEntryPoints.front(); }
  
  void GetSortedCodeTraces (VectCodeTraces& vectCodeTraces);
  void GetCodeTraces (VectCodeTraces& vectCodeTraces);
  
  //  void GetCtrlTransferTargets (SetStackCtrlTransferTargets& setstackCtrlTransferTargets);
  const VectEntryPoints& GetEntryPointsTargets () { return m_vectEntryPoints; }
  
 public:
  void InitCallReturnState ();
  void DeleteCallReturnState ();
  
  /*
    public:
    void GetAllCodeTraces (VectBranchTargets& vectTargets);
  */
};

#endif // !defined(__CALL_FUNCTION_H__INCLUDED_)




