/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// CallTarget.cpp: implementation of the CallTarget class.
//
//////////////////////////////////////////////////////////////////////

#include "CallTarget.h" 
class CallFunction; 
#include "BranchTargetCtrlTransfer.h" 
#include "ExportedObject.h"

#include "ReportDisAsmInfo.h"
#include "JavaDisAsmFile.h"
#include "UrlFile.h"

#include <algorithm>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CallTarget::CallTarget()
{
  m_pFunction = NULL;
}

CallTarget::~CallTarget()
{

}

void CallTarget::AddExportedAsObject(ExportedObject* pExportedAs)
{
  if (pExportedAs->m_pTarget)
    {
      // we should already be known by us
      
      assert (find(vectExportedAs.begin(),
		   vectExportedAs.end(),
		   pExportedAs) != vectExportedAs.end());
      
      return;
    }
  
  vectExportedAs.push_back(pExportedAs);
  pExportedAs->m_pTarget = this;
}

void CallTarget::SetCallFunction (CallFunction* pFunction)
{
  assert (pFunction);
  
  m_pFunction = pFunction;
}

#include "HtmlPage.h" 
#include "HtmlTags.h" // dep_chk_unknown     

void CallTarget::HtmlPlainDisAsm (HtmlPage& page,
				  const ReportDisAsmInfo& reportDisAsmInfo) const
{
  int i = 0;
  
  FOR_EACH_CONST_ITER(VectExportedAs,vectExportedAs,exported)
    {
      page << "[export:";
      
      FOR_EACH_CONST_ITER(VectString,(*exported)->m_vectNames,name)
        page << *name;
      
      page << "]\n";
    }
  
  for (VectFromTrace::const_iterator from = m_vectFromTrace.begin(); 
       from != m_vectFromTrace.end(); ++from, ++i)
    {
      BranchTargetCtrlTransfer* pFromTrace = *from;
      
      // external target
      
      page << '[';
      
      assert (pFromTrace);
      
      CallFunction* pFromFunc = 
	reportDisAsmInfo.GetTraceFunction(pFromTrace->GetTraceHead());
      
      assert (reportDisAsmInfo.GetFunctionUrlFile(pFromFunc,DIS_ASM_URL_HTML));
      
      page << a(reportDisAsmInfo.GetLabel(pFromTrace)).href(reportDisAsmInfo.GetFunctionUrlFile(pFromFunc,DIS_ASM_URL_HTML));
      
      page << (i % 4 == 3 ? "]\n" : "] ");
    }
  if (i % 4 != 0)
    page << '\n';
}

void CallTarget::JdafDisAsm (JdafAsmCodeSnippet& snippet,
			     const ReportDisAsmInfo& reportDisAsmInfo) const
{
  for (VectFromTrace::const_iterator i = m_vectFromTrace.begin(); 
       i != m_vectFromTrace.end(); i++)
    {
      BranchTargetCtrlTransfer* pFromTrace = *i;
      
      // external target
      
      CallFunction* pFromFunc = 
	reportDisAsmInfo.GetTraceFunction(pFromTrace->GetTraceHead());
      
      assert (reportDisAsmInfo.GetFunctionUrlFile(pFromFunc,DIS_ASM_URL_HTML));
      
      JdafExternal* pExternal = 
	new JdafExternal(reportDisAsmInfo.GetFunctionUrlFile(pFromFunc,
							     DIS_ASM_URL_JAVA),
			 reportDisAsmInfo.GetLabel(pFromTrace),
			 reportDisAsmInfo.GetLabel(pFromTrace));
      
      snippet.AddExternalCaller(pExternal);
    }
}


