/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// CodeTrace.cpp: implementation of the codeTrace class.
//
//////////////////////////////////////////////////////////////////////

#include "CodeTrace.h" 
#include "gen/DisAssemblerError.h" 

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CodeTrace::CodeTrace()
{
  m_pRawCode = NULL;
  // m_dwRawCodeLength = 0; (indicated by m_pRawCode = 0)
}

CodeTrace::~CodeTrace()
{
  delete[] m_pRawCode;
}

void CodeTrace::SetRawCode (const byte* pRawCode,dword dwLength)
{
  assert (!m_pRawCode);
  
  assert (dwLength); // something to add
  
  m_pRawCode = new byte[dwLength];
  m_dwRawCodeLength = dwLength;
  
  memcpy (m_pRawCode,pRawCode,dwLength);
}


void CodeTrace::Split (dword dwOffset,CodeTrace* pNew,dword dwCodeTraceRVA)
{
  assert (m_pRawCode);
  
  // split after dwOffset bytes
  
  // make big fuss if dwOffset isn't exactly at start of instruction
  
  assert (dwOffset > 0); // it's not a split to chop of no instructions!
  assert (dwOffset < m_dwRawCodeLength); // cannot split after end
  
  dword dwKeep = 0;
  unsigned int nKeep = 0;
  
  do
    {
      assert (m_vectRawInstrLengths.size () > nKeep); // or we're trying to split after end
      
      dwKeep += m_vectRawInstrLengths[nKeep];
      nKeep++;
      
      if (dwKeep > dwOffset)
	{
	  // oups, we're trying to split in middle of function
	  
	  throw ERROR_DISASM_CRAZY_JUMP_CANT_SPLIT(dwKeep+dwCodeTraceRVA,dwOffset+dwCodeTraceRVA);
	  
	  //			throw "Can't split disassembled codetrace in middle of instruction. (Crazy jump)";
	}
    }
  while (dwKeep < dwOffset);
  
  assert (dwKeep == dwOffset);
  
  pNew->m_dwRawCodeLength = m_dwRawCodeLength-dwKeep;
  pNew->m_pRawCode = new byte[pNew->m_dwRawCodeLength];
  memcpy (pNew->m_pRawCode,m_pRawCode+dwKeep,pNew->m_dwRawCodeLength);
  
  for (unsigned int i = nKeep; i < m_vectRawInstrLengths.size (); i++)
    {
      pNew->m_vectRawInstrLengths.push_back(m_vectRawInstrLengths[i]);
      pNew->m_vectInstructions.push_back(m_vectInstructions[i]);
    }
  
  m_vectRawInstrLengths.resize(nKeep);
  m_vectInstructions.resize(nKeep);
	
  m_dwRawCodeLength = dwKeep;
  
  byte* pOldRawCode = m_pRawCode;
  m_pRawCode = new byte[m_dwRawCodeLength];
  memcpy (m_pRawCode,pOldRawCode,m_dwRawCodeLength);
  delete[] pOldRawCode;
}

/*
	byte* m_pRawCode;

	vector<byte> m_vectRawInstrLengths;
	vector<Instruction*> m_vectInstructions;
*/






