/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// DataDestPtr.cpp: implementation of the DataDestPtr class.
//
//////////////////////////////////////////////////////////////////////

#include "DataDestPtr.h" // dep_chk_needed   
#include "Misc.h" // dep_chk_unknown   

#include <assert.h>
#include <memory.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

DataDestPtr::DataDestPtr()
{	
  m_pStart = NULL;
  m_pCurrent = NULL;
  IFDEBUG(m_pEnd = NULL);
}

DataDestPtr::DataDestPtr(void* pStart IFDEBUGPARAM(void* pEnd) )
{
  m_pStart = (byte*) pStart;
  m_pCurrent = (byte*)pStart;
  IFDEBUG(m_pEnd = (byte*) pEnd);
}

DataDestPtr::~DataDestPtr()
{

}
	
void DataDestPtr::WriteByte (byte bData)
{
  assert (m_pStart <= m_pEnd);
  assert (m_pCurrent >= m_pStart && m_pCurrent <= m_pEnd);
  
  assert (m_pCurrent < m_pEnd); // otherwise too small array allocated (go fix code otherplace)
  
  *((byte*) m_pCurrent) = bData;
  
  m_pCurrent = (m_pCurrent) + 1;
}

void DataDestPtr::WriteWord (word wData)
{
  assert (m_pStart <= m_pEnd);
  assert (m_pCurrent >= m_pStart && m_pCurrent <= m_pEnd);
  
  assert ((byte*) (((word*) m_pCurrent) + 1) <= m_pEnd); // otherwise too small array allocated (go fix code otherplace)
  
  *((word*) m_pCurrent) = wData;
  
  m_pCurrent = (byte*) (((word*) m_pCurrent) + 1);
}

void DataDestPtr::WriteDWord (dword dwData)
{
  assert (m_pStart <= m_pEnd);
  assert (m_pCurrent >= m_pStart && m_pCurrent <= m_pEnd);
  
  assert ((byte*) (((dword*) m_pCurrent) + 1) <= m_pEnd); // otherwise too small array allocated (go fix code otherplace)
  
  *((dword*) m_pCurrent) = dwData;
  
  m_pCurrent = (byte*) (((dword*) m_pCurrent) + 1);
}

void DataDestPtr::WriteDWordOffset (dword dwOffset,dword dwData)
{
  assert (m_pStart <= m_pEnd);
  assert (m_pStart + dwOffset >= m_pStart);
  
  assert ((byte*) (((dword*) (m_pStart+dwOffset)) + 1) <= m_pEnd); // otherwise too small array allocated (go fix code otherplace)
  
  *((dword*) (m_pStart+dwOffset)) = dwData;
}

void DataDestPtr::WriteData (const void* pData,uint nLength)
{
  assert (m_pStart <= m_pEnd);
  assert (m_pCurrent >= m_pStart && m_pCurrent <= m_pEnd);
  
  assert (m_pCurrent + nLength <= m_pEnd); // otherwise too small array allocated (go fix code otherplace)
  
  memcpy (m_pCurrent,pData,nLength);
  
  m_pCurrent += nLength;
}

void* DataDestPtr::WritePtr(uint nLength)
{
  assert (m_pStart <= m_pEnd);
  assert (m_pCurrent >= m_pStart && m_pCurrent <= m_pEnd);
  
  assert (m_pCurrent + nLength <= m_pEnd); // otherwise too small array allocated (go fix code otherplace)
  
  void* pPtr = m_pCurrent;
  
  m_pCurrent += nLength;
  
  return pPtr;
}

void DataDestPtr::WriteZero (uint nLength)
{
  assert (m_pStart <= m_pEnd);
  assert (m_pCurrent >= m_pStart && m_pCurrent <= m_pEnd);
  
  assert (m_pCurrent + nLength <= m_pEnd); // otherwise too small array allocated (go fix code otherplace)
  
  memset (m_pCurrent,0,nLength);
  
  m_pCurrent += nLength;
}

void DataDestPtr::WriteString(const string& str)
{
  WriteData(str.c_str(),str.size()+1);
}

void DataDestPtr::WriteString(const string& str,uint nAlignment)
{
  // make total output bytes a multiple of nAlignment
  
  uint n = str.size()+1;
  
  WriteData(str.c_str(),n);
  
  assert (powerof2(nAlignment)); // so we can use  & (nAlignment-1) , instead of % nAlignment
  
  WriteZero((4 - n) & (nAlignment-1));
}






