/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// DataDestRelocPtr.cpp: implementation of the DataDestRelocPtr class.
//
//////////////////////////////////////////////////////////////////////

#include "DataDestRelocPtr.h"

#include "RelocatableDest.h" 

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


DataDestRelocPtr::DataDestRelocPtr(RelocatableDest* pDest)
  : DataDestPtr (pDest->m_pData IFDEBUGPARAM(pDest->m_pData+pDest->m_dwSize) )
{
  assert (pDest->m_pData || !pDest->m_dwSize);
  
  m_pReloc = &(pDest->m_relocInfo);
  dwRVAOffset = 0;
}

DataDestRelocPtr::DataDestRelocPtr(void* pStart IFDEBUGPARAM(void* pEnd) ,
				   RelocationInfo* pReloc)
  : DataDestPtr (pStart IFDEBUGPARAM(pEnd) )
{
  m_pReloc = pReloc;
  dwRVAOffset = 0;
}

DataDestRelocPtr::DataDestRelocPtr(DataDestRelocPtr* pBase,
				   dword dwOffset IFDEBUGPARAM(dword dwSize) )
  : DataDestPtr (pBase->m_pStart+dwOffset 
		 IFDEBUGPARAM(pBase->m_pStart+dwOffset+dwSize) )
{
  assert (pBase->m_pStart + dwOffset        <= pBase->m_pEnd);
  assert (pBase->m_pStart + dwOffset+dwSize <= pBase->m_pEnd);
  
  m_pReloc = pBase->m_pReloc;
  dwRVAOffset = dwOffset;
}

DataDestRelocPtr::~DataDestRelocPtr()
{
  
}

void DataDestRelocPtr::InsertRelocation(dword dwOffset,word wRelocate)
{
  assert (wRelocate);
  
  assert (dwOffset < (dword) (m_pEnd - m_pStart));
  assert (dwOffset + sizeof (dword) < (dword) (m_pEnd - m_pStart));
  
  //  printf ("add reloc: %d %d\n",dwOffset,wRelocate);

  m_pReloc->InsertRelocation(dwOffset + dwRVAOffset,wRelocate);
}

void DataDestRelocPtr::WriteDWord (dword dwData,word wRelocate)
{
  //  printf ("add reloc: %d %d\n",GetOffset() + dwRVAOffset,wRelocate);
  
  if (wRelocate)
    m_pReloc->InsertRelocation(GetOffset() + dwRVAOffset,wRelocate);
  
  DataDestPtr::WriteDWord (dwData);
}


