/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// DataPtrSrc.h: interface for the DataPtrSrc class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__DATA_PTR_SRC_H__INCLUDED_)
#define __DATA_PTR_SRC_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "DataPtr.h" 

class DataPtrSrc : public DataPtr
{
 public:
  DataPtrSrc(const byte* pStart,const byte* pEnd);
  DataPtrSrc(const byte* pStart,dword dwLength,dword dwCurOffset = 0);
  DataPtrSrc();
  virtual ~DataPtrSrc();
  
 public:
  void SetSrc(const byte* pStart,const byte* pEnd,dword dwCurOffset = 0);

 public:
  const byte* m_pStart;
  const byte* m_pEnd;
  
  const byte* m_pCurrent;
  
 public:
  virtual const byte* GetPtr (uint nBytes);
  virtual const byte* GetPtr (dword dwVirtualAddress,uint nBytes);
  
  virtual bool IsValidAddress (dword dwVirtualAddress);
  
  dword GetOffset();
  virtual dword GetRVA () { return GetOffset(); }
  
  const byte* GetCurrent () const { return m_pCurrent; }
  
  void SetCurrent (const byte* pCur);
  void SetOffset(dword dwOffset);
  virtual void SetRVA(dword dwRVA) { SetOffset(dwRVA); }
  void ResetCurrent () { m_pCurrent = m_pStart; }
  
  const char* GetStringPtr(uint* nLength);
  const char* GetStringPtr(dword dwVirtualAddress,uint* nLength);
  const char* GetStringPtr(dword dwVirtualAddress);
  
 protected:
  uint GetStringLength (const byte* pCur);
  
  /*
    public:
	const char* GetString ();
	const char* GetString (dword dwVirtualAddress);
	
	DataPtrPESectionBody GetRVADataPtr (dword dwVirtualAddress);
	*/
};

//////////////////////////////////////////////////////////////////////

class DataPtrSrcRVA : public DataPtrSrc
{
public:
  DataPtrSrcRVA(const byte* pStart,dword dwRVAOffset,dword dwLength,
		dword dwRVACur);
  
 public:
  dword m_dwRVAOffset;

 public:
  virtual const byte* GetPtr (dword dwVirtualAddress,uint nBytes);
  virtual bool IsValidAddress (dword dwVirtualAddress);
  
  virtual dword GetRVA ();
  virtual void SetRVA(dword dwRVA);
  
};

#endif // !defined(__DATA_PTR_SRC_H__INCLUDED_)
