/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// DestRelocatableSections.cpp: implementation of the DestRelocatableSections class.
//
//////////////////////////////////////////////////////////////////////

#include "DestRelocatableSections.h"
#include "RelocatableDest.h"


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

DestRelocatableSections::DestRelocatableSections()
{

}

DestRelocatableSections::~DestRelocatableSections()
{

  
  FOR_EACH_ITER(MapSectionRelocatableDest,mapSectionRelocatableDest,kill)
    {
      delete kill->second;
    }
}

bool DestRelocatableSections::HasSection(word wRelocate)
{
  return mapSectionRelocatableDest.find(wRelocate) != 
    mapSectionRelocatableDest.end();
}

word DestRelocatableSections::GetSectionFirstInRange (word wRelocMin,
						      word wRelocMax)
{
  MapSectionRelocatableDest::iterator min = 
    mapSectionRelocatableDest.lower_bound(wRelocMin);
  
  if (min != mapSectionRelocatableDest.end() &&
      min->first <= wRelocMax)
    return min->first;
  return 0;
}

dword DestRelocatableSections::AlignSectionFileLength (word wRelocate,
						       dword dwAlign)
{
  assert (HasSection(wRelocate));
  
  MapSectionRelocatableDest::iterator sect = 
    mapSectionRelocatableDest.find(wRelocate);
  
  return sect->second->AlignLength(dwAlign);
}

dword DestRelocatableSections::GetSectionFileLength(word wRelocate)
{
  assert (HasSection(wRelocate));
  MapSectionRelocatableDest::iterator sect = 
    mapSectionRelocatableDest.find(wRelocate);
  return sect->second->m_dwSize;
}

dword DestRelocatableSections::GetSectionFileEnd(word wRelocate)
{
  assert (HasSection(wRelocate));
  MapSectionRelocatableDest::iterator sect = 
    mapSectionRelocatableDest.find(wRelocate);
  return sect->second->m_relocInfo.m_dwFileOffset + sect->second->m_dwSize;
}

dword DestRelocatableSections::GetSectionImageLength(word wRelocate)
{
  assert (HasSection(wRelocate));
  MapSectionRelocatableDest::iterator sect = 
    mapSectionRelocatableDest.find(wRelocate);
  assert (sect->second->m_dwSize /* + plus any uninitialized length */ >= 
	  sect->second->m_dwSize);
  return sect->second->m_dwSize; // plus any uninitialized length
}

dword DestRelocatableSections::GetSectionImageEnd(word wRelocate)
{
  assert (HasSection(wRelocate));
  MapSectionRelocatableDest::iterator sect = 
    mapSectionRelocatableDest.find(wRelocate);
  assert (sect->second->m_dwSize /* + plus any uninitialized length */ >= 
	  sect->second->m_dwSize);
  return sect->second->m_relocInfo.m_dwImageOffset + sect->second->m_dwSize; // plus any uninitialized length
}

void DestRelocatableSections::AddRelocatableDest(word wRelocate,
						 RelocatableDest* pDest)
{
  if (!pDest)
    return;
  
  assert (!(wRelocate & RELATIVE_TYPE_MASK) && 
	  (wRelocate & RELATIVE_SECTION_MASK));
  
  assert (!HasSection(wRelocate));
  
  mapSectionRelocatableDest.insert(MapSectionRelocatableDest::value_type(wRelocate,pDest));
}

const RelocatableDest* DestRelocatableSections::GetRelocatableDest(word wRelocate)
{
  MapSectionRelocatableDest::iterator sect = 
    mapSectionRelocatableDest.find(wRelocate);
  
  if (sect == mapSectionRelocatableDest.end())
    return NULL;
  
  return sect->second;
}

#ifndef NDEBUG
/*
void VerifyMapDoRelocationIntegrity(MapDoRelocation& mapDoRelocation)
{
	printf ("\n");

	int i = 0;
	MapDoRelocation::iterator iter;
	for (iter = mapDoRelocation.begin();
			iter != mapDoRelocation.end(); ++iter, i++)
	{
		printf ("%2d: %04x | %08x %04x %04x\n",i,iter->first,
			iter->second.m_dwMove,
			iter->second.m_wAppendedTo,
			iter->second.m_wChangeRelocation);
	}

	for (iter = mapDoRelocation.begin();
			iter != mapDoRelocation.end(); ++iter)
	{
		word wLoc = iter->first;

		MapDoRelocation::iterator loc = mapDoRelocation.find(wLoc);

		assert (loc != mapDoRelocation.end());
		assert (loc == iter);
	}
}

#define VERIFY_MapDoRelocation_INTEGRITY VerifyMapDoRelocationIntegrity(mapDoRelocation);

void DestRelocatableSections::VerifyIntegrity()
{
	VERIFY_MapDoRelocation_INTEGRITY;
}
*/
/*
void DestRelocatableSections::VerifyIntegrity(uint no)
{
 	printf ("<%d>\n",no);
        FOR_EACH_ITER(MapSectionRelocatableDest,mapSectionRelocatableDest,reloc)
        {

                printf ("> %08x : %04x  (%08x : %d)\n",
                        reloc->second->m_relocInfo.m_dwFileOffset,reloc->first,
                        reloc->second->m_pData,reloc->second->m_relocInfo.m_mapOffsetRelocInfo.size());
        }
}
*/
#define VERIFY_MapDoRelocation_INTEGRITY 
#else
#define VERIFY_MapDoRelocation_INTEGRITY 
#endif

void DestRelocatableSections::AppendToEnd (word wRelocateAppendTo,
					   word wRelocateSrc,dword dwAlign)
{
  // if there is no section like this, then simply rename the existing one
  
  assert (!(wRelocateAppendTo & RELATIVE_TYPE_MASK) && 
	  (wRelocateAppendTo & RELATIVE_SECTION_MASK));
  assert (!(wRelocateAppendTo & RELATIVE_TYPE_MASK) &&
	  (wRelocateAppendTo & RELATIVE_SECTION_MASK));
  
  MapSectionRelocatableDest::iterator src =
    mapSectionRelocatableDest.find(wRelocateSrc);
  
  if (src == mapSectionRelocatableDest.end())
    return; // nothing to append
  
  MapSectionRelocatableDest::iterator appendTo =
    mapSectionRelocatableDest.find(wRelocateAppendTo);
  
  dword dwMoveFile = 0;
  dword dwMoveImage = 0;
  
  if (appendTo == mapSectionRelocatableDest.end())
    {
      mapSectionRelocatableDest.insert(MapSectionRelocatableDest::value_type(wRelocateAppendTo,src->second));
      
      mapSectionRelocatableDest.erase (src);
      
      // redirect any relocations, but there is no move required
    }
  else
    {
      // so we have to append
      
      appendTo->second->AppendToEnd (wRelocateAppendTo,src->second,
				     dwMoveFile,dwMoveImage,
				     dwAlign);
      
      delete src->second;
      mapSectionRelocatableDest.erase (src);
    }
  
  // redirect any relocations
  
  VERIFY_MapDoRelocation_INTEGRITY;
  AddDoRelocate (wRelocateSrc | RELATIVE_FILE ,dwMoveFile ,
		 wRelocateAppendTo | RELATIVE_FILE);
  VERIFY_MapDoRelocation_INTEGRITY;
  AddDoRelocate (wRelocateSrc | RELATIVE_IMAGE,dwMoveImage,
		 wRelocateAppendTo | RELATIVE_IMAGE);
  VERIFY_MapDoRelocation_INTEGRITY;
}

void DestRelocatableSections::AddDoRelocate (word wRelocate,dword dwMove,word wAppendedTo)
{
  //	typedef map<word,DoRelocation> MapDoRelocation;
  
  //	printf ("AddDoRelocate: %x %x %x\n",wRelocate,dwMove,wAppendedTo);
  
  assert ((wRelocate & RELATIVE_TYPE_MASK) &&
	  (wRelocate & RELATIVE_SECTION_MASK));
  
  VERIFY_MapDoRelocation_INTEGRITY;
  MapDoRelocation::iterator reloc = mapDoRelocation.find(wRelocate);
  VERIFY_MapDoRelocation_INTEGRITY;
  
  if (reloc != mapDoRelocation.end())
    {
      VERIFY_MapDoRelocation_INTEGRITY;
      assert (!reloc->second.m_wAppendedTo); // if it has been appended, then it's fixed relative to
      // the section it was appended to and cannot be moved further (but you may relocate the
      // appended_to section
      
      VERIFY_MapDoRelocation_INTEGRITY;
      reloc->second.m_dwMove += dwMove;
      VERIFY_MapDoRelocation_INTEGRITY;
      reloc->second.m_wAppendedTo = wAppendedTo;
      VERIFY_MapDoRelocation_INTEGRITY;
    }
  else
    mapDoRelocation.insert(MapDoRelocation::value_type(wRelocate,
						       DoRelocation(dwMove,
								    wAppendedTo)));
  VERIFY_MapDoRelocation_INTEGRITY;
}


void DestRelocatableSections::RelocateSection(word wRelocate,
					      dword dwNewSectionBase)
{
  // after this, DoRelocations will peform the actual relocations (of
  // data pointing at us).  (so many can be done at the same time),
  // this will just prepare for the relocations
  
  // find the section, and change it's base offset make a note on how
  // to relocate anything pointing into that section
  
  VERIFY_MapDoRelocation_INTEGRITY;
  
  assert ((wRelocate & RELATIVE_TYPE_MASK) &&
	  (wRelocate & RELATIVE_SECTION_MASK));
  
  MapSectionRelocatableDest::iterator reloc = 
    mapSectionRelocatableDest.find(wRelocate & RELATIVE_SECTION_MASK);
  
  assert (reloc != mapSectionRelocatableDest.end());
  
  dword dwMove = reloc->second->Relocate(wRelocate,dwNewSectionBase);
  
  VERIFY_MapDoRelocation_INTEGRITY;
  AddDoRelocate (wRelocate,dwMove);
  VERIFY_MapDoRelocation_INTEGRITY;
  
  

//	typedef map<word,DoRelocation> MapDoRelocation;

/*
    MapSectionRelocatableDest::iterator reloc_self = mapSectionRelocatableDest.find(wRelocate);

  	assert (reloc_self != mapSectionRelocatableDest.end());

        dword dwMove = reloc_self->second->RelocateSelf(wRelocate,dwNewSectionBase);

 	FOR_EACH_ITER(MapSectionRelocatableDest,mapSectionRelocatableDest,reloc)
        {
        reloc->second->Relocate(wRelocate,dwMove);
        }
*/
}

void DestRelocatableSections::DoRelocations ()
{
  // this will perform the actual relocations.
  
  // but first simplify the relocations (i.e., if there are any that
  // has been appended, then you have to follow like a linked list for
  // each relocation. It's more efficient to do that here (note: after
  // this simplification, you can no longer move anything as you would
  // then have to figure out which ones depend on the one you are
  // about to move
  
  FOR_EACH_ITER(MapDoRelocation,mapDoRelocation,iter)
    {
      assert (iter->second.m_wChangeRelocation == 0);

      MapDoRelocation::iterator follow = iter;
      
      while (follow->second.m_wAppendedTo)
        {
          iter->second.m_wChangeRelocation = follow->second.m_wAppendedTo;
          
          if ((follow = mapDoRelocation.find(follow->second.m_wAppendedTo)) ==
	      mapDoRelocation.end())
            break;
          
          iter->second.m_dwMove += follow->second.m_dwMove;
        }
      iter->second.m_wAppendedTo = 0; // as this has been adjusted (and we do not want to adjust stuff twice, should anyone later in the map depend on us)
    }
  
  FOR_EACH_ITER(MapSectionRelocatableDest,mapSectionRelocatableDest,reloc)
    {
      // printf ("reloc: %d\n",reloc->first);
      
      reloc->second->DoRelocate(reloc->first,mapDoRelocation);
    }
  mapDoRelocation.clear();
}


void DestRelocatableSections::GetSectionInFileOrder(MapOffsetRelocatableDest& mapOffsetRelocatableDest)
{
  FOR_EACH_ITER(MapSectionRelocatableDest,mapSectionRelocatableDest,reloc)
    {
      // printf ("%08x (%04x): %x \n",reloc->second->m_relocInfo.m_dwFileOffset,reloc->second->m_dwSize,reloc->first);
      
      assert (!reloc->second->m_dwSize ||
	      mapOffsetRelocatableDest.find(reloc->second->m_relocInfo.m_dwFileOffset) == mapOffsetRelocatableDest.end());
      
      if (reloc->second->m_dwSize) 
        {        
	  mapOffsetRelocatableDest.insert(MapOffsetRelocatableDest::value_type(reloc->second->m_relocInfo.m_dwFileOffset,
									       reloc->second));
        }
    }
}




