/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// DisAssembler.cpp: implementation of the DisAssembler class.
//
//////////////////////////////////////////////////////////////////////

#include "DisAssembler.h" // dep_chk_needed   
#include "InstructionDecoder.h" // dep_chk_needed   
#include "BranchInstr.h" // dep_chk_needed   
#include "BranchTargetTrace.h" // dep_chk_needed
#include "BranchTargetCtrlTransfer.h"
#include "gen/DisAssemblerError.h" // dep_chk_unknown   

#include "Executable.h"
#include "JumpTarget.h"
#include "CallTarget.h"
#include "CallFunction.h"
#include "ExportedObject.h"
#include "DataPtr.h"

#include <stdio.h>

#include <algorithm>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

DisAssembler::DisAssembler()
{
	m_pExecutable = NULL;
}

DisAssembler::~DisAssembler()
{

}

CtrlTransferTarget* DisAssembler::AddJumpTarget (dword dwTargetVA,
						 CallFunction* pFunction)
{
  // printf ("AddJumpTarget(%08x,%08x)\n",dwTargetVA,pFunction);
  
  // so dwTargetVA is a jump target
  // make sure it has an JumpTarget structure to keep all incoming
  
  MapVAJumpTarget::iterator i =
    m_mapVAJumpTarget.find(dwTargetVA);
  
  if (i != m_mapVAJumpTarget.end())
    {
      JumpTarget* pJumpTarget = i->second;
      assert (pJumpTarget);
      
      if (pJumpTarget->m_pTarget)
	SetTraceFunc (pJumpTarget->m_pTarget->GetTraceHead(),pFunction);
      else
	{
	  MapUnsetJumpTargetCallFunction::iterator func = m_mapUnsetJumpTargetCallFunction.find(pJumpTarget);
	  
	  assert (func != m_mapUnsetJumpTargetCallFunction.end());
	  
	  if (func->second != pFunction)
	    {
	      MergeTraceFunc(pFunction,func->second);
	      assert (func->second == pFunction); // should have been moved
	    }
	}
      return pJumpTarget;
    }
  
  BranchTarget* pBranchTarget = AddBranchTarget(dwTargetVA);
  
  // exception will be thrown in case of this is an attempt at a crazy
  // jump and in that case, we'll don't do what's below, so that's ok
  
  // JumpTarget not known, add it
  
  JumpTarget* pJumpTarget = new JumpTarget();
  
  m_mapVAJumpTarget.insert(MapVAJumpTarget::value_type(dwTargetVA,
						       pJumpTarget));
  
  if (pBranchTarget)
    {
      pBranchTarget->SetJumpSource(pJumpTarget);
      SetTraceFunc (pBranchTarget->GetTraceHead(),pFunction);
    }
  else
    {
      assert (m_mapUnsetJumpTargetCallFunction.find(pJumpTarget) == 
	      m_mapUnsetJumpTargetCallFunction.end());
      m_mapUnsetJumpTargetCallFunction.insert(MapUnsetJumpTargetCallFunction::value_type(pJumpTarget,pFunction));
    }
  return pJumpTarget;
}

CtrlTransferTarget* DisAssembler::AddCallTarget (ExportedObject* pExportedObj)
{
  return AddCallTarget(pExportedObj->m_dwEntryRVA,pExportedObj);
}

CtrlTransferTarget* DisAssembler::AddCallTarget (dword dwTargetVA,
						 ExportedObject* pExportedObj)
{
  // printf ("DisAssembler::AddCallTarget(%08x)\n",dwTargetVA);
  
  // so dwTargetVA is a jump target
  // make sure it has an JumpTarget structure to keep all incoming
  
  MapVACallTarget::iterator i =
    m_mapVACallTarget.find(dwTargetVA);
  
  if (i != m_mapVACallTarget.end())
    {
      CallTarget* pCallTarget = i->second;
      assert (pCallTarget);
      if (pExportedObj)
        pCallTarget->AddExportedAsObject(pExportedObj);
      return pCallTarget;
    }
  
  BranchTarget* pBranchTarget = AddBranchTarget(dwTargetVA);
  
  // exception will be thrown in case of this is an attempt at a crazy jump
  // and in that case, we'll don't do what's below, so that's ok
  
  // CallTarget not known, add it
  
  CallTarget* pCallTarget = new CallTarget();
  
  if (pExportedObj)
    pCallTarget->AddExportedAsObject(pExportedObj);
  
  m_mapVACallTarget.insert(MapVACallTarget::value_type(dwTargetVA,
						       pCallTarget));
  
  if (pBranchTarget)
    {
      JumpTarget* pJumpTarget = pBranchTarget->GetJumpSource();
      if (pJumpTarget)
        {
          MapUnsetJumpTargetCallFunction::iterator func = 
	    m_mapUnsetJumpTargetCallFunction.find(pJumpTarget);
          
          if (func != m_mapUnsetJumpTargetCallFunction.end())
            {
              MapCodeTraceCallFunction::iterator iter = 
		m_mapCodeTraceCallFunction.find(pBranchTarget->GetTraceHead());
              
              assert (func->second == iter->second);
              //		pTarget->SetJumpSource(jump->second);
              //
              //		//printf ("func: %x (%d),(%d)\n",func->second,func->second->m_vectEntryPoints.size(),func->second->m_setCodeTraces.size());
              //
              //		SetTraceFunc (pTarget->GetTraceHead(),func->second);
              
              m_mapUnsetJumpTargetCallFunction.erase(func);
            }
                
        }
      pBranchTarget->SetCallSource(pCallTarget);
      AddCallTarget(pCallTarget);
    }
  
  return pCallTarget;
}

void DisAssembler::AddCallTarget (CallTarget* pCallTarget)
{
  // printf ("DisAssembler::AddCallTarget(%08x)\n",pCallTarget);
  
  assert (pCallTarget);
  assert (pCallTarget->m_pTarget);
  
  BranchTarget* pTrace = pCallTarget->m_pTarget->GetTraceHead();
  
  MapCodeTraceCallFunction::iterator iter = 
    m_mapCodeTraceCallFunction.find(pTrace);
  
  CallFunction* pFunction;
  
  if (iter != m_mapCodeTraceCallFunction.end())
    pFunction = iter->second;
  else
    {
      pFunction = new CallFunction ();
      pFunction->AddCodeTrace(pTrace);
      m_mapCodeTraceCallFunction.insert(MapCodeTraceCallFunction::value_type(pTrace,pFunction));
    }
  
  pFunction->AddEntryPoint(pCallTarget/*,NULL*/);
  
  assert (m_pExecutable);
  m_pExecutable->AddCallFunction(pFunction);
}

void DisAssembler::MergeTraceFunc (CallFunction* pDestFunc,
				   CallFunction* pSrcFunc)
{
  // printf ("MergeTraceFunc(%08x,%08x)\n",pDestFunc,pSrcFunc);
  
  // printf ("DisAssembler::MergeTraceFunc(%08x,%08x)\n",pDestFunc,pSrcFunc);
  
  assert (pDestFunc != pSrcFunc);
  
  // the two have to be merged
  
  // all traces in pFunctionAfter are to be transferred as traces of pFunction
  
  FOR_EACH_ITER(SetCodeTraces,pSrcFunc->m_setCodeTraces,i)
    {
      MapCodeTraceCallFunction::iterator j = m_mapCodeTraceCallFunction.find(*i);
      assert (j != m_mapCodeTraceCallFunction.end());
      assert (j->second == pSrcFunc);
      
      j->second = pDestFunc;
    }
  
  pDestFunc->Merge(pSrcFunc);
  m_pExecutable->RemoveCallFunction(pSrcFunc);
  
  // m_mapUnsetJumpTargetCallFunction
  
  FOR_EACH_ITER(MapUnsetJumpTargetCallFunction,
		m_mapUnsetJumpTargetCallFunction,j)
    {
      if (j->second == pSrcFunc)
	j->second = pDestFunc;
    }
  
  // printf ("Deleted: %x\n",pSrcFunc);
  
  delete pSrcFunc;
}

void DisAssembler::SetTraceFunc (BranchTarget* pTrace,CallFunction* pFunction)
{
  // printf ("SetTraceFunc(%08x,%08x)\n",pTrace,pFunction);
  
  // printf ("DisAssembler::SetTraceFunc(%08x,%08x)\n",pTrace,pFunction);
  
  MapCodeTraceCallFunction::iterator iter = 
    m_mapCodeTraceCallFunction.find(pTrace);
  
  if (iter != m_mapCodeTraceCallFunction.end())
    {
      // printf ("prev known\n");
      
      CallFunction* pFunctionOld = iter->second;
      
      if (pFunctionOld != pFunction)
	MergeTraceFunc (pFunction,pFunctionOld);
      
      assert (iter->second == pFunction); // make sure we were moved
    }	
  else
    {
      // printf ("prev unknown\n");
      
      pFunction->AddCodeTrace(pTrace);
      m_mapCodeTraceCallFunction.insert(MapCodeTraceCallFunction::value_type(pTrace,pFunction));
    }	
}

void DisAssembler::BindTrace (BranchTarget* pBefore,BranchTarget* pAfter)
{
  assert (pBefore);
  assert (pAfter);
  
  // we need to fixup CallFunction membership
  
  pBefore->SetNextTarget (pAfter);
  
  BranchTarget* pTrace = pBefore->GetTraceHead();
  
  // the trace head of pBefore must have an target
  
  MapCodeTraceCallFunction::iterator iter = 
    m_mapCodeTraceCallFunction.find(pTrace);
  
  assert (iter != m_mapCodeTraceCallFunction.end());
  
  CallFunction* pFunction = iter->second;
  
  // if pAfter has an function
  // first remove it as an trace from the function
  
  // if it's different than function of trace head, the two must be merged
  
  iter = m_mapCodeTraceCallFunction.find(pAfter);
  
  if (iter != m_mapCodeTraceCallFunction.end())
    {
      CallFunction* pFunctionAfter = iter->second;
      
      pFunctionAfter->RemoveCodeTrace (pAfter);
      m_mapCodeTraceCallFunction.erase(iter);
      
      if (pFunction != pFunctionAfter)
	MergeTraceFunc (pFunction,pFunctionAfter);
    }
}



/*
CtrlTransferTarget* DisAssembler::AddCallTarget (dword dwTargetVA,bool bAddIfNotKnown)
{
	if (m_setCallTargetVA.find (dwTargetVA) != m_setCallTargetVA.end ())
	{
		assert (m_mapDisAsmCodePart.find(dwTargetVA) != m_mapDisAsmCodePart.end());
		return m_mapDisAsmCodePart.find(dwTargetVA)->second.m_pTarget;
	}

	m_setCallTargetVA.insert(dwTargetVA);

	// get the BranchTarget, add it to the executable

	BranchTarget* pTarget = AddBranchTarget (dwTargetVA,bAddIfNotKnown); // will be added if not known

	if (pTarget)
		m_pExecutable->AddCallTarget(dwTargetVA,pTarget);

	return pTarget;
}
*/

//CtrlTransferTarget* DisAssembler::AddJumpTarget (dword dwTargetVA,bool bAddIfNotKnown)
//{
	/*
	if (m_setCallTargetVA.find (dwTargetVA) != m_setCallTargetVA.end ())
	{
		assert (m_mapDisAsmCodePart.find(dwTargetVA) != m_mapDisAsmCodePart.end());
		return m_mapDisAsmCodePart.find(dwTargetVA)->second.m_pTarget;
	}

	m_setCallTargetVA.insert(dwTargetVA);

	// get the BranchTarget, add it to the executable

	BranchTarget* pTarget = AddBranchTarget (dwTargetVA,bAddIfNotKnown); // will be added if not known

	if (pTarget)
		m_pExecutable->AddCallTarget(dwTargetVA,pTarget);

	return pTarget;
	*/
//}

// the disassembler is in charge of decoding assembler instruction traces
// from the binary, and put and cut them according to any jump targets
// but is not to figure out what the code itself do

BranchTarget* DisAssembler::AddBranchTarget (dword dwTargetVA,
					     bool bAddIfNotKnown)
{
  // if we're lucky, it's outside anything yet disassebled,
  // but it could also be inside something, in which case we have to
  // split that something... Arrgh
  
  if (!m_mapDisAsmCodePart.empty ())
    {
      MapDisAsmCodePart::iterator prevTarget = 
	m_mapDisAsmCodePart.lower_bound(dwTargetVA);
      
      // see if we hit ourselves?
      
      if (prevTarget != m_mapDisAsmCodePart.end () &&
	  prevTarget->first == dwTargetVA)
	{
	  // we hit the sameplace as somebody else has hit (yahoo)
	  // nothing to do
	  
	  return prevTarget->second.m_pTarget;
	}
      
      assert (m_mapDisAsmCodePart.find(dwTargetVA) == 
	      m_mapDisAsmCodePart.end()); // if we were in the list, we should have been found above
      
      // ok, so either we found something inside, in which case the
      // found one was > than us or we got .end(), which is also
      // larger than us
      
      // now we want to find out if we hit inside anything
      
      // if we got .begin(), there cannot be anything smaller than us

      if (prevTarget != m_mapDisAsmCodePart.begin ()) // we got the first one, so there can be no smaller
	{	
	  assert (prevTarget == m_mapDisAsmCodePart.end() ||
		  prevTarget->first > dwTargetVA);
	  
	  prevTarget--; // we got the target after the wanted one, back up
	  
	  // dword dwa = prevTarget->first;
	  // dword dwb = dwTargetVA;

	  assert (prevTarget->first < dwTargetVA);
	  
	  if (prevTarget->second.m_dwEndVA > dwTargetVA)
	    {
	      // aouch, we hit somewhere inside an already known
	      // CodeTrace it must be split and let's hope we did not
	      // hit inside an instruction (which can't be split, the
	      // only interpretation being that the actual bytes can
	      // be interpreted in two different ways depending on
	      // where we start, and this is NOT supported by this
	      // prgm)

	      if (!bAddIfNotKnown)
		return NULL;
	      return SplitCodeTrace (prevTarget->first,dwTargetVA);
	    }
	}
    }
  
  if (!bAddIfNotKnown)
    return NULL;
  
  // so we hit somewhere that has not been handled before
  
  m_stackDisAsmVA.push (dwTargetVA);
  
  //	return AddBranchTarget (dwTargetVA,dwTargetVA);
  
  assert (m_mapDisAsmCodePart.find(dwTargetVA) == m_mapDisAsmCodePart.end());
  
  //	BranchTarget* pTarget = new BranchTarget(dwTargetVA);
  
  //	printf ("-%08x(%08x) [%08x]\n",dwTargetVA,dwTargetVA,pTarget);

  DisAsmCodePart newTarget = { dwTargetVA,NULL };
  
  m_mapDisAsmCodePart.insert(MapDisAsmCodePart::value_type(dwTargetVA,newTarget));
  
  //printf ("%08x->%08x (add)\n",dwVA,dwVA);
  
  //	if (m_setDisAsmVA.find(dwVA) != m_setDisAsmVA.end())
  //	{
  //		m_setDisAsmVA.insert(dwVA);  // this is used to not push the same address onto the stack multiple times	}
  //	}
  
  return NULL; // pTarget;
}

void DisAssembler::SetBranchTarget (dword dwStartVA,
				    dword dwEndVA,
				    BranchTarget* pTarget)
{
  // printf ("SetBranchTarget(%08x,%08x,%08x)\n",dwStartVA,dwEndVA,pTarget);
  
  // printf ("DisAssembler::SetBranchTarget(%08x,,)\n",dwStartVA);
  
  MapDisAsmCodePart::iterator addTarget = m_mapDisAsmCodePart.find(dwStartVA);
  
  assert (addTarget != m_mapDisAsmCodePart.end());
  assert (!addTarget->second.m_pTarget);
  
  addTarget->second.m_dwEndVA = dwEndVA;
  addTarget->second.m_pTarget = pTarget;
  
  m_mapBranchTargetVA.insert(MapBranchTargetVA::value_type(pTarget,dwStartVA));
  
  // add jump and call sources if present and
  // add it to executable if it's a call target
  
  MapVAJumpTarget::iterator jump =
    m_mapVAJumpTarget.find(dwStartVA);
  
  if (jump != m_mapVAJumpTarget.end())
    {
      MapUnsetJumpTargetCallFunction::iterator func = 
	m_mapUnsetJumpTargetCallFunction.find(jump->second);
      
      assert (func != m_mapUnsetJumpTargetCallFunction.end());
      pTarget->SetJumpSource(jump->second);

      //printf ("func: %x (%d),(%d)\n",func->second,func->second->m_vectEntryPoints.size(),func->second->m_setCodeTraces.size());
      
      SetTraceFunc (pTarget->GetTraceHead(),func->second);
      
      m_mapUnsetJumpTargetCallFunction.erase(func);
    }

  MapVACallTarget::iterator call =
    m_mapVACallTarget.find(dwStartVA);

  if (call != m_mapVACallTarget.end())
    {
      pTarget->SetCallSource(call->second);
      AddCallTarget(call->second);
    }
}

/*
BranchTarget* DisAssembler::AddBranchTarget (dword dwTargetVA,dword dwEndVA)
{
	assert (m_mapDisAsmCodePart.find(dwTargetVA) == m_mapDisAsmCodePart.end());

	BranchTarget* pTarget = new BranchTarget();
	
	printf ("-%08x %08x  [%08x][%08x]\n",dwTargetVA,dwEndVA,pTarget);

	DisAsmCodePart newTarget = { dwEndVA,pTarget };

	m_mapDisAsmCodePart.insert(MapDisAsmCodePart::value_type(dwTargetVA,newTarget));

	//printf ("%08x->%08x (add)\n",dwVA,dwVA);

//	if (m_setDisAsmVA.find(dwVA) != m_setDisAsmVA.end())
//	{
//		m_setDisAsmVA.insert(dwVA);  // this is used to not push the same address onto the stack multiple times	}
//	}

	return pTarget;
}
*/
/*
void DisAssembler::SetBranchTarget (dword dwStartVA,dword dwEndVA,BranchTarget* pTarget)
{
	MapDisAsmCodePart::iterator setTarget = m_mapDisAsmCodePart.find(dwStartVA);

	if (setTarget != m_mapDisAsmCodePart.end())
	{
		assert (setTarget->second.m_dwEndsAtVA == dwStartVA);
		assert (setTarget->second.m_pTarget == NULL);

		setTarget->second.m_dwEndsAtVA = dwEndVA;
		setTarget->second.m_pTarget = pTarget;

		//printf ("%08x->%08x (set)\n",dwStartVA,dwEndVA);
	}
	else
	{
		DisAsmCodePart addTarget = { dwEndVA,pTarget };

		m_mapDisAsmCodePart.insert(MapDisAsmCodePart::value_type(dwStartVA,addTarget));

		//printf ("%08x->%08x (set)\n",dwStartVA,dwEndVA);
	}

	if (m_setCallTargetVA.find (dwStartVA) != m_setCallTargetVA.end ())
	{
		assert (m_pExecutable);
		m_pExecutable->AddCallTarget(dwStartVA,pTarget);
	}
}
*/
BranchTarget* DisAssembler::SplitCodeTrace (dword dwTargetVA,dword dwSplitAtVA)
{
  MapDisAsmCodePart::iterator splitTarget = 
    m_mapDisAsmCodePart.find(dwTargetVA);
  
  assert (splitTarget != m_mapDisAsmCodePart.end()); // must be found
  
  assert (splitTarget->second.m_dwEndVA > dwSplitAtVA); // must contain split position
  assert (splitTarget->second.m_pTarget); // must contain something to split
  
  // the only thing that can go wrong is that our split position is
  // INSIDE an multi-byte instruction handle that NOT so gracefully,
  // as it would mess up all our code to treat such stupid situations
  // (it's either a sign of self-modification, or a real hack)
  
  // the handling will be: choke at this CodeTrace, but treat the rest
  // of the program as far as possible. If we later can show that
  // either this, or the other guy that came here first actually were
  // unreachable, we'll try again after removing that code, but
  // otherwise we can't continue with editing of the program)

  BranchTargetTrace* pSplit = 
    dynamic_cast<BranchTargetTrace*>(splitTarget->second.m_pTarget);
  
  if (!pSplit)
    {
      assert (dynamic_cast<BranchTargetCtrlTransfer*>(splitTarget->second.m_pTarget)); // if it's not a trace, it has to be a control transfer
      
      // and control transfers only have one instruction, so we're in the middle
      
      throw ERROR_DISASM_CRAZY_JUMP_CANT_SPLIT(dwTargetVA,dwSplitAtVA);
      
      // throw "Can't split disassembled codetrace in middle of instruction. (Crazy jump)";
    }
  
  BranchTargetTrace* pSplitOffTarget = new BranchTargetTrace();

  try
    {
      pSplit->Split (dwSplitAtVA-dwTargetVA,pSplitOffTarget,dwTargetVA);
      
    }
  catch (const Error*)
    {
      delete pSplitOffTarget;
      throw ;
    }
#ifndef NDEBUG
  catch (...)
    {
		assert (false);
    }
#endif
  
  DisAsmCodePart newTarget = { splitTarget->second.m_dwEndVA,pSplitOffTarget };
  
  splitTarget->second.m_dwEndVA = dwSplitAtVA;
  
  m_mapDisAsmCodePart.insert(MapDisAsmCodePart::value_type(dwSplitAtVA,
							   newTarget));
  m_mapBranchTargetVA.insert(MapBranchTargetVA::value_type(pSplitOffTarget,
							   dwSplitAtVA));
  /*
    assert (pSplit->m_pToNext != pSplit);
	assert (pSplit->m_pFromPrevious != pSplit);
	assert (pSplitOffTarget->m_pToNext != pSplitOffTarget);
	assert (pSplitOffTarget->m_pFromPrevious != pSplitOffTarget);
  */
  return pSplitOffTarget;
}

/*
struct DisAsmCodePart
{
	dword dwEndsAtVA;
	BranchTarget* m_pCodeTrace;
};

	MapDisAsmCodePart m_mapDisAsmCodePart; // traces of code already (or about to be) handled
*/


dword DisAssembler::GetNextDisAsmVA ()
{
  dword dwVA = m_stackDisAsmVA.top();
  
  m_stackDisAsmVA.pop ();
  
  return dwVA;
}

Instruction* GetNextInstruction (InstructionDecoder& decoder,DataPtr* pDataPtr)
{
  Instruction* pInstr;
  dword dwInstrVA = pDataPtr->GetRVA();
  
  try
    {
      pInstr = decoder.Decode (*pDataPtr);
    }
  catch (Error* pError)
    {
      dword dwCodeEndVA = pDataPtr->GetRVA();
      
      pError->AddErrorInfoRVA(dwInstrVA);
      pError->AddErrorInfoRVA(dwCodeEndVA);

      char* fmt = pDataPtr->FormatAsHexBytes(dwInstrVA,dwCodeEndVA);
      EMIT_ERROR(ERROR_DISASM_DECODING_INSTRUCTION_FAILED(fmt,dwInstrVA,dwCodeEndVA));
      delete fmt;
      
      errorTracker.AddError(pError);
      
      /*
		printf ("\nException(1): ");
		pError->PrintShortMessage();
		printf ("%08x: ",dwInstrVA);

		dword dwCodeEndVA = pDataPtr->GetRVA();
		for (dword dwByteVA = dwInstrVA; dwByteVA < dwCodeEndVA; dwByteVA++)
		{
			byte bByte = pDataPtr->GetByte(dwByteVA);				
			printf ("%02x ",bByte);
		}
		printf ("\n");
      */
      return NULL;
    }
#ifndef NDEBUG
  catch (...)
    {
      assert (false);
    }
#endif
  
	// disassemble as long as we don't hit a branching instruction
/*
	printf ("  %08x:  ",dwInstrVA);

	dword dwCodeEndVA = pDataPtr->GetRVA();
	int i = 0;
	bool bOut = false;

	char szStr[64];

	pInstr->AsmString (szStr);

	for (dword dwByteVA = dwInstrVA; dwByteVA < dwCodeEndVA; dwByteVA++, i++)
	{
		byte bByte = pDataPtr->GetByte(dwByteVA);

		if (i == 7)
		{
			if (!bOut)
			{
				printf (" %s",szStr);
				bOut = true;
			}
			printf ("\n             ");
			i = 0;
		}
		
		printf ("%02x ",bByte);
	}

	for (; i < 7; i++)
		printf ("   ");

	if (!bOut)
		printf (" %s\n",szStr);
	else
		printf ("\n",szStr);
*/
  return pInstr;
}
/*
dword dwDisAssembled = 0;
*/

struct InstrAndLen
{
  byte			m_bLen;
  Instruction*	m_pInstr;
};

void AddToCodeTrace(InstrAndLen& pInstrAndLen,CodeTrace* pTrace)
{
  pTrace->AddInstruction(pInstrAndLen.m_pInstr,pInstrAndLen.m_bLen);
}

void DisAssembler::DisAsm (dword dwStartVA,DataPtr* pDataPtr)
{
  // how far can we go?
  
  MapDisAsmCodePart::iterator thisTarget = m_mapDisAsmCodePart.find(dwStartVA);
  
  assert (thisTarget != m_mapDisAsmCodePart.end ());
  assert (thisTarget->second.m_dwEndVA >= dwStartVA);
  
  assert (!thisTarget->second.m_pTarget ==
		thisTarget->second.m_dwEndVA != dwStartVA);
  
  if (thisTarget->second.m_pTarget)
    return; // already disassembled
  
  //	BranchTarget* pCodeTrace = thisTarget->second.m_pTarget; 
  //	assert (!pCodeTrace->HasInstructions()); // must not contain any instructions if it han't been disassembled
  
  MapDisAsmCodePart::iterator nextTarget = thisTarget;
  nextTarget++;
  
  BranchTarget* pPrevFallThrough = NULL;
  
  if (thisTarget != m_mapDisAsmCodePart.begin()) // something before us
    {
      MapDisAsmCodePart::iterator prevTarget = thisTarget;
      prevTarget--;
      
      if (prevTarget->second.m_dwEndVA == dwStartVA) // end just before us
	{
	  BranchTargetCtrlTransfer* pCtrlTransfer = 
	    dynamic_cast<BranchTargetCtrlTransfer*>(prevTarget->second.m_pTarget);
	  
	  if (!pCtrlTransfer ||
	      pCtrlTransfer->GetBranchInstr()->CanTraceContinue())
	    pPrevFallThrough = prevTarget->second.m_pTarget; 				
	}
    }

  dword dwNextVA = 
    (nextTarget == m_mapDisAsmCodePart.end() ? 0xffffffff : nextTarget->first);
  
  assert (dwStartVA < dwNextVA); // there must be something to decode before next target, or we shouldn't be here
  
  //	printf ("+%08x(%08x) [%08x][%08x]\n",dwStartVA,dwNextVA,pCodeTrace);
  
  InstructionDecoder decoder;
  
  BranchInstr* pBranchInstr = NULL;
  
  dword dwEndLastInCodeTraceVA = dwStartVA;
  dword dwEndThisVA;
  
  bool bInstrTrashTarget = false;
  
  vector<InstrAndLen> m_vectInstrs;
  
  for ( ; ; )
    {
      // dword dwStartLastVA = pDataPtr->GetRVA();
      //printf ("Working at %08x\n",dwStartLastVA);
      Instruction *pInstr = GetNextInstruction (decoder,pDataPtr);
      
      if (!pInstr)
	{
	  // decoding failed
	  break;
	}
      
      dwEndThisVA = pDataPtr->GetRVA();
      
      if (dwEndThisVA > dwNextVA)
	{
	  // we have decoded instruction passing on a branch target
	  // not allowed
	  
	  // first we must save what was found before,
	  
	  bInstrTrashTarget = true;
	  
	  break;
	  
	  // after that: throw
	}
      
      if ((pBranchInstr = dynamic_cast<BranchInstr*>(pInstr)))
	break;
      
      //		assert (pCodeTrace);
      
      dword dwLength = dwEndThisVA-dwEndLastInCodeTraceVA;
      assert (dwLength < 256);
      
      dwEndLastInCodeTraceVA = dwEndThisVA;
      
      InstrAndLen instr;
      
      instr.m_bLen = (byte) dwLength;
      instr.m_pInstr = pInstr;
      
      m_vectInstrs.push_back(instr);
      
      // pCodeTrace->AddInstruction (pInstr,(byte) dwLength); // keep 
      
      if (dwEndThisVA == dwNextVA)
	{
	  // we landed on next instruction (which has already been disassembled and
	  // therefore is an branch target)
	  // and could just continue into it (we'll link to it as next-ptr)
	  
	  break;
	}
    }
  if (!m_vectInstrs.empty ())
    {
      BranchTargetTrace* pTrace = new BranchTargetTrace();
      
      for_each (m_vectInstrs.begin(),m_vectInstrs.end(),AddToCodeTrace,pTrace);
      
      dword dwLength = dwEndLastInCodeTraceVA-dwStartVA;
      pTrace->SetRawCode (pDataPtr->GetPtr(dwStartVA,dwLength),dwLength);
      
      SetBranchTarget (dwStartVA,dwEndLastInCodeTraceVA,pTrace);
      
      /*
	dwDisAssembled += dwLength;
	printf ("disasm: %d %d\n",dwDisAssembled,dwLength);
      */
      if (pPrevFallThrough)
	{
	  assert (!pPrevFallThrough->GetNextTarget());
	  BindTrace (pPrevFallThrough,pTrace);
	}
      pPrevFallThrough = pTrace;
      
      BranchTarget* pNext = AddBranchTarget (dwEndLastInCodeTraceVA);

      if (pNext) // if next is null (not disassembled yet), we'll have to set it later
	BindTrace (pTrace,pNext);
    }
  if (pBranchInstr)
    {
      BranchTargetCtrlTransfer* pBranch = new BranchTargetCtrlTransfer(); 
      
      SetBranchTarget (dwEndLastInCodeTraceVA,dwEndThisVA,pBranch);
      
      dword dwLength = dwEndThisVA-dwEndLastInCodeTraceVA;
      assert (dwLength < 256);
      
      pBranch->SetInstruction (pBranchInstr); // keep 
      /*
	dwDisAssembled += dwLength;
	printf ("disasm: %d %d*\n",dwDisAssembled,dwLength);
      */
      pBranch->SetRawCode (pDataPtr->GetPtr(dwEndLastInCodeTraceVA,dwLength),
			   dwLength);
      
      if (pPrevFallThrough)
	{
	  assert (!pPrevFallThrough->GetNextTarget());
	  BindTrace (pPrevFallThrough,pBranch);
	}
      
      if (pBranchInstr->CanTraceContinue ())
	{
	  BranchTarget* pNext = AddBranchTarget (dwEndThisVA);
	  if (pNext) // if next it null (not disassembled yet), we'll have to set it later
	    BindTrace (pBranch,pNext);
	}
      
      // sort out what we're pointing to
      
      // if we may continue, set the following instruction as a branch
      // target
      
      if (pBranchInstr->IsTargetRelative ())
	{
	  // the target is encoded using offset in immediate that
	  // means that we can deduce the target here and now (now
	  // code analysis needed)
	  
	  SetRelativeBranchTarget (dwEndLastInCodeTraceVA,
				   pBranch,pBranchInstr);
	}
    }
  /*
    if (pBranchTarget)
    AddBranchTarget (pBranchTarget,dwStartVA,dwStartLastVA,mapHandledCode);
  */
  
  if (bInstrTrashTarget)
    throw ERROR_DISASM_CRAZY_JUMP_DISASM_PAST_TARGET(dwEndLastInCodeTraceVA,dwNextVA);
}

void DisAssembler::SetRelativeBranchTarget (dword dwStartVA,
					    BranchTargetCtrlTransfer* pBranch,
					    BranchInstr* pBranchInstr)
{
  // printf ("SetRelativeBranchTarget(%08x,%08x,%08x)\n",dwStartVA,pBranch,pBranchInstr);
  
  assert (pBranchInstr->m_vectTargets.empty());
  // assert (!pBranchInstr->m_bVectTargetsComplete);

  // the target is encoded using offset in immediate
  // that means that we can deduce the target here and now (no code analysis needed)
  
  // so? get immediate
  
  dword dwTargetVA = dwStartVA+pBranch->GetRawCodeLength()+
    pBranchInstr->GetOffset();
  
  CtrlTransferTarget* pTarget;
  
  if (pBranchInstr->IsJump())
    {
      MapCodeTraceCallFunction::iterator func = 
	m_mapCodeTraceCallFunction.find(pBranch->GetTraceHead());
      
      assert (func != m_mapCodeTraceCallFunction.end());
      
      CallFunction* pFunction = func->second;
      
      pTarget = AddJumpTarget (dwTargetVA,pFunction); // calls AddBranchTarget if necessary
    }
  else
    {
      assert (pBranchInstr->IsCall());
      pTarget = AddCallTarget (dwTargetVA);
    }
  
  //	printf ("Set target (0x%08x) (ptr:%08x)\n",dwTargetVA,pTarget);
  
  if (pTarget) // if next is null (not disassembled yet), we'll have to set it later
    {			
      pBranchInstr->AddTarget (pTarget/*,true*/);
      pTarget->AddTargetSource(pBranch);
    }
}

void DisAssembler::DeleteUnusedTargets ()
{
  // delete all CallTargets and JumpTargets
  // that are not followed by any BranchTargets
  
#ifndef NDEBUG
  // make sure all entries really are unused
  FOR_EACH_ITER(MapUnsetJumpTargetCallFunction,
		m_mapUnsetJumpTargetCallFunction,check)
    {
      //		printf ("Checking jump target (0x%08x) (m_pTarget: 0x%08x, m_pMoreSource: 0x%08x)\n",check->first,check->first->m_pTarget,check->first->m_pMoreSource);
      
      assert (!check->first->m_pTarget);
      assert (!check->first->m_pMoreSource);
    }
#endif

  for (MapVACallTarget::iterator call = m_mapVACallTarget.begin(); 
       call != m_mapVACallTarget.end(); )
    {
      if (!call->second->m_pTarget)
	{
	  // has no target, delete
	  
	  assert (!call->second->m_pMoreSource);
	  
	  call->second->RemoveTargetSources();
	  delete call->second;
	  
	  SET_MAP_ERASE_CONTINUE_ITER(m_mapVACallTarget,call);
	  // call = m_mapVACallTarget.erase(call);
	}
      else
	++call;
    }
  
  for (MapVAJumpTarget::iterator jump = m_mapVAJumpTarget.begin(); 
       jump != m_mapVAJumpTarget.end(); )
    {
      if (!jump->second->m_pTarget)
	{
	  // has no target, delete
	  
	  assert (m_mapUnsetJumpTargetCallFunction.find(jump->second) != 
		  m_mapUnsetJumpTargetCallFunction.end()); // there should be an entry here
	  
	  assert (!jump->second->m_pMoreSource);
	  
	  jump->second->RemoveTargetSources();
	  delete jump->second;
	  
	  SET_MAP_ERASE_CONTINUE_ITER(m_mapVAJumpTarget,jump);
	  //			jump = m_mapVAJumpTarget.erase(jump);
	}
      else 
	++jump;
    }
  
  // now clear out list of unused jump targets
  
  m_mapUnsetJumpTargetCallFunction.clear();
  
  // clear from m_mapDisAsmCodePart addresses that didn't disassemble at all	
  for(MapDisAsmCodePart::iterator target = m_mapDisAsmCodePart.begin(); 
      target != m_mapDisAsmCodePart.end(); )
    {
      if (!target->second.m_pTarget)
	SET_MAP_ERASE_CONTINUE_ITER(m_mapDisAsmCodePart,target);
      //			target = m_mapDisAsmCodePart.erase(target);
      else
			++target;
    }
}

#ifndef NDEBUG
void DisAssembler::ConsistencyCheck ()
{
  FOR_EACH_ITER(MapDisAsmCodePart,m_mapDisAsmCodePart,target)
    {
      BranchTarget* pTarget = target->second.m_pTarget;
      BranchTarget* pTrace = pTarget->GetTraceHead();
      
      MapCodeTraceCallFunction::iterator func = 
	m_mapCodeTraceCallFunction.find(pTrace);
      
      assert (func != m_mapCodeTraceCallFunction.end()); // we should be known
      CallFunction* pFunc = func->second;
      assert (pFunc);
      
      // make sure linking is ok (fall-through lists)
      
      pTarget->ConsistencyCheck(pFunc);		
      
      // make sure all branch targets belong to a function (and that
      // that function know about us). Also make sure that it only is
      // the targets trace head that has the info
      
      if (pTrace == pTarget)
	{
	  assert (pFunc->m_setCodeTraces.find(pTrace) != 
		  pFunc->m_setCodeTraces.end()); // we should be known
	}
      else
	{
	  MapCodeTraceCallFunction::iterator func2 = 
	    m_mapCodeTraceCallFunction.find(pTarget);
	  assert (func2 == m_mapCodeTraceCallFunction.end()); // we should not be known
	}
    }
}
#endif


/*
  void DisAssembler::FixUpTargetPointers ()
{
	// fix target pointers for relative jumps and calls

	for (MapDisAsmCodePart::iterator fixTarget = m_mapDisAsmCodePart.begin();
		fixTarget != m_mapDisAsmCodePart.end(); fixTarget++)
	{
		// sanity check:
		// forward and backward pointers are in sync

		BranchTarget* pFix = fixTarget->second.m_pTarget;

		if (!pFix)
			continue; // dissambly failed

//		printf ("Fixup trace (0x%08x)\n",pFix->m_dwVA);

		assert (!pFix->m_pToNext ||
			pFix->m_pToNext->m_pFromPrevious == pFix);
		assert (!pFix->m_pFromPrevious ||
			pFix->m_pFromPrevious->m_pToNext == pFix);

#ifndef NDEBUG
		for (VectFromTrace::iterator fromTrace = pFix->m_vectFromTrace.begin();
			fromTrace != pFix->m_vectFromTrace.end(); fromTrace++)
		{
			assert (*fromTrace);
				
			BranchTargetCtrlTransfer* pFrom = dynamic_cast<BranchTargetCtrlTransfer*>(*fromTrace);

			assert (pFrom);

			VectTargets::iterator iterFindFix = 
				find(pFrom->GetInstruction()->m_vectTargets.begin(),
					 pFrom->GetInstruction()->m_vectTargets.end(),
					 pFix);

			assert (*iterFindFix == pFix);
		}
#endif // NDEBUG

		BranchTargetCtrlTransfer* pBranch = dynamic_cast<BranchTargetCtrlTransfer*>(pFix);

		if (pBranch)
		{
			BranchInstr* pBranchInstr = pBranch->GetInstruction();

			if (pBranchInstr->IsTargetRelative () &&
				!pBranchInstr->m_bVectTargetsComplete) // does not know target
			{
//				printf ("Fixup target from (0x%08x)\n",pBranch->m_dwVA);

				assert (m_mapBranchTargetVA.find(pBranch) !=
						m_mapBranchTargetVA.end());

				dword dwAddress = m_mapBranchTargetVA.find(pBranch)->second;

				SetRelativeBranchTarget (dwAddress,pBranch,pBranchInstr,false);

#ifndef NDEBUG
				if (!pBranchInstr->m_bVectTargetsComplete)
					printf ("Couldn't fixup jump from (0x%08x) to (0x%08x)\n",dwAddress,
						dwAddress+pBranch->GetRawCodeLength()+
							pBranchInstr->GetOffset());
#endif // NDEBUG
			}
		}
	}
}
*/







