/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// DisAssembler.h: interface for the DisAssembler class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__DIS_ASSEMBLER_H__INCLUDED_)
#define __DIS_ASSEMBLER_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

class Executable; 
class DataPtr; 
class BranchTargetCtrlTransfer; 
class BranchInstr;
class CallTarget; 
class JumpTarget; 
class ExportedObject;

#include "BranchTarget.h"

#include <stack>
#include <map>
#include <set>

using namespace std;

struct DisAsmCodePart
{
  dword m_dwEndVA;
  BranchTarget* m_pTarget;
};

typedef map<dword,DisAsmCodePart>      MapDisAsmCodePart;

typedef map<dword,CallTarget*> MapVACallTarget;
typedef map<dword,JumpTarget*> MapVAJumpTarget;

typedef map<JumpTarget*,CallFunction*> MapUnsetJumpTargetCallFunction;

typedef map<const BranchTarget*,CallFunction*> MapCodeTraceCallFunction;

class ReportDisAsmInfo;
class UrlFolder;

class DisAssembler  
{
 public:
  DisAssembler();
  virtual ~DisAssembler();
  
 protected:
  Executable* m_pExecutable;
  
 protected:
  MapDisAsmCodePart   m_mapDisAsmCodePart;       // traces of code already (or about to be) handled
  MapVACallTarget     m_mapVACallTarget; // where our call targets are
  MapVAJumpTarget     m_mapVAJumpTarget; // where our jump targets are
  MapUnsetJumpTargetCallFunction m_mapUnsetJumpTargetCallFunction;	
  
  MapCodeTraceCallFunction m_mapCodeTraceCallFunction;
  
  MapBranchTargetVA m_mapBranchTargetVA;
	
 protected:
  stack<dword> m_stackDisAsmVA; // the VAs to disassemble, in order to be done
  //	set<dword> m_setDisAsmVA;     // known VAs to disassemble, but not yet handled (sorted)
  
  //set<dword> m_setCallTargetVA; // the VAs of all functions entry points
  
 public:
  const MapBranchTargetVA* GetMapBranchTargetVA () 
    { return &m_mapBranchTargetVA; }
  const MapCodeTraceCallFunction* GetMapCodeTraceCallFunction() 
    { return &m_mapCodeTraceCallFunction; }
  
 public:
  void SetExecutable (Executable* pExec) 
    {
      assert (!m_pExecutable); 
      m_pExecutable = pExec; 
    }
  
 public:
  BranchTarget* AddBranchTarget (dword dwTargetVA,bool bAddIfNotKnown = true);
  //BranchTarget* AddBranchTarget (dword dwTargetVA,dword dwEndVA);
  CtrlTransferTarget* AddCallTarget (dword dwTargetVA,
				     ExportedObject* pExportedObj = NULL);
  CtrlTransferTarget* AddCallTarget (ExportedObject* pExportedObj);
  CtrlTransferTarget* AddJumpTarget (dword dwTargetVA,CallFunction* pFunction);
  
  void SetBranchTarget (dword dwStartVA,dword dwEndVA,BranchTarget* pTarget);
  
  void SetRelativeBranchTarget (dword dwStartVA,
				BranchTargetCtrlTransfer* pBranch,
				BranchInstr* pBranchInstr);
  
 protected:
  void AddCallTarget (CallTarget* pCallTarget);
  void BindTrace (BranchTarget* pBefore,BranchTarget* pAfter);
  void SetTraceFunc (BranchTarget* pTrace,CallFunction* pFunction);
  void MergeTraceFunc (CallFunction* pDestFunc,CallFunction* pSrcFunc);
  
 protected:
  BranchTarget* SplitCodeTrace (dword dwTargetVA,dword dwSplitAtVA);
  
 public:
  bool MoreToDisAsm () { return !m_stackDisAsmVA.empty(); }
  dword GetNextDisAsmVA ();
  void DisAsm (dword dwVA,DataPtr* pDataPtr);
  
  void DeleteUnusedTargets ();
  // void FixUpTargetPointers ();

#ifndef NDEBUG
  void ConsistencyCheck ();
#endif
  
 public:
  void DataAnalysis (UrlFolder* pParentFolder,
		     ReportDisAsmInfo& reportDisAsmInfo); // I don't like the precense of ReportDisAsmInfo here
  
 protected:
  void DataAnalysisFunction(CallFunction* pFunction);
  
 protected:
  void ConvertTracesToSingle (CallFunction* pFunction);
  void ConvertTracesToMultiple (CallFunction* pFunction);
  
  void RedirectPointers (BranchTarget* pFrom,BranchTarget* pTo);
  
  
  
};

#endif // !defined(__DIS_ASSEMBLER_H__INCLUDED_)




