/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ELF_File.h: interface for the ELF_File class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__E_L_F__FILE_H__INCLUDED_)
#define __E_L_F__FILE_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "ExecutableImage.h"
#include "FileMap.h"

//////////////////////////////////////////////////////////////////////

typedef  dword Elf32_Addr;
typedef   word Elf32_Half;
typedef  dword Elf32_Off;
typedef sdword Elf32_Sword;
typedef  dword Elf32_Word;

//////////////////////////////////////////////////////////////////////

// be somewhat careful with the comments here, because this file is
// processed by a script to generate the html-dump of the headers
// (including certain comments a textual info

struct Elf32_Ident // Dumpable
{
  dword dw_magic;				// Magic number
  
  byte  b_class;				// Class
  byte  b_data;					// Data Endianess
  byte  b_version;				// ELF Header Version
  
  byte  b_pad[9];				// Padding
};

//////////////////////////////////////////////////////////////////////

#define EI_NIDENT 16

//////////////////////////////////////////////////////////////////////
// 

#define ELF_ID_MAGIC 0x464c457f // 0x75,E,L,F (in reverse)

//////////////////////////////////////////////////////////////////////
// Elf32_Ident.b_class

#define ELF_ID_CLASSNONE	0		// Invalid class
#define ELF_ID_CLASS32		1		// 32-bit objects
#define ELF_ID_CLASS64		2		// 64-bit objects

//////////////////////////////////////////////////////////////////////
// Elf32_Ident.b_data

#define ELF_ID_DATANONE		0		// Invalid data encoding
#define ELF_ID_DATA2LSB		1		// Little-endian
#define ELF_ID_DATA2MSB		2		// Big-endian

//////////////////////////////////////////////////////////////////////

struct Elf32_Header // Dumpable
{
  // byte e_ident [EI_NIDENT];	// moved into Elf32_Eident structure 

  Elf32_Half e_type;			// File type
  Elf32_Half e_machine;			// Target machine
  Elf32_Word e_version;			// ELF version
  Elf32_Addr e_entry;			// Entry point
  Elf32_Off  e_phoff;			// PHT file offset
  Elf32_Off  e_shoff;			// SHT file offset
  Elf32_Word e_flags;			// Machine flags
  Elf32_Half e_ehsize;			// Header size
  Elf32_Half e_phentsize;		// PHT entry size
  Elf32_Half e_phnum;			// PHT entries
  Elf32_Half e_shentsize;		// SHT entry size
  Elf32_Half e_shnum;			// SHT entries
  Elf32_Half e_shstrndx;		// SHT strings
};

//////////////////////////////////////////////////////////////////////
// Elf32_Header.e_type

#define ELF_ET_NONE			0		// No file type
#define ELF_ET_REL			1		// Relocatable file
#define ELF_ET_EXEC			2		// Executable file
#define ELF_ET_DYN			3		// Shared object file
#define ELF_ET_CORE			4		// Core file

#define ELF_ET_LOPROC		0xff00	// Processor-specific
#define ELF_ET_HIPROC		0xffff	// Processor-specific 

//////////////////////////////////////////////////////////////////////
// Elf32_Header.e_machine

#define ELF_EM_NONE			0		// No machine
#define ELF_EM_386			3		// Intel 80386

//////////////////////////////////////////////////////////////////////
// Elf32_Ident.b_version
// Elf32_Header.e_version

#define ELF_EV_NONE			0		// Invalid version
#define ELF_EV_CURRENT		1		// Current version

//////////////////////////////////////////////////////////////////////

#define ELF_SHN_UNDEF		0
#define ELF_SHN_LORESERVE	0xff00
#define ELF_SHN_LOPROC		0xff00
#define ELF_SHN_HIPROC		0xff1f
#define ELF_SHN_ABS		0xfff1	// Absolute
#define ELF_SHN_COMMON		0xfff2	// Common
#define ELF_SHN_HIRESERVE	0xffff

//////////////////////////////////////////////////////////////////////

struct Elf32_SectionHeader // Dumpable (name_object)
{
  Elf32_Word	sh_name;		// Name index    (fmt:+name_object)
  Elf32_Word	sh_type;		// Type
  Elf32_Word	sh_flags;		// Flags
  Elf32_Addr	sh_addr;		// Im. address
  Elf32_Off 	sh_offset;		// File off.
  Elf32_Word	sh_size;		// Size
  Elf32_Word	sh_link;		// Link
  Elf32_Word	sh_info;		// Info
  Elf32_Word	sh_addralign;	// Align
  Elf32_Word	sh_entsize;		// Ent sz
};

//////////////////////////////////////////////////////////////////////
// Elf32_SectionHeader.sh_type

#define ELF_SHT_NULL		0			// Inactive
#define ELF_SHT_PROGBITS	1			// ProgBits
#define ELF_SHT_SYMTAB		2			// Symbol table
#define ELF_SHT_STRTAB		3			// String table
#define ELF_SHT_RELA		4			// Reloc (RELA)
#define ELF_SHT_HASH		5			// Symbol hash
#define ELF_SHT_DYNAMIC		6			// Dynamic
#define ELF_SHT_NOTE		7			// Note
#define ELF_SHT_NOBITS		8			// NoBits
#define ELF_SHT_REL	    	9			// Reloc (REL)
#define ELF_SHT_SHLIB		10			// Shlib
#define ELF_SHT_DYNSYM		11			// Symtab (min)
#define ELF_SHT_HI_SPEC_DEF (ELF_SHT_DYNSYM)
#define ELF_SHT_LOPROC		0x70000000
#define ELF_SHT_HIPROC		0x7fffffff
#define ELF_SHT_LOUSER		0x80000000
#define ELF_SHT_HIUSER		0xffffffff

//////////////////////////////////////////////////////////////////////
// Elf32_SectionHeader.sh_flags

// Mask: 
#define ELF_SHF_ALLOC		0x2			// Alloc
#define ELF_SHF_WRITE		0x1			// Write 
#define ELF_SHF_EXECINSTR	0x4			// Execute 
#define ELF_SHF_MASKPROC	0xf0000000

#define ELF_SHF_ALLOC_IF_LOAD	0x2000

#define ELF_SHF_ACCESS_MASK (ELF_SHF_WRITE | ELF_SHF_ALLOC | ELF_SHF_EXECINSTR)

//////////////////////////////////////////////////////////////////////

struct Elf32_ProgramHeader // Dumpable
{
  Elf32_Word	p_type;			// Type
  Elf32_Off     p_offset;		// File offset
  Elf32_Addr	p_vaddr;		// Im. address		
  Elf32_Addr	p_paddr;		// Phys. address
  Elf32_Word	p_filesz;		// File size
  Elf32_Word	p_memsz;		// Image size
  Elf32_Word	p_flags;		// Flags
  Elf32_Word	p_align;		// Align
};

//////////////////////////////////////////////////////////////////////
// Elf32_ProgramHeader.p_type

#define ELF_PT_NULL	    	0			// Unused
#define ELF_PT_LOAD	        1			// Loadable
#define ELF_PT_DYNAMIC		2			// Dynamic linking info
#define ELF_PT_INTERP		3			// Program interpreter
#define ELF_PT_NOTE	    	4			// Note section
#define ELF_PT_SHLIB		5			// Shlib (unspecified)
#define ELF_PT_PHDR	    	6			// Program header
#define ELF_PT_LOPROC		0x70000000
#define ELF_PT_HIPROC		0x7fffffff

//////////////////////////////////////////////////////////////////////
// Elf32_ProgramHeader.p_flags

// Mask: 
#define ELF_PT_R		4			// Read
#define ELF_PT_W		2			// Write
#define ELF_PT_X		1			// Execute
#define ELF_PT_MASKOS	0x0ff00000
#define ELF_PT_MASKPROC	0xf0000000	

//////////////////////////////////////////////////////////////////////

struct Elf32_Symbol // Dumpable (name_object)
{
  Elf32_Word	st_name;		// Name    (fmt:+name_object)
  Elf32_Addr	st_value;		// Value
  Elf32_Word	st_size;		// Size
  unsigned char	st_info;		// Info
  unsigned char	st_other;		
  Elf32_Half	st_shndx;		// Section #
};

//////////////////////////////////////////////////////////////////////

#define ELF_STB_BIND_MASK	0xf0
#define ELF_STB_TYPE_MASK	0x0f

#define ELF32_ST_BIND(i)	((i)>>4)
#define ELF32_ST_TYPE(i)	((i)&0xf)
#define ELF32_ST_INFO(b,t)	(((b)<<4)+((t)&0xf))

//////////////////////////////////////////////////////////////////////
// Elf32_Symbol.st_info

#define ELF_STB_LOCAL		0
#define ELF_STB_GLOBAL		1
#define ELF_STB_WEAK		2
#define ELF_STB_LOPROC		13
#define ELF_STB_HIPROC		15

/* these three are just for the dumping */
// Mask:ELF_STB_BIND_MASK
#define ELF_STB_LOCAL_DMP	(ELF_STB_LOCAL<<4)		// Local
#define ELF_STB_GLOBAL_DMP	(ELF_STB_GLOBAL<<4)		// Global
#define ELF_STB_WEAK_DMP	(ELF_STB_WEAK<<4)		// Weak

// Mask:ELF_STB_TYPE_MASK
#define ELF_STT_NOTYPE		0		// NoType
#define ELF_STT_OBJECT		1		// Object
#define ELF_STT_FUNC		2		// Function
#define ELF_STT_SECTION		3		// Section
#define ELF_STT_FILE		4		// File
#define ELF_STT_LOPROC		13
#define ELF_STT_HIPROC		15

//////////////////////////////////////////////////////////////////////

struct Elf32_Hash // Dumpable
{
  Elf32_Word		h_buckets;		// Buckets
  Elf32_Word		h_chains;		// Chains
};

//////////////////////////////////////////////////////////////////////

union dyn_d_un
{
    Elf32_Word d_val;        /*   // Value*/
    Elf32_Addr d_ptr;        /*   // Ptr*/
};

struct Elf32_Dynamic // Dumpable
{
  Elf32_Sword d_tag;              // Tag
  /* this is for fooling the struct dumper (if you don't like this not
     so very nice way of doing things, make the dumper more competent
     so this can be avoided (one really bad thing is that it end the
     structure search at the first end brace it encounters, and not at
     the matching end brace (I can not put the actual character here,
     because then we would end inside this comment :-( ) Elf32_Word
     d_un.d_val; // Value/Ptr */
  dyn_d_un d_un;
  /* 
     {
     Elf32_Word d_val;        
     Elf32_Addr d_ptr;        
     } d_un;
  */
};
// externElf32_Dyn_DYNAMIC[];

//////////////////////////////////////////////////////////////////////
// Elf32_Dynamic.d_tag

#define ELF_DT_NULL     0   /* ignored mandatory mandatory */ // End
#define ELF_DT_NEEDED   1   /* d_val   optional  optional  */ // NeededLib
#define ELF_DT_PLTRELSZ 2   /* d_val   optional  optional  */ // PltRelSize
#define ELF_DT_PLTGOT   3   /* d_ptr   optional  optional  */ // PltGot
#define ELF_DT_HASH     4   /* d_ptr   mandatory mandatory */ // Hash
#define ELF_DT_STRTAB   5   /* d_ptr   mandatory mandatory */ // StrTab
#define ELF_DT_SYMTAB   6   /* d_ptr   mandatory mandatory */ // SymTab
#define ELF_DT_RELA     7   /* d_ptr   mandatory optional  */ // Rela
#define ELF_DT_RELASZ   8   /* d_val   mandatory optional  */ // RelaSize
#define ELF_DT_RELAENT  9   /* d_val   mandatory optional  */ // RelaEntrySize
#define ELF_DT_STRSZ    10  /* d_val   mandatory mandatory */ // StringsSize
#define ELF_DT_SYMENT   11  /* d_val   mandatory mandatory */ // SymEntrySize
#define ELF_DT_INIT     12  /* d_ptr   optional  optional  */ // Init
#define ELF_DT_FINI     13  /* d_ptr   optional  optional  */ // Fini
#define ELF_DT_SONAME   14  /* d_val   ignored   optional  */ // Sonami
#define ELF_DT_RPATH    15  /* d_val   optional  ignored   */ // Rpath
#define ELF_DT_SYMBOLIC 16  /* ignored ignored   optional  */ // Symbolic
#define ELF_DT_REL      17  /* d_ptr   mandatory optional  */ // Rel
#define ELF_DT_RELSZ    18  /* d_val   mandatory optional  */ // Rel size
#define ELF_DT_RELENT   19  /* d_val   mandatory optional  */ // Rel entry size
#define ELF_DT_PLTREL   20  /* d_val   optional  optional  */ // PtlRelType
#define ELF_DT_DEBUG    21  /* d_ptr   optional  ignored   */ // Debug
#define ELF_DT_TEXTREL  22  /* ignored optional  optional  */ // TextRel
#define ELF_DT_JMPREL   23  /* d_ptr   optional  optional  */ // JmpRel
#define ELF_DT_LOPROC   0x70000000 /* unspecified unspecified unspecified */
#define ELF_DT_HIPROC   0x7fffffff /* unspecified unspecified unspecified */

#define ELF_DT_NUM_NUMBERED 24 // last numbered num + 1 */

//////////////////////////////////////////////////////////////////////

struct Elf32_NoteHeader
{
  Elf32_Word		n_namesz;	// Name
  Elf32_Word		n_descsz;	// Value
  Elf32_Sword		n_type;		// Size
};

//////////////////////////////////////////////////////////////////////

struct Elf32_Rel // Dumpable
{
  Elf32_Addr  r_offset;       // Offset
  Elf32_Word  r_info;         // Info
};

//////////////////////////////////////////////////////////////////////

struct Elf32_Rela // Dumpable
  : public Elf32_Rel
{
  /*
    Elf32_Addr  r_offset;       // Offset
    Elf32_Word  r_info;         // Info
  */
  Elf32_Sword r_addend;       // Addend
};

//////////////////////////////////////////////////////////////////////

#define ELF32_R_SYM(i)      ((i) >> 8)
#define ELF32_R_TYPE(i)     ((unsigned char) (i))
#define ELF32_R_INFO(s,t)   (((s) << 8) + (unsigned char) (t))

#define ELF_32_R_SYM_MASK	0xffffff00
#define ELF_32_R_TYPE_MASK	0x000000ff

//////////////////////////////////////////////////////////////////////
// Elf32_Rel.r_info
// Elf32_Rela.r_info

// Mask:ELF_32_R_TYPE_MASK
#define ELF_R_386_NONE     0  /* none   none        */ // R_NONE    
#define ELF_R_386_32       1  /* word32 S + A       */ // R_32      
#define ELF_R_386_PC32     2  /* word32 S + A - P   */ // R_PC32    
#define ELF_R_386_GOT32    3  /* word32 G + A - P   */ // R_GOT32   
#define ELF_R_386_PLT32    4  /* word32 L + A - P   */ // R_PLT32   
#define ELF_R_386_COPY     5  /* none   none        */ // R_COPY    
#define ELF_R_386_GLOB_DAT 6  /* word32 S           */ // R_GLOB_DAT
#define ELF_R_386_JMP_SLOT 7  /* word32 S           */ // R_JMP_SLOT
#define ELF_R_386_RELATIVE 8  /* word32 B + A       */ // R_RELATIVE
#define ELF_R_386_GOTOFF   9  /* word32 S + A - GOT */ // R_GOTOFF  
#define ELF_R_386_GOTPC    10 /* word32 GOT + A - P */ // R_GOTPC   

//////////////////////////////////////////////////////////////////////

#include <vector>

class Elf32_ProgramBlock;
class Elf32_Section;
class Elf32_SectionStringTable;
class Elf32_SectionSymbolTable;

struct FileSectionsLayoutInfo;

typedef vector<Elf32_ProgramBlock*> VectElf32_ProgramBlocks;
typedef vector<Elf32_Section*>      VectElf32_Sections;

struct FileLayoutInfo
{
  dword dwImageBase;
  dword dwFileNextFree;
  dword dwImageNextFree;
};

class DataPtrSrc;
class WriteRelocDest;
class ExportedObject;

typedef vector<ExportedObject*> VectExportedObjects;

class ELF_File : public ExecutableImage
{
 public:
  ELF_File(const string& strFileName);
  virtual ~ELF_File();
  
 public:
  // I first figured to read each section into memory separately, but
  // then realized that it will not be a good idea, as some program
  // headers reference memory outside the sections (e.g. if the
  // sections are not present), so then I would have to treat the rest
  // in some way anyway. So what wo do is to read the entire file, and
  // then when and if there are sections and program blocks, we'll
  // have classes for that to contain the logic, but using the common
  // memory already alloced.

  byte* m_pFileData;
  Elf32_Word m_nSize;
  
 public:
  Elf32_Ident m_elfIdent;
  Elf32_Header m_elfHeader;

  Elf32_Addr e_entryInit;			
  Elf32_Addr e_entryFini;			

  Elf32_ProgramHeader* m_pProgramHeaders;
  Elf32_SectionHeader* m_pSectionHeaders;

  VectElf32_ProgramBlocks m_vectProgramBlocks;
  VectElf32_Sections      m_vectSections;

  VectExportedObjects m_vectExportedObjects;

 public:
  FileMap m_fileMap;

 public:
  FileLayoutInfo m_fileLayout;

 public:
  static bool IsType (int file);
  
 public:
  virtual void ReadFile (int file);
  virtual void PostReadFile ();
  virtual bool ProcessFile (UrlFolder* pFolder);

 public:
  void GetFileData(void* pDest,Elf32_Off nOffset,Elf32_Off nSize);
  DataPtrSrc* GetFileData(Elf32_Off nOffset,Elf32_Off nSize);
 
 public:
  virtual ExecutableImage* GetWriteImage(const string& strFileName);
  
  virtual bool PrepareWrite();
  virtual bool WriteFile();
  
  void DoFileLayout (word wRelocSection,
		     const FileSectionsLayoutInfo& layoutInfo);
  void DoFileLayout (word wRelocSectionStart);
  void DoFileLayout (word wRelocSectionStart,word wRelocSectionEnd);
  void CutFileAfter (word wRelocSectionKeep);

 public:
  void CreateCodeSection(Elf32_Section* pDestSection);
  
 public:
  virtual void HTMLFileInfo (UrlFolder* pParent);

 protected:
  Elf32_Section* GetSection(Elf32_Word sect);
  Elf32_SectionStringTable* GetStringSection(Elf32_Word sect);
  Elf32_SectionSymbolTable* GetSymbolSection(Elf32_Word sect);

 protected:
  virtual DataPtr* GetDisAsmDataPtr (dword dwVirtualAddress);

  virtual void GetExportedCallTargets (DisAssembler& disasm);

  void ReadImportExportSymbols ();

 public:
    Elf32_Section* FindSection (const char* pName,Elf32_Word type);

    uint GetSectionIndex (Elf32_Section* pSection);

 protected:
    Elf32_ProgramBlock* AddProgramBlockLoadSectors(word wRelocMin,
						   word wRelocMax,
						   word wRelocProgramHeader);
 
};

ExecutableImage* CreateSimpleElfExecutable();

#endif // !defined(__E_L_F__FILE_H__INCLUDED_)



