/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Error.cpp: implementation of the Error class.
//
//////////////////////////////////////////////////////////////////////

#include "Error.h" // dep_chk_needed   
#include "gen/ErrorMessages.h" // dep_chk_unknown   
#include "ErrorTracker.h" // dep_chk_needed   
#include "Misc.h"

#include <ctype.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

int compare_dword(const void* p1,const void* p2)
{
  if (*((const dword*) p1) < *((const dword*) p2))
    return -1;
  return *((const dword*) p1) > *((const dword*) p2);
}

Error::Error(dword dwErrorId,ERROR_FILE_LINE_FUNC_ARGS)
{
  m_dwErrorId = dwErrorId;
  m_pFile = file;
  m_nLine = line;
  m_pFunction = function;  
  
  m_pContext = errorTracker.GetCurrentContextSetUsed();
}

Error::~Error() 
{

}

#define COMPARE_LESS(object) {if((object)<(rhs.object)) return true;if((object)>(rhs.object)) return true;}
#define COMPARE_LESS2(lhs,rhs) {if((lhs)<(rhs)) return true;if((lhs)>(rhs)) return true;}

bool Error::operator<(const Error& rhs) const
{
  COMPARE_LESS(m_dwErrorId);
  // COMPARE_LESS(m_pFile); // we do not care where the error was generated,
  // COMPARE_LESS(m_nLine); // and besides, m_pFile and m_pFunction can't be
  // COMPARE_LESS(m_pFunction); // compared like this
  
  COMPARE_LESS(m_setRVA);
  
  VectErrorArg::const_iterator i_lhs = m_vectArg.begin();
  VectErrorArg::const_iterator i_rhs = rhs.m_vectArg.begin();
  
  for ( ; i_lhs != m_vectArg.end() && i_rhs != rhs.m_vectArg.begin() ; ++i_lhs, ++i_rhs)
    {
      // now, they may be of different types
      
      const ErrorArgInteger* pIntLhs = 
	dynamic_cast<const ErrorArgInteger*>(*i_lhs);
      const ErrorArgInteger* pIntRhs = 
	dynamic_cast<const ErrorArgInteger*>(*i_rhs);
      
      if (pIntLhs)
	{
	  if (pIntRhs)
	    {
	      COMPARE_LESS2(pIntLhs->m_t,pIntRhs->m_t);
				continue;
	    }
	  return true;
	}
      else if (pIntRhs)
	return false;
      
      const ErrorArgString* pStrLhs = 
	dynamic_cast<const ErrorArgString*>(*i_lhs);
      const ErrorArgString* pStrRhs = 
	dynamic_cast<const ErrorArgString*>(*i_rhs);
      
      assert (pStrLhs && pStrRhs); // otherwise unknown object types
      
      COMPARE_LESS2(pStrLhs->m_t,pStrRhs->m_t);
      /*
	if (pIntLhs)
	{
	if (pIntRhs)
	{
	COMPARE_LESS2(pIntLhs->m_t,pIntRhs->m_t);
	continue;
	}
	return true;
	}
	else if (pIntRhs)
	return false;
	
	assert (false);
      */
    }
  
  // remaining number of elements
  // lhs   rhs
  //  0     0   =  => !<
  //  1     0   >  => !<
  //  0     1   <  =>  <
  
  return i_rhs != m_vectArg.end();
}

ErrorMessageStrings* FindErrorMessageStrings(dword dwErrorId)
{
  ErrorMessageStrings* pStrings = 
    (ErrorMessageStrings*) bsearch(&dwErrorId,errorMessageStrings,
				   NUMBER_OF_NAMED_ERROR_MESSAGES,
				   sizeof (ErrorMessageStrings),compare_dword);
  
  assert (pStrings); // or non-existent error code
  
  return pStrings;
}

void Error::AddErrorInfoRVA(dword dwRVA) 
{
  m_setRVA.insert(dwRVA);
}

void Error::AddErrorArgInteger(dword dw) 
{ 
  m_vectArg.push_back(new ErrorArgInteger(dw)); 
}

void Error::AddErrorArgString(const string& str) 
{ 
  m_vectArg.push_back(new ErrorArgString(str)); 
}

void Error::AddErrorArgString(const char* str) 
{ 
  m_vectArg.push_back(new ErrorArgString(str)); 
}

void Error::AddErrorArgRVA(dword dwRVA)
{ 
  AddErrorInfoRVA(dwRVA); 
  m_vectArg.push_back(new ErrorArgInteger(dwRVA)); 
}

void ErrorArgInteger::Format(const char* fmt,string& append)
{
  appendsprintf(append,fmt,m_t);
}

void ErrorArgString::Format(const char* fmt,string& append)
{
  appendsprintf(append,fmt,m_t.c_str());
}


const char* Error::GetShortMessage () 
{
  ErrorMessageStrings* pStrings = FindErrorMessageStrings(m_dwErrorId);
  
  return pStrings->m_pShort;
}


void Error::FormatShortMessage (string& str)
{
  const char* pShort = GetShortMessage();

  // ok, so everything we need is inside the string
  
  /*
    int nSize = strlen(pFmt) + 10;
    char* pOut = (char*) malloc(nSize);
  */

  uint narg = 0;
  
  
  for (const char* pFmt = pShort; *pFmt; pFmt++)
    {
      if (*pFmt != '%')
        str += *pFmt;
      else
        {
          // we found a format specifier (or a %%)
	  
          pFmt++;
	  
          if (*pFmt == '%')
            continue;
	  
          // so it's a formating string
	  
          char fmt[10];
          fmt[0] = '%';
          uint i = 1;
          while ((*pFmt >= '0' && *pFmt <= '9') || 
		 *pFmt == '.' || *pFmt == '+' || 
		 *pFmt == '-' || *pFmt == ' ' || *pFmt == '#')
            {
              assert(i+1 < sizeof (fmt));
              fmt[i++] = *(pFmt++);
            }
          assert (*pFmt == 'd' || *pFmt == 'i' || *pFmt == 'c' || 
		  *pFmt == 'u' || *pFmt == 'x' || *pFmt == 'X' || 
		  *pFmt == 's');
          assert(i+2 < sizeof (fmt));
          fmt[i] = *(pFmt++);
          fmt[i+1] = '\0';
	  
          // now, format the argument;
	  
          if (isdigit(*pFmt))
	    {
	      assert ((uint) (*pFmt - '1') < m_vectArg.size()); // if you get stuck here, you do not have enough arguments
	      m_vectArg[*pFmt - '1']->Format(fmt,str);	  
	    }
          else
	    {
	      pFmt--; // 
	      assert ((uint) (narg) < m_vectArg.size()); // if you get stuck here, you do not have enough arguments
	      m_vectArg[narg++]->Format(fmt,str);	  
	    }
        }
    }
  
  
  
  
  
  
  //  SetErrorRVA m_setRVA;
  // VectErrorArg m_vectArg;
  
  /*
    int* a = 0;
    
    int* b = &(*a);
  */
  /*
    dword dwErrorId;
    const char* m_pShort;
    const char* m_pShortArgOrder;
    const char* m_pLong;
    };
    
    #define ARG_ORDER_TYPE_INT 0x10
    #define ARG_ORDER_TYPE_PTR 0x20
    #define ARG_ORDER_TYPE_RVA 0x30
    
    #define ARG_ORDER_INT(n) (ARG_ORDER_TYPE_INT | (n))
    #define ARG_ORDER_PTR(n) (ARG_ORDER_TYPE_PTR | (n))
    #define ARG_ORDER_RVA(n) (ARG_ORDER_TYPE_RVA | (n))
    
    
    
    dword m_dwErrorId;
    const char *m_pFile;
    unsigned int m_nLine;
    const char *m_pFunction;  
  */

	/*
printf


#define va_start(ap,v)  ( ap = (va_list)&v + _INTSIZEOF(v) )
#define va_arg(ap,t)    ( *(t *)((ap += _INTSIZEOF(t)) - _INTSIZEOF(t)) )
#define va_end(ap)      ( ap = (va_list)0 )
*/
}

void Error::PrintShortMessage ()
{
  string str;
  
  FormatShortMessage(str);
  
  printf ("%s:%d(%s)\n %s\n",m_pFile,m_nLine,m_pFunction,str.c_str());
  
  if (!m_setRVA.empty())
    {
      printf (" RVA:");
      FOR_EACH_ITER(SetErrorRVA,m_setRVA,rva)
        printf (" 0x%08x",*rva);
      printf ("\n");
    }
}



/*
 public:
  SetErrorRVA m_setRVA;
  VectErrorArg m_vectArg;
*/

/*
struct ErrorMessageStrings
{
  dword dwErrorId;
  const char* m_pShort;
  const char* m_pLong;
};

extern ErrorMessageStrings errorMessageStrings[];
*/

#define ID_ERROR_GENERIC      0x8000

GenericError::GenericError(const char* szShort,ERROR_FILE_LINE_FUNC_ARGS)
  : Error(ID_ERROR_GENERIC,file,line,function)
{
  m_szShort = szShort;
}

GenericError::GenericError(const char* szShort,ERROR_FILE_LINE_FUNC_ARGS,
			   dword arg1)
  : Error(ID_ERROR_GENERIC,file,line,function)
{
  m_szShort = szShort;
  AddErrorArgInteger(arg1);
}

GenericError::GenericError(const char* szShort,ERROR_FILE_LINE_FUNC_ARGS,
			   dword arg1,dword arg2)
  : Error(ID_ERROR_GENERIC,file,line,function)
{
  m_szShort = szShort;
  AddErrorArgInteger(arg1);
  AddErrorArgInteger(arg2);
}

GenericError::GenericError(const char* szShort,ERROR_FILE_LINE_FUNC_ARGS,
			   dword arg1,dword arg2,dword arg3)
  : Error(ID_ERROR_GENERIC,file,line,function)
{
  m_szShort = szShort;
  AddErrorArgInteger(arg1);
  AddErrorArgInteger(arg2);
  AddErrorArgInteger(arg3);
}

GenericError::GenericError(const char* szShort,ERROR_FILE_LINE_FUNC_ARGS,
			   const char* arg1)
  : Error(ID_ERROR_GENERIC,file,line,function)
{
  m_szShort = szShort;
  AddErrorArgString(arg1);
}

GenericError::GenericError(const char* szShort,ERROR_FILE_LINE_FUNC_ARGS,
			   const char* arg1,const char* arg2)
  : Error(ID_ERROR_GENERIC,file,line,function)
{
  m_szShort = szShort;
  AddErrorArgString(arg1);
  AddErrorArgString(arg2);
}

GenericError::GenericError(const char* szShort,ERROR_FILE_LINE_FUNC_ARGS,
			   const string& arg1)
  : Error(ID_ERROR_GENERIC,file,line,function)
{
  m_szShort = szShort;
  AddErrorArgString(arg1);
}

GenericError::GenericError(const char* szShort,ERROR_FILE_LINE_FUNC_ARGS,
			   const string& arg1,const string& arg2)
  : Error(ID_ERROR_GENERIC,file,line,function)
{
  m_szShort = szShort;
  AddErrorArgString(arg1);
  AddErrorArgString(arg2);
}

bool GenericError::operator<(const Error& rhs) const
{
  COMPARE_LESS(m_dwErrorId);
  
  const GenericError& _rhs = dynamic_cast<const GenericError&>(rhs);
  
  COMPARE_LESS2(m_szShort,_rhs.m_szShort);
  
  return Error::operator<(rhs);  
}

const char* GenericError::GetShortMessage () 
{
  return m_szShort; 
}

