/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Error.h: interface for the Error class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__ERROR_H__INCLUDED_)
#define __ERROR_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "GlobalDefs.h" // dep_chk_needed     

#include <string>

using namespace std;

class ErrorArg
{
 protected:
  ErrorArg() { }
 public:
  virtual ~ErrorArg() { }
  
 public:
  virtual void Format(const char* fmt,string& append) = 0;
};

class ErrorArgInteger: public ErrorArg
{
 public:
  ErrorArgInteger(dword t) { m_t = t; }
 public:
  virtual ~ErrorArgInteger() { }

 public:
  virtual void Format(const char* fmt,string& append);

 public:
  dword m_t;
};

class ErrorArgString: public ErrorArg
{
 public:
  ErrorArgString(const string& t) { m_t = t; }
  ErrorArgString(const char* t) { m_t = t; }
 public:
  virtual ~ErrorArgString() { }

 public:
  virtual void Format(const char* fmt,string& append);

 public:
  string m_t;
};
/*
class ErrorArgRVA
{
 public:
  ErrorArgRVA(dword dwRVA) { m_dwRVA = dwRVA; }

 public:
  dword m_dwRVA;
};
*/
struct ErrorMessageStrings
{
  dword dwErrorId;
  const char* m_pShort;
  const char* m_pLong;
};

#define ARG_ORDER_TYPE_INT  0x10
#define ARG_ORDER_TYPE_PTR  0x20
#define ARG_ORDER_TYPE_RVA  0x30

#define ARG_ORDER_TYPE_MASK 0xf0
#define ARG_ORDER_NUM_MASK  0x0f

#define ARG_ORDER_INT(n) (ARG_ORDER_TYPE_INT | (n))
#define ARG_ORDER_PTR(n) (ARG_ORDER_TYPE_PTR | (n))
#define ARG_ORDER_RVA(n) (ARG_ORDER_TYPE_RVA | (n))

extern ErrorMessageStrings errorMessageStrings[];

#include <set>
#include <vector>

using namespace std;

typedef set<dword> SetErrorRVA;
typedef vector<ErrorArg*> VectErrorArg;

class ErrorContext;

#define ERROR_FILE_LINE_FUNC_ARGS const char *file,unsigned int line,const char *function

class Error  
{
 public:
  Error(dword dwErrorId,ERROR_FILE_LINE_FUNC_ARGS);
  
  virtual ~Error();
  
 public:
  dword m_dwErrorId;
  const char *m_pFile;
  unsigned int m_nLine;
  const char *m_pFunction;  
  
  ErrorContext* m_pContext;
  
 public:
  SetErrorRVA m_setRVA;
  VectErrorArg m_vectArg;
  
 public:
  virtual bool operator<(const Error& rhs) const;

 public:
  void AddErrorInfoRVA(dword dwRVA);
  
 public:
  void AddErrorArgInteger(dword dw);
  void AddErrorArgString(const string& str);
  void AddErrorArgString(const char* str);
  void AddErrorArgRVA(dword dwRVA);
  
 public: 
  Error* RVA(dword dwRVA) { AddErrorInfoRVA(dwRVA); return this; }

 public:
  void FormatShortMessage (string& str);
  void PrintShortMessage ();
  
 protected:
  virtual const char* GetShortMessage ();
};

class GenericError : public Error
{
 public:
  GenericError(const char* szShort,ERROR_FILE_LINE_FUNC_ARGS);
  GenericError(const char* szShort,ERROR_FILE_LINE_FUNC_ARGS,dword arg1);
  GenericError(const char* szShort,ERROR_FILE_LINE_FUNC_ARGS,dword arg1,dword arg2);
  GenericError(const char* szShort,ERROR_FILE_LINE_FUNC_ARGS,dword arg1,dword arg2,dword arg3);
  GenericError(const char* szShort,ERROR_FILE_LINE_FUNC_ARGS,const char* arg1);
  GenericError(const char* szShort,ERROR_FILE_LINE_FUNC_ARGS,const char* arg1,const char* arg2);
  GenericError(const char* szShort,ERROR_FILE_LINE_FUNC_ARGS,const string& arg1);
  GenericError(const char* szShort,ERROR_FILE_LINE_FUNC_ARGS,const string& arg1,const string& arg2);

 public:
  const char *m_szShort;

 public:
  virtual bool operator<(const Error& rhs) const;
  
 protected:
  virtual const char* GetShortMessage ();

};

#ifndef __ASSERT_FUNCTION
#define __ASSERT_FUNCTION "<unknown>"
#endif

// TODO: I want to use the variable arg macros of C99, but I could not
// get g++ to use them without issuing warnings or not using c++ at
// all (but without warning?)  so, figure out the proper way to make
// gcc swallow it (it's able to, but I couldn't).  like:
//
// #define GENERIC_ERROR(short_msg,...) (new GenericError(short_msg,__FILE__,__LINE__,__ASSERT_FUNCTION,...))
// 
// so we do not have to have the variants .._ERROR1, .._ERROR2 etc

#define GENERIC_ERROR(short_msg) (new GenericError(short_msg,__FILE__,__LINE__,__ASSERT_FUNCTION))
#define GENERIC_ERROR1(short_msg,arg1) (new GenericError(short_msg,__FILE__,__LINE__,__ASSERT_FUNCTION,(arg1)))
#define GENERIC_ERROR2(short_msg,arg1,arg2) (new GenericError(short_msg,__FILE__,__LINE__,__ASSERT_FUNCTION,(arg1),(arg2)))
#define GENERIC_ERROR3(short_msg,arg1,arg2,arg3) (new GenericError(short_msg,__FILE__,__LINE__,__ASSERT_FUNCTION,(arg1),(arg2),(arg3)))

#define EMIT_GENERIC_ERROR(short_msg) { errorTracker.AddError(GENERIC_ERROR(short_msg)); }
#define EMIT_GENERIC_ERROR1(short_msg,arg1) { errorTracker.AddError(GENERIC_ERROR1(short_msg,(arg1))); }
#define EMIT_GENERIC_ERROR2(short_msg,arg1,arg2) { errorTracker.AddError(GENERIC_ERROR2(short_msg,(arg1),(arg2))); }
#define EMIT_GENERIC_ERROR3(short_msg,arg1,arg2,arg3) { errorTracker.AddError(GENERIC_ERROR3(short_msg,(arg1),(arg2),(arg3))); }

#define VALIDATE(expr,errmsg) {if(!(expr)){throw GENERIC_ERROR(errmsg);}}
#define VALIDATE1(expr,errmsg,arg1) {if(!(expr)){throw GENERIC_ERROR1(errmsg,(arg1));}}
#define VALIDATE2(expr,errmsg,arg1,arg2) {if(!(expr)){throw GENERIC_ERROR2(errmsg,(arg1),(arg2));}}
#define VALIDATE3(expr,errmsg,arg1,arg2,arg3) {if(!(expr)){throw GENERIC_ERROR3(errmsg,(arg1),(arg2),(arg3));}}

#define WARNING(expr,errmsg) {if(!(expr)){EMIT_GENERIC_ERROR(errmsg);}}
#define WARNING1(expr,errmsg,arg1) {if(!(expr)){EMIT_GENERIC_ERROR1(errmsg,(arg1));}}
#define WARNING2(expr,errmsg,arg1,arg2) {if(!(expr)){EMIT_GENERIC_ERROR2(errmsg,(arg1),(arg2));}}
#define WARNING3(expr,errmsg,arg1,arg2,arg3) {if(!(expr)){EMIT_GENERIC_ERROR3(errmsg,(arg1),(arg2),(arg3));}}

/*
bool CompareLessErrorPtrs(const Error* err1,const Error* err2) { return *err1 < *err2; }
*/
#endif // !defined(__ERROR_H__INCLUDED_)



