/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ErrorTracker.cpp: implementation of the ErrorTracker class.
//
//////////////////////////////////////////////////////////////////////

#include "ErrorTracker.h" 
#include "ErrorContext.h" 

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ErrorTracker errorTracker;

//////////////////////////////////////////////////////////////////////

void CurrentErrorContext::SetUsed()
{
  if (m_bUsed)
    return; // already set
  m_bUsed = true;
  if (m_pParent)
    m_pParent->SetUsed();
}

//////////////////////////////////////////////////////////////////////

ErrorTracker::ErrorTracker()
{
  m_pCurrentErrorContext = NULL;
}

ErrorTracker::~ErrorTracker()
{
  assert (!m_pCurrentErrorContext); // all error contexts should have been properly popped
}

void ErrorTracker::PushContext(ErrorContext* pErrorContext)
{
  // after being sent here, we are responsible for deletion
  
  assert (!pErrorContext->m_pParent);
  
  if (m_pCurrentErrorContext)
    pErrorContext->m_pParent = m_pCurrentErrorContext->m_pContext;
  
  m_pCurrentErrorContext = 
    new CurrentErrorContext(pErrorContext,m_pCurrentErrorContext);
}

void ErrorTracker::PopContext()
{
  // as we are may be called on deletion of automatic PushErrorContext
  // destruction, and an function may contain several of these, and
  // I'm not sure if those are guaranteed to be destroyed in order,
  // we'll pop one context as even if we're not called in order, the
  // number of pops will be the same

  CurrentErrorContext* pDelete = m_pCurrentErrorContext;
  
  m_pCurrentErrorContext = pDelete->m_pParent;
  
  if (!pDelete->m_bUsed)
    delete pDelete->m_pContext;
  
  delete pDelete;
}

ErrorContext* ErrorTracker::GetCurrentContextSetUsed()
{
  if (!m_pCurrentErrorContext)
    return NULL;
  else
    {
      m_pCurrentErrorContext->SetUsed();
      return m_pCurrentErrorContext->m_pContext;
    }
}

void ErrorTracker::AddError(Error* pError)
{
  // ok, so add the error to the list of errors and print it
  
  // however, do not add it if we already have one exactly the same
  // (regardless of context)
  
  if (!(m_setError.insert(pError)).second)
    return; // we saw it before

  // so add it to the list
  
  m_vectError.push_back(pError);
  
  /*	
	SetError m_setError; // to avoid dublets
	VectErrorAndContext m_vectErrorAndContext;
  */
  
  
  pError->PrintShortMessage ();
}


