/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Executable.cpp: implementation of the Executable class.
//
//////////////////////////////////////////////////////////////////////

#ifdef WIN32
#include "GlobalDefs.h" 
#endif
#include "Executable.h" 
#include "BranchTarget.h"
#include "JavaDisAsmFile.h"

#include "ImportedEntries.h"
#include "ExportedEntries.h"
#include "ExportedObject.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Executable::Executable()
{
  m_pImportedEntries = NULL;
  m_pExportedEntries = NULL;
  m_pEntryFunction = NULL;
  m_pInitFunction = NULL;
  m_pFiniFunction = NULL;
}

Executable::~Executable()
{
  delete m_pImportedEntries;
  delete m_pExportedEntries;
  delete m_pEntryFunction;
  delete m_pInitFunction;
  delete m_pFiniFunction;
}
	
void Executable::AddCallFunction (CallFunction* pFunction)
{
  //	printf ("Executable::AddCallFunction(%08x)\n",pFunction);
  
  assert (pFunction);
  
  if (m_setCallFunctions.find(pFunction) == m_setCallFunctions.end())
    {
      // not known before
      
      m_setCallFunctions.insert(pFunction);
    }
  /*
    assert (pTarget);
    
    CallFunction* pFunction = new CallFunction ();
    
    pFunction->AddEntryPoint(pTarget);
  */
  
  // m_mapCallTargets.insert(MapCallTargets::value_type(dwTargetVA,pTarget));
}

void Executable::RemoveCallFunction (CallFunction* pFunction)
{
  //	printf ("Executable::RemoveCallFunction(%08x)\n",pFunction);
  
  assert (pFunction);
  assert (m_setCallFunctions.find(pFunction) != m_setCallFunctions.end());
  m_setCallFunctions.erase(pFunction);
}


/*
void HTML_BranchTargets(HtmlPage& page,BranchTarget* pTarget)
{
	MapVATargets mapShow;
	stack<BranchTarget*> stackInvestigate;

	mapShow.insert(MapVATargets::value_type(pTarget->m_dwVA,pTarget));
	stackInvestigate.push(pTarget);

	while (!stackInvestigate.empty())
	{
		BranchTarget* pCheck = stackInvestigate.top();
		stackInvestigate.pop();

		if (pCheck->m_pToNext)
		{
			BranchTarget* pInsert = pCheck->m_pToNext;			

			if (mapShow.find(pInsert->m_dwVA) ==
				mapShow.end())
			{
				mapShow.insert(MapVATargets::value_type(pInsert->m_dwVA,pInsert));
				stackInvestigate.push(pInsert);
			}
		}

		BranchInstr* pBranchInstr = pCheck->GetBranchInstr ();

		if (pBranchInstr && !pBranchInstr->IsCall())
		{
			for (list<BranchTarget*>::iterator i = pBranchInstr->m_listTargets.begin();
				i != pBranchInstr->m_listTargets.end (); i++)
			{
				BranchTarget* pInsert = *i;			

				if (mapShow.find(pInsert->m_dwVA) ==
					mapShow.end())
				{
					mapShow.insert(MapVATargets::value_type(pInsert->m_dwVA,pInsert));
					stackInvestigate.push(pInsert);
				}
			}
		}
	}

	page.Begin(table().border(1));

	for (MapVATargets::iterator i = mapShow.begin();
		i != mapShow.end (); i++)
	{
		(i->second)->HtmlTable(page);	
		
		page << (tr() << td());
	}

	page.End();



//  pTarget->HtmlTable(page);
	
}
*/

#include "UrlFile.h"
#include "UrlFolder.h"
#include "HtmlPage.h"
#include "HtmlTags.h" // dep_chk_unknown   

void Executable::HtmlDisAsmCallFunction (UrlFile* pFile,
					 UrlFile* pFileJava,
					 CallFunction* /* pFunction */,
					 const VectCodeTraces& vectCodeTraces,
					 ReportDisAsmInfo& reportDisAsmInfo)
{
  // generate file
  
  HtmlPage page(pFile);
  
  page.Header();
  
  if (pFileJava)
    {
      assert (reportDisAsmInfo.m_pUrlHtmlJavaViewer);
      page << 
	(p() << 
	 a("Java-enhanced display").href(reportDisAsmInfo.m_pUrlHtmlJavaViewer,
					 "?file=" + reportDisAsmInfo.m_pUrlHtmlJavaViewer->GetRelativePathToItem (pFileJava)));
    }
  
  page.Begin (new pre());
  
  for (VectCodeTraces::const_iterator i = vectCodeTraces.begin();
       i != vectCodeTraces.end (); i++)
    {
      BranchTarget* pTarget = (*i);
      
      while (pTarget)
        {
          pTarget->HtmlPlainDisAsm(page,reportDisAsmInfo);
          pTarget = pTarget->GetNextTarget();
        }
      page << '\n';
    }
  
  page.End ();
}

void Executable::JavaDisAsmCallFunction (UrlFile* pFile,
					 CallFunction* pFunction,
					 const VectCodeTraces& vectCodeTraces,
					 ReportDisAsmInfo& reportDisAsmInfo)
{
  // generate file
  
  JavaDisAsmFile file(pFile);
  
  CallTarget* pEntryPoint = pFunction->GetFirstEntryPoint()/*.m_pTarget*/;
  assert (pEntryPoint);			
  BranchTarget* pTarget = pEntryPoint->m_pTarget;
  assert (pTarget);			
  string strLabel = reportDisAsmInfo.GetLabel(pTarget);
  
  file.Write(JdafHeader("Function: "+strLabel));
  
  for (VectCodeTraces::const_iterator i = vectCodeTraces.begin();
       i != vectCodeTraces.end (); i++)
    {
      pTarget = (*i);
      
      while (pTarget)
        {
          JdafAsmCodeSnippet snippet;
          
          pTarget->JdafDisAsm(snippet,reportDisAsmInfo);
          file.Write(snippet);
          
          pTarget = pTarget->GetNextTarget();
        }
    }
  
  file.WriteEnd();
}

		/*

#include <stdio.h>

typedef map<dword,BranchTarget*> MapVATargets; // a map with dword the offset, as we want them sorted on address

// // #include "HtmlTags.h" // dep_chk_unknown   

// // #include "BranchInstr.h" // dep_chk_unknown   
// // #include "BranchTargetCtrlTransfer.h" // dep_chk_unknown   

void HtmlCallFunction (UrlFile* pFile,CallFunction* pFunction,
					   const MapBranchTargetVA* pMapBT_VA)
{
	MapVATargets mapShow;
	stack<BranchTarget*> stackInvestigate;

	mapShow.insert(MapVATargets::value_type(pMapBT_VA->find(pTarget)->second,pTarget));
	stackInvestigate.push(pTarget);

	while (!stackInvestigate.empty())
	{
		BranchTarget* pCheck = stackInvestigate.top();
		stackInvestigate.pop();

		if (pCheck->m_pToNext)
		{
			BranchTarget* pInsert = pCheck->m_pToNext;			

			if (mapShow.find(pMapBT_VA->find(pInsert)->second) ==
				mapShow.end())
			{
				mapShow.insert(MapVATargets::value_type(pMapBT_VA->find(pInsert)->second,pInsert));
				stackInvestigate.push(pInsert);
			}
		}

		BranchTargetCtrlTransfer* pCtrlTransfer = dynamic_cast<BranchTargetCtrlTransfer*>(pCheck);

		if (pCtrlTransfer)
		{
			BranchInstr* pBranchInstr = pCtrlTransfer->GetInstruction();

			if (!pBranchInstr->IsCall())
			{
				for (VectTargets::iterator i = pBranchInstr->m_vectTargets.begin();
					i != pBranchInstr->m_vectTargets.end (); i++)
				{
					BranchTarget* pInsert = *i;			

					if (mapShow.find(pMapBT_VA->find(pInsert)->second) ==
						mapShow.end())
					{
						mapShow.insert(MapVATargets::value_type(pMapBT_VA->find(pInsert)->second,pInsert));
						stackInvestigate.push(pInsert);
					}
				}
			}
		}
	}

	HtmlPage page(pFile);

	page.Header();

	char fmt1[256];
	sprintf (fmt1,"../java/view.html?start=0x%08lx&offset=0x%08lx",
		pMapBT_VA->find(mapShow.begin()->second)->second,
		pMapBT_VA->find(mapShow.begin()->second)->second);

	page << (p() << a("Java-enhanced display").href(fmt1));

	page.Begin (new pre());

	bool bLastNotFallThrough = false;

	for (MapVATargets::iterator i = mapShow.begin();
		i != mapShow.end (); i++)
	{
		if (bLastNotFallThrough)
			page << '\n';

		(i->second)->HtmlPlainDisAsm(page,pMapBT_VA);
		
		bLastNotFallThrough = !((i->second)->m_pToNext);		
	}

	page.End ();



}
*/
/*
// // #include "JavaDisAsmFile.h" // dep_chk_unknown   

void JdafCallFunction (UrlFile* pFile,CallFunction* pFunction,
					   const MapBranchTargetVA* pMapBT_VA)
{
	MapVATargets mapShow;
	stack<BranchTarget*> stackInvestigate;

	mapShow.insert(MapVATargets::value_type(pMapBT_VA->find(pTarget)->second,pTarget));
	stackInvestigate.push(pTarget);

	while (!stackInvestigate.empty())
	{
		BranchTarget* pCheck = stackInvestigate.top();
		stackInvestigate.pop();

		if (pCheck->m_pToNext)
		{
			BranchTarget* pInsert = pCheck->m_pToNext;			

			if (mapShow.find(pMapBT_VA->find(pInsert)->second) ==
				mapShow.end())
			{
				mapShow.insert(MapVATargets::value_type(pMapBT_VA->find(pInsert)->second,pInsert));
				stackInvestigate.push(pInsert);
			}
		}

		BranchTargetCtrlTransfer* pCtrlTransfer = dynamic_cast<BranchTargetCtrlTransfer*>(pCheck);

		if (pCtrlTransfer)
		{
			BranchInstr* pBranchInstr = pCtrlTransfer->GetInstruction();

			if (!pBranchInstr->IsCall())
			{
				for (VectTargets::iterator i = pBranchInstr->m_vectTargets.begin();
					i != pBranchInstr->m_vectTargets.end (); i++)
				{
					BranchTarget* pInsert = *i;			

					if (mapShow.find(pMapBT_VA->find(pInsert)->second) ==
						mapShow.end())
					{
						mapShow.insert(MapVATargets::value_type(pMapBT_VA->find(pInsert)->second,pInsert));
						stackInvestigate.push(pInsert);
					}
				}
			}
		}
	}

	JavaDisAsmFile file(pFile);

	char fmt[256];
	sprintf (fmt,"0x%08lx",pMapBT_VA->find(mapShow.begin()->second)->second);

	file.Write(JdafHeader(fmt));

//	char fmt1[256];
//	sprintf (fmt1,"../java/view.html?start=0x%08lx&offset=0x%08lx",
//		mapShow.begin()->second->m_dwVA,
//		mapShow.begin()->second->m_dwVA);
//
//	page << (p() << a("Java-enhanced display").href(fmt1));

//	page.Begin (new pre());

//	bool bLastNotFallThrough = false;

	SetBranchTargets setInternalTargets;

	MapVATargets::iterator i;

	for (i = mapShow.begin(); i != mapShow.end (); i++)
		setInternalTargets.insert((i->second));

	for (i = mapShow.begin(); i != mapShow.end (); i++)
	{
//		if (bLastNotFallThrough)
//			page << '\n';

		JdafAsmCodeSnippet snippet;

		(i->second)->JdafDisAsm(snippet,&setInternalTargets,pMapBT_VA);

		file.Write(snippet);

//		(i->second)->HtmlPlainDisAsm(page);
		
//		bLastNotFallThrough = !((i->second)->m_pToNext);		
	}

//	page.End ();

	file.WriteEnd();
}
*/

struct HtmlJavaUrls
{
public:
  HtmlJavaUrls() { m_pFileTargetHtml = NULL; m_pFileTargetJava = NULL; };
  
public:
  UrlFile* m_pFileTargetHtml;
  UrlFile* m_pFileTargetJava;
};

typedef	map<string,HtmlJavaUrls> MapLabelHtmlJavaUrls;

void Executable::HtmlJavaDisAsmCreateUrls (UrlFolder* pParentFolder,
					   ReportDisAsmInfo& reportDisAsmInfo)
{
  UrlFolder* pFolderPlain = pParentFolder->AddFolder("plain");
  UrlFile* pFilePlain = pFolderPlain->AddIndexFile("Plain Disassembly");
  
  HtmlPage pageHtml(pFilePlain);
  pageHtml.Header();
  
  UrlFolder* pFolderJava = pParentFolder->AddFolder("java");
  UrlFile* pFileJava = 
    pFolderJava->AddIndexFile("Java Enhanched Display of Disassembly");
  
  HtmlPage pageJava(pFileJava);
  pageJava.Header();
  
  UrlFile* pFileJavaView = 
    pFolderJava->AddFile("view.html","Java Viewer Disassembly");
  
  HtmlPage pageJavaView(pFileJavaView);
  pageJavaView.HeaderNoBody();
  
  pageJavaView.Begin(body().
                     attrib("style",
                            "\"margin: 0px;margin-bottom: 0px;border-width: 0px;padding: 0px; overflow: hidden\"").
                     attrib("topmargin","0").
                     attrib("leftmargin","0").
                     attrib("marginwidth","0").
                     attrib("marginheight","0").
                     attrib("border","0").
                     attrib("margin","0"));
  
  pageJavaView << (applet().
                   attrib("code","DisAsmViewer.class").
                   attrib("archive","DisAsmViewer.zip").
                   attrib("width","100%").
                   attrib("height","100%") << p("Java has to be enabled for this to work..."));
  
  reportDisAsmInfo.m_pUrlHtmlJavaViewer = pFileJavaView;
  
  MapLabelHtmlJavaUrls mapLabelHtmlJavaUrls;
  
  for (SetCallFunctions::iterator i = m_setCallFunctions.begin();
       i != m_setCallFunctions.end (); i++)
    {
      CallTarget* pEntryPoint = (*i)->GetFirstEntryPoint()/*.m_pTarget*/;
      assert (pEntryPoint);
      
      BranchTarget* pTarget = pEntryPoint->m_pTarget;
      
      if (pTarget)
        {
          HtmlJavaUrls urls;
          
          string strLabel = reportDisAsmInfo.GetLabel(pTarget);
	  
          urls.m_pFileTargetHtml = 
	    pFolderPlain->AddFile(strLabel+".html",
				  strLabel+" - Plain Disassembly");
          reportDisAsmInfo.AddFunctionUrlFile(*i,
					      urls.m_pFileTargetHtml,
					      DIS_ASM_URL_HTML);
          
          urls.m_pFileTargetJava = pFolderJava->AddFile(strLabel+".jdaf");
          reportDisAsmInfo.AddFunctionUrlFile(*i,
					      urls.m_pFileTargetJava,
					      DIS_ASM_URL_JAVA);
          
          mapLabelHtmlJavaUrls.insert(MapLabelHtmlJavaUrls::value_type(strLabel,urls));
        }		
    }
  
  for (MapLabelHtmlJavaUrls::iterator j = mapLabelHtmlJavaUrls.begin();
       j != mapLabelHtmlJavaUrls.end (); j++)
    {
      pageHtml << (a().href(j->second.m_pFileTargetHtml) << 
		   code(j->first)) << br();
      pageJava << (a().href(reportDisAsmInfo.m_pUrlHtmlJavaViewer,
                            "?file=" + 
                            reportDisAsmInfo.m_pUrlHtmlJavaViewer->GetRelativePathToItem (j->second.m_pFileTargetJava)) 
                   << code(j->first)) << br();
    }
}

void Executable::HtmlJavaDisAsm (ReportDisAsmInfo& reportDisAsmInfo)
{
  for (SetCallFunctions::iterator i = m_setCallFunctions.begin();
       i != m_setCallFunctions.end (); i++)
    {
      UrlFile* pFileHtmlTarget = 
	reportDisAsmInfo.GetFunctionUrlFile(*i,DIS_ASM_URL_HTML);
      UrlFile* pFileJavaTarget = 
	reportDisAsmInfo.GetFunctionUrlFile(*i,DIS_ASM_URL_JAVA);
      
      VectCodeTraces vectCodeTraces;
      
      // sort the traces so they are output in a good order
      
      (*i)->GetSortedCodeTraces(vectCodeTraces);
      
      reportDisAsmInfo.SortCodeTraces (vectCodeTraces);
      
      if (pFileHtmlTarget)
        HtmlDisAsmCallFunction (pFileHtmlTarget,pFileJavaTarget,
                                *i,vectCodeTraces,reportDisAsmInfo);
      
      if (pFileJavaTarget)
        JavaDisAsmCallFunction (pFileJavaTarget,
                                *i,vectCodeTraces,reportDisAsmInfo);
    }
}


/*
void Executable::JavaDisAsm (UrlFolder* pFolder,const MapBranchTargetVA* pMapBT_VA)
{
	UrlFolder* pFolderJava = pFolder->AddFolder("java");

	UrlFile* pFileJava = pFolderJava->AddFile("view.html","Java Viewer Disassembly");

	HtmlPage page(pFileJava);

	page.HeaderNoBody();

	page.Begin(body().
		attrib("style","\"margin: 0px;margin-bottom: 0px;border-width: 0px;padding: 0px; overflow: hidden\"").
		attrib("topmargin","0").
		attrib("leftmargin","0").
		attrib("marginwidth","0").
		attrib("marginheight","0").
		attrib("border","0").
		attrib("margin","0"));

	page << (applet().
		attrib("code","DisAsmViewer.class").
		attrib("archive","DisAsmViewer.zip").
		attrib("width","100%").
		attrib("height","100%") << p("Java has to be enabled for this to work..."));

	for (MapCallTargets::iterator i = m_mapCallTargets.begin();
			i != m_mapCallTargets.end (); i++)
	{
		char fmt[64];

		sprintf (fmt,"0x%08lx.jdaf",i->first);
				
		UrlFile* pFileTarget = pFolderJava->AddFile(fmt);
			
		
		//      HTML_BranchTargets (page,i->second);

		JdafCallFunction (pFileTarget,i->second,pMapBT_VA);
	}
}
*/
/*
void Executable::GenerateReportHTML ()
{
  ofstream out("disasm.html");
  
  HtmlPage page;
  
  page.SetOutStream(&out);
  
  page.Header("Disassembly");

  page << a("Executable").href("executable.html");

  char fmt[64];

  for (MapCallTargets::iterator i = m_mapCallTargets.begin();
       i != m_mapCallTargets.end (); i++)
    {
      sprintf (fmt,"0x%08lx",i->first);
      
      page << h1(fmt);

//      HTML_BranchTargets (page,i->second);

    }


  
 // page.Begin(table().border(1));
//
 // page.End ();
  
}
*/


void Executable::InitVariableState ()
{
  FOR_EACH_ITER(SetCallFunctions,m_setCallFunctions,i)
    {
      (*i)->InitCallReturnState ();
    }
}





