/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Executable.h: interface for the Executable class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__EXECUTABLE_H__INCLUDED_)
#define __EXECUTABLE_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "ExportedEntry.h" 
#include "CallFunction.h" 
#include "ReportDisAsmInfo.h" 

#include "DataAnalyseFunction.h"   

class ImportedEntries;
class ExportedEntries;
class ExportedObject;

class UrlFolder; 

#include <map>
#include <vector>

using namespace std ;

typedef set<CallFunction*> SetCallFunctions;

class Executable  
{
 public:
  Executable();
  virtual ~Executable();
  
 protected:
  //	vector<ExportedEntry> m_vectExportedEntries;
  
  //	vector<BranchTarget> m_vectBranchTargets;
  
  // MapCallTargets m_mapCallTargets; // function entries (note: the same code may be pointed to by many entries here, if some functions have multiple entry adresses)
  SetCallFunctions m_setCallFunctions; // function entries 
  
 public:
  ImportedEntries* m_pImportedEntries;
  ExportedEntries* m_pExportedEntries;
  ExportedObject*	 m_pEntryFunction;
  ExportedObject*	 m_pInitFunction; // initialization code
  ExportedObject*	 m_pFiniFunction; // finalization code
  
 public:
  DataAnalyseFunction m_analysis;


 public:
  void AddCallFunction (CallFunction* pFunction);
  void RemoveCallFunction (CallFunction* pFunction);
  
  const SetCallFunctions* GetSetCallFunctions () 
    { return &m_setCallFunctions; }
  
  void InitVariableState ();
  
 public:
  void HtmlJavaDisAsm (ReportDisAsmInfo& reportDisAsmInfo);
  
  void HtmlJavaDisAsmCreateUrls (UrlFolder* pParentFolder,
				 ReportDisAsmInfo& reportDisAsmInfo);
  
 protected:
  void HtmlDisAsmCallFunction (UrlFile* pFile,
                               UrlFile* pFileJava,
                               CallFunction* pFunction,
                               const VectCodeTraces& vectCodeTraces,
                               ReportDisAsmInfo& reportDisAsmInfo);
  void JavaDisAsmCallFunction (UrlFile* pFile,
                               CallFunction* pFunction,
                               const VectCodeTraces& vectCodeTraces,
                               ReportDisAsmInfo& reportDisAsmInfo);
  
  
};

#endif // !defined(__EXECUTABLE_H__INCLUDED_)

