/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ExecutableImage.h: interface for the ExecutableImage class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__EXECUTABLE_IMAGE_H__INCLUDED_)
#define __EXECUTABLE_IMAGE_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "GlobalDefs.h"

#include <string>
 
using namespace std;

#include "UrlFolder.h" // dep_chk_needed     
#include "DestRelocatableSections.h"

class Executable;
class DisAssembler;
class DataPtr;

class ExecutableImage  
{
 public:
  ExecutableImage(const string& strFileName);
  virtual ~ExecutableImage();
  
 public:
  string	m_strFileName;
  
  DestRelocatableSections m_destSections;
  

  Executable*		 m_pExecutable;

 public:
  virtual void ReadFile (int file) = 0; // all actual file reading goes here
  virtual void PostReadFile (); // process things easily dechiphered (not requiring any intelligence)
  virtual bool ProcessFile (UrlFolder* pFolder) = 0; // process instructions
  
 public:
  virtual ExecutableImage* GetWriteImage(const string& strFileName) = 0;
  
  virtual bool PrepareWrite() = 0;
  virtual bool WriteFile();
  
  virtual bool TuneImage();
  
 public:
  virtual void HTMLFileInfo (UrlFolder* pParent) = 0;
  
 protected:
  bool DisAssemble (UrlFolder* pParent);
  
 protected:
  virtual void GetCallTargets (DisAssembler& disasm);
  virtual void GetExportedCallTargets (DisAssembler& disasm);
  virtual DataPtr* GetDisAsmDataPtr (dword dwVirtualAddress) = 0;
};

#endif // !defined(__EXECUTABLE_IMAGE_H__INCLUDED_)
