/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ExecutableWriter.cpp: implementation of the ExecutableWriter class.
//
//////////////////////////////////////////////////////////////////////

#include "ExecutableWriter.h"

#include "DataDestRelocPtr.h"

#include "InstructionEncoder.h"
#include "CallFunction.h"

#include "BranchTargetTraceSingle.h" 
#include "BranchTargetTrace.h" 
#include "BranchTargetCtrlTransfer.h" 

#include "OperandMemory.h"
#include "OperandImmediate.h"

#include "RelocationInfo.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ExecutableWriter::ExecutableWriter()
{
  
}

ExecutableWriter::~ExecutableWriter()
{
  
}

byte get_lowest_bit_set(word w)
{
  byte r = 0;
  
  while (w && !(w & 1))
    {
      r++;
      w >>= 1;
    }
  return r;
}

typedef map<CtrlTransferTargetBase*,dword> MapCtrlTransferTargetBaseOffset;

void ExecutableWriter::Write(WriteRelocDest* pDest)
{
  printf ("ExecutableWriter::Write\n");
  

  // hi, ho. we'll not use the standard relocator to relocate jump and
  // call relative addresses in the executable code
  //
  // 1st: they are not subject to any external relocation, so the
  // information is not needed later.
  //
  // 2nd: they always occur as immediates for the jump/call
  // instructions they are for (and may be only 8 bits/1 byte), so not
  // dword
  //
  // 3rd: because they may be shortened to 1 byte (as above) if the
  // jump is small enough, we'll anyway have to figure out how to
  // write the code anyway from start to use this as much as possible
  // (i.e. keeping jumps and calls short, especially often used ones,
  // and as it affect alignment and several other things, this should
  // already be decided when we're called to do the writing.
  //
  // For now, we'll make a first run, decide all addresses, and then
  // make the production run. This is not perhaps efficient, but this
  // code is gonna change anyway when we're writing the
  // efficiencyising machine code writer this is done by noting where
  // all targets end up.. we're writing continously, not caring about
  // alignment at all...  so this will probably just make just about
  // any code worse..

  InstructionEncoder encoder;
  
  const SetCallFunctions* pSetCallFunctions = 
    m_pExecutable->GetSetCallFunctions ();
  
  dword dwSizeUsed = 0;
  
  MapCtrlTransferTargetBaseOffset mapCtrlTransferTargetBaseOffset;
  
  FOR_EACH_CONST_ITER(SetCallFunctions,*pSetCallFunctions,func)
    {
      VectCodeTraces vectCodeTraces;
      
      (*func)->GetCodeTraces(vectCodeTraces);
      
      FOR_EACH_CONST_ITER(VectCodeTraces,vectCodeTraces,trace)
	{
	  BranchTarget* pTarget = (*trace);
	  
	  while (pTarget)
	    {
	      CtrlTransferTarget* pSource = pTarget->GetCtrlTransferTarget();

	      while (pSource)
		{
		  printf ("pSource: %p (%p) @ %d\n",
			  (void*) pSource,(void*) pTarget,dwSizeUsed);
		  mapCtrlTransferTargetBaseOffset.insert(MapCtrlTransferTargetBaseOffset::value_type(pSource,dwSizeUsed));
		  pSource = pSource->m_pMoreSource;
		}
	      
	      // printf ("pTarget: %08x \n",pTarget);
	      
	      BranchTargetTraceSingle* pTraceSingle;
	      BranchTargetTrace* pTraceMultiple;
	      BranchTargetCtrlTransfer* pTraceCtrlTransfer;
	      
	      Instruction* pInstr = NULL;
	      
	      if ((pTraceSingle = 
		   dynamic_cast<BranchTargetTraceSingle*>(pTarget)) != NULL)
		pInstr = pTraceSingle->GetInstruction();
	      else if ((pTraceMultiple = 
			dynamic_cast<BranchTargetTrace*>(pTarget)) != NULL)
		assert (false);
	      else if ((pTraceCtrlTransfer = 
			dynamic_cast<BranchTargetCtrlTransfer*>(pTarget)) != NULL)
		{
		  BranchInstr* pBranchInstr = pTraceCtrlTransfer->GetBranchInstr();
		  
		  // if it is an relative jump, make the address size
		  // such that 4 bytes is required, so that we can fit
		  // any value that may be needed
		  
		  Operand* pOp = pBranchInstr->m_pTargetOp;
		  
		  if (pBranchInstr->IsTargetRelative())
		    {
		      OperandImmediate* pOpImm = 
			dynamic_cast<OperandImmediate*>(pOp);
		      
		      assert (pOpImm);
		      
		      // if we already know the target, and it is
		      // close enough, we can use 1-byte encoding
		      
		      CtrlTransferTargetBase* pRelTarget = pBranchInstr->m_vectTargets.front();
		      
		      printf ("pRelTarget: %p @ %d\n",(void*) pRelTarget,dwSizeUsed);
		      
		      MapCtrlTransferTargetBaseOffset::iterator target_offset = mapCtrlTransferTargetBaseOffset.find(pRelTarget);
		      
		      if (pBranchInstr->IsJump() &&
			  target_offset != mapCtrlTransferTargetBaseOffset.end() &&
			  dwSizeUsed - target_offset->second < 128-2) // relative encodings are 2 bytes long, when offset si 1 byte
			{
			  printf ("pRelTarget: %p @ %d -> %d\n",(void*) pRelTarget,dwSizeUsed,target_offset->second);
			  
			  pOpImm->m_dwImm = target_offset->second - (dwSizeUsed + 2);
			  assert ((pOpImm->m_dwImm & 0xffffff00) == 0xffffff00);
			  pOpImm->m_bImmBytes = 1;
			}
		      else
			{
			  pOpImm->m_dwImm = 0xcdcdcdcd;
			  pOpImm->m_bImmBytes = 4;
			}
		    }
		  pInstr = pBranchInstr;
		}
	      
	      assert (pInstr);
	      
	      //encoder.Encode(pInstr,*(pDest->m_pDest));
	      word wLengths = encoder.EncodeLengths(pInstr);

	      dword wLength = get_lowest_bit_set(wLengths)+1;

	      dwSizeUsed += wLength;
	      
	      printf ("Encoded: %04x\n",wLengths,wLength);

	      pTarget = pTarget->GetNextTarget();
	    }
	}
    }









  // production run
  
  FOR_EACH_CONST_ITER(SetCallFunctions,*pSetCallFunctions,func)
    {
      VectCodeTraces vectCodeTraces;
      
      (*func)->GetCodeTraces(vectCodeTraces);
      
      FOR_EACH_CONST_ITER(VectCodeTraces,vectCodeTraces,trace)
	{
	  BranchTarget* pTarget = (*trace);
	  
	  while (pTarget)
	    {
	      // printf ("pTarget: %08x \n",pTarget);
	      
	      pDest->AssureDestMemory(16); // 16 bytes maximum per instruction
	      
	      BranchTargetTraceSingle* pTraceSingle;
	      BranchTargetTrace* pTraceMultiple;
	      BranchTargetCtrlTransfer* pTraceCtrlTransfer;
	      
	      Instruction* pInstr = NULL;
	      
	      bool bHasImportedEntryFixup = false;
	      bool bHasRelativeOffset = false;
	      dword dwRelativeOffsetTo = 0;
	      
	      if ((pTraceSingle = dynamic_cast<BranchTargetTraceSingle*>(pTarget)) != NULL)
		{
		  pInstr = pTraceSingle->GetInstruction();
		  //					DetermineStateChange (&pIncoming,outgoingState,
		  //						pTraceSingle->GetInstruction());
		}
	      else if ((pTraceMultiple = dynamic_cast<BranchTargetTrace*>(pTarget)) != NULL)
		{
		  assert (false);
		}
	      else if ((pTraceCtrlTransfer = dynamic_cast<BranchTargetCtrlTransfer*>(pTarget)) != NULL)
		{
		  //					DetermineStateChangeCtrlTransfer (&pIncoming,outgoingState,
		  //						pTraceCtrlTransfer->GetBranchInstr());
		  
		  // Here we really cheat, we brutally assume it's a pointer into
		  // the inport table and fixes it to point to the new position
		  
		  BranchInstr* pBranchInstr = pTraceCtrlTransfer->GetBranchInstr();
		  
		  Operand* pOp = pBranchInstr->m_pTargetOp;
		  /*
		  OperandMemory* pOpMem;
		  
		  if ((pOpMem = dynamic_cast<OperandMemory*>(pOp)))
		    {
		      assert (peOptHeader.dwImageBase >= 0x80); // so the address won't be shortened by the encoder and therefore not relocatable
		      
		      MapMovedImportAddressTableRVA::iterator fixup = mapMovedImportAddressTableRVA.find(pOpMem->m_dwDisp-peOptHeader.dwImageBase);
		      
		      assert (fixup != mapMovedImportAddressTableRVA.end());
		      
		      pOpMem->m_dwDisp = fixup->second+peOptHeader.dwImageBase;
		      
		      bHasImportedEntryFixup = true;
		    }
		  */

		  if (pBranchInstr->IsTargetRelative())
		    {
		      OperandImmediate* pOpImm = dynamic_cast<OperandImmediate*>(pOp);

		      assert (pOpImm);

		      // we can't set the actual value here, because we do not know where this instruction will end
		      // (and that is where it's relative from)

		      if (pOpImm->m_bImmBytes == 4) // or it's 1 byte, and then it's already calculated
			{
			  // can only have one target
			  assert (pBranchInstr->m_vectTargets.size() == 1); // could be zero if target wasn't successfully decoded,
			  // but then, we should not be here
			  
			  CtrlTransferTargetBase* pRelTarget = pBranchInstr->m_vectTargets.front();
			  
			  MapCtrlTransferTargetBaseOffset::iterator target_offset = mapCtrlTransferTargetBaseOffset.find(pRelTarget);
			  
			  assert (target_offset != mapCtrlTransferTargetBaseOffset.end()); // or will not write target...
			  
			  printf ("pRelTarget: %p @ %d -> %d\n",(void*) pRelTarget,dwSizeUsed,target_offset->second);

			  dwRelativeOffsetTo = target_offset->second;
			  bHasRelativeOffset = true;
			}
		    }

		  pInstr = pBranchInstr;
		}
	      
	      assert (pInstr);
	      
	      word wDispImmLocs;
	      
	      encoder.Encode(pInstr,*(pDest->m_pDest),&wDispImmLocs);
	      
	      printf ("Encoded: %04x\n",wDispImmLocs);

	      if (bHasImportedEntryFixup)
		{
		  assert (DIL_DISP_LEN(wDispImmLocs) == 4);
		  
		  // there is a pointer to import table, somewhere,
		  // at offset DIL_DISP_LOC(wDispImmLocs)
		  // an total length was DIL_LENGTH(wDispImmLocs)
		  
		  pDest->m_pDest->InsertRelocation(pDest->m_pDest->GetOffset()-
						   DIL_LENGTH(wDispImmLocs)+
						   DIL_DISP_LOC(wDispImmLocs),RELATIVE_IMAGE_IMPORT_DIRECTORY);
		  
		}
	      if (bHasRelativeOffset)
		{
		  assert (DIL_IMM_LEN(wDispImmLocs) == 4);
		  pDest->m_pDest->WriteDWordOffset(pDest->m_pDest->GetOffset()-
						   DIL_LENGTH(wDispImmLocs)+
						   DIL_IMM_LOC(wDispImmLocs),dwRelativeOffsetTo - pDest->m_pDest->GetOffset());
		}
	      
	      //				pTarget->HtmlPlainDisAsm(page,reportDisAsmInfo);
	      pTarget = pTarget->GetNextTarget();
	    }
	}
    }


}
