/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// FileMap.cpp: implementation of the FileMap class.
//
//////////////////////////////////////////////////////////////////////

#include "FileMap.h" 
#include "Error.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

FileMap::FileMap()
{
  m_dwLength = 0;
}

FileMap::~FileMap()
{
  
}

void FileMap::SetFileLength(dword dwLength)
{
  assert (dwLength >= m_dwLength); // or we have to clean the map (actually, we will usually only set once)
  m_dwLength = dwLength;
}

void FileMap::AddAreaType(dword dwOffset,dword dwSize,FileMapItem* pItem)
{
  assert (dwOffset < m_dwLength);
  assert (dwOffset + dwSize <= m_dwLength);
    
  // this will throw error if the existing objects do not like the new
  // one, or the new one can't stand the existing ones
  
  ValidateCoExistence(dwOffset,dwSize,pItem);
  
  m_mapFileMapInfo.insert (MapFileMapInfo::value_type(dwOffset,
						      FileMapInfo(dwSize,pItem)));
}

void FileMap::ValidateCoExistence(dword dwOffset,dword dwSize,FileMapItem* pItem)
{
  if (dwOffset >= m_dwLength ||
      dwOffset + dwSize >= m_dwLength)
    return; // outside known length
  
  // there must be nothing in the range specified
  // we have to investigate all map entries until we have passed our range

  dword dwEnd = dwOffset + dwSize;
  
  FOR_EACH_CONST_ITER(MapFileMapInfo,m_mapFileMapInfo,info)
    {
      if (info->first >= dwEnd)
          break; // old ones are past the new one
      
      if (info->first + info->second.m_dwSize > dwOffset)
	{
	  // overlap, so see if it's ok

	  pItem->DoValidateCoExistence(info->second.m_pItem);
	  
	  
	  
	  // nothing thrown, so they mutually agree, continue
	}
    }
  // ok, no clashes
}

bool FileMap::IsAreaUnused(dword dwOffset,dword dwSize)
{
  if (dwOffset >= m_dwLength)
    return false;
  
  // there must be nothing in the range specified
  // we have to investigate all map entries until we have passed our range
  
  dword dwEnd = dwOffset + dwSize;
  
  FOR_EACH_CONST_ITER(MapFileMapInfo,m_mapFileMapInfo,info)
    {
      if (info->first > dwEnd)
        break;
      
      if (info->first + info->second.m_dwSize > dwOffset)
	return false; // overlap
    }
  return true;
}
  

