/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// FileMapItem.cpp: implementation of the FileMapItem class.
//
//////////////////////////////////////////////////////////////////////

#include "FileMapItem.h"
#include "Error.h"
#include "ErrorTracker.h"


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

FileMapItem::FileMapItem()
{

}

FileMapItem::~FileMapItem()
{

}

void FileMapItem::DoValidateCoExistence (FileMapItem* pOtherItem)
{
  try
    {
      // to avoid to have to make them accept each other mutually,
      // it's enough if one says it's ok by returning true,
      // then we accept it.
      // But we still let both get the chance to throw something
      // if they like
      
      if ((ValidateCoExistence(pOtherItem) |
	   pOtherItem->ValidateCoExistence(this)))
	return; // none threw anything and at least one said ok (return true)
    }
  CATCH_EMIT_ERROR;
  
  // so if we're here, there was some trouble,
    // throw an error, and tell what kind of objects could not live together
  
  throw GENERIC_ERROR2("Incompatible objects (%s1 and %s2) use same bytes of file.",GetDescription(),pOtherItem->GetDescription());
}

/*
  bool FileMapItem::ValidateCoExistence (FileMapItem* pOtherItem)
  {
  // we and the other object share bytes in the executable file
  // is that ok?
  
  // default is that we can coexist with nothing
  
  return false; // meaning we have not agreed to this
  }
*/
//////////////////////////////////////////////////////////////////////

GenericFileMapItem::GenericFileMapItem(word wType)
{
  m_wType = wType;
}

bool GenericFileMapItem::ValidateCoExistence (FileMapItem* pOtherItem)
{
  GenericFileMapItem* pGenericOtherItem;
  
  if ((pGenericOtherItem = dynamic_cast<GenericFileMapItem*>(pOtherItem)))
    {
      // let's see if we like that guy
      
      
      
    }
  
  return false;
}

string GenericFileMapItem::GetDescription()
{
  switch (m_wType)
    {
    case GFMI_ELF_IDENT: return "ELF ident";
    case GFMI_ELF_HEADER: return "ELF header";
    case GFMI_ELF_PROGRAM_HEADER: return "ELF program header";
    case GFMI_ELF_SECTION_HEADER: return "ELF section header";
    }
  assert (false);
  return "<unknown>";
}


