/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// FlagsState.h: interface for the FlagsState class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__FLAGS_STATE_H__INCLUDED_)
#define __FLAGS_STATE_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

class Variable; 
class VariableOp; 
#include "StatusFlagsAffected.h" 

// the FlagsState contain information on what we know about the flags
// in the FLAGS register it is owned by an RegisterState object We
// know both any present state (0/1), and perhaps how it came to be
// (e.g. which variables were tested against each other) Please note
// that if an flag is neither known or set as undefined, it's value
// may be known by the flags state in the parent RegisterState if
// present

// FlagsStateSource keep track of what set an particular flag(s) this
// information is needed when we are to constrain the values of
// variables when passing an conditional jump (the two possible
// branches) will have disjoint variable values

// so we need some things: what kind of operation was performed, and
// what variables might be affected: the source operands and the
// destination operand

class FlagsStateSource
{
 public:
  FlagsStateSource();
  virtual ~FlagsStateSource();
  
 public:
  word m_wFlags;
  
  VariableOp* m_pOperation;
  Variable*   m_pResult;
};

class FlagsState
{
 public:
  FlagsState();
  virtual ~FlagsState();
  
 public:
  FlagsStateSource* m_pSource;
  
 public:
  // virtual bool GetCC(word wFlag,bool* pIsTrue,bool *pAffectedOrUndefined) = 0; // returns true if definate value put in pIsTrue
  virtual bool GetFlagCode(word wFlag) const = 0;
  virtual bool IsAffectedOrUndefined(word wMask) const = 0;
  
 public:
  virtual string FlagsString() const { return "FSTATE(error?)"; }
};

class FlagsStateAffected : public FlagsState, StatusFlagsAffected
{
 public:
  FlagsStateAffected();
  FlagsStateAffected(const StatusFlagsAffected& src);
  virtual ~FlagsStateAffected();

 public:
  // virtual bool GetCC(word wFlag,bool* pIsTrue,bool *pAffectedOrUndefined); 
  virtual bool GetFlagCode(word wFlag) const { return StatusFlagsAffected::GetFlagCode(wFlag); }
  virtual bool IsAffectedOrUndefined(word wMask) const { return StatusFlagsAffected::IsAffectedOrUndefined(wMask); }
  
 public:
  virtual string FlagsString() const;
};

class FlagsStateUnknown : public FlagsState
{
 public:
  FlagsStateUnknown();
  virtual ~FlagsStateUnknown();

 public:
  // virtual bool GetCC(word wFlag,bool* pIsTrue,bool *pAffectedOrUndefined); 
  virtual bool GetFlagCode(word wFlag) const { return FLAG_UNDEFINED; }
  virtual bool IsAffectedOrUndefined(word wMask) const { return true; }
  
 public:
  virtual string FlagsString() const { return "unknown"; }
};


class FlagsStateUnaffected : public FlagsState
{
 public:
  FlagsStateUnaffected();
  virtual ~FlagsStateUnaffected();
  
 public:
  // virtual bool GetCC(word wFlag,bool* pIsTrue,bool *pAffectedOrUndefined); 
  virtual bool GetFlagCode(word wFlag) const { return FLAG_UNAFFECTED; }
  virtual bool IsAffectedOrUndefined(word wMask) const { return false; }
  
 public:
  virtual string FlagsString() const { return "unaffected"; }
};

#endif // !defined(__FLAGS_STATE_H__INCLUDED_)
