#!/usr/bin/perl

#  exmhcg - executable program tuner  
#  Copyright (C) 2003  Hkan Johansson
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

use GenCode;

$outfile_cpp = shift @ARGV;
$outfile_h = shift @ARGV;

@infiles = @ARGV;

print "Generating $outfile_cpp, $outfile_h from $infile\n";

open(OUTFILE_CPP,">$outfile_cpp") || die "Can't open $outfile_cpp";
print OUTFILE_CPP GeneratedHeader($outfile_cpp,'Structure dump for @infiles.');

open(OUTFILE_H,">$outfile_h") || die "Can't open $outfile_h";
print OUTFILE_H GeneratedHeaderH($outfile_h,'Structure dump for @infiles.');

print OUTFILE_H   "#include <stddef.h>\n\n";
print OUTFILE_H   "class HtmlPage;\n\n";
print OUTFILE_H   "class HtmlObject;\n\n";
print OUTFILE_CPP "#include \"../$outfile_h\"\n";
print OUTFILE_CPP "#include \"HtmlPage.h\"\n";
print OUTFILE_CPP "#include \"HtmlTags.h\"\n";
print OUTFILE_CPP "#include \"../Misc.h\"\n";

%namedids = ();

foreach $infile (@infiles)
{
    $cont = ReadFile ($infile);
    $included = 0;
    
    # So find ourselves all dumpable structures

    $namedids = $cont;
    
    while ($namedids =~ s/\/\/\/\/\s+(\/\/\s.+?)\/\/\/\///s)
    {
        $nid = $1;
        @strmem = ();

        while ($nid =~ s/^\/\/\s+(\w+\.\w+)\s+//s)
        {
            push @strmem,$1;
        }

        # print "@strmem : $nid\n";

        foreach $strmem (@strmem)
        {
            $namedids{$strmem} = $nid;
        }	
    }

    $structs = $cont;

    while ($structs =~ s/^.*?\sstruct\s+(\w+)\s+\/\/\s+Dumpable(.+?)\{(.+?)\}//s )
    {
        # $structs =~ s/^.*?\s(struct\s)/$1/s; # \

        # print "$1 : \n$2\n";

        $structname = $1;
        $arguments = $2;
        $members = $3;

        if (!$included)
        {
            $included = 1;
            print OUTFILE_CPP "#include \"..\/$infile\"\n";
        }

        # figure out any arguments

        @args = ();

        $hargs = "";
        $cppargs = "";

        if ($arguments =~ /\((.+?)\)/)
        {
            @args = split /,/,$1;
            foreach $arg (@args)
            {
                $hargs .= ",HtmlObject* $arg = NULL";
                $cppargs .= ",HtmlObject* $arg";
            }
            printf "Args: $hargs $cppargs\n";
        }

        #####################################################
        # for dumping the structure as a table
        
        print OUTFILE_H   "struct $structname;\n";
        print OUTFILE_H   "void DumpStruct$structname(const $structname\& st,HtmlPage& page$hargs);\n\n";
        print OUTFILE_CPP "void DumpStruct$structname(const $structname\& st,HtmlPage& page$cppargs)\n{\n";
        
        #print OUTFILE_CPP "  if (pObjTitle) { page << (h1() << pObjTitle); }\n\n";
        #print OUTFILE_CPP "  page << h1(\"$structname\");\n\n";
        print OUTFILE_CPP "  page.Begin(table().border(0));\n\n";

        # find structure members (they are oneliners for this to work)

        @members = split /\n/,$members;

        foreach $member (@members)
        {
            if ($member =~ /\s*(\w+)\s+([\w\.]+)\s*;\s*\/\/\s*(.+?)\s*$/)
            {
                $memtype = $1;
                $memname = $2;
                $meminfo = $3;

                # print ">>$1 $2 $3\n";

                $nidname = "$structname.$memname";

                if ($namedids{$nidname})
                {
                    NamedIdsOutput ($meminfo,$memname,$namedids{$nidname},$memtype,1);
                }
                else { ValueOutput($meminfo,$memname,$memtype,1); }
            }
        }

        print OUTFILE_CPP "\n  page.End();\n";
        print OUTFILE_CPP "};\n\n";

        #####################################################
        # for dumping the structure as a table
        
        print OUTFILE_H   "struct $structname;\n";
        print OUTFILE_H   "void DumpStructTable$structname(const $structname\& st,HtmlPage& page$hargs);\n\n";
        print OUTFILE_CPP "void DumpStructTable$structname(const $structname\& st,HtmlPage& page$cppargs)\n{\n";
        
        #print OUTFILE_CPP "  page << h1(\"$structname\");\n\n";

        # find structure members (they are oneliners for this to work)

        @members = split /\n/,$members;

        foreach $member (@members)
        {
            if ($member =~ /\s*(\w+)\s+([\w\.]+)\s*;\s*\/\/\s*(.+?)\s*$/)
            {
                $memtype = $1;
                $memname = $2;
                $meminfo = $3;

                # print ">>$1 $2 $3\n";

                $nidname = "$structname.$memname";

                if ($namedids{$nidname})
                {
                    NamedIdsOutput ($meminfo,$memname,$namedids{$nidname},$memtype,0);
                }
                else { ValueOutput($meminfo,$memname,$memtype,0); }
            }
        }

        print OUTFILE_CPP "};\n\n";

        #####################################################
        # for dumping an array of the structure as a table

        print OUTFILE_H   "void DumpStructTableHeader$structname(HtmlPage& page);\n\n";
        print OUTFILE_CPP "void DumpStructTableHeader$structname(HtmlPage& page)\n{\n";
        
        
        # find structure members (they are oneliners for this to work)

        @members = split /\n/,$members;

        foreach $member (@members)
        {
            if ($member =~ /\s*(\w+)\s+([\w\.]+)\s*;\s*\/\/\s*(.+?)\s*$/)
            {
                $memtype = $1;
                $memname = $2;
                $meminfo = $3;

#               print ">>$1 $2 $3\n";

				$attrib = "";;

				if ($meminfo =~ s/\s+\(fmt:(.+?)\)//)
				{
					my $fmt = $1;

#					print "FORMAT: <<$meminfo>> <<$fmt>>\n";

					if ($fmt =~ /\+(.+)/)
					{
						$attrib = ".colspan(2)";
					}
				}


                print OUTFILE_CPP "  page << th(\"$meminfo\")$attrib;\n";
            }
        }

        print OUTFILE_CPP "};\n\n";

        #####################################################

    }
    
    

}

print OUTFILE_CPP <<"EndOfText";

//////////////////////////////////////////////////////////////////////
EndOfText

close(OUTFILE_CPP);

print OUTFILE_H GeneratedFooterH();
close(OUTFILE_H );


sub ValueOutput($meminfo,$memname,$memtype)
{
    my $meminfo = shift;
    my $memname = shift;
    my $memtype = shift;
    my $tablerow = shift;

    my $extrahtml = ();

    if ($meminfo =~ s/\s+\(fmt:(.+?)\)//)
    {
                my $fmt = $1;
                print "fmt:   $fmt\n";

                if ($fmt =~ /\+(.+)/)
                { 
                        $extrahtml = $1;	
                        print "efmt:   $fmt $extrahtml\n";
                }
    }

    $formater = "strsprintf(\"0x%x\",st.$memname)";

    if ($extrahtml)
    {	
                # print OUTFILE_CPP "    const HtmlTagPtr& html_tag_ptr = (td($formater) << ' ' << $extrahtml);\n";
                $tdformater = "html_tag_ptr";

                if ($tablerow) 
                { 
                        print OUTFILE_CPP "    page.Begin(tr());\n";
                        print OUTFILE_CPP "    page << td(\"$meminfo\");\n";
                        print OUTFILE_CPP "    page << td($formater);\n";
                        print OUTFILE_CPP "    page << (td() << $extrahtml);\n";
                        print OUTFILE_CPP "    page.End();\n";
                }
                else     
                {
                        print OUTFILE_CPP "    page << td($formater);\n"; 
                        print OUTFILE_CPP "    page << (td() << $extrahtml);\n"; 
                }
    }
    else
    {
                if ($tablerow) { print OUTFILE_CPP "    page << (tr() << td(\"$meminfo\") << td($formater));\n"; }
                else           { print OUTFILE_CPP "    page << td($formater);\n"; }
    }
}

sub NamedIdsOutput($meminfo,$memname,$nid,$memtype)
{
    my $meminfo = shift;
    my $memname = shift;
    my $nid = shift;
    my $memtype = shift;
    my $tablerow = shift;

    my @nid = split /\n/,$nid;

    $formater = "strsprintf(\"0x%x\",varval)";

        print OUTFILE_CPP "    {\n";
        print OUTFILE_CPP "      $memtype varval = st.$memname;\n";

    if ($tablerow) 
    { 
        print OUTFILE_CPP "    page.Begin(tr());\n";
        print OUTFILE_CPP "      page << td(\"$meminfo\") << td($formater);\n";
        print OUTFILE_CPP "      page.Begin(td());\n";
                $setprinted = "";
                $notprinted = "";
    }
    else
    {
        print OUTFILE_CPP "      page.Begin(td());\n";
        print OUTFILE_CPP "      bool bAnyPrinted = false;\n";
                $setprinted = "bAnyPrinted = true; ";
                $notprinted = " || !bAnyPrinted";
    }

    # loop through the named ids

    my $mask = ") ==";
    my $else = ""; my $setelse = "else ";

    foreach $nid (@nid)
    {
        if ($nid =~ /\#define\s+(\w+)\s+(.+?)\s*\/\/\s*(.+)\s*/)
        {
            # print "$1 -> $3\n"; 

            print OUTFILE_CPP "        ${else}if ((st.$memname$mask $1) { page << \"$3 \"; varval -= $1; $setprinted}\n";
            $else = $setelse;
        }	
        if ($nid =~ /\/\/\s*Mask:\s*(\w*)/)
        {
            my $m = $1;

            if ($m)
            {
                $mask = " & $m) ==";
                $setelse = "else ";
            }
            else
            {
                $mask = ") &";
                $setelse = "";
            }

            print "Mask:: $mask\n";
            $else = "";
        }
    }

    print OUTFILE_CPP "        if (varval${notprinted}) { page << ' ' << $formater; }\n";

    print OUTFILE_CPP "      page.End();\n";
    if ($tablerow) 
    { 
        print OUTFILE_CPP "    page.End();\n";
    }
        print OUTFILE_CPP "    }\n";
}

