#!/usr/bin/perl

#  exmhcg - executable program tuner  
#  Copyright (C) 2003  Hkan Johansson
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

use GenCode;

$outfile_cpp = shift @ARGV;
$outfile_h = shift @ARGV;

@infiles = @ARGV;

print "Generating $outfile_cpp, $outfile_h from @infiles\n";

open(OUTFILE_CPP,">$outfile_cpp") || die "Can't open $outfile_cpp";

print OUTFILE_CPP GeneratedHeader($outfile_cpp,'Message strings for error messages.');

open(OUTFILE_H,">$outfile_h") || die "Can't open $outfile_h";

print OUTFILE_H GeneratedHeaderH($outfile_h,'Message strings for error messages.');

    $errorcnt = 1;

@errorids = ();

foreach $infile (@infiles)
{
    open(INFILE,"<$infile") || die "Can't open $infile";

    $infile_out_h = $infile;
    $infile_out_h =~ s/\.txt/\.h/;
    $infile_out_h_in_gen_dir = $infile_out_h;
    $infile_out_h = "gen/".$infile_out_h;

    # print "$infile_out_h_in_gen_dir";
    
print "Generating $infile_out_h from $infile\n";

open(INFILE_OUT_H,">$infile_out_h") || die "Can't open $infile_out_h";

$FULLNAME = $infile_out_h;
$FULLNAME =~ s/.*\///g;
$FULLNAME =~ s/\./\_/g;
$FULLNAME =~ s/([A-Z])/\_$1/g;
$FULLNAME = uc($FULLNAME);

print INFILE_OUT_H GeneratedHeaderH($infile_out_h,'Ids and macros for error messages.');

print INFILE_OUT_H <<"EndOfText"; 

#include <assert.h>
#include "../ErrorTracker.h"

EndOfText

print OUTFILE_CPP "#include \"$infile_out_h_in_gen_dir\"\n";

#    $lastfound = 0;
#    
#    while (<INFILE>)
#    { 
#	$readline = $_;
#	
#	if ($readline =~ /$regexp/)
#	{
#	    push @groups , $1;
#	    push @groups , ($count);
#	    $lastfound = 1;
#	    if (!defined($counter))
#	    { $count++; }
#	}
#	elsif ($lastfound && $readline =~ /$counter/)
#	{
#	    $count++;
#	}
#    }

    $cont = "";

    while (<INFILE>)
    {
        $cont .= $_;
    }

    $cont =~ s/\r\n/\n/g;

    while ($cont)
    {
        # on first line(s) are the identifier, with possibly some arguments,
        # [empty line]
        # on second line(s) are a short error description, which may take parameters, such as %RVA
        # [empty line]
        # on third lines are a long description, not taking parameters, as it is a general
        # description of the error condition
        # [empty line]
        # ... some more errors

        $cont =~ s/^\s*//g;
        if ($cont =~ s/((ERROR|WARN)_.*?)\n\s*\n(.*?)\n\s*\n(.*?)\n\s*\n//s ||
            $cont =~ s/((ERROR|WARN)_.*?)\n\s*\n(.*?)\n\s*\n(.*?)\s*$//s)
#	if ($cont =~ /((ERROR|WARN)_.*?)\n\s*\n(.*?)\n\s*\n(.*?)\n\s*\n/)
	{
	    # $cont =~ s/((ERROR|WARN)_.*?)\n\s*\n(.*?)\n\s*\n(.*?)(\n\s*\n|$)//s)

	    $error = $1;
	    $short = $3;
	    $long = $4;

	    # print "$short\n";

	    $short =~ s/\n/ /g; $short =~ s/  +/ /g;
	    $long =~ s/\n/ /g; $long =~ s/  +/ /g;

	    $short =~ s/%RVA([^0-9])/%RVA1$1/g;
	    
	    # print "$short\n";

	    ($args,$short) = FigureOutArgs($short); # (...,$argorder) = 

	    print INFILE_OUT_H "#define   ID_$error $errorcnt\n";

	    if ($args)
	    {
		$args_parm_list = $args;
		
		$args_parm_list =~ s/([diouxXc][0-9]+)/dword $1/g;
		$args_parm_list =~ s/([s][0-9]+)/const char* $1/g;
		$args_parm_list =~ s/(rva[0-9]+)/dword $1/g;
		
		$args_parm_creators = $args;
		
		$args_parm_creators =~ s/([diouxXc][0-9]+)/AddErrorArgInteger($1)/g;
		$args_parm_creators =~ s/([s][0-9]+)/AddErrorArgString($1)/g;
		$args_parm_creators =~ s/(rva[0-9]+)/AddErrorArgRVA($1)/g;
	
		print INFILE_OUT_H "Error* CreateError_$error(dword dwErrorId,const char *file,unsigned int line,const char *function,$args_parm_list);\n";
		
		print OUTFILE_CPP "Error* CreateError_$error(dword dwErrorId,const char *file,unsigned int line,const char *function,$args_parm_list)\n";
		print OUTFILE_CPP "{\n";
		
		print OUTFILE_CPP "  Error* pError = new Error(dwErrorId,file,line,function);\n";
		foreach $arg_creator (split /,/,$args_parm_creators )
		{
		    print OUTFILE_CPP "  pError->$arg_creator;\n";
		}
		print OUTFILE_CPP "  return pError;\n";
		print OUTFILE_CPP "}\n\n";
		
		print INFILE_OUT_H "#define      $error($args) CreateError_$error(ID_$error,__FILE__,__LINE__,__ASSERT_FUNCTION,$args)\n";		


	    }
	    else
	    {
		print INFILE_OUT_H "#define      $error($args) (new Error(ID_$error,__FILE__,__LINE__,__ASSERT_FUNCTION))\n";		    
	    }


	    
	    # print INFILE_OUT_H "#define EMIT_$error($args) {errorTracker.AddError($error($args));}\n";

	    # print INFILE_OUT_H "\n";

	    # print "$short\n\n";

	    print OUTFILE_CPP "const char* strSHORT_$error = \"$short\";\n";
	    # print OUTFILE_CPP "const char argorderSHORT_$error\[\] = { $argorder };\n";

	    if ($long eq "<as above>")
	    {
		print OUTFILE_CPP "const char* strLONG_$error = $lastlong;\n\n";
	    }
	    else
	    {
		print OUTFILE_CPP "const char* strLONG_$error = \"$long\";\n\n";
		$lastlong = "strLONG_$error";
	    }

	    push @errors,$error;

	    $errorcnt++;
	}
	else
	{
	    print "Error: $cont\n"; die;
	}
    }

    print INFILE_OUT_H GeneratedFooterH();
    
    close(INFILE_OUT_H );

    close(INFILE);
}

print OUTFILE_CPP "ErrorMessageStrings errorMessageStrings[] = {\n";

foreach $error (@errors)
{
    print OUTFILE_CPP "  { ID_$error,strSHORT_$error,strLONG_$error },\n"; # ...,argorderSHORT_$error,...
}
print OUTFILE_CPP "};\n";

print OUTFILE_CPP <<"EndOfText";

//////////////////////////////////////////////////////////////////////
EndOfText

close(OUTFILE_CPP);

$numerrmess = $#errors+1;

print OUTFILE_H "#define NUMBER_OF_NAMED_ERROR_MESSAGES $numerrmess\n\n";

print OUTFILE_H GeneratedFooterH();
close(OUTFILE_H );


sub FigureOutArgs()
{
    my $short = shift;
    # my @argorder = ();
    # my $argorder = "";
    my $shortout = "";

    my $shortorig = $short;
    
    # print "\n$short\n";

    my @normalargs = ();
    my @specialargs = ();

    # Now see what we try to print in the error message.
    # The variables printed must come as arguments

  find_args:
    while ( 1 )
    {
        if ($short =~ s/^(.*?)%/%/) { $shortout .= $1; }
        # print "(($shortout,$1))\n";
        if ($short =~ s/^(%[0-9\+\-\ \.]*([diouxXcs]))([0-9]+)//)
        {
            # print "ARG: $1  $2\n";
            $shortout .= $1.$3;
            if ($normalargs[$3-1]) { print "Arg $3 already used: $shortorig\n"; die; }
            $normalargs[$3-1] = $2.$3;	    
            # $argtype = $2;
            # $argnum = $3;
            # if ($argtype =~ /[diouxXc]/) { push @argorder,"ARG_ORDER_INT($argnum)"; }
            # else { push @argorder,"ARG_ORDER_PTR($argnum)"; }
        }
        elsif ($short =~ s/^%RVA([0-9]+)//)
        {
            # print "ARG: RVA  $1\n";
            $shortout .= "0x%08x$1";
            # push @argorder,"ARG_ORDER_RVA($1)";
            if ($normalargs[$1-1]) { print "Arg $1 already used: $shortorig\n"; die; }
            $normalargs[$1-1] = 'rva'.$1;
            # push @specialargs , 'rva'.$1;
        }
        elsif ($short =~ s/^(%%)//)
        {
            $shortout .= $1;	    
        }
        elsif ($short =~ /^%/)
        {
            print "Contain bad format specifier (%) (probably missing format or number):\n$short\n"; die;
        }
        else { last find_args; }
        # print "(($shortout <> $short))\n";
    }
    $shortout .= $short;

    my @args = (@normalargs,@specialargs);

    my $args = join(',',@args);

    # $argorder = join(',',(@argorder,'0'));

    # print "\n$shortout  (($short))\n$argorder\n";

    return ($args,$shortout); # ,$argorder);
}
