/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ImportedEntries.cpp: implementation of the ImportedEntries class.
//
//////////////////////////////////////////////////////////////////////

#include "ImportedEntries.h" 
#include "gen/ImportedEntriesError.h" 
#include "DataDestRelocPtr.h" 
#include "DataPtrPESectionBody.h"  

#include "RelocatableDest.h" 

//#include <assert.h>
#include <algorithm>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

MapMovedImportAddressTableRVA mapMovedImportAddressTableRVA;

ImportedEntries::ImportedEntries()
{
  
}

ImportedEntries::~ImportedEntries()
{
  
}

void ImportedEntries::ReadPEForwarderChain(set<dword>& nForwFuncs,
					   DataPtrPESectionBody* pDataPtr,
					   dword dwForwarder,
					   dword dwForwarderRVA,
					   dword dwTableLength,
					   dword dwImportAddressTableRVA)
{
  while (dwForwarder != (dword) -1)
    {
      if (dwForwarder >= dwTableLength)
        throw WARN_IMP_FNC_FORWARDER_OUTSIDE_RANGE(dwForwarder,
						   dwTableLength)->RVA(dwForwarderRVA);
      
      // if we've had it before, it's an error as this is an loop
      
      if (!(nForwFuncs.insert(dwForwarder)).second)
        {	
          // it was already in list, abort
	  
          throw WARN_IMP_FNC_FORWARDER_CHAIN_CONTAIN_LOOP(dwForwarder)->RVA(dwForwarderRVA);
        }

      // GetDWord may throw if attempting to read outside source

      dwForwarderRVA = dwImportAddressTableRVA + sizeof (dword) * dwForwarder;
      dwForwarder = pDataPtr->GetDWord(dwForwarderRVA);
    }
}

struct ImportDirectoryTable
{
  dword dwImportLookupTableRVA;
  dword dwTimeDate;
  dword dwForwarderChain;
  dword dwNameRVA;
  dword dwImportAddressTableRVA;
};

void ImportedEntries::ReadPEImports (DataPtrPESectionBody* pDataPtr)
{
  // ok, so read the imports
  
  for ( ; ; )
    {
      dword dwIDT_RVA = pDataPtr->GetRVA();
      const ImportDirectoryTable* pIDT = 
	(const ImportDirectoryTable*) pDataPtr->GetPtr(sizeof (ImportDirectoryTable));
      
      if (!pIDT->dwImportLookupTableRVA)
        {
          // entire array 0 or just this member?
          
          for (const char* ptr = (const char*) pIDT; 
	       ptr < (const char*) (pIDT + 1); ptr++)
            if (*ptr)
              {
                // buggy directory with only address-table ???
                throw ERROR_IMP_FNC_IMPORT_DIR_NOT_TERMINATED_BY_ALL_ZERO_STRUCT()->RVA(dwIDT_RVA);
              }
          
          // end of imports
          return;
        }
      
      if (pIDT->dwTimeDate == 0xffffffff)
        throw ERROR_IMP_FNC_NEW_TYPE_BINDING_NOT_SUPPORTED()->RVA(dwIDT_RVA);
      
      ImportedLibrary* pImpLib = new ImportedLibrary;
      
      pImpLib->m_strLibraryName = pDataPtr->GetString (pIDT->dwNameRVA);
      pImpLib->m_dwTimeDate = pIDT->dwTimeDate;
      
      m_vectLibraries.push_back(pImpLib);
      
      // first determine the length of the lookup and address tables,
      // so we can see if an forwarder is outside range
      
      DataPtrPESectionBody* pDataPtrImportLookupTable = 
	pDataPtr->GetRVADataPtr(pIDT->dwImportLookupTableRVA);
      DataPtrPESectionBody* pDataPtrImportAddressTable = 
	pDataPtr->GetRVADataPtr(pIDT->dwImportAddressTableRVA);
      
      //		printf ("Library: %s\n",impLib.m_strLibraryName.c_str());
      
      dword nTableLength = 0;
      
      for ( ; ; nTableLength++)
        {
          // dword dwPtrAddressRVA = pDataPtrImportAddressTable->GetRVA ();
          
          dword dwLookupRVA = pDataPtrImportLookupTable->GetDWord();
          dword dwAddressRVA = pDataPtrImportAddressTable->GetDWord();
          
          if (!dwLookupRVA || !dwAddressRVA)
            {
              if (dwLookupRVA || dwAddressRVA)
                throw ERROR_IMP_FNC_LOOKUP_AND_ADDRESS_TABLES_NOT_TERMINATED_SAME(dwLookupRVA,
                                                                                  dwAddressRVA)->
                  RVA(pDataPtrImportLookupTable->GetRVA()-sizeof (dword))->
                  RVA(pDataPtrImportAddressTable->GetRVA()-sizeof (dword));
              break;
            }
        }
      
      set<dword> nForwFuncs;
      
      // read forwarder chain
      
      // pIDT->dwForwarderChain = 0xffffffff , no forwarder chain,
      // otherwise, it points to first forwarded, which of course also
      // can be zero I've however also found that sometimes a zero is
      // ejected, even though it's not forwarded, so in case it
      // started at zero, we'll silently ignore any errors, that is:
      // impossible indexes
      
      if (pIDT->dwForwarderChain == 0)
        {
          // errors will come out as exceptions
          try
            {
              ReadPEForwarderChain(nForwFuncs,pDataPtr,
				   pIDT->dwForwarderChain,
				   dwIDT_RVA+2*sizeof(dword),
				   nTableLength,pIDT->dwImportAddressTableRVA);
            }
          catch (const Error*)
            {
              nForwFuncs.clear();
                                // throw ;
            }
        }
      else // if we get any errors here, they are considered fatal
        ReadPEForwarderChain(nForwFuncs,pDataPtr,
			     pIDT->dwForwarderChain,dwIDT_RVA+2*sizeof(dword),
			     nTableLength,pIDT->dwImportAddressTableRVA);
      
      delete pDataPtrImportLookupTable;
      delete pDataPtrImportAddressTable;
      
      pDataPtrImportLookupTable = 
	pDataPtr->GetRVADataPtr(pIDT->dwImportLookupTableRVA);
      pDataPtrImportAddressTable = 
	pDataPtr->GetRVADataPtr(pIDT->dwImportAddressTableRVA);
      
      //		printf ("Library: %s\n",impLib.m_strLibraryName.c_str());
      
      unsigned int nIndex;
      
      for (nIndex = 0; ; nIndex++)
        {
          dword dwPtrAddressRVA = pDataPtrImportAddressTable->GetRVA ();
          
          dword dwLookupRVA = pDataPtrImportLookupTable->GetDWord();
          dword dwAddressRVA = pDataPtrImportAddressTable->GetDWord();
          
          assert (!dwLookupRVA == !dwAddressRVA);
          
          if (!dwLookupRVA) // end of table?
            break;

          ImportedEntry* pImpFnc = new ImportedEntry;

          if (dwLookupRVA & IMPORT_FUNC_BY_ORDINAL_ONLY)
            {
              if (dwLookupRVA & 0x7fff0000)
                throw ERROR_IMP_FNC_ORDINAL_OUTSIDE_WORD_RANGE(dwLookupRVA & 0x7fff0000)->
                  RVA(pDataPtrImportLookupTable->GetRVA()-sizeof (dword));
	      
              pImpFnc->m_nFunctionOrdinal = dwLookupRVA;
              //				printf ("  Fcn: %d\n",impFnc.m_nFunctionOrdinal);
            }
          else
            {
              DataPtrPESectionBody* pDataPtrOrdinalName = 
		pDataPtr->GetRVADataPtr(dwLookupRVA);
	      
              pImpFnc->m_nFunctionOrdinal = pDataPtrOrdinalName->GetWord ();
              pImpFnc->m_strName = pDataPtrOrdinalName->GetString();
              //				printf ("  Fcn: %d,%s\n",impFnc.m_nFunctionOrdinal,impFnc.m_strFunctionName.c_str());
	      
              delete pDataPtrOrdinalName;
            }
	  
          pImpFnc->m_dwImportAddressTableRVA = dwPtrAddressRVA;
	  
          pImpFnc->m_bForwarded = 
	    (nForwFuncs.find(nIndex) != nForwFuncs.end());
	  
          // if the library was bound, we should get the adress
	  
          if (!(pImpFnc->m_bForwarded)) // if forwarded, the address-table contain the forwarder-next-index, so it can't be the address
            {
              if (pIDT->dwTimeDate) // bound if non-zero)
                {
                  pImpFnc->m_dwBoundAddress = dwAddressRVA;
                }
              else
                {
                  if (dwAddressRVA != dwLookupRVA)
                    throw ERROR_IMP_FNC_LOOKUP_AND_ADDRESS_NOT_EQUAL_FOR_UNBOUND(dwLookupRVA,
                                                                                 dwAddressRVA)->
                      RVA(pDataPtrImportLookupTable->GetRVA()-sizeof (dword))->
                      RVA(pDataPtrImportAddressTable->GetRVA()-sizeof (dword));
                }
            }
          AddFnc(pImpFnc,pImpLib);
        }
      
      delete pDataPtrImportLookupTable;
      delete pDataPtrImportAddressTable;
      

      //		if (!nForwFuncs.empty() && *(--(nForwFuncs.end())) >= nIndex)
      //			throw "One or more forwarders outside range.";				

    }
  
  //	return true;
}

void ImportedEntries::AddFnc (ImportedEntry* pFnc,ImportedLibrary* pLib)
{
  pFnc->m_pLibrary = pLib;
  
  pLib->m_vectImportedEntry.push_back(pFnc);
  m_vectImportedEntry.push_back (pFnc);
}

void ImportedEntries::DeleteFnc (ImportedEntry* pFnc)
{
  ImportedLibrary* pLib = pFnc->m_pLibrary;
  
  VectImportedEntry::iterator erase = 
    find(pLib->m_vectImportedEntry.begin(),
	 pLib->m_vectImportedEntry.end(),pFnc);
  
  assert (erase != pLib->m_vectImportedEntry.end());
  pLib->m_vectImportedEntry.erase(erase);
  
  erase = find(m_vectImportedEntry.begin(),m_vectImportedEntry.end(),pFnc);
  
  assert (erase != m_vectImportedEntry.end());
  m_vectImportedEntry.erase(erase);
  
  delete pFnc;
}

dword FindNextForward(VectImportedEntry& vectFnc,dword dwLookNext)
{
  uint n = vectFnc.size();
  
  for ( ; dwLookNext < n; dwLookNext++)
    {
      if (vectFnc[dwLookNext]->m_bForwarded)
	return dwLookNext;
    }
  return (dword) -1;
}

RelocatableDest* ImportedEntries::WritePEImports ()
{
  // we'll rewrite the import table
  //
  // although, one could also fix an existing import table actually,
  // there would be nothing to fix, except adding the offset by which
  // the import table was moved in the executable image and add the
  // offset to all the imported functions' m_dwImportAddressTableRVA
  //
  // that would be the FixImports() or perhaps RelocateImports()
  //
  // It would be useful if we somehow haven't figured out enough about
  // the image to dare to touch (or do a major rewrite of the import
  // table), but if we do not dare to do that, then we probably do not
  // dare to do anything anyway, so...

  /*
    vector<ImportedLibrary*> m_vectLibraries;
    vector<ImportedEntry*> m_vectImportedEntry;
  */
  
  // all functions are in m_vectImportedEntry, and all libraries are
  // in m_vectLibraries
  
  // at this point, we assume that all functions that are pointed to
  // by an library are needed. So if you want to delete any, do it
  // before we're called)

  // To be written are
  // One ImportDirectoryTable for each library, and one (all-zero) one as terminator
  // For each library, two tables:
  // One look-up table with RVA's to the strings (or ordinals & 0x80000000)
  // One address table (initialized to the same as the look-up table if unbound)
  // And one array with the hint and names
  
  // find out number of libraries and also the number of functions, so
  // we can calculate we're to place the strings
  
  uint nLibraries = 0;
  uint nFunctions = 0;
  
  dword dwHintStringSize = 0;
  dword dwLibNameStringSize = 0;
  
  FOR_EACH_CONST_ITER(VectImportedLibrary,m_vectLibraries,library)
    {
      if ((*library)->m_vectImportedEntry.size())
        {
          nLibraries++;
          nFunctions += (*library)->m_vectImportedEntry.size();
	  
          dwLibNameStringSize += (*library)->m_strLibraryName.size()+1;
	  
          FOR_EACH_CONST_ITER(VectImportedEntry,
			      (*library)->m_vectImportedEntry,func)
            {
              if (!((*func)->m_nFunctionOrdinal & IMPORT_FUNC_BY_ORDINAL_ONLY))
                {
                  int str_size = (*func)->m_strName.size();
		  
                  str_size = (str_size + 2) & ~0x01; // +1 to include terminating '\0', then (...+1)&~1 to make it word-aligned
		  
                  dwHintStringSize += sizeof (word) + str_size;
                }
            }
        }
    }
  
  if (!nLibraries)
    {
      // no functions imported
      
      return NULL;
    }
  
  // ok, so how much space is needed?
  
  dword dwIDTSize = (nLibraries + 1) * sizeof (ImportDirectoryTable); // + 1 for terminator
  dword dwLookUpTablesSize = (nFunctions + nLibraries) * sizeof (dword);  // + nLibraries for terminators
  
  // and space for the hint's and strings

  RelocatableDest* pDest = new RelocatableDest();
  
//  char* pImportTable = new char[dwIDTSize+2*dwLookUpTablesSize+dwLibNameStringSize+dwHintStringSize];

  pDest->AllocDataBuffer(dwIDTSize+
			 2*dwLookUpTablesSize+
			 dwLibNameStringSize+
			 dwHintStringSize);
  
  //  RelocationInfo* pRelocation = new RelocationInfo();
	
  //  DataDestRelocPtr destPtr((void*) pImportTable,pImportTable+dwIDTSize+2*dwLookUpTablesSize+dwLibNameStringSize+dwHintStringSize,pRelocation);

  DataDestRelocPtr destPtr(pDest);
  
  //DataDestRelocPtr(void* pStart IFDEBUGPARAM(void* pEnd) ,RelocationInfo* pReloc);
  
#define IMP_DIR_PTR (0)
#define IMP_LOOKUP_PTR (IMP_DIR_PTR+dwIDTSize)
#define IMP_ADDRESS_PTR (IMP_LOOKUP_PTR+dwLookUpTablesSize)
#define IMP_LIB_NAME_PTR (IMP_ADDRESS_PTR+dwLookUpTablesSize)
#define IMP_HINT_NAME_PTR (IMP_LIB_NAME_PTR+dwLibNameStringSize)
#define IMP_HINT_NAME_END_PTR (IMP_HINT_NAME_PTR+dwHintStringSize)
  
  DataDestRelocPtr destPtrDirectoryTables(&destPtr,IMP_DIR_PTR       IFDEBUGPARAM(dwIDTSize          ) );
  DataDestRelocPtr destPtrLookupTables   (&destPtr,IMP_LOOKUP_PTR    IFDEBUGPARAM(dwLookUpTablesSize ) );
  DataDestRelocPtr destPtrAddressTables  (&destPtr,IMP_ADDRESS_PTR   IFDEBUGPARAM(dwLookUpTablesSize ) );
  DataDestRelocPtr destPtrLibNameTable   (&destPtr,IMP_LIB_NAME_PTR  IFDEBUGPARAM(dwLibNameStringSize) );
  DataDestRelocPtr destPtrHintNameTables (&destPtr,IMP_HINT_NAME_PTR IFDEBUGPARAM(dwHintStringSize   ) );
  
  FOR_EACH_CONST_ITER(VectImportedLibrary,m_vectLibraries,lib)
    {
      ImportedLibrary* pLib = *lib;
      ImportDirectoryTable IDT;
      
      dword dwForward = FindNextForward(pLib->m_vectImportedEntry,0);
      
      IDT.dwTimeDate = pLib->m_dwTimeDate;
      IDT.dwNameRVA = destPtrLibNameTable.GetRVA();
      IDT.dwImportAddressTableRVA = destPtrAddressTables.GetRVA();
      IDT.dwImportLookupTableRVA = destPtrLookupTables.GetRVA();
      IDT.dwForwarderChain = (dword) -1;
      
      destPtrDirectoryTables.WriteDWord(IDT.dwImportLookupTableRVA,
					RELATIVE_IMAGE_CURRENT);
      destPtrDirectoryTables.WriteDWord(IDT.dwTimeDate);
      destPtrDirectoryTables.WriteDWord(IDT.dwForwarderChain);
      destPtrDirectoryTables.WriteDWord(IDT.dwNameRVA,
					RELATIVE_IMAGE_CURRENT);
      destPtrDirectoryTables.WriteDWord(IDT.dwImportAddressTableRVA,
					RELATIVE_IMAGE_CURRENT);
      
      destPtrLibNameTable.WriteString(pLib->m_strLibraryName);
      
      if (pLib->m_vectImportedEntry.size())
        {
          FOR_EACH_CONST_ITER(VectImportedEntry,pLib->m_vectImportedEntry,func)
            {
              ImportedEntry* pFnc = *func;
	      
              dword dwHintNameRVA;
              word wHintNameRVARelocate;
	      
              if (!(pFnc->m_nFunctionOrdinal & IMPORT_FUNC_BY_ORDINAL_ONLY))
                {
                  dwHintNameRVA = destPtrHintNameTables.GetRVA();
                  wHintNameRVARelocate = RELATIVE_IMAGE_CURRENT;
		  
                  assert (!(pFnc->m_nFunctionOrdinal & 0xffff0000)); // so it fits in a word
		  
                  destPtrHintNameTables.WriteWord(pFnc->m_nFunctionOrdinal);
                  destPtrHintNameTables.WriteString(pFnc->m_strName,
						    sizeof (word));
                }
              else
                {
                  dwHintNameRVA = pFnc->m_nFunctionOrdinal;
                  wHintNameRVARelocate = 0;
                }
	      
              destPtrLookupTables.WriteDWord(dwHintNameRVA,
					     wHintNameRVARelocate);
	      
              dword dwNewImportAddressTableRVA = destPtrAddressTables.GetRVA();
	      
              assert (pFnc->m_dwImportAddressTableRVA);
              assert (mapMovedImportAddressTableRVA.find(pFnc->m_dwImportAddressTableRVA) == 
                      mapMovedImportAddressTableRVA.end());
	      
              mapMovedImportAddressTableRVA.
                insert(MapMovedImportAddressTableRVA::value_type(pFnc->m_dwImportAddressTableRVA,
                                                                 dwNewImportAddressTableRVA));
	      
              pFnc->m_dwImportAddressTableRVA = dwNewImportAddressTableRVA;
	      
              if (pFnc->m_bForwarded)
                {
                  dwForward = FindNextForward(pLib->m_vectImportedEntry,
					      dwForward+1);
                  destPtrAddressTables.WriteDWord(dwForward);
                }
              else if (pFnc->m_dwBoundAddress)
                destPtrAddressTables.WriteDWord(pFnc->m_dwBoundAddress);
              else
                destPtrAddressTables.WriteDWord(dwHintNameRVA,
						wHintNameRVARelocate);
            }
          destPtrLookupTables.WriteDWord(0);
          destPtrAddressTables.WriteDWord(0);
        }
    }
  
  destPtrDirectoryTables.WriteZero(sizeof (ImportDirectoryTable));
  
  assert (destPtrDirectoryTables.ReacedEnd());
  assert (destPtrLookupTables.ReacedEnd());
  assert (destPtrAddressTables.ReacedEnd());
  assert (destPtrLibNameTable.ReacedEnd());
  assert (destPtrHintNameTables.ReacedEnd());
  
  return pDest;
}


