/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ImportedFncs.cpp: implementation of the ImportedFncs class.
//
//////////////////////////////////////////////////////////////////////

#include "ImportedFncs.h"
#include <assert.h>

#include <set>

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ImportedFncs::ImportedFncs()
{

}

ImportedFncs::~ImportedFncs()
{

}

struct ImportDirectoryTable
{
  dword dwImportLookupTableRVA;
  dword dwTimeDate;
  dword dwForwarderChain;
  dword dwNameRVA;
  dword dwImportAddressTableRVA;
};

bool ImportedFncs::ReadImports (DataPtrPESectionBody* pDataPtr)
{
  // ok, so read the imports
  
  for ( ; ; )
    {
      ImportDirectoryTable* pIDT = 
	(ImportDirectoryTable*) pDataPtr->GetPtr(sizeof (ImportDirectoryTable));
      
      if (!pIDT->dwImportLookupTableRVA)
	{
	  // entire array 0 or just this member?
	  
	  // end of imports
	  return true;
	}
      
      ImportedLibrary* pImpLib = new ImportedLibrary;

      pImpLib->m_strLibraryName = pDataPtr->GetString (pIDT->dwNameRVA);
      pImpLib->m_dwTimeDate = pIDT->dwTimeDate;
      
      m_vectLibraries.push_back(pImpLib);
      
      set<dword> nForwFuncs;
      
      // read forwarder chain
      
      if (0 /* flag to avoid nutty forwarder chains*/)
	{
	  dword dwForwarder = pIDT->dwForwarderChain;
	  
	  while (dwForwarder != (dword) -1)
	    {
	      nForwFuncs.insert(dwForwarder);
	      
	      dwForwarder = pDataPtr->GetDWord(pIDT->dwImportAddressTableRVA +
					       sizeof (dword) * dwForwarder);
	    }
	}
      
      DataPtrPESectionBody dataPtrImportLookupTable = 
	pDataPtr->GetRVADataPtr(pIDT->dwImportLookupTableRVA);
      DataPtrPESectionBody dataPtrImportAddressTable = 
	pDataPtr->GetRVADataPtr(pIDT->dwImportAddressTableRVA);
      
      //		printf ("Library: %s\n",impLib.m_strLibraryName.c_str());
      
      unsigned int nIndex;
      
      for (nIndex = 0; ; nIndex++)
	{
	  dword dwPtrAddressRVA = dataPtrImportAddressTable.GetRVA ();
	  
	  dword dwLookupRVA = dataPtrImportLookupTable.GetDWord();
	  dword dwAddressRVA = dataPtrImportAddressTable.GetDWord();
	  
	  if (!dwLookupRVA || !dwAddressRVA)
	    {
	      if (dwLookupRVA || dwAddressRVA)
		throw "Import lookup and address table not terminated simultaneously.";				
	      break;
	    }
	  
	  ImportedFnc* pImpFnc = new ImportedFnc;
	  
	  pImpFnc->m_pLibrary = pImpLib;
	  
	  if (dwLookupRVA & 0x80000000)
	    {
	      pImpFnc->m_nFunctionOrdinal = dwLookupRVA & 0x7fffffff;
	      //				printf ("  Fcn: %d\n",impFnc.m_nFunctionOrdinal);
	    }
	  else
	    {
	      DataPtrPESectionBody dataPtrOrdinalName = 
		pDataPtr->GetRVADataPtr(dwLookupRVA);
	      
	      pImpFnc->m_nFunctionOrdinal = dataPtrOrdinalName.GetWord ();
	      pImpFnc->m_strName = dataPtrOrdinalName.GetString();
	      //				printf ("  Fcn: %d,%s\n",impFnc.m_nFunctionOrdinal,impFnc.m_strFunctionName.c_str());
	    }
	  
	  pImpFnc->m_dwImportAddressTableRVA = dwPtrAddressRVA;
	  
	  pImpFnc->m_bForwarded = 
	    (nForwFuncs.find(nIndex) != nForwFuncs.end());
	  
	  m_vectImportedFnc.push_back(pImpFnc);
	}
      
      if (!nForwFuncs.empty() && *(--(nForwFuncs.end())) >= nIndex)
	throw "One or more forwarders outside range.";
    }
  
  //	return true;
}

