#!/usr/bin/perl

#  exmhcg - executable program tuner  
#  Copyright (C) 2003  Hkan Johansson
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

$infile = $ARGV[0];
$outfile = $ARGV[1];
$outfile2 = $ARGV[2];

$shortname = $ARGV[3];
$fullname = $ARGV[4];

$prefix = $shortname;

$Shortname = $ARGV[5]; # $shortname;
$Fullname = $ARGV[6]; # $fullname;
$SHORTNAME = $ARGV[7]; # $shortname;
$FULLNAME = $ARGV[8]; # $fullname;

print "Generating $outfile, $outfile2 from $infile\nPrefix is '$prefix'.\n\n";
 
print "$prefix , $shortname , $fullname , $Shortname , $Fullname , $SHORTNAME , $FULLNAME\n";

open(INFILE,"<$infile") || die "Can't open $infile";

@nmbrs = ();

while (<INFILE>)
{ 
    $readline = $_;

    if ($readline =~ /\s*\#\s*define\s+(\w+)\s+(.*)/)
    {
        $name = $1;
        $macro = $2;

        if ($macro =~ /(\/\/)/) { $macro = $`; }
        if ($macro =~ /(\/\*)/) { $macro = $`; }

        $macro=~s/^\s+//g;
        $macro=~s/\s+$//g;

        $macro = macro_expand($macro);

        $macros{$name} = $macro;

        # printf "                     $name -> <$macro>\n";
 
    }

    if ($readline =~ /\s*\#\s*define\s+$shortname([A-Z0-9_][a-zA-Z0-9_]*)/)
    {
        $lineend = $';
        $instr_name = $1;
        $instr_id = "$shortname$1";

        # print "$readline\n";
        # print "$lineend\n";

        if (!($lineend =~ /no_$shortname_name/))
        {
            if (($lineend =~ /$shortname_name="(.*)"/))
            {
                $instr_name = $1;
            }

            push @nmbrs,$macros{$instr_id};
            $instr_nmbrs{$macros{$instr_id}} = $instr_id;
            $instrs{$instr_id} = $instr_name;
 
            # print "$instr_id , $macros{$instr_id} => $instr_name\n";
            # print "$'\n";
        }
    }
}

close(INFILE);

open(OUTFILE,">$outfile") || die "Can't open $outfile";

print OUTFILE <<"EndOfText";
// $outfile: strings for $fullname names.
//
//////////////////////////////////////////////////////////////////////
//
// This file is automatically generated by a script.
// Do not edit - useless.
//
//////////////////////////////////////////////////////////////////////

#include "../$infile"

//////////////////////////////////////////////////////////////////////

char c${Shortname}Names[] = {\
EndOfText

# print join(' ',@nmbrs)."\n\n";

@nmbrs = sort {$a <=> $b} @nmbrs;

# print join(' ',@nmbrs)."\n\n";

###

$offset = 0;

$lastname = "";

@names = ();

foreach $no (@nmbrs)
{
    $id = $instr_nmbrs{$no};
    $name = $instrs{$id};
    $len = (length($name)+1);

    @length = (@length,$len);
    @offset = (@offset,$offset);

    # print "$no $id $name $offset\n";

    ($lastname ne $name) || die "Error: Multiple $name\n";

    $dataout = "\"$name\\0\"";
    $out = sprintf "%-12s // %-15s = %-5d, offset = %d",($dataout,$id,$no,$offset);

    print OUTFILE "$out\n";

    $offset = $offset + $len;

    $lastname = $name;

    push @names,$name;
}

# print join(' ',@names)."\n\n";
#
#foreach $big (@names)
#{
#    foreach $small (@names)
#    {
#	if (($big ne $small) && ($big =~ /$small/))
#	{
#	    print "$small in $big\n";
#	}
#    }
#}

###

print OUTFILE <<"EndOfText";
};

$Shortname\IdNameIndex index$Shortname\IdName[] = {
EndOfText

$i = 0;

foreach $no (@nmbrs)
{
    $id = $instr_nmbrs{$no};

    $offset = $offset[$i];

    $out = sprintf "{ %-15s, %-5d }, // %-15s = %d",($id,$offset,$id,$no);

    print OUTFILE "$out\n";

    $offset = $offset + $len;

    $i++;
}

$no_instr_names = $i;

print OUTFILE <<"EndOfText";
};

//////////////////////////////////////////////////////////////////////
EndOfText

close(OUTFILE);

open(OUTFILE,">$outfile2") || die "Can't open $outfile2";

print OUTFILE <<"EndOfText";
// $outfile2: defines for Instruction names.
//
//////////////////////////////////////////////////////////////////////
//
// This file is automatically generated by a script.
// Do not edit - useless.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__$FULLNAME\_NAMES_H__INCLUDED_)
#define __$FULLNAME\_NAMES_H__INCLUDED_

#define NUMBER_OF_$SHORTNAME\_NAMES      $no_instr_names
#define SIZEOF_$SHORTNAME\_NAMES_STRING  $offset

#endif // !defined(__$FULLNAME\_NAMES_H__INCLUDED_)

//////////////////////////////////////////////////////////////////////

EndOfText

#	$macro = macro_expand($name,$macro,$macros);

sub macro_expand {
    $m = $_[0];

    # remove parens

    $m=~s/^\(//g;
    $m=~s/\)$//g;  

    @exprs = split(/\|/,$m);

    $result = 0;

    foreach $expr (@exprs)
    {
        $expr=~s/^\s+//g;
        $expr=~s/\s+$//g;

        if ($value = @macros{$expr})
        {
            $expr = $value;
        }
        elsif ($expr =~ /0x/)
        {
            $expr = hex($expr);
        }
        $result |= $expr;
    }

    return $result;
} 


