/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Instruction.cpp: implementation of the Instruction class.
//
//////////////////////////////////////////////////////////////////////

#include "Instruction.h" 
#include "gen/Instruction_Names.h" 
#include "OperandRegister.h" 

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Instruction::Instruction()
{
  m_nInstruction = 0;
  //	m_wInfo = 0;
}

Instruction::~Instruction()
{

}

//////////////////////////////////////////////////////////////////////

bool Instruction::OpDestSrcSameRegister () const
{
  const OperandRegister* pOpRegDest;
  const OperandRegister* pOpRegSrc;
  
  if (!(pOpRegDest = dynamic_cast<const OperandRegister*>(GetDestOp())))
    return false;
  if (!(pOpRegSrc = dynamic_cast<const OperandRegister*>(GetSrcOp())))
    return false;
  
  return pOpRegDest->m_bReg == pOpRegSrc->m_bReg;
}

// extern char cInstrNames[];
// extern InstrIdNameIndex indexInstrIdName[];

int compare_short (const void* p1,const void* p2)
{
  return ((short) *((const short*) p1) - (short) *((const short*) p2));
}

const char* GetInstructionName (short nInstr)
{
  InstrIdNameIndex* pIndex = 
    (InstrIdNameIndex*) bsearch (&nInstr,
				 indexInstrIdName,
				 NUMBER_OF_INSTR_NAMES,
				 sizeof (InstrIdNameIndex),
				 compare_short);
  
  // printf ("%d %d    | %d %x\n",NUMBER_OF_INSTR_NAMES,sizeof (InstrIdNameIndex),nInstr,nInstr);
  
  assert (pIndex); // otherwise we're missing name of the instruction
  
  assert (pIndex->m_nOffset >= 0 &&
	  pIndex->m_nOffset < SIZEOF_INSTR_NAMES_STRING);
  
  return cInstrNames+pIndex->m_nOffset;
}

//////////////////////////////////////////////////////////////////////
/*
void Instruction::AsmString (char* pDest) const
{
	pDest = GenAsmString (pDest);
	*pDest = '\0';
}
*/
void Instruction::InstrAsmString (char* pDest) const
{
  pDest = GenInstrAsmString (pDest);
  *pDest = '\0';
}

void Instruction::OpAsmString (char* pDest) const
{
  pDest = GenOpAsmString (pDest);
  *pDest = '\0';
}
/*
  char* Instruction::GenAsmString (char* pDest) const
{
	const char* pSrc = GetInstructionName(m_nInstruction);

	while (*pSrc)
		*(pDest++) = *(pSrc++);
	return pDest;
}
*/
char* Instruction::GenInstrAsmString (char* pDest) const
{
  const char* pSrc = GetInstructionName(m_nInstruction);

  while (*pSrc)
    *(pDest++) = *(pSrc++);
  return pDest;
}

char* Instruction::GenOpAsmString (char* pDest) const
{
  *pDest = '\0';
  return pDest;
}

//////////////////////////////////////////////////////////////////////




