/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Instruction.h: interface for the Instruction class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__INSTRUCTION_H__INCLUDED_)
#define __INSTRUCTION_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

class Operand; 
#include "GlobalDefs.h" 

#include <assert.h>

//////////////////////////////////////////////////////////////////////

// this halfway but not fully logical separation of instruction
// numbering cause trouble when trying to get the names of
// instructions from the codes, as even though there is a very small
// number of instructions they span a large number range

// I've therefore written a little script that generate two lists
// (into a src-file) that hold the names without to much fuss. Run it
// whenever an instruction has been changed

// The program is a perl script (almost my first one) named
// InstrNamesFromIds.pl and generates the file Instruction_Names.cpp

//////////////////////////////////////////////////////////////////////

#define instrgMASK			0x7800 // (bits 12-15)

#define INSTR_GROUP(instr) ((instr) & instrgMASK)

//////////////////////////////////////////////////////////////////////

#define instrgCTRL_TRANSFER		0x0800

#define instrmCTRL_TRANSFER_TYPE_MASK	0x0070

#define instrmJMP			0x0010
#define instrmCALL			0x0020
#define instrmRET			0x0030
#define instrmJMP_COND			0x0040
#define instrmINT			0x0050
#define instrmIRET			0x0060

#define instrmNEAR_FAR_MASK		0x0003

#define instrmNEAR			0x0001
#define instrmFAR			0x0002

#define instrmABS_REL_MASK		0x000c

#define instrmREL			0x0004
#define instrmABS			0x0008

#define instrJMPN_REL			(instrgCTRL_TRANSFER | instrmJMP  | instrmNEAR | instrmREL) // instr_name="JMPnr"
#define instrJMPN_ABS			(instrgCTRL_TRANSFER | instrmJMP  | instrmNEAR | instrmABS) // instr_name="JMPna"
#define instrJMPF_ABS			(instrgCTRL_TRANSFER | instrmJMP  | instrmFAR  | instrmABS) // instr_name="JMPfa"

#define instrCALLN_REL			(instrgCTRL_TRANSFER | instrmCALL | instrmNEAR | instrmREL) // instr_name="CALLnr"
#define instrCALLN_ABS			(instrgCTRL_TRANSFER | instrmCALL | instrmNEAR | instrmABS) // instr_name="CALLna"
#define instrCALLF_ABS			(instrgCTRL_TRANSFER | instrmCALL | instrmFAR  | instrmABS) // instr_name="CALLfa"

#define instrRETN			(instrgCTRL_TRANSFER | instrmRET  | instrmNEAR) // instr_name="RETn"
#define instrRETF			(instrgCTRL_TRANSFER | instrmRET  | instrmFAR) // instr_name="RETf"

#define instrmCOND_COUNT    		0x0400

#define instrmJCC           		0x0100

#define instrmJECXZ         		(instrmCOND_COUNT | 0x0001)

#define instrmLOOP          		(instrmCOND_COUNT | 0x0002)
#define instrmLOOPZ         		(instrmCOND_COUNT | 0x0003)
#define instrmLOOPNZ        		(instrmCOND_COUNT | 0x0004)

#define INSTR_IS_COND_COUNT(instr) ((instr) & (instrmCTRL_TRANSFER_TYPE_MASK | instrmCOND_COUNT)) == (instrmJMP_COND | instrmCOND_COUNT)

#define instrmCOUNT_ECX         	0x0008  // when addr-size is 16, so count is CX and not ECX

#define instrmJMP_COND_MASK		0x000f

#define instrJCC			(instrgCTRL_TRANSFER | instrmJMP_COND | instrmJCC) // always relative // no_instr_name
#define instrJCXZ			(instrgCTRL_TRANSFER | instrmJMP_COND | instrmJECXZ) // always relative
#define instrJECXZ			(instrgCTRL_TRANSFER | instrmJMP_COND | instrmJECXZ | instrmCOUNT_ECX) // always relative

#define instrmCC_O			(0x0000) // O       - NO      
#define instrmCC_B			(0x0002) // B/NAE/C - NB/AE/NC
#define instrmCC_Z			(0x0004) // Z/E	    - NZ/NE   
#define instrmCC_BE			(0x0006) // BE/NA   - NBE/A   
#define instrmCC_S			(0x0008) // S	    - NS      
#define instrmCC_P			(0x000a) // P/PE    - NP/PO   
#define instrmCC_L			(0x000c) // L/NGE   - NL/GE   
#define instrmCC_LE			(0x000e) // LE/NG   - NLE/G   

#define instrmCC_COND_MASK		0x000f
#define instrmCC_COND_NEG_MASK		0x0001
#define instrmCC_COND_MASK_BOTH		0x000e // both conditions O and NO, Z and NZ and so on

#define INSTR_CC_GET_COND(instr) 	((instr) & instrmCC_COND_MASK)
#define INSTR_CC_IS_NEG(instr) 	 	(!!((instr) & instrmCC_COND_NEG_MASK))
#define INSTR_CC_GET_COND_BOTH(instr) 	((instr) & instrmCC_COND_MASK_BOTH)

#define instrJCC_0			(instrJCC | 0x0000) // instr_name="JO"
#define instrJCC_1			(instrJCC | 0x0001) // instr_name="JNO"
#define instrJCC_2			(instrJCC | 0x0002) // instr_name="JB/NAE/C"
#define instrJCC_3			(instrJCC | 0x0003) // instr_name="JNB/AE/NC"
#define instrJCC_4			(instrJCC | 0x0004) // instr_name="JZ/E"
#define instrJCC_5			(instrJCC | 0x0005) // instr_name="JNZ/NE"
#define instrJCC_6			(instrJCC | 0x0006) // instr_name="JBE/NA"
#define instrJCC_7			(instrJCC | 0x0007) // instr_name="JNBE/A"
#define instrJCC_8			(instrJCC | 0x0008) // instr_name="JS"
#define instrJCC_9			(instrJCC | 0x0009) // instr_name="JNS"
#define instrJCC_A			(instrJCC | 0x000a) // instr_name="JP/PE"
#define instrJCC_B			(instrJCC | 0x000b) // instr_name="JNP/PO"
#define instrJCC_C			(instrJCC | 0x000c) // instr_name="JL/NGE"
#define instrJCC_D			(instrJCC | 0x000d) // instr_name="JNL/GE"
#define instrJCC_E			(instrJCC | 0x000e) // instr_name="JLE/NG"
#define instrJCC_F			(instrJCC | 0x000f) // instr_name="JNLE/G"

#define instrLOOP16			(instrgCTRL_TRANSFER | instrmJMP_COND | instrmLOOP)

#define instrLOOP16Z			(instrgCTRL_TRANSFER | instrmJMP_COND | instrmLOOPZ)
#define instrLOOP16NZ			(instrgCTRL_TRANSFER | instrmJMP_COND | instrmLOOPNZ)

#define instrLOOP			(instrgCTRL_TRANSFER | instrmJMP_COND | instrmLOOP | instrmCOUNT_ECX)

#define instrLOOPZ			(instrgCTRL_TRANSFER | instrmJMP_COND | instrmLOOPZ | instrmCOUNT_ECX)
#define instrLOOPNZ			(instrgCTRL_TRANSFER | instrmJMP_COND | instrmLOOPNZ | instrmCOUNT_ECX)

#define instrINT3			(instrgCTRL_TRANSFER | instrmINT  | 0x0001)
#define instrINT			(instrgCTRL_TRANSFER | instrmINT  | 0x0002)
#define instrINTO			(instrgCTRL_TRANSFER | instrmINT  | 0x0003)
#define instrIRET			(instrgCTRL_TRANSFER | instrmIRET | 0x0004)

//////////////////////////////////////////////////////////////////////

#define instrgNOP			0x1000

#define instrNOP			(instrgNOP | 0x0001)
#define instrUD2			(instrgNOP | 0x0002) // not to be used as nop... (but is halfway a nop)
#define instrFNOP			(instrgNOP | 0x0003)

//////////////////////////////////////////////////////////////////////

#define instrgINT_ARITH			0x1800

#define instrgmINT_ARITH_MEMORY_LOCK	0x0100

#define instrADD			(instrgINT_ARITH | 0x0000) 
#define instrOR				(instrgINT_ARITH | 0x0001) 
#define instrADC			(instrgINT_ARITH | 0x0002) 
#define instrSBB			(instrgINT_ARITH | 0x0003) 
#define instrAND			(instrgINT_ARITH | 0x0004) 
#define instrSUB			(instrgINT_ARITH | 0x0005) 
#define instrXOR			(instrgINT_ARITH | 0x0006) 
#define instrCMP			(instrgINT_ARITH | 0x0007) 

#define instrINC			(instrgINT_ARITH | 0x0009) 
#define instrDEC			(instrgINT_ARITH | 0x000a) 

#define instrTEST			(instrgINT_ARITH | 0x000b) 

#define instrNOT			(instrgINT_ARITH | 0x000c)
#define instrNEG			(instrgINT_ARITH | 0x000d)

#define instrXADD			(instrgINT_ARITH | 0x000e)

#define instrAAA			(instrgINT_ARITH | 0x0011)
#define instrAAD			(instrgINT_ARITH | 0x0012)
#define instrAAM			(instrgINT_ARITH | 0x0013)
#define instrAAS			(instrgINT_ARITH | 0x0014)
#define instrDAA			(instrgINT_ARITH | 0x0015)
#define instrDAS			(instrgINT_ARITH | 0x0016)

#define instrMUL			(instrgINT_ARITH | 0x0021) 
#define instrIMUL			(instrgINT_ARITH | 0x0022) 
#define instrDIV			(instrgINT_ARITH | 0x0023) 
#define instrIDIV			(instrgINT_ARITH | 0x0024) 

#define instrLEA			(instrgINT_ARITH | 0x0031) 

#define instrCBWCWDE			(instrgINT_ARITH | 0x0041)
#define instrCWDCDQ			(instrgINT_ARITH | 0x0042)

#define instrROL			(instrgINT_ARITH | 0x0051)
#define instrROR			(instrgINT_ARITH | 0x0052)
#define instrRCL			(instrgINT_ARITH | 0x0053)
#define instrRCR			(instrgINT_ARITH | 0x0054)
#define instrSHL			(instrgINT_ARITH | 0x0055)
#define instrSHR			(instrgINT_ARITH | 0x0056)
//#define instrSAL			(instrgINT_ARITH | 0x0057) // no_instr_name
#define instrSAR			(instrgINT_ARITH | 0x0058)

#define instrSHLD			(instrgINT_ARITH | 0x0059)
#define instrSHRD			(instrgINT_ARITH | 0x005a)

#define instrBT				(instrgINT_ARITH | 0x0060)
#define instrBTR			(instrgINT_ARITH | 0x0061)
#define instrBTC			(instrgINT_ARITH | 0x0062)
#define instrBTS			(instrgINT_ARITH | 0x0063)

#define instrBSF			(instrgINT_ARITH | 0x0064)
#define instrBSR			(instrgINT_ARITH | 0x0065)

#define instrBSWAP			(instrgINT_ARITH | 0x0071)

#define instrLOCK_ADD			(instrgmINT_ARITH_MEMORY_LOCK | instrADD) 
#define instrLOCK_OR			(instrgmINT_ARITH_MEMORY_LOCK | instrOR) 
#define instrLOCK_ADC			(instrgmINT_ARITH_MEMORY_LOCK | instrADC) 
#define instrLOCK_SBB			(instrgmINT_ARITH_MEMORY_LOCK | instrSBB) 
#define instrLOCK_AND			(instrgmINT_ARITH_MEMORY_LOCK | instrAND) 
#define instrLOCK_SUB			(instrgmINT_ARITH_MEMORY_LOCK | instrSUB) 
#define instrLOCK_XOR			(instrgmINT_ARITH_MEMORY_LOCK | instrXOR) 

#define instrLOCK_INC			(instrgmINT_ARITH_MEMORY_LOCK | instrINC) 
#define instrLOCK_DEC			(instrgmINT_ARITH_MEMORY_LOCK | instrDEC) 

#define instrLOCK_NOT			(instrgmINT_ARITH_MEMORY_LOCK | instrNOT)
#define instrLOCK_NEG			(instrgmINT_ARITH_MEMORY_LOCK | instrNEG)

#define instrLOCK_XADD			(instrgmINT_ARITH_MEMORY_LOCK | instrXADD)

#define instrLOCK_BTR			(instrgmINT_ARITH_MEMORY_LOCK | instrBTR)
#define instrLOCK_BTC			(instrgmINT_ARITH_MEMORY_LOCK | instrBTC)
#define instrLOCK_BTS			(instrgmINT_ARITH_MEMORY_LOCK | instrBTS)

//////////////////////////////////////////////////////////////////////

#define instrgSTACK			0x2000

#define instrmPUSH			0x0001
#define instrmPOP			0x0002

#define instrPUSH			(instrgSTACK | instrmPUSH | 0x0010) 
#define instrPOP			(instrgSTACK | instrmPOP  | 0x0010) 

#define instrPUSHA			(instrgSTACK | instrmPUSH | 0x0020) 
#define instrPOPA			(instrgSTACK | instrmPOP  | 0x0020) 

#define instrPUSHF			(instrgSTACK | instrmPUSH | 0x0030) 
#define instrPOPF			(instrgSTACK | instrmPOP  | 0x0030) 

#define instrENTER			(instrgSTACK | 0x0041)
#define instrLEAVE			(instrgSTACK | 0x0042)

//////////////////////////////////////////////////////////////////////

#define instrgMEMORY			0x2800

#define instrMOV			(instrgMEMORY | 0x0001) 
#define instrXCHG			(instrgMEMORY | 0x0002) 
#define instrBOUND			(instrgMEMORY | 0x0003) 
#define instrXLAT			(instrgMEMORY | 0x0004) 

#define instrLDS			(instrgMEMORY | 0x0011) 
#define instrLES			(instrgMEMORY | 0x0012) 
#define instrLFS			(instrgMEMORY | 0x0013) 
#define instrLGS			(instrgMEMORY | 0x0014) 
#define instrLSS			(instrgMEMORY | 0x0015) 

#define instrMOVZX			(instrgMEMORY | 0x0021)
#define instrMOVSX			(instrgMEMORY | 0x0022)

#define instrCMPXCHG			(instrgMEMORY | 0x0031)
#define instrCMPXCH8B			(instrgMEMORY | 0x0032)

#define instrVERR			(instrgMEMORY | 0x0041)
#define instrVERW			(instrgMEMORY | 0x0042)

#define instrLOCK_XCHG			(instrgmINT_ARITH_MEMORY_LOCK | instrXCHG) 

#define instrLOCK_CMPXCHG		(instrgmINT_ARITH_MEMORY_LOCK | instrCMPXCHG)
#define instrLOCK_CMPXCH8B		(instrgmINT_ARITH_MEMORY_LOCK | instrCMPXCH8B)

//////////////////////////////////////////////////////////////////////

#define instrgSTRING			0x3000

#define instrgmSTRING_REP     		0x10
#define instrgmSTRING_REPE    		0x20
#define instrgmSTRING_REPNE   		0x40

#define instrCMPS			(instrgSTRING | 0x0001) 
#define instrMOVS			(instrgSTRING | 0x0002) 

#define instrINS			(instrgSTRING | 0x0003) 
#define instrOUTS			(instrgSTRING | 0x0004) 

#define instrLODS			(instrgSTRING | 0x0005) 
#define instrSTOS			(instrgSTRING | 0x0006) 
#define instrSCAS			(instrgSTRING | 0x0007) 

#define instrREPNE_CMPS			(instrgmSTRING_REPNE | instrCMPS) 
#define instrREPNE_SCAS			(instrgmSTRING_REPNE | instrSCAS) 

#define instrREPE_CMPS			(instrgmSTRING_REPE | instrCMPS) 
#define instrREPE_SCAS			(instrgmSTRING_REPE | instrSCAS) 

#define instrREP_MOVS			(instrgmSTRING_REP | instrMOVS) 

#define instrREP_INS			(instrgmSTRING_REP | instrINS) 
#define instrREP_OUTS			(instrgmSTRING_REP | instrOUTS) 

#define instrREP_LODS			(instrgmSTRING_REP | instrLODS) 
#define instrREP_STOS			(instrgmSTRING_REP | instrSTOS) 

//////////////////////////////////////////////////////////////////////

#define instrgFP			0x3800

#define instrmFP_ADD			0x0001
#define instrmFP_MUL			0x0002
#define instrmFP_SUB			0x0003
#define instrmFP_SUBR			0x0004
#define instrmFP_DIV			0x0005
#define instrmFP_DIVR			0x0006
#define instrmFP_COM			0x0007

#define instrmFP_LD			0x0008
#define instrmFP_ST			0x0009

#define instrmFP_LD_SPEC		0x000a

#define instrmFP_XCH			0x000b

#define instrmFP_ONE_OP			0x000c

#define instrmFP_OTHER			0x000d

#define instrmFP_FCMOV			0x000f

#define instrmFP_INTEGER_OP		0x0010  // modifiers for instrmFP_ADD -- instrmFP_COM
#define instrmFP_BCD_OP			0x0020  // if not integer or bcd -> real
#define instrmFP_POP			0x0040
#define instrmFP_POP_POP		0x0080
#define instrmFP_EFLAGS			0x0100
#define instrmFP_UNORDERED		0x0200

#define instrFXCH			(instrgFP | instrmFP_XCH)

#define instrFADD			(instrgFP | instrmFP_ADD  )
#define instrFMUL			(instrgFP | instrmFP_MUL  )
#define instrFSUB			(instrgFP | instrmFP_SUB  )
#define instrFSUBR			(instrgFP | instrmFP_SUBR )
#define instrFDIV			(instrgFP | instrmFP_DIV  )
#define instrFDIVR			(instrgFP | instrmFP_DIVR )

#define instrFIADD			(instrgFP | instrmFP_ADD  | instrmFP_INTEGER_OP)
#define instrFIMUL			(instrgFP | instrmFP_MUL  | instrmFP_INTEGER_OP)
#define instrFISUB			(instrgFP | instrmFP_SUB  | instrmFP_INTEGER_OP)
#define instrFISUBR			(instrgFP | instrmFP_SUBR | instrmFP_INTEGER_OP)
#define instrFIDIV			(instrgFP | instrmFP_DIV  | instrmFP_INTEGER_OP)
#define instrFIDIVR			(instrgFP | instrmFP_DIVR | instrmFP_INTEGER_OP)

#define instrFADDP			(instrgFP | instrmFP_ADD  | instrmFP_POP)
#define instrFMULP			(instrgFP | instrmFP_MUL  | instrmFP_POP)
#define instrFSUBP			(instrgFP | instrmFP_SUB  | instrmFP_POP)
#define instrFSUBRP			(instrgFP | instrmFP_SUBR | instrmFP_POP)
#define instrFDIVP			(instrgFP | instrmFP_DIV  | instrmFP_POP)
#define instrFDIVRP			(instrgFP | instrmFP_DIVR | instrmFP_POP)

#define instrFCOM			(instrgFP | instrmFP_COM)
#define instrFCOMP			(instrgFP | instrmFP_COM                      | instrmFP_POP)
#define instrFCOMPP			(instrgFP | instrmFP_COM                      | instrmFP_POP_POP)

#define instrFCOMI			(instrgFP | instrmFP_COM                                     | instrmFP_EFLAGS)
#define instrFCOMIP			(instrgFP | instrmFP_COM                      | instrmFP_POP | instrmFP_EFLAGS)

#define instrFUCOM			(instrgFP | instrmFP_COM | instrmFP_UNORDERED)
#define instrFUCOMP			(instrgFP | instrmFP_COM | instrmFP_UNORDERED | instrmFP_POP)
#define instrFUCOMPP			(instrgFP | instrmFP_COM | instrmFP_UNORDERED | instrmFP_POP_POP)

#define instrFUCOMI			(instrgFP | instrmFP_COM | instrmFP_UNORDERED                | instrmFP_EFLAGS)
#define instrFUCOMIP			(instrgFP | instrmFP_COM | instrmFP_UNORDERED | instrmFP_POP | instrmFP_EFLAGS)

#define instrFICOM			(instrgFP | instrmFP_COM                                     | instrmFP_INTEGER_OP)
#define instrFICOMP			(instrgFP | instrmFP_COM                      | instrmFP_POP | instrmFP_INTEGER_OP)

#define instrFLD			(instrgFP | instrmFP_LD)
#define instrFILD			(instrgFP | instrmFP_LD               | instrmFP_INTEGER_OP)
#define instrFBLD			(instrgFP | instrmFP_LD               | instrmFP_BCD_OP)

#define instrFST			(instrgFP | instrmFP_ST)
#define instrFSTP			(instrgFP | instrmFP_ST | instrmFP_POP)
#define instrFIST			(instrgFP | instrmFP_ST                | instrmFP_INTEGER_OP)
#define instrFISTP			(instrgFP | instrmFP_ST | instrmFP_POP | instrmFP_INTEGER_OP)
#define instrFBSTP			(instrgFP | instrmFP_ST | instrmFP_POP | instrmFP_BCD_OP)

#define instrFLD1			(instrgFP | instrmFP_LD_SPEC | 0x0010)
#define instrFLDL2T			(instrgFP | instrmFP_LD_SPEC | 0x0020)
#define instrFLDL2E			(instrgFP | instrmFP_LD_SPEC | 0x0030)
#define instrFLDPI			(instrgFP | instrmFP_LD_SPEC | 0x0040)
#define instrFLDLG2			(instrgFP | instrmFP_LD_SPEC | 0x0050)
#define instrFLDLN2			(instrgFP | instrmFP_LD_SPEC | 0x0060)
#define instrFLDZ			(instrgFP | instrmFP_LD_SPEC | 0x0070)

#define instrFCMOVB			(instrgFP | instrmFP_FCMOV | 0x0000)
#define instrFCMOVE			(instrgFP | instrmFP_FCMOV | 0x0010)
#define instrFCMOVBE			(instrgFP | instrmFP_FCMOV | 0x0020)
#define instrFCMOVU			(instrgFP | instrmFP_FCMOV | 0x0030)
#define instrFCMOVNB			(instrgFP | instrmFP_FCMOV | 0x0040)
#define instrFCMOVNE			(instrgFP | instrmFP_FCMOV | 0x0050)
#define instrFCMOVNBE			(instrgFP | instrmFP_FCMOV | 0x0060)
#define instrFCMOVNU			(instrgFP | instrmFP_FCMOV | 0x0070)

#define instrFCHS			(instrgFP | instrmFP_ONE_OP | 0x0010)
#define instrFABS			(instrgFP | instrmFP_ONE_OP | 0x0020)
#define instrFTST			(instrgFP | instrmFP_ONE_OP | 0x0030)
#define instrFXAM			(instrgFP | instrmFP_ONE_OP | 0x0040)

#define instrF2XM1			(instrgFP | instrmFP_ONE_OP | 0x0100)
#define instrFYL2X			(instrgFP | instrmFP_ONE_OP | 0x0110)
#define instrFPTAN			(instrgFP | instrmFP_ONE_OP | 0x0120)
#define instrFPATAN			(instrgFP | instrmFP_ONE_OP | 0x0130)
#define instrFXTRACT			(instrgFP | instrmFP_ONE_OP | 0x0140)
#define instrFPREM1			(instrgFP | instrmFP_ONE_OP | 0x0150)
#define instrFDECSTP			(instrgFP | instrmFP_ONE_OP | 0x0160)
#define instrFINCSTP			(instrgFP | instrmFP_ONE_OP | 0x0170)
#define instrFPREM			(instrgFP | instrmFP_ONE_OP | 0x0180)
#define instrFYL2XP1			(instrgFP | instrmFP_ONE_OP | 0x0190)
#define instrFSQRT			(instrgFP | instrmFP_ONE_OP | 0x01a0)
#define instrFSINCOS			(instrgFP | instrmFP_ONE_OP | 0x01b0)
#define instrFRNDINT			(instrgFP | instrmFP_ONE_OP | 0x01c0)
#define instrFSCALE			(instrgFP | instrmFP_ONE_OP | 0x01d0)
#define instrFSIN			(instrgFP | instrmFP_ONE_OP | 0x01e0)
#define instrFCOS			(instrgFP | instrmFP_ONE_OP | 0x01f0)

#define instrFLDENV			(instrgFP | instrmFP_OTHER | 0x0010)
#define instrFSTENV			(instrgFP | instrmFP_OTHER | 0x0020)

#define instrFLDCW			(instrgFP | instrmFP_OTHER | 0x0030)
#define instrFSTCW			(instrgFP | instrmFP_OTHER | 0x0040)

#define instrFRSTOR			(instrgFP | instrmFP_OTHER | 0x0050)
#define instrFSAVE			(instrgFP | instrmFP_OTHER | 0x0060)

#define instrFSTSW			(instrgFP | instrmFP_OTHER | 0x0070)

#define instrFCLEX			(instrgFP | instrmFP_OTHER | 0x0080)
#define instrFNINIT			(instrgFP | instrmFP_OTHER | 0x0090)
#define instrFFREE			(instrgFP | instrmFP_OTHER | 0x00a0)

#define instrFWAIT			(instrgFP | instrmFP_OTHER | 0x00b0) 

//////////////////////////////////////////////////////////////////////

#define instrgMMXSSE			0x4000

#define instrADDPD		        (instrgMMXSSE | 0x0001)
#define instrADDPS		        (instrgMMXSSE | 0x0002)
#define instrADDSD		        (instrgMMXSSE | 0x0003)
#define instrADDSS		        (instrgMMXSSE | 0x0004)
#define instrANDNPD		        (instrgMMXSSE | 0x0005)
#define instrANDNPS		        (instrgMMXSSE | 0x0006)
#define instrANDPD		        (instrgMMXSSE | 0x0007)
#define instrANDPS		        (instrgMMXSSE | 0x0008)
#define instrCMPPD		        (instrgMMXSSE | 0x0009)
#define instrCMPPS		        (instrgMMXSSE | 0x000a)
#define instrCMPSD		        (instrgMMXSSE | 0x000b)
#define instrCMPSS		        (instrgMMXSSE | 0x000c)
#define instrCOMISD		        (instrgMMXSSE | 0x000d)
#define instrCOMISS		        (instrgMMXSSE | 0x000e)
#define instrCVTDQ2PD		        (instrgMMXSSE | 0x000f)
#define instrCVTDQ2PS		        (instrgMMXSSE | 0x0010)
#define instrCVTPD2DQ		        (instrgMMXSSE | 0x0011)
#define instrCVTPD2PI		        (instrgMMXSSE | 0x0012)
#define instrCVTPD2PS		        (instrgMMXSSE | 0x0013)
#define instrCVTPI2PD		        (instrgMMXSSE | 0x0014)
#define instrCVTPI2PS		        (instrgMMXSSE | 0x0015)
#define instrCVTPS2DQ		        (instrgMMXSSE | 0x0016)
#define instrCVTPS2PD		        (instrgMMXSSE | 0x0017)
#define instrCVTPS2PI		        (instrgMMXSSE | 0x0018)
#define instrCVTSD2SI		        (instrgMMXSSE | 0x0019)
#define instrCVTSD2SS		        (instrgMMXSSE | 0x001a)
#define instrCVTSI2SD		        (instrgMMXSSE | 0x001b)
#define instrCVTSI2SS		        (instrgMMXSSE | 0x001c)
#define instrCVTSS2SD		        (instrgMMXSSE | 0x001d)
#define instrCVTSS2SI		        (instrgMMXSSE | 0x001e)
#define instrCVTTPD2DQ		        (instrgMMXSSE | 0x001f)
#define instrCVTTPD2PI		        (instrgMMXSSE | 0x0020)
#define instrCVTTPS2DQ		        (instrgMMXSSE | 0x0021)
#define instrCVTTPS2PI		        (instrgMMXSSE | 0x0022)
#define instrCVTTSD2SI		        (instrgMMXSSE | 0x0023)
#define instrCVTTSS2SI		        (instrgMMXSSE | 0x0024)
#define instrDIVPD			(instrgMMXSSE | 0x0025)
#define instrDIVPS			(instrgMMXSSE | 0x0026)
#define instrDIVSD			(instrgMMXSSE | 0x0027)
#define instrDIVSS			(instrgMMXSSE | 0x0028)
#define instrEMMS			(instrgMMXSSE | 0x0029)
#define instrMASKMOVDQU		        (instrgMMXSSE | 0x002a)
#define instrMASKMOVQ		        (instrgMMXSSE | 0x002b)
#define instrMAXPD			(instrgMMXSSE | 0x002c)
#define instrMAXPS			(instrgMMXSSE | 0x002d)
#define instrMAXSD			(instrgMMXSSE | 0x002e)
#define instrMAXSS			(instrgMMXSSE | 0x002f)
#define instrMINPD			(instrgMMXSSE | 0x0030)
#define instrMINPS			(instrgMMXSSE | 0x0031)
#define instrMINSD			(instrgMMXSSE | 0x0032)
#define instrMINSS			(instrgMMXSSE | 0x0033)
#define instrMOVAPD			(instrgMMXSSE | 0x0034)
#define instrMOVAPS			(instrgMMXSSE | 0x0035)
#define instrMOVD			(instrgMMXSSE | 0x0036)
#define instrMOVDQ2Q		        (instrgMMXSSE | 0x0037)
#define instrMOVDQA			(instrgMMXSSE | 0x0038)
#define instrMOVDQU			(instrgMMXSSE | 0x0039)
#define instrMOVHLPS		        (instrgMMXSSE | 0x003a)
#define instrMOVHPD			(instrgMMXSSE | 0x003b)
#define instrMOVHPS			(instrgMMXSSE | 0x003c)
#define instrMOVLHPS		        (instrgMMXSSE | 0x003d)
#define instrMOVLPD			(instrgMMXSSE | 0x003e)
#define instrMOVLPS			(instrgMMXSSE | 0x003f)
#define instrMOVMSKPD		        (instrgMMXSSE | 0x0042)
#define instrMOVMSKPS		        (instrgMMXSSE | 0x0043)
#define instrMOVNTDQ		        (instrgMMXSSE | 0x0044)
#define instrMOVNTI			(instrgMMXSSE | 0x0045)
#define instrMOVNTPD		        (instrgMMXSSE | 0x0046)
#define instrMOVNTPS		        (instrgMMXSSE | 0x0047)
#define instrMOVNTQ			(instrgMMXSSE | 0x0048)
#define instrMOVQ			(instrgMMXSSE | 0x0049)
#define instrMOVQ2DQ		        (instrgMMXSSE | 0x004c)
#define instrMOVSD			(instrgMMXSSE | 0x004d)
#define instrMOVSS			(instrgMMXSSE | 0x004e)
#define instrMOVUPD			(instrgMMXSSE | 0x004f)
#define instrMOVUPS			(instrgMMXSSE | 0x0050)
#define instrMULPD			(instrgMMXSSE | 0x0051)
#define instrMULPS			(instrgMMXSSE | 0x0052)
#define instrMULSD			(instrgMMXSSE | 0x0053)
#define instrMULSS			(instrgMMXSSE | 0x0054)
#define instrORPD			(instrgMMXSSE | 0x0055)
#define instrORPS			(instrgMMXSSE | 0x0056)
#define instrPACKSSDW		        (instrgMMXSSE | 0x0057)
#define instrPACKSSWB		        (instrgMMXSSE | 0x0058)
#define instrPACKUSWB		        (instrgMMXSSE | 0x0059)
#define instrPADDB			(instrgMMXSSE | 0x005a)
#define instrPADDD			(instrgMMXSSE | 0x005b)
#define instrPADDQ			(instrgMMXSSE | 0x005c)
#define instrPADDSB			(instrgMMXSSE | 0x005d)
#define instrPADDSW			(instrgMMXSSE | 0x005e)
#define instrPADDUSB		        (instrgMMXSSE | 0x005f)
#define instrPADDUSW		        (instrgMMXSSE | 0x0060)
#define instrPADDW			(instrgMMXSSE | 0x0061)
#define instrPAND			(instrgMMXSSE | 0x0062)
#define instrPANDN			(instrgMMXSSE | 0x0063)
#define instrPAVGB			(instrgMMXSSE | 0x0064)
#define instrPAVGW			(instrgMMXSSE | 0x0065)
#define instrPCMPEQB		        (instrgMMXSSE | 0x0066)
#define instrPCMPEQD		        (instrgMMXSSE | 0x0067)
#define instrPCMPEQW		        (instrgMMXSSE | 0x0068)
#define instrPCMPGTB		        (instrgMMXSSE | 0x0069)
#define instrPCMPGTD		        (instrgMMXSSE | 0x006a)
#define instrPCMPGTW		        (instrgMMXSSE | 0x006b)
#define instrPEXTRW			(instrgMMXSSE | 0x006c)
#define instrPINSRW			(instrgMMXSSE | 0x006d)
#define instrPMADDWD		        (instrgMMXSSE | 0x006f)
#define instrPMAXSW			(instrgMMXSSE | 0x0070)
#define instrPMAXUB			(instrgMMXSSE | 0x0071)
#define instrPMINSW			(instrgMMXSSE | 0x0072)
#define instrPMINUB			(instrgMMXSSE | 0x0073)
#define instrPMOVMSKB		        (instrgMMXSSE | 0x0074)
#define instrPMULHUW		        (instrgMMXSSE | 0x0075)
#define instrPMULHW			(instrgMMXSSE | 0x0076)
#define instrPMULLW			(instrgMMXSSE | 0x0077)
#define instrPMULUDQ		        (instrgMMXSSE | 0x0078)
#define instrPOR			(instrgMMXSSE | 0x0079)
#define instrPSADBW			(instrgMMXSSE | 0x007a)
#define instrPSHUFD			(instrgMMXSSE | 0x007b)
#define instrPSHUFHW		        (instrgMMXSSE | 0x007c)
#define instrPSHUFLW		        (instrgMMXSSE | 0x007d)
#define instrPSHUFW			(instrgMMXSSE | 0x007e)
#define instrPSLLD			(instrgMMXSSE | 0x007f)
#define instrPSLLDQ			(instrgMMXSSE | 0x0081)
#define instrPSLLQ			(instrgMMXSSE | 0x0082)
#define instrPSLLW			(instrgMMXSSE | 0x0083)
#define instrPSRAD			(instrgMMXSSE | 0x0084)
#define instrPSRAW			(instrgMMXSSE | 0x0086)
#define instrPSRLD			(instrgMMXSSE | 0x0087)
#define instrPSRLDQ			(instrgMMXSSE | 0x0089)
#define instrPSRLQ			(instrgMMXSSE | 0x008a)
#define instrPSRLW			(instrgMMXSSE | 0x008b)
#define instrPSUBB			(instrgMMXSSE | 0x008c)
#define instrPSUBD			(instrgMMXSSE | 0x008d)
#define instrPSUBQ			(instrgMMXSSE | 0x008e)
#define instrPSUBSB			(instrgMMXSSE | 0x008f)
#define instrPSUBSW			(instrgMMXSSE | 0x0090)
#define instrPSUBUSB		        (instrgMMXSSE | 0x0091)
#define instrPSUBUSW		        (instrgMMXSSE | 0x0092)
#define instrPSUBW			(instrgMMXSSE | 0x0093)
#define instrPUNPCKHBW		        (instrgMMXSSE | 0x0094)
#define instrPUNPCKHDQ		        (instrgMMXSSE | 0x0095)
#define instrPUNPCKHQDQ		        (instrgMMXSSE | 0x0096)
#define instrPUNPCKHWD		        (instrgMMXSSE | 0x0097)
#define instrPUNPCKLBW		        (instrgMMXSSE | 0x0098)
#define instrPUNPCKLDQ		        (instrgMMXSSE | 0x0099)
#define instrPUNPCKLQDQ		        (instrgMMXSSE | 0x009a)
#define instrPUNPCKLWD		        (instrgMMXSSE | 0x009b)
#define instrPXOR			(instrgMMXSSE | 0x009c)
#define instrRCPPS			(instrgMMXSSE | 0x009d)
#define instrRCPSS			(instrgMMXSSE | 0x009e)
#define instrRSQRTPS		        (instrgMMXSSE | 0x009f)
#define instrRSQRTSS		        (instrgMMXSSE | 0x00a0)
#define instrSHUFPD			(instrgMMXSSE | 0x00a1)
#define instrSHUFPS			(instrgMMXSSE | 0x00a2)
#define instrSQRTPD			(instrgMMXSSE | 0x00a3)
#define instrSQRTPS			(instrgMMXSSE | 0x00a4)
#define instrSQRTSD			(instrgMMXSSE | 0x00a5)
#define instrSQRTSS			(instrgMMXSSE | 0x00a6)
#define instrSUBPD			(instrgMMXSSE | 0x00a7)
#define instrSUBPS			(instrgMMXSSE | 0x00a8)
#define instrSUBSD			(instrgMMXSSE | 0x00a9)
#define instrSUBSS			(instrgMMXSSE | 0x00aa)
#define instrUCOMISD		        (instrgMMXSSE | 0x00ab)
#define instrUCOMISS		        (instrgMMXSSE | 0x00ac)
#define instrUNPCKHPD		        (instrgMMXSSE | 0x00ad)
#define instrUNPCKHPS		        (instrgMMXSSE | 0x00ae)
#define instrUNPCKLPD		        (instrgMMXSSE | 0x00af)
#define instrUNPCKLPS		        (instrgMMXSSE | 0x00b0)
#define instrXORPD			(instrgMMXSSE | 0x00b1)
#define instrXORPS			(instrgMMXSSE | 0x00b2)

//////////////////////////////////////////////////////////////////////

#define instrgCOND			0x4800

#define instrCMOVCC			(instrgCOND | 0x0010) // no_instr_name
#define instrSETCC			(instrgCOND | 0x0020) // no_instr_name

#define instrCMOVCC_0		        (instrCMOVCC | 0x0000) // instr_name="CMOVO"        
#define instrCMOVCC_1		        (instrCMOVCC | 0x0001) // instr_name="CMOVNO"
#define instrCMOVCC_2		        (instrCMOVCC | 0x0002) // instr_name="CMOVB/NAE/C"
#define instrCMOVCC_3		        (instrCMOVCC | 0x0003) // instr_name="CMOVNB/AE/NC"
#define instrCMOVCC_4		        (instrCMOVCC | 0x0004) // instr_name="CMOVZ/E"
#define instrCMOVCC_5		        (instrCMOVCC | 0x0005) // instr_name="CMOVNZ/NE"
#define instrCMOVCC_6		        (instrCMOVCC | 0x0006) // instr_name="CMOVBE/NA"
#define instrCMOVCC_7		        (instrCMOVCC | 0x0007) // instr_name="CMOVNBE/A"
#define instrCMOVCC_8		        (instrCMOVCC | 0x0008) // instr_name="CMOVS"
#define instrCMOVCC_9		        (instrCMOVCC | 0x0009) // instr_name="CMOVNS"
#define instrCMOVCC_A		        (instrCMOVCC | 0x000A) // instr_name="CMOVP/PE"
#define instrCMOVCC_B		        (instrCMOVCC | 0x000B) // instr_name="CMOVNP/PO"
#define instrCMOVCC_C		        (instrCMOVCC | 0x000C) // instr_name="CMOVL/NGE"
#define instrCMOVCC_D		        (instrCMOVCC | 0x000D) // instr_name="CMOVNL/GE"
#define instrCMOVCC_E		        (instrCMOVCC | 0x000E) // instr_name="CMOVLE/NG"
#define instrCMOVCC_F		        (instrCMOVCC | 0x000F) // instr_name="CMOVNLE/G"

#define instrSETCC_0		        (instrSETCC | 0x0000) // instr_name="SETO"
#define instrSETCC_1		        (instrSETCC | 0x0001) // instr_name="SETNO"
#define instrSETCC_2		        (instrSETCC | 0x0002) // instr_name="SETB/NAE/C"
#define instrSETCC_3		        (instrSETCC | 0x0003) // instr_name="SETNB/AE/NC"
#define instrSETCC_4		        (instrSETCC | 0x0004) // instr_name="SETZ/E"
#define instrSETCC_5		        (instrSETCC | 0x0005) // instr_name="SETNZ/NE"
#define instrSETCC_6		        (instrSETCC | 0x0006) // instr_name="SETBE/NA"
#define instrSETCC_7		        (instrSETCC | 0x0007) // instr_name="SETNBE/A"
#define instrSETCC_8		        (instrSETCC | 0x0008) // instr_name="SETS"
#define instrSETCC_9		        (instrSETCC | 0x0009) // instr_name="SETNS"
#define instrSETCC_A		        (instrSETCC | 0x000A) // instr_name="SETP/PE"
#define instrSETCC_B		        (instrSETCC | 0x000B) // instr_name="SETNP/PO"
#define instrSETCC_C		        (instrSETCC | 0x000C) // instr_name="SETL/NGE"
#define instrSETCC_D		        (instrSETCC | 0x000D) // instr_name="SETNL/GE"
#define instrSETCC_E		        (instrSETCC | 0x000E) // instr_name="SETLE/NG"
#define instrSETCC_F		        (instrSETCC | 0x000F) // instr_name="SETNLE/G"

//////////////////////////////////////////////////////////////////////

#define instrgOTHER			0x5000

#define instrARPL			(instrgOTHER | 0x0003) 

#define instrSAHF			(instrgOTHER | 0x0004) 
#define instrLAHF			(instrgOTHER | 0x0005) 

#define instrIN				(instrgOTHER | 0x0007) 
#define instrOUT			(instrgOTHER | 0x0008) 

#define instrHLT			(instrgOTHER | 0x0009)

#define instrCMC			(instrgOTHER | 0x0011)

#define instrCLC			(instrgOTHER | 0x0012)
#define instrSTC			(instrgOTHER | 0x0013)
#define instrCLI			(instrgOTHER | 0x0014)
#define instrSTI			(instrgOTHER | 0x0015)
#define instrCLD			(instrgOTHER | 0x0016)
#define instrSTD			(instrgOTHER | 0x0017)

#define instrCLTS			(instrgOTHER | 0x0018)

#define instrINVD			(instrgOTHER | 0x0021)
#define instrWBINVD			(instrgOTHER | 0x0022)

#define instrCPUID			(instrgOTHER | 0x0023)

#define instrLAR			(instrgOTHER | 0x0024)

#define instrLSL			(instrgOTHER | 0x0025)

#define instrRDPMC			(instrgOTHER | 0x0031)

#define instrRDTSC			(instrgOTHER | 0x0032)

#define instrRDMSR			(instrgOTHER | 0x0033)
#define instrWRMSR			(instrgOTHER | 0x0034)

#define instrSYSENTER		        (instrgOTHER | 0x0035)
#define instrSYSEXIT		        (instrgOTHER | 0x0036)

#define instrRSM			(instrgOTHER | 0x0037)

#define instrSLDT			(instrgOTHER | 0x0041)
#define instrLLDT			(instrgOTHER | 0x0042)

#define instrSTR			(instrgOTHER | 0x0043)
#define instrLTR			(instrgOTHER | 0x0044)

#define instrSGDT			(instrgOTHER | 0x0045)
#define instrLGDT			(instrgOTHER | 0x0046)
#define instrSIDT			(instrgOTHER | 0x0047)
#define instrLIDT			(instrgOTHER | 0x0048)
#define instrSMSW			(instrgOTHER | 0x0049)
#define instrLMSW			(instrgOTHER | 0x004a)

#define instrINVLPG			(instrgOTHER | 0x0051)

#define instrPREFETCHNTA	        (instrgOTHER | 0x0061)
#define instrPREFETCHT0		        (instrgOTHER | 0x0062)
#define instrPREFETCHT1		        (instrgOTHER | 0x0063)
#define instrPREFETCHT2		        (instrgOTHER | 0x0064)

#define instrCLFLUSH		        (instrgOTHER | 0x0071)

#define instrMFENCE			(instrgOTHER | 0x0081)
#define instrLFENCE			(instrgOTHER | 0x0082)
#define instrSFENCE			(instrgOTHER | 0x0083)

#define instrFXSAVE			(instrgOTHER | 0x0091)
#define instrFXRSTOR		        (instrgOTHER | 0x0092)
#define instrLDMXCSR		        (instrgOTHER | 0x0093)
#define instrSTMXCSR		        (instrgOTHER | 0x0094)

//////////////////////////////////////////////////////////////////////

#define instrgNOT_INSTR		        0x8000 // no_instr_name

#define instrRESERVED		        (instrgNOT_INSTR | 0x0001) // no_instr_name, instruction is not to be used
#define instrPREFIX			(instrgNOT_INSTR | 0x0002) // no_instr_name
#define instrESCAPE2BYTE	        (instrgNOT_INSTR | 0x0003) // no_instr_name

#define instrgGROUP			(instrgNOT_INSTR | 0x0010)

#define instrmIS_GROUP_MASK		(0xfff0)

#define instrmGROUP_TYPE_MASK	        (0x000f)

#define instrmGROUP_MOD_MEM		(0x0001)
#define instrmGROUP_MOD_11		(0x0002)
#define instrmGROUP_MOD_BOTH	        (0x0004)
#define instrmGROUP_COPROC		(0x0008)

#define instrGROUP			(instrgGROUP | instrmGROUP_MOD_BOTH) // no_instr_name
#define instrGROUP_MOD_MEM		(instrgGROUP | instrmGROUP_MOD_MEM) // no_instr_name
#define instrGROUP_MOD_11		(instrgGROUP | instrmGROUP_MOD_11) // no_instr_name
#define instrGROUP_MOD_MEM_11	        (instrgGROUP | instrmGROUP_MOD_MEM | instrmGROUP_MOD_11) // no_instr_name

#define instrESCAPE_COPROC		(instrgGROUP | instrmGROUP_MOD_MEM | instrmGROUP_MOD_11 | instrmGROUP_COPROC) // no_instr_name
#define instrSUBGRP_COPROC		(instrgNOT_INSTR | 0x0005) // no_instr_name
#define instrPRE_MMX_SSE		(instrgNOT_INSTR | 0x0006) // no_instr_name
#define instrPRE_MMX_SSE_MOD_MEM_11	(instrgNOT_INSTR | 0x0007) // no_instr_name

//////////////////////////////////////////////////////////////////////

struct InstrIdNameIndex
{
  short m_nInstr;
  short m_nOffset;
};

extern char cInstrNames[];
extern InstrIdNameIndex indexInstrIdName[];

//////////////////////////////////////////////////////////////////////

const char* GetInstructionName (short nInstr);

//////////////////////////////////////////////////////////////////////

//#define INSTR_INFO_ADDR_SIZE_32	0x0001

#define MAX_INSTR_CODE_LENGTH   16

class Operand;

class Instruction  
{
 public:
  Instruction();
  virtual ~Instruction();
  
 public:
  short m_nInstruction;
  
  //	word  m_wInfo;
  
  // m_wInfo was used to keep track of addr-size of JECX and LOOP
  // instr, that otherwise had no chance to retain the knowledge os
  // addr-size (other instructions usually have that info implicit in
  // the operands. The jump instructions should actually also keep
  // track of operand-size attribute which they do not do right now...
  // If there are more instructions that need this facility, we should
  // perhaps reinstate m_wInfo and use that for all kinds of
  // instructions needing this support

 public:
  void SetInstruction (short nInstr) { m_nInstruction = nInstr; }
  
  //	void SetAddrSize32 (bool bAddrSize32) { m_wInfo |= (bAddrSize32 ? INSTR_INFO_ADDR_SIZE_32 : 0); }
  /*
public:
	virtual byte GetRepLock () const { return 0; } // see AnyInstruction
*/
 public:
  virtual const Operand* GetDestOp () const { return NULL; }
  virtual const Operand* GetSrcOp () const { return NULL; }
  virtual const Operand* GetSrc2Op () const { return NULL; }
	/*
	bool GetAddrSize32 () { assert (m_nInstruction == instrJECXZ || m_nInstruction == instrLOOP || m_nInstruction == instrLOOPZ || m_nInstruction == instrLOOPNZ); 
							return (m_wInfo & INSTR_INFO_ADDR_SIZE_32); }
	*/
 public:
  bool OpDestSrcSameRegister () const;
  
 public:
  //	void AsmString (char* pDest) const;
  
  void InstrAsmString (char* pDest) const;
  void OpAsmString (char* pDest) const;
  
 protected:
  //	virtual char* GenAsmString (char* pDest) const;
  
  virtual char* GenInstrAsmString (char* pDest) const;
  virtual char* GenOpAsmString (char* pDest) const;
};

//////////////////////////////////////////////////////////////////////

inline char* OutString (char* pDest,char* pSrc)
{
  while (*pSrc)
    *(pDest++) = *(pSrc++);
  return pDest;
}

//////////////////////////////////////////////////////////////////////

#endif // !defined(__INSTRUCTION_H__INCLUDED_)
