/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// InstructionDecoder.cpp: implementation of the InstructionDecoder class.
//
//////////////////////////////////////////////////////////////////////

#include "InstructionDecoder.h"    

#include "gen/InstructionDecoderError.h" 
#include "gen/Instruction_Names.h"

#include "Instruction.h"    
#include "OperandRegister.h"    
#include "OperandMemory.h"    
#include "OperandImmediate.h"    

#include "BranchInstr.h"    
#include "OneOpInstr.h"    
#include "TwoOpInstr.h"    
#include "AnyInstruction.h"    
#include "DataPtr.h"

#ifdef _MSC_VER
#include <typeinfo.h>
#else
#include <typeinfo>
#endif
#include <assert.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

InstructionDecoder::InstructionDecoder()
{
  m_bCodeSize32 = true;
}

InstructionDecoder::~InstructionDecoder()
{

}

//////////////////////////////////////////////////////////////////////

#ifndef NDEBUG
void OpcodeInfoSanityCheck (OpcodeInfo& info)
{
  if (info.m_nInstruction == instrRESERVED) // we may have carried over some garbage when reading group etc, so do not check
    return;
  
  // check that  info  is self-consistent
  
  dword dest_op = (info.m_nArguments & arg_OP_DEST_MASK);
  dword src_op  = (info.m_nArguments & arg_OP_SRC_MASK );
  dword src2_op = (info.m_nArguments & arg_OP_SRC2_MASK);
  
  assert (dest_op == 0 || 
	  ((dest_op >> arg_OP_DEST_SHIFT) != (src_op  >> arg_OP_SRC_SHIFT ) && 
	   (dest_op >> arg_OP_DEST_SHIFT) != (src2_op >> arg_OP_SRC2_SHIFT)));
  assert (src_op == 0 || 
	  (src_op >> arg_OP_SRC_SHIFT) != (src2_op >> arg_OP_SRC2_SHIFT) || 
	  (src_op == arg_SRC_IS_IMM));
  
  assert (!src2_op || src_op);
  
  assert (dest_op == 0 ||
	  dest_op == arg_DEST_IS_MODRM_REG ||
	  dest_op == arg_DEST_IS_MODRM_MODRM ||
	  dest_op == arg_DEST_IS_IMM ||
	  dest_op == arg_DEST_IS_MEM_DISP ||
	  dest_op == arg_DEST_AS_REG_EXTRA ||
	  dest_op == arg_DEST_LOW_3BITS_REG);
  
  assert (src_op == 0 ||
	  src_op == arg_SRC_IS_MODRM_REG ||
	  src_op == arg_SRC_IS_MODRM_MODRM ||
	  src_op == arg_SRC_IS_IMM ||
	  src_op == arg_SRC_IS_MEM_DISP ||
	  src_op == arg_SRC_AS_REG_EXTRA ||
	  src_op == arg_SRC_AS_REG_EXTRA2 ||
	  src_op == arg_SRC_LOW_3BITS_REG ||
	  src_op == arg_SRC_CONST_1);
  
  assert (src2_op == 0 ||
	  src2_op == arg_SRC2_IS_IMM ||
	  src2_op == arg_SRC2_AS_REG_EXTRA);
  
  bool mod_rm_reg = 
    (dest_op == arg_DEST_IS_MODRM_REG || src_op == arg_SRC_IS_MODRM_REG);
  
  if (info.m_nArguments & arg_EXTRA_MODRM_REG_IS_SPECIAL)
    assert (mod_rm_reg);
  
  bool mod_rm_modrm = 
    (dest_op == arg_DEST_IS_MODRM_MODRM || src_op == arg_SRC_IS_MODRM_MODRM);
  
  dword onlymorr = 
    info.m_nArguments & (arg_MODRM_ONLY_M_MODRM | arg_MODRM_ONLY_R_MODRM);
  
  if (onlymorr)
    {
      assert (
	      onlymorr == arg_MODRM_ONLY_M_MODRM ||
	      onlymorr == arg_MODRM_ONLY_R_MODRM);
      assert (mod_rm_modrm);   // modrm must give operand
      
      if (onlymorr == arg_MODRM_ONLY_M_MODRM)
	assert (
		((info.m_nArguments & arg_MODRM_RT_RM_MASK) != 
		 arg_MODRM_RM_FPU_REG) &&
		((info.m_nArguments & arg_MODRM_RT_RM_MASK) !=
		 arg_MODRM_RM_MMX_REG) &&
		((info.m_nArguments & arg_MODRM_RT_RM_MASK) != 
		 arg_MODRM_RM_SSE_REG));
      /*
	assert (
	info.m_nExtra2 == ext2_MODRM_ONLY_MEM_MODRM ||
	info.m_nExtra2 == ext2_MODRM_ONLY_REG_MODRM);
      */
    }

  bool mod_rm_nnn = 
    ((info.m_nInstruction & instrmIS_GROUP_MASK) == instrgGROUP);
  
  assert (!mod_rm_reg || !mod_rm_nnn);
  
  if (mod_rm_modrm || mod_rm_reg || mod_rm_nnn)
    assert (info.m_nArguments & arg_NEED_MODRM_BYTE);
  
  dword opsize = 
    info.m_nArguments & (arg_OPSIZE_8 | arg_OPSIZE_16 | arg_OPSIZE_32 | 
			 arg_OPSIZE_16_32 | arg_OPSIZE_64 | arg_OPSIZE_128 | 
			 arg_OPSIZE_IN_EXTRA2 | arg_OPSIZE_IN_EXTRAS_16_32);
  
  bool reg_extra_need_opsize = 
    ((info.m_nExtra1 & reggTYPE_MASK) == reggSTD8_16_32 ||
     (info.m_nExtra1 & reggTYPE_MASK) == reggSTD16_32);
  
  bool reg_extra2_need_opsize = 
    ((info.m_nExtra2 & reggTYPE_MASK) == reggSTD8_16_32 ||
     (info.m_nExtra2 & reggTYPE_MASK) == reggSTD16_32);
  
  bool modrm_reg_need_opsize = 
    (!(info.m_nArguments & arg_EXTRA_MODRM_REG_IS_SPECIAL) &&
     !((info.m_nArguments & arg_MODRM_RT_REG_MASK) == arg_MODRM_REG_MMX_REG) &&
     !((info.m_nArguments & arg_MODRM_RT_REG_MASK) == arg_MODRM_REG_SSE_REG));
  
  bool modrm_rm_need_opsize = 
    (!((((info.m_nArguments & arg_MODRM_RT_RM_MASK) == arg_MODRM_RM_FPU_REG) ||
	((info.m_nArguments & arg_MODRM_RT_RM_MASK) == arg_MODRM_RM_MMX_REG) ||
	((info.m_nArguments & arg_MODRM_RT_RM_MASK) == arg_MODRM_RM_SSE_REG)) &&
       (info.m_nArguments & arg_MODRM_ONLY_R_MODRM)));
  
  bool dest_need_opsize = 
    ((dest_op == arg_DEST_IS_MODRM_REG && modrm_reg_need_opsize) ||
     (dest_op == arg_DEST_IS_MODRM_MODRM && modrm_rm_need_opsize) ||
     dest_op == arg_DEST_IS_IMM ||
     dest_op == arg_DEST_IS_MEM_DISP ||
     (dest_op == arg_DEST_AS_REG_EXTRA && reg_extra_need_opsize) ||
     dest_op == arg_DEST_LOW_3BITS_REG);

  bool src_need_opsize = 
    ((src_op == arg_SRC_IS_MODRM_REG && modrm_reg_need_opsize) ||
     (src_op == arg_SRC_IS_MODRM_MODRM && modrm_rm_need_opsize) ||
     src_op == arg_SRC_IS_IMM ||
     src_op == arg_SRC_IS_MEM_DISP ||
     (src_op == arg_SRC_AS_REG_EXTRA && reg_extra_need_opsize) ||
     (src_op == arg_SRC_AS_REG_EXTRA2 && reg_extra2_need_opsize) ||
     src_op == arg_SRC_LOW_3BITS_REG);

  bool src2_need_opsize = 
    (src2_op == arg_SRC2_IS_IMM ||
     (src2_op == arg_SRC2_AS_REG_EXTRA && reg_extra_need_opsize));
  
  dword lastopopsize = 
    info.m_nArguments & (arg_LAST_OPRND_OPSIZE_8 | arg_LAST_OPRND_OPSIZE_16 | 
			 arg_LAST_OPRND_DBL_OPSIZE | 
			 arg_LAST_OPRND_HALF_OPSIZE);
  
  assert (lastopopsize == 0 ||
	  lastopopsize == arg_LAST_OPRND_OPSIZE_8 ||
	  lastopopsize == arg_LAST_OPRND_OPSIZE_16 ||
	  lastopopsize == arg_LAST_OPRND_DBL_OPSIZE ||
	  lastopopsize == arg_LAST_OPRND_HALF_OPSIZE);
  
  assert ((opsize == 0 &&
	   (info.m_nArguments == arg_NONE || // it's ok not to have opsize if no operands specified
	    info.m_nArguments == arg_MODRM ||
	    (!dest_need_opsize && !src_need_opsize && !src2_need_opsize))) ||
	  opsize == arg_OPSIZE_8 ||
	  opsize == arg_OPSIZE_16 ||
	  opsize == arg_OPSIZE_32 ||
	  opsize == arg_OPSIZE_16_32 ||
	  opsize == arg_OPSIZE_64 ||
	  opsize == arg_OPSIZE_128 ||
	  opsize == arg_OPSIZE_IN_EXTRA2 ||
	  opsize == arg_OPSIZE_IN_EXTRAS_16_32 ||
	  lastopopsize);
  
  
  if (lastopopsize)
    assert (
	    !!(dest_op) +
	    !!(src_op) +
	    !!(src2_op) >= 2); // at least two operands (so one is last)
  
  if (lastopopsize)
    assert (
	    !!(dest_need_opsize) +
	    !!(src_need_opsize) +
	    !!(src2_need_opsize) >= !!opsize + !!lastopopsize); // even tougher: at least two ops needing opsize
  
  if (lastopopsize && !opsize)
    assert (
	    !!(dest_need_opsize) +
	    !!(src_need_opsize) +
	    !!(src2_need_opsize) == 1); // only one needing opsize if we only have last to offer
  
  if (info.m_nArguments & arg_SIGN_EXTEND_IMMEDIATE)
    {
      assert ((src_op == arg_SRC_IS_IMM) + (src2_op == arg_SRC2_IS_IMM) == 1);
      
      assert (lastopopsize == arg_LAST_OPRND_OPSIZE_8 &&
	      (opsize == arg_OPSIZE_16 ||
	       opsize == arg_OPSIZE_32 ||
	       opsize == arg_OPSIZE_16_32));
    }
  
  int nUseExtra1 = 
    !!(info.m_nArguments & arg_EXTRA_MODRM_REG_IS_SPECIAL) +
    (dest_op == arg_DEST_AS_REG_EXTRA) +
    (src_op  == arg_SRC_AS_REG_EXTRA) +
    (src2_op == arg_SRC2_AS_REG_EXTRA);
  
  int nUseExtra2 = 
    !!(info.m_nArguments & arg_EXTRA2_DEST_SRC_SI_DI) +
    (src_op == arg_SRC_AS_REG_EXTRA2);
  
  if (info.m_nArguments & arg_OPSIZE_IN_EXTRA2)
    {
      assert (nUseExtra2 <= 1);
      nUseExtra2++;
    }
  
  if (info.m_nArguments & arg_OPSIZE_IN_EXTRAS_16_32)
    {
      assert (nUseExtra1 <= 1 && nUseExtra2 <= 1);
      nUseExtra1++;
      nUseExtra2++;
    }
  
////////////////////////

  bool bExtra1MayBeZero = false;
  bool bExtra2MayBeZero = false;
  
  if ((info.m_nInstruction & instrmIS_GROUP_MASK) == instrgGROUP)
    {
      if (info.m_nInstruction & instrmGROUP_MOD_BOTH)
	{
	  nUseExtra1++;
	  bExtra1MayBeZero = true;
	}
      else 
	{
	  if (info.m_nInstruction & instrmGROUP_MOD_MEM)
	    {
	      nUseExtra1++;
	      bExtra1MayBeZero = true;
	    }
	  if (info.m_nInstruction & instrmGROUP_MOD_11)
	    {
	      nUseExtra2++;
	      bExtra2MayBeZero = true;
	    }
	}
    }
  
  if (info.m_nInstruction == instrPRE_MMX_SSE ||
      info.m_nInstruction == instrPRE_MMX_SSE_MOD_MEM_11)
    {
      nUseExtra1++;
      bExtra1MayBeZero = true;
      nUseExtra2++;
    }
  
  ///////////////////////	
	
  assert (nUseExtra1 <= 1);
  assert (nUseExtra2 <= 1);
  
  assert ((nUseExtra1 != 0) == (info.m_nExtra1 != 0) ||
	  (bExtra1MayBeZero && (info.m_nExtra1 == 0)));
  assert ((nUseExtra2 != 0) == (info.m_nExtra2 != 0) ||
	  (bExtra2MayBeZero && (info.m_nExtra2 == 0)));
  
  if (info.m_nArguments & arg_LAST_OP_MEM_OR_IMM_FAR_PTR)
    {
      assert (!src2_op);
      
      assert (src_op  == arg_SRC_IS_MODRM_MODRM ||			
	      ((dest_op == arg_DEST_IS_MODRM_MODRM ||
		dest_op == arg_DEST_IS_IMM) && !src_op));
    }
  
  dword rtregtype = info.m_nArguments & arg_MODRM_RT_REG_MASK;
  dword rtrmtype = info.m_nArguments & arg_MODRM_RT_RM_MASK;
  
  if (rtregtype)
    assert (mod_rm_reg);
  if (rtrmtype)
    assert (mod_rm_modrm &&
	    !(info.m_nArguments & arg_MODRM_ONLY_M_MODRM));
  
  assert (rtregtype == 0 ||
	  rtregtype == arg_MODRM_REG_MMX_REG ||
	  rtregtype == arg_MODRM_REG_SSE_REG);
  
  assert (rtrmtype == 0 ||
	  rtrmtype == arg_MODRM_RM_FPU_REG ||
	  rtrmtype == arg_MODRM_RM_MMX_REG ||
	  rtrmtype == arg_MODRM_RM_SSE_REG);
  
  if (info.m_nArguments & arg_EXTRA2_DEST_SRC_SI_DI)
    {
      assert (info.m_nExtra2 & (ext2_DEST_MEM_DI | ext2_DEST_MEM_SI | 
				ext2_SRC_MEM_DI | ext2_SRC_MEM_SI));
      
      assert (dest_op == 0 ||
	      !(info.m_nExtra2 & (ext2_DEST_MEM_DI | ext2_DEST_MEM_SI)));
      assert (src_op == 0 ||
	      !(info.m_nExtra2 & (ext2_SRC_MEM_DI | ext2_SRC_MEM_SI)));
      
      assert (!(info.m_nExtra2 & ext2_DEST_MEM_DI) || 
	      !(info.m_nExtra2 & ext2_DEST_MEM_SI));
      assert (!(info.m_nExtra2 & ext2_SRC_MEM_DI) || 
	      !(info.m_nExtra2 & ext2_SRC_MEM_SI));
      assert (!(info.m_nExtra2 & ext2_DEST_MEM_DI) || 
	      !(info.m_nExtra2 & ext2_SRC_MEM_DI));
      assert (!(info.m_nExtra2 & ext2_DEST_MEM_SI) || 
	      !(info.m_nExtra2 & ext2_SRC_MEM_SI));
    }
}
#endif

//////////////////////////////////////////////////////////////////////

#define MOD_FROM_MODRM(modrm) (((modrm) >> 6) & 0x03)
#define REG_FROM_MODRM(modrm) (((modrm) >> 3) & 0x07)
#define RM_FROM_MODRM(modrm) ((modrm) & 0x07)
#define NNN_FROM_MODRM(modrm) REG_FROM_MODRM(modrm)

#define SCALE_FROM_SIB(modrm) (((modrm) >> 6) & 0x03)
#define INDEX_FROM_SIB(modrm) (((modrm) >> 3) & 0x07)
#define BASE_FROM_SIB(modrm) ((modrm) & 0x07)

word OperandSizeFromArgExtraOpSize (OpcodeInfo& info,
				    bool bLastOperand = false,
				    bool bOpSize32 = true)
{
  if (bLastOperand && 
      (info.m_nArguments & arg_LAST_OPRND_OPSIZE_MASK))
    {
      if (info.m_nArguments & arg_LAST_OPRND_DBL_OPSIZE)
	return OperandSizeFromArgExtraOpSize (info) << 1;
      if (info.m_nArguments & arg_LAST_OPRND_HALF_OPSIZE)
	{
	  assert (!(OperandSizeFromArgExtraOpSize (info) & 1)); // cannot be odd
	  return OperandSizeFromArgExtraOpSize (info) >> 1;
	}
      assert ((info.m_nArguments & arg_LAST_OPRND_OPSIZE_8) || 
	      (info.m_nArguments & arg_LAST_OPRND_OPSIZE_16));		
      return (info.m_nArguments & arg_LAST_OPRND_OPSIZE_8 ? 1 : 2);
    }
  
  if (info.m_nArguments & arg_OPSIZE_NORMAL_MASK)
    return (info.m_nArguments & arg_OPSIZE_NORMAL_MASK) >> 
      arg_OPSIZE_NORMAL_SHIFT;
  /*
    if (info.m_nArguments & arg_OPSIZE_8)
    return 1;
    //	else if (info.m_nArguments & arg_OPSIZE_16_32)
    //		return (bOpSize32 ? 4 : 2);
	else if (info.m_nArguments & arg_OPSIZE_16)
	return 2;
	else if (info.m_nArguments & arg_OPSIZE_32)
	return 4;
	else if (info.m_nArguments & arg_OPSIZE_64)
	return 8;
	else if (info.m_nArguments & arg_OPSIZE_128)
	return 16;
  */
  else
    {
      if (info.m_nArguments & arg_OPSIZE_IN_EXTRA2)
	{
	  return OPSIZE_FROM_EXTRA(info.m_nExtra2);
	}
      else
	{
	  assert (info.m_nArguments & arg_OPSIZE_IN_EXTRAS_16_32);
	  
	  if (bOpSize32)
	    return OPSIZE_FROM_EXTRA(info.m_nExtra2);
	  else
	    return OPSIZE_FROM_EXTRA(info.m_nExtra1);
	}
      /*		
			assert (Sinfo.m_nArguments & arg_OPSIZE_IN_EXTRA);

		byte nExtra = info.m_nExtra1;

		if ((info.m_nArguments & arg_EXTRA_MODRM_REG_IS_SPECIAL) ||
			((info.m_nArguments & arg_OP_DEST_MASK) == arg_DEST_AS_REG_EXTRA) ||
			((info.m_nArguments & arg_OP_SRC_MASK ) == arg_SRC_AS_REG_EXTRA ) ||
			((info.m_nArguments & arg_OP_SRC2_MASK) == arg_SRC2_AS_REG_EXTRA))
		{
			// extra1 already taken, we're in extra2
			
			nExtra = info.m_nExtra2;
		}
		else if (info.m_nExtra2 &&
			!(info.m_nArguments & arg_EXTRA2_DEST_SRC_SI_DI) &&
			((info.m_nArguments & arg_OP_SRC_MASK ) != arg_SRC_AS_REG_EXTRA2))
		{
			// we may be in extra2 if it is not used by anoyone else,
			// if it is non-zero

			// in this case, it is dependent on the current operand size attribute

			if (bOpSize32)
				nExtra = info.m_nExtra2;
			// else nExtra = info.m_nExtra1; already
		}

		return OPSIZE_FROM_EXTRA(nExtra);
*/
    }
}

byte OperandRegTypeFromArgExtraOpSize (OpcodeInfo& info,
				       byte bModRMRegType = 0,
				       bool bLastOperand = false)
{
  if (bModRMRegType &= arg_MODRM_RT_MASK)
    {
      if (bModRMRegType == arg_MODRM_RT_FPU_REG)
	return reggST;
      else if (bModRMRegType == arg_MODRM_RT_MMX_REG)
	return reggMM;
      {
	assert (bModRMRegType == arg_MODRM_RT_SSE_REG);
	return reggXMM;
      }
    }
	
  if (bLastOperand && 
      (info.m_nArguments & arg_LAST_OPRND_OPSIZE_MASK))
    {
      assert ((info.m_nArguments & arg_LAST_OPRND_OPSIZE_8) || 
	      (info.m_nArguments & arg_LAST_OPRND_OPSIZE_16));
	  return (info.m_nArguments & arg_LAST_OPRND_OPSIZE_8 ? 
		  reggSTD8 : reggSTD16);
    }
  
  if (info.m_nArguments & arg_OPSIZE_8)
    return reggSTD8;
  //	else if (info.m_nArguments & arg_OPSIZE_16_32)
  //		return (bOpSize32 ? reggSTD32 : reggSTD16);
  else if (info.m_nArguments & arg_OPSIZE_16)
    return reggSTD16;
  else
    {
      assert (info.m_nArguments & arg_OPSIZE_32);
      return reggSTD32;
    }
}

OperandRegister* OperandRegisterFromExtraAndOpSize(byte bReg,int nArguments)
{
  byte bRegType = (bReg & reggTYPE_MASK);
  
  if (bRegType == reggSTD8_16_32 ||
      bRegType == reggSTD16_32)
    {
      bReg &= reggNAME_MASK;
      
      if (nArguments & arg_OPSIZE_8)
	{
	  assert (bRegType == reggSTD8_16_32);
	  bReg |= reggSTD8;
	}
      //		else if (nArguments & arg_OPSIZE_16_32)
      //			bReg |= (bOpSize32 ? reggSTD32 : reggSTD16);
      else if (nArguments & arg_OPSIZE_32)
	bReg |= reggSTD32;
      else
	{
	  assert (nArguments & arg_OPSIZE_16);
	  bReg |= reggSTD16;
	}
    }
  return new OperandRegister (bReg);
}

OperandRegister* OperandSpecialRegisterFromRegAndExtra(byte bModRM,byte bExtra)
{
  byte bReg = REG_FROM_MODRM(bModRM);
  
  assert (bExtra == reggSEG ||
          bExtra == reggDBG ||
          bExtra == reggCTRL ||
          bExtra == reggTST);
  
  if (bExtra == reggSEG)
    {
      if (bReg >= 6)
        throw ERROR_INSTR_DECODE_INVALID_SEG_REG(bReg);
      assert (bReg < 6);
    }
  else if (bExtra == reggCTRL)
    {
      if (bReg != 0 && bReg != 2 && bReg != 3 && bReg != 4)
        throw ERROR_INSTR_DECODE_INVALID_CTRL_REG(bReg);
    }
  
  bReg |= bExtra; // bExtra = reggSEG/DBG/CTRL/TST
  
  return new OperandRegister (bReg);
}

OperandRegister* OperandRegisterFromModRMRegAndType(byte bModRM,byte bRegType)
{
  return new OperandRegister (bRegType | REG_FROM_MODRM(bModRM));
}

OperandRegister* OperandRegisterFromRegAndType(byte bRegName,byte bRegType)
{
  return new OperandRegister (bRegType | bRegName);
}

OperandImmediate* ReadOperandImmediateFromOpSize (DataPtr& dataPtr,
						  byte bSize,byte bHigh16 = 0)
{
  OperandImmediate* pOpImm = new OperandImmediate ();
  
  pOpImm->m_bImmBytes = bSize;
  pOpImm->m_bImmHigh16Bytes = bHigh16;
  
  // and read it...
  
  if (bSize == 1)
    pOpImm->m_dwImm = dataPtr.GetByte ();
  else if (bSize == 2)
    pOpImm->m_dwImm = dataPtr.GetWord ();
  else
    {
      assert (bSize == 4);
      pOpImm->m_dwImm = dataPtr.GetDWord ();
    }
  
  if (bHigh16 == 2)
    pOpImm->m_wImmHigh16 = dataPtr.GetWord ();
  
  return pOpImm;
}

OperandMemory* OperandMemoryDispFromAddrSize(DataPtr& dataPtr,
					     OpcodeInfo& info,
					     bool bAddrSize32,
					     byte bSegReg)
{
  OperandMemory* pOpMem = new OperandMemory;
  
  if (bAddrSize32)
    pOpMem->SetDisplacementSizeDWord(dataPtr.GetDWord ());
  else
    pOpMem->SetDisplacementSizeWord(dataPtr.GetWord ());
  
  assert (!(info.m_nArguments & arg_LAST_OPRND_OPSIZE_MASK));
  
  pOpMem->m_wOpSize = OperandSizeFromArgExtraOpSize(info);
  pOpMem->m_bSegmentReg = (bSegReg ? bSegReg : regDS);
  
  return pOpMem;
}

Operand* OperandFromModRMAndOpSize(byte bModRM,
				   OpcodeInfo& info,
				   bool bAddrSize32,
				   bool bOpSize32,
				   byte bSegReg,
				   DataPtr& dataPtr,
				   byte bModRMRegType,
				   bool bLastOperand = false)
{
  if (MOD_FROM_MODRM(bModRM) == 3)
    {
      if ((info.m_nArguments & arg_MODRM_ONLY_M_MODRM))
	throw ERROR_INSTR_DECODE_MODRM_MAY_ONLY_BE_MEM();

      return new OperandRegister (OperandRegTypeFromArgExtraOpSize (info,bModRMRegType,bLastOperand) | RM_FROM_MODRM(bModRM));
    }
  else
    {
      if ((info.m_nArguments & arg_MODRM_ONLY_R_MODRM))
	throw ERROR_INSTR_DECODE_MODRM_MAY_ONLY_BE_REG();
      
      // ptr to memory
      
      OperandMemory* pOpMem = new OperandMemory;
      
      if (bAddrSize32)
	{
	  assert (pOpMem->m_bDispBytes == 0);
	  switch (MOD_FROM_MODRM(bModRM))
	    {
	    case 1: pOpMem->m_bDispBytes = 1; break;
	    case 2: pOpMem->m_bDispBytes = 4; break;
	    }
	  
	  if (RM_FROM_MODRM(bModRM) == 4)
	    {
	      // sib byte
	      
	      byte bSIB = dataPtr.GetByte ();
	      
	      if (BASE_FROM_SIB(bSIB) == 5 &&
		  MOD_FROM_MODRM(bModRM) == 0)
		{
		  pOpMem->m_bDispBytes = 4;
		}
	      else
		{
		  pOpMem->m_bBaseReg = reggSTD32 | BASE_FROM_SIB(bSIB);
		}
	      if (INDEX_FROM_SIB(bSIB) != 4)
		{
		  pOpMem->m_bIndexReg = reggSTD32 | INDEX_FROM_SIB(bSIB);
		  pOpMem->m_bScale = SCALE_FROM_SIB(bSIB);
		}
	    }
	  else if (RM_FROM_MODRM(bModRM) == 5 &&
		   MOD_FROM_MODRM(bModRM) == 0)
	    {
	      pOpMem->m_bDispBytes = 4;
	    }
	  else
	    {
	      pOpMem->m_bBaseReg = reggSTD32 | RM_FROM_MODRM(bModRM);
	    }
	}
      else
	{
	  pOpMem->m_bDispBytes = MOD_FROM_MODRM(bModRM);
	  
	  if (MOD_FROM_MODRM(bModRM) == 0 &&
	      RM_FROM_MODRM(bModRM) == 6)
	    {
	      pOpMem->m_bDispBytes = 2;
	    }
	  else
	    {
	      if (RM_FROM_MODRM(bModRM) < 4)
		{
		  pOpMem->m_bBaseReg = 
		    RM_FROM_MODRM(bModRM) & 2 ? regBP : regBX;
		  pOpMem->m_bIndexReg = 
		    RM_FROM_MODRM(bModRM) & 1 ? regDI : regSI;
		}
	      else
		{
		  switch (RM_FROM_MODRM(bModRM))
		    {
		    case 4: pOpMem->m_bBaseReg = regSI; break;
		    case 5: pOpMem->m_bBaseReg = regDI; break;
		    case 6: pOpMem->m_bBaseReg = regBP; break;
		    case 7: pOpMem->m_bBaseReg = regBX; break;
		    }
		}
	    }
	}

      switch (pOpMem->m_bDispBytes)
	{
	case 1: pOpMem->SetDisplacementByte(dataPtr.GetByte ()); break;
	case 2: pOpMem->SetDisplacementWord(dataPtr.GetWord ()); break;
	case 4: pOpMem->SetDisplacementDWord(dataPtr.GetDWord ()); break;
	default:
	  assert (!pOpMem->m_bDispBytes);
	}
      
      pOpMem->m_wOpSize = OperandSizeFromArgExtraOpSize(info,
							bLastOperand,
							bOpSize32);
      if (info.m_nArguments & arg_LAST_OP_MEM_OR_IMM_FAR_PTR)
	pOpMem->m_wOpSize += 2;
      if (bSegReg)
	pOpMem->m_bSegmentReg = bSegReg;
      else
	pOpMem->m_bSegmentReg = 
	  (IS_REG_NAME_SP_BP(pOpMem->m_bBaseReg) ? regSS : regDS);
      
      return pOpMem;
    }
}

OperandMemory* OperandSIDIMemory (byte bType,
				  bool bAddrSize32,
				  OpcodeInfo& info,
				  byte bPrefixSegment)
{
  OperandMemory* pOpMem = new OperandMemory;
  
  bool bDI = (bType & (ext2_DEST_MEM_DI | ext2_SRC_MEM_DI)) != 0;
  
  pOpMem->m_bIndexReg = 
    (bAddrSize32 ? reggSTD32 : reggSTD16) | (bDI ? reggDI : reggSI);
  
  if (bPrefixSegment && !bDI)
    pOpMem->m_bSegmentReg = bPrefixSegment;
  else
    pOpMem->m_bSegmentReg = (bDI ? regES : regDS);
  
  pOpMem->m_wOpSize = OperandSizeFromArgExtraOpSize(info,false);
  
  return pOpMem;
}

// // #include "gen/Instruction_Names.h" // dep_chk_unknown   
int compare_short (const void* p1,const void* p2);

Instruction* InstructionDecoder::Decode (DataPtr& dataPtr) const
{
  //	byte* pBegin = dataPtr.pData;
  
  bool bOpSize32 = m_bCodeSize32;
  bool bAddrSize32 = m_bCodeSize32;
  
  byte bPrefixSegment = 0;
  
  byte bPrefixMask = 0;
  
  byte bRepeat = 0;
  
  byte bOpcode; // , bModRM, bSIB;
  byte bModRM;
  
  OpcodeInfo info;
  
  byte bSSEPrefix = 0; // for NO prefix
  
  for ( ; ; )
    {
      bOpcode = dataPtr.GetByte ();
      
      info = infoByte[0][bOpcode];
      
      if (info.m_nInstruction == instrPREFIX)
	{
	  // decode a prefix
	  
	  // see if the prefix happen to be instruction code for an SSE instruction
	  
	  if ((bOpcode == 0x66 || bOpcode == 0xf3 || bOpcode == 0xf2) && // it's an possible SSE prefix
	      dataPtr.GetByte(dataPtr.GetRVA()) == 0x0f) // check without advancing for two byte opcode escape
	    {		
	      // there is no problem if dataPtr.GetByte fails (throws)
	      // because os read outside source, as we'd do that
	      // anyway when finished reading prefixes and trying to
	      // read the instruction
	      
	      OpcodeInfo info2;

	      info2 = infoByte[1][dataPtr.GetByte (dataPtr.GetRVA()+1)];
	      
	      if (info2.m_nInstruction == instrGROUP_MOD_11)
		{
		  bModRM = dataPtr.GetByte (dataPtr.GetRVA()+2);
		  info2 = infoGroups[info2.m_nExtra2][NNN_FROM_MODRM(bModRM)];
		}
	      
	      bSSEPrefix = (bOpcode == 0x66 ? 1 : (bOpcode == 0xf3 ? 2 : 3));
	      
	      if ((info2.m_nInstruction == instrPRE_MMX_SSE || // possible
		   info2.m_nInstruction == instrPRE_MMX_SSE_MOD_MEM_11) && // possible
		  (info2.m_nExtra2 & (1 << bSSEPrefix))) // we may have this prefixes
		{
		  // ok, so the prefix was part of the sse opcode,
		  
		  bOpcode = dataPtr.GetByte ();
		  
		  assert (bOpcode == 0x0f); // as we just found
		  
		  info.m_nInstruction = instrESCAPE2BYTE; // so the two byte opcode is decoded
		  
		  break; // abort prefix decoding
		}
	      else
		{
		  // if an SSE instruction for this prefix didn't exist
		  // even if it existed for another SSE prefix, 
		  // assume it's an normal prefix
		  
		  bSSEPrefix = 0;
		}
	    }
	  
	  if (bPrefixMask & info.m_nArguments)
	    throw ERROR_INSTR_DECODE_TWO_PREFIX_FROM_SAME_GROUP(bOpcode);
	  
	  bPrefixMask |= info.m_nArguments;
	  
	  switch ((byte) info.m_nArguments)
	    {
	    case arg_PREFIX_LOCK: 
	      break;
	    case arg_PREFIX_REP: 
	      bRepeat = info.m_nExtra1;
	      break;
	    case arg_PREFIX_SEGMENT: 
	      bPrefixSegment = info.m_nExtra1;
	      break;
	    case arg_PREFIX_OPSIZE: 
	      bOpSize32 ^= true;
	      break;
	    case arg_PREFIX_ADDRSIZE: 
	      bAddrSize32 ^= true;
	      break;
	    }
	}
      else 
	break;
    }
  
  if (info.m_nInstruction == instrESCAPE2BYTE)
    {
      bOpcode = dataPtr.GetByte ();
      
      info = infoByte[1][bOpcode];
    }
  
#ifndef NDEBUG
  OpcodeInfoSanityCheck (info);
#endif
  
  if (info.m_nArguments & arg_NEED_MODRM_BYTE)
    bModRM = dataPtr.GetByte ();
  
  if ((info.m_nInstruction & instrmIS_GROUP_MASK) == instrgGROUP)
    {
      int nGroup;
      
      if (info.m_nInstruction & instrmGROUP_MOD_BOTH)
	{
	  nGroup = info.m_nExtra1;
	}
      else 
	{
	  if (MOD_FROM_MODRM(bModRM) != 3) // MOD = mem
	    {
	      if (!(info.m_nInstruction & instrmGROUP_MOD_MEM))
		throw ERROR_INSTR_DECODE_MODRM_MAY_ONLY_BE_MEM();
	      
	      nGroup = info.m_nExtra1;
	    }
	  else // mod = 11
	    {
	      if (!(info.m_nInstruction & instrmGROUP_MOD_11))
		throw ERROR_INSTR_DECODE_MODRM_MAY_ONLY_BE_REG();
	      
	      nGroup = info.m_nExtra2;
	    }
	  if (info.m_nInstruction & instrmGROUP_MOD_11)
	    info.m_nExtra2 = 0;
	  
	  assert ((info.m_nInstruction & instrmGROUP_MOD_MEM) || 
		  info.m_nExtra1 == 0); // otherwise info will be lost
	  assert ((info.m_nInstruction & instrmGROUP_MOD_11) || 
		  info.m_nExtra2 == 0); // otherwise info will be lost
	}
      
      OpcodeInfo info2 = infoGroups[nGroup][NNN_FROM_MODRM(bModRM)];
      
      if (info2.m_nInstruction == instrSUBGRP_COPROC)
	{
	  assert ((info.m_nArguments & arg_OP_DEST_MASK) != 
		  arg_DEST_IS_MODRM_MODRM); // as subgroup got N from RM, we can't get register or mem from there too
	  assert ((info.m_nArguments & arg_OP_SRC_MASK ) != 
		  arg_SRC_IS_MODRM_MODRM ); // as subgroup got N from RM, we can't get register or mem from there too
	  
	  info2 = infoGroups[info2.m_nExtra1][RM_FROM_MODRM(bModRM)];
	}
      
      assert (info.m_nExtra2 == 0 ||
	      info2.m_nExtra2 == 0); // otherwise info will be jumbled
      
      info.m_nArguments |= info2.m_nArguments;
      info.m_nInstruction = info2.m_nInstruction;
      info.m_nExtra1 = info2.m_nExtra1;
      info.m_nExtra2 |= info2.m_nExtra2;
      
      assert ((info.m_nArguments & arg_OP_DEST_MASK) != 
	      arg_DEST_IS_MODRM_REG); // as group got NNN from REG, we can't get register from there too
      assert ((info.m_nArguments & arg_OP_SRC_MASK ) != 
	      arg_SRC_IS_MODRM_REG ); // as group got NNN from REG, we can't get register from there too
    }
  
  if (info.m_nInstruction == instrPRE_MMX_SSE ||
      info.m_nInstruction == instrPRE_MMX_SSE_MOD_MEM_11)
    {
      // bSSEPrefix contain any prefix
      
      assert (info.m_nExtra2 & (1 << bSSEPrefix) ||
	      bSSEPrefix == 0);
      
      assert (!(info.m_nArguments & ~arg_NEED_MODRM_BYTE));
      
      dword dwOldArgs = info.m_nArguments;
      
      int nOffset = (info.m_nExtra1 << 1) + bSSEPrefix;
      
      if (info.m_nInstruction == instrPRE_MMX_SSE_MOD_MEM_11)
	{
	  assert ( (info.m_nExtra2 & ext2_SSE_PREFIX_NO_66) && 
		   !(info.m_nExtra2 & ext2_SSE_PREFIX_F3_F2)); // so that nOffset += 2 is valid
	  assert (dwOldArgs & arg_NEED_MODRM_BYTE);
	  
	  if (MOD_FROM_MODRM(bModRM) == 3) // MOD = 11 (not mem)
	    nOffset += 2;
	}
      
      assert (nOffset < ext_MMX_SSE_NUM_GROUPS * 2);
      
      info = infoMMX_SSE[nOffset]; // index to be multiplied by two
      
      if ( (info.m_nArguments & arg_NEED_MODRM_BYTE) &&
	   !(dwOldArgs         & arg_NEED_MODRM_BYTE))
	bModRM = dataPtr.GetByte ();
    }
  
  assert (info.m_nInstruction != instrSUBGRP_COPROC &&
	  info.m_nInstruction != instrPRE_MMX_SSE &&
	  info.m_nInstruction != instrPRE_MMX_SSE_MOD_MEM_11 &&
	  (info.m_nInstruction & instrmIS_GROUP_MASK) != instrgGROUP);
  
#ifndef NDEBUG
  OpcodeInfoSanityCheck (info);
#endif
  
  if (info.m_nInstruction == instrRESERVED)
    throw ERROR_INSTR_DECODE_INSTR_RESERVED(bOpcode);
  

  // Check that the prefixes used are valid with the instruction being
  // decoded, before we try to decode the actual parameters...
  
  // Opsize, requires arg_OPSIZE_16_32
  
  if (bPrefixMask & arg_PREFIX_OPSIZE)
    {
      if ((info.m_nArguments & arg_OPSIZE_MASK) != arg_OPSIZE_16_32 &&
	  !(info.m_nArguments & arg_OPSIZE_IN_EXTRAS_16_32) )
	throw ERROR_INSTR_DECODE_WRONG_OPSIZE_PREFIX();
      // hi there!, are you here because this threw something at your
      // program, and it was because there was an op-size prefix with
      // an (conditional) jump instruction?  in that case: well, they
      // are not supported by this program (yet), because I wasn't
      // sure they were legal, and was not sure it would be useful to
      // do so anywhere.  if support is to be added: hmm, two things
      // are necessary which are not elementary: support must be added
      // to the Instruction class to store the information that the
      // instruction has an operand size prefix, support must be added
      // to print this information, support must be added when
      // encoding these instructions again to eject an prefix, support
      // must be added to handle jumps like this, because they mess
      // non-trivially with EIP's high word (like the disassembler's
      // easy detection of relative jumps..) Still think you found a
      // program that need all this?, then ok, fix.
    }
  if ((info.m_nArguments & arg_OPSIZE_MASK) == arg_OPSIZE_16_32)
    {
      // choose depending on opsize
      
      info.m_nArguments &= (~arg_OPSIZE_MASK);
      info.m_nArguments |= (bOpSize32 ? arg_OPSIZE_32 : arg_OPSIZE_16);
    }
  
  dword dwDestOp = (info.m_nArguments & arg_OP_DEST_MASK);
  dword dwSrcOp  = (info.m_nArguments & arg_OP_SRC_MASK );
  dword dwSrc2Op = (info.m_nArguments & arg_OP_SRC2_MASK);
  
  // Address size, requires memory reference
  
  if (bPrefixMask & arg_PREFIX_ADDRSIZE)
    {
      // check that some operand reside in memory
      
      if (!((dwDestOp == arg_DEST_IS_MODRM_MODRM || // *dest, or
	     dwSrcOp == arg_SRC_IS_MODRM_MODRM) &&  // *src modrm-mod, and
	    MOD_FROM_MODRM(bModRM) != 3) &&		  // mem (not reg)
	  dwDestOp != arg_DEST_IS_MEM_DISP &&       // *dest mem(disp)
	  dwSrcOp != arg_SRC_IS_MEM_DISP &&         // *src  mem(disp)
	  !(info.m_nArguments & arg_EXTRA2_DEST_SRC_SI_DI) && // *string instr (implicit memory)
	  !INSTR_IS_COND_COUNT(info.m_nInstruction))   // JECX or LOOP instr (addr-size affect CX or ECX)
	throw ERROR_INSTR_DECODE_WRONG_ADDRSIZE_PREFIX();
    }
  /*
    if (info.m_nArguments & arg_EXTRA2_DEST_SRC_SI_DI)
    if (dwDestOp == arg_DEST_IS_MEM_DISP)
    else if (dwSrcOp == arg_SRC_IS_MEM_DISP)
    
    if (dwDestOp == arg_DEST_IS_MODRM_MODRM)
    else if (dwSrcOp == arg_SRC_IS_MODRM_MODRM)
    
		if (MOD_FROM_MODRM(bModRM) == 3)
  */

  // Segement prefix, requires memory reference, that can have it's
  // segment overridden
  
  if (bPrefixMask & arg_PREFIX_SEGMENT)
    {
      // check that some operand reside in memory
      
      if (!((dwDestOp == arg_DEST_IS_MODRM_MODRM || // *dest, or
	     dwSrcOp == arg_SRC_IS_MODRM_MODRM) &&  // *src modrm-mod, and
	    MOD_FROM_MODRM(bModRM) != 3) &&		  // mem (not reg)
	  dwDestOp != arg_DEST_IS_MEM_DISP &&       // *dest mem(disp)
	  dwSrcOp != arg_SRC_IS_MEM_DISP &&         // *src  mem(disp)
	  !((info.m_nArguments & arg_EXTRA2_DEST_SRC_SI_DI) && // *string instr (implicit memory)
	    (info.m_nExtra2 & (ext2_DEST_MEM_SI | ext2_SRC_MEM_SI)))) // with DS:SI (not ES:DI, which cannot be overridden)
	throw ERROR_INSTR_DECODE_WRONG_SEGMENT_PREFIX();
    }	
  
  // Lock prefix, requires lockable instruction, and memory operand as
  // destination
  
  if ((bPrefixMask & arg_PREFIX_LOCK_REP_MASK) == arg_PREFIX_LOCK)
    {
      if (!(dwDestOp == arg_DEST_IS_MODRM_MODRM &&  // *dest modrm-mod, and
	    MOD_FROM_MODRM(bModRM) != 3) &&		  // mem (not reg)
	  dwDestOp != arg_DEST_IS_MEM_DISP)         // *dest mem(disp)
	throw ERROR_INSTR_DECODE_WRONG_LOCK_PREFIX_NO_MEM_DEST();
      
      if (!IsValidLockInstruction(info.m_nInstruction))
	throw ERROR_INSTR_DECODE_WRONG_LOCK_PREFIX_BAD_INSTR();
      
      // so it's a valid LOCK function, change it to that
      
      info.m_nInstruction |= instrgmINT_ARITH_MEMORY_LOCK;
      
      assert (bsearch (&info.m_nInstruction,
		       indexInstrIdName,
		       NUMBER_OF_INSTR_NAMES,
		       sizeof (InstrIdNameIndex),
		       compare_short));
    }
  else // Rep/n/e prefix, requires (suitable) string instruction
    if ((bPrefixMask & arg_PREFIX_LOCK_REP_MASK) == arg_PREFIX_REP)
      {
	if ((info.m_nInstruction & instrgMASK) != instrgSTRING)
	  throw ERROR_INSTR_DECODE_WRONG_REP_PREFIX_BAD_INSTR();
	
	assert (bRepeat == instrgmSTRING_REPNE ||
		bRepeat == instrgmSTRING_REPE);
	
	if (info.m_nInstruction == instrCMPS ||
	    info.m_nInstruction == instrSCAS)
	  {
	    info.m_nInstruction |= bRepeat;
	  }
	else
	  {
	    /* FIXME: 
	     *
	     * I've found string instructions that normally take REP
	     * as prefix prefixed by repne. This is undefined, but
	     * assume this behaves as simple REP prefix.
	     *
	     * if (bRepeat == instrgmSTRING_REPNE)
	     * throw "REPNE prefix used on illegal instruction."; // this never happens because of that */
	    
	    info.m_nInstruction |= instrgmSTRING_REP;
	  }
  
  assert (bsearch (&info.m_nInstruction,
		   indexInstrIdName,
		   NUMBER_OF_INSTR_NAMES,
		   sizeof (InstrIdNameIndex),
		   compare_short));
      }
 
 Operand* pOpDest = NULL;
 Operand* pOpSrc = NULL;
 Operand* pOpSrc2 = NULL;
	
 if (dwDestOp == arg_DEST_IS_MODRM_REG)
   {
     assert (pOpDest == NULL);
     if (info.m_nArguments & arg_EXTRA_MODRM_REG_IS_SPECIAL)
       pOpDest = OperandSpecialRegisterFromRegAndExtra(bModRM,info.m_nExtra1);
     else
       pOpDest = OperandRegisterFromModRMRegAndType(bModRM,
						    OperandRegTypeFromArgExtraOpSize(info,(byte) (info.m_nArguments >> arg_MODRM_RT_REG_SHIFT)));
   }
 else if (dwSrcOp == arg_SRC_IS_MODRM_REG)
   {
     assert (pOpSrc == NULL);
     if (info.m_nArguments & arg_EXTRA_MODRM_REG_IS_SPECIAL)
       pOpSrc = OperandSpecialRegisterFromRegAndExtra(bModRM,info.m_nExtra1);
     else
       pOpSrc = OperandRegisterFromModRMRegAndType(bModRM,
						   OperandRegTypeFromArgExtraOpSize(info,(byte) (info.m_nArguments >> arg_MODRM_RT_REG_SHIFT),!dwSrc2Op));
   }

 if (dwDestOp == arg_DEST_IS_MODRM_MODRM)
   {
     assert (pOpDest == NULL);
     pOpDest = 
       OperandFromModRMAndOpSize(bModRM,info,
				 bAddrSize32,bOpSize32,bPrefixSegment,dataPtr,
				 (byte) (info.m_nArguments >> 
					 arg_MODRM_RT_RM_SHIFT));
   }
 else if (dwSrcOp == arg_SRC_IS_MODRM_MODRM)
   {
     assert (pOpSrc == NULL);
     pOpSrc = 
       OperandFromModRMAndOpSize(bModRM,info,
				 bAddrSize32,bOpSize32,bPrefixSegment,
				 dataPtr,
				 (byte) (info.m_nArguments >> 
					 arg_MODRM_RT_RM_SHIFT),
				 !dwSrc2Op);
   }
 /*
   OperandMemory* pOpMem;
   
   if ((pOpMem = dynamic_cast<OperandMemory*>(pOpDest)) ||
		(pOpMem = dynamic_cast<OperandMemory*>(pOpSrc)))
	{
		assert (!dynamic_cast<OperandMemory*>(pOpDest) ||
				!dynamic_cast<OperandMemory*>(pOpSrc) ||
				(info.m_nArguments & arg_EXTRA2_DEST_SRC_SI_DI));
		ReadDisplacement (pOpMem,dataPtr);
	}
	assert (!dynamic_cast<OperandMemory*>(pOpSrc2));
*/

 if (dwDestOp == arg_DEST_AS_REG_EXTRA)
   {
     assert (pOpDest == NULL);
     pOpDest = OperandRegisterFromExtraAndOpSize(info.m_nExtra1,
						 info.m_nArguments);
   }
 else if (dwSrcOp == arg_SRC_AS_REG_EXTRA)
   {
     assert (pOpSrc == NULL);
     pOpSrc = OperandRegisterFromExtraAndOpSize(info.m_nExtra1,
						info.m_nArguments);
   }
 else if (dwSrc2Op == arg_SRC2_AS_REG_EXTRA)
   {
     assert (pOpSrc2 == NULL);
     pOpSrc2 = OperandRegisterFromExtraAndOpSize(info.m_nExtra1,
						 info.m_nArguments);
   }
 
 if (dwSrcOp == arg_SRC_AS_REG_EXTRA2)
   {
     assert (pOpSrc == NULL);
     pOpSrc = OperandRegisterFromExtraAndOpSize(info.m_nExtra2,
						info.m_nArguments);
   }
 
 if (dwDestOp == arg_DEST_LOW_3BITS_REG)
   {
     assert (info.m_nInstruction == instrPOP ||
	     info.m_nInstruction == instrINC ||
	     info.m_nInstruction == instrDEC ||
	     info.m_nInstruction == instrMOV ||
	     info.m_nInstruction == instrXCHG ||
	     info.m_nInstruction == instrBSWAP);
     assert (pOpDest == NULL);
     
     pOpDest = OperandRegisterFromRegAndType(bOpcode & 0x07,
					     OperandRegTypeFromArgExtraOpSize(info));
   }
 else if (dwSrcOp == arg_SRC_LOW_3BITS_REG)
   {
     assert (info.m_nInstruction == instrPUSH);
     assert (pOpSrc == NULL);
     
     pOpSrc = OperandRegisterFromRegAndType(bOpcode & 0x07,
					    OperandRegTypeFromArgExtraOpSize(info));
   }
 
 if (info.m_nArguments & arg_EXTRA2_DEST_SRC_SI_DI)
   {
     if (info.m_nExtra2 & (ext2_DEST_MEM_DI | ext2_DEST_MEM_SI))
       {
	 assert (pOpDest == NULL);
	 pOpDest = OperandSIDIMemory (info.m_nExtra2 & (ext2_DEST_MEM_DI | 
							ext2_DEST_MEM_SI),
				      bAddrSize32,info,bPrefixSegment);
       }
     if (info.m_nExtra2 & (ext2_SRC_MEM_DI | ext2_SRC_MEM_SI))
       {
	 assert (pOpSrc == NULL);
	 pOpSrc = OperandSIDIMemory (info.m_nExtra2 & (ext2_SRC_MEM_DI |
						       ext2_SRC_MEM_SI),
				     bAddrSize32,info,bPrefixSegment);
       }
   }
 
 if (dwDestOp == arg_DEST_IS_MEM_DISP)
   {
     assert (pOpDest == NULL);
     pOpDest = OperandMemoryDispFromAddrSize (dataPtr,info,
					      bAddrSize32,bPrefixSegment);
   }
 else if (dwSrcOp == arg_SRC_IS_MEM_DISP)
   {
     assert (pOpSrc == NULL);
     pOpSrc = OperandMemoryDispFromAddrSize (dataPtr,info,
					     bAddrSize32,bPrefixSegment);
   }
 
 if (dwDestOp == arg_DEST_IS_IMM)
   {
     assert (pOpDest == NULL);
     
     pOpDest = 
       ReadOperandImmediateFromOpSize (dataPtr,
				       OperandSizeFromArgExtraOpSize(info),
				       ((info.m_nArguments & 
					 arg_LAST_OP_MEM_OR_IMM_FAR_PTR) ? 
					2 : 0));
   }
 else if (dwSrcOp == arg_SRC_IS_IMM)
   {
     assert (pOpSrc == NULL);
     
     OperandImmediate* pOpImm;
     
     pOpSrc = pOpImm = ReadOperandImmediateFromOpSize (dataPtr,
						       OperandSizeFromArgExtraOpSize(info,!dwSrc2Op));
     
     if (info.m_nArguments & arg_SIGN_EXTEND_IMMEDIATE)
       pOpImm->SignExtend (pOpDest->GetOpSize());
   }
 if (dwSrc2Op == arg_SRC2_IS_IMM)
   {
     assert (pOpSrc2 == NULL);
     
     OperandImmediate* pOpImm;
     
     pOpSrc2 = pOpImm = ReadOperandImmediateFromOpSize (dataPtr,
							OperandSizeFromArgExtraOpSize(info,true));
     
     if (info.m_nArguments & arg_SIGN_EXTEND_IMMEDIATE)
       pOpImm->SignExtend (pOpDest->GetOpSize());
   }
 
 if (dwSrcOp == arg_SRC_CONST_1)
   {
     assert (pOpSrc == NULL);
     pOpSrc = new OperandImmediate (1);
   }
 
 assert ((info.m_nInstruction != instrPUSH &&
	  info.m_nInstruction != instrNOP) ||
	 pOpDest == NULL);
 assert ((info.m_nInstruction != instrPOP &&
	  info.m_nInstruction != instrINC &&
	  info.m_nInstruction != instrDEC &&
	  info.m_nInstruction != instrNOP) ||
	 pOpSrc == NULL);
 
	///////////////////////////////////
	// REMOVE; checked before decoding operands above
/*
	if (bPrefixMask & arg_PREFIX_ADDRSIZE)
	{
		assert (dynamic_cast<OperandMemory*>(pOpDest) ||
				dynamic_cast<OperandMemory*>(pOpSrc));
	}	
	if (bPrefixMask & arg_PREFIX_SEGMENT)
	{
		// check that some operand reside in memory

		if ((!dynamic_cast<OperandMemory*>(pOpDest) &&
			 !dynamic_cast<OperandMemory*>(pOpSrc)) ||
			((info.m_nArguments & arg_EXTRA2_DEST_SRC_SI_DI) &&
			 !(info.m_nExtra2 & (ext2_DEST_MEM_SI | ext2_SRC_MEM_SI))))
			 assert (false);
	}	
	if ((bPrefixMask & arg_PREFIX_LOCK_REP_MASK) == arg_PREFIX_LOCK)
	{
		assert (dynamic_cast<OperandMemory*>(pOpDest));
		assert (IsValidLockInstruction(info.m_nInstruction));
	}
	if ((bPrefixMask & arg_PREFIX_LOCK_REP_MASK) == arg_PREFIX_REP)
	{
		assert ((info.m_nInstruction & instrgMASK) == instrgSTRING);
	}
*/
	// REMOVE
	///////////////////////////////////

 Instruction* pInstr = NULL;
 
 if ((info.m_nInstruction & instrgMASK) == instrgCTRL_TRANSFER)
   {
     // a branch instruction
     
     BranchInstr* pBrInstr = new BranchInstr ();
     
     assert (pOpSrc == NULL);
     
     // if (pOpDest)
     pBrInstr->SetTargetOp (pOpDest);
     
     if (INSTR_IS_COND_COUNT(info.m_nInstruction) &&
	 bAddrSize32)
       info.m_nInstruction |= instrmCOUNT_ECX; // mark as using ECX (and not CX)
     
     pInstr = pBrInstr;
     
   }
 /*
	else if ((info.m_nInstruction == instrPUSH ||
			  info.m_nInstruction == instrPOP ||
			  info.m_nInstruction == instrINC ||
			  info.m_nInstruction == instrDEC) && 
			 !(bPrefixMask & arg_PREFIX_LOCK))
	{
		OneOpInstr* pOneOpInstr = new OneOpInstr ();

		if (pOpDest)
			pOneOpInstr->SetDestOp (pOpDest);
		if (pOpSrc)
			pOneOpInstr->SetSrcOp (pOpSrc);

		assert (!pOpSrc2);

		pInstr = pOneOpInstr;
	}
	else if (pOpSrc2 == NULL && 
			 !(bPrefixMask & arg_PREFIX_LOCK))
	{
		TwoOpInstr* pTwoOpInstr = new TwoOpInstr ();

		if (pOpDest)
			pTwoOpInstr->SetDestOp (pOpDest);
		if (pOpSrc)
			pTwoOpInstr->SetSrcOp (pOpSrc);

		assert (!pOpSrc2);

		pInstr = pTwoOpInstr;
	}
 */
 else if (pOpSrc2 == NULL/* && 
			    !(bPrefixMask & arg_PREFIX_LOCK_REP)*/)
   {
     if (pOpSrc == NULL && pOpDest == NULL)
       {
	 Instruction* pNoOpInstr = new Instruction ();
	 
	 pInstr = pNoOpInstr;
       }
     else if (pOpSrc == NULL || 
	      (pOpDest == NULL && info.m_nInstruction == instrPUSH))
       {
	 OneOpInstr* pOneOpInstr = new OneOpInstr ();
	 
	 // if (pOpDest)
	 pOneOpInstr->SetDestOp (pOpDest);
	 // if (pOpSrc)
	 pOneOpInstr->SetSrcOp (pOpSrc);
	 
	 pInstr = pOneOpInstr;
       }
     else
       {
	 TwoOpInstr* pTwoOpInstr = new TwoOpInstr ();
	 
	 // if (pOpDest)
	 pTwoOpInstr->SetDestOp (pOpDest);
	 // if (pOpSrc)
	 pTwoOpInstr->SetSrcOp (pOpSrc);
	 
	 pInstr = pTwoOpInstr;					
       }
   }
 else
   {
     // do not know how to handle,
     // put the stuff into an general 'garbage' can :-)
     
     AnyInstruction* pAnyInstr = new AnyInstruction ();
     
     // if (pOpDest)
     pAnyInstr->SetDestOp (pOpDest);
     // if (pOpSrc)
     pAnyInstr->SetSrcOp (pOpSrc);
     // if (pOpSrc2)
     pAnyInstr->SetSrc2Op (pOpSrc2);
     /*
       if ((bPrefixMask & arg_PREFIX_LOCK_REP_MASK) == arg_PREFIX_LOCK)
       pAnyInstr->SetLocked ();
       if ((bPrefixMask & arg_PREFIX_LOCK_REP_MASK) == arg_PREFIX_REP)
       {
       if (bRepeat == ANY_INSTR_REP &&
       (info.m_nInstruction == instrCMPS ||
       info.m_nInstruction == instrSCAS))
       bRepeat = ANY_INSTR_REPE;
       
       assert (bRepeat != ANY_INSTR_REPNE ||
       (info.m_nInstruction == instrCMPS ||
       info.m_nInstruction == instrSCAS));
       
       pAnyInstr->SetRepeat (bRepeat);
       }
     */
     pInstr = pAnyInstr;
   }
 
 assert (pInstr);
 
 assert (!(info.m_nInstruction & instrgNOT_INSTR));
 
 pInstr->SetInstruction (info.m_nInstruction);
 
 assert (pInstr->GetDestOp() == pOpDest);
 assert (pInstr->GetSrcOp () == pOpSrc );
 assert (pInstr->GetSrc2Op() == pOpSrc2);
 
 return pInstr;
}

bool IsValidLockInstruction (word nInstr)
{
  static short wValidLockInstr[18] =
  {
    instrADD, instrOR, instrADC, instrSBB,
    instrAND, instrSUB, instrXOR, 
    instrINC, instrDEC,
    instrNOT, instrNEG,
    instrXADD, 
    instrBTR, instrBTC, instrBTS, 
    instrXCHG,
    instrCMPXCHG, instrCMPXCH8B, 
  };
  
  unsigned int i;
  
  for (i = 0; i < sizeof (wValidLockInstr)/sizeof(wValidLockInstr[0]); i++)
    if (wValidLockInstr[i] == nInstr)
      return true;
  return false;
}






