/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// InstructionDecoder.h: interface for the InstructionDecoder class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__INSTRUCTION_DECODER_H__INCLUDED_)
#define __INSTRUCTION_DECODER_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "GlobalDefs.h" // dep_chk_needed     
#include "Instruction.h" // dep_chk_needed     
class DataPtr; // from  // // #include "DataPtr.h" // dep_chk_unknown     by dep_chk 

//////////////////////////////////////////////////////////////////////

#define arg_OP_IS_IMM			0x01L // dest,src,src2 // begin with the ones used by desp,src and src2
#define arg_OP_AS_REG_EXTRA		0x02L //      src,src2

#define arg_OP_IS_MEM_DISP		0x03L // dest,src
#define arg_OP_AS_REG_EXTRA2		0x04L //      src
#define arg_OP_LOW_3BITS_REG		0x05L // dest,src
#define arg_OP_CONST_1			0x06L //      src

#define arg_OP_IS_MODRM_REG		0x08L // dest,src
#define arg_OP_IS_MODRM_MODRM		0x09L //      src

#define arg_OP_MASK			0x0fL

// all are excluding, that is if DEST is one of them, SRC can't be that too
// (except CONST1, but that is only used by SRC anyway)
// (and that SRC and SRC2 both may be arg_OP_IS_IMM)

// dest and src need 3 bits
// src 2 need 2 bits

#define arg_OP_DEST_SHIFT		28 // 32-4
#define arg_OP_SRC_SHIFT		24 // 32-4-4
#define arg_OP_SRC2_SHIFT		22 // 32-4-4-2

#define arg_OP_DEST_MASK		0xf0000000
#define arg_OP_SRC_MASK			0x0f000000
#define arg_OP_SRC2_MASK		0x00c00000
					
#define arg_NEED_MODRM_BYTE		0x00000001
					
#define arg_DEST_IS_MODRM_REG		((dword) (arg_OP_IS_MODRM_REG << arg_OP_DEST_SHIFT))
#define arg_SRC_IS_MODRM_REG		(arg_OP_IS_MODRM_REG << arg_OP_SRC_SHIFT)
					
#define arg_DEST_IS_MODRM_MODRM		((dword) (arg_OP_IS_MODRM_MODRM << arg_OP_DEST_SHIFT))
#define arg_SRC_IS_MODRM_MODRM		(arg_OP_IS_MODRM_MODRM << arg_OP_SRC_SHIFT)
					
#define arg_MODRM_ONLY_M_MODRM		0x00000002 // only refer to memory or register
#define arg_MODRM_ONLY_R_MODRM		0x00000004 // only refer to memory or register

#define	arg_EXTRA_MODRM_REG_IS_SPECIAL	0x00000008
					
#define arg_DEST_IS_IMM			(arg_OP_IS_IMM << arg_OP_DEST_SHIFT)
#define arg_SRC_IS_IMM			(arg_OP_IS_IMM << arg_OP_SRC_SHIFT)
#define arg_SRC2_IS_IMM			(arg_OP_IS_IMM << arg_OP_SRC2_SHIFT)
					
#define arg_DEST_IS_MEM_DISP		(arg_OP_IS_MEM_DISP << arg_OP_DEST_SHIFT)
#define arg_SRC_IS_MEM_DISP		(arg_OP_IS_MEM_DISP << arg_OP_SRC_SHIFT)
					
#define arg_OPSIZE_8			0x00000010
#define arg_OPSIZE_16			0x00000020
#define arg_OPSIZE_32			0x00000040
#define arg_OPSIZE_16_32		0x00000060	// 16/32, depending in opsize-prefix
#define arg_OPSIZE_64			0x00000080
#define arg_OPSIZE_128			0x00000100
					
#define arg_OPSIZE_NORMAL_MASK		0x000001f0
#define arg_OPSIZE_NORMAL_SHIFT		4
					
#define arg_OPSIZE_IN_EXTRA2		0x00000200 // (opsize in extra1)
#define arg_OPSIZE_IN_EXTRAS_16_32	0x00000400 // (opsize32 in extra1, opsize16 in extra2)

#define arg_OPSIZE_MASK			(arg_OPSIZE_8 | arg_OPSIZE_16 | arg_OPSIZE_32 | arg_OPSIZE_16_32 | arg_OPSIZE_64 | arg_OPSIZE_128 | arg_OPSIZE_IN_EXTRA2 | arg_OPSIZE_IN_EXTRAS_16_32)
					
#define arg_SIGN_EXTEND_IMMEDIATE	0x00000800
					
#define arg_LAST_OPRND_OPSIZE_8		0x00001000	
#define arg_LAST_OPRND_OPSIZE_16	0x00002000	
#define arg_LAST_OPRND_DBL_OPSIZE	0x00004000
#define arg_LAST_OPRND_HALF_OPSIZE	0x00008000
					
#define arg_LAST_OPRND_OPSIZE_MASK	(arg_LAST_OPRND_OPSIZE_8 | arg_LAST_OPRND_OPSIZE_16 | arg_LAST_OPRND_DBL_OPSIZE | arg_LAST_OPRND_HALF_OPSIZE)
					
#define	arg_LAST_OP_MEM_OR_IMM_FAR_PTR	0x00010000
					
#define	arg_DEST_AS_REG_EXTRA		(arg_OP_AS_REG_EXTRA << arg_OP_DEST_SHIFT)
#define	arg_SRC_AS_REG_EXTRA		(arg_OP_AS_REG_EXTRA << arg_OP_SRC_SHIFT)
#define	arg_SRC2_AS_REG_EXTRA		(arg_OP_AS_REG_EXTRA << arg_OP_SRC2_SHIFT)
#define	arg_SRC_AS_REG_EXTRA2		(arg_OP_AS_REG_EXTRA2 << arg_OP_SRC_SHIFT)
					
#define	arg_DEST_LOW_3BITS_REG		(arg_OP_LOW_3BITS_REG << arg_OP_DEST_SHIFT)
#define	arg_SRC_LOW_3BITS_REG		(arg_OP_LOW_3BITS_REG << arg_OP_SRC_SHIFT)
					
#define arg_EXTRA2_DEST_SRC_SI_DI	0x00020000

#define arg_SRC_CONST_1			(arg_OP_CONST_1 << arg_OP_SRC_SHIFT)

#define arg_MODRM_RT_FPU_REG		0x01L // ST(i)
#define arg_MODRM_RT_MMX_REG		0x02L // MMi
#define arg_MODRM_RT_SSE_REG		0x03L // XMMi

#define arg_MODRM_RT_MASK		0x03L

#define arg_MODRM_RT_REG_SHIFT		20 // 32-4-4-2-2 
#define arg_MODRM_RT_RM_SHIFT		18 // 32-4-4-2-2-2

#define arg_MODRM_RT_REG_MASK		0x00300000
#define arg_MODRM_RT_RM_MASK		0x000c0000

#define arg_MODRM_REG_MMX_REG		(arg_MODRM_RT_MMX_REG << arg_MODRM_RT_REG_SHIFT)
#define arg_MODRM_REG_SSE_REG		(arg_MODRM_RT_SSE_REG << arg_MODRM_RT_REG_SHIFT)

#define arg_MODRM_RM_FPU_REG		(arg_MODRM_RT_FPU_REG << arg_MODRM_RT_RM_SHIFT)
#define arg_MODRM_RM_MMX_REG		(arg_MODRM_RT_MMX_REG << arg_MODRM_RT_RM_SHIFT)
#define arg_MODRM_RM_SSE_REG		(arg_MODRM_RT_SSE_REG << arg_MODRM_RT_RM_SHIFT)

//////////////////////////////////////////////////////////////////////

#define ext_M512					(0x80 | 9) // 0x80 to mark different notation,  | 9 as 512 is (1 << 9)

#define OPSIZE_FROM_EXTRA(extra) ((extra & 0x80) ? (((word) 1) << ((extra) & 0x1f)) : (word) (extra))

//////////////////////////////////////////////////////////////////////

// we should be able to remove ext2_DEST_MEM_SI and ext2_SRC_MEM_DI as they are not used
// The only problem is that the instruction manual is as far as I can see not
// self consistent regarding placing always DI as destination and SI as source operands,
// (although the names imply that)

#define ext2_DEST_MEM_DI		0x01
#define ext2_SRC_MEM_DI			0x02
#define ext2_DEST_MEM_SI		0x04
#define ext2_SRC_MEM_SI			0x08
					
//#define ext2_X			ext2_DEST_MEM_SI
#define ext2_Y				ext2_DEST_MEM_DI
#define ext2__X				ext2_SRC_MEM_SI
//#define ext2__Y			ext2_SRC_MEM_DI
//#define ext2_X_Y			(ext2_DEST_MEM_SI | ext2_SRC_MEM_DI)
#define ext2_Y_X			(ext2_DEST_MEM_DI | ext2_SRC_MEM_SI)
/*					
#define ext2_SSE_PREFIX_		0x00
#define ext2_SSE_PREFIX_NO		0x01
*/					
#define ext2_SSE_PREFIX_66		0x02
#define ext2_SSE_PREFIX_NO_66	    	0x03
#define ext2_SSE_PREFIX_F3		0x04
#define ext2_SSE_PREFIX_NO_F3		0x05
#define ext2_SSE_PREFIX_66_F3   	0x06
#define ext2_SSE_PREFIX_NO_66_F3   	0x07
#define ext2_SSE_PREFIX_F2		0x08
#define ext2_SSE_PREFIX_NO_F2		0x09
#define ext2_SSE_PREFIX_66_F2		0x0a
#define ext2_SSE_PREFIX_NO_66_F2	0x0b
#define ext2_SSE_PREFIX_F3_F2		0x0c
#define ext2_SSE_PREFIX_NO_F3_F2	0x0d
#define ext2_SSE_PREFIX_66_F3_F2	0x0e
#define ext2_SSE_PREFIX_NO_66_F3_F2	0x0f

//////////////////////////////////////////////////////////////////////

#define arg_MODRM			(arg_NEED_MODRM_BYTE)

#define arg_E				(arg_DEST_IS_MODRM_MODRM | arg_NEED_MODRM_BYTE)
#define arg__E				(arg_SRC_IS_MODRM_MODRM | arg_NEED_MODRM_BYTE)

#define arg_G				(arg_DEST_IS_MODRM_REG | arg_NEED_MODRM_BYTE)
#define arg__G				(arg_SRC_IS_MODRM_REG | arg_NEED_MODRM_BYTE)

#define arg_Q				(arg_DEST_IS_MODRM_MODRM | arg_NEED_MODRM_BYTE | arg_MODRM_RM_MMX_REG)
#define arg_QR				(arg_DEST_IS_MODRM_MODRM | arg_NEED_MODRM_BYTE | arg_MODRM_RM_MMX_REG | arg_MODRM_ONLY_R_MODRM)
#define arg__Q				(arg_SRC_IS_MODRM_MODRM | arg_NEED_MODRM_BYTE | arg_MODRM_RM_MMX_REG)
#define arg__QR				(arg_SRC_IS_MODRM_MODRM | arg_NEED_MODRM_BYTE | arg_MODRM_RM_MMX_REG | arg_MODRM_ONLY_R_MODRM)

#define arg_P				(arg_DEST_IS_MODRM_REG | arg_NEED_MODRM_BYTE | arg_MODRM_REG_MMX_REG)
#define arg__P				(arg_SRC_IS_MODRM_REG | arg_NEED_MODRM_BYTE | arg_MODRM_REG_MMX_REG)

#define arg_W				(arg_DEST_IS_MODRM_MODRM | arg_NEED_MODRM_BYTE | arg_MODRM_RM_SSE_REG)
#define arg_WR				(arg_DEST_IS_MODRM_MODRM | arg_NEED_MODRM_BYTE | arg_MODRM_RM_SSE_REG | arg_MODRM_ONLY_R_MODRM)
#define arg__W				(arg_SRC_IS_MODRM_MODRM | arg_NEED_MODRM_BYTE | arg_MODRM_RM_SSE_REG)
#define arg__WR				(arg_SRC_IS_MODRM_MODRM | arg_NEED_MODRM_BYTE | arg_MODRM_RM_SSE_REG | arg_MODRM_ONLY_R_MODRM)

#define arg_V				(arg_DEST_IS_MODRM_REG | arg_NEED_MODRM_BYTE | arg_MODRM_REG_SSE_REG)
#define arg__V				(arg_SRC_IS_MODRM_REG | arg_NEED_MODRM_BYTE | arg_MODRM_REG_SSE_REG)

#define arg_I				(arg_DEST_IS_IMM)
#define arg__I				(arg_SRC_IS_IMM)
#define arg___I				(arg_SRC2_IS_IMM)

#define arg_M				(arg_DEST_IS_MODRM_MODRM | arg_MODRM_ONLY_M_MODRM | arg_NEED_MODRM_BYTE)
#define arg__M				(arg_SRC_IS_MODRM_MODRM | arg_MODRM_ONLY_M_MODRM | arg_NEED_MODRM_BYTE)

#define arg_O				(arg_DEST_IS_MEM_DISP)
#define arg__O				(arg_SRC_IS_MEM_DISP)

#define arg_R				(arg_DEST_IS_MODRM_MODRM | arg_MODRM_ONLY_R_MODRM | arg_NEED_MODRM_BYTE)
#define arg__R				(arg_SRC_IS_MODRM_MODRM | arg_MODRM_ONLY_R_MODRM | arg_NEED_MODRM_BYTE)

#define arg_S				(arg_DEST_IS_MODRM_REG | arg_EXTRA_MODRM_REG_IS_SPECIAL | arg_NEED_MODRM_BYTE)
#define arg__S				(arg_SRC_IS_MODRM_REG  | arg_EXTRA_MODRM_REG_IS_SPECIAL | arg_NEED_MODRM_BYTE)

#define arg_CDT				(arg_DEST_IS_MODRM_REG | arg_EXTRA_MODRM_REG_IS_SPECIAL | arg_NEED_MODRM_BYTE)
#define arg__CDT			(arg_SRC_IS_MODRM_REG  | arg_EXTRA_MODRM_REG_IS_SPECIAL | arg_NEED_MODRM_BYTE)

#define arg_Y				(arg_EXTRA2_DEST_SRC_SI_DI)	// further info in extra2
//#define arg__Y			(arg_EXTRA2_DEST_SRC_SI_DI)	
//#define arg_X				(arg_EXTRA2_DEST_SRC_SI_DI)	
#define arg__X				(arg_EXTRA2_DEST_SRC_SI_DI)	

#define arg_reg				(arg_DEST_AS_REG_EXTRA)	
#define arg__reg			(arg_SRC_AS_REG_EXTRA)	
#define arg__reg2			(arg_SRC_AS_REG_EXTRA2)	
#define arg___reg			(arg_SRC2_AS_REG_EXTRA)	

#define arg_REG				(arg_DEST_LOW_3BITS_REG)
#define arg__REG			(arg_SRC_LOW_3BITS_REG)

#define arg_STREG			(arg_DEST_IS_MODRM_MODRM | arg_MODRM_RM_FPU_REG | arg_MODRM_ONLY_R_MODRM | arg_NEED_MODRM_BYTE)
#define arg__STREG			(arg_SRC_IS_MODRM_MODRM  | arg_MODRM_RM_FPU_REG | arg_MODRM_ONLY_R_MODRM | arg_NEED_MODRM_BYTE)

//////////////////////////////////////////////////////////////////////

#define arg_Ew				(arg_E | arg_OPSIZE_16)

#define arg_Eb_Gb			(arg_E | arg__G | arg_OPSIZE_8)			
#define arg_Ev_Gv			(arg_E | arg__G | arg_OPSIZE_16_32)	
#define arg_Gb_Eb			(arg_G | arg__E | arg_OPSIZE_8)	
#define arg_Gv_Ev			(arg_G | arg__E | arg_OPSIZE_16_32)	

#define arg_Gv_Eb			(arg_G | arg__E | arg_OPSIZE_16_32 | arg_LAST_OPRND_OPSIZE_8)
#define arg_Gv_Ew			(arg_G | arg__E | arg_OPSIZE_16_32 | arg_LAST_OPRND_OPSIZE_16)

#define arg_Gv_Ev_Iv		(arg_G | arg__E | arg___I | arg_OPSIZE_16_32)	
#define arg_Gv_Ev_Ib		(arg_G | arg__E | arg___I | arg_OPSIZE_16_32 | arg_LAST_OPRND_OPSIZE_8)	
#define arg_Gv_Ev_Ibs		(arg_G | arg__E | arg___I | arg_OPSIZE_16_32 | arg_LAST_OPRND_OPSIZE_8 | arg_SIGN_EXTEND_IMMEDIATE)	

#define arg_Ev_Gv_Ib		(arg_E | arg__G | arg___I | arg_OPSIZE_16_32 | arg_LAST_OPRND_OPSIZE_8)

#define arg_Ev_Gv_reg		(arg_E | arg__G | arg___reg | arg_OPSIZE_16_32)

#define arg_Ew_Gw			(arg_E | arg__G | arg_OPSIZE_16)	

#define arg_Gv_Ma			(arg_G | arg__M | arg_OPSIZE_16_32 | arg_LAST_OPRND_DBL_OPSIZE)	
#define arg_Gv_M			(arg_G | arg__M | arg_OPSIZE_16_32)	
#define arg_Gv_Mp			(arg_G | arg__M | arg_OPSIZE_16_32 | arg_LAST_OP_MEM_OR_IMM_FAR_PTR)

#define arg_Eb_Ib			(arg_E | arg__I | arg_OPSIZE_8) 	
#define arg_Ev_Iv			(arg_E | arg__I | arg_OPSIZE_16_32) 	
#define arg_Ev_Ib			(arg_E | arg__I | arg_OPSIZE_16_32 | arg_LAST_OPRND_OPSIZE_8) 	
#define arg_Ev_Ibs			(arg_E | arg__I | arg_OPSIZE_16_32 | arg_LAST_OPRND_OPSIZE_8 | arg_SIGN_EXTEND_IMMEDIATE) 	

#define arg_Eb_1			(arg_E | arg_SRC_CONST_1 | arg_OPSIZE_8)
#define arg_Ev_1			(arg_E | arg_SRC_CONST_1 | arg_OPSIZE_16_32)

#define arg_Eb_reg2			(arg_E | arg__reg2 | arg_OPSIZE_8)
#define arg_Ev_reg2			(arg_E | arg__reg2 | arg_OPSIZE_16_32)

#define arg_Ms				(arg_M | arg_OPSIZE_IN_EXTRA2)
#define arg__Ms				(arg__M | arg_OPSIZE_IN_EXTRA2)
#define arg_Mvs				(arg_M | arg_OPSIZE_IN_EXTRAS_16_32)
#define arg__Mvs			(arg__M | arg_OPSIZE_IN_EXTRAS_16_32)
#define arg_Mb				(arg_M | arg_OPSIZE_8)
#define arg__Mb				(arg__M | arg_OPSIZE_8)
#define arg_Mw				(arg_M | arg_OPSIZE_16)
#define arg__Mw				(arg__M | arg_OPSIZE_16)
#define arg_Md				(arg_M | arg_OPSIZE_32)
#define arg__Md				(arg__M | arg_OPSIZE_32)
#define arg_Mq				(arg_M | arg_OPSIZE_64)
#define arg__Mq				(arg__M | arg_OPSIZE_64)

#define arg_Md_Gd			(arg_M | arg__G | arg_OPSIZE_32)

#define arg_Ms_reg			(arg_M | arg__reg | arg_OPSIZE_IN_EXTRA2)
#define arg_reg_Ms			(arg_reg | arg__M | arg_OPSIZE_IN_EXTRA2)
#define arg_Mw_reg			(arg_M | arg__reg | arg_OPSIZE_16)
#define arg_reg_Mw			(arg_reg | arg__M | arg_OPSIZE_16)
#define arg_Md_reg			(arg_M | arg__reg | arg_OPSIZE_32)
#define arg_reg_Md			(arg_reg | arg__M | arg_OPSIZE_32)
#define arg_Mq_reg			(arg_M | arg__reg | arg_OPSIZE_64)
#define arg_reg_Mq			(arg_reg | arg__M | arg_OPSIZE_64)

#define arg_reg_reg2			(arg_reg | arg__reg2 | arg_OPSIZE_16_32)

#define arg_reg_Ib			(arg_reg | arg__I | arg_OPSIZE_8)	
#define arg_reg_Iv			(arg_reg | arg__I | arg_OPSIZE_16_32)	

#define arg_reg_Ob			(arg_reg | arg__O | arg_OPSIZE_8)
#define arg_reg_Ov			(arg_reg | arg__O | arg_OPSIZE_16_32)
#define arg_Ob_reg			(arg_O | arg__reg | arg_OPSIZE_8)
#define arg_Ov_reg			(arg_O | arg__reg | arg_OPSIZE_16_32)

#define arg_Ib_reg			(arg_I | arg__reg | arg_OPSIZE_8)
#define arg_Iv_reg			(arg_I | arg__reg | arg_OPSIZE_16_32)				

#define arg_reg_Ib			(arg_reg | arg__I | arg_OPSIZE_8)
#define arg_reg_Iv			(arg_reg | arg__I | arg_OPSIZE_16_32)				

#define arg_Ib				(arg_I | arg_OPSIZE_8)
#define arg_Iv				(arg_I | arg_OPSIZE_16_32)				
#define arg_Iw				(arg_I | arg_OPSIZE_16)				

#define arg__Ib				(arg__I | arg_OPSIZE_8)
#define arg__Iv				(arg__I | arg_OPSIZE_16_32)				
#define arg__Iw				(arg__I | arg_OPSIZE_16)				

#define arg__Iw_Ib			(arg__I | arg___I | arg_OPSIZE_16 | arg_LAST_OPRND_OPSIZE_8)

#define arg_Jb				(arg_I | arg_OPSIZE_8)	
#define arg_Jv				(arg_I | arg_OPSIZE_16_32)

#define arg_Ap				(arg_I | arg_OPSIZE_16_32 | arg_LAST_OP_MEM_OR_IMM_FAR_PTR)

#define arg_Eb				(arg_E | arg_OPSIZE_8)
#define arg_Ev				(arg_E | arg_OPSIZE_16_32)
#define arg_Mp				(arg_M | arg_OPSIZE_16_32 | arg_LAST_OP_MEM_OR_IMM_FAR_PTR)

#define arg__Ev				(arg__E | arg_OPSIZE_16_32)

#define arg_Yb_reg			(arg_Y | arg__reg | arg_OPSIZE_8)	
#define arg_Yv_reg			(arg_Y | arg__reg | arg_OPSIZE_16_32)	

#define arg_Yb_Xb			(arg_Y | arg__X | arg_OPSIZE_8)	
#define arg_Yv_Xv			(arg_Y | arg__X | arg_OPSIZE_16_32)	

//#define arg_Xb_Yb			(arg_X | arg__Y | arg_OPSIZE_8)	
//#define arg_Xv_Yv			(arg_X | arg__Y | arg_OPSIZE_16_32)	

#define arg_reg_Xb			(arg_reg | arg__X | arg_OPSIZE_8)	
#define arg_reg_Xv			(arg_reg | arg__X | arg_OPSIZE_16_32)	

//#define arg_reg_Yb			(arg_reg | arg__Y | arg_OPSIZE_8)	
//#define arg_reg_Yv			(arg_reg | arg__Y | arg_OPSIZE_16_32)	

#define arg_Rd_CDTd			(arg_R | arg__CDT | arg_OPSIZE_32)
#define arg_CDTd_Rd			(arg_CDT | arg__R | arg_OPSIZE_32)

#define arg_Ew_Sw			(arg_E | arg__S | arg_OPSIZE_16)
#define arg_Sw_Ew			(arg_S | arg__E | arg_OPSIZE_16)

#define arg_REG8_Ib			(arg_REG | arg__I | arg_OPSIZE_8)
#define arg_eREG_Iv			(arg_REG | arg__I | arg_OPSIZE_16_32)

#define arg_eREG			(arg_REG | arg_OPSIZE_16_32)
#define arg_eREG_reg			(arg_REG | arg__reg | arg_OPSIZE_16_32)

#define arg__eREG			(arg__REG | arg_OPSIZE_16_32)

#define arg_EREG			(arg_REG | arg_OPSIZE_32)

#define arg_STREG_reg			(arg_STREG | arg__reg)
#define arg_reg_STREG			(arg_reg | arg__STREG)
#define arg_Mq_STREG			(arg_M | arg__STREG | arg_OPSIZE_64)

// don't specify size for MMX/SSE operands, it only applies to any memory operands
// or other register operands

#define arg_P_Q				(arg_P | arg__Q)			
#define arg_P_QR			(arg_P | arg__QR)			
#define arg_Q_P				(arg_Q | arg__P)			
#define arg_V_W				(arg_V | arg__W)			
#define arg_V_WR			(arg_V | arg__WR)			
#define arg_W_V				(arg_W | arg__V)			
#define arg_P_W				(arg_P | arg__W)			
#define arg_P_WR			(arg_P | arg__WR)			
#define arg_V_Q				(arg_V | arg__Q)			
#define arg_V_QR			(arg_V | arg__QR)			

#define arg_Pd_Qd			(arg_P_Q | arg_OPSIZE_32 ) // size applies to Qd if [mem]			
#define arg_Pq_Qq			(arg_P_Q | arg_OPSIZE_64 ) // size applies to Qq if [mem]			
#define arg_Pq_Qd			(arg_P_Q | arg_OPSIZE_32 ) // size applies to Qd if [mem]			

#define arg_Pq_QRq			(arg_P_QR)			

#define arg_Qd_Pd			(arg_Q_P | arg_OPSIZE_32 ) // size applies to Qd if [mem]			
#define arg_Qq_Pq			(arg_Q_P | arg_OPSIZE_64 ) // size applies to Qq if [mem]			

#define arg_Vd_Wd			(arg_V_W | arg_OPSIZE_32 ) // size applies to Wd if [mem]			
#define arg_Vq_Wq			(arg_V_W | arg_OPSIZE_64 ) // size applies to Wq if [mem]			
#define arg_Vo_Wo			(arg_V_W | arg_OPSIZE_128) // size applies to Wo if [mem]			

#define arg_Vq_WRq			(arg_V_WR)			
#define arg_Vo_WRo			(arg_V_WR)			

#define arg_Vd_Wq			(arg_W_V | arg_OPSIZE_64 ) // size applies to Wq if [mem]			
#define arg_Vq_Wd			(arg_W_V | arg_OPSIZE_32 ) // size applies to Wd if [mem]			
#define arg_Vo_Wq			(arg_W_V | arg_OPSIZE_64 ) // size applies to Wq if [mem]			

#define arg_Wd_Vd			(arg_W_V | arg_OPSIZE_32 ) // size applies to Wd if [mem]			
#define arg_Wq_Vq			(arg_W_V | arg_OPSIZE_64 ) // size applies to Wq if [mem]			
#define arg_Wo_Vo			(arg_W_V | arg_OPSIZE_128) // size applies to Wo if [mem]			

#define arg_Pq_Wq			(arg_P_W | arg_OPSIZE_64 ) // size applies to Wq if [mem]			
#define arg_Pq_Wo			(arg_P_W | arg_OPSIZE_128) // size applies to Wo if [mem]			

#define arg_Pq_WRq			(arg_P_WR)			

#define arg_Vo_Qq			(arg_V_Q | arg_OPSIZE_64 ) // size applies to Qq if [mem]			

#define arg_Vo_QRq			(arg_V_QR)			

#define arg_Pq_Ed			(arg_P | arg__E | arg_OPSIZE_32) // size applies to Ed		
#define arg_Ed_Pd			(arg_E | arg__P | arg_OPSIZE_32) // size applies to Ed		

#define arg_Gd_Wd			(arg_G | arg__W | arg_OPSIZE_32) // size applies to Gd			
#define arg_Gd_Wq			(arg_G | arg__W | arg_OPSIZE_32) // size applies to Gd			
#define arg_Vd_Ed			(arg_V | arg__E | arg_OPSIZE_32) // size applies to Ed		
#define arg_Vq_Ed			(arg_V | arg__E | arg_OPSIZE_32) // size applies to Ed
#define arg_Vo_Ed			(arg_V | arg__E | arg_OPSIZE_32) // size applies to Ed
#define arg_Ed_Vd			(arg_G | arg__W | arg_OPSIZE_32) // size applies to Ed			
#define arg_Ed_Vo			(arg_G | arg__W | arg_OPSIZE_32) // size applies to Ed			

#define arg_Rd_Pq			(arg_R | arg__P | arg_OPSIZE_32) // size applies to Rd			
#define arg_Rd_Vo			(arg_R | arg__V | arg_OPSIZE_32) // size applies to Rd			

#define arg_Vq_Mq			(arg_V | arg__M | arg_OPSIZE_64) // size applies to Mq			
#define arg_Mq_Vq			(arg_M | arg__V | arg_OPSIZE_64) // size applies to Mq			

#define arg_Mo_Vo			(arg_M | arg__V | arg_OPSIZE_128)// size applies to Mo			
#define arg_Vo_Mq			(arg_V | arg__M | arg_OPSIZE_64) // size applies to Mq

#define arg_Mq_Pq			(arg_M | arg__P | arg_OPSIZE_64) // size applies to Mq
/*
#define arg_Pq_Ib			(arg_P | arg__I | arg_LAST_OPRND_OPSIZE_8)	
#define arg_Vo_Ib			(arg_V | arg__I | arg_LAST_OPRND_OPSIZE_8)	
*/
#define arg_QRq_Ib			(arg_QR | arg__I | arg_LAST_OPRND_OPSIZE_8)	
#define arg_WRo_Ib			(arg_WR | arg__I | arg_LAST_OPRND_OPSIZE_8)	
/*
#define arg_Pq_Mw_Ib			(arg_P | arg__M | arg___I | arg_OPSIZE_16 | arg_LAST_OPRND_OPSIZE_8) // size applies to Mw	
#define arg_Vo_Mw_Ib			(arg_V | arg__M | arg___I | arg_OPSIZE_16 | arg_LAST_OPRND_OPSIZE_8) // size applies to Mw

#define arg_Pq_Rd_Ib			(arg_P | arg__R | arg___I | arg_OPSIZE_32 | arg_LAST_OPRND_OPSIZE_8) // size applies to Rd
#define arg_Vo_Rd_Ib			(arg_V | arg__R | arg___I | arg_OPSIZE_32 | arg_LAST_OPRND_OPSIZE_8) // size applies to Rd
*/
#define arg_Pq_Ew_Ib			(arg_P | arg__E | arg___I | arg_OPSIZE_16 | arg_LAST_OPRND_OPSIZE_8) // size applies to Mw	
#define arg_Vo_Ew_Ib			(arg_V | arg__E | arg___I | arg_OPSIZE_16 | arg_LAST_OPRND_OPSIZE_8) // size applies to Mw

#define arg_Pq_Qq_Ib			(arg_P_Q | arg___I | arg_OPSIZE_64  | arg_LAST_OPRND_OPSIZE_8) // size applies to Qq if [mem]
#define arg_Vd_Wd_Ib			(arg_V_W | arg___I | arg_OPSIZE_32  | arg_LAST_OPRND_OPSIZE_8) // size applies to Wd if [mem]	
#define arg_Vq_Wq_Ib			(arg_V_W | arg___I | arg_OPSIZE_64  | arg_LAST_OPRND_OPSIZE_8) // size applies to Wq if [mem]
#define arg_Vo_Wo_Ib			(arg_V_W | arg___I | arg_OPSIZE_128 | arg_LAST_OPRND_OPSIZE_8) // size applies to Wo if [mem]	
#define arg_Gd_QRq_Ib			(arg_G | arg__QR | arg___I | arg_OPSIZE_32 | arg_LAST_OPRND_OPSIZE_8) // size applies to Gd
#define arg_Gd_WRo_Ib			(arg_G | arg__WR | arg___I | arg_OPSIZE_32 | arg_LAST_OPRND_OPSIZE_8) // size applies to Gd	

#define arg_Vo_Mq_OR_Vq_Wq		(arg_V_W | arg_OPSIZE_64) // size applies to Wq if [mem]
#define arg_Mq_Vq_OR_Wq_Vq		(arg_W_V | arg_OPSIZE_64) // size applies to Wq if [mem]
/*
#define arg_Pq_MwORRd_Ib		(arg_P | arg__G | arg___I | arg_OPSIZE_16 | arg_OPSIZE_32_IF_REG | arg_LAST_OPRND_OPSIZE_8) // size applies to Gw
#define arg_Vo_MwORRd_Ib		(arg_V | arg__G | arg___I | arg_OPSIZE_16 | arg_OPSIZE_32_IF_REG | arg_LAST_OPRND_OPSIZE_8) // size applies to Gw
*/
// no arguments

#define arg_NONE			0

//////////////////////////////////////////////////////////////////////

#define arg_PREFIX_LOCK_REP 		0x10

#define arg_PREFIX_LOCK			(arg_PREFIX_LOCK_REP | 0x01) // the common 0x0010 is because they may not occur in same opcode  (one prefix from each group each...)
#define arg_PREFIX_REP			(arg_PREFIX_LOCK_REP | 0x02) // the common 0x0010 is because they may not occur in same opcode  (one prefix from each group each...)
#define arg_PREFIX_SEGMENT		0x20
#define arg_PREFIX_OPSIZE		0x40
#define arg_PREFIX_ADDRSIZE		0x80

#define arg_PREFIX_LOCK_REP_MASK 	(arg_PREFIX_LOCK | arg_PREFIX_REP)


#define arg_PREFIX_MASK			0xff

//////////////////////////////////////////////////////////////////////

#include "gen/GroupInstrConstants.h" // dep_chk_unknown      // used in extra byte(s)
#include "gen/MMX_SSE_InstrConstants.h" // dep_chk_unknown     

//////////////////////////////////////////////////////////////////////

struct OpcodeInfo
{
  dword	m_nArguments;	
  word	m_nInstruction;

  byte	m_nExtra1;		
  byte	m_nExtra2;		
};

//////////////////////////////////////////////////////////////////////

extern const OpcodeInfo infoByte[2][256];
extern const OpcodeInfo infoGroups[ext_NUM_GROUPS][8];
extern const OpcodeInfo infoMMX_SSE[ext_MMX_SSE_NUM_GROUPS<<1];

//////////////////////////////////////////////////////////////////////

struct DATAPTR
{
  byte* pData;
  byte* pDataEnd;
};

//////////////////////////////////////////////////////////////////////

#ifndef NDEBUG
void OpcodeInfoSanityCheck (OpcodeInfo& info);
#endif

//////////////////////////////////////////////////////////////////////

bool IsValidLockInstruction (word nInstr);


class InstructionDecoder  
{
 public:
  InstructionDecoder();
  virtual ~InstructionDecoder();
  
 public:
  bool	m_bCodeSize32;
  
 public:
  Instruction* Decode (DataPtr& dataPtr) const;
  
};

#endif // !defined(__INSTRUCTION_DECODER_H__INCLUDED_)

