/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// InstructionDecoder_Info.cpp: implementation of the InstructionDecoder class.
//
//////////////////////////////////////////////////////////////////////

#include "InstructionDecoder.h" 

#include "Instruction.h" 
#include "OperandRegister.h" 

//////////////////////////////////////////////////////////////////////

// s=[];for i=0:255;s=[s;sprintf('	{ 0,					instrRESERVED,		0			},	// 0x%02x',i)];end

const OpcodeInfo infoByte[2][256] =
{
{
	{ arg_Eb_Gb, 			instrADD,			0,0				},	// 0x00	// 1
	{ arg_Ev_Gv, 			instrADD,			0,0				},	// 0x01
	{ arg_Gb_Eb, 			instrADD,			0,0				},	// 0x02
	{ arg_Gv_Ev, 			instrADD,			0,0				},	// 0x03
	{ arg_reg_Ib, 			instrADD,			regAL,0			},	// 0x04
	{ arg_reg_Iv,			instrADD,			regeAX,0		},	// 0x05
	{ arg__reg,				instrPUSH,			regES,0			},	// 0x06
	{ arg_reg,				instrPOP,			regES,0			},	// 0x07
	{ arg_Eb_Gb, 			instrOR,			0,0				},	// 0x08
	{ arg_Ev_Gv, 			instrOR,			0,0				},	// 0x09
	{ arg_Gb_Eb, 			instrOR,			0,0				},	// 0x0a
	{ arg_Gv_Ev, 			instrOR,			0,0				},	// 0x0b
	{ arg_reg_Ib, 			instrOR,			regAL,0			},	// 0x0c
	{ arg_reg_Iv,			instrOR,			regeAX,0		},	// 0x0d
	{ arg__reg,				instrPUSH,			regCS,0			},	// 0x0e
	{ arg_NONE,				instrESCAPE2BYTE,	0,0				},	// 0x0f

	{ arg_Eb_Gb, 			instrADC,			0,0				},	// 0x10	// 1
	{ arg_Ev_Gv, 			instrADC,			0,0				},	// 0x11
	{ arg_Gb_Eb, 			instrADC,			0,0				},	// 0x12
	{ arg_Gv_Ev, 			instrADC,			0,0				},	// 0x13
	{ arg_reg_Ib, 			instrADC,			regAL,0			},	// 0x14
	{ arg_reg_Iv,			instrADC,			regeAX,0		},	// 0x15
	{ arg__reg,				instrPUSH,			regSS,0			},	// 0x16
	{ arg_reg,				instrPOP,			regSS,0			},	// 0x17
	{ arg_Eb_Gb, 			instrSBB,			0,0				},	// 0x18
	{ arg_Ev_Gv,			instrSBB,			0,0				},	// 0x19
	{ arg_Gb_Eb, 			instrSBB,			0,0				},	// 0x1a
	{ arg_Gv_Ev, 			instrSBB,			0,0				},	// 0x1b
	{ arg_reg_Ib, 			instrSBB,			regAL,0			},	// 0x1c
	{ arg_reg_Iv,			instrSBB,			regeAX,0		},	// 0x1d
	{ arg__reg,				instrPUSH,			regDS,0			},	// 0x1e
	{ arg_reg,				instrPOP,			regDS,0			},	// 0x1f

	{ arg_Eb_Gb, 			instrAND,			0,0				},	// 0x20	// 1
	{ arg_Ev_Gv, 			instrAND,			0,0				},	// 0x21
	{ arg_Gb_Eb, 			instrAND,			0,0				},	// 0x22
	{ arg_Gv_Ev, 			instrAND,			0,0				},	// 0x23
	{ arg_reg_Ib, 			instrAND,			regAL,0			},	// 0x24
	{ arg_reg_Iv,			instrAND,			regeAX,0		},	// 0x25
	{ arg_PREFIX_SEGMENT,	instrPREFIX,		regES,0			},	// 0x26
	{ arg_NONE,				instrDAA,			0,0				},	// 0x27
	{ arg_Eb_Gb, 			instrSUB,			0,0				},	// 0x28
	{ arg_Ev_Gv, 			instrSUB,			0,0				},	// 0x29
	{ arg_Gb_Eb, 			instrSUB,			0,0				},	// 0x2a
	{ arg_Gv_Ev, 			instrSUB,			0,0				},	// 0x2b
	{ arg_reg_Ib, 			instrSUB,			regAL,0			},	// 0x2c
	{ arg_reg_Iv,			instrSUB,			regeAX,0		},	// 0x2d
	{ arg_PREFIX_SEGMENT,	instrPREFIX,		regCS,0			},	// 0x2e
	{ arg_NONE,				instrDAS,			0,0				},	// 0x2f

	{ arg_Eb_Gb, 			instrXOR,			0,0				},	// 0x30	// 1
	{ arg_Ev_Gv, 			instrXOR,			0,0				},	// 0x31
	{ arg_Gb_Eb, 			instrXOR,			0,0				},	// 0x32
	{ arg_Gv_Ev, 			instrXOR,			0,0				},	// 0x33
	{ arg_reg_Ib, 			instrXOR,			regAL,0			},	// 0x34
	{ arg_reg_Iv,			instrXOR,			regeAX,0		},	// 0x35
	{ arg_PREFIX_SEGMENT,	instrPREFIX,		regSS,0			},	// 0x36
	{ arg_NONE,				instrAAA,			0,0				},	// 0x37
	{ arg_Eb_Gb, 			instrCMP,			0,0				},	// 0x38
	{ arg_Ev_Gv, 			instrCMP,			0,0				},	// 0x39
	{ arg_Gb_Eb, 			instrCMP,			0,0				},	// 0x3a
	{ arg_Gv_Ev, 			instrCMP,			0,0				},	// 0x3b
	{ arg_reg_Ib, 			instrCMP,			regAL,0			},	// 0x3c
	{ arg_reg_Iv,			instrCMP,			regeAX,0		},	// 0x3d
	{ arg_PREFIX_SEGMENT,	instrPREFIX,		regDS,0			},	// 0x3e
	{ arg_NONE,				instrAAS,			0,0				},	// 0x3f

	{ arg_eREG,				instrINC,			0,0				},	// 0x40	// 1
	{ arg_eREG,				instrINC,			0,0				},	// 0x41
	{ arg_eREG,				instrINC,			0,0				},	// 0x42
	{ arg_eREG,				instrINC,			0,0				},	// 0x43
	{ arg_eREG,				instrINC,			0,0				},	// 0x44
	{ arg_eREG,				instrINC,			0,0				},	// 0x45
	{ arg_eREG,				instrINC,			0,0				},	// 0x46
	{ arg_eREG,				instrINC,			0,0				},	// 0x47
	{ arg_eREG,				instrDEC,			0,0				},	// 0x48
	{ arg_eREG,				instrDEC,			0,0				},	// 0x49
	{ arg_eREG,				instrDEC,			0,0				},	// 0x4a
	{ arg_eREG,				instrDEC,			0,0				},	// 0x4b
	{ arg_eREG,				instrDEC,			0,0				},	// 0x4c
	{ arg_eREG,				instrDEC,			0,0				},	// 0x4d
	{ arg_eREG,				instrDEC,			0,0				},	// 0x4e
	{ arg_eREG,				instrDEC,			0,0				},	// 0x4f

	{ arg__eREG,			instrPUSH,			0,0				},	// 0x50	// 1
	{ arg__eREG,			instrPUSH,			0,0				},	// 0x51
	{ arg__eREG,			instrPUSH,			0,0				},	// 0x52
	{ arg__eREG,			instrPUSH,			0,0				},	// 0x53
	{ arg__eREG,			instrPUSH,			0,0				},	// 0x54
	{ arg__eREG,			instrPUSH,			0,0				},	// 0x55
	{ arg__eREG,			instrPUSH,			0,0				},	// 0x56
	{ arg__eREG,			instrPUSH,			0,0				},	// 0x57
	{ arg_eREG,				instrPOP,			0,0				},	// 0x58
	{ arg_eREG,				instrPOP,			0,0				},	// 0x59
	{ arg_eREG,				instrPOP,			0,0				},	// 0x5a
	{ arg_eREG,				instrPOP,			0,0				},	// 0x5b
	{ arg_eREG,				instrPOP,			0,0				},	// 0x5c
	{ arg_eREG,				instrPOP,			0,0				},	// 0x5d
	{ arg_eREG,				instrPOP,			0,0				},	// 0x5e
	{ arg_eREG,				instrPOP,			0,0				},	// 0x5f

	{ arg_NONE,				instrPUSHA,			0,0				},	// 0x60	// 1
	{ arg_NONE,				instrPOPA,			0,0				},	// 0x61
	{ arg_Gv_Ma,			instrBOUND,			0,0				},	// 0x62
	{ arg_Ew_Gw,			instrARPL,			0,0				},	// 0x63
	{ arg_PREFIX_SEGMENT,	instrPREFIX,		regFS,0			},	// 0x64
	{ arg_PREFIX_SEGMENT,	instrPREFIX,		regGS,0			},	// 0x65
	{ arg_PREFIX_OPSIZE,	instrPREFIX,		0,0				},	// 0x66
	{ arg_PREFIX_ADDRSIZE,	instrPREFIX,		0,0				},	// 0x67
	{ arg__Iv,				instrPUSH,			0,0				},	// 0x68
	{ arg_Gv_Ev_Iv,			instrIMUL,			0,0				},	// 0x69
	{ arg__Ib,				instrPUSH,			0,0				},	// 0x6a
	{ arg_Gv_Ev_Ibs,		instrIMUL,			0,0				},	// 0x6b
	{ arg_Yb_reg,			instrINS,			regDX,	ext2_Y	},	// 0x6c
	{ arg_Yv_reg,			instrINS,			regDX,	ext2_Y	},	// 0x6d
	{ arg_reg_Xb,			instrOUTS,			regDX,	ext2__X	},	// 0x6e
	{ arg_reg_Xv,			instrOUTS,			regDX,	ext2__X	},	// 0x6f

	{ arg_Jb,				instrJCC_0,			0,0				},	// 0x70	// 1
	{ arg_Jb,				instrJCC_1,			0,0				},	// 0x71
	{ arg_Jb,				instrJCC_2,			0,0				},	// 0x72
	{ arg_Jb,				instrJCC_3,			0,0				},	// 0x73
	{ arg_Jb,				instrJCC_4,			0,0				},	// 0x74
	{ arg_Jb,				instrJCC_5,			0,0				},	// 0x75
	{ arg_Jb,				instrJCC_6,			0,0				},	// 0x76
	{ arg_Jb,				instrJCC_7,			0,0				},	// 0x77
	{ arg_Jb,				instrJCC_8,			0,0				},	// 0x78
	{ arg_Jb,				instrJCC_9,			0,0				},	// 0x79
	{ arg_Jb,				instrJCC_A,			0,0				},	// 0x7a
	{ arg_Jb,				instrJCC_B,			0,0				},	// 0x7b
	{ arg_Jb,				instrJCC_C,			0,0				},	// 0x7c
	{ arg_Jb,				instrJCC_D,			0,0				},	// 0x7d
	{ arg_Jb,				instrJCC_E,			0,0				},	// 0x7e
	{ arg_Jb,				instrJCC_F,			0,0				},	// 0x7f

	{ arg_Eb_Ib | arg_MODRM,instrGROUP,			ext_GRP_1,0		},	// 0x80	// 1
	{ arg_Ev_Iv | arg_MODRM,instrGROUP,			ext_GRP_1,0		},	// 0x81
	{ arg_Eb_Ib | arg_MODRM,instrGROUP,			ext_GRP_1,0		},	// 0x82
	{ arg_Ev_Ibs | arg_MODRM,instrGROUP,		ext_GRP_1,0		},	// 0x83
	{ arg_Eb_Gb,			instrTEST,			0,0				},	// 0x84
	{ arg_Ev_Gv,			instrTEST,			0,0				},	// 0x85
	{ arg_Eb_Gb,			instrXCHG,			0,0				},	// 0x86
	{ arg_Ev_Gv,			instrXCHG,			0,0				},	// 0x87
	{ arg_Eb_Gb,			instrMOV,			0,0				},	// 0x88
	{ arg_Ev_Gv,			instrMOV,			0,0				},	// 0x89
	{ arg_Gb_Eb,			instrMOV,			0,0				},	// 0x8a
	{ arg_Gv_Ev,			instrMOV,			0,0				},	// 0x8b
	{ arg_Ew_Sw,			instrMOV,			reggSEG,0		},	// 0x8c
	{ arg_Gv_M,				instrLEA,			0,0				},	// 0x8d
	{ arg_Sw_Ew,			instrMOV,			reggSEG,0		},	// 0x8e
	{ arg_Ev,				instrPOP,			0,0				},	// 0x8f

	{ arg_NONE,				instrNOP,			0,0				},	// 0x90	// 1
	{ arg_eREG_reg,			instrXCHG,			regeAX,0		},	// 0x91
	{ arg_eREG_reg,			instrXCHG,			regeAX,0		},	// 0x92
	{ arg_eREG_reg,			instrXCHG,			regeAX,0		},	// 0x93
	{ arg_eREG_reg,			instrXCHG,			regeAX,0		},	// 0x94
	{ arg_eREG_reg,			instrXCHG,			regeAX,0		},	// 0x95
	{ arg_eREG_reg,			instrXCHG,			regeAX,0		},	// 0x96
	{ arg_eREG_reg,			instrXCHG,			regeAX,0		},	// 0x97
	{ arg_reg |
		arg_OPSIZE_16_32,	instrCBWCWDE,		regeAX,0		},	// 0x98
	{ arg_reg_reg2 |
		arg_OPSIZE_16_32,	instrCWDCDQ,		regeDX,regeAX	},	// 0x99
	{ arg_Ap,				instrCALLF_ABS,		0,0				},	// 0x9a
	{ arg_NONE,				instrFWAIT,			0,0				},	// 0x9b
	{ arg_NONE,				instrPUSHF,			0,0				},	// 0x9c // arg__Fv
	{ arg_NONE,				instrPOPF,			0,0				},	// 0x9d // arg_Fv
	{ arg__reg,				instrSAHF,			regAH,0			},	// 0x9e
	{ arg_reg,				instrLAHF,			regAH,0			},	// 0x9f

	{ arg_reg_Ob,			instrMOV,			regAL,0			},	// 0xa0	// 1
	{ arg_reg_Ov,			instrMOV,			regeAX,0		},	// 0xa1
	{ arg_Ob_reg,			instrMOV,			regAL,0			},	// 0xa2
	{ arg_Ov_reg,			instrMOV,			regeAX,0		},	// 0xa3
	{ arg_Yb_Xb,			instrMOVS,			0,	ext2_Y_X	},	// 0xa4
	{ arg_Yv_Xv,			instrMOVS,			0,	ext2_Y_X	},	// 0xa5
	{ arg_Yb_Xb,			instrCMPS,			0,	ext2_Y_X	},	// 0xa6
	{ arg_Yv_Xv,			instrCMPS,			0,	ext2_Y_X	},	// 0xa7
	{ arg_reg_Ib,			instrTEST,			regAL,0			},	// 0xa8
	{ arg_reg_Iv,			instrTEST,			regeAX,0		},	// 0xa9
	{ arg_Yb_reg,			instrSTOS,			regAL,	ext2_Y	},	// 0xaa
	{ arg_Yv_reg,			instrSTOS,			regeAX,	ext2_Y	},	// 0xab
	{ arg_reg_Xb,			instrLODS,			regAL,	ext2__X	},	// 0xac
	{ arg_reg_Xv,			instrLODS,			regeAX,	ext2__X	},	// 0xad
	{ arg_Yb_reg,			instrSCAS,			regAL,	ext2_Y	},	// 0xae
	{ arg_Yv_reg,			instrSCAS,			regeAX,	ext2_Y	},	// 0xaf

	{ arg_REG8_Ib,			instrMOV,			0,0				},	// 0xb0	// 1
	{ arg_REG8_Ib,			instrMOV,			0,0				},	// 0xb1
	{ arg_REG8_Ib,			instrMOV,			0,0				},	// 0xb2
	{ arg_REG8_Ib,			instrMOV,			0,0				},	// 0xb3
	{ arg_REG8_Ib,			instrMOV,			0,0				},	// 0xb4
	{ arg_REG8_Ib,			instrMOV,			0,0				},	// 0xb5
	{ arg_REG8_Ib,			instrMOV,			0,0				},	// 0xb6
	{ arg_REG8_Ib,			instrMOV,			0,0				},	// 0xb7
	{ arg_eREG_Iv,			instrMOV,			0,0				},	// 0xb8
	{ arg_eREG_Iv,			instrMOV,			0,0				},	// 0xb9
	{ arg_eREG_Iv,			instrMOV,			0,0				},	// 0xba
	{ arg_eREG_Iv,			instrMOV,			0,0				},	// 0xbb
	{ arg_eREG_Iv,			instrMOV,			0,0				},	// 0xbc
	{ arg_eREG_Iv,			instrMOV,			0,0				},	// 0xbd
	{ arg_eREG_Iv,			instrMOV,			0,0				},	// 0xbe
	{ arg_eREG_Iv,			instrMOV,			0,0				},	// 0xbf

	{ arg_Eb_Ib | arg_MODRM,instrGROUP,			ext_GRP_2,0		},	// 0xc0	// 1
	{ arg_Ev_Ib | arg_MODRM,instrGROUP,			ext_GRP_2,0		},	// 0xc1
	{ arg_Iw,				instrRETN,			0,0				},	// 0xc2
	{ arg_NONE,				instrRETN,			0,0				},	// 0xc3
	{ arg_Gv_Mp,			instrLES,			0,0				},	// 0xc4
	{ arg_Gv_Mp,			instrLDS,			0,0				},	// 0xc5
	{ arg_Eb_Ib | arg_MODRM,instrGROUP,			ext_GRP_11,0	},	// 0xc6
	{ arg_Ev_Iv | arg_MODRM,instrGROUP,			ext_GRP_11,0	},	// 0xc7
	{ arg__Iw_Ib,			instrENTER,			0,0				},	// 0xc8
	{ arg_NONE,				instrLEAVE,			0,0				},	// 0xc9
	{ arg_Iw,				instrRETF,			0,0				},	// 0xca
	{ arg_NONE,				instrRETF,			0,0				},	// 0xcb
	{ arg_NONE,				instrINT3,			0,0				},	// 0xcc
	{ arg_Ib,				instrINT,			0,0				},	// 0xcd
	{ arg_NONE,				instrINTO,			0,0				},	// 0xce
	{ arg_NONE,				instrIRET,			0,0				},	// 0xcf

	{ arg_Eb_1 | arg_MODRM,	instrGROUP,			ext_GRP_2,0		},	// 0xd0	// 1
	{ arg_Ev_1 | arg_MODRM,	instrGROUP,			ext_GRP_2,0		},	// 0xd1
	{ arg_Eb_reg2 | arg_MODRM,instrGROUP,		ext_GRP_2,regCL	},	// 0xd2
	{ arg_Ev_reg2 | arg_MODRM,instrGROUP,		ext_GRP_2,regCL	},	// 0xd3
	{ arg_Ib,				instrAAM,			0,0				},	// 0xd4
	{ arg_Ib,				instrAAD,			0,0				},	// 0xd5
	{ 0,					instrRESERVED,		0,0				},	// 0xd6
	{ arg_NONE,				instrXLAT,			0,0				},	// 0xd7
	{ arg_MODRM,			instrESCAPE_COPROC,	ext_FP_D8_mod_mem,ext_FP_D8_mod_11 },	// 0xd8
	{ arg_MODRM,			instrESCAPE_COPROC,	ext_FP_D9_mod_mem,ext_FP_D9_mod_11 },	// 0xd9
	{ arg_MODRM,			instrESCAPE_COPROC,	ext_FP_DA_mod_mem,ext_FP_DA_mod_11 },	// 0xda
	{ arg_MODRM,			instrESCAPE_COPROC,	ext_FP_DB_mod_mem,ext_FP_DB_mod_11 },	// 0xdb
	{ arg_MODRM,			instrESCAPE_COPROC,	ext_FP_DC_mod_mem,ext_FP_DC_mod_11 },	// 0xdc
	{ arg_MODRM,			instrESCAPE_COPROC,	ext_FP_DD_mod_mem,ext_FP_DD_mod_11 },	// 0xdd
	{ arg_MODRM,			instrESCAPE_COPROC,	ext_FP_DE_mod_mem,ext_FP_DE_mod_11 },	// 0xde
	{ arg_MODRM,			instrESCAPE_COPROC,	ext_FP_DF_mod_mem,ext_FP_DF_mod_11 },	// 0xdf

	{ arg_Jb,				instrLOOPNZ,		0,0				},	// 0xe0	// 1
	{ arg_Jb,				instrLOOPZ,			0,0				},	// 0xe1
	{ arg_Jb,				instrLOOP,			0,0				},	// 0xe2
	{ arg_Jb,				instrJECXZ,			0,0				},	// 0xe3
	{ arg_reg_Ib,			instrIN,			regAL,0			},	// 0xe4
	{ arg_reg_Iv,			instrIN,			regeAX,0		},	// 0xe5
	{ arg_Ib_reg,			instrOUT,			regAL,0			},	// 0xe6
	{ arg_Iv_reg,			instrOUT,			regeAX,0		},	// 0xe7
	{ arg_Jv,				instrCALLN_REL,		0,0				},	// 0xe8
	{ arg_Jv,				instrJMPN_REL,		0,0				},	// 0xe9
	{ arg_Ap,				instrJMPF_ABS,		0,0				},	// 0xea
	{ arg_Jb,				instrJMPN_REL,		0,0				},	// 0xeb
	{ arg_reg_reg2,			instrIN,			regAL,	regDX	},	// 0xec
	{ arg_reg_reg2 |
		arg_OPSIZE_16_32,	instrIN,			regeAX,	regDX	},	// 0xed
	{ arg_reg_reg2,			instrOUT,			regDX,	regAL	},	// 0xee
	{ arg_reg_reg2 |
		arg_OPSIZE_16_32,	instrOUT,			regDX,	regeAX	},	// 0xef

	{ arg_PREFIX_LOCK,		instrPREFIX,		0,0				},	// 0xf0	// 1
	{ 0,					instrRESERVED,		0,0				},	// 0xf1
	{ arg_PREFIX_REP,		instrPREFIX,		instrgmSTRING_REPNE,0	},	// 0xf2
	{ arg_PREFIX_REP,		instrPREFIX,		instrgmSTRING_REPE,0	},	// 0xf3
	{ arg_NONE,				instrHLT,			0,0				},	// 0xf4
	{ arg_NONE,				instrCMC,			0,0				},	// 0xf5
	{ arg_Eb | arg_MODRM,	instrGROUP,			ext_GRP_3,0		},	// 0xf6
	{ arg_Ev | arg_MODRM,	instrGROUP,			ext_GRP_3,0		},	// 0xf7
	{ arg_NONE,				instrCLC,			0,0				},	// 0xf8
	{ arg_NONE,				instrSTC,			0,0				},	// 0xf9
	{ arg_NONE,				instrCLI,			0,0				},	// 0xfa
	{ arg_NONE,				instrSTI,			0,0				},	// 0xfb
	{ arg_NONE,				instrCLD,			0,0				},	// 0xfc
	{ arg_NONE,				instrSTD,			0,0				},	// 0xfd
	{ arg_MODRM,			instrGROUP,			ext_GRP_4,0		},	// 0xfe
	{ arg_MODRM,			instrGROUP,			ext_GRP_5,0		},	// 0xff
},

//////////////////////////////////////////////////////////////////////

{
	{ arg_MODRM,			instrGROUP,			ext_GRP_6,0		},	// 0x00	// 2
	{ arg_MODRM,			instrGROUP,			ext_GRP_7,0		},	// 0x01
	{ arg_Gv_Ew,			instrLAR,			0,0				},	// 0x02
	{ arg_Gv_Ew,			instrLSL,			0,0				},	// 0x03
	{ 0,					instrRESERVED,		0,0				},	// 0x04
	{ 0,					instrRESERVED,		0,0				},	// 0x05
	{ arg_NONE,				instrCLTS,			0,0				},	// 0x06
	{ 0,					instrRESERVED,		0,0				},	// 0x07
	{ arg_NONE,				instrINVD,			0,0				},	// 0x08
	{ arg_NONE,				instrWBINVD,		0,0				},	// 0x09
	{ 0,					instrRESERVED,		0,0				},	// 0x0a
	{ arg_NONE,				instrUD2,			0,0				},	// 0x0b
	{ 0,					instrRESERVED,		0,0				},	// 0x0c
	{ 0,					instrRESERVED,		0,0				},	// 0x0d
	{ 0,					instrRESERVED,		0,0				},	// 0x0e
	{ 0,					instrRESERVED,		0,0				},	// 0x0f

	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_10,ext2_SSE_PREFIX_NO_66_F3_F2	},	// 0x10	// 2
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_11,ext2_SSE_PREFIX_NO_66_F3_F2	},	// 0x11
	{ arg_MODRM,			instrPRE_MMX_SSE_MOD_MEM_11,ext_MMX_SSE_12_MOD_MEM,ext2_SSE_PREFIX_NO_66	},	// 0x12
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_13,ext2_SSE_PREFIX_NO_66	},	// 0x13
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_14,ext2_SSE_PREFIX_NO_66	},	// 0x14
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_15,ext2_SSE_PREFIX_NO_66	},	// 0x15
	{ arg_MODRM,			instrPRE_MMX_SSE_MOD_MEM_11,ext_MMX_SSE_16_MOD_MEM,ext2_SSE_PREFIX_NO_66	},	// 0x16
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_17,ext2_SSE_PREFIX_NO_66	},	// 0x17
	{ arg_MODRM,			instrGROUP_MOD_MEM,	ext_GRP_16_mod_mem,0	},	// 0x18
	{ 0,					instrRESERVED,		0,0				},	// 0x19
	{ 0,					instrRESERVED,		0,0				},	// 0x1a
	{ 0,					instrRESERVED,		0,0				},	// 0x1b
	{ 0,					instrRESERVED,		0,0				},	// 0x1c
	{ 0,					instrRESERVED,		0,0				},	// 0x1d
	{ 0,					instrRESERVED,		0,0				},	// 0x1e
	{ 0,					instrRESERVED,		0,0				},	// 0x1f

	{ arg_Rd_CDTd,			instrMOV,			reggCTRL,0		},	// 0x20	// 2
	{ arg_Rd_CDTd,			instrMOV,			reggDBG,0		},	// 0x21
	{ arg_CDTd_Rd,			instrMOV,			reggCTRL,0		},	// 0x22
	{ arg_CDTd_Rd,			instrMOV,			reggDBG,0		},	// 0x23
	{ arg_Rd_CDTd,			instrMOV,			reggTST,0		},	// 0x24
	{ 0,					instrRESERVED,		0,0				},	// 0x25
	{ arg_CDTd_Rd,			instrMOV,			reggTST,0		},	// 0x26
	{ 0,					instrRESERVED,		0,0				},	// 0x27
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_28,ext2_SSE_PREFIX_NO_66	},	// 0x28
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_29,ext2_SSE_PREFIX_NO_66	},	// 0x29
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_2A,ext2_SSE_PREFIX_NO_66_F3_F2	},	// 0x2A
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_2B,ext2_SSE_PREFIX_NO_66	},	// 0x2B
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_2C,ext2_SSE_PREFIX_NO_66_F3_F2	},	// 0x2C
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_2D,ext2_SSE_PREFIX_NO_66_F3_F2	},	// 0x2D
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_2E,ext2_SSE_PREFIX_NO_66	},	// 0x2E
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_2F,ext2_SSE_PREFIX_NO_66	},	// 0x2F

	{ arg_NONE,				instrWRMSR,			0,0				},	// 0x30	// 2
	{ arg_NONE,				instrRDTSC,			0,0				},	// 0x31
	{ arg_NONE,				instrRDMSR,			0,0				},	// 0x32
	{ arg_NONE,				instrRDPMC,			0,0				},	// 0x33
	{ arg_NONE,				instrSYSENTER,		0,0				},	// 0x34
	{ arg_NONE,				instrSYSEXIT,		0,0				},	// 0x35
	{ 0,					instrRESERVED,		0,0				},	// 0x36
	{ 0,					instrRESERVED,		0,0				},	// 0x37
	{ 0,					instrRESERVED,		0,0				},	// 0x38
	{ 0,					instrRESERVED,		0,0				},	// 0x39
	{ 0,					instrRESERVED,		0,0				},	// 0x3a
	{ 0,					instrRESERVED,		0,0				},	// 0x3b
	{ 0,					instrRESERVED,		0,0				},	// 0x3c
	{ 0,					instrRESERVED,		0,0				},	// 0x3d
	{ 0,					instrRESERVED,		0,0				},	// 0x3e
	{ 0,					instrRESERVED,		0,0				},	// 0x3f

	{ arg_Gv_Ev,			instrCMOVCC_0,		0,0				},	// 0x40	// 2
	{ arg_Gv_Ev,			instrCMOVCC_1,		0,0				},	// 0x41
	{ arg_Gv_Ev,			instrCMOVCC_2,		0,0				},	// 0x42
	{ arg_Gv_Ev,			instrCMOVCC_3,		0,0				},	// 0x43
	{ arg_Gv_Ev,			instrCMOVCC_4,		0,0				},	// 0x44
	{ arg_Gv_Ev,			instrCMOVCC_5,		0,0				},	// 0x45
	{ arg_Gv_Ev,			instrCMOVCC_6,		0,0				},	// 0x46
	{ arg_Gv_Ev,			instrCMOVCC_7,		0,0				},	// 0x47
	{ arg_Gv_Ev,			instrCMOVCC_8,		0,0				},	// 0x48
	{ arg_Gv_Ev,			instrCMOVCC_9,		0,0				},	// 0x49
	{ arg_Gv_Ev,			instrCMOVCC_A,		0,0				},	// 0x4a
	{ arg_Gv_Ev,			instrCMOVCC_B,		0,0				},	// 0x4b
	{ arg_Gv_Ev,			instrCMOVCC_C,		0,0				},	// 0x4c
	{ arg_Gv_Ev,			instrCMOVCC_D,		0,0				},	// 0x4d
	{ arg_Gv_Ev,			instrCMOVCC_E,		0,0				},	// 0x4e
	{ arg_Gv_Ev,			instrCMOVCC_F,		0,0				},	// 0x4f

	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_50,ext2_SSE_PREFIX_NO_66	},	// 0x50	// 2
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_51,ext2_SSE_PREFIX_NO_66_F3_F2	},	// 0x51
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_52,ext2_SSE_PREFIX_NO_F3	},	// 0x52
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_53,ext2_SSE_PREFIX_NO_F3	},	// 0x53
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_54,ext2_SSE_PREFIX_NO_66	},	// 0x54
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_55,ext2_SSE_PREFIX_NO_66	},	// 0x55
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_56,ext2_SSE_PREFIX_NO_66	},	// 0x56
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_57,ext2_SSE_PREFIX_NO_66	},	// 0x57
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_58,ext2_SSE_PREFIX_NO_66_F3_F2	},	// 0x58
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_59,ext2_SSE_PREFIX_NO_66_F3_F2	},	// 0x59
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_5A,ext2_SSE_PREFIX_NO_66_F3_F2	},	// 0x5A
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_5B,ext2_SSE_PREFIX_NO_66_F3	},	// 0x5B
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_5C,ext2_SSE_PREFIX_NO_66_F3_F2	},	// 0x5C
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_5D,ext2_SSE_PREFIX_NO_66_F3_F2	},	// 0x5D
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_5E,ext2_SSE_PREFIX_NO_66_F3_F2	},	// 0x5E
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_5F,ext2_SSE_PREFIX_NO_66_F3_F2	},	// 0x5F

	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_60,ext2_SSE_PREFIX_NO_66	},	// 0x60	// 2
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_61,ext2_SSE_PREFIX_NO_66	},	// 0x61
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_62,ext2_SSE_PREFIX_NO_66	},	// 0x62
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_63,ext2_SSE_PREFIX_NO_66	},	// 0x63
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_64,ext2_SSE_PREFIX_NO_66	},	// 0x64
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_65,ext2_SSE_PREFIX_NO_66	},	// 0x65
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_66,ext2_SSE_PREFIX_NO_66	},	// 0x66
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_67,ext2_SSE_PREFIX_NO_66	},	// 0x67
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_68,ext2_SSE_PREFIX_NO_66	},	// 0x68
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_69,ext2_SSE_PREFIX_NO_66	},	// 0x69
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_6A,ext2_SSE_PREFIX_NO_66	},	// 0x6A
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_6B,ext2_SSE_PREFIX_NO_66	},	// 0x6B
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_6C,ext2_SSE_PREFIX_66	},	// 0x6C
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_6D,ext2_SSE_PREFIX_66	},	// 0x6D
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_6E,ext2_SSE_PREFIX_NO_66	},	// 0x6E
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_6F,ext2_SSE_PREFIX_NO_66_F3	},	// 0x6F

	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_70,ext2_SSE_PREFIX_NO_66_F3_F2	},	// 0x70	// 2
	{ arg_MODRM,			instrGROUP_MOD_11,	0,	ext_GRP_12_mod_11	},	// 0x71
	{ arg_MODRM,			instrGROUP_MOD_11,	0,	ext_GRP_13_mod_11	},	// 0x72
	{ arg_MODRM,			instrGROUP_MOD_11,	0,	ext_GRP_14_mod_11	},	// 0x73
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_74,ext2_SSE_PREFIX_NO_66	},	// 0x74
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_75,ext2_SSE_PREFIX_NO_66	},	// 0x75
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_76,ext2_SSE_PREFIX_NO_66	},	// 0x76
	{ arg_NONE,				instrEMMS,			0,0				},	// 0x77
	{ 0,					instrRESERVED,		0,0				},	// 0x78
	{ 0,					instrRESERVED,		0,0				},	// 0x79
	{ 0,					instrRESERVED,		0,0				},	// 0x7a
	{ 0,					instrRESERVED,		0,0				},	// 0x7b
	{ 0,					instrRESERVED,		0,0				},	// 0x7c
	{ 0,					instrRESERVED,		0,0				},	// 0x7d
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_7E,ext2_SSE_PREFIX_NO_66_F3	},	// 0x7E
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_7F,ext2_SSE_PREFIX_NO_66_F3	},	// 0x7F

	{ arg_Jv,				instrJCC_0,			0,0				},	// 0x80	// 2
	{ arg_Jv,				instrJCC_1,			0,0				},	// 0x81
	{ arg_Jv,				instrJCC_2,			0,0				},	// 0x82
	{ arg_Jv,				instrJCC_3,			0,0				},	// 0x83
	{ arg_Jv,				instrJCC_4,			0,0				},	// 0x84
	{ arg_Jv,				instrJCC_5,			0,0				},	// 0x85
	{ arg_Jv,				instrJCC_6,			0,0				},	// 0x86
	{ arg_Jv,				instrJCC_7,			0,0				},	// 0x87
	{ arg_Jv,				instrJCC_8,			0,0				},	// 0x88
	{ arg_Jv,				instrJCC_9,			0,0				},	// 0x89
	{ arg_Jv,				instrJCC_A,			0,0				},	// 0x8a
	{ arg_Jv,				instrJCC_B,			0,0				},	// 0x8b
	{ arg_Jv,				instrJCC_C,			0,0				},	// 0x8c
	{ arg_Jv,				instrJCC_D,			0,0				},	// 0x8d
	{ arg_Jv,				instrJCC_E,			0,0				},	// 0x8e
	{ arg_Jv,				instrJCC_F,			0,0				},	// 0x8f

	{ arg_Eb,				instrSETCC_0,		0,0				},	// 0x90	// 2
	{ arg_Eb,				instrSETCC_1,		0,0				},	// 0x91
	{ arg_Eb,				instrSETCC_2,		0,0				},	// 0x92
	{ arg_Eb,				instrSETCC_3,		0,0				},	// 0x93
	{ arg_Eb,				instrSETCC_4,		0,0				},	// 0x94
	{ arg_Eb,				instrSETCC_5,		0,0				},	// 0x95
	{ arg_Eb,				instrSETCC_6,		0,0				},	// 0x96
	{ arg_Eb,				instrSETCC_7,		0,0				},	// 0x97
	{ arg_Eb,				instrSETCC_8,		0,0				},	// 0x98
	{ arg_Eb,				instrSETCC_9,		0,0				},	// 0x99
	{ arg_Eb,				instrSETCC_A,		0,0				},	// 0x9a
	{ arg_Eb,				instrSETCC_B,		0,0				},	// 0x9b
	{ arg_Eb,				instrSETCC_C,		0,0				},	// 0x9c
	{ arg_Eb,				instrSETCC_D,		0,0				},	// 0x9d
	{ arg_Eb,				instrSETCC_E,		0,0				},	// 0x9e
	{ arg_Eb,				instrSETCC_F,		0,0				},	// 0x9f

	{ arg__reg,				instrPUSH,			regFS,0			},	// 0xa0	// 2
	{ arg_reg,				instrPOP,			regFS,0			},	// 0xa1
	{ arg_NONE,				instrCPUID,			0,0				},	// 0xa2
	{ arg_Ev_Gv,			instrBT,			0,0				},	// 0xa3
	{ arg_Ev_Gv_Ib,			instrSHLD,			0,0				},	// 0xa4
	{ arg_Ev_Gv_reg,		instrSHLD,			regCL,0			},	// 0xa5
	{ 0,					instrRESERVED,		0,0				},	// 0xa6
	{ 0,					instrRESERVED,		0,0				},	// 0xa7
	{ arg__reg,				instrPUSH,			regGS,0			},	// 0xa8
	{ arg_reg,				instrPOP,			regGS,0			},	// 0xa9
	{ arg_NONE,				instrRSM,			0,0				},	// 0xaa
	{ arg_Ev_Gv,			instrBTS,			0,0				},	// 0xab
	{ arg_Ev_Gv_Ib,			instrSHRD,			0,0				},	// 0xac
	{ arg_Ev_Gv_reg,		instrSHRD,			regCL,0			},	// 0xad
	{ arg_MODRM,			instrGROUP_MOD_MEM_11,	ext_GRP_15_mod_mem,	ext_GRP_15_mod_11	},	// 0xae
	{ arg_Gv_Ev,			instrIMUL,			0,0				},	// 0xaf

	{ arg_Eb_Gb,			instrCMPXCHG,		0,0				},	// 0xb0	// 2
	{ arg_Ev_Gv,			instrCMPXCHG,		0,0				},	// 0xb1
	{ arg_Gv_Mp,			instrLSS,			0,0				},	// 0xb2
	{ arg_Ev_Gv,			instrBTR,			0,0				},	// 0xb3
	{ arg_Gv_Mp,			instrLFS,			0,0				},	// 0xb4
	{ arg_Gv_Mp,			instrLGS,			0,0				},	// 0xb5
	{ arg_Gv_Eb,			instrMOVZX,			0,0				},	// 0xb6
	{ arg_Gv_Ew,			instrMOVZX,			0,0				},	// 0xb7
	{ 0,					instrRESERVED,		0,0				},	// 0xb8
	{ arg_MODRM,			instrGROUP,			ext_GRP_10,0	},	// 0xb9
	{ arg_Ev_Ib | arg_MODRM,instrGROUP,			ext_GRP_8,0		},	// 0xba
	{ arg_Ev_Gv,			instrBTC,			0,0				},	// 0xbb
	{ arg_Gv_Ev,			instrBSF,			0,0				},	// 0xbc
	{ arg_Gv_Ev,			instrBSR,			0,0				},	// 0xbd
	{ arg_Gv_Eb,			instrMOVSX,			0,0				},	// 0xbe
	{ arg_Gv_Ew,			instrMOVSX,			0,0				},	// 0xbf

	{ arg_Eb_Gb,			instrXADD,			0,0				},	// 0xc0	// 2
	{ arg_Ev_Gv,			instrXADD,			0,0				},	// 0xc1
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_C2,ext2_SSE_PREFIX_NO_66_F3_F2	},	// 0xC2
	{ arg_Md_Gd,			instrMOVNTI,		0,0				},	// 0xC3
//	{ arg_MODRM,			instrPRE_MMX_SSE_MOD_MEM_11,ext_MMX_SSE_C4_MOD_MEM,ext2_SSE_PREFIX_NO_66	},	// 0xC4
	{ arg_MODRM,			instrPRE_MMX_SSE,	ext_MMX_SSE_C4,ext2_SSE_PREFIX_NO_66	},	// 0xC4
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_C5,ext2_SSE_PREFIX_NO_66	},	// 0xC5
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_C6,ext2_SSE_PREFIX_NO_66	},	// 0xC6
	{ arg_MODRM,			instrGROUP_MOD_MEM,	ext_GRP_9_mod_mem,0 },	// 0xc7
	{ arg_EREG,				instrBSWAP,			0,0				},	// 0xc8
	{ arg_EREG,				instrBSWAP,			0,0				},	// 0xc9
	{ arg_EREG,				instrBSWAP,			0,0				},	// 0xca
	{ arg_EREG,				instrBSWAP,			0,0				},	// 0xcb
	{ arg_EREG,				instrBSWAP,			0,0				},	// 0xcc
	{ arg_EREG,				instrBSWAP,			0,0				},	// 0xcd
	{ arg_EREG,				instrBSWAP,			0,0				},	// 0xce
	{ arg_EREG,				instrBSWAP,			0,0				},	// 0xcf

	{ 0,					instrRESERVED,		0,0				},	// 0xd0	// 2
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_D1,ext2_SSE_PREFIX_NO_66	},	// 0xD1
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_D2,ext2_SSE_PREFIX_NO_66	},	// 0xD2
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_D3,ext2_SSE_PREFIX_NO_66	},	// 0xD3
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_D4,ext2_SSE_PREFIX_NO_66	},	// 0xD4
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_D5,ext2_SSE_PREFIX_NO_66	},	// 0xD5
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_D6,ext2_SSE_PREFIX_66_F3_F2	},	// 0xD6
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_D7,ext2_SSE_PREFIX_NO_66	},	// 0xD7
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_D8,ext2_SSE_PREFIX_NO_66	},	// 0xD8
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_D9,ext2_SSE_PREFIX_NO_66	},	// 0xD9
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_DA,ext2_SSE_PREFIX_NO_66	},	// 0xDA
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_DB,ext2_SSE_PREFIX_NO_66	},	// 0xDB
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_DC,ext2_SSE_PREFIX_NO_66	},	// 0xDC
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_DD,ext2_SSE_PREFIX_NO_66	},	// 0xDD
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_DE,ext2_SSE_PREFIX_NO_66	},	// 0xDE
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_DF,ext2_SSE_PREFIX_NO_66	},	// 0xDF

	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_E0,ext2_SSE_PREFIX_NO_66	},	// 0xE0	// 2
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_E1,ext2_SSE_PREFIX_NO_66	},	// 0xE1
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_E2,ext2_SSE_PREFIX_NO_66	},	// 0xE2
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_E3,ext2_SSE_PREFIX_NO_66	},	// 0xE3
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_E4,ext2_SSE_PREFIX_NO_66	},	// 0xE4
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_E5,ext2_SSE_PREFIX_NO_66	},	// 0xE5
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_E6,ext2_SSE_PREFIX_66_F3_F2	},	// 0xE6
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_E7,ext2_SSE_PREFIX_NO_66	},	// 0xE7
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_E8,ext2_SSE_PREFIX_NO_66	},	// 0xE8
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_E9,ext2_SSE_PREFIX_NO_66	},	// 0xE9
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_EA,ext2_SSE_PREFIX_NO_66	},	// 0xEA
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_EB,ext2_SSE_PREFIX_NO_66	},	// 0xEB
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_EC,ext2_SSE_PREFIX_NO_66	},	// 0xEC
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_ED,ext2_SSE_PREFIX_NO_66	},	// 0xED
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_EE,ext2_SSE_PREFIX_NO_66	},	// 0xEE
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_EF,ext2_SSE_PREFIX_NO_66	},	// 0xEF

	{ 0,					instrRESERVED,		0,0				},	// 0xf0	// 2
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_F1,ext2_SSE_PREFIX_NO_66	},	// 0xF1
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_F2,ext2_SSE_PREFIX_NO_66	},	// 0xF2
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_F3,ext2_SSE_PREFIX_NO_66	},	// 0xF3
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_F4,ext2_SSE_PREFIX_NO_66	},	// 0xF4
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_F5,ext2_SSE_PREFIX_NO_66	},	// 0xF5
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_F6,ext2_SSE_PREFIX_NO_66	},	// 0xF6
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_F7,ext2_SSE_PREFIX_NO_66	},	// 0xF7
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_F8,ext2_SSE_PREFIX_NO_66	},	// 0xF8
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_F9,ext2_SSE_PREFIX_NO_66	},	// 0xF9
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_FA,ext2_SSE_PREFIX_NO_66	},	// 0xFA
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_FB,ext2_SSE_PREFIX_NO_66	},	// 0xFB
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_FC,ext2_SSE_PREFIX_NO_66	},	// 0xFC
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_FD,ext2_SSE_PREFIX_NO_66	},	// 0xFD
	{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_FE,ext2_SSE_PREFIX_NO_66	},	// 0xFE
	{ 0,					instrRESERVED,		0,0				},	// 0xff
}
};

//////////////////////////////////////////////////////////////////////

const OpcodeInfo infoGroups[][8] =
{
	{ // ext_GRP_1
		{ arg_NONE,				instrADD,			0,0				},	// 0x00 // group 1
		{ arg_NONE,				instrOR,			0,0				},	// 0x01
		{ arg_NONE,				instrADC,			0,0				},	// 0x02
		{ arg_NONE,				instrSBB,			0,0				},	// 0x03
		{ arg_NONE,				instrAND,			0,0				},	// 0x04
		{ arg_NONE,				instrSUB,			0,0				},	// 0x05
		{ arg_NONE,				instrXOR,			0,0				},	// 0x06
		{ arg_NONE,				instrCMP,			0,0				},	// 0x07
	},
	{ // ext_GRP_2
		{ arg_NONE,				instrROL,			0,0				},	// 0x00 // group 2
		{ arg_NONE,				instrROR,			0,0				},	// 0x01
		{ arg_NONE,				instrRCL,			0,0				},	// 0x02
		{ arg_NONE,				instrRCR,			0,0				},	// 0x03
		{ arg_NONE,				instrSHL,			0,0				},	// 0x04
		{ arg_NONE,				instrSHR,			0,0				},	// 0x05
		{ 0,					instrRESERVED,		0,0				},	// 0x06
		{ arg_NONE,				instrSAR,			0,0				},	// 0x07
	},
	{ // ext_GRP_3
		{ arg__I,				instrTEST,			0,0				},	// 0x00 // group 3
		{ 0,					instrRESERVED,		0,0				},	// 0x01
		{ arg_NONE,				instrNOT,			0,0				},	// 0x02
		{ arg_NONE,				instrNEG,			0,0				},	// 0x03
		{ arg__reg,				instrMUL,			regALeAX,0		},	// 0x04
		{ arg__reg,				instrIMUL,			regALeAX,0		},	// 0x05
		{ arg__reg,				instrDIV,			regALeAX,0		},	// 0x06
		{ arg__reg,				instrIDIV,			regALeAX,0		},	// 0x07
	},
	{ // ext_GRP_4
		{ arg_Eb,				instrINC,			0,0				},	// 0x00 // group 4
		{ arg_Eb,				instrDEC,			0,0				},	// 0x01
		{ 0,					instrRESERVED,		0,0				},	// 0x02
		{ 0,					instrRESERVED,		0,0				},	// 0x03
		{ 0,					instrRESERVED,		0,0				},	// 0x04
		{ 0,					instrRESERVED,		0,0				},	// 0x05
		{ 0,					instrRESERVED,		0,0				},	// 0x06
		{ 0,					instrRESERVED,		0,0				},	// 0x07
	},
	{ // ext_GRP_5
		{ arg_Ev,				instrINC,			0,0				},	// 0x00 // group 5
		{ arg_Ev,				instrDEC,			0,0				},	// 0x01
		{ arg_Ev,				instrCALLN_ABS,		0,0				},	// 0x02
		{ arg_Mp,				instrCALLF_ABS,		0,0				},	// 0x03
		{ arg_Ev,				instrJMPN_ABS,		0,0				},	// 0x04
		{ arg_Mp,				instrJMPF_ABS,		0,0				},	// 0x05
		{ arg__Ev,				instrPUSH,			0,0				},	// 0x06
		{ 0,					instrRESERVED,		0,0				},	// 0x07
	},
	{ // ext_GRP_6
		{ arg_Ev,				instrSLDT,			0,0				},	// 0x00 // group 6
		{ arg_Ev,				instrSTR,			0,0				},	// 0x01
		{ arg_Ew,				instrLLDT,			0,0				},	// 0x02
		{ arg_Ew,				instrLTR,			0,0				},	// 0x03
		{ arg_Ew,				instrVERR,			0,0				},	// 0x04
		{ arg_Ew,				instrVERW,			0,0				},	// 0x05
		{ 0,					instrRESERVED,		0,0				},	// 0x06
		{ 0,					instrRESERVED,		0,0				},	// 0x07
	},
	{ // ext_GRP_7
		{ arg_Ms,				instrSGDT,			0,6				},	// 0x00 // group 7
		{ arg_Ms,				instrSIDT,			0,6				},	// 0x01
		{ arg__Ms,				instrLGDT,			0,6				},	// 0x02
		{ arg__Ms,				instrLIDT,			0,6				},	// 0x03
		{ arg_Ew,				instrSMSW,			0,0				},	// 0x04
		{ 0,					instrRESERVED,		0,0				},	// 0x05
		{ arg_Ew,				instrLMSW,			0,0				},	// 0x06
		{ arg_Mb,				instrINVLPG,		0,0				},	// 0x07
	},
	{ // ext_GRP_8
		{ 0,					instrRESERVED,		0,0				},	// 0x00 // group 8
		{ 0,					instrRESERVED,		0,0				},	// 0x01
		{ 0,					instrRESERVED,		0,0				},	// 0x02
		{ 0,					instrRESERVED,		0,0				},	// 0x03
		{ arg_NONE,				instrBT,			0,0				},	// 0x04
		{ arg_NONE,				instrBTS,			0,0				},	// 0x05
		{ arg_NONE,				instrBTR,			0,0				},	// 0x06
		{ arg_NONE,				instrBTC,			0,0				},	// 0x07
	},
	{ // ext_GRP_9_mod_mem
		{ 0,					instrRESERVED,		0,0				},	// 0x00 // group 9 mod = mem
		{ arg_Mq,				instrCMPXCH8B,		0,0				},	// 0x01
		{ 0,					instrRESERVED,		0,0				},	// 0x02
		{ 0,					instrRESERVED,		0,0				},	// 0x03
		{ 0,					instrRESERVED,		0,0				},	// 0x04
		{ 0,					instrRESERVED,		0,0				},	// 0x05
		{ 0,					instrRESERVED,		0,0				},	// 0x06
		{ 0,					instrRESERVED,		0,0				},	// 0x07
	},
	{ // ext_GRP_10
		{ 0,					instrRESERVED,		0,0				},	// 0x00 // group 10
		{ 0,					instrRESERVED,		0,0				},	// 0x01
		{ 0,					instrRESERVED,		0,0				},	// 0x02
		{ 0,					instrRESERVED,		0,0				},	// 0x03
		{ 0,					instrRESERVED,		0,0				},	// 0x04
		{ 0,					instrRESERVED,		0,0				},	// 0x05
		{ 0,					instrRESERVED,		0,0				},	// 0x06
		{ 0,					instrRESERVED,		0,0				},	// 0x07
	},
	{ // ext_GRP_11
		{ arg_NONE,				instrMOV,			0,0				},	// 0x00 // group 11
		{ 0,					instrRESERVED,		0,0				},	// 0x01
		{ 0,					instrRESERVED,		0,0				},	// 0x02
		{ 0,					instrRESERVED,		0,0				},	// 0x03
		{ 0,					instrRESERVED,		0,0				},	// 0x04
		{ 0,					instrRESERVED,		0,0				},	// 0x05
		{ 0,					instrRESERVED,		0,0				},	// 0x06
		{ 0,					instrRESERVED,		0,0				},	// 0x07
	},
	{ // ext_GRP_12_mod_11
		{ 0,					instrRESERVED,		0,0				},	// 0x00 // group 12 mod = 11
		{ 0,					instrRESERVED,		0,0				},	// 0x01
		{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_71_2,ext2_SSE_PREFIX_NO_66	},	// 0x71_2
		{ 0,					instrRESERVED,		0,0				},	// 0x03
		{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_71_4,ext2_SSE_PREFIX_NO_66	},	// 0x71_4
		{ 0,					instrRESERVED,		0,0				},	// 0x05
		{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_71_6,ext2_SSE_PREFIX_66	},	// 0x71_6
		{ 0,					instrRESERVED,		0,0				},	// 0x07
	},
	{ // ext_GRP_13_mod_11
		{ 0,					instrRESERVED,		0,0				},	// 0x00 // group 13 mod = 11
		{ 0,					instrRESERVED,		0,0				},	// 0x01
		{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_72_2,ext2_SSE_PREFIX_NO_66	},	// 0x72_2
		{ 0,					instrRESERVED,		0,0				},	// 0x03
		{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_72_4,ext2_SSE_PREFIX_NO_66	},	// 0x72_4
		{ 0,					instrRESERVED,		0,0				},	// 0x05
		{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_72_6,ext2_SSE_PREFIX_66	},	// 0x72_6
		{ 0,					instrRESERVED,		0,0				},	// 0x07
	},
	{ // ext_GRP_14_mod_11
		{ 0,					instrRESERVED,		0,0				},	// 0x00 // group 14 mod = 11
		{ 0,					instrRESERVED,		0,0				},	// 0x01
		{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_73_2,ext2_SSE_PREFIX_NO_66	},	// 0x73_2
		{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_73_3,ext2_SSE_PREFIX_66	},	// 0x73_3
		{ 0,					instrRESERVED,		0,0				},	// 0x04
		{ 0,					instrRESERVED,		0,0				},	// 0x05
		{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_73_6,ext2_SSE_PREFIX_66	},	// 0x73_6
		{ arg_NONE,				instrPRE_MMX_SSE,	ext_MMX_SSE_73_7,ext2_SSE_PREFIX_66	},	// 0x73_7
	},
	{ // ext_GRP_15_mod_mem
		{ arg_Ms,				instrFXSAVE,		0,ext_M512		},	// 0x00 // group 15 mod = mem
		{ arg__Ms,				instrFXRSTOR,		0,ext_M512		},	// 0x01
		{ arg__Md,				instrLDMXCSR,		0,0				},	// 0x02
		{ arg_Md,				instrSTMXCSR,		0,0				},	// 0x03
		{ 0,					instrRESERVED,		0,0				},	// 0x04
		{ 0,					instrRESERVED,		0,0				},	// 0x05
		{ 0,					instrRESERVED,		0,0				},	// 0x06
		{ arg__Mb,				instrCLFLUSH,		0,0				},	// 0x07
	},
	{ // ext_GRP_16_mod_mem
		{ arg__Mb,				instrPREFETCHNTA,	0,0				},	// 0x00 // group 16 mod = mem
		{ arg__Mb,				instrPREFETCHT0,	0,0				},	// 0x01
		{ arg__Mb,				instrPREFETCHT1,	0,0				},	// 0x02
		{ arg__Mb,				instrPREFETCHT2,	0,0				},	// 0x03
		{ 0,					instrRESERVED,		0,0				},	// 0x04
		{ 0,					instrRESERVED,		0,0				},	// 0x05
		{ 0,					instrRESERVED,		0,0				},	// 0x06
		{ 0,					instrRESERVED,		0,0				},	// 0x07
	},
	{ // ext_GRP_15_mod_11
		{ 0,					instrRESERVED,		0,0				},	// 0x00 // group 15 mod = 11
		{ 0,					instrRESERVED,		0,0				},	// 0x01
		{ 0,					instrRESERVED,		0,0				},	// 0x02
		{ 0,					instrRESERVED,		0,0				},	// 0x03
		{ 0,					instrRESERVED,		0,0				},	// 0x04
		{ arg_NONE,				instrMFENCE,		0,0				},	// 0x05
		{ arg_NONE,				instrLFENCE,		0,0				},	// 0x06
		{ arg_NONE,				instrSFENCE,		0,0				},	// 0x07
	},

	{ // ext_FP_D8_mod_mem
		{ arg_reg_Md,			instrFADD, 			regST0,0		},	// 0x00 // fp, d8, mod = mem
		{ arg_reg_Md,			instrFMUL, 			regST0,0		},	// 0x01
		{ arg_reg_Md,			instrFCOM, 			regST0,0		},	// 0x02
		{ arg_reg_Md,			instrFCOMP,			regST0,0		},	// 0x03
		{ arg_reg_Md,			instrFSUB, 			regST0,0		},	// 0x04
		{ arg_reg_Md,			instrFSUBR,			regST0,0		},	// 0x05
		{ arg_reg_Md,			instrFDIV, 			regST0,0		},	// 0x06
		{ arg_reg_Md,			instrFDIVR,			regST0,0		},	// 0x07
	},
	{ // ext_FP_D9_mod_mem
		{ arg_reg_Md,			instrFLD,			regST0,0		},	// 0x00 // fp, d9, mod = mem
		{ 0,        			instrRESERVED,		regST0,0		},	// 0x01
		{ arg_Md_reg,			instrFST,			regST0,0     	},	// 0x02
		{ arg_Md_reg,			instrFSTP,			regST0,0		},	// 0x03
		{ arg__Mvs,				instrFLDENV,		14,28			},	// 0x04
		{ arg__Mw,				instrFLDCW,			0,0				},	// 0x05
		{ arg_Mvs,				instrFSTENV,		14,28			},	// 0x06
		{ arg_Mw,				instrFSTCW,			0,0				},	// 0x07
	},
	{ // ext_FP_DA_mod_mem
		{ arg_reg_Md,			instrFIADD, 		regST0,0		},	// 0x00 // fp, da, mod = mem
		{ arg_reg_Md,			instrFIMUL, 		regST0,0		},	// 0x01
		{ arg_reg_Md,			instrFICOM, 		regST0,0		},	// 0x02
		{ arg_reg_Md,			instrFICOMP,		regST0,0		},	// 0x03
		{ arg_reg_Md,			instrFISUB, 		regST0,0		},	// 0x04
		{ arg_reg_Md,			instrFISUBR,		regST0,0		},	// 0x05
		{ arg_reg_Md,			instrFIDIV, 		regST0,0		},	// 0x06
		{ arg_reg_Md,			instrFIDIVR,		regST0,0		},	// 0x07
	},
	{ // ext_FP_DB_mod_mem
		{ arg_reg_Md,			instrFILD,			regST0,0		},	// 0x00 // fp, db, mod = mem
		{ 0,					instrRESERVED,		0,0				},	// 0x01
		{ arg_Md_reg,			instrFIST,			regST0,0		},	// 0x02
		{ arg_Md_reg,			instrFISTP,			regST0,0		},	// 0x03
		{ 0,					instrRESERVED,		0,0				},	// 0x04
		{ arg_reg_Ms,			instrFLD,			regST0,10		},	// 0x05
		{ 0,					instrRESERVED,		0,0				},	// 0x06
		{ arg_Ms_reg,			instrFSTP,			regST0,10		},	// 0x07
	},
	{ // ext_FP_DC_mod_mem
		{ arg_reg_Mq,			instrFADD, 			regST0,0		},	// 0x00 // fp, dc, mod = mem
		{ arg_reg_Mq,			instrFMUL, 			regST0,0		},	// 0x01
		{ arg_reg_Mq,			instrFCOM, 			regST0,0		},	// 0x02
		{ arg_reg_Mq,			instrFCOMP,			regST0,0		},	// 0x03
		{ arg_reg_Mq,			instrFSUB, 			regST0,0		},	// 0x04
		{ arg_reg_Mq,			instrFSUBR,			regST0,0		},	// 0x05
		{ arg_reg_Mq,			instrFDIV, 			regST0,0		},	// 0x06
		{ arg_reg_Mq,			instrFDIVR,			regST0,0		},	// 0x07
	},
	{ // ext_FP_DD_mod_mem
		{ arg_reg_Mq,			instrFLD,			regST0,0		},	// 0x00 // fp, dd, mod = mem
		{ 0,        			instrRESERVED,		regST0,0		},	// 0x01
		{ arg_Mq_reg,			instrFST,			regST0,0     	},	// 0x02
		{ arg_Mq_reg,			instrFSTP,			regST0,0		},	// 0x03
		{ arg__Mvs,				instrFRSTOR,		94,108			},	// 0x04
		{ 0,					instrRESERVED,		0,0				},	// 0x05
		{ arg_Mvs,				instrFSAVE,			94,108			},	// 0x06
		{ arg_Mw,				instrFSTSW,			0,0				},	// 0x07
	},
	{ // ext_FP_DE_mod_mem
		{ arg_reg_Mw,			instrFIADD, 		regST0,0		},	// 0x00 // fp, de, mod = mem
		{ arg_reg_Mw,			instrFIMUL, 		regST0,0		},	// 0x01
		{ arg_reg_Mw,			instrFICOM, 		regST0,0		},	// 0x02
		{ arg_reg_Mw,			instrFICOMP,		regST0,0		},	// 0x03
		{ arg_reg_Mw,			instrFISUB, 		regST0,0		},	// 0x04
		{ arg_reg_Mw,			instrFISUBR,		regST0,0		},	// 0x05
		{ arg_reg_Mw,			instrFIDIV, 		regST0,0		},	// 0x06
		{ arg_reg_Mw,			instrFIDIVR,		regST0,0		},	// 0x07
	},
	{ // ext_FP_DF_mod_mem
		{ arg_reg_Mw,			instrFILD,			regST0,0		},	// 0x00 // fp, df, mod = mem
		{ 0,					instrRESERVED,		0,0				},	// 0x01
		{ arg_Mw_reg,			instrFIST,			regST0,0		},	// 0x02
		{ arg_Mw_reg,			instrFISTP,			regST0,0		},	// 0x03
		{ arg_reg_Ms,			instrFBLD,			regST0,10		},	// 0x04
		{ arg_reg_Mq,			instrFILD,			regST0,0		},	// 0x05
		{ arg_Ms_reg,			instrFBSTP,			regST0,10		},	// 0x06
		{ arg_Mq_reg,			instrFISTP,			regST0,0		},	// 0x07
	},

	{ // ext_FP_D8_mod_11
		{ arg_reg_STREG,		instrFADD, 			regST0,0		},	// 0x00 // fp, d8, mod = 11
		{ arg_reg_STREG,		instrFMUL, 			regST0,0		},	// 0x01
		{ arg_reg_STREG,		instrFCOM, 			regST0,0		},	// 0x02
		{ arg_reg_STREG,		instrFCOMP,			regST0,0		},	// 0x03
		{ arg_reg_STREG,		instrFSUB, 			regST0,0		},	// 0x04
		{ arg_reg_STREG,		instrFSUBR,			regST0,0		},	// 0x05
		{ arg_reg_STREG,		instrFDIV, 			regST0,0		},	// 0x06
		{ arg_reg_STREG,		instrFDIVR,			regST0,0		},	// 0x07
	},
	{ // ext_FP_D9_mod_11
		{ arg_reg_STREG,		instrFLD,			regST0,0		},	// 0x00 // fp, d9, mod = 11
		{ arg_reg_STREG,		instrFXCH,			regST0,0		},	// 0x01
		{ arg_NONE,				instrSUBGRP_COPROC,	ext_FP_D9_D0,0	},	// 0x02
		{ 0,					instrRESERVED,		0,0				},	// 0x03
		{ arg_NONE,				instrSUBGRP_COPROC,	ext_FP_D9_E0,0	},	// 0x04
		{ arg_NONE,				instrSUBGRP_COPROC, ext_FP_D9_E8,0	},	// 0x05
		{ arg_NONE,				instrSUBGRP_COPROC,	ext_FP_D9_F0,0	},	// 0x06
		{ arg_NONE,				instrSUBGRP_COPROC,	ext_FP_D9_F8,0	},	// 0x07
	},
	{ // ext_FP_DA_mod_11
		{ arg_reg_STREG,		instrFCMOVB,		regST0,0		},	// 0x00 // fp, da, mod = 11
		{ arg_reg_STREG,		instrFCMOVE,		regST0,0		},	// 0x01
		{ arg_reg_STREG,		instrFCMOVBE,		regST0,0		},	// 0x02
		{ arg_reg_STREG,		instrFCMOVU,		regST0,0		},	// 0x03
		{ 0,					instrRESERVED,		0,0				},	// 0x04
		{ arg_NONE,				instrSUBGRP_COPROC,	ext_FP_DA_E8,0	},	// 0x05
		{ 0,					instrRESERVED,		0,0				},	// 0x06
		{ 0,					instrRESERVED,		0,0				},	// 0x07
	},
	{ // ext_FP_DB_mod_11
		{ arg_reg_STREG,		instrFCMOVNB,		regST0,0		},	// 0x00 // fp, db, mod = 11
		{ arg_reg_STREG,		instrFCMOVNE,		regST0,0		},	// 0x01
		{ arg_reg_STREG,		instrFCMOVNBE,		regST0,0		},	// 0x02
		{ arg_reg_STREG,		instrFCMOVNU,		regST0,0		},	// 0x03
		{ arg_NONE,				instrSUBGRP_COPROC,	ext_FP_DB_E0,0	},	// 0x04
		{ arg_reg_STREG,		instrFUCOMI,		regST0,0		},	// 0x05
		{ arg_reg_STREG,		instrFCOMI,			regST0,0		},	// 0x06
		{ 0,					instrRESERVED,		0,0				},	// 0x07
	},
	{ // ext_FP_DC_mod_11
		{ arg_STREG_reg,		instrFADD, 			regST0,0		},	// 0x00 // fp, dc, mod = 11
		{ arg_STREG_reg,		instrFMUL, 			regST0,0		},	// 0x01
		{ 0,					instrRESERVED, 		0,0				},	// 0x02
		{ 0,					instrRESERVED,		0,0				},	// 0x03
		{ arg_STREG_reg,		instrFSUB, 			regST0,0		},	// 0x04
		{ arg_STREG_reg,		instrFSUBR,			regST0,0		},	// 0x05
		{ arg_STREG_reg,		instrFDIV,			regST0,0		},	// 0x06
		{ arg_STREG_reg,		instrFDIVR,			regST0,0		},	// 0x07
	},
	{ // ext_FP_DD_mod_11
		{ arg__STREG,			instrFFREE,			0,0				},	// 0x00 // fp, dd, mod = 11
		{ 0,					instrRESERVED,		0,0				},	// 0x01
		{ arg_STREG_reg,		instrFST,			regST0,0		},	// 0x02
		{ arg_STREG_reg,		instrFSTP,			regST0,0		},	// 0x03
		{ arg_reg_STREG,		instrFUCOM,			regST0,0		},	// 0x04
		{ arg_reg_STREG,		instrFUCOMP,		regST0,0		},	// 0x05
		{ 0,					instrRESERVED,		0,0				},	// 0x06
		{ 0,					instrRESERVED,		0,0				},	// 0x07
	},
	{ // ext_FP_DE_mod_11
		{ arg_STREG_reg,		instrFADDP, 		regST0,0		},	// 0x00 // fp, de, mod = 11
		{ arg_STREG_reg,		instrFMULP, 		regST0,0		},	// 0x01
		{ 0,					instrRESERVED, 		0,0				},	// 0x02
		{ arg_NONE,				instrSUBGRP_COPROC,	ext_FP_DE_D8,0	},	// 0x03
		{ arg_STREG_reg,		instrFSUBP, 		regST0,0		},	// 0x04
		{ arg_STREG_reg,		instrFSUBRP,		regST0,0		},	// 0x05
		{ arg_STREG_reg,		instrFDIVP, 		regST0,0		},	// 0x06
		{ arg_STREG_reg,		instrFDIVRP,		regST0,0		},	// 0x07
	},
	{ // ext_FP_DF_mod_11
		{ 0,					instrRESERVED,		0,0				},	// 0x00 // fp, df, mod = 11
		{ 0,					instrRESERVED,		0,0				},	// 0x01
		{ 0,					instrRESERVED,		0,0				},	// 0x02
		{ 0,					instrRESERVED,		0,0				},	// 0x03
		{ arg_NONE,				instrSUBGRP_COPROC,	ext_FP_DF_E0,0	},	// 0x04
		{ arg_reg_STREG,		instrFUCOMIP,		regST0,0		},	// 0x05
		{ arg_reg_STREG,		instrFCOMIP,		regST0,0		},	// 0x06
		{ 0,					instrRESERVED,		0,0				},	// 0x07
	},

	{ // ext_FP_D9_D0
		{ arg_NONE,				instrFNOP,			0,0				},	// 0x00
		{ 0,					instrRESERVED,		0,0				},	// 0x01
		{ 0,					instrRESERVED,		0,0				},	// 0x02
		{ 0,					instrRESERVED,		0,0				},	// 0x03
		{ 0,					instrRESERVED,		0,0				},	// 0x04
		{ 0,					instrRESERVED,		0,0				},	// 0x05
		{ 0,					instrRESERVED,		0,0				},	// 0x07
		{ 0,					instrRESERVED,		0,0				},	// 0x08
	},
	{ // ext_FP_D9_E0
		{ arg_reg,				instrFCHS,			regST0,0		},	// 0x00
		{ arg_reg,				instrFABS,			regST0,0		},	// 0x01
		{ 0,					instrRESERVED,		0,0				},	// 0x02
		{ 0,					instrRESERVED,		0,0				},	// 0x03
		{ arg_reg,				instrFTST,			regST0,0		},	// 0x04
		{ arg_reg,				instrFXAM,			regST0,0		},	// 0x05
		{ 0,					instrRESERVED,		0,0				},	// 0x07
		{ 0,					instrRESERVED,		0,0				},	// 0x08
	},
	{ // ext_FP_D9_E8
		{ arg_reg,				instrFLD1,			regST0,0		},	// 0x00
		{ arg_reg,				instrFLDL2T,		regST0,0		},	// 0x01
		{ arg_reg,				instrFLDL2E,		regST0,0		},	// 0x02
		{ arg_reg,				instrFLDPI,			regST0,0		},	// 0x03
		{ arg_reg,				instrFLDLG2,		regST0,0		},	// 0x04
		{ arg_reg,				instrFLDLN2,		regST0,0		},	// 0x05
		{ arg_reg,				instrFLDZ,			regST0,0		},	// 0x07
		{ 0,					instrRESERVED,		0,0				},	// 0x08
	},
	{ // ext_FP_D9_F0
		{ arg_reg,				instrF2XM1,			regST0,0		},	// 0x00
		{ arg_reg_reg2,			instrFYL2X,			regST0,regST1	},	// 0x01
		{ arg_reg,				instrFPTAN,			regST0,0		},	// 0x02
		{ arg_reg_reg2,			instrFPATAN,		regST0,regST1	},	// 0x03
		{ arg_reg,				instrFXTRACT,		regST0,0		},	// 0x04
		{ arg_reg_reg2,			instrFPREM1,		regST0,regST1	},	// 0x05
		{ arg_NONE,				instrFDECSTP,		0,0				},	// 0x07
		{ arg_NONE,				instrFINCSTP,		0,0				},	// 0x08
	},
	{ // ext_FP_D9_F8
		{ arg_reg_reg2,			instrFPREM,			regST0,regST1	},	// 0x00
		{ arg_reg_reg2,			instrFYL2XP1,		regST0,regST1	},	// 0x01
		{ arg_reg,				instrFSQRT,			regST0,0		},	// 0x02
		{ arg_reg,				instrFSINCOS,		regST0,0		},	// 0x03
		{ arg_reg,				instrFRNDINT,		regST0,0		},	// 0x04
		{ arg_reg_reg2,			instrFSCALE,		regST0,regST1	},	// 0x05
		{ arg_reg,				instrFSIN,			regST0,0		},	// 0x07
		{ arg_reg,				instrFCOS,			regST0,0		},	// 0x08
	},
	{ // ext_FP_DA_E8
		{ 0,					instrRESERVED,		0,0				},	// 0x00
		{ arg_reg_reg2,			instrFUCOMPP,		regST0,regST1	},	// 0x01
		{ 0,					instrRESERVED,		0,0				},	// 0x02
		{ 0,					instrRESERVED,		0,0				},	// 0x03
		{ 0,					instrRESERVED,		0,0				},	// 0x04
		{ 0,					instrRESERVED,		0,0				},	// 0x05
		{ 0,					instrRESERVED,		0,0				},	// 0x07
		{ 0,					instrRESERVED,		0,0				},	// 0x08
	},
	{ // ext_FP_DB_E0
		{ 0,					instrRESERVED,		0,0				},	// 0x00
		{ 0,					instrRESERVED,		0,0				},	// 0x01
		{ arg_NONE,				instrFCLEX,			0,0				},	// 0x02
		{ arg_NONE,				instrFNINIT,		0,0				},	// 0x03
		{ 0,					instrRESERVED,		0,0				},	// 0x04
		{ 0,					instrRESERVED,		0,0				},	// 0x05
		{ 0,					instrRESERVED,		0,0				},	// 0x07
		{ 0,					instrRESERVED,		0,0				},	// 0x08
	},
	{ // ext_FP_DE_D8
		{ 0,					instrRESERVED,		0,0				},	// 0x00
		{ arg_reg_reg2,			instrFCOMPP,		regST0,regST1	},	// 0x01
		{ 0,					instrRESERVED,		0,0				},	// 0x02
		{ 0,					instrRESERVED,		0,0				},	// 0x03
		{ 0,					instrRESERVED,		0,0				},	// 0x04
		{ 0,					instrRESERVED,		0,0				},	// 0x05
		{ 0,					instrRESERVED,		0,0				},	// 0x06
		{ 0,					instrRESERVED,		0,0				},	// 0x07
	},
	{ // ext_FP_DF_E0
		{ arg_reg,				instrFSTSW,			regAX,0			},	// 0x00
		{ 0,					instrRESERVED,		0,0				},	// 0x01
		{ 0,					instrRESERVED,		0,0				},	// 0x02
		{ 0,					instrRESERVED,		0,0				},	// 0x03
		{ 0,					instrRESERVED,		0,0				},	// 0x04
		{ 0,					instrRESERVED,		0,0				},	// 0x05
		{ 0,					instrRESERVED,		0,0				},	// 0x06
		{ 0,					instrRESERVED,		0,0				},	// 0x07
	},
};

//////////////////////////////////////////////////////////////////////

const OpcodeInfo infoMMX_SSE[] =
{
	// ext_MMX_SSE_10
	{ arg_Vo_Wo,			instrMOVUPS,		0,0			},	// 0x10 (n/a)
	{ arg_Vo_Wo,			instrMOVUPD,		0,0			},	// 0x10 (66h)
	{ arg_Vd_Wd,			instrMOVSS,			0,0			},	// 0x10 (F3h)
	{ arg_Vq_Wq,			instrMOVSD,			0,0			},	// 0x10 (F2h)
	// ext_MMX_SSE_11
	{ arg_Wo_Vo,			instrMOVUPS,		0,0			},	// 0x11 (n/a)
	{ arg_Wo_Vo,			instrMOVUPD,		0,0			},	// 0x11 (66h)
	{ arg_Wd_Vd,			instrMOVSS,			0,0			},	// 0x11 (F3h)
	{ arg_Wq_Vq,			instrMOVSD,			0,0			},	// 0x11 (F2h)
	// ext_MMX_SSE_12_MOD_MEM
	{ arg_Vq_Mq,			instrMOVLPS,		0,0			},	// 0x12 (n/a)
	{ arg_Vq_Mq,			instrMOVLPD,	 	0,0			},	// 0x12 (66h)
	// ext_MMX_SSE_12_MOD_11
	{ arg_Vq_WRq,			instrMOVHLPS,		0,0			},	// 0x12 (n/a)
	{ 0,					instrRESERVED,	 	0,0			},	// 0x12 (66h)
	// ext_MMX_SSE_13
	{ arg_Mq_Vq,			instrMOVLPS,		0,0			},	// 0x13 (n/a)
	{ arg_Mq_Vq,			instrMOVLPD,		0,0			},	// 0x13 (66h)
	// ext_MMX_SSE_14
	{ arg_Vo_Wo,			instrUNPCKLPS,		0,0			},	// 0x14 (n/a)
	{ arg_Vo_Wo,			instrUNPCKLPD,		0,0			},	// 0x14 (66h)
	// ext_MMX_SSE_15
	{ arg_Vo_Wo,			instrUNPCKHPS,		0,0			},	// 0x15 (n/a)
	{ arg_Vo_Wo,			instrUNPCKHPD,		0,0			},	// 0x15 (66h)
	// ext_MMX_SSE_16_MOD_MEM
	{ arg_Vq_Mq,			instrMOVHPS,		0,0			},	// 0x16 (n/a)
	{ arg_Vq_Mq,			instrMOVHPD,		0,0			},	// 0x16 (66h)
	// ext_MMX_SSE_16_MOD_11
	{ arg_Vq_WRq,			instrMOVLHPS,		0,0			},	// 0x16 (n/a)
	{ 0,					instrRESERVED,		0,0			},	// 0x16 (66h)
	// ext_MMX_SSE_17
	{ arg_Mq_Vq,			instrMOVHPS,		0,0			},	// 0x17 (n/a)
	{ arg_Mq_Vq,			instrMOVHPD,		0,0			},	// 0x17 (66h)
	// ext_MMX_SSE_28
	{ arg_Vo_Wo,			instrMOVAPS,		0,0			},	// 0x28 (n/a)
	{ arg_Vo_Wo,			instrMOVAPD,		0,0			},	// 0x28 (66h)
	// ext_MMX_SSE_29
	{ arg_Wo_Vo,			instrMOVAPS,		0,0			},	// 0x29 (n/a)
	{ arg_Wo_Vo,			instrMOVAPD,		0,0			},	// 0x29 (66h)
	// ext_MMX_SSE_2A
	{ arg_Vq_Mq,			instrCVTPI2PS,		0,0			},	// 0x2A (n/a)
	{ arg_Vo_Mq,			instrCVTPI2PD,		0,0			},	// 0x2A (66h)
	{ arg_Vd_Ed,			instrCVTSI2SS,		0,0			},	// 0x2A (F3h)
	{ arg_Vq_Ed,			instrCVTSI2SD,		0,0			},	// 0x2A (F2h)
	// ext_MMX_SSE_2B
	{ arg_Mo_Vo,			instrMOVNTPS,		0,0			},	// 0x2B (n/a)
	{ arg_Mo_Vo,			instrMOVNTPD,		0,0			},	// 0x2B (66h)
	// ext_MMX_SSE_2C
	{ arg_Pq_Wq,			instrCVTTPS2PI,		0,0			},	// 0x2C (n/a)
	{ arg_Pq_Wo,			instrCVTTPD2PI,		0,0			},	// 0x2C (66h)
	{ arg_Gd_Wd,			instrCVTTSS2SI,		0,0			},	// 0x2C (F3h)
	{ arg_Gd_Wq,			instrCVTTSD2SI,		0,0			},	// 0x2C (F2h)
	// ext_MMX_SSE_2D
	{ arg_Pq_Wq,			instrCVTPS2PI,		0,0			},	// 0x2D (n/a)
	{ arg_Pq_Wo,			instrCVTPD2PI,		0,0			},	// 0x2D (66h)
	{ arg_Gd_Wd,			instrCVTSS2SI,		0,0			},	// 0x2D (F3h)
	{ arg_Gd_Wq,			instrCVTSD2SI,		0,0			},	// 0x2D (F2h)
	// ext_MMX_SSE_2E
	{ arg_Vd_Wd,			instrUCOMISS,		0,0			},	// 0x2E (n/a)
	{ arg_Vq_Wq,			instrUCOMISD,		0,0			},	// 0x2E (66h)
	// ext_MMX_SSE_2F
	{ arg_Vd_Wd,			instrCOMISS,		0,0			},	// 0x2F (n/a)
	{ arg_Vq_Wq,			instrCOMISD,		0,0			},	// 0x2F (66h)
	// ext_MMX_SSE_50
	{ arg_Rd_Vo,			instrMOVMSKPS,		0,0			},	// 0x50 (n/a)
	{ arg_Rd_Vo,			instrMOVMSKPD,		0,0			},	// 0x50 (66h)
	// ext_MMX_SSE_51
	{ arg_Vo_Wo,			instrSQRTPS,		0,0			},	// 0x51 (n/a)
	{ arg_Vo_Wo,			instrSQRTPD,		0,0			},	// 0x51 (66h)
	{ arg_Vd_Wd,			instrSQRTSS,		0,0			},	// 0x51 (F3h)
	{ arg_Vq_Wq,			instrSQRTSD,		0,0			},	// 0x51 (F2h)
	// ext_MMX_SSE_52
	{ arg_Vo_Wo,			instrRSQRTPS,		0,0			},	// 0x52 (n/a)
	{ 0,					instrRESERVED,		0,0			},	// 0x52
	{ arg_Vd_Wd,			instrRSQRTSS,		0,0			},	// 0x52 (F3h)
	{ 0,					instrRESERVED,		0,0			},	// 0x52
	// ext_MMX_SSE_53
	{ arg_Vo_Wo,			instrRCPPS,			0,0			},	// 0x53 (n/a)
	{ 0,					instrRESERVED,		0,0			},	// 0x53
	{ arg_Vd_Wd,			instrRCPSS,			0,0			},	// 0x53 (F3h)
	{ 0,					instrRESERVED,		0,0			},	// 0x53
	// ext_MMX_SSE_54
	{ arg_Vo_Wo,			instrANDPS,			0,0			},	// 0x54 (n/a)
	{ arg_Vo_Wo,			instrANDPD,			0,0			},	// 0x54 (66h)
	// ext_MMX_SSE_55
	{ arg_Vo_Wo,			instrANDNPS,		0,0			},	// 0x55 (n/a)
	{ arg_Vo_Wo,			instrANDNPD,		0,0			},	// 0x55 (66h)
	// ext_MMX_SSE_56
	{ arg_Vo_Wo,			instrORPS,			0,0			},	// 0x56 (n/a)
	{ arg_Vo_Wo,			instrORPD,			0,0			},	// 0x56 (66h)
	// ext_MMX_SSE_57
	{ arg_Vo_Wo,			instrXORPS,			0,0			},	// 0x57 (n/a)
	{ arg_Vo_Wo,			instrXORPD,			0,0			},	// 0x57 (66h)
	// ext_MMX_SSE_58
	{ arg_Vo_Wo,			instrADDPS,			0,0			},	// 0x58 (n/a)
	{ arg_Vo_Wo,			instrADDPD,			0,0			},	// 0x58 (66h)
	{ arg_Vd_Wd,			instrADDSS,			0,0			},	// 0x58 (F3h)
	{ arg_Vq_Wq,			instrADDSD,			0,0			},	// 0x58 (F2h)
	// ext_MMX_SSE_59
	{ arg_Vo_Wo,			instrMULPS,			0,0			},	// 0x59 (n/a)
	{ arg_Vo_Wo,			instrMULPD,			0,0			},	// 0x59 (66h)
	{ arg_Vd_Wd,			instrMULSS,			0,0			},	// 0x59 (F3h)
	{ arg_Vq_Wq,			instrMULSD,			0,0			},	// 0x59 (F2h)
	// ext_MMX_SSE_5A
	{ arg_Vo_Wq,			instrCVTPS2PD,		0,0			},	// 0x5A (n/a)
	{ arg_Vo_Wo,			instrCVTPD2PS,		0,0			},	// 0x5A (66h)
	{ arg_Vq_Wd,			instrCVTSS2SD,		0,0			},	// 0x5A (F3h)
	{ arg_Vd_Wq,			instrCVTSD2SS,		0,0			},	// 0x5A (F2h)
	// ext_MMX_SSE_5B
	{ arg_Vo_Wo,			instrCVTDQ2PS,		0,0			},	// 0x5B (n/a)
	{ arg_Vo_Wo,			instrCVTPS2DQ,		0,0			},	// 0x5B (66h)
	{ arg_Vo_Wo,			instrCVTTPS2DQ,		0,0			},	// 0x5B (F3h)
	{ 0,					instrRESERVED,		0,0			},	// 0x5B
	// ext_MMX_SSE_5C
	{ arg_Vo_Wo,			instrSUBPS,			0,0			},	// 0x5C (n/a)
	{ arg_Vo_Wo,			instrSUBPD,			0,0			},	// 0x5C (66h)
	{ arg_Vd_Wd,			instrSUBSS,			0,0			},	// 0x5C (F3h)
	{ arg_Vq_Wq,			instrSUBSD,			0,0			},	// 0x5C (F2h)
	// ext_MMX_SSE_5D
	{ arg_Vo_Wo,			instrMINPS,			0,0			},	// 0x5D (n/a)
	{ arg_Vo_Wo,			instrMINPD,			0,0			},	// 0x5D (66h)
	{ arg_Vd_Wd,			instrMINSS,			0,0			},	// 0x5D (F3h)
	{ arg_Vq_Wq,			instrMINSD,			0,0			},	// 0x5D (F2h)
	// ext_MMX_SSE_5E
	{ arg_Vo_Wo,			instrDIVPS,			0,0			},	// 0x5E (n/a)
	{ arg_Vo_Wo,			instrDIVPD,			0,0			},	// 0x5E (66h)
	{ arg_Vd_Wd,			instrDIVSS,			0,0			},	// 0x5E (F3h)
	{ arg_Vq_Wq,			instrDIVSD,			0,0			},	// 0x5E (F2h)
	// ext_MMX_SSE_5F
	{ arg_Vo_Wo,			instrMAXPS,			0,0			},	// 0x5F (n/a)
	{ arg_Vo_Wo,			instrMAXPD,			0,0			},	// 0x5F (66h)
	{ arg_Vd_Wd,			instrMAXSS,			0,0			},	// 0x5F (F3h)
	{ arg_Vq_Wq,			instrMAXSD,			0,0			},	// 0x5F (F2h)
	// ext_MMX_SSE_60
	{ arg_Pq_Qd,			instrPUNPCKLBW,		0,0			},	// 0x60 (n/a)
	{ arg_Vo_Wq,			instrPUNPCKLBW,		0,0			},	// 0x60 (66h)
	// ext_MMX_SSE_61
	{ arg_Pq_Qd,			instrPUNPCKLWD,		0,0			},	// 0x61 (n/a)
	{ arg_Vo_Wq,			instrPUNPCKLWD,		0,0			},	// 0x61 (66h)
	// ext_MMX_SSE_62
	{ arg_Pq_Qd,			instrPUNPCKLDQ,		0,0			},	// 0x62 (n/a)
	{ arg_Vo_Wq,			instrPUNPCKLDQ,		0,0			},	// 0x62 (66h)
	// ext_MMX_SSE_63
	{ arg_Pq_Qq,			instrPACKSSWB,		0,0			},	// 0x63 (n/a)
	{ arg_Vo_Wo,			instrPACKSSWB,		0,0			},	// 0x63 (66h)
	// ext_MMX_SSE_64
	{ arg_Pq_Qq,			instrPCMPGTB,		0,0			},	// 0x64 (n/a)
	{ arg_Vo_Wo,			instrPCMPGTB,		0,0			},	// 0x64 (66h)
	// ext_MMX_SSE_65
	{ arg_Pq_Qq,			instrPCMPGTW,		0,0			},	// 0x65 (n/a)
	{ arg_Vo_Wo,			instrPCMPGTW,		0,0			},	// 0x65 (66h)
	// ext_MMX_SSE_66
	{ arg_Pq_Qq,			instrPCMPGTD,		0,0			},	// 0x66 (n/a)
	{ arg_Vo_Wo,			instrPCMPGTD,		0,0			},	// 0x66 (66h)
	// ext_MMX_SSE_67
	{ arg_Pq_Qq,			instrPACKUSWB,		0,0			},	// 0x67 (n/a)
	{ arg_Vo_Wo,			instrPACKUSWB,		0,0			},	// 0x67 (66h)
	// ext_MMX_SSE_68
	{ arg_Pq_Qq,			instrPUNPCKHBW,		0,0			},	// 0x68 (n/a)
	{ arg_Vo_Wo,			instrPUNPCKHBW,		0,0			},	// 0x68 (66h)
	// ext_MMX_SSE_69
	{ arg_Pq_Qq,			instrPUNPCKHWD,		0,0			},	// 0x69 (n/a)
	{ arg_Vo_Wo,			instrPUNPCKHWD,		0,0			},	// 0x69 (66h)
	// ext_MMX_SSE_6A
	{ arg_Pq_Qq,			instrPUNPCKHDQ,		0,0			},	// 0x6A (n/a)
	{ arg_Vo_Wo,			instrPUNPCKHDQ,		0,0			},	// 0x6A (66h)
	// ext_MMX_SSE_6B
	{ arg_Pq_Qq,			instrPACKSSDW,		0,0			},	// 0x6B (n/a)
	{ arg_Vo_Wo,			instrPACKSSDW,		0,0			},	// 0x6B (66h)
	// ext_MMX_SSE_6C
	{ 0,					instrRESERVED,		0,0			},	// 0x6C
	{ arg_Vo_Wq,			instrPUNPCKLQDQ,	0,0			},	// 0x6C (66h)
	// ext_MMX_SSE_6D
	{ 0,					instrRESERVED,		0,0			},	// 0x6D
	{ arg_Vo_Wo,			instrPUNPCKHQDQ,	0,0			},	// 0x6D (66h)
	// ext_MMX_SSE_6E
	{ arg_Pq_Ed,			instrMOVD,			0,0			},	// 0x6E (n/a)
	{ arg_Vo_Ed,			instrMOVD,			0,0			},	// 0x6E (66h)
	// ext_MMX_SSE_6F
	{ arg_Pq_Qq,			instrMOVQ,			0,0			},	// 0x6F (n/a)
	{ arg_Vo_Wo,			instrMOVDQA,		0,0			},	// 0x6F (66h)
	{ arg_Vo_Wo,			instrMOVDQU,		0,0			},	// 0x6F (F3h)
	{ 0,					instrRESERVED,		0,0			},	// 0x6F
	// ext_MMX_SSE_70
	{ arg_Pq_Qq_Ib,			instrPSHUFW,		0,0			},	// 0x70 (n/a)
	{ arg_Vo_Wo_Ib,			instrPSHUFD,		0,0			},	// 0x70 (66h)
	{ arg_Vo_Wo_Ib,			instrPSHUFHW,		0,0			},	// 0x70 (F3h)
	{ arg_Vo_Wo_Ib,			instrPSHUFLW,		0,0			},	// 0x70 (F2h)

	
/*
	Replaced this with what is below, because the P and Q accesses the REG field,
	but that is already used for the group NNN access (as we came here), so
	P should be Q and V should be W, which access the RM field, that has not been used.

	// ext__MMX_SSE_71_2
	 arg_Pq_Ib,			instrPSRLW,			0,0			},	// 0x71_2 (n/a)
	 arg_Vo_Ib,			instrPSRLW,			0,0			},	// 0x71_2 (66h)
	// ext__MMX_SSE_71_4
	 arg_Pq_Ib,			instrPSRAW,			0,0			},	// 0x71_4 (n/a)
	 arg_Vo_Ib,			instrPSRAW,			0,0			},	// 0x71_4 (66h)
	// ext__MMX_SSE_71_6
	 0,					instrRESERVED,		0,0			},	// 0x71_6
	 arg_Vo_Ib,			instrPSLLW,			0,0			},	// 0x71_6 (66h)
	// ext__MMX_SSE_72_2
	 arg_Pq_Ib,			instrPSRLD,			0,0			},	// 0x72_2 (n/a)
	 arg_Vo_Ib,			instrPSRLD,			0,0			},	// 0x72_2 (66h)
	// ext__MMX_SSE_72_4
	 arg_Pq_Ib,			instrPSRAD,			0,0			},	// 0x72_4 (n/a)
	 arg_Vo_Ib,			instrPSRAD,			0,0			},	// 0x72_4 (66h)
	// ext__MMX_SSE_72_6
	 0,					instrRESERVED,		0,0			},	// 0x72_6
	 arg_Vo_Ib,			instrPSLLD,			0,0			},	// 0x72_6 (66h)
	// ext__MMX_SSE_73_2
	 arg_Pq_Ib,			instrPSRLQ,			0,0			},	// 0x73_2 (n/a)
	 arg_Vo_Ib,			instrPSRLQ,			0,0			},	// 0x73_2 (66h)
	// ext__MMX_SSE_73_3
	 0,					instrRESERVED,		0,0			},	// 0x73_3
	 arg_Vo_Ib,			instrPSRLDQ,		0,0			},	// 0x73_3 (66h)
	// ext__MMX_SSE_73_6
	 0,					instrRESERVED,		0,0			},	// 0x73_6
	 arg_Vo_Ib,			instrPSLLQ,			0,0			},	// 0x73_6 (66h)
	// ext__MMX_SSE_73_7
	 0,					instrRESERVED,		0,0			},	// 0x73_7
	 arg_Vo_Ib,			instrPSLLDQ,		0,0			},	// 0x73_7 (66h)

*/

	// ext_MMX_SSE_71_2
	{ arg_QRq_Ib,			instrPSRLW,			0,0			},	// 0x71_2 (n/a)
	{ arg_WRo_Ib,			instrPSRLW,			0,0			},	// 0x71_2 (66h)
	// ext_MMX_SSE_71_4
	{ arg_QRq_Ib,			instrPSRAW,			0,0			},	// 0x71_4 (n/a)
	{ arg_WRo_Ib,			instrPSRAW,			0,0			},	// 0x71_4 (66h)
	// ext_MMX_SSE_71_6
	{ 0,					instrRESERVED,		0,0			},	// 0x71_6
	{ arg_WRo_Ib,			instrPSLLW,			0,0			},	// 0x71_6 (66h)
	// ext_MMX_SSE_72_2
	{ arg_QRq_Ib,			instrPSRLD,			0,0			},	// 0x72_2 (n/a)
	{ arg_WRo_Ib,			instrPSRLD,			0,0			},	// 0x72_2 (66h)
	// ext_MMX_SSE_72_4
	{ arg_QRq_Ib,			instrPSRAD,			0,0			},	// 0x72_4 (n/a)
	{ arg_WRo_Ib,			instrPSRAD,			0,0			},	// 0x72_4 (66h)
	// ext_MMX_SSE_72_6
	{ 0,					instrRESERVED,		0,0			},	// 0x72_6
	{ arg_WRo_Ib,			instrPSLLD,			0,0			},	// 0x72_6 (66h)
	// ext_MMX_SSE_73_2
	{ arg_QRq_Ib,			instrPSRLQ,			0,0			},	// 0x73_2 (n/a)
	{ arg_WRo_Ib,			instrPSRLQ,			0,0			},	// 0x73_2 (66h)
	// ext_MMX_SSE_73_3
	{ 0,					instrRESERVED,		0,0			},	// 0x73_3
	{ arg_WRo_Ib,			instrPSRLDQ,		0,0			},	// 0x73_3 (66h)
	// ext_MMX_SSE_73_6
	{ 0,					instrRESERVED,		0,0			},	// 0x73_6
	{ arg_WRo_Ib,			instrPSLLQ,			0,0			},	// 0x73_6 (66h)
	// ext_MMX_SSE_73_7
	{ 0,					instrRESERVED,		0,0			},	// 0x73_7
	{ arg_WRo_Ib,			instrPSLLDQ,		0,0			},	// 0x73_7 (66h)


	
	// ext_MMX_SSE_74
	{ arg_Pq_Qq,			instrPCMPEQB,		0,0			},	// 0x74 (n/a)
	{ arg_Vo_Wo,			instrPCMPEQB,		0,0			},	// 0x74 (66h)
	// ext_MMX_SSE_75
	{ arg_Pq_Qq,			instrPCMPEQW,		0,0			},	// 0x75 (n/a)
	{ arg_Vo_Wo,			instrPCMPEQW,		0,0			},	// 0x75 (66h)
	// ext_MMX_SSE_76
	{ arg_Pq_Qq,			instrPCMPEQD,		0,0			},	// 0x76 (n/a)
	{ arg_Vo_Wo,			instrPCMPEQD,		0,0			},	// 0x76 (66h)
	// ext_MMX_SSE_7E
	{ arg_Ed_Pd,			instrMOVD,			0,0			},	// 0x7E (n/a)
	{ arg_Ed_Vd,			instrMOVD,			0,0			},	// 0x7E (66h)
	{ arg_Vo_Mq_OR_Vq_Wq,	instrMOVQ,			0,0			},	// 0x7E (F3h)
	{ 0,					instrRESERVED,		0,0			},	// 0x7E
	// ext_MMX_SSE_7F
	{ arg_Qq_Pq,			instrMOVQ,			0,0			},	// 0x7F (n/a)
	{ arg_Wo_Vo,			instrMOVDQA,		0,0			},	// 0x7F (66h)
	{ arg_Wo_Vo,			instrMOVDQU,		0,0			},	// 0x7F (F3h)
	{ 0,					instrRESERVED,		0,0			},	// 0x7F
	// ext_MMX_SSE_C2
	{ arg_Vo_Wo_Ib,			instrCMPPS,			0,0			},	// 0xC2 (n/a)
	{ arg_Vo_Wo_Ib,			instrCMPPD,			0,0			},	// 0xC2 (66h)
	{ arg_Vd_Wd_Ib,			instrCMPSS,			0,0			},	// 0xC2 (F3h)
	{ arg_Vq_Wq_Ib,			instrCMPSD,			0,0			},	// 0xC2 (F2h)


/*
	Replaced this with what is below, because the R access is to the
	low 16 bits anyway, read access

	// ext___MMX_SSE_C4_MOD_MEM
	  arg_Pq_Mw_Ib,			instrPINSRW,		0,0			},	// 0xC4 (n/a)
	  arg_Vo_Mw_Ib,			instrPINSRW,		0,0			},	// 0xC4 (66h)
	// ext___MMX_SSE_C4_MOD_11
	  arg_Pq_Rd_Ib,			instrPINSRW,		0,0			},	// 0xC4 (n/a)
	  arg_Vo_Rd_Ib,			instrPINSRW,		0,0			},	// 0xC4 (66h)

    and this line

          arg_MODRM,			instrPRE_MMX_SSE_MOD_MEM_11,ext___MMX_SSE_C4_MOD_MEM,ext2_SSE_PREFIX_NO_66	},	// 0xC4

    with

  	  arg_MODRM,			instrPRE_MMX_SSE,ext___MMX_SSE_C4,ext2_SSE_PREFIX_NO_66	},	// 0xC4

*/

	// ext_MMX_SSE_C4
	{ arg_Pq_Ew_Ib,			instrPINSRW,		0,0			},	// 0xC4 (n/a)
	{ arg_Vo_Ew_Ib,			instrPINSRW,		0,0			},	// 0xC4 (66h)


	// ext_MMX_SSE_C5
	{ arg_Gd_QRq_Ib,		instrPEXTRW,		0,0			},	// 0xC5 (n/a)
	{ arg_Gd_WRo_Ib,		instrPEXTRW,		0,0			},	// 0xC5 (66h)
	// ext_MMX_SSE_C6
	{ arg_Vo_Wo_Ib,			instrSHUFPS,		0,0			},	// 0xC6 (n/a)
	{ arg_Vo_Wo_Ib,			instrSHUFPD,		0,0			},	// 0xC6 (66h)
	// ext_MMX_SSE_D1
	{ arg_Pq_Qq,			instrPSRLW,			0,0			},	// 0xD1 (n/a)
	{ arg_Vo_Wo,			instrPSRLW,			0,0			},	// 0xD1 (66h)
	// ext_MMX_SSE_D2
	{ arg_Pq_Qq,			instrPSRLD,			0,0			},	// 0xD2 (n/a)
	{ arg_Vo_Wo,			instrPSRLD,			0,0			},	// 0xD2 (66h)
	// ext_MMX_SSE_D3
	{ arg_Pq_Qq,			instrPSRLQ,			0,0			},	// 0xD3 (n/a)
	{ arg_Vo_Wo,			instrPSRLQ,			0,0			},	// 0xD3 (66h)
	// ext_MMX_SSE_D4
	{ arg_Pq_Qq,			instrPADDQ,			0,0			},	// 0xD4 (n/a)
	{ arg_Vo_Wo,			instrPADDQ,			0,0			},	// 0xD4 (66h)
	// ext_MMX_SSE_D5
	{ arg_Pq_Qq,			instrPMULLW,		0,0			},	// 0xD5 (n/a)
	{ arg_Vo_Wo,			instrPMULLW,		0,0			},	// 0xD5 (66h)
	// ext_MMX_SSE_D6
	{ 0,					instrRESERVED,		0,0			},	// 0xD6
	{ arg_Mq_Vq_OR_Wq_Vq,	instrMOVQ,			0,0			},	// 0xD6 (66h)
	{ arg_Vo_QRq,			instrMOVQ2DQ,		0,0			},	// 0xD6 (F3h)
	{ arg_Pq_WRq,			instrMOVDQ2Q,		0,0			},	// 0xD6 (F2h)
	// ext_MMX_SSE_D7
	{ arg_Rd_Pq,			instrPMOVMSKB,		0,0			},	// 0xD7 (n/a)
	{ arg_Rd_Vo,			instrPMOVMSKB,		0,0			},	// 0xD7 (66h)
	// ext_MMX_SSE_D8
	{ arg_Pq_Qq,			instrPSUBUSB,		0,0			},	// 0xD8 (n/a)
	{ arg_Vo_Wo,			instrPSUBUSB,		0,0			},	// 0xD8 (66h)
	// ext_MMX_SSE_D9
	{ arg_Pq_Qq,			instrPSUBUSW,		0,0			},	// 0xD9 (n/a)
	{ arg_Vo_Wo,			instrPSUBUSW,		0,0			},	// 0xD9 (66h)
	// ext_MMX_SSE_DA
	{ arg_Pq_Qq,			instrPMINUB,		0,0			},	// 0xDA (n/a)
	{ arg_Vo_Wo,			instrPMINUB,		0,0			},	// 0xDA (66h)
	// ext_MMX_SSE_DB
	{ arg_Pq_Qq,			instrPAND,			0,0			},	// 0xDB (n/a)
	{ arg_Vo_Wo,			instrPAND,			0,0			},	// 0xDB (66h)
	// ext_MMX_SSE_DC
	{ arg_Pq_Qq,			instrPADDUSB,		0,0			},	// 0xDC (n/a)
	{ arg_Vo_Wo,			instrPADDUSB,		0,0			},	// 0xDC (66h)
	// ext_MMX_SSE_DD
	{ arg_Pq_Qq,			instrPADDUSW,		0,0			},	// 0xDD (n/a)
	{ arg_Vo_Wo,			instrPADDUSW,		0,0			},	// 0xDD (66h)
	// ext_MMX_SSE_DE
	{ arg_Pq_Qq,			instrPMAXUB,		0,0			},	// 0xDE (n/a)
	{ arg_Vo_Wo,			instrPMAXUB,		0,0			},	// 0xDE (66h)
	// ext_MMX_SSE_DF
	{ arg_Pq_Qq,			instrPANDN,			0,0			},	// 0xDF (n/a)
	{ arg_Vo_Wo,			instrPANDN,			0,0			},	// 0xDF (66h)
	// ext_MMX_SSE_E0
	{ arg_Pq_Qq,			instrPAVGB,			0,0			},	// 0xE0 (n/a)
	{ arg_Vo_Wo,			instrPAVGB,			0,0			},	// 0xE0 (66h)
	// ext_MMX_SSE_E1
	{ arg_Pq_Qq,			instrPSRAW,			0,0			},	// 0xE1 (n/a)
	{ arg_Vo_Wo,			instrPSRAW,			0,0			},	// 0xE1 (66h)
	// ext_MMX_SSE_E2
	{ arg_Pq_Qq,			instrPSRAD,			0,0			},	// 0xE2 (n/a)
	{ arg_Vo_Wo,			instrPSRAD,			0,0			},	// 0xE2 (66h)
	// ext_MMX_SSE_E3
	{ arg_Pq_Qq,			instrPAVGW,			0,0			},	// 0xE3 (n/a)
	{ arg_Vo_Wo,			instrPAVGW,			0,0			},	// 0xE3 (66h)
	// ext_MMX_SSE_E4
	{ arg_Pq_Qq,			instrPMULHUW,		0,0			},	// 0xE4 (n/a)
	{ arg_Vo_Wo,			instrPMULHUW,		0,0			},	// 0xE4 (66h)
	// ext_MMX_SSE_E5
	{ arg_Pq_Qq,			instrPMULHW,		0,0			},	// 0xE5 (n/a)
	{ arg_Vo_Wo,			instrPMULHW,		0,0			},	// 0xE5 (66h)
	// ext_MMX_SSE_E6
	{ 0,					instrRESERVED,		0,0			},	// 0xE6
	{ arg_Vo_Wo,			instrCVTTPD2DQ,		0,0			},	// 0xE6 (66h)
	{ arg_Vo_Wq,			instrCVTDQ2PD,		0,0			},	// 0xE6 (F3h)
	{ arg_Vo_Wo,			instrCVTPD2DQ,		0,0			},	// 0xE6 (F2h)
	// ext_MMX_SSE_E7
	{ arg_Mq_Pq,			instrMOVNTQ,		0,0			},	// 0xE7 (n/a)
	{ arg_Mo_Vo,			instrMOVNTDQ,		0,0			},	// 0xE7 (66h)
	// ext_MMX_SSE_E8
	{ arg_Pq_Qq,			instrPSUBSB,		0,0			},	// 0xE8 (n/a)
	{ arg_Vo_Wo,			instrPSUBSB,		0,0			},	// 0xE8 (66h)
	// ext_MMX_SSE_E9
	{ arg_Pq_Qq,			instrPSUBSW,		0,0			},	// 0xE9 (n/a)
	{ arg_Vo_Wo,			instrPSUBSW,		0,0			},	// 0xE9 (66h)
	// ext_MMX_SSE_EA
	{ arg_Pq_Qq,			instrPMINSW,		0,0			},	// 0xEA (n/a)
	{ arg_Vo_Wo,			instrPMINSW,		0,0			},	// 0xEA (66h)
	// ext_MMX_SSE_EB
	{ arg_Pq_Qq,			instrPOR,			0,0			},	// 0xEB (n/a)
	{ arg_Vo_Wo,			instrPOR,			0,0			},	// 0xEB (66h)
	// ext_MMX_SSE_EC
	{ arg_Pq_Qq,			instrPADDSB,		0,0			},	// 0xEC (n/a)
	{ arg_Vo_Wo,			instrPADDSB,		0,0			},	// 0xEC (66h)
	// ext_MMX_SSE_ED
	{ arg_Pq_Qq,			instrPADDSW,		0,0			},	// 0xED (n/a)
	{ arg_Vo_Wo,			instrPADDSW,		0,0			},	// 0xED (66h)
	// ext_MMX_SSE_EE
	{ arg_Pq_Qq,			instrPMAXSW,		0,0			},	// 0xEE (n/a)
	{ arg_Vo_Wo,			instrPMAXSW,		0,0			},	// 0xEE (66h)
	// ext_MMX_SSE_EF
	{ arg_Pq_Qq,			instrPXOR,			0,0			},	// 0xEF (n/a)
	{ arg_Vo_Wo,			instrPXOR,			0,0			},	// 0xEF (66h)
	// ext_MMX_SSE_F1
	{ arg_Pq_Qq,			instrPSLLW,			0,0			},	// 0xF1 (n/a)
	{ arg_Vo_Wo,			instrPSLLW,			0,0			},	// 0xF1 (66h)
	// ext_MMX_SSE_F2
	{ arg_Pq_Qq,			instrPSLLD,			0,0			},	// 0xF2 (n/a)
	{ arg_Vo_Wo,			instrPSLLD,			0,0			},	// 0xF2 (66h)
	// ext_MMX_SSE_F3
	{ arg_Pq_Qq,			instrPSLLQ,			0,0			},	// 0xF3 (n/a)
	{ arg_Vo_Wo,			instrPSLLQ,			0,0			},	// 0xF3 (66h)
	// ext_MMX_SSE_F4
	{ arg_Pq_Qq,			instrPMULUDQ,		0,0			},	// 0xF4 (n/a)
	{ arg_Vo_Wo,			instrPMULUDQ,		0,0			},	// 0xF4 (66h)
	// ext_MMX_SSE_F5
	{ arg_Pq_Qq,			instrPMADDWD,		0,0			},	// 0xF5 (n/a)
	{ arg_Vo_Wo,			instrPMADDWD,		0,0			},	// 0xF5 (66h)
	// ext_MMX_SSE_F6
	{ arg_Pq_Qq,			instrPSADBW,		0,0			},	// 0xF6 (n/a)
	{ arg_Vo_Wo,			instrPSADBW,		0,0			},	// 0xF6 (66h)
	// ext_MMX_SSE_F7
	{ arg_Pq_QRq,			instrMASKMOVQ,		0,0			},	// 0xF7 (n/a)
	{ arg_Vo_WRo,			instrMASKMOVDQU,	0,0			},	// 0xF7 (66h)
	// ext_MMX_SSE_F8
	{ arg_Pq_Qq,			instrPSUBB,			0,0			},	// 0xF8 (n/a)
	{ arg_Vo_Wo,			instrPSUBB,			0,0			},	// 0xF8 (66h)
	// ext_MMX_SSE_F9
	{ arg_Pq_Qq,			instrPSUBW,			0,0			},	// 0xF9 (n/a)
	{ arg_Vo_Wo,			instrPSUBW,			0,0			},	// 0xF9 (66h)
	// ext_MMX_SSE_FA
	{ arg_Pq_Qq,			instrPSUBD,			0,0			},	// 0xFA (n/a)
	{ arg_Vo_Wo,			instrPSUBD,			0,0			},	// 0xFA (66h)
	// ext_MMX_SSE_FB
	{ arg_Pq_Qq,			instrPSUBQ,			0,0			},	// 0xFB (n/a)
	{ arg_Vo_Wo,			instrPSUBQ,			0,0			},	// 0xFB (66h)
	// ext_MMX_SSE_FC
	{ arg_Pq_Qq,			instrPADDB,			0,0			},	// 0xFC (n/a)
	{ arg_Vo_Wo,			instrPADDB,			0,0			},	// 0xFC (66h)
	// ext_MMX_SSE_FD
	{ arg_Pq_Qq,			instrPADDW,			0,0			},	// 0xFD (n/a)
	{ arg_Vo_Wo,			instrPADDW,			0,0			},	// 0xFD (66h)
	// ext_MMX_SSE_FE
	{ arg_Pq_Qq,			instrPADDD,			0,0			},	// 0xFE (n/a)
	{ arg_Vo_Wo,			instrPADDD,			0,0			},	// 0xFE (66h)
};

//////////////////////////////////////////////////////////////////////









