/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// InstructionEncoder.cpp: implementation of the InstructionEncoder class.
//
//////////////////////////////////////////////////////////////////////

#include "InstructionEncoder.h" 
#include "InstructionDecoder.h" 
#include "TryDoInstructionEncoding.h"

#include "DataDestPtr.h" 

#include <assert.h>
#include <stdlib.h>
#include <memory.h>

#include <list>
#include <map>

using namespace std;

#ifdef _MSC_VER
#include <typeinfo.h>
#else
#include <typeinfo>
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

InverseOpcodeInfo*      InstructionEncoder::m_pInverseOpcodeInfo = NULL;
InverseOpcodeInfoIndex* InstructionEncoder::m_pInverseOpcodeInfoIndex = NULL;
int                     InstructionEncoder::m_nInverseOpcodeInfoIndex = 0;

InstructionEncoder::InstructionEncoder()
{
  m_bCodeSize32 = true;
  
}

InstructionEncoder::~InstructionEncoder()
{
  // delete[] m_pInverseOpcodeInfo; // do not delete, it's a static
}
/*
  struct InverseOpcodeInfo
{
	byte m_nByte;		// opcode byte
	byte m_nGroupNNN;	// index into group structure
	byte m_nSubGroupN;	// index into subgroup structure or mmx/sse structure
	byte m_nFlags;
};
*/

typedef multimap<word,InverseOpcodeInfo> MapInstrInverseOpcodeInfo;

int compare_word(const void* p1,const void* p2)
{
  if (*((const word*) p1) < *((const word*) p2))
    return -1;
  return *((const word*) p1) > *((const word*) p2);
}
/*
typedef multimap<unsigned int,byte> MapGroupExtraAccessBytes;
*/
word OpMayBeType (dword dwArg,dword dwOpArg,byte bExtra2 = 0)
{
  switch (dwOpArg)
    {
    case arg_OP_IS_IMM: return INVERSE_OPCODE_OP_MAY_BE_IMM;
    case arg_OP_AS_REG_EXTRA: return INVERSE_OPCODE_OP_MAY_BE_REG;
      
    case arg_OP_IS_MEM_DISP: return INVERSE_OPCODE_OP_MAY_BE_MEM;
    case arg_OP_AS_REG_EXTRA2: return INVERSE_OPCODE_OP_MAY_BE_REG;
    case arg_OP_LOW_3BITS_REG: return INVERSE_OPCODE_OP_MAY_BE_REG;
    case arg_OP_CONST_1: return INVERSE_OPCODE_OP_MAY_BE_IMM;
      
    case arg_OP_IS_MODRM_REG: return INVERSE_OPCODE_OP_MAY_BE_REG;
    case arg_OP_IS_MODRM_MODRM:
      if (dwArg & arg_MODRM_ONLY_M_MODRM)
	{
	  assert (!(dwArg & arg_MODRM_ONLY_R_MODRM)); // cannot be both only mem and only reg, because then we can only be nothing, which is wrong
	  return INVERSE_OPCODE_OP_MAY_BE_MEM;
	}
      else if (dwArg & arg_MODRM_ONLY_R_MODRM)
	return INVERSE_OPCODE_OP_MAY_BE_REG;
      return INVERSE_OPCODE_OP_MAY_BE_MEM | INVERSE_OPCODE_OP_MAY_BE_REG;
    }
  assert (dwOpArg == 0);
  
  if ((dwArg & arg_EXTRA2_DEST_SRC_SI_DI) && bExtra2)
    return INVERSE_OPCODE_OP_MAY_BE_MEM;
  return 0;
}

word OpsMayBeType (dword dwArg,byte bExtra2)
{
  return
    (OpMayBeType(dwArg,
		 (dwArg & arg_OP_DEST_MASK) >> 
		 arg_OP_DEST_SHIFT,
		 bExtra2 & (ext2_DEST_MEM_DI | ext2_DEST_MEM_SI)) << 
     INVERSE_OPCODE_OP_DEST_SHIFT) |
    (OpMayBeType(dwArg,
		 (dwArg & arg_OP_SRC_MASK ) >> 
		 arg_OP_SRC_SHIFT ,
		 bExtra2 & (ext2_SRC_MEM_DI  | ext2_SRC_MEM_SI )) << 
     INVERSE_OPCODE_OP_SRC_SHIFT ) |
    (OpMayBeType(dwArg,
		 (dwArg & arg_OP_SRC2_MASK) >> arg_OP_SRC2_SHIFT) << 
     INVERSE_OPCODE_OP_SRC2_SHIFT);
}
/*
dword OnlyMOrRFromMod (word wFlags)
{
	if ((wFlags & (INVERSE_OPCODE_TAG_GROUP_MOD_11 | INVERSE_OPCODE_TAG_GROUP_MOD_MEM)) == INVERSE_OPCODE_TAG_GROUP_MOD_11)
		return arg_MODRM_ONLY_R_MODRM;
	if ((wFlags & (INVERSE_OPCODE_TAG_GROUP_MOD_11 | INVERSE_OPCODE_TAG_GROUP_MOD_MEM)) == INVERSE_OPCODE_TAG_GROUP_MOD_MEM)
		return arg_MODRM_ONLY_M_MODRM;
	return 0;
}
*/
/*
void SetGroupInfo(const OpcodeInfo* pInfo,
				  InverseOpcodeInfo* groupInfo,
				  MapGroupExtraAccessBytes& mapGroupExtraAccessBytes,
				  InverseOpcodeInfo* mmxsseInfo,
				  int i,word wFlags,byte nGroupNNN = 0)
{
	if ((pInfo->m_nInstruction & instrmIS_GROUP_MASK) == instrgGROUP)
	{
		int nGroupType = pInfo->m_nInstruction & instrmGROUP_TYPE_MASK;

		assert ((pInfo->m_nArguments & arg_OP_DEST_MASK) != arg_DEST_IS_MODRM_REG); // as group got NNN from REG, we can't get register from there too
		assert ((pInfo->m_nArguments & arg_OP_SRC_MASK ) != arg_SRC_IS_MODRM_REG ); // as group got NNN from REG, we can't get register from there too

		if ((nGroupType & instrmGROUP_MOD_BOTH) ||
				(nGroupType & instrmGROUP_MOD_MEM))
		{
			assert (pInfo->m_nExtra1 < sizeof (infoGroups)/(sizeof(infoGroups[0])));
			InverseOpcodeInfo* pGroupInfo = groupInfo+pInfo->m_nExtra1;

			if (pGroupInfo->m_nFlags)
			{
				assert (pGroupInfo->m_nByte || pInfo->m_nExtra1 == ext_GRP_6);
				mapGroupExtraAccessBytes.insert(MapGroupExtraAccessBytes::value_type(pInfo->m_nExtra1,i));

				assert ((pGroupInfo->m_nFlags & (wFlags | INVERSE_OPCODE_TAG_GROUP_MOD_MEM | INVERSE_OPCODE_TAG_GROUP_MOD_11)) ==
					(wFlags | 
					 (nGroupType & instrmGROUP_MOD_BOTH ? 
					 (INVERSE_OPCODE_TAG_GROUP_MOD_MEM | INVERSE_OPCODE_TAG_GROUP_MOD_11) :
					 INVERSE_OPCODE_TAG_GROUP_MOD_MEM)));
			}
			else
			{
				assert (!pGroupInfo->m_nByte);
				pGroupInfo->m_nByte = i;

				pGroupInfo->m_nFlags = wFlags | 
					(nGroupType & instrmGROUP_MOD_BOTH ? 
					(INVERSE_OPCODE_TAG_GROUP_MOD_MEM | INVERSE_OPCODE_TAG_GROUP_MOD_11) :
					INVERSE_OPCODE_TAG_GROUP_MOD_MEM) | 
					OpsMayBeType(pInfo->m_nArguments | (nGroupType & instrmGROUP_MOD_BOTH ? 0 : arg_MODRM_ONLY_M_MODRM));
			}
		}
		if (nGroupType & instrmGROUP_MOD_11)
		{
			assert (!(nGroupType & instrmGROUP_MOD_BOTH));

			assert (pInfo->m_nExtra2 < sizeof (infoGroups)/(sizeof(infoGroups[0])));
			InverseOpcodeInfo* pGroupInfo = groupInfo+pInfo->m_nExtra2;

			assert (!pGroupInfo->m_nFlags);
			assert (!pGroupInfo->m_nByte);

			pGroupInfo->m_nByte = i;
			pGroupInfo->m_nFlags = wFlags | 
				INVERSE_OPCODE_TAG_GROUP_MOD_11 |
				OpsMayBeType(pInfo->m_nArguments | arg_MODRM_ONLY_R_MODRM);
		}
	}
	else if (pInfo->m_nInstruction == instrSUBGRP_COPROC)
	{
		assert (pInfo->m_nExtra1 < sizeof (infoGroups)/(sizeof(infoGroups[0])));
		InverseOpcodeInfo* pGroupInfo = groupInfo+pInfo->m_nExtra1;
		InverseOpcodeInfo* pSrcGroupInfo = &(groupInfo[i]);

		assert (!pGroupInfo->m_nFlags);
		assert (mapGroupExtraAccessBytes.find(i) == mapGroupExtraAccessBytes.end());

		*pGroupInfo = *pSrcGroupInfo; //set byte and m_nFlags
		pGroupInfo->m_nFlags |= wFlags;

		assert (!(pInfo->m_nArguments & (arg_OP_DEST_MASK | arg_OP_SRC_MASK | arg_OP_SRC2_MASK))); // so there is no need to call OpsMayBeType

		pGroupInfo->m_nGroupNNN = nGroupNNN;
	}
	else if (pInfo->m_nInstruction == instrPRE_MMX_SSE ||
			 pInfo->m_nInstruction == instrPRE_MMX_SSE_MOD_MEM_11)
	{
		InverseOpcodeInfo m_info;

		if (wFlags & INVERSE_OPCODE_TAG_SUBGROUP)
		{
			assert (mapGroupExtraAccessBytes.find(i) == mapGroupExtraAccessBytes.end());
			assert (groupInfo[i].m_nFlags);

			m_info = (groupInfo[i]);
			m_info.m_nFlags |= INVERSE_OPCODE_TAG_MMX_SSE_PREFIX;
		}
		else
		{
			m_info.m_nByte = i;
			m_info.m_nFlags = wFlags | INVERSE_OPCODE_TAG_MMX_SSE_PREFIX;
		}

		assert (pInfo->m_nExtra1 < sizeof (infoMMX_SSE)/(2*sizeof(infoMMX_SSE[0])));
		InverseOpcodeInfo* pMMXSSEInfo = mmxsseInfo+pInfo->m_nExtra1;

		m_info.m_nGroupNNN = nGroupNNN;
		m_info.m_nSubGroupN = 0;
		*pMMXSSEInfo = m_info;
		if (pInfo->m_nExtra2 & (ext2_SSE_PREFIX_F3_F2))
		{
			assert (pInfo->m_nInstruction == instrPRE_MMX_SSE);

			assert ((byte) (pInfo->m_nExtra1+1) < sizeof (infoMMX_SSE)/(2*sizeof(infoMMX_SSE[0])));
			pMMXSSEInfo++; // = mmxsseInfo+(pInfo->m_nExtra1+1);

			m_info.m_nSubGroupN = 1 << 1;
			*pMMXSSEInfo = m_info;
		}
		else if (pInfo->m_nInstruction == instrPRE_MMX_SSE_MOD_MEM_11)
		{
			pMMXSSEInfo->m_nFlags |= INVERSE_OPCODE_TAG_GROUP_MOD_MEM;

			assert ((byte) (pInfo->m_nExtra1+1) < sizeof (infoMMX_SSE)/(2*sizeof(infoMMX_SSE[0])));
			pMMXSSEInfo++; // = mmxsseInfo+(pInfo->m_nExtra1+1);

			*pMMXSSEInfo = m_info;
			pMMXSSEInfo->m_nFlags |= INVERSE_OPCODE_TAG_GROUP_MOD_11;
		}
	}
}
*/
/*
bool DestOpMayBeMem(dword dwArg)
{
	dword dwDestOp = (dwArg & arg_OP_DEST_MASK);

	return ((dwDestOp == arg_DEST_IS_MODRM_MODRM &&  // *dest modrm-mod, and
			 !(dwArg & arg_MODRM_ONLY_R_MODRM)) ||	 // can be mem (not only reg)
			dwDestOp == arg_DEST_IS_MEM_DISP);       // *dest mem(disp)
}
*/
/*
void FillInSetupInverseInsert(MapInstrInverseOpcodeInfo& mapInstrInverseOpcodeInfo,
							  word nInstr,byte nByte,word wFlags)
{
	InverseOpcodeInfo info;

	info.m_nByte = nByte;
	info.m_nFlags = wFlags;
//	info.m_nGroupNNN = info.m_nSubGroupN = 0;
	
	mapInstrInverseOpcodeInfo.insert(MapInstrInverseOpcodeInfo::value_type(nInstr,info));
}

void SetupInverseInsert(MapInstrInverseOpcodeInfo& mapInstrInverseOpcodeInfo,
						word nInstr,byte nByte,word wFlags)
{
	FillInSetupInverseInsert(mapInstrInverseOpcodeInfo,nInstr,nByte,wFlags);

	if ((wFlags & INVERSE_OPCODE_DEST_MAY_BE_MEM) && // tested first, as easier, and most likely false
		IsValidLockInstruction (nInstr)) 
	{
		// if it may have lock prefix
		// and can encode dest as memory reference 

		FillInSetupInverseInsert(mapInstrInverseOpcodeInfo,
			nInstr | instrgmINT_ARITH_MEMORY_LOCK,
			nByte,
			wFlags | INVERSE_OPCODE_TAG_LOCK_REP_PREFIX);
	}
	else // can't have both lock and rep
	if ((nInstr & instrgMASK) == instrgSTRING)
	{
		if (nInstr == instrCMPS ||
			nInstr == instrSCAS)
		{
			FillInSetupInverseInsert(mapInstrInverseOpcodeInfo,nInstr | instrgmSTRING_REPNE,nByte,wFlags | INVERSE_OPCODE_TAG_LOCK_REP_PREFIX);
			FillInSetupInverseInsert(mapInstrInverseOpcodeInfo,nInstr | instrgmSTRING_REPE ,nByte,wFlags | INVERSE_OPCODE_TAG_LOCK_REP_PREFIX);
		}
		else
			FillInSetupInverseInsert(mapInstrInverseOpcodeInfo,nInstr | instrgmSTRING_REP  ,nByte,wFlags | INVERSE_OPCODE_TAG_LOCK_REP_PREFIX);
	}
}

void SetupInverseInsert(MapInstrInverseOpcodeInfo& mapInstrInverseOpcodeInfo,
						const MapGroupExtraAccessBytes& mapGroupExtraAccessBytes,
						word nInstr,
						const InverseOpcodeInfo* pGroupInfo,
						int nGroup,int nGroupN,
						word wFlags)
{
	InverseOpcodeInfo info;

	info = *pGroupInfo; // set byte, flags and groupNNN (if subgroup)
	info.m_nFlags |= wFlags;

	if (pGroupInfo->m_nFlags & INVERSE_OPCODE_TAG_SUBGROUP)
		info.m_nSubGroupN = nGroupN;
	else
		info.m_nGroupNNN = nGroupN;

	mapInstrInverseOpcodeInfo.insert(MapInstrInverseOpcodeInfo::value_type(nInstr,info));

	MapGroupExtraAccessBytes::const_iterator extraAccessByte = mapGroupExtraAccessBytes.find(nGroup);
	for ( ; extraAccessByte != mapGroupExtraAccessBytes.end() && extraAccessByte->first == nGroup; ++extraAccessByte)
	{
		info.m_nByte = extraAccessByte->second;

		mapInstrInverseOpcodeInfo.insert(MapInstrInverseOpcodeInfo::value_type(nInstr,info));
	}

	if ((info.m_nFlags & INVERSE_OPCODE_DEST_MAY_BE_MEM) && // tested first, as easier, and most likely false
		IsValidLockInstruction (nInstr)) // tested last 
	{
		// if it may have lock prefix
		// and can encode dest as memory reference 

		assert (!(wFlags & INVERSE_OPCODE_TAG_LOCK_REP_PREFIX)); // so we were not called recursively

		SetupInverseInsert(mapInstrInverseOpcodeInfo,
			mapGroupExtraAccessBytes,
			nInstr | instrgmINT_ARITH_MEMORY_LOCK,
			pGroupInfo,
			nGroup,nGroupN,
			wFlags | INVERSE_OPCODE_TAG_LOCK_REP_PREFIX);
	}
}
*/

void AddInverseInfo(word nInstruction,
		    const InverseOpcodeInfo& inverseInfo,
		    MapInstrInverseOpcodeInfo& mapInstrInverseOpcodeInfo)
{
#ifndef NDEBUG
  if (inverseInfo.m_nFlags & (INVERSE_OPCODE_TAG_GROUP_MOD_MEM | 
			      INVERSE_OPCODE_TAG_GROUP_MOD_11))
    {
      assert ((inverseInfo.m_nArguments & arg_OP_DEST_MASK) != 
	      arg_DEST_IS_MODRM_REG); // as group got NNN from REG, we can't get register from there too
      assert ((inverseInfo.m_nArguments & arg_OP_SRC_MASK ) != 
	      arg_SRC_IS_MODRM_REG ); // as group got NNN from REG, we can't get register from there too
    }
  if (inverseInfo.m_nFlags & (INVERSE_OPCODE_TAG_SUBGROUP))
    {
      assert ((inverseInfo.m_nArguments & arg_OP_DEST_MASK) != 
	      arg_DEST_IS_MODRM_MODRM); // as subgroup got N from RM, we can't get register or mem from there too
      assert ((inverseInfo.m_nArguments & arg_OP_SRC_MASK ) != 
	      arg_SRC_IS_MODRM_MODRM ); // as subgroup got N from RM, we can't get register or mem from there too
    }
  
  OpcodeInfo info;
  
  info.m_nArguments = inverseInfo.m_nArguments;
  info.m_nExtra1 = inverseInfo.m_nExtra1;
  info.m_nExtra2 = inverseInfo.m_nExtra2;
  info.m_nInstruction = nInstruction;
  
  OpcodeInfoSanityCheck (info);
#endif

  mapInstrInverseOpcodeInfo.insert(MapInstrInverseOpcodeInfo::value_type(nInstruction,inverseInfo));
}


void FindInverseInfo(word nInstruction,
		     const InverseOpcodeInfo& inverseInfo,
		     MapInstrInverseOpcodeInfo& mapInstrInverseOpcodeInfo)
{
  if (!(nInstruction & instrgNOT_INSTR))
    {
      InverseOpcodeInfo inverseInfo2 = inverseInfo;
      
      if ((inverseInfo2.m_nArguments & arg_OP_DEST_MASK) != 
	  arg_DEST_IS_MODRM_MODRM &&
	  (inverseInfo2.m_nArguments & arg_OP_SRC_MASK ) != 
	  arg_SRC_IS_MODRM_MODRM )
	inverseInfo2.m_nArguments &= ~(arg_MODRM_ONLY_M_MODRM | 
				       arg_MODRM_ONLY_R_MODRM); // in order not to shock OpcodeInfoSanityCheck
      
      inverseInfo2.m_nFlags |= OpsMayBeType (inverseInfo2.m_nArguments,
					     inverseInfo2.m_nExtra2);
      
      AddInverseInfo(nInstruction,inverseInfo2,mapInstrInverseOpcodeInfo);
      
      // now see if it can be base of lock/rep prefixes ?
      
      if ((inverseInfo2.m_nFlags & INVERSE_OPCODE_DEST_MAY_BE_MEM) && // tested first, as easier, and most likely false
	  IsValidLockInstruction (nInstruction)) // tested last 
	{
	  // if it may have lock prefix
	  // and can encode dest as memory reference 
	  
	  inverseInfo2.m_nFlags |= INVERSE_OPCODE_TAG_LOCK_REP_PREFIX;
	  inverseInfo2.m_nArguments |= arg_MODRM_ONLY_M_MODRM;
	  AddInverseInfo(nInstruction | instrgmINT_ARITH_MEMORY_LOCK,
			 inverseInfo2,mapInstrInverseOpcodeInfo);
	}
      else // can't have both lock and rep
	if ((nInstruction & instrgMASK) == instrgSTRING)
	  {
	    inverseInfo2.m_nFlags |= INVERSE_OPCODE_TAG_LOCK_REP_PREFIX;
	    
	    if (nInstruction == instrCMPS ||
		nInstruction == instrSCAS)
	      {
		AddInverseInfo(nInstruction | instrgmSTRING_REPNE,
			       inverseInfo2,mapInstrInverseOpcodeInfo);
		AddInverseInfo(nInstruction | instrgmSTRING_REPE,
			       inverseInfo2,mapInstrInverseOpcodeInfo);
	      }
	    else
	      AddInverseInfo(nInstruction | instrgmSTRING_REP,
			     inverseInfo2,mapInstrInverseOpcodeInfo);
	  }
    }
  else if ((nInstruction & instrmIS_GROUP_MASK) == instrgGROUP)
    {
      int nGroupType = nInstruction & instrmGROUP_TYPE_MASK;
      
      if ((nGroupType & instrmGROUP_MOD_BOTH) ||
	  (nGroupType & instrmGROUP_MOD_MEM))
	for (int n = 0; n < 8; n++)
	  {
	    InverseOpcodeInfo inverseInfo2 = inverseInfo;
	    
	    if (nGroupType & instrmGROUP_MOD_11)
	      inverseInfo2.m_nExtra2 = 0;
	    
	    const OpcodeInfo* pInfo = &infoGroups[inverseInfo.m_nExtra1][n];
	    
	    inverseInfo2.m_nGroupNNN = n;
	    inverseInfo2.m_nFlags |= 
	      ((nGroupType & instrmGROUP_MOD_BOTH) ? 
	       (INVERSE_OPCODE_TAG_GROUP_MOD_MEM | 
		INVERSE_OPCODE_TAG_GROUP_MOD_11) : 
	       INVERSE_OPCODE_TAG_GROUP_MOD_MEM);
	    
	    inverseInfo2.m_nArguments |= 
	      pInfo->m_nArguments | ((nGroupType & instrmGROUP_MOD_BOTH) ? 
				     0 : arg_MODRM_ONLY_M_MODRM);
	    inverseInfo2.m_nExtra1  = pInfo->m_nExtra1;
	    inverseInfo2.m_nExtra2 |= pInfo->m_nExtra2;
	    
	    FindInverseInfo(pInfo->m_nInstruction,inverseInfo2,
			    mapInstrInverseOpcodeInfo);
	  }	
      if (nGroupType & instrmGROUP_MOD_11)
	for (int n = 0; n < 8; n++)
	  {
	    InverseOpcodeInfo inverseInfo2 = inverseInfo;
	    
	    const OpcodeInfo* pInfo = &infoGroups[inverseInfo.m_nExtra2][n];
	    
	    inverseInfo2.m_nGroupNNN = n;
	    inverseInfo2.m_nFlags |= INVERSE_OPCODE_TAG_GROUP_MOD_11;
	    
	    inverseInfo2.m_nArguments |= 
	      pInfo->m_nArguments | arg_MODRM_ONLY_R_MODRM;
	    inverseInfo2.m_nExtra1  = pInfo->m_nExtra1;
	    inverseInfo2.m_nExtra2  = pInfo->m_nExtra2;
	    
	    FindInverseInfo(pInfo->m_nInstruction,inverseInfo2,
			    mapInstrInverseOpcodeInfo);
	  }	
    }
  else if (nInstruction == instrSUBGRP_COPROC)
    {
      for (int n = 0; n < 8; n++)
	{
	  InverseOpcodeInfo inverseInfo2 = inverseInfo;
	  
	  const OpcodeInfo* pInfo = &infoGroups[inverseInfo.m_nExtra1][n];
	  
	  inverseInfo2.m_nSubGroupN = n;
	  inverseInfo2.m_nFlags |= INVERSE_OPCODE_TAG_SUBGROUP;
	  
	  inverseInfo2.m_nArguments |= pInfo->m_nArguments;
	  inverseInfo2.m_nExtra1  = pInfo->m_nExtra1;
	  inverseInfo2.m_nExtra2 |= pInfo->m_nExtra2;
	  
	  FindInverseInfo(pInfo->m_nInstruction,
			  inverseInfo2,mapInstrInverseOpcodeInfo);
	}
    }
  else if(nInstruction == instrPRE_MMX_SSE )
    {
      int nOffset = (inverseInfo.m_nExtra1 << 1);
      int nNumPrefix = 
	((inverseInfo.m_nExtra2 & ext2_SSE_PREFIX_F3_F2) ? 4 : 2);
      
      for (int nPrefix = 0 ; nPrefix < nNumPrefix; nPrefix++)
	{
	  InverseOpcodeInfo inverseInfo2 = inverseInfo;
	  
	  const OpcodeInfo* pInfo = &infoMMX_SSE[nOffset+nPrefix];
	  
	  inverseInfo2.m_nSubGroupN = nPrefix;
	  inverseInfo2.m_nFlags |= INVERSE_OPCODE_TAG_MMX_SSE_PREFIX;
	  
	  inverseInfo2.m_nArguments = pInfo->m_nArguments;
	  inverseInfo2.m_nExtra1  = pInfo->m_nExtra1;
	  inverseInfo2.m_nExtra2  = pInfo->m_nExtra2;			
			
	  FindInverseInfo(pInfo->m_nInstruction,
			  inverseInfo2,mapInstrInverseOpcodeInfo);
	}
    }
  else if(nInstruction == instrPRE_MMX_SSE_MOD_MEM_11)
    {
      int nOffset = (inverseInfo.m_nExtra1 << 1);
      int nNumPrefix = 4;
      
      for (int nPrefix = 0 ; nPrefix < nNumPrefix; nPrefix++)
	{
	  InverseOpcodeInfo inverseInfo2 = inverseInfo;
	  
	  const OpcodeInfo* pInfo = &infoMMX_SSE[nOffset+nPrefix];
	  
	  inverseInfo2.m_nSubGroupN = nPrefix & 0x01;
	  inverseInfo2.m_nFlags |= INVERSE_OPCODE_TAG_MMX_SSE_PREFIX;
	  
	  inverseInfo2.m_nArguments = 
	    pInfo->m_nArguments | 
	    (nPrefix & 0x02 ? arg_MODRM_ONLY_R_MODRM : arg_MODRM_ONLY_M_MODRM);
	  inverseInfo2.m_nExtra1  = pInfo->m_nExtra1;
	  inverseInfo2.m_nExtra2  = pInfo->m_nExtra2;			
	  
	  FindInverseInfo(pInfo->m_nInstruction,
			  inverseInfo2,mapInstrInverseOpcodeInfo);
	}
    }
  else
    {
      assert (nInstruction == instrRESERVED ||
	      nInstruction == instrPREFIX ||
	      nInstruction == instrESCAPE2BYTE);
    }
}



void InstructionEncoder::InitInverseOpcodeInfo ()
{
  assert (!m_pInverseOpcodeInfo);
  
  MapInstrInverseOpcodeInfo mapInstrInverseOpcodeInfo;
  
  InverseOpcodeInfo inverseInfo;
  
  inverseInfo.m_nGroupNNN = inverseInfo.m_nSubGroupN = 0;
  
  for (int nTwoByte = 0; nTwoByte < 2; nTwoByte++)
    {
      inverseInfo.m_nFlags = (nTwoByte ? INVERSE_OPCODE_TAG_2_BYTE : 0);
      
      for (int nByte = 0; nByte < 0x100; nByte++)
	{
	  inverseInfo.m_nByte = nByte;
	  
	  const OpcodeInfo* pInfo = &infoByte[nTwoByte][nByte];
	  
	  inverseInfo.m_nArguments = pInfo->m_nArguments;
	  inverseInfo.m_nExtra1 = pInfo->m_nExtra1;
	  inverseInfo.m_nExtra2 = pInfo->m_nExtra2;
	  
	  FindInverseInfo(pInfo->m_nInstruction,inverseInfo,mapInstrInverseOpcodeInfo);
	}
    }
  
  //////////////////////////////////////////////
  //////////////////////////////////////////////
  //////////////////////////////////////////////

//	unsigned int nNum = 0; 
/*
	for (j = 0; j < 2; j++)
		for (i = 0; i < sizeof (infoByte[0]) / sizeof (infoByte[0][0]); i++)
		{
			if (!(infoByte[j][i].m_nInstruction & instrgNOT_INSTR))
				nNum++;
			else if ((infoByte[j][i].m_nInstruction & instrmIS_GROUP_MASK) == instrgGROUP)
				nNum += 8; // this will overcount, (as we count instructions when in a group (for loop below), but is necessary when group has several access bytes
		}

	assert (sizeof (infoGroups[0]) / sizeof (infoGroups[0][0]) == 8);
	for (i = 0; i < sizeof (infoGroups) / sizeof (infoGroups[0]); i++)
		for (j = 0; j < 8; j++)
			if (!(infoGroups[i][j].m_nInstruction & instrgNOT_INSTR))
				nNum++;

	for (i = 0; i < sizeof (infoMMX_SSE) / sizeof (infoMMX_SSE[0]); i++)
		if (!(infoMMX_SSE[i].m_nInstruction & instrgNOT_INSTR))
			nNum++;
*/
//	assert (nNum > 1); // or new[] (at least msvc) get into trouble ?
/*
	unsigned int i, j;

	MapInstrInverseOpcodeInfo mapInstrInverseOpcodeInfo;

	// int nInsert = 0;

	InverseOpcodeInfo groupInfo[sizeof (infoGroups) / sizeof (infoGroups[0])];
	InverseOpcodeInfo mmxsseInfo[sizeof (infoMMX_SSE) / (2*sizeof (infoMMX_SSE[0]))];

	memset (groupInfo,0,sizeof (groupInfo));
	memset (mmxsseInfo,0,sizeof (mmxsseInfo));

	MapGroupExtraAccessBytes mapGroupExtraAccessBytes;

	for (j = 0; j < 2; j++)
	{
		byte nTag2ByteFlag = (j ? INVERSE_OPCODE_TAG_2_BYTE : 0);

		for (i = 0; i < sizeof (infoByte[0]) / sizeof (infoByte[0][0]); i++)
		{
			const OpcodeInfo* pInfo = &infoByte[j][i];

			if (!(pInfo->m_nInstruction & instrgNOT_INSTR))
			{
				SetupInverseInsert(mapInstrInverseOpcodeInfo,
					pInfo->m_nInstruction,
					(byte) i,
					nTag2ByteFlag | OpsMayBeType(pInfo->m_nArguments,pInfo->m_nExtra2));
			}
			else if (pInfo->m_nInstruction & instrgGROUP)
				SetGroupInfo (pInfo,groupInfo,mapGroupExtraAccessBytes,mmxsseInfo,i,nTag2ByteFlag);
		}
	}

	for (i = 0; i < sizeof (infoGroups) / sizeof (infoGroups[0]); i++)
		for (j = 0; j < 8; j++)
		{
			const OpcodeInfo* pInfo = &infoGroups[i][j];

			if (!(pInfo->m_nInstruction & instrgNOT_INSTR))
			{
				InverseOpcodeInfo* pGroupInfo = &groupInfo[i];

				assert (pGroupInfo->m_nFlags); 

				SetupInverseInsert(mapInstrInverseOpcodeInfo,
					mapGroupExtraAccessBytes,
					pInfo->m_nInstruction,
					pGroupInfo,
					i,j,
					OpsMayBeType(pInfo->m_nArguments | OnlyMOrRFromMod(pGroupInfo->m_nFlags)));


//				pSetupInsert->m_nInstruction = pInfo->m_nInstruction;
//				pSetupInsert->m_info = *pGroupInfo; // set byte, flags and groupNNN (if subgroup)
////				m_pInverseOpcodeInfo[nNum].m_info.m_nByte = pGroupInfo->m_nByte;
//				if (pGroupInfo->m_nFlags & INVERSE_OPCODE_TAG_SUBGROUP)
//				{
//					pSetupInsert->m_info.m_nSubGroupN = j;
////					m_pInverseOpcodeInfo[nNum].m_info.m_nGroupNNN = pGroupInfo->m_nGroupNNN;
//				}
//				else
//					pSetupInsert->m_info.m_nGroupNNN = j;
////				m_pInverseOpcodeInfo[nNum].m_info.m_nFlags = pGroupInfo->m_nFlags;
//
//				MapGroupExtraAccessBytes::iterator extraAccessByte = mapGroupExtraAccessBytes.find(i);
//
//				for ( ; extraAccessByte != mapGroupExtraAccessBytes.end() && extraAccessByte->first == i; ++extraAccessByte)
//				{
//					*(pSetupInsert+1) = *pSetupInsert;
//					pSetupInsert++;
//					pSetupInsert->m_info.m_nByte = extraAccessByte->second;
//				}
//
//				pSetupInsert++;

			}
			else if (pInfo->m_nInstruction & instrgGROUP)
				SetGroupInfo (pInfo,groupInfo,mapGroupExtraAccessBytes,mmxsseInfo,i,INVERSE_OPCODE_TAG_SUBGROUP,j);
		}

	for (i = 0; i < sizeof (infoMMX_SSE) / sizeof (infoMMX_SSE[0]); i++)
	{
		const OpcodeInfo* pInfo = &infoMMX_SSE[i];
		
		if (!(pInfo->m_nInstruction & instrgNOT_INSTR))
		{
			InverseOpcodeInfo* pMMXSSEInfo = &mmxsseInfo[i >> 1];

			assert (pMMXSSEInfo->m_nByte || pMMXSSEInfo->m_nFlags); // well opcode 0 could be a group, but it isn't, so this check is a valid test if we know the opcode

#ifndef NDEBUG
			if (i > ext_MMX_SSE_17*2 && 
				pMMXSSEInfo->m_nFlags & (INVERSE_OPCODE_TAG_GROUP_MOD_11 | INVERSE_OPCODE_TAG_GROUP_MOD_MEM))
			{
				assert ((pInfo->m_nArguments & arg_OP_DEST_MASK) != arg_DEST_IS_MODRM_REG); // as group got NNN from REG, we can't get register from there too
				assert ((pInfo->m_nArguments & arg_OP_SRC_MASK ) != arg_SRC_IS_MODRM_REG ); // as group got NNN from REG, we can't get register from there too
			}
#endif

			InverseOpcodeInfo info;

			info = *pMMXSSEInfo; // set byte, flags and groupNNN (if subgroup)
			info.m_nSubGroupN |= (i & 1); // i has low bit of needed Prefix (bit 2 was set when pGroupInfo initialized)
			info.m_nFlags |= OpsMayBeType(pInfo->m_nArguments | OnlyMOrRFromMod(pMMXSSEInfo->m_nFlags));

			mapInstrInverseOpcodeInfo.insert(MapInstrInverseOpcodeInfo::value_type(pInfo->m_nInstruction,info));
		}
	}
*/
  assert (sizeof (infoMMX_SSE) / sizeof (infoMMX_SSE[0]) == 
	  2*ext_MMX_SSE_NUM_GROUPS);

  //////////////////////////////////
		
  // so now they are sorted in pSetupInverseOpcodeInfo insert into
  // m_pInverseOpcodeInfo, count number of different and generate
  // m_pInverseOpcodeInfoIndex

  int nNum = mapInstrInverseOpcodeInfo.size();
  
  m_pInverseOpcodeInfo = new InverseOpcodeInfo[nNum];
  int nDifferent = 0;
  int nLastInstruction = 0;
  
  InverseOpcodeInfo* pInsert = m_pInverseOpcodeInfo;
  
  FOR_EACH_CONST_ITER(MapInstrInverseOpcodeInfo,mapInstrInverseOpcodeInfo,iter)
    {
      *(pInsert++) = iter->second;
      if (iter->first != nLastInstruction)
	{
	  assert (iter->first > nLastInstruction);
	  nLastInstruction = iter->first;
	  nDifferent++;
	}
      else
	{
	  assert (iter != mapInstrInverseOpcodeInfo.begin());
	  (pInsert-2)->m_nFlags |= INVERSE_OPCODE_TAG_HAS_MORE_VARIANTS;
	}
    }
  
  m_pInverseOpcodeInfoIndex = 
    new InverseOpcodeInfoIndex[m_nInverseOpcodeInfoIndex = nDifferent];
  
  InverseOpcodeInfoIndex* pInsertIndex = m_pInverseOpcodeInfoIndex;
  
  nLastInstruction = 0;
  int i = 0;
  FOR_EACH_CONST_ITER(MapInstrInverseOpcodeInfo,
		      mapInstrInverseOpcodeInfo,iter2)
    {
      if (iter2->first != nLastInstruction)
	{
	  nLastInstruction = iter2->first;
	  pInsertIndex->m_nInstruction = iter2->first;
	  pInsertIndex->m_nIndex = (word) i;
	  pInsertIndex++;
	}
      i++;
    }
  
  assert (pInsertIndex-m_pInverseOpcodeInfoIndex == m_nInverseOpcodeInfoIndex);
  
	//////////////////////////////////

/* FAST_ENCODER_STUFF:
 *
        assert (instrCMP - instrADD + 1 == 8); 

        for (i = 0; i < 9; i++)
        {
                InverseOpcodeInfo* pInfo = FindFirstInverseOpcodeInfo((i == 8) ? instrMOV : i + instrADD);
                assert (pInfo);

                pInfo->m_nSubGroupN |= 2 << INVERSE_OPCODE_QUICK_ENCODER_N_SHIFT;
                pInfo->m_nFlags |= INVERSE_OPCODE_TAG_HAS_QUICK_ENCODER;
        }

        assert (instrDEC == instrINC + 1);
        assert (instrPOP == instrPUSH + 1);
                
        for (i = 0; i < 4; i++)
        {
                InverseOpcodeInfo* pInfo = FindFirstInverseOpcodeInfo(i < 2 ? (i + instrPUSH) : (i-2+instrINC));
                assert (pInfo);

                pInfo->m_nSubGroupN |= 1 << INVERSE_OPCODE_QUICK_ENCODER_N_SHIFT;
                pInfo->m_nFlags |= INVERSE_OPCODE_TAG_HAS_QUICK_ENCODER;
        }
        */
}

InverseOpcodeInfo* InstructionEncoder::FindFirstInverseOpcodeInfo (word nInstr)
{
  InverseOpcodeInfoIndex* pIndex = 
    (InverseOpcodeInfoIndex*) bsearch(&nInstr,
				      m_pInverseOpcodeInfoIndex,
				      m_nInverseOpcodeInfoIndex,
				      sizeof (m_pInverseOpcodeInfoIndex[0]),
				      compare_word);
  
  if (!pIndex)
    return NULL;
  
  return m_pInverseOpcodeInfo+pIndex->m_nIndex;
  /*
	int nFirstPossible = 0;
	int nLastPossible = m_nInverseOpcodeInfo-1;

	while (nFirstPossible <= nLastPossible)
	{	
		int nTry = (nFirstPossible + nLastPossible) >> 1;

		if (m_pInverseOpcodeInfo[nTry].m_nInstruction > nInstr)
			nLastPossible = nTry - 1;
		else if (m_pInverseOpcodeInfo[nTry].m_nInstruction < nInstr)
			nFirstPossible = nTry + 1;
		else
		{
			assert (m_pInverseOpcodeInfo[nTry].m_nInstruction == nInstr);
			if (nTry == 0 ||
				m_pInverseOpcodeInfo[nTry-1].m_nInstruction < nInstr)
				return &m_pInverseOpcodeInfo[nTry];

			assert (m_pInverseOpcodeInfo[nTry-1].m_nInstruction == nInstr);

			nLastPossible = nTry - 1;
		}
	}
	return NULL;
*/
}

//////////////////////////////////////////////////////////////////////

/*
struct tagInverseOpcodeInfo
{
	word m_nInstruction;

	byte m_nByte;
	byte m_nTagAndNNN;
} InverseOpcodeInfo;

#define INVERSE_OPCODE_NNN_MASK			0x07
#define INVERSE_OPCODE_2_BYTE			0x08
#define INVERSE_OPCODE_GROUP_MOD_11		0x01
#define INVERSE_OPCODE_GROUP_MOD_MEM	0x02
*/
/*
extern const OpcodeInfo infoByte1[];
extern const OpcodeInfo infoByte2[];
extern const OpcodeInfo infoGroups[16][8];
*/



/*
void OutputOpcode (DataDestPtr& dataPtr,byte bOpCode)
{
	dataPtr.WriteByte (bOpCode);
}
*/
/*
void OutputModRM_ModRegRM (DataDestPtr& dataPtr,byte bMod,byte bReg,byte bRM)
{
	dataPtr.WriteByte ((bMod << 6) | (bReg << 3) | bRM);
}

void OutputModRM_RegRM (DataDestPtr& dataPtr,byte bReg,byte bRM)
{
	dataPtr.WriteByte (0xc0 | (bReg << 3) | bRM);
}
*/




/*
void OutputModRM_MemReg (DataDestPtr& dataPtr,const OperandMemory* pOpMem,byte bReg,bool bAddrSize32)
{
	byte bBaseRegType = pOpMem->m_bBaseReg & reggTYPE_MASK;
	byte bIndexRegType = pOpMem->m_bIndexReg & reggTYPE_MASK;

	assert (!bBaseRegType || !bIndexRegType || (bBaseRegType == bIndexRegType)); // mixed address register sizes
	
	byte bRegType = bBaseRegType | bBaseRegType;

	assert (bRegType == reggSTD16 || bRegType == reggSTD32 || bRegType == 0); // address register size not 16 or 32

	assert (bAddrSize32 == (bRegType == reggSTD32) || // correct size
		(bRegType == 0 && (pOpMem->DisplacementSize () == (bAddrSize32 ? 4 : 2)))); // or only displacement

	if (bAddrSize32)
	{
		// these asserts should catch all invalid address modes
		// later asserts should just tell that inappropriate mod/rm (sib) bytes are about to be used

		assert (pOpMem->DisplacementSize () == 0 ||
			pOpMem->DisplacementSize () == 1 ||
			pOpMem->DisplacementSize () == 4);

		assert (pOpMem->GetScale () || 
			( ((pOpMem->m_bBaseReg != regESP || pOpMem->m_bIndexReg != regESP)) &&
			( (pOpMem->DisplacementSize () == 0 && ((pOpMem->m_bBaseReg && pOpMem->m_bBaseReg != regEBP) || (pOpMem->m_bIndexReg && pOpMem->m_bIndexReg != regEBP)) ) ||
			  (pOpMem->DisplacementSize () == 1 && (pOpMem->m_bBaseReg || pOpMem->m_bIndexReg) ) ||
			  (pOpMem->DisplacementSize () == 4) )));

		assert (!pOpMem->GetScale () ||
			((pOpMem->m_bIndexReg && pOpMem->m_bIndexReg != regESP)) &&
			((pOpMem->DisplacementSize () == 0 && (pOpMem->m_bBaseReg && pOpMem->m_bBaseReg != regEBP)) ||
			 (pOpMem->DisplacementSize () == 1 && pOpMem->m_bBaseReg) ||
			 (pOpMem->DisplacementSize () == 4)));


		// 32 bit address

		byte bMod = 0;
		byte bRM = 0;

		assert (!pOpMem->GetScale () || pOpMem->m_bIndexReg || pOpMem->m_bBaseReg);
 
                assert (pOpMem->DisplacementSize () == 0 ||
                        pOpMem->DisplacementSize () == 1 ||
                        pOpMem->DisplacementSize () == 4);

                if (pOpMem->DisplacementSize () == 1)
                        bMod = 1;
                else if (pOpMem->DisplacementSize () == 4)
                {
                        if (pOpMem->m_bBaseReg || pOpMem->m_bIndexReg)
                                bMod = 2;
                        else
                                bRM = 5; // bMod already 0
                }

                //bool bNeedSib = false;

                if ((!pOpMem->m_bIndexReg || !pOpMem->m_bBaseReg) && !pOpMem->GetScale ())
                {
                        // only one reg and no scale

                        if (!pOpMem->m_bIndexReg && !pOpMem->m_bBaseReg)
                        {
                                assert (pOpMem->DisplacementSize () == 4);

                                assert (bMod == 0);
                                assert (bRM == 5);
                        }
                        else
                        {
                                byte bRegName = (pOpMem->m_bIndexReg | pOpMem->m_bBaseReg) & reggNAME_MASK;
                                
                                assert (((pOpMem->m_bIndexReg | pOpMem->m_bBaseReg) & reggTYPE_MASK) == reggSTD32);

                                assert (bRegName != 5 || pOpMem->DisplacementSize () != 0);

                                bRM = bRegName; // if bRegName == 4 => need sib-byte
                        }
                }
                else
                        bRM = 4;

                dataPtr.WriteByte(bMod | (bReg << 3) | bRM);

                if (bRM == 4)
                {
                        byte bIndex = pOpMem->m_bIndexReg;
                        byte bBase = pOpMem->m_bBaseReg;

                        assert (bBase || bIndex);  // otherwise sib shouldn't be needed

                        assert (((bBase | bIndex) & reggTYPE_MASK) == reggSTD32);

                        if (!bBase && !pOpMem->GetScale ())
                        {
                                assert (bIndex);

                                bBase = bIndex;
                                bIndex = 0;
                        }

                        if ((bBase == regEBP && !pOpMem->GetScale () && !pOpMem->DisplacementSize ()) ||
                                bIndex == regESP)
                                swap (bBase,bIndex);

                        bBase &= reggNAME_MASK;
                        bIndex &= reggNAME_MASK;

                        assert (bIndex != 4 || !pOpMem->GetScale ());
                        
                        dataPtr.WriteByte((pOpMem->GetScale () << 6) | (bIndex << 3) | bBase);
                }


                // actually, sometimes, when only base (or index with scale 1 used), or no displacement
                // we could increase instruction size by 1 and/or 1, 3 or 4 bytes (by issuing unnecessary sib and displacement of 1 or 4 bytes being 0, or enlarge disp of 1 to 4)
        }
        else
        {
                // 16 bit address

                byte bMod = 0;
                byte bRM = 0;

                assert (pOpMem->GetScale () == 0);		

                assert (pOpMem->DisplacementSize () == 0 ||
                        pOpMem->DisplacementSize () == 1 ||
                        pOpMem->DisplacementSize () == 2);

                byte bIndex = pOpMem->m_bIndexReg;
                byte bBase = pOpMem->m_bBaseReg;

                if (pOpMem->DisplacementSize () == 1)
                        bMod = 1;
                else if (pOpMem->DisplacementSize () == 2)
                {
                        if (bBase || bIndex)
                                bMod = 2;
                        else
                                bRM = 6; // bMod already 0
                }

                if (bIndex && !bBase)
                {
                        bBase = bIndex;
                        bIndex = 0;
                }
                switch (((word) bBase << 8) | bIndex)
                {
                case ((((word) regSI) << 8) | regBX):
                case ((((word) regBX) << 8) | regSI): bRM = 0; break;
                case ((((word) regDI) << 8) | regBX):
                case ((((word) regBX) << 8) | regDI): bRM = 1; break;
                case ((((word) regSI) << 8) | regBP):
                case ((((word) regBP) << 8) | regSI): bRM = 2; break;
                case ((((word) regDI) << 8) | regBP):
                case ((((word) regBP) << 8) | regDI): bRM = 3; break;

                case (((word) regSI)):
                case (((word) regSI) << 8): bRM = 4; break;
                case (((word) regDI)):
                case (((word) regDI) << 8): bRM = 5; break;
                case (((word) regBP)):
                case (((word) regBP) << 8): 
                        assert (pOpMem->DisplacementSize () != 0);
                        bRM = 6; 
                        break;
                case (((word) regBX)):
                case (((word) regBX) << 8): bRM = 7; break;
                
                case 0:
                        assert (pOpMem->DisplacementSize () == 2);
                        break;

                default:
                        assert (false); // base/index combination impossible to encode, and should not have been set anywhere
                        break;
                }

                dataPtr.WriteByte((bMod << 6) | (bReg << 3) | bRM);

                // actually, sometimes when no displacement in use
                // we could increase instruction size by 1 or 2 bytes (by issuing unnecessary displacement of 1 or 2 bytes being 0, or enlarge disp from 1 to 2)
        }

        // output displacement

        if (pOpMem->DisplacementSize ())
                OutputDisplacement (dataPtr,pOpMem);
}
*/

/* FAST_ENCODER_STUFF:
 *
// If we have special encoders for arithmetic, mov, push/pop, we should probably
// also have it for control-transfer

// The fast encoders do not take care of segment overrides properly
//
// If it's determined that much execution time is spent encoding affected instructions,
// the uncomments this stuff again and fix it
// and some code in other places to... Marked by FAST_ENCODER_STUFF

bool InstructionEncoder::FastStd2Op (const Instruction* pInstr,DataDestPtr& dataPtr) const
{
	// the code assume that the order of any Prefixes does not affect
	// execution speed on any processor
	// if it does, rewrite! to take that into account

	// this should always return shortest possible form
	// however, it is allowed to fail on strange operands
	// (then try the general encoder)

	assert (pInstr->m_nInstruction == instrMOV ||
		pInstr->m_nInstruction == instrADD ||
		pInstr->m_nInstruction == instrOR ||
		pInstr->m_nInstruction == instrADC ||
		pInstr->m_nInstruction == instrSBB ||
		pInstr->m_nInstruction == instrAND ||
		pInstr->m_nInstruction == instrSUB ||
		pInstr->m_nInstruction == instrXOR ||
		pInstr->m_nInstruction == instrCMP);

	const Operand* pOpDest = pInstr->GetDestOp ();
	const Operand* pOpSrc = pInstr->GetSrcOp ();
	assert (pOpDest);
	assert (pOpSrc);

	byte bMemoryDestOpcodeFlag = 2; // this is NOT boolean, it is for encoding...

	if (typeid (pOpSrc) == typeid (OperandRegister) &&
		typeid (pOpDest) == typeid (OperandMemory))
	{
		swap (pOpDest,pOpSrc);

		bMemoryDestOpcodeFlag = 0;
	}

	byte bSize16_32 = 1;

	int nDestSize = pOpDest->GetOpSize ();

	if (nDestSize == 1)
		bSize16_32 = 0;
	else
	{
		if (nDestSize != 2 && nDestSize != 4)
			return false;
		
		if ((nDestSize == 4) != m_bCodeSize32)
			OutputOpSizePrefix (dataPtr);
	}

	if (typeid (pOpDest) == typeid (OperandRegister))
	{
		byte bDestReg = ((OperandRegister*) pOpDest)->m_bReg;
		byte bDestRegType = bDestReg & reggTYPE_MASK;

		assert (bDestRegType == reggSTD8 ||
			bDestRegType == reggSTD16 ||
			bDestRegType == reggSTD32);

		if (typeid (pOpSrc) == typeid (OperandRegister))
		{
			byte bSrcReg = ((OperandRegister*) pOpSrc)->m_bReg;
			byte bSrcRegType = bSrcReg & reggTYPE_MASK;

			if (bDestRegType != bSrcRegType) // same size
				return false;

			if (pInstr->m_nInstruction == instrMOV)
				OutputOpcode (dataPtr,0x88 | bSize16_32);
			else
			{
				assert ((instrADD & 0x07) == 0x00); // make sure the codes internal to this prgm are what we assume
				assert ((instrCMP & 0x07) == 0x07);
				
				OutputOpcode (dataPtr,((pInstr->m_nInstruction & 0x07) << 3) | bSize16_32);
			}
			OutputModRM_RegRM (dataPtr,bSrcReg & reggNAME_MASK,bDestReg & reggNAME_MASK);
			return true;
		}
		else if (typeid (pOpSrc) == typeid (OperandMemory))
		{
			if (((const OperandMemory*) pOpSrc)->GetOpSize () != 
				((const OperandRegister*) pOpDest)->GetOpSize ())
				return false;

			bool bAddrSize32 = m_bCodeSize32;

			if (IsAddrSize32 ((OperandMemory*) pOpSrc) != bAddrSize32)
			{
				OutputAddrSizePrefix (dataPtr);
				bAddrSize32 = !bAddrSize32;
			}

			if (pInstr->m_nInstruction == instrMOV &&
				!(bDestReg & reggNAME_MASK) &&  // AL/eAX
				IsPureDisplacement((const OperandMemory*) pOpSrc))
			{
				OutputOpcode (dataPtr,0xA0 | (0x02 ^ bMemoryDestOpcodeFlag) | bSize16_32);
				OutputDisplacement (dataPtr,(const OperandMemory*) pOpSrc);
			}
			else
			{
				if (pInstr->m_nInstruction == instrMOV)
					OutputOpcode (dataPtr,0x88 | bSize16_32);
				else
				{
					assert ((instrADD & 0x07) == 0x00);
					assert ((instrCMP & 0x07) == 0x07);
					
					OutputOpcode (dataPtr,((pInstr->m_nInstruction & 0x07) << 3) | bMemoryDestOpcodeFlag | bSize16_32);
				}
				
				OutputModRM_MemReg (dataPtr,(OperandMemory*) pOpSrc,bDestReg & reggNAME_MASK,bAddrSize32);
			}
			return true;
		}
		else if (typeid (pOpSrc) == typeid (OperandImmediate))
		{
			if (pInstr->m_nInstruction == instrMOV)
			{
				if (((const OperandImmediate*) pOpSrc)->GetOpSize () != 
					((const OperandRegister*) pOpDest)->GetOpSize ())
					return false;

				OutputOpcode (dataPtr, 0xB0 | (bSize16_32 << 3) | (bDestReg & reggNAME_MASK));

				OutputImmediate (dataPtr,(const OperandImmediate*) pOpSrc);
				return true;
			}
			else if (!(bDestReg & reggNAME_MASK)) // AL/eAX
			{
				if (((const OperandImmediate*) pOpSrc)->GetOpSize () != 
					((const OperandRegister*) pOpDest)->GetOpSize ())
					return false;

				OutputOpcode (dataPtr,((pInstr->m_nInstruction & 0x07) << 3) | 0x04 | bSize16_32);
				OutputImmediate (dataPtr,(const OperandImmediate*) pOpSrc);
				return true;
			}
		}
	}

	if (typeid (pOpSrc) == typeid (OperandImmediate) &&
		(typeid (pOpDest) == typeid (OperandRegister) ||
		 typeid (pOpDest) == typeid (OperandMemory)))
	{
		if (((const OperandImmediate*) pOpSrc)->GetOpSize () != 
			((const OperandRegister*) pOpDest)->GetOpSize ())
		{
			if (pInstr->m_nInstruction == instrMOV ||
				((const OperandImmediate*) pOpSrc)->GetOpSize () != 1)
				return false;
		}

		bool bAddrSize32 = m_bCodeSize32;

		if (typeid (pOpDest) == typeid (OperandMemory) &&
			IsAddrSize32 ((OperandMemory*) pOpSrc) != bAddrSize32)
		{
			OutputAddrSizePrefix (dataPtr);
			bAddrSize32 = !bAddrSize32;
		}

		byte bOpCode;
		byte bNNN;

		if (pInstr->m_nInstruction == instrMOV)
		{
			bOpCode = 0xC3;
			bNNN = 0;
		}
		else
		{
			bNNN = (pInstr->m_nInstruction & 0x07);
			if (((const OperandImmediate*) pOpSrc)->GetOpSize () == 1 && bSize16_32)
				bOpCode = 0x82; // sign extend
			else
				bOpCode = 0x80;
		}	
			
		OutputOpcode (dataPtr,bOpCode | bSize16_32);

		if (typeid (pOpDest) == typeid (OperandRegister))
			OutputModRM_RegRM (dataPtr,bNNN,((const OperandRegister*) pOpDest)->m_bReg & reggNAME_MASK);
		else
			OutputModRM_MemReg (dataPtr,(OperandMemory*) pOpDest,bNNN,bAddrSize32);
		
		OutputImmediate (dataPtr,(const OperandImmediate*) pOpSrc);
		return true;
	}

	// XXX   reg,reg       (x)
	// XXX   reg,mem       (x)
	// XXX   reg,imm       
	// XXX   mem,reg       (x)
	// XXX   mem,imm
	// XXX   AL/AX/EAX,imm (x)

	return false;
}
*/
/* FAST_ENCODER_STUFF:
 *
bool InstructionEncoder::FastStd1Op (const Instruction* pInstr,DataDestPtr& dataPtr) const
{
	// the code assume that the order of any Prefixes does not affect
	// execution speed on any processor
	// if it does, rewrite! to take that into account

	// this should always return shortest possible form
	// however, it is allowed to fail on strange operands
	// (then try the general encoder)

	assert (pInstr->m_nInstruction == instrPUSH ||
		pInstr->m_nInstruction == instrPOP ||
		pInstr->m_nInstruction == instrINC ||
		pInstr->m_nInstruction == instrDEC);

	const Operand* pOp = pInstr->GetDestOp ();
	assert (pOp);

	if (typeid (*pOp) == typeid (OperandRegister))
	{
		byte bReg = ((OperandRegister*) pOp)->m_bReg;
		byte bRegType = bReg & reggTYPE_MASK;

		if (bRegType == reggSTD16 || 
			bRegType == reggSTD32)
		{
			if ((bRegType == reggSTD32) != m_bCodeSize32)
				OutputOpSizePrefix (dataPtr);

			byte bOpCode;

			switch (pInstr->m_nInstruction)
			{
			case instrINC:  bOpCode = 0x40; break;
			case instrDEC:  bOpCode = 0x48; break;
			case instrPUSH: bOpCode = 0x50; break;
			case instrPOP:  bOpCode = 0x58; break;
			}

			assert ((bReg & reggNAME_MASK) < 8);
			OutputOpcode (dataPtr,bOpCode | (bReg & reggNAME_MASK));
			return true;
		}
		else if (bRegType == reggSTD8)
		{
			byte bOpCode;

			switch (pInstr->m_nInstruction)
			{
			case instrINC:  bOpCode = 0xC0; break;
			case instrDEC:  bOpCode = 0xC8; break;
			case instrPUSH: 
			case instrPOP:  return false;
			}

			OutputOpcode (dataPtr,0xFE);
			OutputOpcode (dataPtr,bOpCode | (bReg & reggNAME_MASK));
			return true;
		}
	}
	else if (typeid (*pOp) == typeid (OperandMemory))
	{
		byte bOpSize = ((const OperandImmediate*) pOp)->GetOpSize ();

		if (bOpSize == 2 || bOpSize == 4)
		{
			if ((bOpSize == 4) != m_bCodeSize32)
				OutputOpSizePrefix (dataPtr);

			byte bOpCode, bNNN;

			switch (pInstr->m_nInstruction)
			{
			case instrINC:  bOpCode = 0xFF; bNNN = 0; break;
			case instrDEC:  bOpCode = 0xFF; bNNN = 1;  break;
			case instrPUSH: bOpCode = 0xFF; bNNN = 6;  break;
			case instrPOP:  bOpCode = 0x8F; bNNN = 0;  break;
			}

			OutputOpcode (dataPtr,bOpCode);
			OutputModRM_MemReg (dataPtr,(const OperandMemory*) pOp,bNNN,m_bCodeSize32);
			return true;
		}
		else if (bOpSize == 1)
		{
			byte bNNN;

			switch (pInstr->m_nInstruction)
			{
			case instrINC:  bNNN = 0; break;
			case instrDEC:  bNNN = 1; break;
			case instrPUSH: 
			case instrPOP:  return false;
			}

			OutputOpcode (dataPtr,0xFE);
			OutputModRM_MemReg (dataPtr,(const OperandMemory*) pOp,bNNN,m_bCodeSize32);
			return true;
		}
	}
	else if (typeid (*pOp) == typeid (OperandImmediate))
	{
		if (pInstr->m_nInstruction == instrPUSH)
		{
			byte bOpSize = ((const OperandImmediate*) pOp)->GetOpSize ();

			byte bOpCode;

			switch (bOpSize)
			{
			case 1:
				bOpCode = 0x6A;
				break;
			case 2:
			case 4:
				if ((bOpSize == 4) != m_bCodeSize32)
					OutputOpSizePrefix (dataPtr);
				bOpCode = 0x68;
				break;
			default:
				return false;
			}
			OutputOpcode (dataPtr,bOpCode);
			OutputImmediate (dataPtr,(const OperandImmediate*) pOp);
			return true;
		}
	}

	return false;
}
*/

/*
bool ValidateAddressSize (bool bAddrSize32,PrefixDecision& prefix)
{
	if (prefix.bDecided & PD_ADDRSIZE32)
		return bAddrSize32 == !!(prefix.bSet & PD_ADDRSIZE32);

	prefix.bSet |= (bAddrSize32 ? PD_ADDRSIZE32 : 0);
	prefix.bDecided |= PD_ADDRSIZE32;
	return true;
}
*/


//////////////////////////////////////////////////////////////////////

#ifndef NDEBUG
bool InstructionEncoder::CheckFastEncoder (const Instruction* pInstr,
					   DataDestPtr& dataPtr,
					   word* pDispImmLocs,
					   const void* m_pStart) const
{
  byte cNewCode[MAX_INSTR_CODE_LENGTH]; // four Prefixes, two opcodes, mod/rm, sib, displacement, immediate
  
  DataDestPtr newPtr(cNewCode
#ifndef NDEBUG
		     ,cNewCode+sizeof (cNewCode)
#endif
		     );
  
  verify (Encode (pInstr,newPtr,pDispImmLocs,false));
  
  // this should always be true, as Encode is to return the shortest code,
  // and so if the fast encoders
  
  if ((const byte*) newPtr.m_pCurrent - cNewCode != 
      (const byte*) dataPtr.m_pCurrent - (const byte*) m_pStart)
    return false;
  
  // this should also work, as either the encoding of a certain length
  // is unique or both encoders should default to the same behavior
  // (non-used bits set to same value)
  
  return memcmp (cNewCode,m_pStart,(byte*) newPtr.m_pCurrent - cNewCode) == 0;
}
#endif

bool InstructionEncoder::Encode (const Instruction* pInstr,
				 DataDestPtr& dataPtr,
				 word* pDispImmLocs
				 IFDEBUGPARAM(bool /* bUseFastEncoderIfPossible */)
				 ) const
{
  const InverseOpcodeInfo* inverseOpcodeInfo = 
    FindFirstInverseOpcodeInfo (pInstr->m_nInstruction);
  
  assert (inverseOpcodeInfo); // or instruction completely unknown

/* FAST_ENCODER_STUFF:
 *
        if (inverseOpcodeInfo->m_nFlags & INVERSE_OPCODE_TAG_HAS_QUICK_ENCODER)
        {
                // try quick encoder first
#ifndef NDEBUG
		if (bUseFastEncoderIfPossible)
		{
#endif

		void* pOldCur = dataPtr.GetCurrent();
	
		switch (inverseOpcodeInfo->m_nSubGroupN >> INVERSE_OPCODE_QUICK_ENCODER_N_SHIFT)
		{
		case 1:
			if (FastStd1Op(pInstr,dataPtr))
			{
				assert (CheckFastEncoder (pInstr,dataPtr,pOldCur));
				return true;
			}
			break;
		case 2:
			if (FastStd2Op(pInstr,dataPtr))
			{
				assert (CheckFastEncoder (pInstr,dataPtr,pOldCur));
				return true;
			}
			break;			
		}

		dataPtr.SetCurrent (pOldCur);

#ifndef NDEBUG
		}
#endif
	}
	*/

  void* pDataBegin = dataPtr.GetCurrent ();
  
  TryDoInstructionEncoding trydo(m_bCodeSize32);
  
  TRY_DO_INSTR_ENCODE_PREPARE(pInstr);
  
  for ( ; ; )
    {
      if (TRY_DO_INSTR_ENCODE_MAY_ATTEMPT(inverseOpcodeInfo) &&
	  trydo.TryEncoding (inverseOpcodeInfo,pInstr))
	{
	  if (dataPtr.GetCurrent () == pDataBegin || // no valid opcode found so far
	      trydo.GetShortestLength() < 
	      (byte*) dataPtr.m_pCurrent - (byte*) pDataBegin) // shorter
	    {
	      dataPtr.SetCurrent(pDataBegin);
	      trydo.DoEncoding(dataPtr,pDispImmLocs);
	    }
	}
      
      if (!(inverseOpcodeInfo->m_nFlags & 
	    INVERSE_OPCODE_TAG_HAS_MORE_VARIANTS))
	break;
      inverseOpcodeInfo++;
    }
  
  assert (dataPtr.GetCurrent () >= pDataBegin);
  
  return dataPtr.GetCurrent () != pDataBegin; // could it be encoded?
}

word InstructionEncoder::EncodeLengths (const Instruction* pInstr/*,word* pDispImmLocs*/) const
{
  const InverseOpcodeInfo* inverseOpcodeInfo = 
    FindFirstInverseOpcodeInfo (pInstr->m_nInstruction);
  
  assert (inverseOpcodeInfo); // or instruction completely unknown
  
  word wLengths = 0;
  
  assert (sizeof (wLengths)*8 >= MAX_INSTR_CODE_LENGTH);
  
  TryDoInstructionEncoding trydo(m_bCodeSize32);
  
  TRY_DO_INSTR_ENCODE_PREPARE(pInstr);
  
  for ( ; ; )
    {
      if (TRY_DO_INSTR_ENCODE_MAY_ATTEMPT(inverseOpcodeInfo) &&
	  trydo.TryEncoding (inverseOpcodeInfo,pInstr))
	wLengths |= trydo.GetLengths();
      
      if (!(inverseOpcodeInfo->m_nFlags & 
	    INVERSE_OPCODE_TAG_HAS_MORE_VARIANTS))
	break;
      inverseOpcodeInfo++;
    }
  return wLengths;
}

bool InstructionEncoder::Encode (const Instruction* pInstr,
				 DataDestPtr& dataPtr,
				 byte bLength,word* pDispImmLocs) const
{
  const InverseOpcodeInfo* inverseOpcodeInfo = 
    FindFirstInverseOpcodeInfo (pInstr->m_nInstruction);
  
  assert (inverseOpcodeInfo); // or instruction completely unknown
  
  assert (bLength);
  
  TryDoInstructionEncoding trydo(m_bCodeSize32);
  
  TRY_DO_INSTR_ENCODE_PREPARE(pInstr);
  
  for ( ; ; )
    {
      if (TRY_DO_INSTR_ENCODE_MAY_ATTEMPT(inverseOpcodeInfo) &&
	  trydo.TryEncoding (inverseOpcodeInfo,pInstr) &&
	  trydo.DoEncodingLength(dataPtr,bLength,pDispImmLocs))
	return true;
      
      if (!(inverseOpcodeInfo->m_nFlags & 
	    INVERSE_OPCODE_TAG_HAS_MORE_VARIANTS))
	break;
      inverseOpcodeInfo++;
    }
  return false;
}

bool InstructionEncoder::Encode (const Instruction* pInstr,
				 VectorInstrCode& vectInstrCode,
				 byte bLength) const
{
  const InverseOpcodeInfo* inverseOpcodeInfo = 
    FindFirstInverseOpcodeInfo (pInstr->m_nInstruction);
  
  assert (inverseOpcodeInfo); // or instruction completely unknown
  assert (bLength);
  
  vectInstrCode.clear ();
  
  InstrCode instrCode;
  
  DataDestPtr newPtr(instrCode.m_cCode 
		     IFDEBUGPARAM(instrCode.m_cCode+
				  sizeof(instrCode.m_cCode)));
  
  TryDoInstructionEncoding trydo(m_bCodeSize32);
  
  TRY_DO_INSTR_ENCODE_PREPARE(pInstr);
  
  for ( ; ; )
    {
      if (TRY_DO_INSTR_ENCODE_MAY_ATTEMPT(inverseOpcodeInfo) &&
	  trydo.TryEncoding (inverseOpcodeInfo,pInstr) &&
	  trydo.DoEncodingLength(newPtr,bLength,
				 &instrCode.m_wDispImmLocs,
				 &instrCode.m_bLengthenTricksUsed))
	{
	  vectInstrCode.push_back (instrCode);						
	  newPtr.ResetCurrent();
	}
      
      if (!(inverseOpcodeInfo->m_nFlags & 
	    INVERSE_OPCODE_TAG_HAS_MORE_VARIANTS))
	break;
      inverseOpcodeInfo++;
    }
  return !vectInstrCode.empty();
}

bool InstructionEncoder::Encode (const Instruction* pInstr,
				 VectorInstrCodeLength& vectInstrCodeLength) const
{
  const InverseOpcodeInfo* inverseOpcodeInfo = 
    FindFirstInverseOpcodeInfo (pInstr->m_nInstruction);
  
  assert (inverseOpcodeInfo); // or instruction completely unknown
  
  vectInstrCodeLength.clear ();
  
  InstrCodeLength instrCodeLength;
  
  DataDestPtr newPtr(instrCodeLength.m_cCode 
		     IFDEBUGPARAM(instrCodeLength.m_cCode+
				  sizeof(instrCodeLength.m_cCode)));
  
  TryDoInstructionEncoding trydo(m_bCodeSize32);
  
  TRY_DO_INSTR_ENCODE_PREPARE(pInstr);
  
  for ( ; ; )
    {
      if (TRY_DO_INSTR_ENCODE_MAY_ATTEMPT(inverseOpcodeInfo) &&
	  trydo.TryEncoding (inverseOpcodeInfo,pInstr))
	{
	  for (instrCodeLength.m_bLength = trydo.GetShortestLength(); 
	       instrCodeLength.m_bLength < MAX_INSTR_CODE_LENGTH; 
	       instrCodeLength.m_bLength++)
	    {
	      if (trydo.DoEncodingLength(newPtr,
					 instrCodeLength.m_bLength,
					 &instrCodeLength.m_wDispImmLocs,
					 &instrCodeLength.m_bLengthenTricksUsed))
		{
		  vectInstrCodeLength.push_back (instrCodeLength);						
		  newPtr.ResetCurrent();
		}
	    }
	  
	  // and see if we can get more lengths from trydo
	}
      
      if (!(inverseOpcodeInfo->m_nFlags & 
	    INVERSE_OPCODE_TAG_HAS_MORE_VARIANTS))
	break;
      inverseOpcodeInfo++;
    }
  return !vectInstrCodeLength.empty();
}

/* TODO:

  unhandled directives:

  arg_MEM_OR_IMM_OP_FAR_PTR

  */
