/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// JavaDisAsmFile.cpp: implementation of the JavaDisAsmFile class.
//
//////////////////////////////////////////////////////////////////////

#include "JavaDisAsmFile.h" 
#include "UrlFile.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

JdafHeader::JdafHeader()
{

}

JdafHeader::JdafHeader(string strName)
{
  m_strName = strName;

}

JdafHeader::~JdafHeader()
{

}

//////////////////////////////////////////////////////////////////////

JdafOperand::JdafOperand()
{

}

JdafOperand::JdafOperand(const string& strName)
{
  m_strName = strName;
}

JdafOperand::~JdafOperand()
{

}


//////////////////////////////////////////////////////////////////////

JdafExternal::JdafExternal()
{

}

JdafExternal::JdafExternal(const UrlFile* pFile,
			   const string& id,
			   const string& name)
{
  m_pJdafFile = pFile;
  m_strId = id;
  m_strName = name;
}

JdafExternal::JdafExternal(const string& file,
			   const string& id,
			   const string& name)
{	
  m_strJdafFile = file;
  m_strId = id;
  m_strName = name;
}

JdafExternal::~JdafExternal()
{

}

//////////////////////////////////////////////////////////////////////

JdafAsmInstr::JdafAsmInstr()
{
  m_bLength = 0;
  m_pCode = NULL;
  m_pOpDest = NULL;
  m_pOpSrc = NULL;
  m_pOpSrc2 = NULL;
}

JdafAsmInstr::~JdafAsmInstr()
{
  delete m_pCode;
  delete m_pOpDest;
  delete m_pOpSrc;
  delete m_pOpSrc2;
}

void JdafAsmInstr::SetCode (byte bLength,const byte* pCode)
{
  assert (bLength);
  
  m_bLength = bLength;
  m_pCode = new byte[bLength];
  memcpy (m_pCode,pCode,bLength);
}

void JdafAsmInstr::SetInstr (const string& strInstr)
{
  m_strInstr = strInstr;
}


//////////////////////////////////////////////////////////////////////

JdafAsmBranchInstr::JdafAsmBranchInstr()
{

}

template<class __II> inline
void for_each_delete_ptr(__II __F, __II __L)
{for (; __F != __L; ++__F)
  delete *__F;
}

JdafAsmBranchInstr::~JdafAsmBranchInstr()
{
  for_each_delete_ptr(m_vectExternalTargets.begin(),
		      m_vectExternalTargets.end());
}

void JdafAsmBranchInstr::AddInternalTarget(const string& strId)
{
  m_vectInternalTargets.push_back (strId);
}

void JdafAsmBranchInstr::AddExternalTarget(const JdafExternal* pExternal)
{
  m_vectExternalTargets.push_back (pExternal);
}

//////////////////////////////////////////////////////////////////////

JdafCodeSnippet::JdafCodeSnippet()
{
  m_wInfoMask = 0;
}

JdafCodeSnippet::~JdafCodeSnippet()
{

}

void JdafCodeSnippet::AddExternalCaller(const JdafExternal* pExternal)
{
  m_wInfoMask |= JDAF_CS_EXTERNAL_TARGET;
  
  m_vectExternalCallers.push_back(pExternal);
}

//////////////////////////////////////////////////////////////////////

JdafAsmCodeSnippet::JdafAsmCodeSnippet()
{
  m_dwOffset = 0;
  m_bHasOffset = false;
}

JdafAsmCodeSnippet::~JdafAsmCodeSnippet()
{
  for_each_delete_ptr(m_vectInstructions.begin(),m_vectInstructions.end());
}

void JdafAsmCodeSnippet::SetOffset (dword dwOffset)
{
  m_dwOffset = dwOffset;
  m_bHasOffset = true;
}

void JdafAsmCodeSnippet::AddInstruction (const JdafAsmInstr* pInstr)
{
  m_vectInstructions.push_back(pInstr);
}

//////////////////////////////////////////////////////////////////////

JavaDisAsmFile::JavaDisAsmFile(UrlFile* pFile)
{
  m_pFile = pFile;
  m_pOutStream = pFile->GetOutStream ();
#ifndef NDEBUG
  m_bHeaderAndEndWritten = 0;
#endif
}

JavaDisAsmFile::~JavaDisAsmFile()
{
#ifndef NDEBUG
  assert (m_bHeaderAndEndWritten == (0x01 | 0x02));
#endif
  delete m_pOutStream;
}

void JavaDisAsmFile::WriteByte (byte b)
{
  *m_pOutStream << b;
}

void JavaDisAsmFile::WriteWord (word w)
{
  //WriteByte (0xcc);
  
  WriteByte(w >> 8);
  WriteByte(w);
  
  /*
	*m_pOutStream
		<< ((byte) (w >> 8))
		<< ((byte)  w);
*/

	/*WriteByte (0xcc);
	*m_pOutStream
		<< ((byte) (w >> 8));
	WriteByte (0xcc);
	*m_pOutStream
		<< (byte) ((byte)  w);
		WriteByte (0xcc); */
}

void JavaDisAsmFile::WriteDWord (dword dw)
{
  WriteByte(dw >> 24);
  WriteByte(dw >> 16);
  WriteByte(dw >> 8);
  WriteByte(dw);
/*
	*m_pOutStream
		<< ((byte) (dw >> 24))
		<< ((byte) (dw >> 16))
		<< ((byte) (dw >> 8))
		<< ((byte)  dw);
*/
}

void JavaDisAsmFile::WriteBytes (int len,const byte* pData)
{
  m_pOutStream->write((const char*) pData,len);
}

void JavaDisAsmFile::WriteString (const string& str)
{
  assert (m_pOutStream);
  assert (str.size() <= 0xffff);
  
  WriteWord (str.size());
  
#ifndef NDEBUG
  for (unsigned int i = 0; i < str.size(); i++)
    {
      assert (((byte) str[i]) && ((byte) str[i]) < 0x80); // otherwise handle java UTF encoding (see DataInputStream::readUTF)
    }
#endif
  
  //WriteByte (0xcd);
  *m_pOutStream << str;
  //WriteByte (0xcd);
}

void JavaDisAsmFile::Write (const JdafHeader& header)
{
#ifndef NDEBUG
  assert (m_bHeaderAndEndWritten == (0));
  m_bHeaderAndEndWritten |= 0x01;
#endif
  
  WriteString(header.m_strName);
  WriteString(header.m_strComment);
  
  WriteWord(0);
}

void JavaDisAsmFile::Write (const JdafCodeSnippet& code)
{
#ifndef NDEBUG
  assert (m_bHeaderAndEndWritten == (0x01));
#endif

  WriteString(code.m_strId);
  
  word wInclude = code.m_wInfoMask;
  
  WriteWord (wInclude);
  
  if (code.m_wInfoMask & JDAF_CS_EXTERNAL_TARGET)
    {
      WriteDWord (code.m_vectExternalCallers.size());
      
      for (uint i = 0; i < code.m_vectExternalCallers.size(); i++)
	{
	  const JdafExternal* pInstr = code.m_vectExternalCallers[i];
	  Write (*pInstr);
	}
    }
}

void JavaDisAsmFile::Write (const JdafAsmCodeSnippet& code)
{
#ifndef NDEBUG
  assert (m_bHeaderAndEndWritten == (0x01));
#endif

  WriteWord(1);
  
  Write((const JdafCodeSnippet&) code);
  
  dword wInclude = 0;
  
  if (code.m_bHasOffset)
    wInclude |= 0x01;
  
  uint i;
  
  for (i = 0; i < code.m_vectInstructions.size(); i++)
    {
      const JdafAsmInstr* pInstr = code.m_vectInstructions[i];
      
      if (pInstr->m_bLength)
	wInclude |= 0x02;
    }
  
  WriteWord (wInclude);
  WriteDWord (code.m_vectInstructions.size());
  
  if (wInclude & 0x01)
    {
      assert (code.m_bHasOffset);
      WriteDWord(code.m_dwOffset);
    }
  
  for (i = 0; i < code.m_vectInstructions.size(); i++)
    {
      const JdafAsmInstr* pInstr = code.m_vectInstructions[i];
      const JdafAsmBranchInstr* pBranchInstr = 
	dynamic_cast<const JdafAsmBranchInstr*>(pInstr);
      if (pBranchInstr)
	Write (*pBranchInstr,wInclude);
      else
	Write (*pInstr,wInclude);
    }
}

void JavaDisAsmFile::Write (const JdafAsmInstr& instr,
			    word wInclude,
			    byte bOpIncluded)
{
  if (instr.m_pOpDest) bOpIncluded |= 0x01;
  if (instr.m_pOpSrc ) bOpIncluded |= 0x02;
  if (instr.m_pOpSrc2) bOpIncluded |= 0x04;
  
  WriteByte (bOpIncluded);
  
  if (wInclude & 0x02)
    {
      WriteByte (instr.m_bLength);
      WriteBytes(instr.m_bLength,instr.m_pCode);
    }
  
  WriteString (instr.m_strInstr);
  
  if (instr.m_pOpDest) Write(*instr.m_pOpDest);
  if (instr.m_pOpSrc ) Write(*instr.m_pOpSrc );
  if (instr.m_pOpSrc2) Write(*instr.m_pOpSrc2);
}

void JavaDisAsmFile::Write (const JdafAsmBranchInstr& instr,word wInclude)
{
  byte bOpIncluded = 0;
  
  if (instr.m_vectInternalTargets.size()) bOpIncluded |= 0x08;
  if (instr.m_vectExternalTargets.size()) bOpIncluded |= 0x10;
  
//	assert (m_bType & ~(0x20 | 0x40) == 0);
//	bOpIncluded |= m_bType; // for being a branching instr
  
  Write ((JdafAsmInstr&) instr,wInclude,bOpIncluded);
  
  if (instr.m_vectInternalTargets.size())
    {
      WriteDWord (instr.m_vectInternalTargets.size());
      
      for (uint i = 0; i < instr.m_vectInternalTargets.size(); i++)
	WriteString (instr.m_vectInternalTargets[i]);
    }
  
  if (instr.m_vectExternalTargets.size())
    {
      WriteDWord (instr.m_vectExternalTargets.size());
      
      for (uint i = 0; i < instr.m_vectExternalTargets.size(); i++)
	{
	  const JdafExternal* pExt = instr.m_vectExternalTargets[i];
	  Write (*pExt);
	}
    }
}

void JavaDisAsmFile::Write (const JdafOperand& op)
{
  WriteString(op.m_strName);
}

void JavaDisAsmFile::Write (const JdafExternal& ext)
{
  if (ext.m_pJdafFile)
    WriteString(m_pFile->GetRelativePathToItem(ext.m_pJdafFile));
  else
    WriteString(ext.m_strJdafFile);
  
  WriteString(ext.m_strId);
  if (ext.m_strId != ext.m_strName)
    WriteString(ext.m_strName);
  else
    WriteString("");
}


void JavaDisAsmFile::WriteEnd ()
{
#ifndef NDEBUG
  assert (m_bHeaderAndEndWritten == (0x01));
  m_bHeaderAndEndWritten |= 0x02;
#endif

  WriteWord(0); // nothing more
}













