/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// JavaDisAsmFile.h: interface for the JavaDisAsmFile class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__JAVA_DIS_ASM_FILE_H__INCLUDED_)
#define __JAVA_DIS_ASM_FILE_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "GlobalDefs.h" 

class UrlFile;

#include <string>
#include <vector>

using namespace std;

class JdafHeader
{
 public:
  JdafHeader();
  JdafHeader(string strName);
  virtual ~JdafHeader();
  
 public:
  string	m_strName;
  string	m_strComment;
};

class JdafOperand
{
 public:
  JdafOperand();
  JdafOperand(const string& strName);
  virtual ~JdafOperand();
  
 public:
  string	m_strName;
};

class JdafExternal
{
 public:
  JdafExternal();
  JdafExternal(const UrlFile* pFile,const string& id,const string& name);
  JdafExternal(const string& file,const string& id,const string& name);
  virtual ~JdafExternal();
  
 public:
  const UrlFile* m_pJdafFile;
  
  string m_strJdafFile;
  string m_strId;
  string m_strName;
};

class JdafAsmInstr
{
 public:
  JdafAsmInstr();
  virtual ~JdafAsmInstr();
  
 public:
  byte	m_bLength;
  byte*	m_pCode;
  
  string	m_strInstr;
  
  JdafOperand*	m_pOpDest;	
  JdafOperand*	m_pOpSrc;	
  JdafOperand*	m_pOpSrc2;	
  
 public:
  void SetCode (byte bLength,const byte* pCode);
  void SetInstr (const string& strInstr);
};
/*
#define JDAF_ABI_CAN_CONTINUE	0x20
#define JDAF_ABI_ENDS_TRACE		0x40
*/
class JdafAsmBranchInstr : public JdafAsmInstr
{
 public:
  JdafAsmBranchInstr();
  virtual ~JdafAsmBranchInstr();
  
 public:
  //	byte m_bType;
  
  vector<string> m_vectInternalTargets;
  vector<const JdafExternal*> m_vectExternalTargets;
  
 public:
  void AddInternalTarget(const string& strId);
  void AddExternalTarget(const JdafExternal* pExternal);
  
//	void SetType (byte bType) { m_bType = bType; }

};

#define JDAF_CS_FALLS_THROUGH	0x01
#define JDAF_CS_FALLED_THROUGH	0x02
#define JDAF_CS_INTERNAL_TARGET	0x04
#define JDAF_CS_EXTERNAL_TARGET	0x08

class JdafCodeSnippet
{
 public:
  JdafCodeSnippet();
  virtual ~JdafCodeSnippet();
  
 public:
  string	m_strId;
  
  vector<const JdafExternal*> m_vectExternalCallers;
  
  word	m_wInfoMask;
  
 public:
  void SetId(const string& id) { m_strId = id; }
  
 public:
  void SetFallsThrough()   { m_wInfoMask |= JDAF_CS_FALLS_THROUGH; }
  void SetFalledThrough()  { m_wInfoMask |= JDAF_CS_FALLED_THROUGH; }
  void SetInternalTarget() { m_wInfoMask |= JDAF_CS_INTERNAL_TARGET; }
  void AddExternalCaller(const JdafExternal* pExternal);
  
};

class JdafAsmCodeSnippet : public JdafCodeSnippet
{
 public:
  JdafAsmCodeSnippet();
  virtual ~JdafAsmCodeSnippet();
  
 public:
  dword	m_dwOffset;
  bool	m_bHasOffset;
  
  vector<const JdafAsmInstr*> m_vectInstructions;
	
 public:
  void AddInstruction (const JdafAsmInstr* pInstr);
  
  void SetOffset (dword dwOffset);
};

class JavaDisAsmFile  
{
 public:
  JavaDisAsmFile(UrlFile* pFile);
  virtual ~JavaDisAsmFile();
  
 protected:
  ostream* m_pOutStream;
  
  UrlFile* m_pFile;
  
#ifndef NDEBUG
  byte m_bHeaderAndEndWritten;
#endif
  
 protected:
  void WriteString (const string& str);
  void WriteByte (byte b);
  void WriteWord (word w);
  void WriteDWord (dword dw);
  void WriteBytes (int len,const byte* pData);
  
 public:
  void Write (const JdafHeader& header);
  void Write (const JdafCodeSnippet& code);
  void Write (const JdafAsmCodeSnippet& code);
  void Write (const JdafAsmInstr& instr,word wInclude,byte bOpIncluded = 0);
  void Write (const JdafAsmBranchInstr& instr,word wInclude);
  void Write (const JdafOperand& op);
  void Write (const JdafExternal& ext);
  
  void WriteEnd ();
  
};

#endif // !defined(__JAVA_DIS_ASM_FILE_H__INCLUDED_)
