/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LiveRegisterAnalysis.cpp: implementation of the LiveRegisterAnalysis class.
//
//////////////////////////////////////////////////////////////////////

#include "LiveRegisterAnalysis.h" 
#include "OperandRegister.h" 
class Instruction; 
#include "BranchInstr.h" 

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

void LiveRegisters::SetUnUsedOpReg (byte bReg)
{
  byte bRegType = bReg & reggTYPE_MASK;
  
  bReg &= reggNAME_MASK;
  
  switch (bRegType)
    {
    case reggSTD8:  SetUnUsedRegNrml8 (bReg); return;
    case reggSTD16: SetUnUsedRegNrml16(bReg); return;
    case reggSTD32: SetUnUsedRegNrml32(bReg); return;
    case reggSEG:   SetUnUsedSegReg   (bReg); return;
    case reggST:    SetUnUsedSTn      (bReg); return;
    case reggMM:    SetUnUsedMMn      (bReg); return;
    case reggXMM:   SetUnUsedXMMn     (bReg); return;
    }
  assert (false);
}

void LiveRegisters::SetUsedOpReg (byte bReg)
{
  byte bRegType = bReg & reggTYPE_MASK;
  
  bReg &= reggNAME_MASK;
  
  switch (bRegType)
    {
    case reggSTD8:  SetUsedRegNrml8 (bReg); return;
    case reggSTD16: SetUsedRegNrml16(bReg); return;
    case reggSTD32: SetUsedRegNrml32(bReg); return;
    case reggSEG:   SetUsedSegReg   (bReg); return;
    case reggST:    SetUsedSTn      (bReg); return;
    case reggMM:    SetUsedMMn      (bReg); return;
    case reggXMM:   SetUsedXMMn     (bReg); return;
    }
  assert (false);
}

//////////////////////////////////////////////////////////////////////

LiveRegisterAnalysis::LiveRegisterAnalysis()
{

}

LiveRegisterAnalysis::~LiveRegisterAnalysis()
{

}

void LiveRegisterAnalysis::UpdateRegisterUsage (LiveRegisters& liveReg,BranchInstr* pInstr)
{
  if ((pInstr->m_nInstruction & instrmCTRL_TRANSFER_TYPE_MASK) == instrmJMP_COND)
    {
      switch (pInstr->m_nInstruction & instrmJMP_COND_MASK)
	{
	case instrmJCC:
	  switch (pInstr->m_nInstruction & instrmCC_COND_MASK_TRUE_AND_FALSE)
	    {
	    case instrmCC_O:  liveReg.SetUsedStsFlag(LIVE_REG_STS_EFLAG_OF); break;
	    case instrmCC_B:  liveReg.SetUsedStsFlag(LIVE_REG_STS_EFLAG_CF); break;
	    case instrmCC_Z:  liveReg.SetUsedStsFlag(LIVE_REG_STS_EFLAG_ZF); break;
	    case instrmCC_BE: liveReg.SetUsedStsFlag(LIVE_REG_STS_EFLAG_ZF | LIVE_REG_STS_EFLAG_CF); break;
	    case instrmCC_S:  liveReg.SetUsedStsFlag(LIVE_REG_STS_EFLAG_SF); break;
	    case instrmCC_P:  liveReg.SetUsedStsFlag(LIVE_REG_STS_EFLAG_PF); break;
	    case instrmCC_L:  liveReg.SetUsedStsFlag(LIVE_REG_STS_EFLAG_SF | LIVE_REG_STS_EFLAG_OF); break;
	    case instrmCC_LE: liveReg.SetUsedStsFlag(LIVE_REG_STS_EFLAG_ZF | LIVE_REG_STS_EFLAG_SF | LIVE_REG_STS_EFLAG_OF); break;
	      DEFAULT_ASSERT_FALSE;
	    }
	  return; // done
	case instrJECXZ:
	case instrLOOPZ:
	case instrLOOPNZ:
	  liveReg.SetUsedStsFlag(LIVE_REG_STS_EFLAG_ZF);
	case instrLOOP:
	  if (pInstr->GetAddrSize32())
	    liveReg.SetUnUsedRegNrml32(regmNAME_ECX);
	  else
	    liveReg.SetUnUsedRegNrml16(regmNAME_CX);
	  return;
	  DEFAULT_ASSERT_FALSE;
	}
    }
  
  liveReg.SetAllUsed(); // default behaviour
}


void LiveRegisterAnalysis::UpdateRegisterUsage (LiveRegisters& liveReg,Instruction* pInstr)
{
  assert (!dynamic_cast<BranchInstr*>(pInstr));
  
  // update register usage,
  // this is operated in a backwards fashion
  
  // incoming data (liveReg) is what is alive after this instruction
  
  // outgoing data (liveReg) is what is alive before this instruction
  
  // a register is declared unused if it is redefined in this instruction
  // and does not depend on the value of this instruction
  
  // a register is declared used if it is used by the instruction
  
  // default behaviour is to declare all registers as used (we do not now anything special about the instruction)
  
  
  
  // if ((pInstr->m_nInstruction & instrmCTRL_TRANSFER_TYPE_MASK) == instrmJMP_COND)
  
  
  // if (pInstr->m_nInstruction ==
  //	instrgMASK
  
  
  
  
  // for a normal instruction, find out what registers are incoming and outgoing
  // 
  













  liveReg.SetAllUsed(); // default behaviour
}
