/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LiveRegisterAnalysis.h: interface for the LiveRegisterAnalysis class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__LIVE_REGISTER_ANALYSIS_H__INCLUDED_)
#define __LIVE_REGISTER_ANALYSIS_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "GlobalDefs.h" // dep_chk_needed     

#define LIVE_REG_NRML_REG_AL		0x00000001
#define LIVE_REG_NRML_REG_AH		0x00000100
#define LIVE_REG_NRML_REG_AX		0x00000101
#define LIVE_REG_NRML_REG_EAX		0x01010101

#define LIVE_REG_MORE_REG_8(n)		(n < 4 ? (LIVE_REG_NRML_REG_AL << (n)) : (LIVE_REG_NRML_REG_AH << (n)))
#define LIVE_REG_MORE_REG_16(n)		(LIVE_REG_NRML_REG_AX << (n))
#define LIVE_REG_MORE_REG_32(n)		(LIVE_REG_NRML_REG_EAX << (n))

#define LIVE_REG_NRML_REG_ALL		0xffffffff

#define LIVE_REG_MORE_REG_SEG_FIRST 0x00000001
#define LIVE_REG_MORE_REG_ST0		0x00000100
#define LIVE_REG_MORE_REG_MM0		0x00010000
#define LIVE_REG_MORE_REG_XMM0		0x01000000

#define LIVE_REG_MORE_REG_SEG(n)	(LIVE_REG_MORE_REG_SEG_FIRST << (n))
#define LIVE_REG_MORE_REG_ST(n)		(LIVE_REG_MORE_REG_ST0 << (n))
#define LIVE_REG_MORE_REG_MM(n)		(LIVE_REG_MORE_REG_MM0 << (n))
#define LIVE_REG_MORE_REG_XMM(n)	(LIVE_REG_MORE_REG_XMM0 << (n))

#define LIVE_REG_MORE_REG_SEG_ALL	0x0000003f
#define LIVE_REG_MORE_REG_ST_ALL	0x0000ff00
#define LIVE_REG_MORE_REG_MM_ALL	0x00ff0000
#define LIVE_REG_MORE_REG_XMM_ALL	0xff000000

#define LIVE_REG_MORE_REG_ALL		0xffffff3f

#define LIVE_REG_STS_EFLAG_CF		0x0001
#define LIVE_REG_STS_EFLAG_PF		0x0002
#define LIVE_REG_STS_EFLAG_AF		0x0004
#define LIVE_REG_STS_EFLAG_ZF		0x0008
#define LIVE_REG_STS_EFLAG_SF		0x0010
#define LIVE_REG_STS_EFLAG_DF		0x0020
#define LIVE_REG_STS_EFLAG_OF		0x0040

#define LIVE_REG_STS_WD_FPU_C0		0x0100
#define LIVE_REG_STS_WD_FPU_C1		0x0200
#define LIVE_REG_STS_WD_FPU_C2		0x0400
#define LIVE_REG_STS_WD_FPU_C3		0x0800

#define LIVE_REG_STATUS_FLAGS_ALL	0x0f7f

class Instruction;
class BranchInstr;

struct LiveRegisters
{
  public:
LiveRegisters(const LiveRegisters& src)
  {
    m_dwStdReg = src.m_dwStdReg;
    m_dwMoreReg = src.m_dwMoreReg;
    m_wStatusFlags = src.m_wStatusFlags;
  }
  LiveRegisters() { SetAllUsed(); }
  
  public:
  dword	m_dwStdReg;     // eAX,eCX,eDX,eBX,eSP,eBP,eSI,EDI
  dword	m_dwMoreReg;    // segregs,st0-7,mm0-7,xmm0-7
  word    m_wStatusFlags;	// normal and fpu
  
  public:
  void SetUnUsedRegNrml8  (int n) { assert (n >= 0 && n < 8); m_dwStdReg &= ~LIVE_REG_MORE_REG_8 (n); }
  void SetUnUsedRegNrml16 (int n) { assert (n >= 0 && n < 8); m_dwStdReg &= ~LIVE_REG_MORE_REG_16(n); }
  void SetUnUsedRegNrml32 (int n) { assert (n >= 0 && n < 8); m_dwStdReg &= ~LIVE_REG_MORE_REG_32(n); }
  
  void SetUsedRegNrml8    (int n) { assert (n >= 0 && n < 8); m_dwStdReg |=  LIVE_REG_MORE_REG_8 (n); }
  void SetUsedRegNrml16   (int n) { assert (n >= 0 && n < 8); m_dwStdReg |=  LIVE_REG_MORE_REG_16(n); }
  void SetUsedRegNrml32   (int n) { assert (n >= 0 && n < 8); m_dwStdReg |=  LIVE_REG_MORE_REG_32(n); }
  
  void SetUnUsedSegReg (int n) { assert (n >= 0 && n < 6); m_dwMoreReg &= ~LIVE_REG_MORE_REG_SEG(n); }
  void SetUnUsedSTn    (int n) { assert (n >= 0 && n < 8); m_dwMoreReg &= ~LIVE_REG_MORE_REG_ST(n); }
  void SetUnUsedMMn    (int n) { assert (n >= 0 && n < 8); m_dwMoreReg &= ~LIVE_REG_MORE_REG_MM(n); }
  void SetUnUsedXMMn   (int n) { assert (n >= 0 && n < 8); m_dwMoreReg &= ~LIVE_REG_MORE_REG_XMM(n); }
  
  void SetUsedSegReg   (int n) { assert (n >= 0 && n < 6); m_dwMoreReg |=  LIVE_REG_MORE_REG_SEG(n); }
  void SetUsedSTn      (int n) { assert (n >= 0 && n < 8); m_dwMoreReg |=  LIVE_REG_MORE_REG_ST(n); }
  void SetUsedMMn      (int n) { assert (n >= 0 && n < 8); m_dwMoreReg |=  LIVE_REG_MORE_REG_MM(n); }
  void SetUsedXMMn     (int n) { assert (n >= 0 && n < 8); m_dwMoreReg |=  LIVE_REG_MORE_REG_XMM(n); }
  
  void SetUnUsedOpReg (byte bReg);
  void SetUsedOpReg (byte bReg);
  
  void SetUnUsedStsFlag (word wFlag) { assert (!(wFlag & (~LIVE_REG_STATUS_FLAGS_ALL))); m_wStatusFlags &= ~wFlag; }
  void SetUsedStsFlag   (word wFlag) { assert (!(wFlag & (~LIVE_REG_STATUS_FLAGS_ALL))); m_wStatusFlags |=  wFlag; }
  
  void SetAllUsed ()
  {
    m_dwStdReg = LIVE_REG_NRML_REG_ALL;
    m_dwMoreReg = LIVE_REG_MORE_REG_ALL;
    m_wStatusFlags = LIVE_REG_STATUS_FLAGS_ALL;
  }
  
};

class LiveRegisterAnalysis  
{
 public:
  LiveRegisterAnalysis();
  virtual ~LiveRegisterAnalysis();
  
  
  
  
 public:
  void UpdateRegisterUsage (LiveRegisters& liveReg,Instruction* pInstr);
  void UpdateRegisterUsage (LiveRegisters& liveReg,BranchInstr* pInstr);
  
  
};

#endif // !defined(__LIVE_REGISTER_ANALYSIS_H__INCLUDED_)
