/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Misc.cpp: miscellaneous.
//
//////////////////////////////////////////////////////////////////////

#include "Misc.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

bool powerof2 (unsigned long x)
{
  // ehh, is this the best way to test this???, it seem like a long
  // way... Suggestions anybody

  // BSF and BSR could be put to use, but not now, only if
  // optimization is needed...
  
  while (x > 1)
    {
      if (x & 1)
	return false;
      x >>= 1;
    }
  return true;
}

//////////////////////////////////////////////////////////////////////

bool iszero(const void* ptr,size_t sz)
{
  for (const char* p = (const char*) ptr; sz; sz--)
    {
      if (*(p++))
        return false;
    }
  return true;
}

//////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#ifdef WIN32
#define vsnprintf _vsnprintf
#else
#endif

char* make_message(const char *fmt, va_list ap) 
{
  int n, size = strlen(fmt)+1; // +1 for trailing '\0'
  char *p;
  va_list ap_use;
  if ((p = (char *) malloc (size)) == NULL)
    return NULL;
  while (1) {
    /* Try to print in the allocated space. */
    
    ap_use = ap;
    
    // now, va_start should have been called by our caller,
    // but we may use ap (via ap_use) multiple times,
    // without any call to va_end, hopefully, this won't
    // cause any problems

    n = vsnprintf (p, size, fmt, ap_use);

    /* If that worked, return the string. */
    if (n > -1 && n < size)
      break;
    /* Else try again with more space. */
    if (n > -1)    /* glibc 2.1 */
      size = n+1; /* precisely what is needed */
    else           /* glibc 2.0 */
      size *= 2;  /* twice the old size */
    if ((p = (char *) realloc (p, size)) == NULL)
      {
        free(p);
        return NULL;
      }
  }
/*
  struct 
  {
   int start;
   char data[64];
  } str;

  va_list ap2;
  va_start(ap2,str.start);
*/
  return p;
}

string strsprintf(const char *fmt, ...)
{
  va_list ap;
  va_start(ap,fmt);
  char* p = make_message(fmt,ap);
  va_end(ap);
  
  // now, I seem to do a lot copying of the string here while returning.
  // perhaps you have any suggestions for making this more efficient
  // (while still being able to use strsprintf as an argument to functions and so...)
  
  string s = p ? p : "";
  free(p);
  return s;
}

void appendsprintf(string& append,const char *fmt, ...)
{
  va_list ap;
  va_start(ap,fmt);
  char* p = make_message(fmt,ap);
  va_end(ap);
  
  if (p)
    append += p;
  free(p);
}






