/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// OperandImmediate.cpp: implementation of the OperandImmediate class.
//
//////////////////////////////////////////////////////////////////////

#include "OperandImmediate.h" // dep_chk_needed 
#include <stdio.h>
#include <assert.h>

#include "VariableNConst.h" // dep_chk_unknown 


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

OperandImmediate::OperandImmediate()
{
  m_bImmHigh16Bytes = 0;
}

OperandImmediate::OperandImmediate(byte bImm)
{
  m_bImmHigh16Bytes = 0;
  
  m_bImmBytes = 1;
  m_dwImm = bImm;
}

OperandImmediate::OperandImmediate(dword dwImm,byte	bImmBytes)
{
  m_bImmHigh16Bytes = 0;
  
  assert (bImmBytes == 1 || bImmBytes == 2 || bImmBytes == 4);
  
  m_dwImm = dwImm;
  m_bImmBytes = bImmBytes;
}

OperandImmediate::OperandImmediate(dword dwImm,word	wImmHigh16,
				   byte	bImmBytes)
{
  m_bImmHigh16Bytes = 2;
  
  assert (bImmBytes == 2 || bImmBytes == 4);
  
  m_dwImm = dwImm;
  m_wImmHigh16 = wImmHigh16;
  m_bImmBytes = bImmBytes;
}

OperandImmediate::~OperandImmediate()
{

}

void OperandImmediate::SignExtend (int nDestSize)
{
  assert (m_bImmHigh16Bytes == 0);
  assert (m_bImmBytes == 1);
  
  if (nDestSize == 2)
    {
      m_dwImm = (word) ((sword) ((sbyte) m_dwImm));
    }
  else
    {
      assert (nDestSize == 4);
      m_dwImm = (dword) ((sdword) ((sbyte) m_dwImm));
    }
  m_bImmBytes = nDestSize;
}

bool OperandImmediate::SignExtended () const
{
  assert (m_bImmHigh16Bytes == 0);
  
  if (m_bImmBytes == 1)
    return true; // it's already sign-extended
  else if (m_bImmBytes == 2)
    return m_dwImm == (word) ((sword) ((sbyte) m_dwImm));
  else
    {
      assert (m_bImmBytes == 4);
      return m_dwImm == (dword) ((sdword) ((sbyte) m_dwImm));
    }
}

char* OperandImmediate::GenAsmString (char* pDest) const
{
  assert (m_bImmHigh16Bytes == 0 || m_bImmHigh16Bytes == 2);
  
  if (m_bImmHigh16Bytes)
    {
      assert (m_bImmHigh16Bytes == 2);
      pDest += sprintf (pDest,"%04X:",(word) m_wImmHigh16);
    }
  
  if (m_bImmBytes == 1)
    {
      if (m_dwImm < 10)
	pDest += sprintf (pDest,"%d",(byte) m_dwImm);
      else
	pDest += sprintf (pDest,"%02Xh",(byte) m_dwImm);
    }
  else if (m_bImmBytes == 2)
    pDest += sprintf (pDest,"%04Xh",(word) m_dwImm);
  else
    {
      assert (m_bImmBytes == 4);
      pDest += sprintf (pDest,"%08Xh",(dword) m_dwImm);
    }
  return pDest;
}

Variable* OperandImmediate::GetConstVar () const
{
  assert (m_bImmHigh16Bytes == 0);
  
  assert (m_bImmBytes == 1 || m_bImmBytes == 2 || m_bImmBytes == 4);
  
  if (m_bImmBytes == 1)
    return new Variable8Const((byte) m_dwImm);
  if (m_bImmBytes == 2)
    return new Variable16Const((word) m_dwImm);
  assert (m_bImmBytes == 4);
  return new Variable32Const((dword) m_dwImm);
}


