/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// OperandMemory.cpp: implementation of the OperandMemory class.
//
//////////////////////////////////////////////////////////////////////

#include "OperandMemory.h" 
#include "OperandRegister.h" 
#include <stdio.h>
#include <assert.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

OperandMemory::OperandMemory()
{
  m_bBaseReg = 0;
  m_bIndexReg = 0;
  m_bScale = 0; 
  m_bSegmentReg = 0;
  m_wOpSize = 0; 
  m_bDispBytes = 0;
  m_dwDisp = 0;
}

OperandMemory::~OperandMemory()
{

}

char* OperandMemory::GenAsmString (char* pDest) const
{
  if (m_wOpSize == 1)
    pDest += sprintf (pDest,"byte ptr ");
  else if (m_wOpSize == 2)
    pDest += sprintf (pDest,"word ptr ");
  else if (m_wOpSize == 4)
    pDest += sprintf (pDest,"dword ptr ");
  else if (m_wOpSize == 8)
    pDest += sprintf (pDest,"qword ptr ");
  else if (m_wOpSize == 16)
    pDest += sprintf (pDest,"oword ptr ");
  else 
    pDest += sprintf (pDest,"%d-byte ptr ",m_wOpSize);
  
  if (m_bSegmentReg != (IS_REG_NAME_SP_BP(m_bBaseReg) ? regSS : regDS))
    {
      // display register name if it's not default, only ugly thing is
      // that we for string instructions will show SI , ES:DI instead
      // of DS:SI , ES:DI
      
      const char* pSrc = GetRegisterName(m_bSegmentReg);
      
      while (*pSrc)
	*(pDest++) = *(pSrc++);
      *(pDest++) = ':';
    }	
  *(pDest++) = '[';
  if (m_bBaseReg)
    {
      const char* pSrc = GetRegisterName(m_bBaseReg);
      
      while (*pSrc)
	*(pDest++) = *(pSrc++);
    }
  if (m_bIndexReg)
    {
      if (m_bBaseReg)
	*(pDest++) = '+';
      const char* pSrc = GetRegisterName(m_bIndexReg);
      
      while (*pSrc)
	*(pDest++) = *(pSrc++);
      if (m_bScale)
	{
	  *(pDest++) = '*';
	  *(pDest++) = '0' + (1 << m_bScale);
	}
    }
  
  assert (m_bDispBytes || m_dwDisp == 0);
  assert (m_bDispBytes > 1 || (m_dwDisp >= -0x80   && m_dwDisp < 0x80));
  assert (m_bDispBytes > 2 || (m_dwDisp >= -0x8000 && m_dwDisp < 0x8000));
  
  if (m_bDispBytes)
    {
      sdword dwDisp = m_dwDisp;
      
      if (dwDisp & 0x80000000)
	{
	  *(pDest++) = '-';
	  dwDisp = -dwDisp;
	}
      else if (m_bBaseReg || m_bIndexReg)
	*(pDest++) = '+';
      
      if (m_bDispBytes == 1)
	pDest += sprintf (pDest,"%02Xh",(byte) dwDisp);
      else if (m_bDispBytes == 2)
	pDest += sprintf (pDest,"%04Xh",(word) dwDisp);
      else
	{
	  assert (m_bDispBytes == 4);
	  pDest += sprintf (pDest,"%08Xh",(dword) dwDisp);
	}
    }
  *(pDest++) = ']';
  return pDest;
}

bool OperandMemory::IsAddrSize32() const
{
  assert (!m_bBaseReg || !m_bIndexReg || 
	  REG_TYPE(m_bBaseReg) == REG_TYPE(m_bIndexReg));
  
  if (m_bBaseReg)
    return REG_TYPE(m_bBaseReg) == reggSTD32;
  if (m_bIndexReg)
    return REG_TYPE(m_bIndexReg) == reggSTD32;
  assert (m_bDispBytes == 2 || m_bDispBytes == 4);
  
  return m_bDispBytes == 4;
}



