/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// OperandMemory.h: interface for the OperandMemory class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__OPERAND_MEMORY_H__INCLUDED_)
#define __OPERAND_MEMORY_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Operand.h" // dep_chk_needed     
// // #include "OperandRegister.h" // dep_chk_unknown     

#define SCALE_1		0x00
#define SCALE_2		0x01
#define SCALE_4		0x02
#define SCALE_8		0x03

class OperandMemory : public Operand
{
 public:
  OperandMemory();
  virtual ~OperandMemory();
  
 public:
  byte	m_bBaseReg;
  byte	m_bIndexReg;
  byte	m_bScale;    // scale = 0,1,2 or 3  (as x 1,2,4 or 8) (low nibble)
  byte	m_bSegmentReg; // 0 if default (no override)
  word	m_wOpSize;
  byte	m_bDispBytes; // strictly not necessary as m_dwDisp contain 0 when unused,
                      // and 0xssssss?? when 8-bit, where s is the sign bit and
                      // 0xssss???? when 16-bit
                      // 
                      // actually, it's necessary when both base and index are 0
                      // so this is the only way of differentiating between 16 and 32-
                      // bit addresses
                      // the instruction encoder may choose a smaller (or larger)
                      // displacement (sign-extended) if that does not change the address
  sdword	m_dwDisp;
  
 public:
  virtual char* GenAsmString (char* pDest) const;
  
 public:
  virtual int DisplacementSize () const { return m_bDispBytes; }
  
  virtual void SetDisplacement (sdword dwDisp) { m_dwDisp = dwDisp; }
  virtual sdword GetDisplacement () const { return m_dwDisp; }
  
 public:
  virtual void SetDisplacementSizeByte (sbyte bDisp) 
    { m_dwDisp = bDisp; m_bDispBytes = 1; }
  virtual void SetDisplacementSizeWord (sword wDisp) 
    { m_dwDisp = wDisp; m_bDispBytes = 2; }
  virtual void SetDisplacementSizeDWord (sdword dwDisp) 
    { m_dwDisp = dwDisp; m_bDispBytes = 4; }
  
  virtual void SetDisplacementByte (sbyte bDisp) 
    { m_dwDisp = bDisp; assert (m_bDispBytes == 1); }
  virtual void SetDisplacementWord (sword wDisp) 
    { m_dwDisp = wDisp; assert (m_bDispBytes == 2); }
  virtual void SetDisplacementDWord (sdword dwDisp) 
    { m_dwDisp = dwDisp; assert (m_bDispBytes == 4); }
 public:
  virtual bool IsMemoryPtr () const { return true; }
  
 public:
  virtual word GetOpSize () const { return m_wOpSize; }
  
  bool IsAddrSize32() const;
  
  int GetScale () const { return m_bScale; }
  
};

#endif // !defined(__OPERAND_MEMORY_H__INCLUDED_)
