/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// OperandRegister.cpp: implementation of the OperandRegister class.
//
//////////////////////////////////////////////////////////////////////

#include "OperandRegister.h" 
#include "gen/Register_Names.h"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
/*
OperandRegister::OperandRegister()
{
	m_bReg = 0;
}
*/
OperandRegister::OperandRegister(byte bReg)
{
  assert (bReg);
  m_bReg = bReg;
}

OperandRegister::~OperandRegister()
{

}

//////////////////////////////////////////////////////////////////////

// extern char cRegNames[];
// extern RegIdNameIndex indexRegIdName[];

int compare_byte (const void* p1,const void* p2)
{
  return ((byte) *((const byte*) p1) - (byte) *((const byte*) p2));
}

const char* GetRegisterName (byte bReg)
{
  RegIdNameIndex* pIndex = 
    (RegIdNameIndex*) bsearch (&bReg,
			       indexRegIdName,
			       NUMBER_OF_REG_NAMES,
			       sizeof (RegIdNameIndex),
			       compare_byte);
  
  assert (pIndex); // otherwise we're missing name of the register
  
  assert (pIndex->m_nOffset >= 0 &&
	  pIndex->m_nOffset < SIZEOF_REG_NAMES_STRING);
  
  return cRegNames+pIndex->m_nOffset;
}

//////////////////////////////////////////////////////////////////////

char* OperandRegister::GenAsmString (char* pDest) const
{
  const char* pSrc = GetRegisterName(m_bReg);
  
  while (*pSrc)
    *(pDest++) = *(pSrc++);
  return pDest;
}

word GetRegisterSize(byte bReg)
{
  byte bRegType = bReg & reggTYPE_MASK;
  
  switch (bRegType)
    {
    case reggSTD8: return 1;
    case reggSTD16: 
    case reggSEG:  return 2;
    case reggSTD32: 
    case reggDBG: 
    case reggCTRL: 
    case reggTST:  return 4;
    case reggST: return 10;
    case reggMM: return 18;
    case reggXMM: return 16;
      DEFAULT_ASSERT_FALSE;
    }
  
  return 0; // which is not very likely size of anything existing
  
  // observe trouble with MMX/SSE instructions that give an MMX/SSE
  // reg as operand, but specify another operand size, perhaps store
  // op-size in OperandRegister
}


