/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// OperandRegister.h: interface for the OperandRegister class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__OPERAND_REGISTER_H__INCLUDED_)
#define __OPERAND_REGISTER_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Operand.h" 
#include "GlobalDefs.h" 

//////////////////////////////////////////////////////////////////////

#define reggSTD8			0x10
#define reggSTD16			0x20
#define reggSTD32			0x30

#define reggSEG				0x40

#define reggDBG				0x80
#define reggCTRL			0x90
#define reggTST				0xa0

#define reggST				0xb0 // floating point stack
#define reggMM				0xc0
#define reggXMM				0xd0

#define reggSTD8_16_32			0xe0 // only used when decoding instructions
#define reggSTD16_32			0xf0 // only used when decoding instructions

#define reggTYPE_MASK			0xf0

#define reggNAME_MASK			0x07

#define REG_TYPE(reg) (reg & reggTYPE_MASK)
#define REG_NAME(reg) (reg & reggNAME_MASK)

#define IS_REG_TYPE_REAL(reg) (REG_TYPE(reg) <= reggXMM) // an existing type, not one of 16_32 or 8_16_32 that do not really know it's size

#define regAL				(reggSTD8 | 0x00)
#define regCL				(reggSTD8 | 0x01)
#define regDL				(reggSTD8 | 0x02)
#define regBL				(reggSTD8 | 0x03)
#define regAH				(reggSTD8 | 0x04)
#define regCH				(reggSTD8 | 0x05)
#define regDH				(reggSTD8 | 0x06)
#define regBH				(reggSTD8 | 0x07)

#define reggAX				0x00
#define reggCX				0x01
#define reggDX				0x02
#define reggBX				0x03
#define reggSP				0x04
#define reggBP				0x05
#define reggSI				0x06
#define reggDI				0x07

#define IS_REG_NAME_SP_BP(reg) ((reg & 0x06) == 0x04)

#define regAX				(reggSTD16 | 0x00)
#define regCX				(reggSTD16 | 0x01)
#define regDX				(reggSTD16 | 0x02)
#define regBX				(reggSTD16 | 0x03)
#define regSP				(reggSTD16 | 0x04)
#define regBP				(reggSTD16 | 0x05)
#define regSI				(reggSTD16 | 0x06)
#define regDI				(reggSTD16 | 0x07)

#define regEAX				(reggSTD32 | 0x00)
#define regECX				(reggSTD32 | 0x01)
#define regEDX				(reggSTD32 | 0x02)
#define regEBX				(reggSTD32 | 0x03)
#define regESP				(reggSTD32 | 0x04)
#define regEBP				(reggSTD32 | 0x05)
#define regESI				(reggSTD32 | 0x06)
#define regEDI				(reggSTD32 | 0x07)

#define regALeAX			(reggSTD8_16_32 | 0x00) // no_reg_name

#define regmNAME_EAX			(0x00)
#define regmNAME_ECX			(0x01)

#define regmNAME_AX			(0x00)
#define regmNAME_CX			(0x01)

#define regmNAME_AL			(0x00)
#define regmNAME_AH			(0x04)

#define regeAX				(reggSTD16_32 | 0x00) // no_reg_name
#define regeCX				(reggSTD16_32 | 0x01) // no_reg_name
#define regeDX				(reggSTD16_32 | 0x02) // no_reg_name
#define regeBX				(reggSTD16_32 | 0x03) // no_reg_name
#define regeSP				(reggSTD16_32 | 0x04) // no_reg_name
#define regeBP				(reggSTD16_32 | 0x05) // no_reg_name
#define regeSI				(reggSTD16_32 | 0x06) // no_reg_name
#define regeDI				(reggSTD16_32 | 0x07) // no_reg_name

#define regES				(reggSEG | 0x00)
#define regCS				(reggSEG | 0x01)
#define regSS				(reggSEG | 0x02)
#define regDS				(reggSEG | 0x03)
#define regFS				(reggSEG | 0x04)
#define regGS				(reggSEG | 0x05)

#define reggSEG_MASK (0x3f) // (1 | (1<<2) | ... | (1<<5))

#define regCR0				(reggCTRL | 0x00)
#define regCR2				(reggCTRL | 0x02)
#define regCR3				(reggCTRL | 0x03)
#define regCR4				(reggCTRL | 0x04)

#define reggCTRL_MASK (1 | (1<<2) | (1<<3) | (1<<4))

#define regDR0				(reggDBG | 0x00)
#define regDR1				(reggDBG | 0x01)
#define regDR2				(reggDBG | 0x02)
#define regDR3				(reggDBG | 0x03)
#define regDR4				(reggDBG | 0x04)
#define regDR5				(reggDBG | 0x05)
#define regDR6				(reggDBG | 0x06)
#define regDR7				(reggDBG | 0x07)

#define regTST0				(reggTST | 0x00)
#define regTST1				(reggTST | 0x01)
#define regTST2				(reggTST | 0x02)
#define regTST3				(reggTST | 0x03)
#define regTST4				(reggTST | 0x04)
#define regTST5				(reggTST | 0x05)
#define regTST6				(reggTST | 0x06)
#define regTST7				(reggTST | 0x07)

#define regST0				(reggST | 0x00)
#define regST1				(reggST | 0x01)
#define regST2				(reggST | 0x02)
#define regST3				(reggST | 0x03)
#define regST4				(reggST | 0x04)
#define regST5				(reggST | 0x05)
#define regST6				(reggST | 0x06)
#define regST7				(reggST | 0x07)

#define regMM0				(reggMM | 0x00)
#define regMM1				(reggMM | 0x01)
#define regMM2				(reggMM | 0x02)
#define regMM3				(reggMM | 0x03)
#define regMM4				(reggMM | 0x04)
#define regMM5				(reggMM | 0x05)
#define regMM6				(reggMM | 0x06)
#define regMM7				(reggMM | 0x07)

#define regXMM0				(reggXMM | 0x00)
#define regXMM1				(reggXMM | 0x01)
#define regXMM2				(reggXMM | 0x02)
#define regXMM3				(reggXMM | 0x03)
#define regXMM4				(reggXMM | 0x04)
#define regXMM5				(reggXMM | 0x05)
#define regXMM6				(reggXMM | 0x06)
#define regXMM7				(reggXMM | 0x07)

//////////////////////////////////////////////////////////////////////

// These are so the 'variables' corresponding to the memory addresses expressed by memory operands
// also can be stored in RegisterState along with the real registers.
// This is a not-so-nice solution (in that it is a hack), so any
// good alternative ideas are welcome. Except unreasonably memory-consuming ones.

#define reg_fakeMEMADDR_DEST   0xfe
#define reg_fakeMEMADDR_SRC    0xff

//////////////////////////////////////////////////////////////////////

#define REGNAME16_32FROM8(name)	(name & 0x03)  // al/ah->(e)ax, cl/ch->(e)cx...
#define ISREGNAME8LOWER(name) (!(name & 0x04))
#define ISREGNAME8HIGHER(name) (!!(name & 0x04))

struct RegIdNameIndex
{
  byte  m_nReg;
  short m_nOffset;
};

extern char cRegNames[];
extern RegIdNameIndex indexRegIdName[];

//////////////////////////////////////////////////////////////////////

const char* GetRegisterName (byte bReg);

word GetRegisterSize(byte bReg);

//////////////////////////////////////////////////////////////////////

class OperandRegister : public Operand
{
 public:
  OperandRegister(byte bReg);
  virtual ~OperandRegister();
  
 public:
  byte	m_bReg;

 public:
  virtual char* GenAsmString (char* pDest) const;
  
 public:
  virtual word GetOpSize () const { return GetRegisterSize(m_bReg); }
};

#endif // !defined(__OPERAND_REGISTER_H__INCLUDED_)
