/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// PEFile.h: interface for the PEFile class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__P_E_FILE_H__INCLUDED_)
#define __P_E_FILE_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "GlobalDefs.h" 
#include "ExecutableImage.h" 
struct IMAGE_MZ_HEADER; 
class PEFileSectionBody;
class DataPtrPESectionBody;
class GUIResources; 

#ifdef WIN32
#pragma pack(push)
#pragma pack(2)
#endif

// The definitions for the file format were gotten from winnt.h of
// the wine project

#define IMAGE_NT_SIGNATURE     0x00004550 /* PE00 */

//////////////////////////////////////////////////////////////////////

#define BYTE byte
#define WORD word
#define DWORD dword

struct IMAGE_PE_FILE_HEADER // Dumpable
{
  /* signature this is not part of this struct according to the spec,
     but making it so simplifies my program a lot (because I do not
     have to care about a loose dword) (and does not break
     compatibility) */

  dword dwSignature;               // Signature
  WORD  wMachine;                  // Machine
  WORD  wNumberOfSections;         // # Sections
  DWORD dwTimeDateStamp;           // Time & Date
  DWORD dwPointerToSymbolTable;    // Ptr To Symbol Table
  DWORD dwNumberOfSymbols;         // # Symbols
  WORD  wSizeOfOptionalHeader;     // Opt Header Size
  WORD  wCharacteristics;          // Characteristics
};

//////////////////////////////////////////////////////////////////////
// IMAGE_PE_FILE_HEADER.wMachine

#define IMAGE_PE_MACHINE_I386			0x014c
#define IMAGE_PE_MACHINE_I486			0x014d
#define IMAGE_PE_MACHINE_I586			0x014e

//////////////////////////////////////////////////////////////////////

/* Directory Entries, indices into the DataDirectory array */

#define IMAGE_DIRECTORY_ENTRY_EXPORT            0
#define IMAGE_DIRECTORY_ENTRY_IMPORT            1
#define IMAGE_DIRECTORY_ENTRY_RESOURCE          2
#define IMAGE_DIRECTORY_ENTRY_EXCEPTION         3
#define IMAGE_DIRECTORY_ENTRY_SECURITY          4
#define IMAGE_DIRECTORY_ENTRY_BASERELOC         5
#define IMAGE_DIRECTORY_ENTRY_DEBUG             6
#define IMAGE_DIRECTORY_ENTRY_COPYRIGHT         7
#define IMAGE_DIRECTORY_ENTRY_GLOBALPTR         8   /* (MIPS GP) */
#define IMAGE_DIRECTORY_ENTRY_TLS               9
#define IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG       10
#define IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT      11
#define IMAGE_DIRECTORY_ENTRY_IAT               12  /* Import Address Table */
#define IMAGE_DIRECTORY_ENTRY_DELAY_IMPORT      13
#define IMAGE_DIRECTORY_ENTRY_COM_DESCRIPTOR    14

#define IMAGE_NUMBEROF_DIRECTORY_ENTRIES 16

//////////////////////////////////////////////////////////////////////

struct IMAGE_PE_DATA_DIRECTORY // Dumpable
{
  DWORD dwVirtualAddress;           // Virtual address
  DWORD dwSize;                     // Size
};

//////////////////////////////////////////////////////////////////////

#define IMAGE_NT_OPT_MAGIC              0x010b

struct IMAGE_PE_OPTIONAL_HEADER // Dumpable
{  

  /* Standard fields */

  WORD    wMagic;                         // Magic
  BYTE    bMajorLinkerVersion;            // Linker version (major) 
  BYTE    bMinorLinkerVersion;            // Linker version (minor)
  DWORD   dwSizeOfCode;                   // Size of code 
  DWORD   dwSizeOfInitializedData;        // Size of initialized data 
  DWORD   dwSizeOfUninitializedData;      // Size of uninitialized data
  DWORD   dwAddressOfEntryPoint;          // Entry point
  DWORD   dwBaseOfCode;                   // Base of code
  DWORD   dwBaseOfData;                   // Base of data  

  /* NT additional fields */

  DWORD   dwImageBase;                    // Image base 
  DWORD   dwSectionAlignment;             // Section alignment
  DWORD   dwFileAlignment;                // File alignment
  WORD    wMajorOperatingSystemVersion;   // OS version (major) 
  WORD    wMinorOperatingSystemVersion;   // OS version (minor)
  WORD    wMajorImageVersion;             // Image version (major) 
  WORD    wMinorImageVersion;             // Image version (minor)
  WORD    wMajorSubsystemVersion;         // Subsystem version (major) 
  WORD    wMinorSubsystemVersion;         // Subsystem version (minor)
  DWORD   dwReserved1;               
  DWORD   dwSizeOfImage;                  // Size of Image
  DWORD   dwSizeOfHeaders;                // Size of headers   
  DWORD   dwCheckSum;                     // Checksum 
  WORD    wSubsystem;                     // Subsystem
  WORD    wDllCharacteristics;            // Dll Characteristics
  DWORD   dwSizeOfStackReserve;           // Stack reserve
  DWORD   dwSizeOfStackCommit;            // Stack commit 
  DWORD   dwSizeOfHeapReserve;            // Heap reserve
  DWORD   dwSizeOfHeapCommit;             // Heap commit
  DWORD   dwLoaderFlags;                  // Loader flags
  DWORD   dwNumberOfRvaAndSizes;          // # Rva and sizes
  IMAGE_PE_DATA_DIRECTORY DataDirectory[IMAGE_NUMBEROF_DIRECTORY_ENTRIES];
};

//#define IMAGE_SIZEOF_FILE_HEADER             20

#define IMAGE_SIZEOF_SHORT_NAME 8

//////////////////////////////////////////////////////////////////////

struct IMAGE_SECTION_HEADER // Dumpable
{
  byte   bName[IMAGE_SIZEOF_SHORT_NAME];  // Name
  union {
    DWORD   dwPhysicalAddress;            // Physical address
    DWORD   dwVirtualSize;                // Virtual size
  } Misc;
  DWORD   dwVirtualAddress;               // Virtual address
  DWORD   dwSizeOfRawData;                // Size of raw data
  DWORD   dwPointerToRawData;             // Ptr to raw data
  DWORD   dwPointerToRelocations;         // Ptr to relocatoins
  DWORD   dwPointerToLinenumbers;         // Ptr to linenumbers
  WORD    wNumberOfRelocations;           // # Relocations
  WORD    wNumberOfLinenumbers;           // # Line numbers
  DWORD   dwCharacteristics;              // Characteristics
};

//////////////////////////////////////////////////////////////////////

#ifdef WIN32
#pragma pack(pop)
#endif

#include "UrlFolder.h" // dep_chk_needed     

class PEFile : public ExecutableImage
{
 public:
  PEFile(const string& strFileName);
  virtual ~PEFile();

 protected:
  IMAGE_MZ_HEADER* mzHeaderAndData;
  
  IMAGE_PE_FILE_HEADER peHeader;
  IMAGE_PE_OPTIONAL_HEADER peOptHeader;
  
  IMAGE_SECTION_HEADER* pSectionHeaders;

  PEFileSectionBody* pSectionBodies;
  
  GUIResources*	 m_pGUIResources;
  
 public:
  void PrintInfo ();
  
 public:
  virtual void HTMLFileInfo (UrlFolder* pParent);
  
 public:
  static bool IsType (int file);
  
 public:
  /*
    byte GetByte (dword nVirtualAddress);
    word GetWord (dword nVirtualAddress);
    unsigned int GetDWord (dword nVirtualAddress);
  */
  DataPtrPESectionBody* GetDataPtr (dword dwVirtualAddress);
  
 public:
  virtual void ReadFile (int file);
  virtual void PostReadFile (); // process things easily dechiphered (not requiring any intelligence)
  virtual bool ProcessFile (UrlFolder* pFolder); // process instructions
  
 public:
  virtual ExecutableImage* GetWriteImage(const string& strFileName);
  
  virtual bool PrepareWrite();
  virtual bool WriteFile();
  
 public:
  RelocatableDest* CreateCodeSection();
  
 protected:
  void ReadImports ();
  void ReadExports ();
  
  void ReadResources ();
  
 public:

 protected:
  //	bool DisAssemble (UrlFolder* pParent);
  
 protected:
  virtual void GetExportedCallTargets (DisAssembler& disasm);
  //	virtual void GetCallTargets (DisAssembler& disasm);

 protected:
  virtual DataPtr* GetDisAsmDataPtr (dword dwVirtualAddress);
  
 protected:
  void DecodeInstruction (unsigned int *IP);
};

#endif // !defined(__P_E_FILE_H__INCLUDED_)
