/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// RegisterState.cpp: implementation of the RegisterState class.
//
//////////////////////////////////////////////////////////////////////

#include "RegisterState.h" 
#include "OperandRegister.h"
#include "VariableN.h" 
#include "VariableNSplit.h" 
#include "VariableNConst.h" 
#include "FlagsState.h"

#include "Misc.h"

//////////////////////////////////////////////////////////////////////

FlagsStateUnknown flagsStateUnknown; // used for all unknown
FlagsStateUnaffected flagsStateUnaffected; // used for all unaffected

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

RegisterState::RegisterState()
{
  // m_pParent = NULL;
  m_pFlags = &flagsStateUnaffected; // be careful not to delete this one
}

/*
void RegisterState::SetFlagsUnaffected()
{
  assert (m_pFlags != &flagsStateUnaffected);

  assert (m_pFlags == &flagsStateUnknown || m_pFlags == &flagsStateUnaffected); // or we need to delete previous
  m_pFlags = &flagsStateUnaffected;
}
*/
void RegisterState::SetFlagsUnknown()
{
  assert (m_pFlags == &flagsStateUnknown || 
	  m_pFlags == &flagsStateUnaffected); // or we need to delete previous
  m_pFlags = &flagsStateUnknown;
}

//RegisterState::RegisterState(/*RegisterState* pParent*/)
//{
//  // m_pParent = pParent;
//  m_pFlags = &flagsStateUnknown; // be careful not to delete this one
//}

RegisterState::~RegisterState()
{
  
}

//////////////////////////////////////////////////////////////////////

void RegisterState::SetFlags(FlagsState* pFlags)
{
  assert (m_pFlags == &flagsStateUnknown || 
	  m_pFlags == &flagsStateUnaffected); 
  // do not delete flagsStateUnknown
  
  m_pFlags = pFlags;
}

bool RegisterCollision(byte bReg1,byte bReg2)
{
  if (bReg1 == bReg2)
    return true;

  assert (REG_TYPE(bReg1) >= reggSTD8);
  assert (REG_TYPE(bReg2) >= reggSTD8);
  
  if (REG_TYPE(bReg1) > reggSTD32 ||
      REG_TYPE(bReg2) > reggSTD32)
    return false;

  if (REG_TYPE(bReg1) == REG_TYPE(bReg2))
    return false; // there is no collision

  if (REG_TYPE(bReg1) < REG_TYPE(bReg2))
    swap(bReg1,bReg2); // so we have any 8-bit var in reg 1

  if (REG_TYPE(bReg1) == reggSTD8)
    return REGNAME16_32FROM8(bReg1) == REG_NAME(bReg2); // AL/AH - eAX ?, etc

  return REG_NAME(bReg1) == REG_NAME(bReg2); // AX - EAX ?, etc
}

void RegisterState::SetRegisterVar (byte bReg,Variable* pVar)
{
  assert (pVar);

  byte bRegType = bReg & reggTYPE_MASK;
  
  assert (bRegType == reggSTD8 ||
	  bRegType == reggSTD16 ||
	  bRegType == reggSTD32 ||
	  bRegType == reggSEG ||
	  bReg == reg_fakeMEMADDR_DEST ||
	  bReg == reg_fakeMEMADDR_SRC);
  
  assert ((bRegType != reggSTD8)  || IS_TYPE(Variable8*,pVar));
  assert ((bRegType != reggSTD16) || IS_TYPE(Variable16*,pVar));
  assert ((bRegType != reggSTD32) || IS_TYPE(Variable32*,pVar));
  
  assert ((bRegType != reggSEG)   || IS_TYPE(Variable16*,pVar));	

  // make sure no overwrite happen

#ifndef NDEBUG
  FOR_EACH_ITER_REVERSE(VectRegVar,m_vectRegVars,regvar)
    assert (!RegisterCollision(bReg,regvar->m_bReg));
#endif

  m_vectRegVars.push_back(RegVar(bReg,pVar));

/*	
#define reggSTD8			0x10
#define reggSTD16			0x20
#define reggSTD32			0x30

#define reggSEG				0x40

#define reggDBG				0x80
#define reggCTRL			0x90
#define reggTST				0xa0

#define reggST				0xb0 // floating point stack
#define reggMM				0xc0
#define reggXMM				0xd0

#define reggSTD8_16_32		0xe0 // only used when decoding instructions
#define reggSTD16_32		0xf0 // only used when decoding instructions

#define reggTYPE_MASK		0xf0

#define reggNAME_MASK		0x07
*/
}

void RegisterState::SetRegisterVarContent(byte bReg,VarContent* pVarContent)
{
  // must not have a variable content before or SetRegisterVar will
  // 'overwrite' the variable in question also not therefore that
  // VarContent cannot be set before variable

  // ok, so see if we're already in the list

  FOR_EACH_ITER(VectRegVar,m_vectRegVars,regvar)
    {
      if (regvar->m_bReg == bReg)
	{
	  regvar->m_pVar->SetContent(pVarContent);
	  return; // we're done
	}      
    }
  
  // aha, we do not have an variable, so we'll have to set one (that
  // has unknown content) for the integer registers we'll use
  // VariableNUnknown
  
  Variable* pVar = NULL;
  
  switch (REG_TYPE(bReg))
    {
    case reggSTD8:
      pVar = new Variable8Unknown();
      break;
    case reggSTD16:
    case reggSEG:
      pVar = new Variable16Unknown();
      break;
    case reggSTD32:
      pVar = new Variable32Unknown();
      break;
    default:
      assert (false); // not supported yet
    }
  
  assert (pVar);
  
  pVar->SetContent(pVarContent);
  SetRegisterVar(bReg,pVar);
}


void RegisterState::SetStackVar (sdword dwOffset,Variable* pVar)
{
  // get len from var
  
  assert (pVar); // may not be NULL to define an unknown value, because we need to get the size from it  
  assert (pVar->GetSize() <= 0xff);
  
  
  // we should check that the newly added variable does not overlap
  // with any of the other ones set!  this cannot be relaxed as the
  // info is put into a map and we could not figure out who came there
  // first
  
#ifndef NDEBUG
  MapSOffsetVar::const_iterator iter = m_mapStackVars.lower_bound(dwOffset);
  
  dword dwSize = pVar->GetSize();

  // we'll get an iterator that has offset >= dwOffset,
  // or an end iterator, if there is nothing after us
  
  // so if the iterator is valid, it should be after us

  if (iter != m_mapStackVars.end())
    assert (iter->first >= dwOffset + (sdword) dwSize); 
  
  // and now backup one step, also ok if iter was .end()
  // anything should be before us in memory
  
  if (iter != m_mapStackVars.begin())
    {
      --iter;
      assert (iter->first + (sdword) iter->second->GetSize() <= dwOffset); 
    }
#endif 
  
  m_mapStackVars.insert (MapSOffsetVar::value_type(dwOffset,pVar));
}

bool RegisterState::HasSameContent(RegisterState* pCompare)
{
  // TODO: compare contents of flags

  // TODO: compare that contents of variables are the same (i.e. same
  // range or const and so)

  // for now we just make sure the same registers are set (which would
  // usually be the case, so the above is necessary)

  // as the variables are not in a set (perhaps they should) we'll
  // have to iterate through the vectors twice. But usually the
  // vectors have only one or two elements..

  // we must make sure both are the same, and none have any extra, not
  // that just all vars in this are also in the other

  if (m_vectRegVars.size() != pCompare->m_vectRegVars.size()) 
    return false; 
  
  // comparing number of items works together with the detailed checks
  // below because no overlaps are allowed
  
  FOR_EACH_ITER(VectRegVar,m_vectRegVars,regvar)
    {
      FOR_EACH_ITER(VectRegVar,pCompare->m_vectRegVars,regvar_cmp)
	{
	  if (regvar->m_bReg == regvar_cmp->m_bReg)
	    {
	      // compare the varaibles contents...

	      // if mismatch, return false


	      // same...
	      goto next_variable; // we could really use named loops in C, so I could use instead continue next_variable,
	      // and next_variable was the name of the outermost loop..
	    }
	}
      // aha, so variable not found.
      return false;
    next_variable:;      
    }
  return true;
}

Variable* RegisterState::GetRegisterNoFixUnknown(byte bReg)
{
  FOR_EACH_ITER_REVERSE(VectRegVar,m_vectRegVars,regvar)
    {
      if (regvar->m_bReg == bReg)
	return regvar->m_pVar;
    }
  return NULL;
}

Variable* RegisterState::GetRegister (byte bReg)
{
  FOR_EACH_ITER_REVERSE(VectRegVar,m_vectRegVars,regvar)
    {
      if (regvar->m_bReg == bReg)
	return regvar->m_pVar;
    }
  return NULL;
}

struct GeneralPurposeRegs
{
 public:
  GeneralPurposeRegs() { bReg32 = 0; bReg16 = 0; bReg8High = 0; bReg8Low = 0; pVar32 = NULL; pVar16 = 0; pVar8High = 0; pVar8Low = 0; }
  
 public:
  byte bReg32;
  byte bReg16;
  byte bReg8High;
  byte bReg8Low;
  
  Variable32* pVar32;
  Variable16* pVar16;
  Variable8* pVar8High;
  Variable8* pVar8Low;
};

#define GEN_REG_LOW8	0x01 //    x
#define GEN_REG_HIGH8	0x02 //   x
#define GEN_REG_LOW16   0x04 //   xx
#define GEN_REG_LOW32	0x08 // xxxx

byte RegisterState::GetRegisterGeneralPurpose (GeneralPurposeRegs& getregs)
{
  assert (!getregs.pVar32);
  assert (!getregs.pVar16);
  assert (!getregs.pVar8High);
  assert (!getregs.pVar8Low);

  byte bFound = 0;
  FOR_EACH_ITER_REVERSE(VectRegVar,m_vectRegVars,regvar)
    {
      assert (regvar->m_pVar);

      if      (regvar->m_bReg == getregs.bReg32)    
	{ 
	  ASSERT_IS_TYPE(Variable32*,regvar->m_pVar); 
	  getregs.pVar32    = (Variable32*) regvar->m_pVar; 
	  assert (!bFound); // or would have overlap
	  return GEN_REG_LOW32;
	}
      else if (regvar->m_bReg == getregs.bReg16)    
	{ 
	  ASSERT_IS_TYPE(Variable16*,regvar->m_pVar); 
	  getregs.pVar16    = (Variable16*) regvar->m_pVar; 
	  assert (!bFound); // or would have overlap
	  return GEN_REG_LOW16;
	}
      else if (regvar->m_bReg == getregs.bReg8High) 
	{ 
	  ASSERT_IS_TYPE(Variable8*,regvar->m_pVar); 
	  getregs.pVar8High = (Variable8*) regvar->m_pVar; 
	  assert (!getregs.pVar16);
	  bFound |= GEN_REG_HIGH8;
	}
      else if (regvar->m_bReg == getregs.bReg8Low)  
	{ 
	  ASSERT_IS_TYPE(Variable8*,regvar->m_pVar); 
	  getregs.pVar8Low  = (Variable8*) regvar->m_pVar; 
	  assert (!getregs.pVar16);
	  bFound |= GEN_REG_LOW8;
	}
    }
  return bFound;
}

//////////////////////////////////////////////////////////////////////

RegisterStateUnknown::RegisterStateUnknown()
{
  m_pFlags = &flagsStateUnknown;
}

RegisterStateUnknown::~RegisterStateUnknown()
{

}

//////////////////////////////////////////////////////////////////////

Variable* RegisterStateUnknown::GetRegister (byte bReg)
{
  Variable* pVar = RegisterState::GetRegister(bReg);
  
  if (pVar)
    return pVar;
  
  // not created before, add it...
  
  // TODO: fix for registers not being integers, and not of size 1, 2
  // or 4...  VariableN_Unknown will assert false as it is now
  
  pVar = VariableN_Unknown(GetRegisterSize(bReg));
  
  m_vectRegVars.push_back(RegVar(bReg,pVar));
  
  return pVar;
}

byte RegisterStateUnknown::GetRegisterGeneralPurpose (GeneralPurposeRegs& getregs)
{
  assert (!getregs.pVar32);
  assert (!getregs.pVar16);
  assert (!getregs.pVar8High);
  assert (!getregs.pVar8Low);
  
  byte bFound = RegisterState::GetRegisterGeneralPurpose(getregs);
  
  if (bFound == GEN_REG_LOW32 || bFound == GEN_REG_LOW16)
    return bFound;
  
  if (getregs.bReg32) // looking for 32-reg, so give them that...
    {
      getregs.pVar32 = new Variable32Unknown();
      m_vectRegVars.push_back(RegVar(getregs.bReg32,getregs.pVar32));
      return GEN_REG_LOW32;
    }
  if (getregs.bReg16) // looking for 16-reg, so give them that...
    {
      getregs.pVar16 = new Variable16Unknown();
      m_vectRegVars.push_back(RegVar(getregs.bReg16,getregs.pVar16));
      return GEN_REG_LOW16;
    }
  if (getregs.bReg8High && !(bFound & GEN_REG_HIGH8)) // looking for 16-reg, so give them that...
    {
      getregs.pVar8High = new Variable8Unknown();
      m_vectRegVars.push_back(RegVar(getregs.bReg8High,getregs.pVar8High));
      bFound |= GEN_REG_HIGH8;
    }
  if (getregs.bReg8Low && !(bFound & GEN_REG_LOW8)) // looking for 16-reg, so give them that...
    {
      getregs.pVar8Low = new Variable8Unknown();
      m_vectRegVars.push_back(RegVar(getregs.bReg8Low,getregs.pVar8Low));
      bFound |= GEN_REG_LOW8;
    }
  return bFound;  
}

//////////////////////////////////////////////////////////////////////

RegisterStateEntry::RegisterStateEntry()
{

}

RegisterStateEntry::~RegisterStateEntry()
{

}

//////////////////////////////////////////////////////////////////////

Variable* RegisterStateEntry::GetRegister (byte bReg)
{
  m_used.AddReg(bReg);

  Variable* pVar = RegisterState::GetRegister(bReg);
  
  if (pVar)
    return pVar;
  
  // see if we have created it before..
  
  FOR_EACH_ITER_REVERSE(VectRegVar,m_vectRegVarsEntry,regvar)
    {
      if (regvar->m_bReg == bReg)
	return regvar->m_pVar;
    }
  
  // not created before, add it...
  
  // TODO: fix for registers not being integers, and not of size 1, 2 or 4...
  // VariableN_Unknown will assert false as it is now
  
  pVar = VariableN_Unknown(GetRegisterSize(bReg));
  
  m_vectRegVarsEntry.push_back(RegVar(bReg,pVar));
  
  return pVar;
}

byte RegisterStateEntry::GetRegisterGeneralPurpose (GeneralPurposeRegs& getregs)
{
  if (getregs.bReg32)
    m_used.AddReg(getregs.bReg32);
  else if (getregs.bReg16)
    m_used.AddReg(getregs.bReg16);
  else if (getregs.bReg8High)
    {
      assert (!getregs.bReg8Low); // noone asks for AL and AH simultaneously, without also asking for AX
      m_used.AddReg(getregs.bReg8High);
    }
  else 
    {
      assert (getregs.bReg8Low);
      m_used.AddReg(getregs.bReg8Low);
    }
  
  // TODO: RegisterStateEntry::GetRegisterGeneralPurpose *need*
  // rewrite, because unlike other GetRegisterGeneralPurpose it *must*
  // return something at least as wide as was requested. So even if
  // somebody was helpful where the call occured and set AX, if we
  // request EAX, then EAX must be returned, with the low part set to
  // the actual AX and the high part as unknown, this is necessary so
  // we can trace what happen to variables
  //
  // however the code for setting m_used above is ok. It will record
  // what is requested anyway

  assert (!getregs.pVar32);
  assert (!getregs.pVar16);
  assert (!getregs.pVar8High);
  assert (!getregs.pVar8Low);

  byte bFound = RegisterState::GetRegisterGeneralPurpose(getregs);

  if (bFound == GEN_REG_LOW32 || bFound == GEN_REG_LOW16)
    return bFound;

  FOR_EACH_ITER_REVERSE(VectRegVar,m_vectRegVarsEntry,regvar)
    {
      assert (regvar->m_pVar);

      if      (regvar->m_bReg == getregs.bReg32)    
	{ 
	  ASSERT_IS_TYPE(Variable32*,regvar->m_pVar); 
	  getregs.pVar32    = (Variable32*) regvar->m_pVar; 
	  assert (!bFound); // or would have overlap
	  return GEN_REG_LOW32;
	}
      else if (regvar->m_bReg == getregs.bReg16)    
	{ 
	  ASSERT_IS_TYPE(Variable16*,regvar->m_pVar); 
	  getregs.pVar16    = (Variable16*) regvar->m_pVar; 
	  assert (!bFound); // or would have overlap
	  return GEN_REG_LOW16;
	}
      else if (regvar->m_bReg == getregs.bReg8High) 
	{ 
	  ASSERT_IS_TYPE(Variable8*,regvar->m_pVar); 
	  getregs.pVar8High = (Variable8*) regvar->m_pVar; 
	  assert (!getregs.pVar32 && !getregs.pVar16);
	  bFound |= GEN_REG_HIGH8;
	}
      else if (regvar->m_bReg == getregs.bReg8Low)  
	{ 
	  ASSERT_IS_TYPE(Variable8*,regvar->m_pVar); 
	  getregs.pVar8Low  = (Variable8*) regvar->m_pVar; 
	  assert (!getregs.pVar32 && !getregs.pVar16);
	  bFound |= GEN_REG_LOW8;
	}
    }
  if (getregs.bReg32) // looking for 32-reg, so give them that...
    {
      getregs.pVar32 = new Variable32Unknown();
      m_vectRegVarsEntry.push_back(RegVar(getregs.bReg32,getregs.pVar32));
      return GEN_REG_LOW32;
    }
  if (getregs.bReg16) // looking for 16-reg, so give them that...
    {
      getregs.pVar16 = new Variable16Unknown();
      m_vectRegVarsEntry.push_back(RegVar(getregs.bReg16,getregs.pVar16));
      return GEN_REG_LOW16;
    }
  if (getregs.bReg8High && !(bFound & GEN_REG_HIGH8)) // looking for 16-reg, so give them that...
    {
      getregs.pVar8High = new Variable8Unknown();
      m_vectRegVarsEntry.push_back(RegVar(getregs.bReg8High,getregs.pVar8High));
      bFound |= GEN_REG_HIGH8;
    }
  if (getregs.bReg8Low && !(bFound & GEN_REG_LOW8)) // looking for 16-reg, so give them that...
    {
      getregs.pVar8Low = new Variable8Unknown();
      m_vectRegVarsEntry.push_back(RegVar(getregs.bReg8Low,getregs.pVar8Low));
      bFound |= GEN_REG_LOW8;
    }
  return bFound;  
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

RegisterStateCombine::RegisterStateCombine()
{
  printf ("RegisterStateCombine::RegisterStateCombine\n");

}

RegisterStateCombine::~RegisterStateCombine()
{
  printf ("RegisterStateCombine::~1RegisterStateCombine\n");
  
}

//////////////////////////////////////////////////////////////////////

Variable* RegisterStateCombine::GetRegister (byte bReg)
{
  printf ("RegisterStateCombine::GetRegister: %s\n",GetRegisterName(bReg));
  
  // first see if we're suppose to have it
  
  if (m_affected.m_setReg.find(bReg) == m_affected.m_setReg.end() &&
      m_unknown.m_setReg.find(bReg)  == m_unknown.m_setReg.end())
    {
      // we're not suppose to provide it
      return NULL;
    }
  
  // ok, so we're to provide it first see if we have provided it
  // before (i.e. already done the merging)

  Variable* pVar = RegisterState::GetRegister(bReg);
  
  if (pVar)
    return pVar;

  // bad luck. We'll have to merge again...

  // so know we'll have to get every source and put them
  // together. Humpty Dumpty.
  
  
  

  
  pVar = VariableN_Unknown(GetRegisterSize(bReg));
  
  m_vectRegVars.push_back(RegVar(bReg,pVar));
  
  return pVar;
}

byte RegisterStateCombine::GetRegisterGeneralPurpose (GeneralPurposeRegs& getregs)
{
  printf ("RegisterStateCombine::GetRegister: %d %d\n",getregs.bReg32,getregs.bReg16);
  
  // first see what we've already decided to provide
  
  byte bFound = RegisterState::GetRegisterGeneralPurpose(getregs);
  
  if (bFound == GEN_REG_LOW32 || bFound == GEN_REG_LOW16)
    return bFound; // no need to check further 

  // ok, so are we instructed to fix any remaining requests?
  
  if (getregs.bReg32) // looking for 32-reg, so give them that...
    {
      if (m_affected.m_setReg.find(getregs.bReg32) != 
	  m_affected.m_setReg.end())
	getregs.pVar32 = new Variable32Unknown(); // TODO: merge!
      else if (m_unknown.m_setReg.find(getregs.bReg32)  == 
	       m_unknown.m_setReg.end())
	getregs.pVar32 = new Variable32Unknown();
      else 
	goto try_reg_16; // is this nasty or??
      m_vectRegVars.push_back(RegVar(getregs.bReg32,getregs.pVar32));
      return GEN_REG_LOW32;
    }
 try_reg_16:
  if (getregs.bReg16) // looking for 32-reg, so give them that...
    {
      if (m_affected.m_setReg.find(getregs.bReg16) != 
	  m_affected.m_setReg.end())
	getregs.pVar16 = new Variable16Unknown(); // TODO: merge!
      else if (m_unknown.m_setReg.find(getregs.bReg16)  == 
	       m_unknown.m_setReg.end())
	getregs.pVar16 = new Variable16Unknown();
      else 
	goto try_reg_8high; // is this nasty or??
      m_vectRegVars.push_back(RegVar(getregs.bReg16,getregs.pVar16));
      return GEN_REG_LOW16;
    }
 try_reg_8high:
  if (getregs.bReg8High && !(bFound & GEN_REG_HIGH8)) // looking for 32-reg, so give them that...
    {
      if (m_affected.m_setReg.find(getregs.bReg8High) != 
	  m_affected.m_setReg.end())
	getregs.pVar8High = new Variable8Unknown(); // TODO: merge!
      else if (m_unknown.m_setReg.find(getregs.bReg8High)  == 
	       m_unknown.m_setReg.end())
	getregs.pVar8High = new Variable8Unknown();
      else 
	goto try_reg_8low; // is this nasty or??
      m_vectRegVars.push_back(RegVar(getregs.bReg8High,getregs.pVar8High));
      bFound |= GEN_REG_HIGH8;
    }
 try_reg_8low:
  if (getregs.bReg8Low && !(bFound & GEN_REG_HIGH8)) // looking for 32-reg, so give them that...
    {
      if (m_affected.m_setReg.find(getregs.bReg8Low) != 
	  m_affected.m_setReg.end())
	getregs.pVar8Low = new Variable8Unknown(); // TODO: merge!
      else if (m_unknown.m_setReg.find(getregs.bReg8Low)  == 
	       m_unknown.m_setReg.end())
	getregs.pVar8Low = new Variable8Unknown();
      else 
	goto done_trying; // is this nasty or??
      m_vectRegVars.push_back(RegVar(getregs.bReg8Low,getregs.pVar8Low));
      bFound |= GEN_REG_LOW8;
    }
 done_trying:
  return bFound;
  
  /*
  assert (!getregs.pVar32);
  assert (!getregs.pVar16);
  assert (!getregs.pVar8High);
  assert (!getregs.pVar8Low);

  byte bFound = RegisterState::GetRegisterGeneralPurpose(getregs);

  if (bFound == GEN_REG_LOW32)
    return bFound;

  FOR_EACH_ITER_REVERSE(VectRegVar,m_vectRegVarsEntry,regvar)
    {
      assert (regvar->m_pVar);

      if      (regvar->m_bReg == getregs.bReg32)    
	{ 
	  ASSERT_IS_TYPE(Variable32*,regvar->m_pVar); 
	  getregs.pVar32    = (Variable32*) regvar->m_pVar; 
	  assert (!bFound); // or would have overlap
	  return GEN_REG_LOW32;
	}
      else if (regvar->m_bReg == getregs.bReg16)    
	{ 
	  ASSERT_IS_TYPE(Variable16*,regvar->m_pVar); 
	  getregs.pVar16    = (Variable16*) regvar->m_pVar; 
	  assert (!bFound); // or would have overlap
	  return GEN_REG_LOW16;
	}
      else if (regvar->m_bReg == getregs.bReg8High) 
	{ 
	  ASSERT_IS_TYPE(Variable8*,regvar->m_pVar); 
	  getregs.pVar8High = (Variable8*) regvar->m_pVar; 
	  assert (!getregs.pVar32 && !getregs.pVar16);
	  bFound |= GEN_REG_HIGH8;
	}
      else if (regvar->m_bReg == getregs.bReg8Low)  
	{ 
	  ASSERT_IS_TYPE(Variable8*,regvar->m_pVar); 
	  getregs.pVar8Low  = (Variable8*) regvar->m_pVar; 
	  assert (!getregs.pVar32 && !getregs.pVar16);
	  bFound |= GEN_REG_LOW8;
	}
    }
  if (getregs.bReg32) // looking for 32-reg, so give them that...
    {
      getregs.pVar32 = new Variable32Unknown();
      m_vectRegVarsEntry.push_back(RegVar(getregs.bReg32,getregs.pVar32));
      return GEN_REG_LOW32;
    }
  if (getregs.bReg16) // looking for 16-reg, so give them that...
    {
      getregs.pVar16 = new Variable16Unknown();
      m_vectRegVarsEntry.push_back(RegVar(getregs.bReg16,getregs.pVar16));
      return GEN_REG_LOW16;
    }
  if (getregs.bReg8High && !(bFound & GEN_REG_HIGH8)) // looking for 16-reg, so give them that...
    {
      getregs.pVar8High = new Variable8Unknown();
      m_vectRegVarsEntry.push_back(RegVar(getregs.bReg8High,getregs.pVar8High));
      bFound |= GEN_REG_HIGH8;
    }
  if (getregs.bReg8Low && !(bFound & GEN_REG_LOW8)) // looking for 16-reg, so give them that...
    {
      getregs.pVar8Low = new Variable8Unknown();
      m_vectRegVarsEntry.push_back(RegVar(getregs.bReg8Low,getregs.pVar8Low));
      bFound |= GEN_REG_LOW8;
    }
  return bFound;  
*/
}

bool RegisterStateAffected::HasSameContent(RegisterStateAffected* pCompare)
{
  printf ("RegisterStateAffected::HasSameContent   %d %d\n",
	  m_setReg.size(),pCompare->m_setReg.size());
  
  if (m_setReg != pCompare->m_setReg)
    return false;
  
  // TODO: compare stack

  return true;
}

bool RegisterStateCombine::HasSameContent(RegisterStateCombine* pCompare)
{
  printf ("RegisterStateCombine::HasSameContent   \n");

  if (!m_affected.HasSameContent(&pCompare->m_affected))
    return false;
  if (!m_unknown.HasSameContent(&pCompare->m_unknown))
    return false;

  // TODO: must compare and see that any variables that MAY be merged
  // would be merged into the same and note that the source states are
  // not the same (pointerwise)

  printf ("RegisterStateCombine::HasSameContent   true\n");

  return true;
}

//////////////////////////////////////////////////////////////////////

Variable8* GetVarGeneralPurpose8Reg (const RegisterStateParentPtr* pStateHolder,
				     byte bReg)
{
  // byte bRegName = bReg & reggNAME_MASK;
  
  // simply find the last occurance of the wanted register
  
  assert (pStateHolder);
  
  GeneralPurposeRegs getregs;
  
  getregs.bReg8Low = bReg;
  getregs.bReg16 = REGNAME16_32FROM8(bReg) | reggSTD16;
  getregs.bReg32 = REGNAME16_32FROM8(bReg) | reggSTD32;
  
  for ( ; ; )
    {
      RegisterState* pState = pStateHolder->GetRegisterState();
      
      //if (!pState)
      //return NULL;
      
      if (pState)
	{
	  byte bFound = pState->GetRegisterGeneralPurpose(getregs);
	  if (bFound)
	    switch (bFound)
	      {
	      case GEN_REG_LOW8:
		return getregs.pVar8Low;
	      case GEN_REG_LOW16:
		return getregs.pVar16->GetHalfPart(ISREGNAME8HIGHER(bReg));
	      case GEN_REG_LOW32:
		return getregs.pVar32->GetQuarterPart(ISREGNAME8HIGHER(bReg));
		DEFAULT_ASSERT_FALSE;
	      }
	}
      
      pStateHolder = pStateHolder->GetParent();
      if (!pStateHolder)
	return NULL;
    }
}

#define SET16_LOW8	0x01
#define SET16_HIGH8	0x02
#define SET16_BOTH8 (SET16_LOW8 | SET16_HIGH8)
#define SET32_LOW16		0x04

Variable16* UnionVar8And16(Variable8* pVar8Low,Variable8* pVar8High,
			   Variable16* pVar16,byte bSet)
{
  if (!bSet)
    return pVar16;
  
  // assert (bSet != SET16_BOTH8);
  
  assert (!!pVar8Low + !!pVar8High + !!pVar16 < 3);
  
  if (pVar16 && !(bSet & SET16_LOW8))
    pVar8Low = pVar16->GetHalfPart(0);
  if (pVar16 && !(bSet & SET16_HIGH8))
    pVar8High = pVar16->GetHalfPart(1);
  
  if (pVar8Low || pVar8High)
    return new Variable16Split(pVar8Low,pVar8High);
  return NULL;
}

Variable16* GetVarGeneralPurpose16Reg (const RegisterStateParentPtr* pStateHolder,
				       byte bReg)
{
  assert (pStateHolder);
  
  GeneralPurposeRegs getregs;
  
  byte bRegName = REG_NAME(bReg);
  
  getregs.bReg16 = bReg;
  getregs.bReg32 = bRegName | reggSTD32;
  if (bRegName < 4)
    {
      getregs.bReg8Low  =  bRegName      | reggSTD8;
      getregs.bReg8High = (bRegName + 4) | reggSTD8;
    }

  //byte bRegName = (bReg & reggNAME_MASK);
  //byte b32RegName = reggSTD32 | bRegName;
  
  Variable8* pVar8Low = NULL;
  Variable8* pVar8High = NULL;
  
  byte bSet = 0;
  
  for ( ; ; )
    {
      RegisterState* pState = pStateHolder->GetRegisterState();
      
      //if (!pState)
      //return NULL;
      
      if (pState)
	{
	  byte bFound = pState->GetRegisterGeneralPurpose(getregs);
	  if (bFound)
	    switch (bFound)
	      {
	      case GEN_REG_LOW16:
		return UnionVar8And16(pVar8Low,pVar8High,getregs.pVar16,bSet);
	      case GEN_REG_LOW32:
		assert (getregs.pVar32);
		return UnionVar8And16(pVar8Low,pVar8High,
				      getregs.pVar32->GetHalfPart(0),bSet);
	      default:
		assert (bFound & (GEN_REG_LOW8 | GEN_REG_HIGH8));
		if (getregs.pVar8Low)
		  {
		    assert (!pVar8Low && !(bSet & SET16_LOW8));
		    pVar8Low = getregs.pVar8Low;
		    getregs.pVar8Low = NULL;
		    getregs.bReg8Low = 0; // no need to look again
		    bSet |= SET16_LOW8;
		  }
		if (getregs.pVar8High)
		  {
		    assert (!pVar8High && !(bSet & SET16_HIGH8));
		    pVar8High = getregs.pVar8High;
		    getregs.pVar8High = NULL;
		    getregs.bReg8High = 0; // no need to look again
		    bSet |= SET16_HIGH8;
		  }
		if (bSet == SET16_BOTH8)
		  {
		    // all set, look no further...
		    return UnionVar8And16(pVar8Low,pVar8High,NULL,bSet);
		  }
		break;
	      }
	}
      pStateHolder = pStateHolder->GetParent();
      if (!pStateHolder)
	break;
    }
  
  return UnionVar8And16(pVar8Low,pVar8High,NULL,bSet);
}

#define SET32_LOW8		SET16_LOW8
#define SET32_HIGH8		SET16_HIGH8
#define SET32_BOTH8		SET16_BOTH8
#define SET32_LOW16		0x04

#define ISSET32_LOW8(set)  ((set) & (SET32_LOW8  | SET32_LOW16))
#define ISSET32_HIGH8(set) ((set) & (SET32_HIGH8 | SET32_LOW16))
#define ISSET32_BOTH8(set) (((set) & SET32_BOTH8) == SET32_BOTH8)

#define ISSET32_ANYLOW16(set) ((set) & (SET32_LOW8 | SET32_HIGH8 | SET32_LOW16))

#define ISSET32_ALL_LOW16(set) (((set) & SET32_LOW16) || (((set) & SET32_BOTH8) == SET32_BOTH8))

Variable32* UnionVar8And16And32(Variable8* pVar8Low,Variable8* pVar8High,
				Variable16* pVar16Low,
				Variable32* pVar32,byte bSet)
{
  if (!bSet)
    return pVar32;
  
  if (ISSET32_ANYLOW16(bSet))
    {
      if (!(bSet & SET32_LOW16))
	{
	  // if any 8 is missing, it might be gotten from 32-reg
	  
	  if (!(bSet & SET32_LOW8) && pVar32)
	    {
	      pVar8Low = pVar32->GetQuarterPart(0);
	      bSet |= SET32_LOW8;
	    }
	  if (!(bSet & SET32_HIGH8) && pVar32)
	    {
	      pVar8High = pVar32->GetQuarterPart(1);
	      bSet |= SET32_HIGH8;
	    }
	}
      pVar16Low = UnionVar8And16(pVar8Low,pVar8High,
				 pVar16Low,bSet & SET32_BOTH8);
    }
  else
    {
      // low 16 is pVar16Low
      // nothing to do
    }

  Variable16* pVar16High = (pVar32 ? pVar32->GetHalfPart(1) : NULL);

  if (pVar16Low || pVar16High)
    return new Variable32Split(pVar16Low,pVar16High);
  return NULL;
}

Variable32* GetVarGeneralPurpose32Reg (const RegisterStateParentPtr* pStateHolder,
				       byte bReg)
{
  assert (pStateHolder);
  
  GeneralPurposeRegs getregs;
  
  byte bRegName = REG_NAME(bReg);

  getregs.bReg16 = bRegName | reggSTD16;
  getregs.bReg32 = bReg;
  if (bRegName < 4)
    {
      getregs.bReg8Low  =  bRegName      | reggSTD8;
      getregs.bReg8High = (bRegName + 4) | reggSTD8;
    }
  
  //byte bRegName = REGNAME16_32FROM8(bReg);
  //byte b16RegName = reggSTD16 | bRegName;
  
  Variable8* pVar8Low = NULL;
  Variable8* pVar8High = NULL;
  Variable16* pVar16Low = NULL;
  
  byte bSet = 0;

  for ( ; ; )
    {
      RegisterState* pState = pStateHolder->GetRegisterState();
      
      //if (!pState)
      //return NULL;
      
      if (pState)
	{
	  byte bFound = pState->GetRegisterGeneralPurpose(getregs);
	  if (bFound)
	    switch (bFound)
	      {
	      case GEN_REG_LOW32:
		return UnionVar8And16And32(pVar8Low,pVar8High,
					   pVar16Low,getregs.pVar32,bSet);
	      case GEN_REG_LOW16:
		if (!pVar16Low && !ISSET32_ALL_LOW16(bSet))
		  {
		    pVar16Low = (Variable16*) getregs.pVar16;
		    getregs.pVar16 = NULL;
		    getregs.pVar8Low = NULL;
		    getregs.pVar8High = NULL;
		    getregs.bReg16 = 0; // no need to look again
		    getregs.bReg8Low = 0; // no need to look again
		    getregs.bReg8High = 0; // no need to look again
		    bSet |= SET32_LOW16;
		  }
		break;
	      default:
		assert (bFound & (GEN_REG_LOW8 | GEN_REG_HIGH8));
		if (getregs.pVar8Low)
		  {
		    assert (!pVar8Low && !ISSET32_LOW8(bSet));
		    pVar8Low = getregs.pVar8Low;
		    getregs.pVar8Low = NULL;
		    getregs.bReg8Low = 0; // no need to look again
		    bSet |= SET32_LOW8;
		  }
		if (getregs.pVar8High)
		  {
		    assert (!pVar8High && !ISSET32_HIGH8(bSet));
		    pVar8High = getregs.pVar8High;
		    getregs.pVar8High = NULL;
		    getregs.bReg8High = 0; // no need to look again
		    bSet |= SET32_HIGH8;
		  }
		if (ISSET32_BOTH8(bSet))
		  getregs.bReg16 = 0; // no need to look again
		break;
	      }
	}
      pStateHolder = pStateHolder->GetParent();
      if (!pStateHolder)
	break;
    }

  return UnionVar8And16And32(pVar8Low,pVar8High,pVar16Low,NULL,bSet);
}

Variable* GetVarSimpleReg (const RegisterStateParentPtr* pStateHolder,
			   byte bReg)
{
  // simply find the last occurance of the wanted register
  
  assert (pStateHolder);
  
  for ( ; ; )
    {
      RegisterState* pState = pStateHolder->GetRegisterState();
      
      //if (!pState)
      //return NULL;

      if (pState)
	{
	  Variable* pVar = pState->GetRegister (bReg);
	  
	  if (pVar)
	    return pVar;
	}
      pStateHolder = pStateHolder->GetParent();
      if (!pStateHolder)
	return NULL;
    }
}

/*
Variable* RegisterState::GetVarGeneralPurpose8Reg (byte bReg)
{
	FOR_EACH_ITER_REVERSE(VectRegVar,m_vectRegVars,regvar)
	{
		if (regvar->m_bReg == bReg)
			return regvar->m_pVar;
	}
	if (m_pParent)
		return m_pParent->GetVarSimpleReg(bReg);
	return NULL;
}

Variable* RegisterState::GetVarGeneralPurpose16Reg (byte bReg)
{


	return NULL;
}

Variable* RegisterState::GetVarGeneralPurpose32Reg (byte bReg)
{


	return NULL;
}

Variable* RegisterState::GetVarSimpleReg (byte bReg)
{
	// simply find the last occurance of the wanted register

	FOR_EACH_ITER_REVERSE(VectRegVar,m_vectRegVars,regvar)
	{
		if (regvar->m_bReg == bReg)
			return regvar->m_pVar;
	}
	if (m_pParent)
		return m_pParent->GetVarSimpleReg(bReg);
	return NULL;
}
*/
Variable* RegisterStateParentPtr::GetRegisterVar (byte bReg) const
{
  byte bRegType = bReg & reggTYPE_MASK;

  // now try to find the variable that was asked for note the extra
  // trouble we have with the partial registers (32/16/8)
  
  // printf ("Looking for: %02x\n",bReg);
  
  switch (bRegType)
    {
    case reggSTD8:
      {
	Variable8* pVar = GetVarGeneralPurpose8Reg(this,bReg);
	ASSERT_IS_TYPE_OR_NULL(Variable8*,(Variable*) pVar);
	return pVar;
      }
    case reggSTD16:
      {
	Variable16* pVar = GetVarGeneralPurpose16Reg(this,bReg);
	ASSERT_IS_TYPE_OR_NULL(Variable16*,(Variable*) pVar);
	return pVar;
      }
    case reggSTD32:
      {
	Variable32* pVar = GetVarGeneralPurpose32Reg(this,bReg);
	ASSERT_IS_TYPE_OR_NULL(Variable32*,(Variable*) pVar);
	return pVar;
      }
    default:
      assert (bRegType == reggSEG);
      return GetVarSimpleReg(this,bReg);
    }
}



Variable* GetNonNullVariablePart(Variable* pSrc,dword dwOffset,dword dwSize)
{
  assert (pSrc);
  assert (dwSize);
  assert (dwOffset+dwSize <= pSrc->GetSize());

  assert (powerof2(dwSize));
  assert ((dwOffset & (dwSize-1)) == 0); // dwOffset aligned on dwSize boundary

  if (dwSize == pSrc->GetSize())
    return pSrc; // so that we never ask for parts which are the entire src

  Variable32* pVar32 = dynamic_cast<Variable32*>(pSrc);

  if (pVar32)
    {
      if (dwSize == 1)
	{
	  Variable8* pRet = NULL;
	  
	  assert (dwOffset < 4);
	  
	  pRet = pVar32->GetQuarterPart(dwOffset); // argument to quarterpart is 0=low byte,3=high byte, and because little endian this is right(?)
	  
	  return (pRet ? pRet : (new Variable8Unknown()));
	}  
      if (dwSize == 2)
	{
	  Variable16* pRet = NULL;
	  
	  assert (dwOffset == 0 || dwOffset == 2);
	  
	  pRet = pVar32->GetHalfPart(dwOffset >> 1);
	  
	  return pRet ? pRet : new Variable16Unknown();
	}  
      assert (dwSize == 4); // although should not be reached as long as we only have types with size max 4
    }
  Variable16* pVar16 = dynamic_cast<Variable16*>(pSrc);
  if (pVar16)
    {
      if (dwSize == 1)
	{
	  Variable8* pRet = NULL;
	  
	  assert (dwOffset < 2);
	  
	  pRet = pVar16->GetHalfPart(dwOffset);

	  return pRet ? pRet : new Variable8Unknown();
	}  
      assert (dwSize == 2); // although should not be reached as long as we only have types with size max 4
    }
  assert (dwSize == pSrc->GetSize()); // redundant (as above)...
  return pSrc;  
}

typedef map<dword,Variable*> MapOffsetVar;

Variable8* GetVarFromMap8(MapOffsetVar::iterator const& iterFirst,
			  MapOffsetVar::iterator const& iterLast 
			  IFDEBUGPARAM(dword dwOffset))
{
  if (iterFirst == iterLast)
    return NULL; // there is nothing
  
  // for a one byte variable, there can only be one, and it has to be
  // of the correct size
  
  
  IFDEBUG(MapOffsetVar::iterator check = iterFirst);
  
  assert (++check == iterLast);
  assert (iterFirst->first == dwOffset);
  assert (iterFirst->second);
  assert (iterFirst->second->GetSize() == 1);

  Variable8* pRet = dynamic_cast<Variable8*>(iterFirst->second);
  assert (pRet);
  return pRet;
}

Variable16* GetVarFromMap16(MapOffsetVar& mapOffsetVarParts,
			    MapOffsetVar::iterator const& iterFirst,
			    MapOffsetVar::iterator const& iterLast,
			    dword dwOffset)
{
  if (iterFirst == iterLast)
    return NULL; // there is nothing
  
  assert (iterFirst->second);
  
  Variable16* pRet = dynamic_cast<Variable16*>(iterFirst->second);
  
  if (pRet) // already 16 byte
    {
      IFDEBUG(MapOffsetVar::iterator check = iterFirst);
      
      assert (++check == iterLast);
      assert (iterFirst->first == dwOffset);

      return pRet;
    }
  
  // ok so we have two parts of 8 bytes each

  // MapOffsetVar::iterator iterMiddle = mapOffsetVarParts.lower_bound(dwOffset+1);
  // MapOffsetVar::iterator iterMiddle = lower_bound(iterFirst,iterLast,dwOffset+1);
  
  MapOffsetVar::iterator iterMiddle = iterFirst;
  while (iterMiddle != iterLast && iterMiddle->first < dwOffset+1)
    ++iterMiddle;
  
  Variable8* pHigh = 
    GetVarFromMap8(iterFirst, iterMiddle IFDEBUGPARAM(dwOffset  ));
  Variable8* pLow  = 
    GetVarFromMap8(iterMiddle,iterLast   IFDEBUGPARAM(dwOffset+1));
  
  assert (pHigh || pLow);

  return new Variable16Split(pHigh,pLow);
}

Variable32* GetVarFromMap32(MapOffsetVar& mapOffsetVarParts,
			    MapOffsetVar::iterator const& iterFirst,
			    MapOffsetVar::iterator const& iterLast,
			    dword dwOffset)
{
  if (iterFirst == iterLast)
    return NULL; // there is nothing
  
  assert (iterFirst->second);
  
  Variable32* pRet = dynamic_cast<Variable32*>(iterFirst->second);

  if (pRet) // already 32 byte
    {
      IFDEBUG(MapOffsetVar::iterator check = iterFirst);
      
      assert (++check == iterLast);
      assert (iterFirst->first == dwOffset);
      
      return pRet;
    }
  
  // ok so we have two parts of 16 bytes
  
  // MapOffsetVar::iterator iterMiddle = mapOffsetVarParts.lower_bound(dwOffset+2);
  
  MapOffsetVar::iterator iterMiddle = iterFirst;
  while (iterMiddle != iterLast && iterMiddle->first < dwOffset+2)
    ++iterMiddle;
  
  Variable16* pHigh = 
    GetVarFromMap16(mapOffsetVarParts,iterFirst ,iterMiddle,dwOffset  );
  Variable16* pLow  = 
    GetVarFromMap16(mapOffsetVarParts,iterMiddle,iterLast  ,dwOffset+2);
  
  assert (pHigh || pLow);

  return new Variable32Split(pHigh,pLow);
}

MapOffsetVar::const_iterator InsertVariablePart(MapOffsetVar& mapOffsetVarParts,
						MapSOffsetVar::const_iterator from,sdword dwFoundOffset,
						dword dwInsertAt,dword dwNextAt)
{
  //printf (":FoundOffset 0x%08x\n",dwFoundOffset);
  //printf ("Insert At:   0x%08x\n",dwInsertAt);
  //printf ("Next filled: 0x%08x\n",dwNextAt);
  //printf ("Max size:    0x%08x\n",dwNextAt-dwInsertAt);

  // figure out 
  // i) how much we can accept (and can get), 
  // ii) how badly misaligned we are (relative to destination
  // location, src location and size requested)

  dword dwAcceptSize = min(dwNextAt-dwInsertAt,
			   (dwFoundOffset + from->second->GetSize())- 
			   dwInsertAt);
  assert (dwAcceptSize);
  
  dword dwGetAtOffset = dwInsertAt - dwFoundOffset;
  
  assert (dwGetAtOffset >= 0);
  dword dwAlign = dwInsertAt | dwGetAtOffset;
  
  // we might want to think about 10-byte doubles (internal x86
  // floating.pt format), they will align down to 2-2-2-2-2 word
  // patterns if shadowed anywhere (if alone, they will however
  // short-circuit at the 'are we lucky'-statement above
  
  // alignment constraints
  if (dwAlign & 1 || dwAcceptSize <= 1)
    dwAcceptSize = 1;
  else if (dwAlign & 2 || dwAcceptSize <= 3)
    dwAcceptSize = 2;
  else if (dwAlign & 4 || dwAcceptSize <= 7)
    dwAcceptSize = 4;
    
  //printf ("GetAtOffset: 0x%08x\n",dwGetAtOffset);
  
  //printf ("AcceptSize: 0x%08x\n",dwAcceptSize);
  //printf ("Align:  0x%08x\n",dwAlign);
  //printf ("AcceptSize:  0x%08x\n",dwAcceptSize);
  
  // ok , so get dwAcceptSize bytes from the variable, at offset dwGetAtOffset
  
  Variable* pGetVar = 
    GetNonNullVariablePart(from->second,dwGetAtOffset,dwAcceptSize);
  
  pair<MapOffsetVar::iterator, bool> doins = 
    mapOffsetVarParts.insert(MapOffsetVar::value_type(dwInsertAt,pGetVar));
  
  assert (doins.second);  

  return doins.first;
}

Variable* GetStackVar (const RegisterStateParentPtr* pStateHolder,
		       sdword dwOffset,dword dwSize)
{
  // simply find the last occurance of the wanted register
  
  assert (pStateHolder);
  
  MapOffsetVar mapOffsetVarParts;  
  
  //printf ("Want stack at 0x%08x - %d\n",dwOffset,dwSize);

  for ( ; ; )
    {
      RegisterState* pState = pStateHolder->GetRegisterState();

      //if (!pState)
      //	break;
      
      if (pState)
	{
	  // see if find anything within the range requested that has
	  // not been set yet
	  
	  if (!pState->m_mapStackVars.empty())
	    {
	      MapSOffsetVar::const_iterator iter = 
		pState->m_mapStackVars.lower_bound(dwOffset); // returns first element starting at >= dwOffset
	      if (iter != pState->m_mapStackVars.begin())
		{
		  --iter; // because the element before may start so that it's offset+size > dwOffset
		  
		  //printf ("%x (%x) %x\n",iter->first,iter->first + (sdword) iter->second->GetSize(),dwOffset);
		  
		  
		  // if the prev element did not have anything to do
		  // with us, discard it again (so we know it is
		  // within range in the loop below)
		  
		  assert (iter != pState->m_mapStackVars.end());
		  
		  if (iter->first + (sdword) iter->second->GetSize() <= 
		      (sdword) dwOffset)
		    ++iter;
		  
		  //if (iter == pState->m_mapStackVars.end())
		  //printf ("END");
		  //else
		  //printf ("%x (%x) %x\n",iter->first,iter->first + (sdword) iter->second->GetSize(),dwOffset);
		  
		  assert (iter == pState->m_mapStackVars.end() || // no element
			  iter->first + (sdword) iter->second->GetSize() > 
			  (sdword) dwOffset); // or after start of searched for element
		}
	      MapSOffsetVar::const_iterator stop = 
		pState->m_mapStackVars.upper_bound(dwOffset+dwSize-1); // -1 so we do not get any 
	      
	      assert (stop == pState->m_mapStackVars.end() ||
		      stop->first >= dwOffset + (sdword) dwSize);
	      
	      for ( ; iter != stop; ++iter)
		{
		  //printf ("Found item at 0x%08x - %d\n",iter->first,iter->second->GetSize());
		  
		  // the found item is within range!
		  
		  assert(iter->first + (sdword) iter->second->GetSize() >= 
			 dwOffset);
		  assert(iter->first < dwOffset + (sdword) dwSize);
		  
		  // are we lucky? (perfect match, and no shadows)
		  if (iter->first == dwOffset && iter->second->GetSize() == 
		      dwSize && mapOffsetVarParts.empty())
		    return iter->second;
		  
		  // so it wasn't a perfect match, or we were shadowed
		  
		  // so we'll (probably) have to cut the found
		  // variable to pieces in order to get it into the
		  // map this is like merging

		  // first go through our map to get to the point
		  // where something can be inserted (and where we may
		  // have something to insert)

		  sdword dwFoundOffset = iter->first - dwOffset;

		  //printf ("Found offset (relative wanted): 0x%08x\n",dwFoundOffset);

		  // printf ("0x%08x - 0x%08x => 0x%08x\n",iter->first,dwOffset,dwFoundOffset);
		  // assert (dwFoundOffset >= 0);
		  /*
		    dword dwFoundUsableSize = iter->second->GetSize();
		    if (dwFoundOffset < 0)
		    dwFoundUsableSize += dwFoundOffset; // subtracting 
		  */
		  // the following uses a linear search, even though
		  // we could attack directly (a binary search) using
		  // lower_bound or something, but the logic seem to
		  // become messy, and besides mapOffsetVarParts never
		  // have many elements

		  MapOffsetVar::const_iterator ins = mapOffsetVarParts.begin();

		  if (ins == mapOffsetVarParts.end() || 
		      (sdword) ins->first > max(dwFoundOffset,0))
		    {
		      dword dwInsertAt = max(dwFoundOffset,0);
		      dword dwNextAt = 
			(ins == mapOffsetVarParts.end() ? dwSize : ins->first);
		      ins = InsertVariablePart(mapOffsetVarParts,
					       iter,dwFoundOffset,
					       dwInsertAt,dwNextAt);
		      assert (ins == mapOffsetVarParts.begin()); // we were inserted first
		    }
		  
		  // so ins is to be the item after which we insert

		  while (ins != mapOffsetVarParts.end())
		    {
		      dword dwInsertAt = ins->first + ins->second->GetSize();
		      dword dwNextAt = dwSize;
		  
		      MapOffsetVar::const_iterator next = ins;
		      ++next;
		      
		      while (next != mapOffsetVarParts.end())
			{
			  dwNextAt = next->first;
			  assert (dwNextAt >= dwInsertAt);
			  
			  if (dwNextAt > dwInsertAt)
			    break;
			  ins = next;
			  dwInsertAt = ins->first + ins->second->GetSize();
			  ++next;
			}
		      
		      //printf ("Insert At:   0x%08x\n",dwInsertAt);
		      //printf ("Next filled: 0x%08x\n",dwNextAt);
		      
		      // assert (dwNextAt >= dwInsertAt);
		      
		      if (dwNextAt <= dwInsertAt)
			break; // we're beyond end of request
		      
		      if ((sdword) dwInsertAt - dwFoundOffset >= 
			  (sdword) iter->second->GetSize())
			break; // we're beyond range of found item
		      
		      MapOffsetVar::const_iterator tmp_ins = ins;
		      
		      ins = InsertVariablePart(mapOffsetVarParts,
					       iter,dwFoundOffset,
					       dwInsertAt,dwNextAt);
		      assert (ins == ++tmp_ins); // we were inserted after previous one
		    }
		}   
	    }
      	  // m_mapStackVars
	  /*
	    FOR_EACH_ITER_REVERSE(VectRegVar,pState->m_vectRegVars,regvar)
	    {
	    if (regvar->m_bReg == bReg)
	    return regvar->m_pVar;
	    }
	  */
      	}
      
      pStateHolder = pStateHolder->GetParent();
      if (!pStateHolder)
	break;
    }
  
  if (mapOffsetVarParts.empty())
    {
      // we found nothing...
      return NULL;
    }
  
  // and now we have to put the pieces back together. using VariableNSplit to glue things together (considering alignment)
  
  if (dwSize == 1)
    return GetVarFromMap8(mapOffsetVarParts.begin(),mapOffsetVarParts.end()
			  IFDEBUGPARAM(0));
  else if (dwSize == 2)
    return GetVarFromMap16(mapOffsetVarParts,mapOffsetVarParts.begin(),
			   mapOffsetVarParts.end(),0);
  else if (dwSize == 4)
    return GetVarFromMap32(mapOffsetVarParts,mapOffsetVarParts.begin(),
			   mapOffsetVarParts.end(),0);
  
  assert (false); // no support for larger types yet (as they do not exist)
}



Variable* RegisterStateParentPtr::GetStackVar(dword dwOffset,
					      dword dwSize) const
{
  // ok, so a variable of size dwSize is wanted
  //
  // it's not a problem yet, but what happen when we start supporting
  // flotaing point values?  I suppose we should perhaps say when we
  // want something as not an integer, but a floating point value ant
  // let this function convert the varible if needed before it is
  // returned (basically this conversion destroy the content unless it
  // is an constant)

  // so, we'll have to get the data from where it was stored.  if the
  // request is for an argument to the function, it will normally not
  // have been stored, so we instead should mark that something will
  // be used down there (so the the caller can figure out what is used
  // as arguements and what's not -- hmmm, here one of the really big
  // problems of this project turn up: variable length arguments
  // lists. like printf (which is a very nice thing, so it has to be
  // supported as well as possible)
  
  // Value retrieval will be quite easy when the value requested has
  // been set previously with a variable of the same size. I.e., so
  // that they are aligned. But we must also take care of other
  // situatations, e.g. a byte request from a previously set dword, a
  // word request from a dword where possibly a byte was written
  // later, and an unaligned (d)word request

  // my idea is to keep a mask of the bytes that we so far has found
  // while traversing the RegisterStates backwards, and when we find
  // something that fills any blanks we have left we set the covered
  // part as filled and get the (partial) value from the something
  // (this is something like what is done at the register
  // reconstructors GetVarGeneralPurpose??Reg above, but they are
  // helped by the fact that certain alignments are always observed)

  // if something unaligned is found, or it is partially shadowed by
  // something set later (but then found earlier by us here), we get
  // it in parts, that are aligned to their size and not larger that
  // the window they are suppose to fill), we then put them back
  // together with VariableNSplit's

  // because floating points values are anyway made strange by
  // unalignment, all this can without any(?) loss be done with
  // integers. So the bottom line will be: if the floating point value
  // is untouched it will be returned as is, but it it has been messed
  // with, the mess will be dealt with in the normal fashion, and then
  // a floating point value will be reconstructed

  return ::GetStackVar(this,dwOffset,dwSize);
}

void RegisterStateAffected::AddReg(byte bReg)
{
  m_setReg.insert(bReg);
}

void RegisterStateAffected::AddStack(dword dwOffset,dword dwSize)
{
  // ok, first see if we have any entry that either has part of this
  // or could be extended to include us

  
  

  // m_mapStackCover
}

void RegisterStateAffected::RemoveUnneedOverlap()
{
  // if EAX affected     , then ignore AX, AL, AH
  //     AX || (AL && AH), the  ignore     AL, AH
  //     AL and AH treated as normal
  
  // E{A,B,C,D}X
  
  for (byte nStdRegName = 0; nStdRegName < 4; nStdRegName++)
    {
      if (m_setReg.find(nStdRegName | reggSTD32) != m_setReg.end())
	{
	  m_setReg.erase ( nStdRegName | reggSTD16);
	  goto erase_std_8; // this goto could be avoided by copying the erase code for reggSTD8 below, but I've programmed to much BASIC to be able to resist it...
	}      
      if (m_setReg.find(nStdRegName | reggSTD16) == m_setReg.end())
	continue;
    erase_std_8:
      m_setReg.erase ( nStdRegName      | reggSTD8);
      m_setReg.erase ((nStdRegName + 4) | reggSTD8);
    }
  // E{S,B}P, E{S,D}I
  for (byte nStdRegName = 4; nStdRegName < 8; nStdRegName++)
    {
      if (m_setReg.find(nStdRegName | reggSTD32) != m_setReg.end())
	m_setReg.erase (nStdRegName | reggSTD16);
    }
}

void RegisterState::GetAffected(RegisterStateAffected& affected)
{
  FOR_EACH_ITER_REVERSE(VectRegVar,m_vectRegVars,regvar)
    {
      if (regvar->m_bReg != reg_fakeMEMADDR_DEST &&
	  regvar->m_bReg != reg_fakeMEMADDR_SRC) // these are not affected(?), they are just remembered here..
	affected.AddReg(regvar->m_bReg);
    }
}
 
void RegisterStateAffected::GetAffected(RegisterStateAffected& affected)
{
  FOR_EACH_ITER(SetReg,m_setReg,reg)
    affected.AddReg(*reg);
}

void RegisterStateCombine::GetAffected(RegisterStateAffected& affected)
{
  m_affected.GetAffected(affected);
  m_unknown.GetAffected(affected);
}

void GetAffected (const RegisterStateParentPtr* pStateHolder,
		  const RegisterStateParentPtr* pStopAt,
		  RegisterStateAffected& affected)
{
  // simply find the last occurance of the wanted register
  
  assert (pStateHolder);
  
  for ( ; ; )
    {
      if (pStateHolder == pStopAt)
	return;

      RegisterState* pState = pStateHolder->GetRegisterState();
      
      if (!pState)
	return;
      /*
        VectRegVar m_vectRegVars;
	MapSOffsetVar m_mapStackVars;
	
	FlagsState* m_pFlags;
      */
      pState->GetAffected(affected);

      pStateHolder = pStateHolder->GetParent();
      if (!pStateHolder) 
	return;
    }
}

void RegisterStateParentPtr::GetAffected(const RegisterStateParentPtr* pStopAt,
					 RegisterStateAffected& affected)
{
  ::GetAffected(this,pStopAt,affected);
  
}


/*
bool GetCC(RegisterState* pState,word wFlag,bool* pIsTrue)
{
  // try to find the value of the requested flag
  // if we do not have any FlagsState, check with parent(s)
  // if it is an combined state that is asked for, and we can't get the state
  // try to get the components and figure out state (if possible)

  do
    {
      if (pState->m_pFlags) // if null, go to parent (if any)
	{
	  bool bAffectedOrUndefined;

	  if (pState->m_pFlags->GetCC(wFlag,pIsTrue,&bAffectedOrUndefined))
	    return true; // found value

	  if (bAffectedOrUndefined)
	    break; // cannot go further back
	}
      pState = pState->m_pParent;
    }
  while(pState);

  // found nothing

  return false;
}

bool GetCC_CForZF(RegisterState* pState,bool* pIsTrue)
{
  bool bCF, bKnowCF = GetCC(pState,FLAGS_MASK_CF,&bCF);

  if (bKnowCF && bCF)	{ *pIsTrue = true; return true; }

  bool bZF, bKnowZF = GetCC(pState,FLAGS_MASK_ZF,&bZF);

  if (bKnowZF && bZF)	{ *pIsTrue = true; return true; }

  if (bKnowCF && bKnowZF)
    {
      assert (!bCF && !bZF);
      *pIsTrue = false;
      return true;
    }
  return false;
}

bool GetCC_SFxorOF(RegisterState* pState,bool* pIsTrue)
{
  bool bSF, bOF;

  if (GetCC(pState,FLAGS_MASK_SF,&bSF) && GetCC(pState,FLAGS_MASK_OF,&bOF))
    {
      *pIsTrue = bSF ^ bOF;
      return true;
    }
  return false;
}

bool GetCC_ZForSFxorOF(RegisterState* pState,bool* pIsTrue)
{
  bool bZF, bKnowZF = GetCC(pState,FLAGS_MASK_ZF,&bZF);

  if (bKnowZF && bZF)	{ *pIsTrue = true; return true; }

  bool bSFxorOF, bKnowSFxorOF = GetCC_SFxorOF(pState,&bSFxorOF);

  if (bKnowSFxorOF && bSFxorOF)	{ *pIsTrue = true; return true; }

  if (bKnowZF && bKnowSFxorOF)
    {
      assert (!bZF && !bSFxorOF);
      *pIsTrue = false;
      return true;
    }
  return false;
}

bool RegisterStateParentPtr::GetCC(word wFlag,bool* pIsTrue) const
{
  // try to find the value of the requested flag
  // if we do not have any FlagsState, check with parent(s)
  // if it is an combined state that is asked for, and we can't get the state
  // try to get the components and figure out state (if possible)

  if (::GetCC(this,wFlag,pIsTrue))
    return true;

  if (wFlag & FLAGS_MASK_COMBINED)
    {
      // let's try to figure it out by parts

      switch (wFlag)
	{
	case FLAGS_MASK_CForZF:
	  return ::GetCC_CForZF(this,pIsTrue);
	case FLAGS_MASK_SFneqOF:
	  return ::GetCC_SFxorOF(this,pIsTrue);
	case FLAGS_MASK_ZForSFneqOF:
	  return ::GetCC_ZForSFxorOF(this,pIsTrue);
	}
    }
  return false; // do not know value
}
*/

byte GetFlagCode (const RegisterStateParentPtr* pStateHolder,word wFlag)
{
  // simply find the last occurance of the wanted flag
  
  assert (pStateHolder);
  
  for ( ; ; )
    {
      RegisterState* pState = pStateHolder->GetRegisterState();

      //if (!pState)
      //	return FLAG_UNDEFINED; // found no info, so undefined...
      // changed to: found no info, so untouched

      if (pState)
	{
	  assert (pState->m_pFlags);
	  
	  byte bCode = pState->m_pFlags->GetFlagCode(wFlag);
	  
	  if (bCode != FLAG_UNAFFECTED)
	    return bCode;
	}
      pStateHolder = pStateHolder->GetParent();
      if (!pStateHolder)
	return FLAG_UNDEFINED; // found top item so we know nothing
    }
}

byte GetFlagCodeCombined (const RegisterStateParentPtr* pStateHolder,
			  word wFlag,
			  word wFlagStopMask)
{
  // simply find the last occurance of the wanted flag

  assert (pStateHolder);

  for ( ; ; )
    {
      RegisterState* pState = pStateHolder->GetRegisterState();
      
      //if (!pState)
      //	return FLAG_UNDEFINED; // found no info, so undefined...
      // changed to: found no info, so untouched

      if (pState)
	{
	  assert (pState->m_pFlags);
	  
	  byte wCode = pState->m_pFlags->GetFlagCode(wFlag);
	  
	  if (wCode != FLAG_UNAFFECTED)
	    return wCode;
	  if (pState->m_pFlags->IsAffectedOrUndefined(wFlagStopMask))
	    return FLAG_SPLIT_DEP; // something messed with (part of) the flags, so it has been split
	}
      
      pStateHolder = pStateHolder->GetParent();
      if (!pStateHolder)
	return FLAG_UNDEFINED;
    }
}

byte RegisterStateParentPtr::GetFlagCode (word wFlag) // 0,1,unknown or undefined ???
{
  // we return one of (0,1,FLAG_UNKNOWN,FLAG_UNDEFINED)
  //
  // as long as we encounter FLAG_UNAFFECTED we go further back...  if
  // we do not find anything conclusive, we return FLAG_UNDEFINED (as
  // it may have come from anywhere...)

  assert (powerof2(wFlag)); // only one flag

  // if it is one of the combined flags, we have to take care not to
  // return a result set before any of the participating flags were
  // changed

  if (wFlag & FLAGS_MASK_COMBINED)
    {
      // so do special handling here we make no attempt at combing
      // flags to form a result if possible, we let the caller do such
      // things...  but perhaps we should have it here instead (if
      // several callers would like it...)

      // first try to find the full flag anyway
      
      byte bCode = 
	GetFlagCodeCombined(this,wFlag,
			    ((wFlag == FLAGS_MASK_CForZF) ? 
			     (FLAGS_MASK_CF | FLAGS_MASK_ZF) :
			     ((wFlag == FLAGS_MASK_SFneqOF) ? 
			      (FLAGS_MASK_SF | FLAGS_MASK_OF) : 
			      (FLAGS_MASK_ZF | FLAGS_MASK_SF | FLAGS_MASK_OF))));
      
      if (bCode != FLAG_SPLIT_DEP)
	return bCode;
      
      // so some part of the flag was set somewhere, and some perhaps
      // was undefined...  so try to pick up the pieces

      static const byte bFLorFLtable[4][4] = {
	{ FLAG_0,         FLAG_1, FLAG_UNKNOWN,   FLAG_UNDEFINED  },
	{ FLAG_1,         FLAG_1, FLAG_1,         FLAG_1          },
	{ FLAG_UNKNOWN,   FLAG_1, FLAG_UNKNOWN,   FLAG_UNDEFINED  },
	{ FLAG_UNDEFINED, FLAG_1, FLAG_UNDEFINED, FLAG_UNDEFINED  } };
      
      static const byte bFLxorFLtable[4][4] = {
	{ FLAG_0,         FLAG_1,         FLAG_UNKNOWN,   FLAG_UNDEFINED  },
	{ FLAG_1,         FLAG_0,         FLAG_UNKNOWN,   FLAG_UNDEFINED  },
	{ FLAG_UNKNOWN,   FLAG_UNKNOWN,   FLAG_UNKNOWN,   FLAG_UNDEFINED  },
	{ FLAG_UNDEFINED, FLAG_UNDEFINED, FLAG_UNDEFINED, FLAG_UNDEFINED  } };
      
      switch (wFlag)
	{
	case FLAGS_MASK_CForZF:
	  {
	    byte bCodeCF = ::GetFlagCode (this,FLAGS_MASK_CF);
	    byte bCodeZF = ::GetFlagCode (this,FLAGS_MASK_CF);
	    
	    assert (bCodeCF <= FLAG_UNDEFINED &&
		    bCodeZF <= FLAG_UNDEFINED);

	    return bFLorFLtable[bCodeCF][bCodeZF];
	  }
	case FLAGS_MASK_SFneqOF:
	  {
	    byte bCodeSF = ::GetFlagCode (this,FLAGS_MASK_SF);
	    byte bCodeOF = ::GetFlagCode (this,FLAGS_MASK_OF);
	    
	    assert (bCodeSF <= FLAG_UNDEFINED &&
		    bCodeOF <= FLAG_UNDEFINED);

	    return bFLxorFLtable[bCodeSF][bCodeOF];
	  }
	case FLAGS_MASK_ZForSFneqOF:
	  {
	    byte bCodeSF = ::GetFlagCode (this,FLAGS_MASK_SF);
	    byte bCodeOF = ::GetFlagCode (this,FLAGS_MASK_OF);
	    
	    assert (bCodeSF <= FLAG_UNDEFINED &&
		    bCodeOF <= FLAG_UNDEFINED);

	    byte bSFnewOF = bFLxorFLtable[bCodeSF][bCodeOF];

	    byte bCodeZF = ::GetFlagCode (this,FLAGS_MASK_CF); // one could have gotten this first and short circuited on it...1
	    
	    assert (bSFnewOF <= FLAG_UNDEFINED &&
		    bCodeZF  <= FLAG_UNDEFINED);

	    return bFLorFLtable[bSFnewOF][bCodeZF];
	  }
	}
      assert (false); // should never end up here
    }
  
  // just get the most recent value, if available
  
  return ::GetFlagCode (this,wFlag);
  
  // return (m_wKnown & wFlag) ? FLAG_UNDEFINED : FLAG_UNAFFECTED;
}

#include "HtmlPage.h"
#include "HtmlTags.h" // dep_chk_unknown   

void RegisterState::HtmlPlainDisAsm (HtmlPage& page,
				     const ReportDisAsmInfo& reportDisAsmInfo) const
{
  // printf ("RegisterState::HtmlPlainDisAsm  \n");
  
  FOR_EACH_CONST_ITER(VectRegVar,m_vectRegVars,var)
    {
      // printf ("%s=\n",GetRegisterName (var->m_bReg));

      page << "* " << GetRegisterName (var->m_bReg) << "=";
      
      page << var->m_pVar->ValueString();
      if (var->m_pVar->m_pVarContent)
	page << '\'' << var->m_pVar->m_pVarContent->ContentString() << '\'';
      
      /*
	page << var->m_pVar->ValueString();
      
	page << typeid(var->m_pVar).name() << ',';
	page << typeid(*(var->m_pVar)).name();
	
	page << '\n';
	
	page << var->m_pVar->ValueString();
      */
      page << '\n';
      
      /*
	public:
	byte m_bReg;
	Variable* m_pVar;
      */
    }
  
  if (m_pFlags)
    {
      page << "* flags=" << m_pFlags->FlagsString() << '\n';
      //printf ("%s=\n",GetRegisterName (var->m_bReg));
    }
  
  FOR_EACH_CONST_ITER(MapSOffsetVar,m_mapStackVars,var)
    {
      
      page << "* stack" << 
	(var->first > 0 ? 
	 strsprintf("+0x%08x",var->first) : 
	 strsprintf("-0x%08x",-var->first)) << "=";
      
      assert (var->second);
      
      page << var->second->ValueString();
      
      page << '\n';
      
    }
  

  /*
  VectRegVar m_vectRegVars;
  
  FlagsState* m_pFlags;
  */
  /*
  if (pRegState)
    {
      page << "OutgoingState::HtmlPlainDisAsm";

      pRegState->HtmlPlainDisAsm (page,reportDisAsmInfo);
      
    }
  */
}

void RegisterStateEntry::HtmlPlainDisAsm (HtmlPage& page,
					  const ReportDisAsmInfo& reportDisAsmInfo) const
{
  RegisterState::HtmlPlainDisAsm(page,reportDisAsmInfo);

  FOR_EACH_CONST_ITER(VectRegVar,m_vectRegVarsEntry,var)
    {
      page << "ARG * " << GetRegisterName (var->m_bReg) << "=";
      
      page << var->m_pVar->ValueString();
      if (var->m_pVar->m_pVarContent)
	page << '\'' << var->m_pVar->m_pVarContent->ContentString() << '\'';
      
      page << '\n';
    }
  
  page << "USED: ";m_used.HtmlPlainDisAsm(page,reportDisAsmInfo); page << '\n';
}


void RegisterStateAffected::HtmlPlainDisAsm (HtmlPage& page,
					     const ReportDisAsmInfo& reportDisAsmInfo) const
{
  FOR_EACH_CONST_ITER(SetReg,m_setReg,reg)
    page << "  " << GetRegisterName (*reg);
}

void RegisterStateCombine::HtmlPlainDisAsm (HtmlPage& page,
					    const ReportDisAsmInfo& reportDisAsmInfo) const
{
  RegisterState::HtmlPlainDisAsm(page,reportDisAsmInfo);
  
  printf ("RegisterStateCombine::HtmlPlainDisAsm\n");
  
  page << "AFFECTED: "; m_affected.HtmlPlainDisAsm(page,reportDisAsmInfo); page << '\n';
  page << "UNKNOWN : "; m_unknown .HtmlPlainDisAsm(page,reportDisAsmInfo); page << '\n';
}











