/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// RegisterState.h: interface for the RegisterState class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__REGISTER_STATE_H__INCLUDED_)
#define __REGISTER_STATE_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "GlobalDefs.h" // dep_chk_needed     

// The RegisterState is a record of the state of the cpu registers (in
// a function) after some certain operations, e.g:
//
// * Directly after entry
// * After a code trace (one element of BranchTarget)
// * Just before a trace if the trace has multiple entry points (jumped to 
//   from different locations) (a merge of the states after each jump)
//
// If a trace is the ancester of only one other trace (at a
// fall-through target or a single source jump), more information may
// be located in the RegisterState of the parent, and the information
// is then not duplicated here...)

// The contents of the registers are kept in an map (indexed by regEAX
// etc) All non-set positions in the map are considered to be unknown
// values, however, if there is an parent RegisterState, that is also
// to be checked for any information so, if a value is known in the
// parent, but has been invalidated here, it must be set with an
// object, e.g. of type VariableNUnknown, to make sure info is not
// taken from the parent

class Variable; 
class VarContent;
#include "DataState.h"
class FlagsState; 

class HtmlPage;
class ReportDisAsmInfo;

/*
  #include <map>
  
  using namespace std;
  
  struct StackData
  {
  Variable* m_pVar;
  };
  
  typedef map<int,StackData> MapStackData;
*/

class RegVar
{
 public:
  RegVar(byte bReg,Variable* pVar) { m_bReg = bReg; m_pVar = pVar; }
  
  
 public:
  byte m_bReg;
  Variable* m_pVar;
};
/*
class MemVar
{
 public:
	MemVar(Variable* pVar) { m_dwOffset = dwOffsetm m_bLength = bLength, m_pVar = pVar; }


	// When used for a location on the stack, m_dwOffset is relative to the frame of the
	// current function, such that 0 is the return address, 4 is the first argument (i.e.
	// positive and arguments),
	// and any negative values are variables on the stack
	// bad things will happen if the program does strange things to the stack... (I have not
	// figured out how and why yet, but it should be carefully considered)
	//
	// So offset 0 is where eSP point just after the CALL has been executed, and just before
	// a corresponding RET is executed (pointing to the return address)

 public:
	dword m_dwOffset;
	byte  m_bLength;
	Variable* m_pVar;
};
*/
#include <vector>
#include <map>
#include <set>

using namespace std;

typedef vector<RegVar> VectRegVar;
typedef map<sdword,Variable*> MapSOffsetVar;

// TODO: One may perhaps be able to avoid a fair amount of memory
// comsumption by making two classes of RegisterState, one that like
// the present can handle several registers and stack positions set at
// the same time, and one that only can handle one The several are
// needed when setting incoming information for a function or when
// combining for several entry points to a trace. But most
// instructions only set one register or memory location, so there are
// no need for vectors and their overhead exception: pop which may
// affect a register and eSP

struct GeneralPurposeRegs;
class RegisterStateAffected;

class RegisterState : public DataState  
{
 public:
  RegisterState();
  //  RegisterState(/*RegisterState* pParent*/);
  virtual ~RegisterState();
  
 public:
  // RegisterState* m_pParent;
  
 public:
  VectRegVar m_vectRegVars;
  MapSOffsetVar m_mapStackVars;
  
  FlagsState* m_pFlags;
  
 public:
  void SetRegisterVar (byte bReg,Variable* pVar);
  void SetRegisterVarContent(byte bReg,VarContent* pVarContent); // must not have a variable before
  
  void SetStackVar (sdword dwOffset,Variable* pVar); // get len from var
  // void SetStackVar (dword dwOffset,byte bLength,Variable* pVar);
  void SetStackVarContent (sdword dwOffset,VarContent* pVarContent); // get len from var
  
 public:
  virtual Variable* GetRegister (byte bReg);
  virtual byte GetRegisterGeneralPurpose (GeneralPurposeRegs& getregs);

  Variable* GetRegisterNoFixUnknown (byte bReg); // TODO: remove or fix better, used by cheating code in FixStackArgument
  
 public:
  virtual void GetAffected(RegisterStateAffected& affected);
  
 public:
  void SetFlags(FlagsState* pFlags);
  
  //void SetFlagsUnaffected(); // this is the default... (but perhaps you need it anyway??)
  void SetFlagsUnknown();
  
 public:
  bool HasSameContent(RegisterState* pCompare);
  
 public:
  virtual void HtmlPlainDisAsm (HtmlPage& page,
				const ReportDisAsmInfo& reportDisAsmInfo) const;
};

class RegisterStateUnknown : public RegisterState
{
 public:
  RegisterStateUnknown();
  //  RegisterState(/*RegisterState* pParent*/);
  virtual ~RegisterStateUnknown();

 public:
  virtual Variable* GetRegister (byte bReg);
  virtual byte GetRegisterGeneralPurpose (GeneralPurposeRegs& getregs);
};

typedef set<byte> SetReg;
typedef map<sdword,dword> MapStackCover;

class RegisterStateAffected
{
 public:
  RegisterStateAffected() { }
  virtual ~RegisterStateAffected() { }

 public:
  SetReg m_setReg;
  MapStackCover m_mapStackCover;
  
 public:
  void AddReg(byte bReg);
  void AddStack(dword dwOffset,dword dwSize);

 public:
  void RemoveUnneedOverlap(); // remove dependence on AX, if also EAX and so on
  
 public:
  void GetAffected(RegisterStateAffected& affected);
  
 public:
  bool HasSameContent(RegisterStateAffected* pCompare);
  
 public:
  virtual void HtmlPlainDisAsm (HtmlPage& page,
				const ReportDisAsmInfo& reportDisAsmInfo) const;
};

class RegisterStateEntry : public RegisterState
{
  // this class is used by the EntryState, so when anybody asks for a
  // register, we provide it, and make a note of it, so that we know
  // that it is used as a argument in the function, (unless of course,
  // we later determine that it is *only* propageted through the
  // function and restored untouched (via push and pops) after the
  // function) (i.e., it has to come out untouched, and nobody may
  // make any other use of the value)

 public:
  RegisterStateEntry();
  //  RegisterState(/*RegisterState* pParent*/);
  virtual ~RegisterStateEntry();
  
 public:
  RegisterStateAffected m_used;
  
 public:
  VectRegVar m_vectRegVarsEntry;
  
 public:
  virtual Variable* GetRegister (byte bReg);
  virtual byte GetRegisterGeneralPurpose (GeneralPurposeRegs& getregs);
  
 public:
  virtual void HtmlPlainDisAsm (HtmlPage& page,
				const ReportDisAsmInfo& reportDisAsmInfo) const;
};


class RegisterStateCombine : public RegisterState
{
 public:
  RegisterStateCombine();
  //  RegisterState(/*RegisterState* pParent*/);
  virtual ~RegisterStateCombine();
  
 public:
  RegisterStateAffected m_affected;
  RegisterStateAffected m_unknown;  // perhaps we should have a RegisterStateCombUnknown, to take care of this one (which only exist for loops)
  
 public:
  virtual Variable* GetRegister (byte bReg);
  virtual byte GetRegisterGeneralPurpose (GeneralPurposeRegs& getregs);
  
 public:
  virtual void GetAffected(RegisterStateAffected& affected);
  
 public:
  virtual void HtmlPlainDisAsm (HtmlPage& page,
				const ReportDisAsmInfo& reportDisAsmInfo) const;
  
 public:
  bool HasSameContent(RegisterStateCombine* pCompare);
  
  //public:
  //bool HasSameContent(RegisterStateCombine* pCompare);
};

class RegisterStateParentPtr
{
  /*
    public:
    RegisterStateParentPtr() { };
    ~RegisterStateParentPtr() { };
  */
  
 public:
  virtual const RegisterStateParentPtr* GetParent() const = 0;
  virtual RegisterState* GetRegisterState() const = 0;
  
 public:
  Variable* GetRegisterVar (byte bReg) const;
  
  Variable* GetStackVar(dword sdwOffset,dword dwSize) const;
  
  byte GetFlagCode (word wFlag); // 0,1,unknown or undefined ???
  // bool GetCC(word wFlag,bool* pIsTrue) const;
  
  void GetAffected(const RegisterStateParentPtr* pStopAt,
		   RegisterStateAffected& affected);
};

#endif // !defined(__REGISTER_STATE_H__INCLUDED_)
