/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// RelocationInfo.h: interface for the RelocationInfo class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__RELOCATION_INFO_H__INCLUDED_)
#define __RELOCATION_INFO_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "GlobalDefs.h" // dep_chk_needed     

#include <map>

using namespace std;

//////////////////////////////////////////////////////////////////////

#define RELATIVE_FILE					0x4000
#define RELATIVE_IMAGE					0x8000

#define RELATIVE_TYPE_MASK				0xc000


#define RELATIVE_CURRENT				0x1000 // points into affected segment (into itself)

//////////////////////////////////////////////////////////////////////

#define RELATIVE_PE                     0x0400 // not that it matters, but let's use different numbers
#define RELATIVE_ELF                    0x0800

//////////////////////////////////////////////////////////////////////

#define REL_TYPE_PE_HEADER              (0x0100 | RELATIVE_PE)
#define REL_TYPE_PE_SECT_HEADER         (0x0200 | RELATIVE_PE)
#define REL_TYPE_PE_BODY                (0x0300 | RELATIVE_PE)

//////////////////////////////////////////////////////////////////////

#define RELATIVE_MZ_HEADER				(0x0011 | REL_TYPE_PE_HEADER)
#define RELATIVE_PE_HEADER				(0x0012 | REL_TYPE_PE_HEADER)
#define RELATIVE_PE_OPT_HEADER			(0x0013 | REL_TYPE_PE_HEADER)

#define RELATIVE_CODE_SECTION_HEADER	(0x0031 | REL_TYPE_PE_HEADER)
#define RELATIVE_DATA_SECTION_HEADER	(0x0041 | REL_TYPE_PE_HEADER)
#define RELATIVE_RSRC_SECTION_HEADER	(0x0051 | REL_TYPE_PE_HEADER)
#define RELATIVE_RELOC_SECTION_HEADER	(0x0061 | REL_TYPE_PE_HEADER)

#define RELATIVE_CODE_SECTION			(0x0031 | REL_TYPE_PE_BODY)
							     
#define RELATIVE_DATA_SECTION			(0x0041 | REL_TYPE_PE_BODY)
#define RELATIVE_IMPORT_DIRECTORY		(0x0042 | REL_TYPE_PE_BODY)
#define RELATIVE_EXPORT_DIRECTORY		(0x0043 | REL_TYPE_PE_BODY)
							     
#define RELATIVE_RSRC_SECTION			(0x0051 | REL_TYPE_PE_BODY)
							     
#define RELATIVE_RELOC_SECTION			(0x0061 | REL_TYPE_PE_BODY) 

//////////////////////////////////////////////////////////////////////

#define REL_TYPE_ELF_HEADER             (0x0100 | RELATIVE_ELF)
#define REL_TYPE_ELF_PRGM_HEADER        (0x0200 | RELATIVE_ELF)
#define REL_TYPE_ELF_BODY               (0x0300 | RELATIVE_ELF) // section body
#define REL_TYPE_ELF_SECT_HEADER        (0x0400 | RELATIVE_ELF) // section header

#define RELOC_SECT_HEADER_FROM_BODY(body) ((body) - REL_TYPE_ELF_BODY + REL_TYPE_ELF_SECT_HEADER)
#define RELOC_SECT_BODY_FROM_HEADER(hdr)  ((hdr)  + REL_TYPE_ELF_BODY - REL_TYPE_ELF_SECT_HEADER)
	   
//////////////////////////////////////////////////////////////////////
 
#define REL_ELF_IDENT		        (0x0001 | REL_TYPE_ELF_HEADER)
#define REL_ELF_HEADER		        (0x0002 | REL_TYPE_ELF_HEADER)
	   				     
#define REL_ELF_PRGM_HEADER_PHDR        (0x0001 | REL_TYPE_ELF_PRGM_HEADER)
#define REL_ELF_PRGM_HEADER_INTERP      (0x0002 | REL_TYPE_ELF_PRGM_HEADER)
#define REL_ELF_PRGM_HEADER_NOTE_ABI    (0x0003 | REL_TYPE_ELF_PRGM_HEADER)
#define REL_ELF_PRGM_HEADER_RX_LOAD1    (0x0004 | REL_TYPE_ELF_PRGM_HEADER) // seldom-access (mainly on init)
#define REL_ELF_PRGM_HEADER_RX_LOAD2    (0x0005 | REL_TYPE_ELF_PRGM_HEADER) // .text .plt (and any in LOAD1 if that's omitted)
#define REL_ELF_PRGM_HEADER_RW_LOAD     (0x0006 | REL_TYPE_ELF_PRGM_HEADER)
#define REL_ELF_PRGM_HEADER_DYNAMIC     (0x0007 | REL_TYPE_ELF_PRGM_HEADER)
#define REL_ELF_PRGM_HEADER_NOTE        (0x0008 | REL_TYPE_ELF_PRGM_HEADER)
	   				     
#define REL_TYPE_FIRST_PRGM_HEADER      (REL_ELF_PRGM_HEADER_PHDR)
#define REL_TYPE_LAST_PRGM_HEADER       (REL_ELF_PRGM_HEADER_NOTE)

#define REL_TYPE_ELF_SECTION_0          (0x0001 | REL_TYPE_ELF_BODY)

#define REL_TYPE_ELF_READ_ONLY_FIRST    (0x0002 | REL_TYPE_ELF_BODY) // first read-only
#define REL_TYPE_ELF_READ_ONLY_LAST     (0x003f | REL_TYPE_ELF_BODY) // last  user-defined read-only

#define REL_TYPE_ELF_READ_EXECUTE_FIRST (0x0040 | REL_TYPE_ELF_BODY) // first read-execute
#define REL_TYPE_ELF_READ_EXECUTE_LAST  (0x005f | REL_TYPE_ELF_BODY) // last read-execute

#define REL_TYPE_ELF_READ_ONLY2_FIRST   (0x0060 | REL_TYPE_ELF_BODY) // first read-only2 (to be placed after executable code)
#define REL_TYPE_ELF_READ_ONLY2_LAST    (0x007f | REL_TYPE_ELF_BODY) // last read-only2

#define REL_TYPE_ELF_READ_WRITE_FIRST   (0x0080 | REL_TYPE_ELF_BODY) // first read-write
#define REL_TYPE_ELF_READ_WRITE_LAST    (0x00bf | REL_TYPE_ELF_BODY) // last  user-defined read-write

#define REL_TYPE_ELF_UNALLOC_FIRST      (0x00c0 | REL_TYPE_ELF_BODY) // first read-write
#define REL_TYPE_ELF_UNALLOC_LAST       (0x00ff | REL_TYPE_ELF_BODY) // last  user-defined read-write

#define REL_TYPE_FIRST_SECTION_HEADER   (RELOC_SECT_HEADER_FROM_BODY(REL_TYPE_ELF_SECTION_0))
#define REL_TYPE_LAST_SECTION_HEADER    (RELOC_SECT_HEADER_FROM_BODY(REL_TYPE_ELF_UNALLOC_LAST))

#define REL_ELF_INTERP_SECTION          (REL_TYPE_ELF_READ_ONLY_FIRST+1)
#define REL_ELF_NOTE_ABI_TAG_SECTION    (REL_ELF_INTERP_SECTION+1)
#define REL_ELF_HASH_SECTION            (REL_ELF_NOTE_ABI_TAG_SECTION+1)
#define REL_ELF_DYNSYM_SECTION          (REL_ELF_HASH_SECTION+1)
#define REL_ELF_DYNSTR_SECTION          (REL_ELF_DYNSYM_SECTION+1)
#define REL_ELF_RELDYN_SECTION          (REL_ELF_DYNSTR_SECTION+1)
#define REL_ELF_RELPLT_SECTION          (REL_ELF_RELDYN_SECTION+1)
#define REL_TYPE_ELF_READ_ONLY_UNKNOWN1 (REL_ELF_RELPLT_SECTION+1) // first user-defined read-only
	   				     
#define REL_ELF_INIT_SECTION            (REL_TYPE_ELF_READ_EXECUTE_FIRST+1)
#define REL_ELF_FINI_SECTION            (REL_ELF_INIT_SECTION+1)
#define REL_ELF_PLT_SECTION             (REL_ELF_FINI_SECTION+1)
#define REL_ELF_TEXT_SECTION            (REL_ELF_PLT_SECTION+1)
// there is NO REL_TYPE_ELF_READ_EXECUTE_UNKNOWN1, because copyingh executable code unseen is not a good idea for us,
// as we have not checked it for jumps, data addresses, anything... 
	   				     
#define REL_ELF_RODATA_SECTION          (REL_TYPE_ELF_READ_ONLY2_FIRST+1)
	   				     
#define REL_ELF_DATA_SECTION            (REL_TYPE_ELF_READ_WRITE_FIRST+1)
#define REL_ELF_DYNAMIC_SECTION         (REL_ELF_DATA_SECTION+1)
#define REL_ELF_GOT_SECTION             (REL_ELF_DYNAMIC_SECTION+1)
#define REL_ELF_BSS_SECTION             (REL_ELF_GOT_SECTION+1)
#define REL_TYPE_ELF_READ_WRITE_UNKNOWN1 (REL_ELF_BSS_SECTION+1) // first user-defined read-write

#define REL_ELF_COMMENT                 (REL_TYPE_ELF_UNALLOC_FIRST+1) // first read-write
#define REL_ELF_SECT_HDR_STRING_TABLE   (REL_ELF_COMMENT+1) // first user-defined read-write
#define REL_TYPE_ELF_UNALLOC_UNKNOWN1   (REL_ELF_SECT_HDR_STRING_TABLE+1) // first user-defined read-write

// there is NO read-write-execute sections supported by us, because that implies self-modifying code, something
// which we cannot handle (the plt-got is as close as you get)

//////////////////////////////////////////////////////////////////////

    // r    .hash  .dynsym  .dynstr  .rel.dyn  .rel.plt   and UNKNOWNS
    // rx   .init  .plt  .text  .fini
    // r    .rodata
    // rw   .data  .dynamic  .got  .bss(last) UNKNOWNS
    // ---- stuff not to be in memory image


#define RELATIVE_SECTION_MASK			0x1fff

//////////////////////////////////////////////////////////////////////

#define RELATIVE_FILE_CURRENT			(RELATIVE_FILE | RELATIVE_CURRENT)

#define RELATIVE_FILE_MZ_HEADER			(RELATIVE_FILE | RELATIVE_MZ_HEADER)
#define RELATIVE_FILE_PE_HEADER			(RELATIVE_FILE | RELATIVE_PE_HEADER)
#define RELATIVE_FILE_PE_OPT_HEADER		(RELATIVE_FILE | RELATIVE_PE_OPT_HEADER)

#define RELATIVE_FILE_CODE_SECTION		(RELATIVE_FILE | RELATIVE_CODE_SECTION) 
#define RELATIVE_FILE_DATA_SECTION		(RELATIVE_FILE | RELATIVE_DATA_SECTION) 
#define RELATIVE_FILE_IMPORT_DIRECTORY		(RELATIVE_FILE | RELATIVE_IMPORT_DIRECTORY)
#define RELATIVE_FILE_EXPORT_DIRECTORY		(RELATIVE_FILE | RELATIVE_EXPORT_DIRECTORY)

#define RELATIVE_IMAGE_CURRENT		    	(RELATIVE_IMAGE | RELATIVE_CURRENT)

#define RELATIVE_IMAGE_CODE_SECTION	    	(RELATIVE_IMAGE | RELATIVE_CODE_SECTION) 
#define RELATIVE_IMAGE_DATA_SECTION	    	(RELATIVE_IMAGE | RELATIVE_DATA_SECTION) 
#define RELATIVE_IMAGE_IMPORT_DIRECTORY		(RELATIVE_IMAGE | RELATIVE_IMPORT_DIRECTORY)
#define RELATIVE_IMAGE_EXPORT_DIRECTORY		(RELATIVE_IMAGE | RELATIVE_EXPORT_DIRECTORY)

//////////////////////////////////////////////////////////////////////

#define REL_ELF_IMAGE_TEXT_SECTION            	(RELATIVE_IMAGE | REL_ELF_TEXT_SECTION)

#define REL_ELF_IMAGE_HASH_SECTION           	(RELATIVE_IMAGE | REL_ELF_HASH_SECTION)
#define REL_ELF_IMAGE_DYNSYM_SECTION          	(RELATIVE_IMAGE | REL_ELF_DYNSYM_SECTION)
#define REL_ELF_IMAGE_DYNSTR_SECTION         	(RELATIVE_IMAGE | REL_ELF_DYNSTR_SECTION)
 
#define REL_ELF_IMAGE_INIT_SECTION            	(RELATIVE_IMAGE | REL_ELF_INIT_SECTION)
#define REL_ELF_IMAGE_FINI_SECTION            	(RELATIVE_IMAGE | REL_ELF_FINI_SECTION)
#define REL_ELF_IMAGE_PLT_SECTION              	(RELATIVE_IMAGE | REL_ELF_PLT_SECTION)
 
#define REL_ELF_IMAGE_DYNAMIC_SECTION           (RELATIVE_IMAGE | REL_ELF_DYNAMIC_SECTION)
#define REL_ELF_IMAGE_GOT_SECTION              	(RELATIVE_IMAGE | REL_ELF_GOT_SECTION)

#define REL_ELF_IMAGE_RELDYN_SECTION                  (RELATIVE_IMAGE | REL_ELF_RELDYN_SECTION)
#define REL_ELF_IMAGE_RELPLT_SECTION              	(RELATIVE_IMAGE | REL_ELF_RELPLT_SECTION)




//////////////////////////////////////////////////////////////////////

struct DoRelocation
{
 public:
  DoRelocation(dword dwMove,word wAppendedTo) 
  { m_dwMove = dwMove; m_wAppendedTo = wAppendedTo; m_wChangeRelocation = 0; }
  
 public:
  dword m_dwMove;
  word  m_wAppendedTo;
  word  m_wChangeRelocation;
};

typedef map<word,DoRelocation> MapDoRelocation;

struct RelocInfo
{
 public:
  RelocInfo(word wRelocate) { assert (wRelocate); m_wRelocate = wRelocate; }
  
 public:
  word  m_wRelocate;
  
};

typedef map<dword,RelocInfo> MapOffsetRelocInfo;

#define RELOCATION_IMAGE_NOT_INCLUDED ((dword) -1)

class RelocationInfo  
{
 public:
  RelocationInfo();
  virtual ~RelocationInfo();
  
 public:
  dword m_dwFileOffset;  // where it will be located in the file
  dword m_dwImageOffset; // the current offset of the image data we're holding information for
  // when created, this could be set to 0, but when inserted into the image, it will hold the value where we should be loaded,
  // and when this is changed, all relocations to this piece of data should add the offset we're adding here
  
 public:
  MapOffsetRelocInfo m_mapOffsetRelocInfo;
  
 public:
  void InsertRelocation(dword dwOffset,word wRelocate);
  
  
};

#endif // !defined(__RELOCATION_INFO_H__INCLUDED_)
