/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ReportDisAsmInfoOffset.cpp: implementation of the ReportDisAsmInfoOffset class.
//
//////////////////////////////////////////////////////////////////////

#include "ReportDisAsmInfoOffset.h" 
#include <stdio.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ReportDisAsmInfoOffset::ReportDisAsmInfoOffset(const MapCodeTraceCallFunction* pMapCodeTraceCallFunction,
					       const MapBranchTargetVA* pMapBranchTargetVA)
  : ReportDisAsmInfo(pMapCodeTraceCallFunction)
{
  assert (pMapBranchTargetVA);
  m_pMapBranchTargetVA = pMapBranchTargetVA;
}

ReportDisAsmInfoOffset::~ReportDisAsmInfoOffset()
{

}

string ReportDisAsmInfoOffset::GetLabel (const BranchTarget* pTarget) const
{
  dword dwOffset;
  
  bool bOffset = GetOffset(pTarget,dwOffset);
  
  assert (bOffset);
  
  char fmt[16];
  
  sprintf(fmt,"%08x",dwOffset);
  
  return string(fmt);
}

bool ReportDisAsmInfoOffset::GetOffset (const BranchTarget* pTarget,
					dword& dwOffset) const
{
  MapBranchTargetVA::const_iterator iter = m_pMapBranchTargetVA->find(pTarget);
  
  if (iter != m_pMapBranchTargetVA->end())
    {
      dwOffset = iter->second;
      return true;
    }

  // not found directly, see if we can find by going back via fall-throughs
  
  dwOffset = 0;
  
  while ((pTarget = pTarget->GetPrevTarget()))
    {
      dword dwAdd = pTarget->GetLengthOfRawCode();
      
      if (!dwAdd) // length cannot be 0 => indicates unknown
	return false;
      
      dwOffset += dwAdd;
      
      iter = m_pMapBranchTargetVA->find(pTarget);
      
      if (iter != m_pMapBranchTargetVA->end())
	{
	  dwOffset += iter->second;
	  return true;
	}
    }
  
  // failed
  
  return false;
}

#include <algorithm>

class BranchTargetVACompare
{
 public:
  BranchTargetVACompare(const MapBranchTargetVA* pMapBranchTargetVA)
  { m_pMapBranchTargetVA = pMapBranchTargetVA; }
  
 public:
  const MapBranchTargetVA* m_pMapBranchTargetVA;
  
 public:
  bool operator()(BranchTarget* lhs,BranchTarget* rhs);
};

bool BranchTargetVACompare::operator()(BranchTarget* lhs,BranchTarget* rhs)
{
  MapBranchTargetVA::const_iterator iter_lhs = m_pMapBranchTargetVA->find(lhs);
  
  if (iter_lhs == m_pMapBranchTargetVA->end())
    return true;
  
  MapBranchTargetVA::const_iterator iter_rhs = m_pMapBranchTargetVA->find(rhs);
  
  if (iter_rhs == m_pMapBranchTargetVA->end())
    return true;
  
  return iter_lhs->second < iter_rhs->second;
}

void ReportDisAsmInfoOffset::SortCodeTraces(VectCodeTraces& vectCodeTraces) const
{
  BranchTargetVACompare compare(m_pMapBranchTargetVA);
  
  sort (vectCodeTraces.begin(),vectCodeTraces.end(),compare);
}




