/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// StackState.h: interface for the StackState class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__STACK_STATE_H__INCLUDED_)
#define __STACK_STATE_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#ifdef WIN32
#include "GlobalDefs.h" // for avoiding warnings     
#endif

// The StackState is a record of the state of the stack (in a
// function) after some certain operations, e.g:
//
// * Directly after entry
//
// * After a code trace (one element of BranchTarget)
//
// * Just before a trace if the trace has multiple entry points
// (jumped to from different locations)
//
// If a trace is the ancester of only one other trace (at a
// fall-through target or a single source jump), more information may
// be located in the StackTrace of the parent, and the information is
// then not duplictaed here...)

// The origin of the trace (offset zero) is usually where the stack
// pointer was at function entry, the size is the (negative) offset to
// ESP

// The contents of the stack are kept in an map (indexed by the offset
// to the object) All non-set positions in the map are considered to
// be unknown values, however, if there is an parent StackState, that
// is also to be checked for any information Which may be troublesome
// when asking for an object of a length that is just partially known
// to one StackState and may be more fully known by another

#include "Variable.h" 
#include "DataState.h" 

#include <map>

using namespace std;

struct StackData
{
  Variable* m_pVar;
};

typedef map<int,StackData> MapStackData;

class StackState : public DataState 
{
 public:
  StackState();
  virtual ~StackState();
  
 public:
  StackState* m_pParent; 
  // state at entry to BranchTarget is single source entry
  
  int m_nSize;
  
  MapStackData m_mapData;
  
 public:
  Variable* GetData(int nOffset,int nSize);
  void SetData (int nOffset,Variable* pVar);
  void SetDataUnknown (int nOffset,int nSize);
  
  void Push(Variable* pVar);
  void PushUnknown(int nSize);
  
  void Pop(int nSize);
  
};

#endif // !defined(__STACK_STATE_H__INCLUDED_)

