/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// StatusFlagsAffected.cpp: implementation of the StatusFlagsAffected class.
//
//////////////////////////////////////////////////////////////////////

#include "StatusFlagsAffected.h" 

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

bool StatusFlagsAffected::IsCompatible(const StatusFlagsKnown& known) const
{
  // if a flag is affected and known by this StatusFlagsAffected, then
  // is must agree with any flags known

  // (m_wValue ^ known.m_wValue) // is 1 for any flag not agreeing
  
  return !(m_wAffected & m_wKnown & known.m_wKnown & 
	   (m_wValue ^ known.m_wValue));
}

void StatusFlagsAffected::Join(const StatusFlagsAffected& src)
{
  assert (m_wAffected == src.m_wAffected); // both must be affected in same way
  assert ((~m_wAffected & m_wKnown) == (~src.m_wAffected & src.m_wKnown)); // and have same values undefined
  
  m_wKnown = (m_wKnown & src.m_wKnown) & // values that both know
    ~(m_wValue ^ src.m_wValue); // and agree upon
  m_wValue &= m_wKnown; // normalize
}

void StatusFlagsAffected::CombinedFromZSO()
{
  assert (((m_wAffected | m_wKnown) & 
	   (FLAGS_MASK_ZF | FLAGS_MASK_SF | FLAGS_MASK_OF)) ==
          (FLAGS_MASK_ZF | FLAGS_MASK_SF | FLAGS_MASK_OF));
  assert (((m_wAffected | m_wKnown) & 
	   (FLAGS_MASK_SFneqOF | FLAGS_MASK_ZForSFneqOF)) ==
          (FLAGS_MASK_SFneqOF | FLAGS_MASK_ZForSFneqOF)); // to be set
  assert (!(m_wValue & (FLAGS_MASK_SFneqOF | FLAGS_MASK_ZForSFneqOF)));
  
  if ((!(m_wValue & FLAGS_MASK_SF)) != (!(m_wValue & FLAGS_MASK_OF)))
    m_wValue |= (FLAGS_MASK_SFneqOF | FLAGS_MASK_ZForSFneqOF);
  else if (m_wValue & FLAGS_MASK_ZF)
    m_wValue |= FLAGS_MASK_ZForSFneqOF;
}

void StatusFlagsAffected::CombinedFromCZSO()
{
  assert (((m_wAffected | m_wKnown) & (FLAGS_MASK_CF | FLAGS_MASK_ZF)) ==
          (FLAGS_MASK_CF | FLAGS_MASK_ZF));
  assert ((m_wAffected | m_wKnown) & FLAGS_MASK_CForZF); // to be set
  assert (!(m_wValue & FLAGS_MASK_CForZF));
  
  if (m_wValue & (FLAGS_MASK_CF | FLAGS_MASK_ZF))
    m_wValue |= FLAGS_MASK_CForZF;
  
  CombinedFromZSO();
}
/*
  void CombinedFromZSOifKnown()
  {
assert (((m_wAffected) & (FLAGS_MASK_ZF | FLAGS_MASK_SF | FLAGS_MASK_OF)) ==
(FLAGS_MASK_ZF | FLAGS_MASK_SF | FLAGS_MASK_OF));
assert (((m_wAffected) & (FLAGS_MASK_SFneqOF | FLAGS_MASK_ZForSFneqOF)) ==
(FLAGS_MASK_SFneqOF | FLAGS_MASK_ZForSFneqOF)); // to be set
assert (!(m_wValue & (FLAGS_MASK_SFneqOF | FLAGS_MASK_ZForSFneqOF)));

		if ((m_wKnown & (FLAGS_MASK_SF | FLAGS_MASK_OF)) == (FLAGS_MASK_SF | FLAGS_MASK_OF))
		{
		bool bSFneqOF = ((!(m_wValue & FLAGS_MASK_SF)) != (!(m_wValue & FLAGS_MASK_OF)));			
		
		  SetKnown(FLAGS_MASK_SFneqOF,bSFneqOF);
		  
			if (bSFneqOF || (m_wValue & FLAGS_MASK_ZF))
			SetKnown1(FLAGS_MASK_ZForSFneqOF);
			else if (m_wKnown && FLAGS_MASK_ZF)
			SetKnown0(FLAGS_MASK_ZForSFneqOF);
			}
			else if ((m_wValue & FLAGS_MASK_ZF))
			SetKnown1(FLAGS_MASK_ZForSFneqOF);
			}
*/
/*
void CombinedFromCZSOifKnown()
{
assert (((m_wAffected) & (FLAGS_MASK_CF | FLAGS_MASK_ZF)) ==
(FLAGS_MASK_CF | FLAGS_MASK_ZF));
assert ((m_wAffected) & FLAGS_MASK_CForZF); // to be set
assert (!(m_wValue & FLAGS_MASK_CForZF));

		if (m_wValue & (FLAGS_MASK_CF | FLAGS_MASK_ZF))
		{
		assert(m_wKnown & (FLAGS_MASK_CF | FLAGS_MASK_ZF));
		SetKnown1(FLAGS_MASK_CForZF);
		}
		else if ((m_wKnown & (FLAGS_MASK_CF | FLAGS_MASK_ZF)) == (FLAGS_MASK_CF | FLAGS_MASK_ZF))
		SetKnown0(FLAGS_MASK_CForZF);
		
		  CombinedFromZSOifKnown();
		  }
*/
void StatusFlagsAffected::ConstrainAffectedByCPAZSO()
{
  m_wAffected = FLAGS_MASK_AFFECT_CPAZSO;
  m_wKnown    = FLAGS_MASK_KNOWN_CPAZSO;
        
  m_wValue   &= FLAGS_MASK_CPAZSO; // get rid of unused flags
        
  CombinedFromCZSO();
}

void StatusFlagsAffected::ConstrainAffectedByCPZSO()
{
  m_wAffected = FLAGS_MASK_AFFECT_CPZSO;
  m_wKnown    = FLAGS_MASK_KNOWN_CPZSO;
        
  m_wValue   &= FLAGS_MASK_CPZSO; // get rid of unused flags
        
  CombinedFromCZSO();
}

void StatusFlagsAffected::ConstrainAffectedByPAZSO()
{
  m_wAffected = FLAGS_MASK_AFFECT_PAZSO;
  m_wKnown    = FLAGS_MASK_KNOWN_PAZSO;
        
  m_wValue   &= FLAGS_MASK_PAZSO; // get rid of unused flags
        
  CombinedFromZSO();
}

void StatusFlagsAffected::AffectedByCPAZSO()
{
  m_wAffected = FLAGS_MASK_AFFECT_CPAZSO;
}

void StatusFlagsAffected::AffectedByCPZSO()
{
  m_wAffected = FLAGS_MASK_AFFECT_CPZSO;
}

void StatusFlagsAffected::AffectedByPAZSO()
{
  m_wAffected = FLAGS_MASK_AFFECT_PAZSO;
}

void StatusFlagsAffected::Undefined(word wMask)
{
  assert (!((m_wAffected | m_wKnown | m_wValue) & wMask));
  m_wKnown |= wMask; // known but not affected, nah, but actually unknown (undefined)
}

bool StatusFlagsAffected::AnyAffectedOrUndefined() const
{
  return !!(m_wAffected | m_wKnown);
}

bool StatusFlagsAffected::IsAffectedOrUndefined(word wMask) const
{
  return !!((m_wAffected | m_wKnown) & wMask);
}

void StatusFlagsAffected::SetKnown0(word wMask)
{
  assert ((m_wAffected & wMask) == wMask);
  assert (!(m_wKnown & wMask) ||
          !(m_wValue & wMask));
  m_wKnown |= wMask; 
}
void StatusFlagsAffected::SetKnown1(word wMask)
{
  assert ((m_wAffected & wMask) == wMask);
  assert (!(m_wKnown & wMask) ||
          (m_wValue & wMask) == (m_wKnown & wMask));
  m_wKnown |= wMask;
  m_wValue |= wMask;
}

void StatusFlagsAffected::SetKnown(word wMask,bool bSet)
{
  assert ((m_wAffected & wMask) == wMask);
  assert (!(m_wKnown & wMask) ||
          (m_wValue & wMask) == (bSet ? (m_wKnown & wMask) : 0));
  assert (!((m_wKnown | m_wValue) & wMask));
  m_wKnown |= wMask;
  if (bSet)
    m_wValue |= wMask;
}

void StatusFlagsAffected::SetFlags(word wMask,word wValue)
{
  assert (!(m_wAffected & wMask));
  assert (!(m_wKnown & wMask));
  assert (!(m_wValue & wMask));

  m_wAffected |= wMask;
  m_wKnown |= wMask;
  m_wValue |= wValue;
}

bool StatusFlagsAffected::IsAffected (word wFlag) const
{
  assert ((m_wAffected & (m_wKnown & wFlag)) == (m_wKnown & wFlag)); // valid value
  return !!(m_wAffected & wFlag);
}

bool StatusFlagsAffected::IsUndefined (word wFlag) const
{
  assert ((m_wAffected & (m_wKnown & wFlag)) == (m_wKnown & wFlag)); // valid value
  return !!(~m_wAffected & m_wKnown & wFlag);
}

bool StatusFlagsAffected::IsKnown (word wFlag) const
{
  assert ((m_wAffected & (m_wKnown & wFlag)) == (m_wKnown & wFlag)); // valid value
  return !!(m_wAffected & m_wKnown & wFlag);
}

bool StatusFlagsAffected::AreKnown (word wMask) const
{
  assert ((m_wAffected & (m_wKnown & wMask)) == (m_wKnown & wMask)); // valid value
  return (m_wAffected & m_wKnown & wMask) == wMask;
}

bool StatusFlagsAffected::GetFlag (word wFlag) const
{
  assert (m_wAffected & wFlag); // not undefined
  assert (m_wKnown & wFlag); // not undefined
  return !!(m_wValue & wFlag);
}

bool StatusFlagsAffected::GetFlagsOr (word wMask) const
{
  assert ((m_wAffected & wMask) == wMask); // not undefined
  assert ((m_wKnown & wMask) == wMask); // not undefined
  return !!(m_wValue & wMask);
}

// returns true if definate value put in pIsTrue
bool StatusFlagsAffected::GetFlagIfKnown(word wFlag,bool* pIsTrue,bool *pAffectedOrUndefined) const
{
  if (!IsKnown(wFlag))
    {
      assert (pAffectedOrUndefined);
      
      *pAffectedOrUndefined = !!((m_wAffected | m_wKnown) & wFlag);
      return false; // value not known
    }
  
  assert (pIsTrue);
  
  *pIsTrue = GetFlag(wFlag);
  
  return true;	
}

byte StatusFlagsAffected::GetFlagCode (word wFlag) const
{
  if (m_wAffected & wFlag)
    {
      if (m_wKnown & wFlag)
	return !!(m_wValue & wFlag); // value is known
      return FLAG_UNKNOWN;           // it was affected, but value is unknown (most common...) (perhaps along with unaffected)
    }
  
  assert (!(m_wValue & wFlag));
  
  return (m_wKnown & wFlag) ? FLAG_UNDEFINED : FLAG_UNAFFECTED;
}

string StatusFlagsAffected::FlagsString() const
{
  string str;

  for (word f = 0x0800; f; f >>= 1)
    {
      str += ((m_wAffected & f) ? 
	      ((m_wKnown & f) ? ((m_wValue & f) ? '1' : '0') : '?') : 
	      ((m_wKnown & f) ? 'x' : '-')); // known but not affected = undefined
    }
  /*
word

#define FLAGS_MASK_CF				0x0001
#define FLAGS_MASK_CForZF			0x0002 // CF|ZF
#define FLAGS_MASK_PF				0x0004
#define FLAGS_MASK_ZForSFneqOF	 		0x0008 // ZF|(SF!=OF) = ZF|(SF^OF)
#define FLAGS_MASK_AF				0x0010
#define FLAGS_MASK_SFneqOF			0x0020 // SF!=OF      = SF^OF
#define FLAGS_MASK_ZF				0x0040
#define FLAGS_MASK_SF				0x0080
#define FLAGS_MASK_DF				0x0400
#define FLAGS_MASK_OF				0x0800
  */

  return str;
}

