/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// StatusFlagsAffected.h: interface for the StatusFlagsAffected class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__STATUS_FLAGS_AFFECTED_H__INCLUDED_)
#define __STATUS_FLAGS_AFFECTED_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "GlobalDefs.h" 

#include <string>

using namespace std;

//////////////////////////////////////////////////////////////////////

// flags set by arithmetic instructions, and used as conditions for
// jumps
//
// we keep special track of some combinations (CForZF, ZForSFneqOF and
// SFneqOF) because they are used as conditions if for example two
// vars of some consts have been operated on, and some gave CF and
// some gave ZF but not always, we can't be sure if we always had the
// condition CF|ZF, but as we keep special track of that we might know
// (or course, we might also get the result that they actually
// sometimes was, but not always)

#define FLAGS_MASK_CF				0x0001
#define FLAGS_MASK_CForZF			0x0002 // CF|ZF
#define FLAGS_MASK_PF				0x0004
#define FLAGS_MASK_ZForSFneqOF	 		0x0008 // ZF|(SF!=OF) = ZF|(SF^OF)
#define FLAGS_MASK_AF				0x0010
#define FLAGS_MASK_SFneqOF			0x0020 // SF!=OF      = SF^OF
#define FLAGS_MASK_ZF				0x0040
#define FLAGS_MASK_SF				0x0080
#define FLAGS_MASK_DF				0x0400
#define FLAGS_MASK_OF				0x0800

#define FLAGS_MASK_COMBINED			(FLAGS_MASK_CForZF | FLAGS_MASK_ZForSFneqOF | FLAGS_MASK_SFneqOF)

#define FLAGS_MASK_CPAZSO			(FLAGS_MASK_CF | FLAGS_MASK_PF | FLAGS_MASK_AF | \
                                     		FLAGS_MASK_ZF | FLAGS_MASK_SF | FLAGS_MASK_OF)
#define FLAGS_MASK_PAZSO			(FLAGS_MASK_PF | FLAGS_MASK_AF | \
                                     		FLAGS_MASK_ZF | FLAGS_MASK_SF | FLAGS_MASK_OF)
#define FLAGS_MASK_CPZSO			(FLAGS_MASK_CF | FLAGS_MASK_PF | \
                                     		FLAGS_MASK_ZF | FLAGS_MASK_SF | FLAGS_MASK_OF)

#define FLAGS_MASK_AFFECT_CPAZSO	(FLAGS_MASK_CF | FLAGS_MASK_PF | FLAGS_MASK_AF | \
                                     	FLAGS_MASK_ZF | FLAGS_MASK_SF | FLAGS_MASK_OF | \
                                     	FLAGS_MASK_CForZF | FLAGS_MASK_ZForSFneqOF | FLAGS_MASK_SFneqOF)
#define FLAGS_MASK_KNOWN_CPAZSO		(FLAGS_MASK_AFFECT_CPAZSO)
#define FLAGS_MASK_AFFECT_PAZSO  	(FLAGS_MASK_PF | FLAGS_MASK_AF | \
                                     	FLAGS_MASK_ZF | FLAGS_MASK_SF | FLAGS_MASK_OF | \
                                    	 FLAGS_MASK_CForZF | FLAGS_MASK_ZForSFneqOF | FLAGS_MASK_SFneqOF)
#define FLAGS_MASK_KNOWN_PAZSO  	(FLAGS_MASK_PF | FLAGS_MASK_AF | \
                                     	FLAGS_MASK_ZF | FLAGS_MASK_SF | FLAGS_MASK_OF | \
                                     	FLAGS_MASK_ZForSFneqOF | FLAGS_MASK_SFneqOF)
#define FLAGS_MASK_AFFECT_CPZSO 	(FLAGS_MASK_CF | FLAGS_MASK_PF | \
                                     	FLAGS_MASK_ZF | FLAGS_MASK_SF | FLAGS_MASK_OF | \
                                     	FLAGS_MASK_CForZF | FLAGS_MASK_ZForSFneqOF | FLAGS_MASK_SFneqOF)
#define FLAGS_MASK_KNOWN_CPZSO		(FLAGS_MASK_AFFECT_CPZSO)

#define FLAGS_MASK_KNOWN_CO	        (FLAGS_MASK_CF | FLAGS_MASK_OF)
#define FLAGS_MASK_UNDEF_AF	        (FLAGS_MASK_AF)

#define FLAG_0          0
#define FLAG_1          1
#define FLAG_UNKNOWN    2
#define FLAG_UNDEFINED  3
#define FLAG_UNAFFECTED 4
#define FLAG_SPLIT_DEP  5 // used by helper function in registerstate to sort out combined flag dependencies

struct StatusFlagsKnown
{
 public:
  StatusFlagsKnown()
  {
    m_wKnown = 0;
    m_wValue = 0;
  }
  
 public:
  word m_wKnown;    // the flags with unambigous values
  word m_wValue;
};

struct StatusFlagsAffected
{
 public:
  StatusFlagsAffected()
    {
      m_wAffected = 0;
      m_wKnown = 0;
      m_wValue = 0;
    }

  StatusFlagsAffected(word wKnown,word wValue)
  {
    m_wAffected = wKnown;
    m_wKnown = wKnown;
    m_wValue = wValue;    
  }
  
  
 public:
  word m_wAffected;
  word m_wKnown;    // the flags with unambigous values
  word m_wValue;
  
  // If a flag is marked as known, but not affected, it means it is
  // undefined so it might (or might) not be old value, processor
  // dependent
  
 public:
  bool IsCompatible(const StatusFlagsKnown& known) const;
  
 public:
  void Join(const StatusFlagsAffected& src);
  
 public:
  void CombinedFromZSO();
  void CombinedFromCZSO();
  
 public:
  void ConstrainAffectedByCPAZSO();
  void ConstrainAffectedByCPZSO();
  void ConstrainAffectedByPAZSO();
  
 public:
  void AffectedByCPAZSO();
  void AffectedByCPZSO();
  void AffectedByPAZSO();
  
  void Undefined(word wMask);
  
  bool AnyAffectedOrUndefined() const;
  
  bool IsAffectedOrUndefined(word wMask) const;

 public:
  void SetKnown0(word wMask);
  void SetKnown1(word wMask);
  void SetKnown(word wMask,bool bSet);
  
  void SetFlags(word wMask,word wValue);

 public:
  bool IsAffected (word wFlag) const;
  // bool AreKnown (word wMask);

  bool IsUndefined (word wFlag) const;
  // bool AreKnown (word wMask);

  bool IsKnown (word wFlag) const;
  bool AreKnown (word wMask) const;
  
  bool GetFlag (word wFlag) const;
  bool GetFlagsOr (word wMask) const;

  byte GetFlagCode (word wFlag) const;
  
 public:
  bool GetFlagIfKnown(word wFlag,
		      bool* pIsTrue,
		      bool *pAffectedOrUndefined) const; // returns true if definate value put in pIsTrue
  
 public:
  string FlagsString() const;
  
};

#endif // !defined(__STATUS_FLAGS_AFFECTED_H__INCLUDED_)
