/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// TryDoInstructionEncoding.h: interface for the TryDoInstructionEncoding class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__TRY_DO_INSTRUCTION_ENCODING_H__INCLUDED_)
#define __TRY_DO_INSTRUCTION_ENCODING_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "GlobalDefs.h"

struct InverseOpcodeInfo; 

class OperandMemory; 
class OperandImmediate; 
class Operand;
class DataDestPtr;
class Instruction;

//////////////////////////////////////////////////////////////////////

#define PD_OPSIZE32    0x01
#define PD_ADDRSIZE32  0x02
#define PD_SEGMENT     0x04
#define PD_REPLOCK     0x08

struct PrefixDecision
{
 public:
  PrefixDecision() { Reset(); }
  
 public:
  void Reset() { m_bDecided = m_bSet = 0; m_bSegmentReg = 0; m_bRepLock = 0; };
  
 public:
  byte m_bDecided;
  byte m_bSet;
  byte m_bSegmentReg; // if != 0 but not decided, then it's the default value (which can harmlessly be emitted - will enlarge opcode by 1 byte)
  byte m_bRepLock;
  
 public:
  bool IsDecidedAddrSize32() const { return !!(m_bDecided & PD_ADDRSIZE32); }
  
  bool IsDecidedAndSetAddrSize32() 
       const 
  { 
    assert (!(m_bSet & PD_ADDRSIZE32) || 
	    (m_bDecided & PD_ADDRSIZE32)); 
    return !!(m_bSet & PD_ADDRSIZE32); 
  }
  
  bool IsAddrSize32() const 
  { 
    assert (m_bDecided & PD_ADDRSIZE32); 
    return !!(m_bSet & PD_ADDRSIZE32); 
  }
  bool IsOpSize32() const   
  { 
    assert (m_bDecided & PD_OPSIZE32);   
    return !!(m_bSet & PD_OPSIZE32); 
  }
  
  bool ValidateAddressSize(bool bAddrSize32);
  bool ValidateSegment (const OperandMemory* pOpMem,byte bDefaultSegReg);
  void ValidateRepLock (byte bLock);
  
 public:
  byte OutputPrefixesLength(bool bCodeSize32) const;
  void OutputPrefixes(DataDestPtr& dataPtr,
		      bool bCodeSize32,
		      bool bExtraSeg = false) const;
  /*
    public:
	byte m_bDecided;
	byte m_bSet;
	byte m_bSegmentReg;
	byte m_bRepLock;

public:
	void Decide(byte bRule,bool bSet) { m_bDecided |= bRule; if (bSet) m_bSet |= bRule; }
	void DecideSegmentReg(byte bSegmentReg) { m_bDecided |= PD_SEGMENT; m_bSegmentReg = bSegmentReg; }
	void DecideLockRep(byte bRepLock) { m_bDecided |= PD_REPLOCK; m_bRepLock = bRepLock; }
#ifndef NDEBUG
	void CanEncode() { m_bDecided |= PD_CAN_ENCODE;
#endif

public:
	bool Decided(byte bRule) { return !!(bDecided & bRule); }
*/

	/*
	bool bOpSize32;
	bool bAddrSize32;

	bool bOpSize32Decided;
	bool bAddrSize32Decided;

	bool bSegmentDecided;

	bool bRepLockDecided;
	*/
};

//////////////////////////////////////////////////////////////////////

// this class first determine if the InverseOpcodeInfo has the
// capability of encoding the Instruction. And if so, it can write the
// actual bytes
//
// It's also capable of a few tricks to enlarge the instruction
// size. Normally, it chooses the smallest size, but simetimes it's
// possible to emit a few extra bytes that does now change the
// instrucion:
//
// * If the instruction has memory access, we might eject an segment
// prefix, if the default would be used anyway
//
// * If the instruction uses a mod/rm and does not need a sib byte, we
// can perhaps make it with an sib byte anyway
//
// * If the instruction uses an 8-bit displacement, it could also be
// an 16/32-bit
//
// * Or if it does not use an displacement, we could eject an
// displacement of size 8 or 16/32, being 0x00, giving the same effect
//
// This gives possibilities of enlargement [1] + [1] + [1|3]/[1|2/4]
// => 1,2,3,4,5 / 1,2,3,4,5,6 But not necessarilly all at the same
// time
//
// Note that the order in which they are used is important, because
// some things are worse to decode (for the cpu), extra prefixes
// should be used as last resort, and we should probably begin with
// unnecessary sib byte => sib,disp,prefix
//
// These were tricks with memory operands. The system itself is by
// design capable of finding other instruction encodings with
// different opcodes for the same thing that might have different
// lengths. The only other way I can think of to make instructions
// longer is rewriting JCC rel8 as JCC rel16/32, which would increase
// instruction size by 2 or 4 bytes (2 bytes is for 16-bit code, which
// is not very much supported anyway), and 4 bytes is just moving a
// dword-boudary so I have not bothered making the code for creating
// such a beast, because it would involve allowing an opcode that's
// different based on length of imm and instruction name and then
// messing inside VerifyOperandSize

//////////////////////////////////////////////////////////////////////

#define MODRM_MAY_ENLARGE_DISP_SIZE 0x80

struct ModRMDecision
{
 public:
  ModRMDecision() { Reset(); }
  
 public:
  void Reset()
  {
    bModRMExtraSIB = bModRM = bSIB = 0; 
    bMinDispSize = (byte) 0;
    // dwDisp = 0; 
  }
  
 public:
  byte bModRM;
  byte bSIB;
  byte bMinDispSize;
  byte bModRMExtraSIB;
  sdword dwDisp;
  
  // if there is a possibility of enlarging by issuing unnecessary
  // SIB, then modrm and sib are in bModRMExtraSIB and bSIB note that
  // if we want to increase displacement size, rm and SIB never
  // changes, only mod changes
  
 public:
  void SetDisplacement(const OperandMemory* pOpMem);
  
  byte OutputDisplacementMinLength() const 
  { return bMinDispSize & ~MODRM_MAY_ENLARGE_DISP_SIZE; }
  byte OutputDisplacement(DataDestPtr& dataPtr,byte bLengthenTricks = 0) const;
};

//////////////////////////////////////////////////////////////////////

class ImmDecision
{
 public:
  ImmDecision() { Reset(); }

 public:
  void Reset()
    {
      m_bImmHigh16Bytes = m_bImmBytes = m_bScndImmBytes = 0; 
    }
  
  
  dword	m_dwImm;
  word	m_wImmHigh16; // for far ptrs
  
  byte	m_bImmBytes;
  byte	m_bImmHigh16Bytes;
  
  byte	m_bScndImmBytes;
  
  byte	m_bScndImm;
  
 public:
  void SetImmediate(const OperandImmediate* pOpImm,bool bSignExtended = false);
  
  byte OutputImmediatesLength() const;
  byte OutputImmediates(DataDestPtr& dataPtr) const;
  
};

//////////////////////////////////////////////////////////////////////

#define LENGTHEN_TRICK_ISSUE_SIB		0x01
#define LENGTHEN_TRICK_DISP_0_TO_8		0x02
#define LENGTHEN_TRISK_DISP_0_TO_16		0x04
#define LENGTHEN_TRISK_DISP_8_TO_16		0x08
#define LENGTHEN_TRISK_DISP_0_TO_32		0x10
#define LENGTHEN_TRISK_DISP_8_TO_32		0x20
#define LENGTHEN_TRICK_ISSUE_SEG		0x40

#define LENGTHEN_TRICK_DISP_TO_8		(LENGTHEN_TRICK_DISP_0_TO_8)
#define LENGTHEN_TRICK_DISP_TO_16		(LENGTHEN_TRISK_DISP_0_TO_16 | LENGTHEN_TRISK_DISP_8_TO_16)
#define LENGTHEN_TRICK_DISP_TO_32		(LENGTHEN_TRISK_DISP_0_TO_32 | LENGTHEN_TRISK_DISP_8_TO_32)

#define LENGTHEN_TRICK_DISP				(LENGTHEN_TRICK_DISP_TO_8 | LENGTHEN_TRICK_DISP_TO_16 | LENGTHEN_TRICK_DISP_TO_32)

class TryDoInstructionEncoding
{
 public:
  TryDoInstructionEncoding(bool bCodeSize32);
  
 protected:
  bool m_bCodeSize32;
  
 public:
  void Reset();
  
 protected:
  // this information is filled in while TryEncoding,
  // and used while DoEncoding (if try was successful)
  
  const InverseOpcodeInfo* m_pInverseOpcodeInfo;
  
  //	OpcodeInfo m_info;
  PrefixDecision m_prefix;
  ModRMDecision m_modrm;
  ImmDecision m_imm;
  
  byte m_bShortestLength;
  
  /*
    const Operand* m_pOpDest;
    const Operand* m_pOpSrc;
    const Operand* m_pOpSrc2;
    
    dword m_dwDestOp;
    dword m_dwSrcOp;
    dword m_dwSrc2Op;
  */
 protected:
  bool ValidateOperandTypeSize (const Operand* pOp,dword dwOp,
				dword dwOpMask,bool bIsLastOperand);
  
  void CalcShortestLength();
  
 public:
  static void CalcNeedOpTypeMask(const Instruction* pInstr,
				 word& wNeedMask, // bits that has to be one
				 word& wNeedNotMask); // bits that has to be zero
  
  bool TryEncoding (const InverseOpcodeInfo* pInverseOpcodeInfo,
		    const Instruction* pInstr);
  
  void DoEncoding (DataDestPtr& dataPtr,word* pDispImmLocs = NULL,
		   byte bLengthenTricks = 0) const;
  
  bool DoEncodingLength (DataDestPtr& dataPtr,byte bLength,
			 word* pDispImmLocs = NULL,
			 byte* pLengthenTricksUsed = NULL) const;
  
  word GetLengths ();
  
  byte GetShortestLength() { return m_bShortestLength; }

  
};

#define TRY_DO_INSTR_ENCODE_PREPARE(pInstr)   word wNeedMask, wNeedNotMask; TryDoInstructionEncoding::CalcNeedOpTypeMask(pInstr,wNeedMask,wNeedNotMask);
#define TRY_DO_INSTR_ENCODE_MAY_ATTEMPT(pInverseOpcodeInfo) (!(~(pInverseOpcodeInfo)->m_nFlags & wNeedMask) && !((pInverseOpcodeInfo)->m_nFlags & wNeedNotMask))

//////////////////////////////////////////////////////////////////////

#endif // !defined(__TRY_DO_INSTRUCTION_ENCODING_H__INCLUDED_)





