/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// VarContentOp.cpp: implementation of the VarContentOp class.
//
//////////////////////////////////////////////////////////////////////

#include "VarContentOp.h"
#include "Variable.h"
#include "VariableOp.h"
#include "VarContStackPtr.h"
#include "VariableNConst.h"


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

void CalcOpVarContent(Variable* pVar,VariableOp* pVarOp)
{
  // attempt to calculate pVar->m_pContent from pVar->pSource, in a
  // recursive manner... i.e.  if we do not have it, try to see if any
  // of our parents may have something

  assert(pVar);
  assert(pVarOp);

  assert(!pVar->m_pVarContent);

  // TODO: eventually, this function should be done something like how
  // VariableN objects are pushed through the calculations with
  // automatic class detection and choosing of appropriate functions
  // to do the calculations...

  // for now we use a simpler approach

  VariableNBinaryOpAdd<dword>* pVarOpAdd;
  
  if ((pVarOpAdd = dynamic_cast<VariableNBinaryOpAdd<dword>*>(pVarOp)))
    {
      VarContStackPtr* pStackPtr;
      Variable32Const* pVarConst;
      
      if (pVarOpAdd->m_pSrc1 &&
	  (pStackPtr = dynamic_cast<VarContStackPtr*>(pVarOpAdd->m_pSrc1->m_pVarContent)) &&
	  (pVarConst = dynamic_cast<Variable32Const*>(pVarOpAdd->m_pSrc2)))
	{
	  pVar->m_pVarContent = 
	    new VarContStackPtr(*pStackPtr,pVarConst->FirstValue());
	}
    }
  
  VariableNBinaryOpSub<dword>* pVarOpSub;
  
  if ((pVarOpSub = dynamic_cast<VariableNBinaryOpSub<dword>*>(pVarOp)))
    {
      VarContStackPtr* pStackPtr;
      Variable32Const* pVarConst;
      
      if (pVarOpSub->m_pSrc1 &&
	  (pStackPtr = dynamic_cast<VarContStackPtr*>(pVarOpSub->m_pSrc1->m_pVarContent)) &&
	  (pVarConst = dynamic_cast<Variable32Const*>(pVarOpSub->m_pSrc2)))
	{
	  pVar->m_pVarContent = 
	    new VarContStackPtr(*pStackPtr,-pVarConst->FirstValue());
	}
    }
  
  
  





}
