/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// VariableCombine.cpp: implementation of the VariableCombine class.
//
//////////////////////////////////////////////////////////////////////

#include "VariableCombine.h" 
#include "VariableConvert.h" 
			     
#include <stdio.h>	     
			     
#include "VariableNConst.h" 
#include "VariableNConsts.h" 
#include "VariableNRange.h" 
#include "VariableNPattern.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

/*
  For anybody reading this: This file is (by comparison to the others)
  heavily commented. That does not mean that it's the only one well
  documented. It just means that it is a nasty piece af bad code. Bad
  because I could find no nice way of putting things. So I choose one
  at random and simply hope it's at least self-consistent...  */

//////////////////////////////////////////////////////////////////////

/*
VariableOp::VariableOp()
{

}

VariableOp::~VariableOp()
{ 

}
*/


// tests to see if 'inside' is entirely covered by 'container'

// try this first if full test requires a loop
#define SHORT_CIRCUIT_IS_VAR_INSIDE_BY_NUMBERS_INCLUDED \
{ \
  if (container->NumbersIncluded() < inside->NumbersIncluded()) \
    return false; \
}

template<class unsignedT>
bool IsVariableInside(const VariableNConst<unsignedT>* container,
		      const VariableNConst<unsignedT>* inside)
{
  return container->m_value == inside->m_value;
}

template<class unsignedT>
bool IsVariableInside(const VariableNConsts<unsignedT>* container,
		      const VariableNConst<unsignedT>* inside)
{
  return container->IsInside (inside->m_value);
}

template<class unsignedT>
bool IsVariableInside(const VariableNConst<unsignedT>* container,
		      const VariableNConsts<unsignedT>* inside)
{
  if (inside->m_vectValues.size() > 1)
    {
      assert (inside->m_vectValues[0] != inside->m_vectValues[1]);
      return false;
    }
  assert (inside->m_vectValues.size() > 0); // must have at least one value
  return container->m_value == inside->m_vectValues[0];
}

template<class unsignedT>
bool IsVariableInside(const VariableNConsts<unsignedT>* container,
		      const VariableNConsts<unsignedT>* inside)
{
  SHORT_CIRCUIT_IS_VAR_INSIDE_BY_NUMBERS_INCLUDED;
  
  typename VariableNConsts<unsignedT>::VectValues::const_iterator container_value = container->m_vectValues.begin();
  typename VariableNConsts<unsignedT>::VectValues::const_iterator inside_value = inside->m_vectValues.begin();
  
  while (inside_value != inside->m_vectValues.end())
    {
      for ( ; ; )
        {
          if (container_value == container->m_vectValues.end())
            return false;
          if (*container_value >= *inside_value)
            break;
          container_value++; // try next
        }
      if (*container_value > *inside_value)
        return false;      
      // ok, so inside_value was found, try next
      container_value++;
      inside_value++;
    }
  return true;
}

template<class unsignedT>
bool IsVariableInside(const VariableNRange<unsignedT>* container,
		      const VariableNConst<unsignedT>* inside)
{
  return container->IsInside (inside->m_value);
}

template<class unsignedT>
bool IsVariableInside(const VariableNConst<unsignedT>* container,
		      const VariableNRange<unsignedT>* inside)
{
  return container->m_value == inside->m_min &&
    container->m_value == inside->m_max;
}

template<class unsignedT>
bool IsVariableInside(const VariableNRange<unsignedT>* container,
		      const VariableNConsts<unsignedT>* inside)
{
  SHORT_CIRCUIT_IS_VAR_INSIDE_BY_NUMBERS_INCLUDED;
  
  FOR_EACH_CONST_ITER(typename VariableNConsts<unsignedT>::VectValues,inside->m_vectValues,inside_value)
    if (!container->IsInside (*inside_value))
      return false;
  return true;
}

template<class unsignedT>
bool IsVariableInside(const VariableNConsts<unsignedT>* container,
		      const VariableNRange<unsignedT>* inside)
{
  typename VariableNConsts<unsignedT>::VectValues::const_iterator min = 
    lower_bound(container->m_vectValues.begin(),
		container->m_vectValues.end(),inside->m_min);
  
  if (min == container->m_vectValues.end() || *min != inside->m_min)
    return false;
  
  typename VariableNConsts<unsignedT>::VectValues::const_iterator max = 
    lower_bound(container->m_vectValues.begin(),
		container->m_vectValues.end(),inside->m_max);
  
  if (max == container->m_vectValues.end() || *(--max) != inside->m_max)
    return false;
  
  if (inside->m_min <= inside->m_max) // proper range
    {
      assert ((unsignedT) distance(min,max) <= 
	      (inside->m_max - inside->m_min + 1)); // otherwise we have doublets
      
      return (unsignedT) distance(min,max) == 
	(inside->m_max - inside->m_min + 1);
    }

  // unproper distance the number of values in the list minus the ones
  // not in range must be the number of values in the range

  assert (container->m_vectValues.size() - (distance(min,max)-1) < 
	  inside->NumbersIncluded());

  return container->m_vectValues.size() - (distance(min,max)-1) == 
    inside->NumbersIncluded();
}

template<class unsignedT>
bool IsVariableInside(const VariableNPattern<unsignedT>* container,
		      const VariableNConst<unsignedT>* inside)
{
  return container->IsInside (inside->m_value);
}

template<class unsignedT>
bool IsVariableInside(const VariableNConst<unsignedT>* container,
		      const VariableNPattern<unsignedT>* inside)
{
  return (inside->m_known == ALL_ONES(unsignedT)) && 
    (container->m_value == inside->m_value);
}

template<class unsignedT>
bool IsVariableInside(const VariableNPattern<unsignedT>* container,
		      const VariableNConsts<unsignedT>* inside)
{
  SHORT_CIRCUIT_IS_VAR_INSIDE_BY_NUMBERS_INCLUDED;

  FOR_EACH_CONST_ITER(typename VariableNConsts<unsignedT>::VectValues,
		      inside->m_vectValues,inside_value)
    if (!container->IsInside (*inside_value))
      return false;
  return true;
}

template<class unsignedT>
bool IsVariableInside(const VariableNConsts<unsignedT>* container,
		      const VariableNPattern<unsignedT>* inside)
{
  // we'll have to generate all possible values, and check them

  // but first, check if it is at all possible
  
  SHORT_CIRCUIT_IS_VAR_INSIDE_BY_NUMBERS_INCLUDED;
  
  // every value that we generate must be part of the consts
  /*
    #define FOR_EACH_CONST_ITER(conttype,container,iter) for(conttype::const_iterator iter = (container).begin(); iter != (container).end(); ++iter)
  */
  assert (container->IsSortedConsistencyCheck());
  typename VariableNConsts<unsignedT>::VectValues::const_iterator iter = 
    container->m_vectValues.begin();
  assert (iter != container->m_vectValues.end()); // must contain at least one value (or would at least have short-
  // circuited, as pattern has at least on value)
  
  unsignedT value = inside->FirstValue();
  do
    {
      while (*iter < value)
        {
          // get next

          ++iter;
          if (iter == container->m_vectValues.end())
            return false; // out of values
        }
      if (*iter > value)
        return false; // missing value      
      // ok, so we found value, get next
    }
  while (inside->NextValue(value));  
  
  /*
  FOR_EACH_CONST_ITER(VariableNConsts<unsignedT>::VectValues,inside->m_vectValues,inside_value)
    if (!container->IsInside (*inside_value))
      return false;
  */
  return true;
}

template<class unsignedT>
bool IsVariableInside(const VariableNPattern<unsignedT>* container,
		      const VariableNPattern<unsignedT>* inside)
{
  return ((container->m_known & inside->m_known) == inside->m_known) &&
    (container->m_value & inside->m_known) == inside->m_value;
}

template<class unsignedT>
bool IsVariableInside(const VariableNRange<unsignedT>* container,
		      const VariableNRange<unsignedT>* inside)
{
  if (container->m_min == container->m_max + 1)
    return true; // covers entire range (so 'inside' has to be inside)
  if (container->m_min <= container->m_max) // container proper range (not wrapping)
    {
      return inside->m_min <= inside->m_max && // inside proper range (not wrapping)
        container->m_min <= inside->m_min && container->m_max >= inside->m_max;
    }
  else
    {
      if (inside->m_min <= inside->m_max) // inside proper range (not wrapping)
        return 
	  (container->m_min <= inside->m_min && 
	   container->m_min <= inside->m_max) ||
          (container->m_max >= inside->m_min && 
	   container->m_max >= inside->m_max);
      else // inside non proper range (wrapping)
        return 
	  container->m_min <= inside->m_min && 
	  container->m_max >= inside->m_max;
    }
}

template<class unsignedT>
bool IsVariableInside(const VariableNPattern<unsignedT>* container,
		      const VariableNRange<unsignedT>* inside)
{
  // if the range is non-proper (wrapping) then it contain both 00 and
  // ff, so pattern must allow everything
  
  if (inside->m_min > inside->m_max)
    return !(container->m_known); // if any bit is known, pattern can't be anything

  // range is proper, so we require that min value is part of pattern,
  // and we must be able to flip all bits from the left up to the
  // rightmost bit which differs between min and max
  
  if (!container->IsInside(inside->m_min))
    return false;
  
  unsignedT flip_bits = inside->m_min ^ inside->m_max;
  
  // flip bits may now be 00110111, but it is the 00100000 we're interested in
  
  unsignedT all_flip = flip_bits;
  
  while (flip_bits >>= 1)
    all_flip |= flip_bits;

  // if pattern is to take this, then all bits of known matching
  // all_flip must be zero

  return !(container->m_known & all_flip);
}

template<class unsignedT>
bool IsVariableInside(const VariableNRange<unsignedT>* container,
		      const VariableNPattern<unsignedT>* inside)
{
  // if the pattern is to fit in a proper range, the minimum and
  // maximum possible values from the pattern must fit into the range

  if (container->m_min > container->m_max)
    return (inside->m_value >= container->m_min) &&
      ((inside->m_value | inside->m_known) <= container->m_max);
  
  // if the pattern is to fit in a non-proper range, any values NOT in
  // the range, must not be in the pattern
  
  if (container->m_max + 1 == container->m_min)
    return true; // range covers anything
  
  // [container->m_max+1,container->m_min-1] is the range that must
  // NOT be described by the pattern
  
  if (inside->IsInside(container->m_max+1))
    return false;
  
  unsignedT flip_bits = (container->m_max+1) ^ (container->m_min-1);
  
  // flip bits may now be 00110111, but it is the 00100000 we're
  // interested in
  
  unsignedT all_flip = flip_bits;
  
  while (flip_bits >>= 1)
    all_flip |= flip_bits;
  
  // if pattern is NOT to take this, then all bits of known matching
  // all_flip must be one
  
  return !(~inside->m_known & all_flip);
}

template<class unsignedT>
bool IsVariableInside(const VariableNSplit<unsignedT>* container,
		      const VariableNSplit<unsignedT>* inside);

//////////////////////////////////////////////////////////////////////

#define GENERATE_VAR_INT_BINARY_ORDERED_OP(name)         // rettype=bool argtype=const VariableN<unsignedT>* funcnameprefix=VariableCombine defret=return false; mayconvert=0,1

GENERATE_VAR_INT_BINARY_ORDERED_OP(IsVariableInside)        
  
//////////////////////////////////////////////////////////////////////

template<class unsignedT>
bool GenericIsVariableInside(const VariableN<unsignedT>* contain,
			     const VariableN<unsignedT>* inside)
{
  if (!contain || contain->NothingKnown())
    return true;
  // so container at least has some constraint
  
  if (!inside || inside->NothingKnown())
    return false; // but false is all-inclusive, so it can't be inside

  return VariableCombineIsVariableInside(contain,inside);
}

template<class unsignedT>
bool IsVariableInside(const VariableNSplit<unsignedT>* container,
		      const VariableNSplit<unsignedT>* inside)
{
  // both high and low parts must be inside
  
  return GenericIsVariableInside(container->m_high,inside->m_high) &&
    GenericIsVariableInside(container->m_low,inside->m_low);
}

//////////////////////////////////////////////////////////////////////

template<class unsignedT>
VariableN<unsignedT>* DoCombine(const VariableNConst<unsignedT>* lhs,
				const VariableNConst<unsignedT>* rhs)
{
  if (lhs->m_value == rhs->m_value)
    return new VariableNConst<unsignedT>(*lhs);
  return new VariableNConsts<unsignedT>(lhs->m_value,rhs->m_value);
}

template<class unsignedT>
unsignedT CostOfCombine(const VariableNConst<unsignedT>* /* lhs */,
			const VariableNConst<unsignedT>* /* rhs */)
{
  return 0; // no cost, no extra fill in
}

template<class unsignedT>
VariableN<unsignedT>* DoCombine(const VariableNConsts<unsignedT>* lhs,
				const VariableNConst<unsignedT>* rhs)
{
  VariableNConsts<unsignedT>* pNew = new VariableNConsts<unsignedT>(*lhs);
  pNew->InsertValue (rhs->m_value);
  return pNew;
}

template<class unsignedT>
unsignedT CostOfCombine(const VariableNConsts<unsignedT>* /* lhs */,
			const VariableNConst<unsignedT>* /* rhs */)
{
  return 0; // no cost, no extra fill in
}

template<class unsignedT>
VariableN<unsignedT>* DoCombine(const VariableNConsts<unsignedT>* lhs,
				const VariableNConsts<unsignedT>* rhs)
{
  VariableNConsts<unsignedT>* pNew = new VariableNConsts<unsignedT>(*lhs);
  
  FOR_EACH_CONST_ITER(typename VariableNConsts<unsignedT>::VectValues,
		      rhs->m_vectValues,rhs_value)
    pNew->InsertValue (*rhs_value);
  return pNew;
}

template<class unsignedT>
unsignedT CostOfCombine(const VariableNConsts<unsignedT>* /* lhs */,
			const VariableNConsts<unsignedT>* /* rhs */)
{
  return 0; // no cost, no extra fill in
}

template<class unsignedT>
VariableN<unsignedT>* DoCombine(const VariableNPattern<unsignedT>* lhs,
				const VariableNPattern<unsignedT>* rhs)
{
  VariableNPattern<unsignedT>* pNew = new VariableNPattern<unsignedT>(*lhs);
  pNew->AddPattern(rhs);
  return pNew;
}

template<class unsignedT>
unsignedT CostOfCombine(const VariableNPattern<unsignedT>* lhs,
			const VariableNPattern<unsignedT>* rhs)
{
  return lhs->CostOfPatternMerge(rhs);
}

template<class unsignedT>
VariableN<unsignedT>* DoCombine(const VariableNRange<unsignedT>* lhs,
				const VariableNConst<unsignedT>* rhs)
{
  VariableNRange<unsignedT>* pNew = new VariableNRange<unsignedT>(*lhs);
  pNew->InsertValue(rhs->m_value);
  return pNew;
}

template<class unsignedT>
unsignedT CostOfCombine(const VariableNRange<unsignedT>* lhs,
			const VariableNConst<unsignedT>* rhs)
{
  return lhs->CostOfInsertValue (rhs->m_value);
}

template<class unsignedT>
VariableN<unsignedT>* DoCombine(const VariableNRange<unsignedT>* lhs,
				const VariableNRange<unsignedT>* rhs)
{
  assert (!IsVariableInside(lhs,rhs)); // otherwise useless (and we shouldn't have been called)
  // so we know something has to be added
  
  if (lhs->IsInside (rhs->m_min) && lhs->IsInside (rhs->m_max)) // and as something had to be added, we're now covering entire range
    return NULL; // all range
  
  VariableNRange<unsignedT>* pNew = new VariableNRange<unsignedT>(*lhs);
  
  if (lhs->IsInside (rhs->m_min)) 
    pNew->ExpandRangeMax (rhs->m_max); 
  else if (lhs->IsInside (rhs->m_max))
    pNew->ExpandRangeMin (rhs->m_min);
  
  // ok so the two ranges are not covering each other i.e., we have
  // two separate ranges
  // 
  // if we had a class covering that case, that would of course be
  // used, but right now we'll find cheapest add
  
  else if ((rhs->m_max - lhs->m_min) < (lhs->m_max - rhs->m_min))
    pNew->ExpandRangeMax (rhs->m_max);
  else
    pNew->ExpandRangeMin (rhs->m_min);
  
  return pNew;
}

template<class unsignedT>
unsignedT CostOfCombine(const VariableNRange<unsignedT>* lhs,
			const VariableNRange<unsignedT>* rhs)
{
  assert (!IsVariableInside(lhs,rhs)); // otherwise useless (and we shouldn't have been called)
  // so we know something has to be added
  
  if (lhs->IsInside (rhs->m_min) || lhs->IsInside (rhs->m_max))
    return 0; // the ranges already overlapped, so nothing was added 'by accident'
  
  // ok so the two ranges are not covering each other i.e., we have
  // two separate ranges
  // 
  // if we had a class covering that case, that would of course be
  // used, but right now we'll find cheapest add
  
  else if ((rhs->m_max - lhs->m_min) < (lhs->m_max - rhs->m_min))
    return rhs->m_min - lhs->m_max - 1;
  else
    return lhs->m_min - rhs->m_max - 1;
}

template<class unsignedT>
VariableN<unsignedT>* DoCombine(const VariableNPattern<unsignedT>* lhs,
				const VariableNConst<unsignedT>* rhs)
{
  VariableNPattern<unsignedT>* pNew = new VariableNPattern<unsignedT>(*lhs);
  pNew->InsertValue(rhs->m_value);
  return pNew;
}


template<class unsignedT>
unsignedT CostOfCombine(const VariableNPattern<unsignedT>* lhs,
			const VariableNConst<unsignedT>* rhs)
{
  return lhs->CostOfInsertValue (rhs->m_value);
}

template<class unsignedT>
VariableN<unsignedT>* DoCombine(const VariableNPattern<unsignedT>* lhs,
				const VariableNConsts<unsignedT>* rhs)
{
  // if necessary (i.e., if combining vars that will be a function
  // pointer), we should not create any fill in at all (by creating a
  // new consts), but we as now do not now if that is necessary, so we
  // make a pattern
  
  VariableNPattern<unsignedT>* pNew = new VariableNPattern<unsignedT>(*lhs);
  FOR_EACH_CONST_ITER(typename VariableNConsts<unsignedT>::VectValues,
		      rhs->m_vectValues,rhs_value)
    pNew->InsertValue (*rhs_value);
  return pNew;
}

template<class unsignedT>
unsignedT CostOfCombine(const VariableNPattern<unsignedT>* lhs,
			const VariableNConsts<unsignedT>* rhs)
{
  // fill in are the number of values represented in the expanded
  // pattern, that were not in the consts list
  
  unsignedT known = lhs->m_known;
  unsignedT num_consts_in_old_pattern = 0;
  
  FOR_EACH_CONST_ITER(typename VariableNConsts<unsignedT>::VectValues,
		      rhs->m_vectValues,rhs_value)
    {
      known &= ~(lhs->m_value ^ *rhs_value);
      if (lhs->IsInside (*rhs_value))
        num_consts_in_old_pattern++;
    }
  
  return VariableNPattern<unsignedT>::NumbersIncluded(known) - 
    lhs->NumbersIncluded() - 
    rhs->NumbersIncluded() + num_consts_in_old_pattern;
}

template<class unsignedT>
VariableN<unsignedT>* DoCombine(const VariableNSplit<unsignedT>* lhs,
				const VariableNSplit<unsignedT>* rhs);

template<class unsignedT>
unsignedT CostOfCombine(const VariableNSplit<unsignedT>* /* lhs */,
			const VariableNSplit<unsignedT>* /* rhs */)
{
  // the cost of an combination like this is too messy to calculate
  // (possible, but a little messy)
  //
  // it's not so bad anyway, because this will make split/split pairs
  // combined quickly, and at least we're of same type, so
  
  return 0;
}

// TODO: Combine and CostOfCombine for const-range  and  range-pattern

//////////////////////////////////////////////////////////////////////

// Note: when conversion takes place, the cost of that is not included
// in CostOfCombine!  
// OUTDATED: Note: may not convert, as conversion
// cost not included in CostOfCombine!
//

#define GENERATE_VAR_INT_BINARY_NOT_ORDERED_OP_COMB(name)         // rettype=VariableN<unsignedT>* argtype=const VariableN<unsignedT>* funcnameprefix=VariableCombine defret=return NULL; invops=1

GENERATE_VAR_INT_BINARY_NOT_ORDERED_OP_COMB(DoCombine)
  
#define ASSERT_FALSE assert(false);
  
#define GENERATE_VAR_INT_BINARY_NOT_ORDERED_OP_COMBCOST(name)     // rettype=unsignedT argtype=const VariableN<unsignedT>* funcnameprefix=VariableCombine defret=return ALL_ONES(unsignedT); invops=1
  
GENERATE_VAR_INT_BINARY_NOT_ORDERED_OP_COMBCOST(CostOfCombine) 
  
//////////////////////////////////////////////////////////////////////
  
#include "gen/VariableCombine_TypedCalls.h" // dep_chk_unknown 
  
  //////////////////////////////////////////////////////////////////////
  
  template<class unsignedT>
VariableN<unsignedT>* DoCombine(const VariableNSplit<unsignedT>* lhs,
				const VariableNSplit<unsignedT>* rhs)
{
  typename VariableN<unsignedT>::half_var_type* pLow  = 
    VariableCombineDoCombine(lhs->m_low,rhs->m_low);
  typename VariableN<unsignedT>::half_var_type* pHigh = 
    VariableCombineDoCombine(lhs->m_high,rhs->m_high);
  
  if (pLow || pHigh)
    return new VariableNSplit<unsignedT>(pLow,pHigh);
  return NULL;
}

//////////////////////////////////////////////////////////////////////

#include <set>

using namespace std;

template<typename unsignedT>
class InputIntermediateVariableN
{
public:
  InputIntermediateVariableN(bool bIntermediate,
			     const VariableN<unsignedT>* pInput)
  {
    assert (pInput);
    
    m_bIntermediate = bIntermediate;
    m_var.m_pInput = pInput;
  }
  
public:
  union
  {
    const VariableN<unsignedT>* m_pInput;        // for use when input
    VariableN<unsignedT>*	m_pIntermediate; // for use when result of combine
  } m_var;
  bool			      m_bIntermediate;	
  
public:
  bool operator<(const InputIntermediateVariableN& rhs) const 
  { return m_var.m_pInput->CompareComb(rhs.m_var.m_pInput) < 0; } 
};

template<typename unsignedT>
struct MultiSetInputIntermediateVariableN
{
  typedef multiset< InputIntermediateVariableN<unsignedT> > Type;
};
/*
template<typename unsignedT>
struct CombineVarNums
{
public:
  CombineVarNums(int i,int j)
  {
    assert (i > j);

    m_i = i;
    m_j = j;
  }
public:
  int m_i, m_j;
};
*/

template<typename unsignedT>
class CostOfCombineVars
{
public:
  CostOfCombineVars(unsignedT f,
                    const typename MultiSetInputIntermediateVariableN<unsignedT>::Type::iterator& i,
                    const typename MultiSetInputIntermediateVariableN<unsignedT>::Type::iterator& j)
  {
    // assert (i > j); // can't compare these iterators (efficiently)
    // assert (pIJ->m_i > pIJ->m_j);
    
    m_fillin = f;
    m_i = i;
    m_j = j;
  }
  
public:
  unsignedT m_fillin;
  typename MultiSetInputIntermediateVariableN<unsignedT>::Type::iterator m_i;
  typename MultiSetInputIntermediateVariableN<unsignedT>::Type::iterator m_j;
  /*
  CombineVarNums<unsignedT>* m_pIJ; // we have a pointer here instead of the real values, because I want
  // to be able to change the values while this object is a part of a (multi)set, which I may
  // not do otherwise. I want to be able to do it because it does not change the ordering, and
  // so does not destroy the ordered set. The only other way would be to delete the object from
  // the set, and insert and altered version
  */
public:
  bool operator<(const CostOfCombineVars& rhs) const 
  { return m_fillin < rhs.m_fillin; } 
};

template<typename unsignedT>
struct MultiSetCostOfCombineVars
{
  typedef multiset< CostOfCombineVars<unsignedT> > Type;
};
/*
  template<typename unsignedT>
void RemoveMultiSetCostOfCombineVarsReferences(MultiSetCostOfCombineVars<unsignedT>::Type& multisetCombiCost,
					       int j,int i,int n)
{
  // removes elements that point to j
  // removes elements that point to i
  // redirect pointers from n-1 to i
  // j and i may be the same

  for(MultiSetCostOfCombineVars<unsignedT>::Type::iterator iter = multisetCombiCost.begin();
      iter != multisetCombiCost.end(); )
    {
      // ok, so does it have anything with i or j to do?
      
      CombineVarNums<unsignedT>* iter_pIJ = iter->m_pIJ;

      if (iter_pIJ->m_i == i || iter_pIJ->m_i == j || 
          iter_pIJ->m_j == i || iter_pIJ->m_j == j)
        {
          // item is to be deleted
          MultiSetCostOfCombineVars<unsignedT>::Type::iterator erase = iter;
          ++iter;
          delete iter_pIJ;
          multisetCombiCost.erase (erase);
        }
      else
        {
          // is it the last element (n-1)?
          // the redirect it to be i
          
          if (iter_pIJ->m_i == n-1)
            {
              if (iter_pIJ->m_j > i)
                {
                  iter_pIJ->m_i = iter_pIJ->m_j;
                  iter_pIJ->m_j = i;
                }
              else
                iter_pIJ->m_i = i;
            }
          ++iter;
        }
    }
}
*/
/*
template<class unsignedT>
int CompareCombTypes(const VariableN<unsignedT>* pVarA1,const VariableN<unsignedT>* pVarA2,
		     const VariableN<unsignedT>* pVarB1,const VariableN<unsignedT>* pVarB2)
{
  // returns < 0 for A best, > 0 for B best, 0 for a draw

  int nTypeA1 = pVarA1->VariableNCompareCombType();
  int nTypeA2 = pVarA2->VariableNCompareCombType();
  int nTypeB1 = pVarB1->VariableNCompareCombType();
  int nTypeB2 = pVarB2->VariableNCompareCombType();

  bool bSameA = nTypeA1 == nTypeA2;
  bool bSameB = nTypeA1 == nTypeA2;

  if (bSameA != bSameB)
    return bSameA ? -1 : 1;

  // so both same or both not same

  if (bSameA) 
    {
      // both same;

      return nTypeA1 - nTypeB1;
    }

  // so compare the ones with smallest number, least number winning

  int nMinA = min(nTypeA1,nTypeA2);
  int nMinB = min(nTypeB1,nTypeB2);

  if (nMinA - nMinB)
    return nMinA-nMinB;

  // (smallest numbers same) compare largest numbers, least number winning

  int nMaxA = max(nTypeA1,nTypeA2);
  int nMaxB = max(nTypeB1,nTypeB2);

  return nMaxA-nMaxB;  
}
*/

template<class unsignedT>
int CompareCombTypesSame(const VariableN<unsignedT>* pVarA1,
			 const VariableN<unsignedT>* pVarA2,
			 const VariableN<unsignedT>* pVarB1,
			 const VariableN<unsignedT>* pVarB2)
{
  // returns < 0 for A best, > 0 for B best, 0 for a draw
  
  int nTypeA1 = pVarA1->VariableNCompareCombType();
  int nTypeA2 = pVarA2->VariableNCompareCombType();
  int nTypeB1 = pVarB1->VariableNCompareCombType();
  int nTypeB2 = pVarB2->VariableNCompareCombType();
  
  bool bSameA = (nTypeA1 == nTypeA2);
  bool bSameB = (nTypeB1 == nTypeB2);

  // is one same and the other not?

  if (bSameA != bSameB)
    return bSameA ? -1 : 1;

  return 0;
}


template<class unsignedT>
int CompareCombNumbersIncluded(const VariableN<unsignedT>* pVarA1,
			       const VariableN<unsignedT>* pVarA2,
			       const VariableN<unsignedT>* pVarB1,
			       const VariableN<unsignedT>* pVarB2)
{
  // returns < 0 for A best, > 0 for B best, 0 for a draw
  
  int nNumA1 = pVarA1->NumbersIncluded();
  int nNumA2 = pVarA2->NumbersIncluded();
  int nNumB1 = pVarB1->NumbersIncluded();
  int nNumB2 = pVarB2->NumbersIncluded();

  int nSumA = nNumA1 + nNumA2;
  int nSumB = nNumB1 + nNumB2;

  assert (nNumA1);
  assert (nNumA2);
  assert (nNumB1);
  assert (nNumB2);

  bool bOverflowA = nSumA < nNumA1;
  bool bOverflowB = nSumB < nNumB1;

  if (bOverflowA != bOverflowB) // best to have many numbers included
    return bOverflowA ? -1 : 1;

  return -(nSumA - nSumB); // best to have many numbers included
}


template<class unsignedT>
void CreateCombinedVar(VariableN<unsignedT>*& pCombinedVar,
		       bool& bHasCombinedVar,
		       /* const typename MultiSetInputIntermediateVariableN<unsignedT>::Type& setVars ,*/
		       const typename MultiSetCostOfCombineVars<unsignedT>::Type::const_iterator& comb)
{
  if (bHasCombinedVar)
    return;
  
  pCombinedVar = VariableCombineDoCombine((comb->m_j)->m_var.m_pInput,
                                          (comb->m_i)->m_var.m_pInput);
  
  // TODO: simplify pCombinedVar if possible (call simplification routine)
  
  bHasCombinedVar = true;
}


//#define CALC_INSERT_FILL_IN_INFO(i,j) 
//{ 
//  unsignedT nFillIn = VariableCombineCostOfCombine(vectVars[(i)].m_var.m_pInput, 
//						   vectVars[(j)].m_var.m_pInput);  
//  /* make sure VariableCombineCostOfCombine is symmetric */ 
//  assert (nFillIn == VariableCombineCostOfCombine(vectVars[(j)].m_var.m_pInput, 
//						  vectVars[(i)].m_var.m_pInput));  
//  CombineVarNums<unsignedT>* pIJ = new CombineVarNums<unsignedT>((i),(j));  
//  multisetCombiCost.insert(CostOfCombineVars<unsignedT> (nFillIn,pIJ));	
//} 

// CostOfCombineVars(v1,v2) will return 0 if the combination is not
// possible (not implemented) this is both good and bad.  IF we
// enventually will have to do this combination it, it will return
// NULL anyway, so we can as well have it directly not to make any
// unnecessary combining before getting the all-inclusive variable
// anyway.  It's bad if the combination routine would get around the
// problem somehow Anyway, it's at least stable. But it should be
// implemented..., so it's a TODO above

template<class unsignedT>
VariableN<unsignedT>* VariableCombine(typename ListVariableN<unsignedT>::Type listVars)
{
  assert (listVars.size ());
  
  // first get rid of all items that are inside each other
  // but we must do it in a stable way, so first the list must be sorted
  
  typename MultiSetInputIntermediateVariableN<unsignedT>::Type setVars;
  
  // insert everything
  
  assert (listVars.size());
  
  for(typename ListVariableN<unsignedT>::Type::const_iterator iter = 
	listVars.begin();
      iter != listVars.end(); ++iter)
    {
      if (!*iter) // nothing known by this, so we can know nothing after combine
        return NULL;
      
      setVars.insert (InputIntermediateVariableN<unsignedT>(false,*iter));
    }
  
  assert (setVars.size ()); // it must be something as listVars.size() was something, and something was added
  
  // ok, so the (multi)setVars is already sorted (on insertation)

  // now we can remove items, and as the vector is sorted, items will
  // be removed in a stable way (unfortunately, removing from a vector
  // is more troublesome than removing from a list as was done before,
  // but, it has to be stable)

  //typename MultiSetInputIntermediateVariableN<unsignedT>::Type::reverse_iterator a = setVars.rbegin();
  //typename MultiSetInputIntermediateVariableN<unsignedT>::Type::reverse_iterator b = setVars.rbegin();
  //typename MultiSetInputIntermediateVariableN<unsignedT>::Type::reverse_iterator c = setVars.rbegin();
  //typename MultiSetInputIntermediateVariableN<unsignedT>::Type::reverse_iterator d = setVars.rbegin();
   
  FOR_EACH_CONST_ITER(typename MultiSetInputIntermediateVariableN<unsignedT>::Type,setVars,contain)
    {
      assert (contain->m_var.m_pInput); // now, it would be ok if contain is null, which represents an undetermined
      // variable, but this should have been taken care of
      
      for(typename MultiSetInputIntermediateVariableN<unsignedT>::Type::iterator inside = setVars.begin(); 
          inside != setVars.end(); )
        {
          if (contain != inside &&
              VariableCombineIsVariableInside(contain->m_var.m_pInput,
                                                                                  inside->m_var.m_pInput))
            {
              // we can remove inside from list as it is already in contain
              
              SET_MAP_ERASE_CONTINUE_ITER(setVars,inside);
                  // inside = setVars.erase (inside);
            }
          else
            ++inside;
        }
    }
  
  assert (setVars.size ()); // one element must still remain

  // so now we got rid of the easy ones we now have to combine
  // everything into one Variable
  
  // it should be done so that the resulting Variable is as small as
  // possible, and in a stable way, so that the same input (in another
  // order) results in the same output. It is also so that different
  // orderings would cause different output if used directly, so use
  // some stable 'sort' when deciding how to combine
  
  // we'll use a test-all approach: try to combine all pairs of
  // Variables, and at each step combine the pair giving the least
  // fill-in after each step, the newly combined variable is inserted
  // as an original variable and tested against everyone else
  
  // we're also not suppose to kill or change any variables that came
  // as input to the function, but should of course remove all
  // intermediate results
  
  // all vars are in vectVars

  // let's compute the fill-in of combining any two input variables
  // and then combine the variables with the least fill in

  // we do it using a lazy approach, keep a set of all possible
  // combinations sorted by the fill in they would cause. For each
  // step, combine on of the ones with least fill in. If there are
  // several with the same fill in, determine which one is most
  // suitable (favour same type combinations, and so on)

  typename MultiSetCostOfCombineVars<unsignedT>::Type multisetCombiCost;
  
  FOR_EACH_ITER(typename MultiSetInputIntermediateVariableN<unsignedT>::Type,
		setVars,i)
    {
      for (typename MultiSetInputIntermediateVariableN<unsignedT>::Type::iterator j = setVars.begin();
           j != i; ++j)
        {
          unsignedT nFillIn = VariableCombineCostOfCombine(i->m_var.m_pInput, 
                                                           j->m_var.m_pInput);
          /* make sure VariableCombineCostOfCombine is symmetric */ 
          assert (nFillIn == VariableCombineCostOfCombine(j->m_var.m_pInput, 
                                                          i->m_var.m_pInput)); 
          multisetCombiCost.insert(CostOfCombineVars<unsignedT> (nFillIn,i,j));
        }
    }
  
  // now, while we have more than one variable left
  
  while (setVars.size() > 1)
    {
      /*
	MultiSetCostOfCombineVars<unsignedT>::Type::const_iterator docomb = multisetCombiCost.begin();
      MultiSetCostOfCombineVars<unsignedT>::Type::const_iterator trycomb = docomb;

      ++trycomb;

      VariableN<unsignedT>* pCombinedVar = NULL;
      bool bHasCombinedVar = false;

      for( ; trycomb != multisetCombiCost.end(); ++trycomb)
        {
          assert (trycomb->m_fillin >= docomb->m_fillin);
          if (trycomb->m_fillin > docomb->m_fillin)
            break;

          VariableN<unsignedT>* pTryCombinedVar = NULL;
          bool bHasTryCombinedVar = false;

          // trycomb and docomb would cause the same fill-in
          
          // now the i and j point into a sorted list, so we could be satisfied with the
          // first one found, unfortunately, 

          // first try some simple compare so we usually do not have to generate the actual combination

          const VariableN<unsignedT>* m_pInputDo1  = pVars[ docomb->m_pIJ->m_i].m_var.m_pInput;
          const VariableN<unsignedT>* m_pInputDo2  = pVars[ docomb->m_pIJ->m_j].m_var.m_pInput;
          const VariableN<unsignedT>* m_pInputTry1 = pVars[trycomb->m_pIJ->m_i].m_var.m_pInput;
          const VariableN<unsignedT>* m_pInputTry2 = pVars[trycomb->m_pIJ->m_j].m_var.m_pInput;

          // if one of the suggested combinations is from two wars of same kind, prefer that one

          int nComp = CompareCombTypesSame (m_pInputDo1,m_pInputDo2,m_pInputTry1,m_pInputTry2);

          if (nComp < 0)
            goto keep_docomb;
          if (nComp > 0)
            goto use_trycomb;

          // compare by number of incoming variables

          nComp = CompareCombNumbersIncluded (m_pInputDo1,m_pInputDo2,m_pInputTry1,m_pInputTry2);

          if (nComp < 0)
            goto keep_docomb;
          if (nComp > 0)
            goto use_trycomb;

          // so still even, arrgh... We'll have to make a stable compare based on the contents,
          // not on how good the combination is... We could of course just make a random choice,
          // but to be stable, we'll do a compare

          // the nasty thing about this compare is that it compares the variables not based on content
          // but rather based on what class they come in. I do not like this. However, as there are many
          // classes it's a mess writing comparision functions for all possible combinations.
          // And the real number of possible comparisions come from using VariableNSplit (Which comes from
          // for example an word put together from two (possibly) different classes of Byte. And compare
          // all those variants by the variants of everything else.
          // One way of comparing everything would be to use the First/NextValue or First/NextValueRange
          // functions and iterate through the CONTENTS of the variables making a GOOD comparision. But
          // unfortunately the iterations may be through at least 2e9 values if getting an pattern as
          // input being having known bits 0x10000001 and another being known 0x00000001 and the first one
          // having the most significant bit 0. So I can't think of a general way of comparing different
          // patterns,ranges,consts etc without the possibility of a BIIIG execution time for a little
          // and frequent compare. So therefore, i resort to the UGLY method of comparing containers
          // instead of content.
          // a litte upside: It the containers are of the same kind, we'll compare them by content,
          // as it's possible to write comparison by same container efficiently
          
          //
          // int nComp = CompareCombTypes (m_pInputDo1,m_pInputDo2,m_pInputTry1,m_pInputTry2);
          //
          // if (nComp < 0)
          //   goto keep_docomb;
          // if (nComp > 0)
          //   goto use_trycomb;
          // 

          // Now if we've come this far, I do create the combined object, because then I only have to
          // compare two objects, instead of two from each input to the other two in which case I must
          // first determine which of the two from each side is the better so I compare in a stable way

          CreateCombinedVar<unsignedT>(pCombinedVar,bHasCombinedVar,pVars,docomb);
          CreateCombinedVar<unsignedT>(pTryCombinedVar,bHasTryCombinedVar,pVars,trycomb);

          assert (bHasCombinedVar);
          assert (bHasTryCombinedVar);

          // ok, so now we can compare them on type and (possibly) content
          // if any ended up a null, then it's a worse choice than the other
          // if both null, we can as well keep to old one

          if (!pTryCombinedVar)
            goto keep_docomb;
          if (!pCombinedVar)
            goto use_trycomb;

          assert (pCombinedVar && pTryCombinedVar);

          nComp = pCombinedVar->CompareComb(pTryCombinedVar);

          if (nComp > 0)
            goto use_trycomb;
          assert (nComp != 0); // otherwise of same type and same content, 
          // one SHOULD have been discarded by container/inside call of VariableCombineIsVariableInside 
          // in beginning of this routine or after each combine

          goto keep_docomb; // one line down, he he

        keep_docomb: // so new one was not better, delete any temporary created for comparison
          delete pTryCombinedVar;
          continue;
        use_trycomb: // new try was better, delete old
          delete pCombinedVar;
          pCombinedVar = pTryCombinedVar;
          bHasCombinedVar = bHasTryCombinedVar;
          docomb = trycomb;
        }

      // ok, so combine using docomb
      */

      // create the new element (if we do not have it yet)
      /*
        CreateCombinedVar<unsignedT>(pCombinedVar,bHasCombinedVar,pVars,docomb);
        assert (bHasCombinedVar);
        // TODO: try to simplify variable... (in CreateCombinedVar)
      */

      // find combination with smallest fill-in as multisetCombiCost
      // is sorted, it's normally the first one, but, if there are
      // several with the same fill-in, we must check them too, as
      // they are not sorted within the same fill-in thay could be
      // sorted by their order in setVars, but that would force us to
      // compare the iterators (m_i, m_j), and because they are not
      // random access iterators, their compare would be lengthy, so
      // instead, we compare the actual variables and result

      typename MultiSetCostOfCombineVars<unsignedT>::Type::const_iterator docomb = multisetCombiCost.begin();
      typename MultiSetCostOfCombineVars<unsignedT>::Type::const_iterator trycomb = docomb;
      
      ++trycomb;
      
      VariableN<unsignedT>* pCombinedVar = NULL;
      bool bHasCombinedVar = false;
      
      for( ; trycomb != multisetCombiCost.end(); ++trycomb)
        {
          assert (trycomb->m_fillin >= docomb->m_fillin);
          if (trycomb->m_fillin > docomb->m_fillin)
            break;
	  
          VariableN<unsignedT>* pTryCombinedVar = NULL;
          bool bHasTryCombinedVar = false;

          // trycomb and docomb would cause the same fill-in
          
          // now the i and j point into a sorted list, so we could be
          // satisfied with the first one found, unfortunately,
	  
          // first try some simple compare so we usually do not have
          // to generate the actual combination
	  
          const VariableN<unsignedT>* m_pInputDo1  = 
	    ( docomb->m_i)->m_var.m_pInput;
          const VariableN<unsignedT>* m_pInputDo2  = 
	    ( docomb->m_j)->m_var.m_pInput;
          const VariableN<unsignedT>* m_pInputTry1 = 
	    (trycomb->m_i)->m_var.m_pInput;
          const VariableN<unsignedT>* m_pInputTry2 = 
	    (trycomb->m_j)->m_var.m_pInput;
	  
          // if one of the suggested combinations is from two wars of
          // same kind, prefer that one

          int nComp = CompareCombTypesSame (m_pInputDo1,m_pInputDo2,
					    m_pInputTry1,m_pInputTry2);
	  
          if (nComp < 0)
            goto keep_docomb;
          if (nComp > 0)
            goto use_trycomb;
	  
          // compare by number of incoming variables
	  
          nComp = CompareCombNumbersIncluded (m_pInputDo1,m_pInputDo2,
					      m_pInputTry1,m_pInputTry2);
	  
          if (nComp < 0)
            goto keep_docomb;
          if (nComp > 0)
            goto use_trycomb;
	  
          // so still even, arrgh... We'll have to make a stable
          // compare based on the contents, not on how good the
          // combination is... We could of course just make a random
          // choice, but to be stable, we'll do a compare
	  
          // the nasty thing about this compare is that it compares
          // the variables not based on content but rather based on
          // what class they come in. I do not like this. However, as
          // there are many classes it's a mess writing comparision
          // functions for all possible combinations.  And the real
          // number of possible comparisions come from using
          // VariableNSplit (Which comes from for example an word put
          // together from two (possibly) different classes of
          // Byte. And compare all those variants by the variants of
          // everything else.  One way of comparing everything would
          // be to use the First/NextValue or First/NextValueRange
          // functions and iterate through the CONTENTS of the
          // variables making a GOOD comparision. But unfortunately
          // the iterations may be through at least 2e9 values if
          // getting an pattern as input being having known bits
          // 0x10000001 and another being known 0x00000001 and the
          // first one having the most significant bit 0. So I can't
          // think of a general way of comparing different
          // patterns,ranges,consts etc without the possibility of a
          // BIIIG execution time for a little and frequent
          // compare. So therefore, i resort to the UGLY method of
          // comparing containers instead of content.  a litte upside:
          // It the containers are of the same kind, we'll compare
          // them by content, as it's possible to write comparison by
          // same container efficiently
          
          //
          // int nComp = CompareCombTypes (m_pInputDo1,m_pInputDo2,m_pInputTry1,m_pInputTry2);
          //
          // if (nComp < 0)
          //   goto keep_docomb;
          // if (nComp > 0)
          //   goto use_trycomb;
          // 

          // Now if we've come this far, I do create the combined
          // object, because then I only have to compare two objects,
          // instead of two from each input to the other two in which
          // case I must first determine which of the two from each
          // side is the better so I compare in a stable way

          CreateCombinedVar<unsignedT>(pCombinedVar,bHasCombinedVar,
				       /*setVars,*/docomb);
          CreateCombinedVar<unsignedT>(pTryCombinedVar,bHasTryCombinedVar,
				       /*setVars,*/trycomb);
	  
          assert (bHasCombinedVar);
          assert (bHasTryCombinedVar);
	  
          // ok, so now we can compare them on type and (possibly)
          // content if any ended up a null, then it's a worse choice
          // than the other if both null, we can as well keep to old
          // one
	  
          if (!pTryCombinedVar)
            goto keep_docomb;
          if (!pCombinedVar)
            goto use_trycomb;
	  
          assert (pCombinedVar && pTryCombinedVar);

          nComp = pCombinedVar->CompareComb(pTryCombinedVar);
	  
          if (nComp > 0)
            goto use_trycomb;
          // nComp <= 0, < 0 would be goto keep_docomb, but equal can
          // be that as well, as the result is the same anyway (we
          // must just hope that now all possible FOUR parents are
          // removed as they should by IsInside below) Otherwise, we
          // won't be stable

          goto keep_docomb; // one line down, he he
	  
        keep_docomb: // so new one was not better, delete any temporary created for comparison
          delete pTryCombinedVar;
          continue;
        use_trycomb: // new try was better, delete old
          delete pCombinedVar;
          pCombinedVar = pTryCombinedVar;
          bHasCombinedVar = bHasTryCombinedVar;
          docomb = trycomb;
        }
      
      // ok, so combine using docomb
      
      CreateCombinedVar<unsignedT>(pCombinedVar,bHasCombinedVar,
				   /*setVars,*/docomb);
      assert (bHasCombinedVar);
      
      // TODO: try to simplify variable... (in CreateCombinedVar)
      
      if (!pCombinedVar)
        {
          // combination includes everything, so we may just as well
          // abort (NULL is ok as variable, representing 0 knowledge)

          // destroy all local structures

          return NULL; // all knowledge gone
        }

      // we must delete entries in the set coming from any of i and j
      // insert the new entry at position j (which is always < i) move
      // the last entry to position i

      // as the object *docomb will be deleted somewhere during this
      // operation, we can't use that

      typename MultiSetInputIntermediateVariableN<unsignedT>::Type::iterator i = docomb->m_i;
      typename MultiSetInputIntermediateVariableN<unsignedT>::Type::iterator j = docomb->m_j;
      
      typename MultiSetCostOfCombineVars<unsignedT>::Type::iterator itercmb;
      
      for(itercmb = multisetCombiCost.begin();
          itercmb != multisetCombiCost.end(); )
        {
          // ok, so does it have anything with i or j to do?
	  
          if (itercmb->m_i == i || itercmb->m_i == j ||
              itercmb->m_j == i || itercmb->m_j == j)
            {
              // item is to be deleted
              SET_MAP_ERASE_CONTINUE_ITER(multisetCombiCost,itercmb);
              //itercmb = multisetCombiCost.erase (itercmb);
            }
          else
            ++itercmb;
        }
      /*
	RemoveMultiSetCostOfCombineVarsReferences<unsignedT>(multisetCombiCost,j,i,n);
      */

      // remove whatever was at position i
      // remove whatever was at position j
      
      if (i->m_bIntermediate)
        delete i->m_var.m_pIntermediate;
      setVars.erase (i);
      
      if (j->m_bIntermediate)
        delete j->m_var.m_pIntermediate;
      setVars.erase (j);
      
      // the new element should not be able to fit into any other
      // element but if any other element now fits into the new one
      // (which is possible), kill it

      // (the new element would not fit into any other - this would be
      // true if isinside was perfect (it isn't, e.g. splits are
      // converted before compare), and therefore we must deal with
      // the possibilty of creating something that fits into something
      // else) but we'll test that last (we can't do both tests
      // interleaved, as that would not be stable, if many things
      // could be killed, which should go first?)

      // first see if any of the old elements fit into the new one

      bool bCanRemoveNewlyCombined = false;

      for(typename MultiSetInputIntermediateVariableN<unsignedT>::Type::iterator iter = setVars.begin(); 
          iter != setVars.end(); )
        {
          if (VariableCombineIsVariableInside(pCombinedVar,
					      iter->m_var.m_pInput))
            {
              // we can remove inside from list as it is in pCombinedVar
              
              for(itercmb = multisetCombiCost.begin();
                  itercmb != multisetCombiCost.end(); )
                {
                  // ok, so does it have anything with i or j to do?
		  
                  if (itercmb->m_i == iter || itercmb->m_j == iter)
		    {
                      // item is to be deleted
                      SET_MAP_ERASE_CONTINUE_ITER(multisetCombiCost,itercmb);
                      //itercmb = multisetCombiCost.erase (itercmb);
                    }
                  else
                    ++itercmb;
                }
	      
	      SET_MAP_ERASE_CONTINUE_ITER(setVars,iter);
	      //iter = setVars.erase (iter);
            }
          else
            {
              if (!bCanRemoveNewlyCombined && // no need to find two containers
                  VariableCombineIsVariableInside(iter->m_var.m_pInput,
						  pCombinedVar))
                bCanRemoveNewlyCombined = true;
              ++iter;
            }
        }
      
      if (bCanRemoveNewlyCombined)
        {
          // this was tested against an variable (*iter), that
          // survived, so we can be removed, no matter if any
          // variables has been killed because we're containing them,
          // because then iter will contain them too, even if IsInside
          // wouldn't say so (because it isn't perfect)
	  
          delete pCombinedVar;
        }
      else
        {
          // insert us into vector, and calculate combination
          // information with new element and old and insert into
          // multisetCombiCost
	  
          typename MultiSetInputIntermediateVariableN<unsignedT>::Type::iterator k = 
            setVars.insert (InputIntermediateVariableN<unsignedT>(true,pCombinedVar));
	  
          FOR_EACH_ITER(typename MultiSetInputIntermediateVariableN<unsignedT>::Type,setVars,l)
            {
              if (l != k)
                {
                  unsignedT nFillIn = VariableCombineCostOfCombine(k->m_var.m_pInput, 
                                                                   l->m_var.m_pInput);  
                  /* make sure VariableCombineCostOfCombine is symmetric */ 
                  assert (nFillIn == VariableCombineCostOfCombine(l->m_var.m_pInput, 
                                                                  k->m_var.m_pInput));  
                  multisetCombiCost.insert(CostOfCombineVars<unsignedT> (nFillIn,k,l));	
                }
            }
        }
    }
  
  // ok, so we're done the only one left should now be an constructed
  // one
  
  assert (multisetCombiCost.size() == 0); // nothing should be left, otherwise be careful, delete members' m_pIJ
  
  assert (setVars.size() == 1);
  assert (setVars.begin()->m_bIntermediate);
  
  VariableN<unsignedT>* pReturn = setVars.begin()->m_var.m_pIntermediate;
  
  return pReturn;
}

// force instantiation
#define VARIABLE_N_COMBINE(type) \
template VariableN<type>* VariableCombine(ListVariableN<type>::Type listVars);

VARIABLE_N_COMBINE(byte)
  VARIABLE_N_COMBINE(word)
  VARIABLE_N_COMBINE(dword)
  
/*
template<class unsignedT>
VariableN<unsignedT>* VariableCombineDoCombine(const VariableN<unsignedT>* arg1,
                                               const VariableN<unsignedT>* arg2)

template<class unsignedT>
bool VariableCombineIsVariableInside(const VariableN<unsignedT>* arg1,
                                     const VariableN<unsignedT>* arg2)

template<class unsignedT>
unsignedT VariableCombineCostOfCombine(const VariableN<unsignedT>* arg1,
                                       const VariableN<unsignedT>* arg2)
*/

