/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// VariableConvert.h: implementation of conversions of VariableN classes.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__VARIABLE_CONVERT_H__INCLUDED_)
#define __VARIABLE_CONVERT_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "VariableNConst.h" // dep_chk_unknown     
#include "VariableNConsts.h" // dep_chk_unknown     
#include "VariableNRange.h" // dep_chk_unknown     
#include "VariableNPattern.h" // dep_chk_unknown     
#include "VariableNSplit.h" // dep_chk_unknown     

//////////////////////////////////////////////////////////////////////
// define how 'good' different arguments are, i.e. how little fill in they have
// which should preferably be used when doing conversions
// numbers must be different so that sorting is stable and will get same type args after each other
// if possible

// argprio(VariableNConst<unsignedT>*)=0
// argprio(VariableNConsts<unsignedT>*)=1
// argprio(VariableNRange<unsignedT>*)=2
// argprio(VariableNPattern<unsignedT>*)=3
// argprio(VariableNSplit<unsignedT>*)=4
// argprio(ANY_PARAMETER)=1000

//////////////////////////////////////////////////////////////////////

template<class unsignedT>
VariableNRange<unsignedT>* ConvertConstsToRange(const VariableNConsts<unsignedT>* src)
{
  // find min and max values and generate range

  // find largest range between two values and use that as the unused range in the range
  
  assert (src->m_vectValues.size());

  typename VariableNConsts<unsignedT>::VectValues::const_iterator iter = 
    src->m_vectValues.begin();
  
  assert (*iter <= src->m_vectValues.back()); // must be ordered

  unsignedT nMin = *iter; 
  unsignedT nMax = src->m_vectValues.back();
  unsignedT nLargestRange = nMin - nMax - 1;
  
  unsignedT nLast = *iter;
  
  ++iter;
  
  for ( ; iter != src->m_vectValues.end(); ++iter)
    {
      assert (*iter > nLast); // must be ordered
      if ((*iter - nLast - 1) > nLargestRange)
        {
          nLargestRange = (*iter - nLast - 1);
          nMin = *iter;
          nMax = nLast;
        }   
      nLast = *iter;
    }
  if (nMin == nMax + 1) // entire range
    return NULL; // could also be new VariableN<unsignedT> ( which also means no knowledge)
  return new VariableNRange<unsignedT>(nMin,nMax);
}

template<class unsignedT>
VariableNPattern<unsignedT>* ConvertConstsToPattern(const VariableNConsts<unsignedT>* src)
{
  // add all values to the pattern

  assert (src->m_vectValues.size());

  /*
    #define FOR_EACH_CONST_ITER(conttype,container,iter) 
    
    FOR_EACH_CONST_ITER(VariableNConsts<unsignedT>::VectValues,lhs->m_vectValues,lhs_value) \
    for(conttype::const_iterator iter = (container).begin(); iter != (container).end(); ++iter)
  */
  
  typename VariableNConsts<unsignedT>::VectValues::const_iterator iter = 
    src->m_vectValues.begin();
  
  VariableNPattern<unsignedT>* result = new VariableNPattern<unsignedT>(*iter); 
  
  ++iter;
  
  for ( ; iter != src->m_vectValues.end(); ++iter)
    {
      result->InsertValue(*iter);
    }
  if (!result->m_known)
    {
      delete result; // nothing known anyway
      return NULL;
    }
  return result;
}

template<class unsignedT>
VariableNPattern<unsignedT>* ConvertRangeToPattern(const VariableNRange<unsignedT>* src)
{
  unsignedT nFiller = 1;
  
  unsignedT nKnown = ~((unsignedT) 0);
  
  while (nFiller && src->IsInside (src->m_min + nFiller))
    {
      nKnown ^= nFiller;
      nFiller <<= 1;
    }
  unsignedT nValue = src->m_min & nKnown;
  
  if (!nKnown)
    return NULL; // nothing known  
  return new VariableNPattern<unsignedT>(nKnown,nValue);
}

template<class unsignedT>
VariableNRange<unsignedT>* ConvertPatternToRange(const VariableNPattern<unsignedT>* src)
{
  if (!src->m_known)
    return NULL; // nothing known
  return new VariableNRange<unsignedT>(src->FirstValue(),src->LastValue());
}

template<class unsignedT>
VariableNPattern<unsignedT>* ConvertConstToPattern(const VariableNConst<unsignedT>* src)
{
  return new VariableNPattern<unsignedT>(src->m_value);
}

template<class unsignedT>
VariableNRange<unsignedT>* ConvertConstToRange(const VariableNConst<unsignedT>* src)
{
  return new VariableNRange<unsignedT>(src->m_value);
}
 
template<class unsignedT>
VariableNPattern<unsignedT>* ConvertSplitToPattern(const VariableNSplit<unsignedT>* src);

template<class unsignedT>
VariableNRange<unsignedT>* ConvertSplitToRange(const VariableNSplit<unsignedT>* src);

//////////////////////////////////////////////////////////////////////

#define GENERATE_VAR_INT_UNARY_CONVERSION_TO_PATTERN(name)            // rettype=VariableNPattern<unsignedT>* argtype=const VariableN<unsignedT>* mayconvert=0,0 
#define GENERATE_VAR_INT_UNARY_CONVERSION_TO_RANGE(name)              // rettype=VariableNRange<unsignedT>* argtype=const VariableN<unsignedT>* mayconvert=0,0

GENERATE_VAR_INT_UNARY_CONVERSION_TO_PATTERN(ConvertToPattern)        // defret=return NULL;
GENERATE_VAR_INT_UNARY_CONVERSION_TO_RANGE(ConvertToRange)            // defret=return NULL;

//////////////////////////////////////////////////////////////////////

#include "gen/VariableConvert_TypedCalls.h" // dep_chk_unknown     

//////////////////////////////////////////////////////////////////////

template<class unsignedT>
VariableNPattern<unsignedT>* ConvertSplitToPattern(const VariableNSplit<unsignedT>* src)
{
  // so we've got to get the pattern of the parts

  typename VariableNPattern<unsignedT>::half_var_type_pattern* pPatternLow = 
    ConvertToPattern(src->m_low);
  typename VariableNPattern<unsignedT>::half_var_type_pattern* pPatternHigh = 
    ConvertToPattern(src->m_high);
  
  // so now we make a full pattern of the two parts

  VariableNPattern<unsignedT>* result = 
    new VariableNPattern<unsignedT>(pPatternLow,pPatternHigh);

  delete pPatternLow;
  delete pPatternHigh;

  if (result->NothingKnown())
    {
      delete result;
      return NULL;
    }
  return result;
}

template<class unsignedT>
VariableNRange<unsignedT>* ConvertSplitToRange(const VariableNSplit<unsignedT>* src)
{
  // so we've got to get the pattern of the parts

  typename VariableNRange<unsignedT>::half_var_type_range* pRangeHigh =
    ConvertToRange(src->m_high);
  
  if (!pRangeHigh)
    {
      if (!src->m_low)
	return NULL; // we know nothing
      
      pRangeHigh = 
	new VariableNRange<VariableNRange<unsignedT>::half_type>(0,ALL_ONES(typename VariableNRange<unsignedT>::half_type));
    }
  
  typename VariableNRange<unsignedT>::half_type low_min = 0;
  typename VariableNRange<unsignedT>::half_type low_max = ALL_ONES(typename VariableNRange<unsignedT>::half_type);
  
  if (src->m_low)
    {
      low_min = src->m_low->FirstValue ();
      low_max = src->m_low->LastValue ();
    }
  
  // so now we make a full range of the two parts
  
  // the range is primarily determined by the high values.

  VariableNRange<unsignedT>* result = new VariableNRange<unsignedT>(low_min,low_max,pRangeHigh);
  
  delete pRangeHigh;
  
  if (!result->NothingKnown()) 
    {
      delete result;
      return NULL;
    }
  return result;
}

#endif // !defined(__VARIABLE_CONVERT_H__INCLUDED_)
