/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// VariableN.h: interface for the Variable class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__VARIABLE_N_H__INCLUDED_)
#define __VARIABLE_N_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "GlobalDefs.h"
#include "Variable.h" 
/*
#define VAR_N_INCLUDE_FILE "VariableNBase_Inc.h"
// // #include "VariableN_InfoAndInclude.h" // dep_chk_unknown     

#define VAR_N_INCLUDE_FILE "VariableN_Inc.h"
// // #include "VariableN_InfoAndInclude.h" // dep_chk_unknown     
*/

template<class unsignedT>
struct int_var_type_info { };

template<>
struct int_var_type_info<byte>
{
  typedef sbyte signedT;
  
  //
  typedef byte  halfT;		// bogus
  typedef byte   quarterT;	// bogus
  //
};

template<>
struct int_var_type_info<word>
{
  typedef sword signedT;
  typedef byte  halfT;		// type of high or low part of this object
  
  //
  typedef byte   quarterT;	// bogus
  //
};

template<>
struct int_var_type_info<dword>
{
  typedef sdword signedT;
  typedef word   halfT;		// type of high or low part of this object
  typedef byte   quarterT;	// type of high or low part of this object
};

#define BITS_IN_TYPE(type)		(sizeof(type)*8)
#define TYPE_RANGE(type)		(1 << BITS_IN_TYPE(type))
#define TWICE_TYPE_RANGE(type)	        (1 << (2*BITS_IN_TYPE(type)))
#define MULT_TYPE_RANGE(type,m)		(1 << (m*BITS_IN_TYPE(type)))

#define ALL_ONES(type)                  (((type) -1)) // 0xffff
#define MOST_SIGNIFICANT_ONE(type)      ((type) 1) << (BITS_IN_TYPE(type)-1) // 0x8000

#define TO_HALF_TYPE(n,p) 		((half_type) ((n) >> (p * BITS_IN_TYPE(half_type))))
#define TO_QUARTER_TYPE(n,p) 		((quarter_type) ((n) >> (p * BITS_IN_TYPE(quarter_type))))

#define LOW_TO_HALF_TYPE(n)		((half_type) (n))
#define HIGH_TO_HALF_TYPE(n)	        ((half_type) ((n) >> BITS_IN_TYPE(half_type)))

#define HALF_TYPE_TO_HIGH(n)	(((value_utype) (n)) << BITS_IN_TYPE(half_type))
#define HALF_TYPE_TO_TYPE(l,h)  (((value_utype) (l)) | (((value_utype) (h)) << BITS_IN_TYPE(half_type)))

#define FIRST_TO_QUARTER_TYPE(n)	((quarter_type) (n))
#define SECOND_TO_QUARTER_TYPE(n)	((quarter_type) ((n) >> BITS_IN_TYPE(quarter_type)))
#define THIRD_TO_QUARTER_TYPE(n)	((quarter_type) ((n) >> 2*BITS_IN_TYPE(quarter_type)))
#define FOURTH_TO_QUARTER_TYPE(n)	((quarter_type) ((n) >> 3*BITS_IN_TYPE(quarter_type)))

// base class of integer variable of certain size

#define VARIABLE_N_BASE_TYPEDEF_VALUE_TYPES \
typedef unsignedT value_utype; \
typedef typename int_var_type_info<unsignedT>::signedT  value_stype; \
typedef typename int_var_type_info<unsignedT>::halfT    half_type; \
typedef typename int_var_type_info<unsignedT>::quarterT quarter_type; \

template<class unsignedT>
class VariableNBase : public Variable
{
 public:
  VARIABLE_N_BASE_TYPEDEF_VALUE_TYPES
    
 protected:	
  VariableNBase() { } // use VariableNUnknown instead
 public:
  // VariableNBase(const VariableNBase& src) : Variable<unsignedT>(src) { }
  VariableNBase(const VariableNBase& src,bool bDep) : Variable(src,bDep) { }
  
  typedef unsignedT value_utype;
  typedef typename int_var_type_info<unsignedT>::signedT value_stype;
  
  typedef typename int_var_type_info<unsignedT>::halfT half_type;
  typedef typename int_var_type_info<unsignedT>::quarterT quarter_type;
  
 public:
  virtual dword GetSize() { return sizeof (unsignedT); };
  
  
};

template<class unsignedT>
struct ValueRange
{
  unsignedT m_min;
  unsignedT m_max;
};

#define VAR_N_COMP_TYPE_CONST    1
#define VAR_N_COMP_TYPE_CONSTS   2
#define VAR_N_COMP_TYPE_RANGE    3
#define VAR_N_COMP_TYPE_PATTERN  4
#define VAR_N_COMP_TYPE_SPLIT    6
#define VAR_N_COMP_TYPE_UNKNOWN  7

//template<class unsignedT>

#define COMPARE_COMB_BY_TYPE \
{ \
  int cmp = VariableNCompareCombType()-(rhs)->VariableNCompareCombType(); \
  if (cmp) \
    return cmp; \
}

#define VARIABLE_N_TYPEDEF_VAR_TYPES(classname,var_type) \
typedef classname<value_utype> var_type; \
typedef classname<half_type> half_##var_type; \
typedef classname<quarter_type> quarter_##var_type;

template<class unsignedT>
class VariableN : public VariableNBase<unsignedT>
{
 public:
  VARIABLE_N_BASE_TYPEDEF_VALUE_TYPES
  VARIABLE_N_TYPEDEF_VAR_TYPES(VariableN,var_type);
  //typedef VariableN<value_utype> var_type;
  //typedef VariableN<half_type> half_var_type;
  //typedef VariableN<quarter_type> quarter_var_type;
  // typedef typename ValueRange<unsignedT> value_range;
  
  // typedef typename VariableRange<unsignedT> var_range_type;

  //	static VariableTreatment& m_treatment;
  
 protected:	
  VariableN() { } // use VariableNUnknown instead
 public:
  // VariableN(const VariableN& src) : VariableNBase<unsignedT>(src) { }
  VariableN(const VariableN& src,bool bDep) 
    : VariableNBase<unsignedT>(src,bDep) { }

 public:	
  virtual ~VariableN() { }
  
 public:
  virtual var_type* CopyVarContent() const = 0; // copy content (i.e., possible values, but not dependencies stored in Variable)
  
 public:
  virtual int VariableNCompareCombType () const = 0;
  virtual int CompareComb(const var_type* rhs) const = 0;
  
 public:
  virtual value_utype FirstValue() const = 0;
  virtual value_utype LastValue() const = 0;
  virtual bool NextValue(value_utype& /* value */) const { return false; } // default: one value only
  
  // NextValueRange must not return a min value that is <= previous range's max+1 (i.e., they may
  // not be adjacent) (This is relied upon by CompareByValueRanges)
  
  virtual void FirstValueRange(ValueRange<unsignedT>& /* rng */) const = 0;
  virtual bool NextValueRange(ValueRange<unsignedT>& /* rng */) const { return false; } // default: one range only
  virtual void FirstValue(value_utype& value) const { value = FirstValue(); }
  virtual void LastValue(value_utype& value) const { value = LastValue(); }
  
  int CompareByValueRanges(const var_type* rhs) const
    {
      ValueRange<unsignedT> rngThis;
      ValueRange<unsignedT> rngRhs;
      
      FirstValueRange (rngThis);
      rhs->FirstValueRange (rngRhs);
      
      for ( ; ; )
        {
          // when we end up here, everything before compared equal,  **EQUAL**--gap--*THIS*****
          // and there was a gap up to these two                     **EQUAL**--gap---*RHS***
          
          if (rngThis.m_min < rngRhs.m_min) return -1;
          if (rngThis.m_min > rngRhs.m_min) return 1;
          if (rngThis.m_max < rngRhs.m_max) return -1;
          if (rngThis.m_max > rngRhs.m_max) return 1;
          
          // so let's get two new ranges
	  
          bool bNextThis = NextValueRange(rngThis);
          bool bNextRhs = rhs->NextValueRange(rngRhs);
	  
          if (!bNextThis)
            return bNextRhs ? -1 : 0;
          if (!bNextRhs)
            return 1;
          
          assert (bNextThis && bNextRhs);
        }
    }
  
 public:
  virtual value_utype NumbersIncluded () const { return 0; } // 0 means all (overflow, as we'd like a 1 in the more than most significant bit)
  virtual bool NothingKnown () const { return true; } // returns true if containing all, false if constraining
  
 public:
  virtual var_type* SimplifyVariable() { return NULL; } // cannot simplify
  
 public:
  virtual half_var_type*    GetHalfPart(byte p)    { assert (sizeof(value_utype) >= 2); assert (p < 2); return NULL; }
  virtual quarter_var_type* GetQuarterPart(byte p) { assert (sizeof(value_utype) >= 4); assert (p < 4); return NULL; }
  
  /*
    virtual half_var_type* GetLowerHalf() { assert (sizeof(value_utype) >= 2); return NULL; }
    virtual half_var_type* GetUpperHalf() { assert (sizeof(value_utype) >= 2); return NULL; }
    
    virtual quarter_var_type* GetFirstQuarter() { assert (sizeof(value_utype) >= 4); return NULL; }  // least signifcant, bits 0-7
  virtual quarter_var_type* GetSecondQuarter() { assert (sizeof(value_utype) >= 4); return NULL; } // least signifcant, bits 8-15
  virtual quarter_var_type* GetThirdQuarter() { assert (sizeof(value_utype) >= 4); return NULL; }  // least signifcant, bits 0-7
  virtual quarter_var_type* GetFourthQuarter() { assert (sizeof(value_utype) >= 4); return NULL; } // least signifcant, bits 8-15
  */
 public:
  virtual string ValueString() const { return "VARIABLE_N(error?)"; }
  
};

// VariableNUnknown
// can be used when value is unknown (thus does not need to contain any data)
// it will however (via base classes) keep track of dependency chains
// but you should use a null pointer instead to use less memory
template<class unsignedT>
class VariableNUnknown : public VariableN<unsignedT>
{
 public:
  VARIABLE_N_BASE_TYPEDEF_VALUE_TYPES
    VARIABLE_N_TYPEDEF_VAR_TYPES(VariableN,var_type);
  
 public:	
  VariableNUnknown() { }
  //  VariableNUnknown(const VariableNUnknown& src) : VariableN<unsignedT>(src) {  }
  VariableNUnknown(const VariableNUnknown& src,bool bDep) : VariableN<unsignedT>(src,bDep) {  }
 public:	
  virtual ~VariableNUnknown() { }
  
 public:
  virtual var_type* CopyVarContent() const { return new VariableNUnknown(*this,false); }
  
 public:
  virtual int VariableNCompareCombType () const { return VAR_N_COMP_TYPE_UNKNOWN; }
  virtual int CompareComb(const var_type* /* rhs */) const { return 0; }
  
 public:
  virtual value_utype FirstValue() const { return 0; }
  virtual value_utype LastValue() const { return ALL_ONES(value_utype); }
  virtual bool NextValue(value_utype& value) const 
    { 
      value++; 
      assert (false); // when this describes a 32 bit number we'll iterate through 4e9 numbers, thios should not happen
      // so rewrite the code that caused this (use First/Next ValueRange instead)
      return !value;
    }
  virtual void FirstValueRange(ValueRange<unsignedT>& rng) const { rng.m_min = 0; rng.m_max = ALL_ONES(value_utype); }
  
 public:
  virtual value_utype NumbersIncluded () { return 0; } // 0 means all (overflow, as we'd like a 1 in the more than most significant bit)
  
 public:
  /*
  virtual half_var_type*    GetHalfPart(byte n)    { assert (sizeof(value_utype) >= 2); assert (n < 2); return NULL; }
  virtual quarter_var_type* GetQuarterPart(byte n) { assert (sizeof(value_utype) >= 4); assert (n < 4); return NULL; }
  */
  /*
  virtual half_var_type* GetLowerHalf() { assert (sizeof(value_utype) >= 2); return NULL; }
  virtual half_var_type* GetUpperHalf() { assert (sizeof(value_utype) >= 2); return NULL; }
  
  virtual quarter_var_type* GetFirstQuarter() { assert (sizeof(value_utype) >= 4); return NULL; }  // least signifcant, bits 0-7
  virtual quarter_var_type* GetSecondQuarter() { assert (sizeof(value_utype) >= 4); return NULL; } // least signifcant, bits 8-15
  virtual quarter_var_type* GetThirdQuarter() { assert (sizeof(value_utype) >= 4); return NULL; }  // least signifcant, bits 0-7
  virtual quarter_var_type* GetFourthQuarter() { assert (sizeof(value_utype) >= 4); return NULL; } // least signifcant, bits 8-15
  */

 public:
  virtual string ValueString() const { return "unknown"; }
};

// VariableNUnreachable
// used when determining result or possible sources for an operation with
// some result (usually flags set) of the outcome known
// eg comparing two unequal numbers and knowing that they comapred as equal
// means that the code after the je was unreachable, but to have some
// variable two pass along anyway, without using VariableNUnknown (which include
// all values), we have this one (which include NO value)

template<class unsignedT>
class VariableNUnreachable : public VariableN<unsignedT>
{
 public:
  VARIABLE_N_BASE_TYPEDEF_VALUE_TYPES
  VARIABLE_N_TYPEDEF_VAR_TYPES(VariableN,var_type);

 public:	
  VariableNUnreachable() { }
 public:	
  virtual ~VariableNUnreachable() { }
  
 public:
  virtual value_utype NumbersIncluded () { return 0; } // 0 means all (overflow, as we'd like a 1 in the more than most significant bit), but this is better than 1 or anything elese
  
 public:
  virtual half_var_type*    GetHalfPart(byte p)    
    { 
      assert (sizeof(value_utype) >= 2); 
      assert (p < 2); 
      return new VariableNUnreachable<half_type>; 
    }
  virtual quarter_var_type* GetQuarterPart(byte p) 
    { 
      assert (sizeof(value_utype) >= 4); 
      assert (p < 4); 
      return new VariableNUnreachable<quarter_type>; 
    }
  /*
  virtual half_var_type* GetLowerHalf() { assert (sizeof(value_utype) >= 2); return new VariableNUnreachable<half_type>; }
  virtual half_var_type* GetUpperHalf() { assert (sizeof(value_utype) >= 2); return new VariableNUnreachable<half_type>; }
  
  virtual quarter_var_type* GetFirstQuarter() { assert (sizeof(value_utype) >= 4); return new VariableNUnreachable<quarter_type>; }  // least signifcant, bits 0-7
  virtual quarter_var_type* GetSecondQuarter() { assert (sizeof(value_utype) >= 4); return new VariableNUnreachable<quarter_type>; } // least signifcant, bits 8-15
  virtual quarter_var_type* GetThirdQuarter() { assert (sizeof(value_utype) >= 4); return new VariableNUnreachable<quarter_type>; }  // least signifcant, bits 0-7
  virtual quarter_var_type* GetFourthQuarter() { assert (sizeof(value_utype) >= 4); return new VariableNUnreachable<quarter_type>; } // least signifcant, bits 8-15
  */

 public:
  virtual string ValueString() const { return "UNREACHABLE(error?)"; } // or actually it has no possible value (he, no electrons at all, perhaps)

};

//////////////////////////////////////////////////////////////////////

typedef VariableN<byte>  Variable8;
typedef VariableN<word>  Variable16;
typedef VariableN<dword> Variable32;

#define DECL_INT_VARS(name) \
typedef VariableN##name<byte>  Variable8##name; \
typedef VariableN##name<word>  Variable16##name; \
typedef VariableN##name<dword> Variable32##name;

DECL_INT_VARS(Unknown)

#endif // !defined(__VARIABLE_N_H__INCLUDED_)
