/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// VariableNConst.h: interface for the Variable class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__VARIABLE_N_CONST_H__INCLUDED_)
#define __VARIABLE_N_CONST_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "VariableN.h" // dep_chk_unknown     
#include "Misc.h"

template<class unsignedT>
class VariableNConst : public VariableN<unsignedT>
{
 public:
  VARIABLE_N_BASE_TYPEDEF_VALUE_TYPES
  VARIABLE_N_TYPEDEF_VAR_TYPES(VariableN     ,var_type);
  VARIABLE_N_TYPEDEF_VAR_TYPES(VariableNConst,var_type_const);


 public:
  //	typedef VariableN<unsignedT> variableT;
  //	typedef VariableN<unsignedT> parentT;
  
  // typedef VariableNConst<half_type> half_var_type_const;
  // typedef VariableNConst<quarter_type> quarter_var_type_const;
  
 public:
  VariableNConst(value_utype value) 
    { m_value = value; }
  VariableNConst(const VariableNConst& src) 
    : VariableN<unsignedT>(src) 
    { m_value = src.m_value; }
  VariableNConst(const VariableNConst& src,bool bDep) 
    : VariableN<unsignedT>(src,bDep) 
    { m_value = src.m_value; }
  
 public:
  virtual var_type* CopyVarContent() const 
    { return new VariableNConst(*this,false); }
  
  virtual int VariableNCompareCombType () const 
    { return VAR_N_COMP_TYPE_CONST; }
  virtual int CompareComb(const var_type* rhs) const
    {
      COMPARE_COMB_BY_TYPE;
      const VariableNConst* pRhs = dynamic_cast<const VariableNConst*>(rhs);
      assert (rhs && pRhs);
      if (m_value < pRhs->m_value)
        return -1;
      return m_value > pRhs->m_value;
    }

 public:
  virtual value_utype FirstValue() const { return m_value; }
  virtual value_utype LastValue() const 
    { return m_value; }
  virtual void FirstValueRange(ValueRange<unsignedT>& rng) const
    { rng.m_min = rng.m_max = m_value; }
  
 public:
  value_utype m_value;
  
 public:
  virtual value_utype NumbersIncluded () { return 1; }
  virtual bool NothingKnown () const { return false; }
  
 public:
  virtual half_var_type*    GetHalfPart(byte p)    
    { 
      assert (sizeof(value_utype) >= 2); 
      assert (p < 2); 
      return new half_var_type_const(TO_HALF_TYPE(m_value,p)); 
    }
  virtual quarter_var_type* GetQuarterPart(byte p) 
    { 
      assert (sizeof(value_utype) >= 4); 
      assert (p < 4); 
      return new quarter_var_type_const(TO_QUARTER_TYPE(m_value,p)); 
    }
  /*
  virtual half_var_type* GetLowerHalf() { return new half_var_type_const(LOW_TO_HALF_TYPE(m_value)); }
  virtual half_var_type* GetUpperHalf() { return new half_var_type_const(HIGH_TO_HALF_TYPE(m_value)); }
  
  virtual quarter_var_type* GetFirstQuarter() { return new quarter_var_type_const(FIRST_TO_QUARTER_TYPE(m_value)); }
  virtual quarter_var_type* GetSecondQuarter() { return new quarter_var_type_const(SECOND_TO_QUARTER_TYPE(m_value)); }
  virtual quarter_var_type* GetThirdQuarter() { return new quarter_var_type_const(THIRD_TO_QUARTER_TYPE(m_value)); }
  virtual quarter_var_type* GetFourthQuarter() { return new quarter_var_type_const(FOURTH_TO_QUARTER_TYPE(m_value)); }
  */
 public:
  virtual string ValueString() const { return strsprintf("0x%x",m_value); }

/*
public:
	virtual variableT* doAdd    (const variableT* rhs) const
	{
		IF_VAR_TYPE(Const){ return new VariableNConst(m_value + pConstRhs->m_value); }
		else return parentT::doAdd(rhs); // we do not know how to handle
	}
	virtual variableT* doOr     (const variableT* rhs) const
	{
		IF_VAR_TYPE(Const){ return new VariableNConst(m_value | pConstRhs->m_value); }
		else return parentT::doOr (rhs); // we do not know how to handle
	}
*/
};

DECL_INT_VARS(Const)

Variable* VariableNZero(int nBytes);
Variable* VariableNNonZero(int nBytes);
Variable* VariableN_Unknown(int nBytes);

#endif // !defined(__VARIABLE_N_CONST_H__INCLUDED_)
