/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// VariableNConsts.h: interface for the Variable class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__VARIABLE_N_CONSTS_H__INCLUDED_)
#define __VARIABLE_N_CONSTS_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "VariableN.h" 
#include <algorithm>

#include "Misc.h"

#ifdef _MSC_VER
// because the compiler wont give me 'min', so...
// even though it's suppose to come with <algorithm>

template <class _Tp>
inline const _Tp& min(const _Tp& __a, const _Tp& __b) {
  if (__b < __a) return __b; return __a;
}
#endif

#define VAR_CONSTS_MAX_NO_OF_VALUES 1024  // if more than this, please use another type (e.g. range or pattern)

template<class unsignedT>
class VariableNConsts : public VariableN<unsignedT>
{
 public:
  VARIABLE_N_BASE_TYPEDEF_VALUE_TYPES
  VARIABLE_N_TYPEDEF_VAR_TYPES(VariableN      ,var_type);
  VARIABLE_N_TYPEDEF_VAR_TYPES(VariableNConsts,var_type_consts);

 public:
  //	typedef VariableN<unsignedT> variableT;
  //	typedef VariableN<unsignedT> parentT;
  
  typedef vector<unsignedT> VectValues;
  
  typedef VariableNConsts<half_type> half_var_type_consts;
  typedef VariableNConsts<quarter_type> quarter_var_type_consts;
  
 public:
  VariableNConsts() {  }
  VariableNConsts(value_utype value) 
    { m_vectValues.push_back(value); }
  VariableNConsts(value_utype value1,value_utype value2) 
    { m_vectValues.push_back(value1); m_vectValues.push_back(value2); }
  VariableNConsts(const VariableNConsts& src)
    : VariableN<unsignedT>(src) , m_vectValues(src.m_vectValues)
    { }
   VariableNConsts(const VariableNConsts& src,bool bDep)
    : VariableN<unsignedT>(src,bDep) , m_vectValues(src.m_vectValues)
    { }
 
 public:
  virtual var_type* CopyVarContent() const 
    { return new VariableNConsts(*this,false); }
  
  virtual int VariableNCompareCombType () const 
    { return VAR_N_COMP_TYPE_CONSTS; }
  virtual int CompareComb(const var_type* rhs) const
    {
      COMPARE_COMB_BY_TYPE;
      const VariableNConsts* pRhs = dynamic_cast<const VariableNConsts*>(rhs);
      assert (rhs && pRhs);
      
      int imax = min(m_vectValues.size(),pRhs->m_vectValues.size());
      
      for (int i = 0; i < imax; i++)
        {
          if (m_vectValues[i] < pRhs->m_vectValues[i]) return -1;
          if (m_vectValues[i] > pRhs->m_vectValues[i]) return 1;
        }
      
      assert (m_vectValues.size() < INT_MAX); // or you've got a really lot of memory, but this class shouldn't be used in that case anyway, see note at definition of VAR_CONSTS_MAX_NO_OF_VALUES
      assert (pRhs->m_vectValues.size() < INT_MAX);

      return m_vectValues.size() - pRhs->m_vectValues.size();
    }
  
 public:
  virtual value_utype FirstValue() const 
    { assert (m_vectValues.size()); return m_vectValues.front(); }
  virtual value_utype LastValue() const 
    { return m_vectValues.back(); }
  virtual bool NextValue(value_utype& value) const 
    {
      // find out where we are in vector and get next NOTE: you really
      // do not want to use this function, as each call includes an
      // binary search as we do not keep any information about where
      // in the vector we get our last value
      
      typename VectValues::const_iterator value_at = 
	lower_bound(m_vectValues.begin(),m_vectValues.end(),value);
      
      assert (value_at != m_vectValues.end());
      assert (*value_at == value);
      
      if (++value_at == m_vectValues.end())
        return false;
      
      value = *value_at;
      return true;
    }
  
 protected:
  void FindValueRange (ValueRange<unsignedT>& rng,
                       typename VectValues::const_iterator value_at) const
    {
      value_utype end_val = rng.m_min = *value_at;
      
      for (++value_at ; 
	   value_at != m_vectValues.end() && *value_at == end_val+1; 
	   ++value_at)
        end_val = *value_at;
      
      rng.m_max = end_val;
    }

 public:
  virtual void FirstValueRange(ValueRange<unsignedT>& rng) const
    {
      assert (m_vectValues.size()); 
      
      typename VectValues::const_iterator value_at = m_vectValues.begin();
      
      FindValueRange(rng,value_at);
    }
  virtual bool NextValueRange(ValueRange<unsignedT>& rng) const
    {
      typename VectValues::const_iterator value_at = 
	lower_bound(m_vectValues.begin(),m_vectValues.end(),rng.m_max);
      
      assert (value_at != m_vectValues.end());
      assert (*value_at == rng.m_max);
      
      if (++value_at == m_vectValues.end())
        return false;
      
      FindValueRange (rng,value_at);
      return true;
    }
  
 public:
  VectValues m_vectValues;
  
 public:
  int size () const { return m_vectValues.size(); }
  
 public:
  virtual value_utype NumbersIncluded () const 
    { return m_vectValues.size(); }
  virtual bool NothingKnown () const 
    { assert (m_vectValues.size()); 
      return ((unsignedT) m_vectValues.size()) == 0; } 
  
  virtual VariableN<unsignedT>* SimplifyVariable();
  
 public:
  virtual half_var_type* GetHalfPart(byte p)
    {
      assert (p < 2);
      half_var_type_consts* pNew = new half_var_type_consts();
      for (uint i = 0; i < m_vectValues.size(); i++)
        pNew->InsertValue(TO_HALF_TYPE(m_vectValues[i],p));
      return pNew;
    }
  virtual quarter_var_type* GetQuarterPart(byte p)
    {
      assert (p < 4);
      quarter_var_type_consts* pNew = new quarter_var_type_consts();
      for (uint i = 0; i < m_vectValues.size(); i++)
        pNew->InsertValue(TO_QUARTER_TYPE(m_vectValues[i],p));
      return pNew;
    }

  /*
  virtual half_var_type* GetLowerHalf()
    {
      half_var_type_consts* pNew = new half_var_type_consts();
      for (uint i = 0; i < m_vectValues.size(); i++)
        pNew->InsertValue(LOW_TO_HALF_TYPE(m_vectValues[i]));
      return pNew;
    }
  virtual half_var_type* GetUpperHalf()
    {
      half_var_type_consts* pNew = new half_var_type_consts();
      for (uint i = 0; i < m_vectValues.size(); i++)
                        pNew->InsertValue(HIGH_TO_HALF_TYPE(m_vectValues[i]));
      return pNew;
    }
  
  virtual quarter_var_type* GetFirstQuarter ()
    {
      quarter_var_type_consts* pNew = new quarter_var_type_consts();
      for (uint i = 0; i < m_vectValues.size(); i++)
        pNew->InsertValue(FIRST_TO_QUARTER_TYPE(m_vectValues[i]));
      return pNew;
    }
  virtual quarter_var_type* GetSecondQuarter()
    {
      quarter_var_type_consts* pNew = new quarter_var_type_consts();
      for (uint i = 0; i < m_vectValues.size(); i++)
        pNew->InsertValue(SECOND_TO_QUARTER_TYPE(m_vectValues[i]));
      return pNew;
    }
  */
  
 public:
#ifndef NDEBUG
  // check if we're sorted and only have one of every value
  bool IsSortedConsistencyCheck() const
    {
      for (uint i = 1; i < m_vectValues.size(); i++)
        if (m_vectValues[i-1] >= m_vectValues[i])
          return false;
      return true;
    }
#endif// NDEBUG
  
 public:
  bool IsInside (value_utype value) const
    { 
      assert (IsSortedConsistencyCheck());
      return binary_search(m_vectValues.begin(),m_vectValues.end(),value);
    }
  
 public:
  void InsertValue(value_utype value)
    {
      typename VectValues::iterator insert_at = 
	lower_bound(m_vectValues.begin(),m_vectValues.end(),value);
      if (insert_at != m_vectValues.end() && *insert_at == value)
        return; // already known		
      m_vectValues.insert(insert_at,value);
    }
  

/*
public:
	virtual variableT* doAdd    (const variableT* rhs) const
	{
		IF_VAR_TYPE(Const){ return new VariableNConst(m_value + pConstRhs->m_value); }
		else return parentT::doAdd(rhs); // we do not know how to handle
	}
	virtual variableT* doOr     (const variableT* rhs) const
	{
		IF_VAR_TYPE(Const){ return new VariableNConst(m_value | pConstRhs->m_value); }
		else return parentT::doOr (rhs); // we do not know how to handle
	}
*/

 public:
  virtual string ValueString() const 
    { 
      assert (m_vectValues.size()); 
      
      string str = strsprintf("{0x%x",m_vectValues[0]);      
      for (uint i = 1; i < m_vectValues.size(); i++)
	str += strsprintf(",0x%x",m_vectValues[i]);
      
      return str + '}';
    }

};

DECL_INT_VARS(Consts)

#endif // !defined(__VARIABLE_N_CONSTS_H__INCLUDED_)
