/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// VariableNPattern.h: interface for the Variable class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__VARIABLE_N_PATTERN_H__INCLUDED_)
#define __VARIABLE_N_PATTERN_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Misc.h"

// // #include "VariableN.h" // dep_chk_unknown     

template<class unsignedT>
class VariableNPattern : public VariableN<unsignedT>
{
 public:
  VARIABLE_N_BASE_TYPEDEF_VALUE_TYPES
  VARIABLE_N_TYPEDEF_VAR_TYPES(VariableN       ,var_type);
  VARIABLE_N_TYPEDEF_VAR_TYPES(VariableNPattern,var_type_pattern);
  
 public:
  VariableNPattern(value_utype known,value_utype value) 
    { m_known = known; m_value = value; }
  VariableNPattern(value_utype value) 
    { m_known = ALL_ONES(value_utype); m_value = value; }
  VariableNPattern(const VariableNPattern& src) 
    : VariableN<unsignedT>(src) 
    { m_known = src.m_known; m_value = src.m_value; } 
  VariableNPattern(const VariableNPattern& src,bool bDep) 
    : VariableN<unsignedT>(src,bDep) 
    { m_known = src.m_known; m_value = src.m_value; } 
  
  VariableNPattern(const half_var_type_pattern* pLow,
		   const half_var_type_pattern* pHigh)
    {
      m_known = 0;
      m_value = 0;
      if (pLow)
        {
          m_known = pLow->m_known;
          m_value = pLow->m_value;
        }
      if (pHigh)
        {
          m_known |= HALF_TYPE_TO_HIGH(pHigh->m_known);
          m_value |= HALF_TYPE_TO_HIGH(pHigh->m_value);
        }
   }
  
 public:
  virtual var_type* CopyVarContent() const 
    { return new VariableNPattern(*this,false); }
  
  virtual int VariableNCompareCombType () const 
    { return VAR_N_COMP_TYPE_PATTERN; }
  virtual int CompareComb(const var_type* rhs) const
    {
      COMPARE_COMB_BY_TYPE;
      const VariableNPattern* pRhs = 
	dynamic_cast<const VariableNPattern*>(rhs);
      assert (rhs && pRhs);
      
      // first compare by minimum value, which is equal to m_value
      
      if (m_value < pRhs->m_value) return -1;
      if (m_value > pRhs->m_value) return 1;
      
      // that failed
      
      // so the first one missing a value is the one that is the
      // smallest as the m_values are the same, all known must be
      // same, but and all unknown would give same values, but if an
      // value is known to be zero, and for the other one it's
      // unknown, the other one will also generate a one there, being
      // bigger. The one generating a one, is the one having a zero in
      // the known position, so m_known has to be compared 'backwards'
      
      if (m_known < pRhs->m_known) return 1;
      if (m_known > pRhs->m_known) return -1;

      return 0;
    }
  
 public:
  value_utype m_known; // bit pattern of known bits
  value_utype m_value; // the values of the known bits
  
  // known could for ex be 0xf0 and value be 0x00, in which case we know
  // that the variable is 0x0? (the four high bits known to be zero)
  
 public:
  bool IsInside (value_utype value) const
    { 
      return !((value ^ m_value) & m_known);
    }
  
  virtual VariableN<unsignedT>* SimplifyVariable();
  
  static value_utype NumbersIncluded (value_utype known) 
    {
      value_utype nIncluded = 1;
      value_utype nUnknownBits = ~known;
      while (nUnknownBits)
        {
          if (nUnknownBits & 0x01)
            nIncluded <<= 1;
          nUnknownBits >>= 1;
        }
      return nIncluded; 
    }
  
 public:
  value_utype FirstValue() const
    {
      assert (!(~m_known & m_value)); // m_value should be zero for bits not known
      return m_value; // all known to be one (it can't be smaller)
    }
  virtual value_utype LastValue() const { return m_value | // all known to be one
                                            (~m_known); } // and all not known (might be one) (so it can't get bigger)
  bool NextValue(value_utype& value) const
    {
      // get next value represented by this pattern normally, it would
      // be one added to the value, except, when that produces a value
      // not included
      
      if (!(++value))
        return false; // we wrapped to zero, so we're finished
      
      if ((value & m_known) == m_value)
        return true;
      
      // we produced something thats not in the pattern first of all,
      // any bit that got set to zero, but that should be one, get
      // fixed
      
      value |= m_value; // this can only increase the value
      
      // the faulty bit (if any remaining know) is the one being one
      // without being said to be so by m_known and m_value

      // when we add one to any number, only one bit can ber flipped
      // from zero to one, but many can be flipped from one to zero
      
      value_utype wrong_bit = value & m_known & ~m_value; // bit being one, that is known, and that should be zero
      
      if (!wrong_bit)
        return true; // the only error was that some bit(s) were set to zero
      
      // ok, so set the bad bit back to zero
      
      value ^= wrong_bit;
      
      // and try to shift the bad bit to the left to see where it
      // would fit, the position where we try to insert it must of
      // course be zero
      
      wrong_bit <<= 1;
      
      for ( ; wrong_bit; wrong_bit <<= 1)
        {
          if (value & wrong_bit)
            continue; // no insert at position not being zero
          
          if (wrong_bit & m_known & ~m_value)
            continue; // that bit cannot be one

          // aha, so we can insert it

          value |= wrong_bit;
          assert ((value & m_known) == m_value);
          return true;
        }

      // oups, out of positions, so
      
      return false;
    }
  
  virtual void FirstValueRange(ValueRange<unsignedT>& /* rng */) const 
    { assert (false); } // not implemented
  virtual bool NextValueRange(ValueRange<unsignedT>& /* rng */) const  
    { assert (false); return false; } // not implemented
  
 public:
  virtual value_utype NumbersIncluded () const
    {
      return NumbersIncluded(m_known);
    }
  virtual bool NothingKnown () const { return !m_known; }
  
 public:
  void InsertValue(value_utype value)
    {
      // all bits that are the same are still known
      // other ones will be unknown
      
      m_known &= ~(m_value ^ value);
      m_value &= m_known; // normalize
    }
  
  value_utype CostOfInsertValue(value_utype value) const
    {
      value_utype nNumOverlap = 0; // # of numbers which both this and src include
      
      if (!((m_known) & (m_value ^ value))) // no disagree where bits known ?
	{
	  nNumOverlap = NumbersIncluded(m_known);
	}
      
      value_utype merged_known = m_known & ~(m_value ^ value);
      
      return NumbersIncluded(merged_known)-(NumbersIncluded()+1-nNumOverlap);
    }
  
  void AddPattern(const VariableNPattern* src)
    {
      m_known &= src->m_known & ~(m_value ^ src->m_value);
      m_value &= m_known; // normalize
    }
  
  value_utype CostOfPatternMerge(const VariableNPattern* src) const
    {
      value_utype nNumOverlap = 0; // # of numbers which both this and src include
      
      if (!((m_known & src->m_known) & (m_value ^ src->m_value))) // no disagree where bits known ?
	{
	  nNumOverlap = NumbersIncluded(m_known & src->m_known);
	}
      
      value_utype merged_known = m_known & src->m_known & ~(m_value ^ src->m_value);
      
      return NumbersIncluded(merged_known)-(NumbersIncluded()+src->NumbersIncluded()-nNumOverlap);
    }
  
public:
  virtual half_var_type*    GetHalfPart(byte p)    
    { 
      assert (sizeof(value_utype) >= 2); 
      assert (p < 2); 
      return new half_var_type_pattern(TO_HALF_TYPE(m_known,p),TO_HALF_TYPE(m_value,p)); 
    }
  virtual quarter_var_type* GetQuarterPart(byte p) 
    { 
      assert (sizeof(value_utype) >= 4); 
      assert (p < 4); 
      return new quarter_var_type_pattern(TO_QUARTER_TYPE(m_known,p),TO_QUARTER_TYPE(m_value,p)); 
    }
  
  /*
	virtual half_var_type* GetLowerHalf() { return new half_var_type_pattern(LOW_TO_HALF_TYPE(m_known),LOW_TO_HALF_TYPE(m_value)); }
	virtual half_var_type* GetUpperHalf() { return new half_var_type_pattern(HIGH_TO_HALF_TYPE(m_known),HIGH_TO_HALF_TYPE(m_value)); }

	virtual quarter_var_type* GetFirstQuarter() { return new quarter_var_type_pattern(FIRST_TO_QUARTER_TYPE(m_known),FIRST_TO_QUARTER_TYPE(m_value)); }
	virtual quarter_var_type* GetSecondQuarter() { return new quarter_var_type_pattern(SECOND_TO_QUARTER_TYPE(m_known),SECOND_TO_QUARTER_TYPE(m_value)); }
	virtual quarter_var_type* GetThirdQuarter() { return new quarter_var_type_pattern(THIRD_TO_QUARTER_TYPE(m_known),THIRD_TO_QUARTER_TYPE(m_value)); }
	virtual quarter_var_type* GetFourthQuarter() { return new quarter_var_type_pattern(FOURTH_TO_QUARTER_TYPE(m_known),FOURTH_TO_QUARTER_TYPE(m_value)); }
  */
 public:
  virtual string ValueString() const 
    { 
      // TODO: I'd rather like this to be like 0xfff?, or
      // 111111111111??10 if necessary because not 4-bit groups same..
      
      return strsprintf("(v=0x%x,k=0x%x)",m_value,m_known);
    }

};

DECL_INT_VARS(Pattern)

#endif // !defined(__VARIABLE_N_PATTERN_H__INCLUDED_)
