/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// VariableNRange.h: interface for the Variable class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__VARIABLE_N_RANGE_H__INCLUDED_)
#define __VARIABLE_N_RANGE_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

// // #include "VariableN.h" // dep_chk_unknown     

#include "Misc.h"

template<class unsignedT>
class VariableNRange : public VariableN<unsignedT>
{
 public:
  VARIABLE_N_BASE_TYPEDEF_VALUE_TYPES
  VARIABLE_N_TYPEDEF_VAR_TYPES(VariableN     ,var_type);
  VARIABLE_N_TYPEDEF_VAR_TYPES(VariableNRange,var_type_range);
  
 public:
  VariableNRange(value_utype min,value_utype max) 
    { m_min = min; m_max = max; }
  VariableNRange(value_utype value) 
    { m_min = value; m_max = value; }
  VariableNRange(const VariableNRange& src) : VariableN<unsignedT>(src)  
    { m_min = src.m_min; m_max = src.m_max; } 
  VariableNRange(const VariableNRange& src,bool bDep) 
    : VariableN<unsignedT>(src,bDep)  
    { m_min = src.m_min; m_max = src.m_max; } 
  
  VariableNRange(half_type nLowMin,half_type nLowMax,
		 const half_var_type_range* pHigh)
    {
      m_min = HALF_TYPE_TO_HIGH(pHigh->m_min) + nLowMin;
      m_max = HALF_TYPE_TO_HIGH(pHigh->m_max) + nLowMax;
    }
  
 public:
  virtual var_type* CopyVarContent() const 
    { return new VariableNRange(*this,false); }
  
  virtual int VariableNCompareCombType () const 
    { return VAR_N_COMP_TYPE_RANGE; }
  virtual int CompareComb(const var_type* rhs) const
    {
      COMPARE_COMB_BY_TYPE;
      const VariableNRange* pRhs = dynamic_cast<const VariableNRange*>(rhs);
      assert (rhs && pRhs);
      return CompareByValueRanges(pRhs);
    }
  
 public:
  virtual value_utype FirstValue() const 
    { return (m_min > m_max) ? 0 : m_min; }
  virtual value_utype LastValue() const 
    { return (m_min > m_max) ? ALL_ONES(value_utype) : m_max; }
  virtual bool NextValue(value_utype& value) const 
    { 
      assert (false); // when this describes a 32 bit range we'll iterate through 4e9 numbers, this should not happen
      // so rewrite the code that caused this (use First/Next ValueRange instead)
      
      if (value == m_max)
        {
          if (m_min > m_max)
            {
              value = m_min; // jump over the unused part
              return true;
            }
          return false; // we just had the end	  
        }
      if (!(++value))
        return false;
      return true;
    } // default: one value only
  virtual void FirstValueRange(ValueRange<unsignedT>& rng) const
    {
      if (m_min > m_max)
        {
          rng.m_min = 0; 
          rng.m_max = m_max;
        }
      else
        {
          rng.m_min = m_min; 
          rng.m_max = m_max;
        }
    }
  virtual bool NextValueRange(ValueRange<unsignedT>& rng) const
    {
      if (m_min > m_max && rng.m_min == 0)
        {
          rng.m_min = m_min;
          rng.m_max = ALL_ONES(value_utype);
          return true;
        }
      return false;
    }
  
 public:
  value_utype m_min; // inclusive
  value_utype m_max; // inclusive
  
 public:
  // if m_min > m_max, we mean the range that wraps (which is useful for signed numbers)
  // if m_min = m_max, we mean only one (if we meant entire range, the number m_min would
  // be included twice)
  
 public:
  bool IsInside (value_utype value) const
    { 
      if (m_min <= m_max)
        return value >= m_min && value <= m_max;
      else
        return value >= m_min || value <= m_max;
    }
  
 public:
  bool IsNormalRange() const { return m_min <= m_max; }
  bool IsWrappingRange() const { return !IsNormalRange(); }
  bool IsSignedRange() const { return (value_stype) m_min <= (value_stype) m_max; } // wrapping or not for unsigned, it must not wrap for signed
  
  bool IsNegNormalRange() const { return IsNormalRange(); }
  bool IsNegSignedRange() const { return (value_stype) -m_max <= (value_stype) -m_min; } // special test when using minus range, because -128 maps to -128 (not 128, as 128 does not exist (for byte, and similar for other types)) wrapping or not for unsigned, it must not wrap for signed
 
 public:
  bool IsBelow   (const VariableNRange* rhs) const { assert(IsNormalRange()); assert(rhs->IsNormalRange()); return m_max < rhs->m_min; }
  bool IsNotBelow(const VariableNRange* rhs) const { assert(IsNormalRange()); assert(rhs->IsNormalRange()); return m_min >= rhs->m_max; }
  bool IsNotAbove(const VariableNRange* rhs) const { assert(IsNormalRange()); assert(rhs->IsNormalRange()); return m_max <= rhs->m_min; }
  bool IsAbove   (const VariableNRange* rhs) const { assert(IsNormalRange()); assert(rhs->IsNormalRange()); return m_min > rhs->m_max; }
  
  bool IsLess      (const VariableNRange* rhs) const { assert(IsSignedRange()); assert(rhs->IsSignedRange()); return (value_stype) m_max < (value_stype) rhs->m_min; }
  bool IsNotLess   (const VariableNRange* rhs) const { assert(IsSignedRange()); assert(rhs->IsSignedRange()); return (value_stype) m_min >= (value_stype) rhs->m_max; }
  bool IsNotGreater(const VariableNRange* rhs) const { assert(IsSignedRange()); assert(rhs->IsSignedRange()); return (value_stype) m_max <= (value_stype) rhs->m_min; }
  bool IsGreater   (const VariableNRange* rhs) const { assert(IsSignedRange()); assert(rhs->IsSignedRange()); return (value_stype) m_min > (value_stype) rhs->m_max; }
  
 public:
  bool IsBelow   (const VariableNConst<unsignedT>* rhs) const { assert(IsNormalRange()); return m_max < rhs->m_value; }
  bool IsNotBelow(const VariableNConst<unsignedT>* rhs) const { assert(IsNormalRange()); return m_min >= rhs->m_value; }
  bool IsNotAbove(const VariableNConst<unsignedT>* rhs) const { assert(IsNormalRange()); return m_max <= rhs->m_value; }
  bool IsAbove   (const VariableNConst<unsignedT>* rhs) const { assert(IsNormalRange()); return m_min > rhs->m_value; }
  
  bool IsLess      (const VariableNConst<unsignedT>* rhs) const { assert(IsSignedRange()); return (value_stype) m_max < (value_stype) rhs->m_value; }
  bool IsNotLess   (const VariableNConst<unsignedT>* rhs) const { assert(IsSignedRange()); return (value_stype) m_min >= (value_stype) rhs->m_value; }
  bool IsNotGreater(const VariableNConst<unsignedT>* rhs) const { assert(IsSignedRange()); return (value_stype) m_max <= (value_stype) rhs->m_value; }
  bool IsGreater   (const VariableNConst<unsignedT>* rhs) const { assert(IsSignedRange()); return (value_stype) m_min > (value_stype) rhs->m_value; }
  
         // for rhs->m_value == 0
  bool IsLess0      () const { assert(IsSignedRange()); return (value_stype) m_max < (value_stype) 0; }
  bool IsNotLess0   () const { assert(IsSignedRange()); return (value_stype) m_min >= (value_stype) 0; }
  bool IsNotGreater0() const { assert(IsSignedRange()); return (value_stype) m_max <= (value_stype) 0; }
  bool IsGreater0   () const { assert(IsSignedRange()); return (value_stype) m_min > (value_stype) 0; }
  
 public:
  bool IsBelowNeg   (const VariableNRange* rhs) const { assert(IsNormalRange()); assert(rhs->IsNormalRange()); return m_max < -rhs->m_max; }
  bool IsNotBelowNeg(const VariableNRange* rhs) const { assert(IsNormalRange()); assert(rhs->IsNormalRange()); return m_min >= -rhs->m_min; }
  bool IsNotAboveNeg(const VariableNRange* rhs) const { assert(IsNormalRange()); assert(rhs->IsNormalRange()); return m_max <= -rhs->m_max; }
  bool IsAboveNeg   (const VariableNRange* rhs) const { assert(IsNormalRange()); assert(rhs->IsNormalRange()); return m_min > -rhs->m_min; }
  
  bool IsLessNeg      (const VariableNRange* rhs) const { assert(IsSignedRange()); assert(rhs->IsSignedRange()); return (value_stype) m_max < (value_stype) -rhs->m_max; }
  bool IsNotLessNeg   (const VariableNRange* rhs) const { assert(IsSignedRange()); assert(rhs->IsSignedRange()); return (value_stype) m_min >= (value_stype) -rhs->m_min; }
  bool IsNotGreaterNeg(const VariableNRange* rhs) const { assert(IsSignedRange()); assert(rhs->IsSignedRange()); return (value_stype) m_max <= (value_stype) -rhs->m_max; }
  bool IsGreaterNeg   (const VariableNRange* rhs) const { assert(IsSignedRange()); assert(rhs->IsSignedRange()); return (value_stype) m_min > (value_stype) -rhs->m_min; }
  
 public:
  bool IsBelowNeg   (const VariableNConst<unsignedT>* rhs) const { assert(IsNormalRange()); return m_max < -rhs->m_value; }
  bool IsNotBelowNeg(const VariableNConst<unsignedT>* rhs) const { assert(IsNormalRange()); return m_min >= -rhs->m_value; }
  bool IsNotAboveNeg(const VariableNConst<unsignedT>* rhs) const { assert(IsNormalRange()); return m_max <= -rhs->m_value; }
  bool IsAboveNeg   (const VariableNConst<unsignedT>* rhs) const { assert(IsNormalRange()); return m_min > -rhs->m_value; }
  
  bool IsLessNeg      (const VariableNConst<unsignedT>* rhs) const { assert(IsSignedRange()); return (value_stype) m_max < (value_stype) -rhs->m_value; }
  bool IsNotLessNeg   (const VariableNConst<unsignedT>* rhs) const { assert(IsSignedRange()); return (value_stype) m_min >= (value_stype) -rhs->m_value; }
  bool IsNotGreaterNeg(const VariableNConst<unsignedT>* rhs) const { assert(IsSignedRange()); return (value_stype) m_max <= (value_stype) -rhs->m_value; }
  bool IsGreaterNeg   (const VariableNConst<unsignedT>* rhs) const { assert(IsSignedRange()); return (value_stype) m_min > (value_stype) -rhs->m_value; }
  
  // for rhs->m_value == 0
  bool IsLessNeg0      () const { assert(IsSignedRange()); return (value_stype) m_max < (value_stype) 0; }
  bool IsNotLessNeg0   () const { assert(IsSignedRange()); return (value_stype) m_min >= (value_stype) 0; }
  bool IsNotGreaterNeg0() const { assert(IsSignedRange()); return (value_stype) m_max <= (value_stype) 0; }
  bool IsGreaterNeg0   () const { assert(IsSignedRange()); return (value_stype) m_min > (value_stype) 0; }
  
 public:
  virtual VariableN<unsignedT>* SimplifyVariable();
  
 public:
  virtual value_utype NumbersIncluded () const { return m_max-m_min+1; }
  virtual bool NothingKnown () const { return m_min == m_max + 1; }
  
public:
  void InsertValue(value_utype value)
    {
      if (IsInside (value))
        return; // already done
      
      // find out if it cheaper to add to the beginning or to the end
      
      assert ((value - m_min) > (m_max - m_min)); // if including with max, make sure we don't loose anything
      assert ((m_max - value) > (m_max - m_min)); // if including with min, make sure we don't loose anything
      
      if ((value - m_min) < (m_max - value))
        m_max = value; // include with max
      else
        m_min = value; // include with min
    }

  value_utype CostOfInsertValue(value_utype value) const
    {
      if (IsInside (value))
        return 0; // already known
      
      // find out if it cheaper to add to the beginning or to the end
      
      assert ((value - m_min) > (m_max - m_min)); // if including with max, make sure we don't loose anything
      assert ((m_max - value) > (m_max - m_min)); // if including with min, make sure we don't loose anything
      
      if ((value - m_min) < (m_max - value))
        return (value - m_max) - 1; // include with max
      else
        return (m_min - value) - 1; // include with min
    }
  
  void ExpandRangeMax (value_utype max)
    {
      assert (!IsInside(max) || max == m_max);
      m_max = max;
    }
  
  void ExpandRangeMin (value_utype min)
    {
      assert (!IsInside(min) || min == m_min);
      m_min = min;
    }
  
 public:
  // TODO: there are some cases that aren't covered by GetHalfPart and
  // GetQuarterPart like the high part being something fixed and then
  // returning the variantion in the low part but we should just miss
  // these and return NULL for now, so nothing wrong should come out
  
  virtual half_var_type*    GetHalfPart(byte p)    
    { 
      assert (sizeof(value_utype) >= 2); 
      assert (p < 2); 
      if (p)
	{
	  // high part
	  if (m_min < m_max)
	    return new half_var_type_range(HIGH_TO_HALF_TYPE(m_min),HIGH_TO_HALF_TYPE(m_max));
	  return NULL;
	}
      else
	{
	  // low part
	  if (m_min < m_max &&
	      m_max < TYPE_RANGE(half_type))
	    return new half_var_type_range(LOW_TO_HALF_TYPE(m_min),LOW_TO_HALF_TYPE(m_max));
	  return NULL;
	}
    }
  virtual quarter_var_type* GetQuarterPart(byte p) 
    { 
      assert (sizeof(value_utype) >= 4); 
      assert (p < 4); 
      if (p == 3)
	{
	  // high part
	  if (m_min < m_max)
	    return new quarter_var_type_range(FOURTH_TO_QUARTER_TYPE(m_min),FOURTH_TO_QUARTER_TYPE(m_max));
	  return NULL;
	}
      else
	{
	  // low part
	  if (m_min < m_max &&
	      m_max < (value_utype) MULT_TYPE_RANGE(quarter_type,p))
	    return new quarter_var_type_range(TO_QUARTER_TYPE(m_min,p),TO_QUARTER_TYPE(m_max,p));
	  return NULL;
	}
      
    }
  /*
        virtual half_var_type* GetLowerHalf()
        {
                if (m_min < m_max &&
                        m_max < TYPE_RANGE(half_type))
                        return new half_var_type_range(LOW_TO_HALF_TYPE(m_min),LOW_TO_HALF_TYPE(m_max));
                return NULL;
        }
        virtual half_var_type* GetUpperHalf()
        {
                if (m_min < m_max)
                        return new half_var_type_range(0,HIGH_TO_HALF_TYPE(m_max));
                return NULL;
        }

        virtual quarter_var_type* GetFirstQuarter()
        {
                if (m_min < m_max &&
                        m_max < TYPE_RANGE(quarter_type)) 
                        return new quarter_var_type_range(FIRST_TO_QUARTER_TYPE(m_min),FIRST_TO_QUARTER_TYPE(m_max));
                return NULL;
        }
        virtual quarter_var_type* GetSecondQuarter()
        {
                if (m_min < m_max &&
                        m_max < TWICE_TYPE_RANGE(quarter_type)) 
                        return new quarter_var_type_range(0,SECOND_TO_QUARTER_TYPE(m_max));
                return NULL;
        }
  */
 public:
  virtual string ValueString() const 
    { 
      if (IsNormalRange())
	return strsprintf("[0x%x-0x%x]",m_min,m_max); 
      else
	return strsprintf("[..-0x%x,0x%x-..]",m_max,m_min);
      }

};

DECL_INT_VARS(Range)

#endif // !defined(__VARIABLE_N_RANGE_H__INCLUDED_)
