/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// VariableNSplit.h: interface for the Variable class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__VARIABLE_N_SPLIT_H__INCLUDED_)
#define __VARIABLE_N_SPLIT_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

// // #include "VariableN.h" // dep_chk_unknown     

template<class unsignedT>
class VariableNSplit : public VariableN<unsignedT>
{
 public:
  VARIABLE_N_BASE_TYPEDEF_VALUE_TYPES
  VARIABLE_N_TYPEDEF_VAR_TYPES(VariableN      ,var_type);
  VARIABLE_N_TYPEDEF_VAR_TYPES(VariableNSplit ,var_type_split);
  
 public:
  VariableNSplit() { m_low = NULL; m_high = NULL; }
  VariableNSplit(half_var_type* low,half_var_type* high = NULL) 
    { m_low = low; m_high = high; }
  VariableNSplit(const VariableNSplit& src) : VariableN<unsignedT>(src)  
    { m_low = src.m_low->CopyVarContent(); 
      m_high = src.m_high->CopyVarContent(); }
  VariableNSplit(const VariableNSplit& src,bool bDep) 
    : VariableN<unsignedT>(src,bDep)  
    { m_low = src.m_low->CopyVarContent(); 
      m_high = src.m_high->CopyVarContent(); }
  
 public:
  virtual var_type* CopyVarContent() const 
    { return new VariableNSplit(*this,false); }
  
  virtual int VariableNCompareCombType () const 
    { return VAR_N_COMP_TYPE_SPLIT; }
  virtual int CompareComb(const var_type* rhs) const
    {
      COMPARE_COMB_BY_TYPE;
      const VariableNSplit* pRhs = dynamic_cast<const VariableNSplit*>(rhs);
      assert (rhs && pRhs);
      int nComp = CompareCombLowOrHigh(m_high,pRhs->m_high);
      if (nComp) return nComp;
      return CompareCombLowOrHigh(m_low,pRhs->m_low);
    }
 protected:
  static int CompareCombLowOrHigh(half_var_type* lhs,half_var_type* rhs) 
    {
      if (!lhs)
        return !rhs ? 0 : 1;
      if (!rhs)
        return -1;
      return lhs->CompareComb(rhs);
    }
  
 public:
  virtual value_utype FirstValue() const
    {
      return (m_low ? m_low->FirstValue() : (half_type) 0) +
        (m_high ? HALF_TYPE_TO_HIGH((half_type) m_high->FirstValue()) : 0);
    }
  virtual value_utype LastValue() const
    {
      return (m_low ? m_low->LastValue() : ALL_ONES(half_type)) +
        (m_high ? HALF_TYPE_TO_HIGH((half_type) m_high->LastValue()) : 
	 HALF_TYPE_TO_HIGH(ALL_ONES(half_type)));
    }
  virtual bool NextValue(value_utype& value) const
    {
      half_type val_low  =  LOW_TO_HALF_TYPE(value);
      half_type val_high = HIGH_TO_HALF_TYPE(value);
      
      // first get next of low part.
      // if that has no more, see if high part can give more and restart
      // low part
      
      bool bHasMore = false;
      
      if (m_low)
        bHasMore = m_low->NextValue(val_low);
      else
        bHasMore = ((++val_low) != 0);
      
      if (bHasMore)
        {
          value = HALF_TYPE_TO_TYPE(val_low,val_high);			
          return true;
        }
      
      // low part had no more, see if high part can do more
      
      bHasMore = false;
      
      if (m_high)
        bHasMore = m_high->NextValue(val_high);
      else
        bHasMore = ((++val_high) != 0);
      
      if (bHasMore)
        {
          value = HALF_TYPE_TO_TYPE((m_low ? m_low->FirstValue() : 0),
				    val_high);			
          return true;
        }
      return false;
    }
  
  virtual void FirstValueRange(ValueRange<unsignedT>& /* rng */) const 
    { assert (false); } // not implemented
  virtual bool NextValueRange(ValueRange<unsignedT>& /* rng */) const  
    { assert (false); return false; } // not implemented
  
 public:
  half_var_type* m_low;
  half_var_type* m_high;
  
 public:
  virtual value_utype NumbersIncluded () const
    { 
      return 
	(m_low ? 
	 (value_utype) m_low->NumbersIncluded() : ALL_ONES(half_type)) *
        (m_high ? 
	 (value_utype) m_high->NumbersIncluded() : ALL_ONES(half_type));
    }
  virtual bool NothingKnown () const 
    { 
      if (m_low && !m_low->NothingKnown())
        return false; // something known by low
      if (m_high && !m_high->NothingKnown())
        return false; // something known by high
      return true; // noone knows nothing
    }
  
public:
  // virtual VariableN<unsignedT>* SimplifyVariable();
  
 public:
  virtual half_var_type*    GetHalfPart(byte p)    
    { 
      assert (sizeof(value_utype) >= 2); 
      assert (p < 2); 
      return p ? m_high : m_low;
    }
  virtual quarter_var_type* GetQuarterPart(byte p) 
    { 
      assert (sizeof(value_utype) >= 4); 
      assert (p < 4); 

      half_var_type* part = (p & 2) ? m_high : m_low;
      
      return part ? part->GetHalfPart(p & 1) : NULL;
    }
  /*
  virtual half_var_type* GetLowerHalf() { return m_low; }
  virtual half_var_type* GetUpperHalf() { return m_high; }
  
  virtual quarter_var_type* GetFirstQuarter() { return m_low->GetLowerHalf(); }
  virtual quarter_var_type* GetSecondQuarter() { return m_low->GetUpperHalf(); }
  */
  
  virtual string ValueString() const 
    { 
      return "(" + (m_high ? m_high->ValueString() : "?") + "," + (m_low ? m_low->ValueString() : "?") + ")"; 
    }
  
};



typedef VariableNSplit<word>  Variable16Split;
typedef VariableNSplit<dword> Variable32Split;

#endif // !defined(__VARIABLE_N_SPLIT_H__INCLUDED_)


