/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// VariableOp.cpp: implementation of the VariableOp class.
//
//////////////////////////////////////////////////////////////////////

#include "StatusFlagsAffected.h"
#include "VariableOp.h" 
#include "VarContentOp.h" 
#include "VariableConvert.h" 
#include "Instruction.h" 
#include <stdio.h>

#include "OperandMemory.h"

#include "VariableNConst.h" 
#include "VariableNConsts.h"
#include "VariableNRange.h" 
#include "VariableNPattern.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

VariableOp::VariableOp()
{

}

VariableOp::~VariableOp()
{ 

}

//////////////////////////////////////////////////////////////////////
/*
word value_flags_CPAZSO(bool bCF,bool bPF,bool bAF,bool bZF,bool bSF,bool bOF)
{
	return ((word) bCF) |
		(((word) (bCF | bZF)) << 1) | 
		(((word) bPF) << 2) | 
		(((word) (bZF | (bSF ^ bOF))) << 3) | 
		(((word) bAF) << 4) | 
		(((word) (bSF ^ bOF)) << 4) | 
		(((word) bZF) << 6) | 
		(((word) bSF) << 7) | 
		(((word) bOF) << 10);
};
*/

bool parity (byte value)
{
  value ^= (value >> 4); // 4 least significant bits 0 if even numbers of 1's (or 0's)
  value ^= (value >> 2); // 2 least significant bits 0 if even numbers of 1's (or 0's)
  value ^= (value >> 1); // 1 least significant bit  0 if even numbers of 1's (or 0's)
  
  return !(value & 0x01);
}

//////////////////////////////////////////////////////////////////////

template<class unsignedT> unsignedT do_Add   (const unsignedT& lhs,const unsignedT& rhs) { return (lhs + rhs); }
template<class unsignedT> unsignedT do_Sub   (const unsignedT& lhs,const unsignedT& rhs) { return (lhs - rhs); }
template<class unsignedT> unsignedT do_InvSub(const unsignedT& lhs,const unsignedT& rhs) { return (rhs - lhs); } // note swapped args
template<class unsignedT> unsignedT do_And   (const unsignedT& lhs,const unsignedT& rhs) { return (lhs & rhs); }
template<class unsignedT> unsignedT do_Or    (const unsignedT& lhs,const unsignedT& rhs) { return (lhs | rhs); }
template<class unsignedT> unsignedT do_Xor   (const unsignedT& lhs,const unsignedT& rhs) { return (lhs ^ rhs); }

template<class unsignedT> unsignedT do_Neg   (const unsignedT& lhs) { return (-lhs); }
template<class unsignedT> unsignedT do_Not   (const unsignedT& lhs) { return (~lhs); }
template<class unsignedT> unsignedT do_Inc   (const unsignedT& lhs) { return (lhs+1); }
template<class unsignedT> unsignedT do_Dec   (const unsignedT& lhs) { return (lhs-1); }

#include "VariableOp_CalcResult.h" // dep_chk_unknown 

#define lhs_may_swap lhs // to allow swapping for InvSub, but not for Add or Sub
#define rhs_may_swap rhs // CANNOT be inside COPY_OP_FUNC, because then it would be swapped too, nulling any effect
/* COPY_OP_FUNC @replace=({'Add'=>'Sub','add'=>'sub'},
                           {'Add'=>'InvSub','add'=>'sub','lhs_may_swap'=>'rhs_may_swap'}); */
template<class unsignedT> unsignedT do_Add   (const unsignedT& lhs,const unsignedT& rhs,StatusFlagsAffected& flags) 
{ 
  unsignedT result;
  calc_result_with_flags_type_info<unsignedT>::
    CalcResultWithFlags_add(&lhs_may_swap,&rhs_may_swap,&result,&flags.m_wValue);
  flags.ConstrainAffectedByCPAZSO();
  return result; 
}
/* END_COPY_OP_FUNC */
/* COPY_OP_FUNC @replace=({'And'=>'Or', 'and'=>'or'},
			   {'And'=>'Xor','and'=>'xor'}); */
template<class unsignedT> unsignedT do_And   (const unsignedT& lhs,const unsignedT& rhs,StatusFlagsAffected& flags) 
{ 
  unsignedT result;
  calc_result_with_flags_type_info<unsignedT>::
    CalcResultWithFlags_and(&lhs,&rhs,&result,&flags.m_wValue);
  flags.ConstrainAffectedByCPZSO();
  return result; 
}
/* END_COPY_OP_FUNC */

template<class unsignedT> unsignedT do_Neg   (const unsignedT& lhs,StatusFlagsAffected& flags) 
{ 
  unsignedT result;
  calc_result_with_flags_type_info<unsignedT>::
    CalcResultWithFlags_neg(&lhs,&result,&flags.m_wValue);
  flags.ConstrainAffectedByCPAZSO();
  return result; 
}
template<class unsignedT> unsignedT do_Not   (const unsignedT& lhs,StatusFlagsAffected& /* flags */) { return (~lhs); } // does not affect flags
/* COPY_OP_FUNC @replace=({'Inc'=>'Dec','inc'=>'dec'}); */
template<class unsignedT> unsignedT do_Inc   (const unsignedT& lhs,StatusFlagsAffected& flags) 
{ 
  unsignedT result;
  calc_result_with_flags_type_info<unsignedT>::
    CalcResultWithFlags_inc(&lhs,&result,&flags.m_wValue);
  flags.ConstrainAffectedByPAZSO();
  return result; 
}
/* END_COPY_OP_FUNC */

//////////////////////////////////////////////////////////////////////

void SetFlagsUnknownAddSubNeg(StatusFlagsAffected& flags)
{
  flags.AffectedByCPAZSO();
}

void SetFlagsUnknownAndOrXor(StatusFlagsAffected& flags)
{
  flags.AffectedByCPZSO();
  flags.Undefined(FLAGS_MASK_AF);
  flags.SetKnown0(FLAGS_MASK_CF | FLAGS_MASK_OF);
}

void SetFlagsUnknownIncDec(StatusFlagsAffected& flags)
{
  flags.AffectedByPAZSO();
}

//////////////////////////////////////////////////////////////////////

/*
template<class unsignedT> unsignedT do_Mul    (const unsignedT& lhs,const unsignedT& rhs) { return (lhs * rhs); }
template<class unsignedT> unsignedT ok_Div    (const unsignedT& lhs,const unsignedT& rhs) { return (rhs != 0);  }
template<class unsignedT> unsignedT do_Div    (const unsignedT& lhs,const unsignedT& rhs) { return (lhs / rhs); }
template<class unsignedT> unsignedT ok_InvDiv (const unsignedT& lhs,const unsignedT& rhs) { return (lhs != 0);  }
template<class unsignedT> unsignedT do_InvDiv (const unsignedT& lhs,const unsignedT& rhs) { return (rhs / lhs); }
template<class unsignedT> unsignedT do_IMul   (const unsignedT& lhs,const unsignedT& rhs) { return (lhs ^ rhs); }
template<class unsignedT> unsignedT do_IDiv   (const unsignedT& lhs,const unsignedT& rhs) { return (lhs ^ rhs); }
template<class unsignedT> unsignedT do_InvIDiv(const unsignedT& lhs,const unsignedT& rhs) { return (lhs ^ rhs); }
*/

//////////////////////////////////////////////////////////////////////
// define how 'good' different arguments are, i.e. how little fill in they have
// which should preferably be used when doing conversions

//////////////////////////////////////////////////////////////////////
// CONST_CONST

/* COPY_OP_FUNC @replace=({'Add'=>'Sub'},{'Add'=>'And'},{'Add'=>'Or'},{'Add'=>'Xor'}); */
template<class unsignedT>
VariableNConst<unsignedT>* Add(const VariableNConst<unsignedT>* lhs, 
			       const VariableNConst<unsignedT>* rhs, 
			       StatusFlagsAffected& flags) 
{ 
  unsignedT nResult = do_Add(lhs->m_value,rhs->m_value,flags); 
  return new VariableNConst<unsignedT>(nResult); 
}

template<class unsignedT> 
word AddConstrained(const VariableNConst<unsignedT>* lhs, 
		    const VariableNConst<unsignedT>* rhs, 
		    const StatusFlagsKnown& flags_known, 
		    VariableN<unsignedT>** /* out_lhs */, 
		    VariableN<unsignedT>** /* out_rhs */, 
		    VariableN<unsignedT>** /* out_result */, 
		    StatusFlagsAffected& flags)
{ 
  /* unsignedT nResult = */ do_Add(lhs->m_value,rhs->m_value,flags); 
  if (flags.IsCompatible(flags_known))
    return CONSTRAINED_NO_CONSTRAINTS; // result same as calculated unconstrained
  else 
    return CONSTRAINED_UNREACHABLE; // calculation couldn't give the flags asked for
} 
/* END_COPY_OP_FUNC */

/* COPY_OP_FUNC @replace=({'Neg'=>'Not'},{'Neg'=>'Inc'},{'Neg'=>'Dec'}); */
template<class unsignedT> 
VariableNConst<unsignedT>* Neg(const VariableNConst<unsignedT>* lhs, 
			       StatusFlagsAffected& flags) 
{ 
  unsignedT nResult = do_Neg(lhs->m_value,flags); 
  return new VariableNConst<unsignedT>(nResult); 
} 

template<class unsignedT> 
word NegConstrained(const VariableNConst<unsignedT>* lhs, 
		    const StatusFlagsKnown& flags_known, 
		    VariableN<unsignedT>** /* out_lhs */, 
		    VariableN<unsignedT>** /* out_result */, 
		    StatusFlagsAffected& flags) 
{ 
  /* unsignedT nResult = */ do_Neg(lhs->m_value,flags); 
  if (flags.IsCompatible(flags_known)) 
    return CONSTRAINED_NO_CONSTRAINTS; // result same as calculated unconstrained 
  else 
    return CONSTRAINED_UNREACHABLE; // calculation couldn't give the flags asked for
} 
/* END_COPY_OP_FUNC */

// // binary operator, but one of the operands may be unknown
// template<class unsignedT>
// VariableNPattern<unsignedT>* And_NOT_IN_USE(const VariableNConst<unsignedT>* lrhs/*,ANY_PARAMETER*/, 
// // {
//   return new VariableNPattern<unsignedT>(~lrhs->m_value, // all that are zero will be zero (when anded with 0 or 1)
// 					 0);
// } 
// 
// // binary operator, but one of the operands may be unknown
// template<class unsignedT>
// VariableNPattern<unsignedT>* Or_NOT_IN_USE(const VariableNConst<unsignedT>* lrhs/*,ANY_PARAMETER*/, 
// 					StatusFlagsAffected& flags)
// {
// 
// 	return new VariableNPattern<unsignedT>(lrhs->m_value, // all that are one will be one (when ored with 0 or 1)
// 					 lrhs->m_value);
// } 

//////////////////////////////////////////////////////////////////////
// CONSTS_CONST

#define DELETE3_AND_RETURN(del1,del2,del3,retval) { delete (del1);delete (del2);delete (del3);return (retval); }
#define DELETE2_AND_RETURN(del1,del2,retval) { delete (del1);delete (del2);return (retval); }

/* COPY_OP_FUNC @replace=({'Add'=>'Sub'},
                           {'Add'=>'InvSub'},
                           {'Add'=>'And'},
                           {'Add'=>'Or' },
                           {'Add'=>'Xor'}); */
template<class unsignedT>
VariableNConsts<unsignedT>* Add(const VariableNConsts<unsignedT>* lhs,
				const VariableNConst<unsignedT>* rhs,
				StatusFlagsAffected& flags)
{
  VariableNConsts<unsignedT>* result = new VariableNConsts<unsignedT>;
  
  typename VariableNConsts<unsignedT>::VectValues::const_iterator lhs_value = (lhs->m_vectValues).begin();
  assert (lhs_value != (lhs->m_vectValues).end());
  result->InsertValue(do_Add(*lhs_value,rhs->m_value,flags)); /* First value, set flags */
  ++lhs_value;
  
  for( ; lhs_value != (lhs->m_vectValues).end(); ++lhs_value)
    {
      StatusFlagsAffected thisflags;
      result->InsertValue(do_Add(*lhs_value,rhs->m_value,thisflags));
      flags.Join (thisflags);
    }
  return result;
}

template<class unsignedT>
word AddConstrained(const VariableNConsts<unsignedT>* lhs,
		    const VariableNConst<unsignedT>* rhs,
		    const StatusFlagsKnown& flags_known, 
		    VariableN<unsignedT>** out_lhs, 
		    VariableN<unsignedT>** /* out_rhs */, 
		    VariableN<unsignedT>** out_result, 
		    StatusFlagsAffected& flags)
{
  VariableNConsts<unsignedT>* result = new VariableNConsts<unsignedT>;
  VariableNConsts<unsignedT>* allowed_lhs = new VariableNConsts<unsignedT>;
  
  FOR_EACH_CONST_ITER(typename VariableNConsts<unsignedT>::VectValues,
		      lhs->m_vectValues,lhs_value)
    {
      StatusFlagsAffected thisflags;
      unsignedT nResult = do_Add(*lhs_value,rhs->m_value,thisflags); 
      if (thisflags.IsCompatible(flags_known)) 
        {
          if (result->size()) // already have some
            flags.Join (thisflags);
          else
            flags = thisflags;
          result->InsertValue(nResult);
          allowed_lhs->InsertValue(*lhs_value);
        }      
    }
  
  if (!result->size())
    DELETE2_AND_RETURN(result,allowed_lhs,CONSTRAINED_UNREACHABLE);
  if (allowed_lhs->size() == lhs->size())
    DELETE2_AND_RETURN(result,allowed_lhs,CONSTRAINED_NO_CONSTRAINTS);
  *out_result = result;
  *out_lhs = allowed_lhs;
  // rhs is necessarily same
  return CONSTRAINED_LHS_RESULT_NOT_SAME;
}
/* END_COPY_OP_FUNC */

//////////////////////////////////////////////////////////////////////
// CONSTS_CONSTS

/* COPY_OP_FUNC @replace=({'Range'=>'Pattern'}); */
template<class unsignedT>
void InsertConstsResult(unsignedT nResult,
			VariableNConsts<unsignedT>*& result,
			VariableNRange<unsignedT>*& result2)
{
  if (result)
    {
      result->InsertValue(nResult);
      if (result->m_vectValues.size() > VAR_CONSTS_MAX_NO_OF_VALUES)
        {
          result2 = ConvertToRange(result);
          delete result;
          result = NULL;
        }
    }
  else
    result2->InsertValue(nResult);
}
/* END_COPY_OP_FUNC */

/* COPY_OP_FUNC @replace=({'Add'=>'Sub'},
                           {'Add'=>'And','Range'=>'Pattern'},
                           {'Add'=>'Or' ,'Range'=>'Pattern'},
                           {'Add'=>'Xor','Range'=>'Pattern'}); */
template<class unsignedT>
VariableN<unsignedT>* Add(const VariableNConsts<unsignedT>* lhs,
			  const VariableNConsts<unsignedT>* rhs,
			  StatusFlagsAffected& flags)
{
  VariableNConsts<unsignedT>* result = new VariableNConsts<unsignedT>;
  VariableNRange<unsignedT>* result2 = NULL;
  
  assert (lhs->m_vectValues.size());
  assert (rhs->m_vectValues.size());
  result->InsertValue(do_Add(lhs->m_vectValues[0],rhs->m_vectValues[0],flags)); /* Get first flags */
  
  FOR_EACH_CONST_ITER(typename VariableNConsts<unsignedT>::VectValues,
		      lhs->m_vectValues,lhs_value)
    FOR_EACH_CONST_ITER(typename VariableNConsts<unsignedT>::VectValues,
			rhs->m_vectValues,rhs_value)
    {
      StatusFlagsAffected thisflags;
      InsertConstsResult(do_Add(*lhs_value,*rhs_value,thisflags),
			 result,result2);
      flags.Join (thisflags);
    }
  assert (result || result2);
  return result ? 
    (VariableN<unsignedT>*) result : (VariableN<unsignedT>*) result2;
}

template<class unsignedT>
word AddConstrained(const VariableNConsts<unsignedT>* lhs,
		    const VariableNConsts<unsignedT>* rhs,
		    const StatusFlagsKnown& flags_known, 
		    VariableN<unsignedT>** out_lhs, 
		    VariableN<unsignedT>** out_rhs, 
		    VariableN<unsignedT>** out_result, 
		    StatusFlagsAffected& flags)
{
  VariableNConsts<unsignedT>* result = new VariableNConsts<unsignedT>;
  VariableNRange<unsignedT>* result2 = NULL;
  VariableNConsts<unsignedT>* allowed_lhs = new VariableNConsts<unsignedT>;
  VariableNConsts<unsignedT>* allowed_rhs = new VariableNConsts<unsignedT>;
  
  FOR_EACH_CONST_ITER(typename VariableNConsts<unsignedT>::VectValues,
		      lhs->m_vectValues,lhs_value)
    FOR_EACH_CONST_ITER(typename VariableNConsts<unsignedT>::VectValues,
			rhs->m_vectValues,rhs_value)
    {
      StatusFlagsAffected thisflags;
      unsignedT nResult = do_Add(*lhs_value,*rhs_value,thisflags); 
      
      if (thisflags.IsCompatible(flags_known)) 
        {
          if (result->size()) // already have some
            flags.Join (thisflags);
          else
            flags = thisflags;
          InsertConstsResult(nResult,result,result2);
          allowed_lhs->InsertValue(*lhs_value);
          allowed_rhs->InsertValue(*rhs_value);
        }      
    }
  assert (result || result2);

  if (result && !result->size()) // if result2, then it has some size
    DELETE3_AND_RETURN(result,allowed_lhs,allowed_rhs,CONSTRAINED_UNREACHABLE);
  // ok so we were reachable, see if we have been constrained in any way
  if (allowed_lhs->size() == lhs->size() &&
      allowed_rhs->size() == rhs->size())
    DELETE3_AND_RETURN(result,allowed_lhs,allowed_rhs,
		       CONSTRAINED_NO_CONSTRAINTS);
  // reachable but either lhs and/or rhs were constrained
  // assume (although not necessarily correct) that result was also constrained
  word wReturn = CONSTRAINED_RESULT_NOT_SAME;
  *out_result = result;
  if (allowed_lhs->size() < lhs->size())
    {
      assert (allowed_lhs->size()); // otherwise unreachable
      *out_lhs = allowed_lhs;
      wReturn |= CONSTRAINED_LHS_NOT_SAME;
    }
  else
    delete allowed_lhs;
  if (allowed_rhs->size() < rhs->size())
    {
      assert (allowed_rhs->size()); // otherwise unreachable
      *out_rhs = allowed_rhs;
      wReturn |= CONSTRAINED_RHS_NOT_SAME;
    }
  else
    delete allowed_rhs;

  return wReturn;
}
/* END_COPY_OP_FUNC */

/* COPY_OP_FUNC @replace=({'Neg'=>'Not'},{'Neg'=>'Inc'},{'Neg'=>'Dec'}); */
template<class unsignedT>
VariableNConsts<unsignedT>* Neg(const VariableNConsts<unsignedT>* lhs,
				StatusFlagsAffected& flags)
{
  VariableNConsts<unsignedT>* result = new VariableNConsts<unsignedT>;
  
  typename VariableNConsts<unsignedT>::VectValues::const_iterator lhs_value = 
    (lhs->m_vectValues).begin();
  assert (lhs_value != (lhs->m_vectValues).end());
  result->InsertValue(do_Neg(*lhs_value,flags)); /* First value, set flags */
  ++lhs_value;
  
  for( ; lhs_value != (lhs->m_vectValues).end(); ++lhs_value)
    {
      StatusFlagsAffected thisflags;
      result->InsertValue(do_Neg(*lhs_value,thisflags));
      flags.Join (thisflags);
    }
  return result;
}

template<class unsignedT>
word NegConstrained(const VariableNConsts<unsignedT>* lhs,
		    const StatusFlagsKnown& flags_known, 
		    VariableN<unsignedT>** out_lhs, 
		    VariableN<unsignedT>** out_result, 
		    StatusFlagsAffected& flags)
{
  VariableNConsts<unsignedT>* result = new VariableNConsts<unsignedT>;
  VariableNConsts<unsignedT>* allowed_lhs = new VariableNConsts<unsignedT>;

  FOR_EACH_CONST_ITER(typename VariableNConsts<unsignedT>::VectValues,
		      lhs->m_vectValues,lhs_value)
    {
      StatusFlagsAffected thisflags;
      unsignedT nResult = do_Neg(*lhs_value,thisflags); 
      if (thisflags.IsCompatible(flags_known)) 
        {
          if (result->size()) // already have some
            flags.Join (thisflags);
          else
            flags = thisflags;
          result->InsertValue(nResult);
          allowed_lhs->InsertValue(*lhs_value);
        }      
    }

  if (!result->size())
    DELETE2_AND_RETURN(result,allowed_lhs,CONSTRAINED_UNREACHABLE);
  if (allowed_lhs->size() == lhs->size())
    DELETE2_AND_RETURN(result,allowed_lhs,CONSTRAINED_NO_CONSTRAINTS);
  *out_result = result;
  *out_lhs = allowed_lhs;
  return CONSTRAINED_LHS_RESULT_NOT_SAME;
}
/* END_COPY_OP_FUNC */

//////////////////////////////////////////////////////////////////////
// CONST_RANGE

template<class unsignedT>
VariableNRange<unsignedT>* SetFlagsForRangeSZP(StatusFlagsAffected& flags,
					       VariableNRange<unsignedT>* range)
{
  if (range->m_min <= range->m_max)
    {
      // all with same sign?
      if ((range->m_min & MOST_SIGNIFICANT_ONE(unsignedT)) ==
	  (range->m_max & MOST_SIGNIFICANT_ONE(unsignedT)))
	flags.SetKnown(FLAGS_MASK_SF,
		       !!(range->m_min & MOST_SIGNIFICANT_ONE(unsignedT)));
      
      if (range->m_min > 0)
	flags.SetKnown0(FLAGS_MASK_ZF);
      else if (range->m_max == 0)
	{
	  assert (range->m_min == 0);
	  flags.SetKnown1(FLAGS_MASK_ZF);
	}
    }
  // parity mess is probably never used, but who cares...
  // we'll check it anyway
  bool bParity = parity(range->m_min);
  unsignedT nCheck = range->m_min;
  for ( ; nCheck != range->m_max; nCheck++)
    {
      if (parity(nCheck) != bParity)
	goto parity_unknown;
    }
  flags.SetKnown(FLAGS_MASK_PF,bParity);
 parity_unknown:;
  return range;
}

template<class unsignedT>
VariableNRange<unsignedT>* Add(const VariableNRange<unsignedT>* lhs,
			       const VariableNConst<unsignedT>* rhs,
			       StatusFlagsAffected& flags)
{
  SetFlagsUnknownAddSubNeg(flags);
  
  if (lhs->IsNormalRange())
    {
      if (lhs->IsBelowNeg(rhs))	     flags.SetKnown1(FLAGS_MASK_CF);
      if (lhs->IsNotBelowNeg(rhs))   flags.SetKnown0(FLAGS_MASK_CF);
      if (lhs->IsNotAboveNeg(rhs))   flags.SetKnown1(FLAGS_MASK_CForZF);
      if (lhs->IsAboveNeg(rhs))	     flags.SetKnown0(FLAGS_MASK_CForZF);
    }				     
  if (lhs->IsSignedRange())	     
    {				     
      if (lhs->IsLessNeg(rhs))	     flags.SetKnown1(FLAGS_MASK_SFneqOF);
      if (lhs->IsNotLessNeg(rhs))    flags.SetKnown0(FLAGS_MASK_SFneqOF);
      if (lhs->IsNotGreaterNeg(rhs)) flags.SetKnown1(FLAGS_MASK_ZForSFneqOF);
      if (lhs->IsGreaterNeg(rhs))    flags.SetKnown0(FLAGS_MASK_ZForSFneqOF);
    }
  
  // TODO: calc AF and OF flag if possible
  
  return SetFlagsForRangeSZP(flags,
			     new VariableNRange<unsignedT>(lhs->m_min + rhs->m_value,
							   lhs->m_max + rhs->m_value));
} 

template<class unsignedT>
VariableNRange<unsignedT>* Sub(const VariableNRange<unsignedT>* lhs,
			       const VariableNConst<unsignedT>* rhs,
			       StatusFlagsAffected& flags)
{
  SetFlagsUnknownAddSubNeg(flags);
  
  if (lhs->IsNormalRange())
    {
      if (lhs->IsBelow(rhs))	     flags.SetKnown1(FLAGS_MASK_CF);
      if (lhs->IsNotBelow(rhs))	     flags.SetKnown0(FLAGS_MASK_CF);
      if (lhs->IsNotAbove(rhs))	     flags.SetKnown1(FLAGS_MASK_CForZF);
      if (lhs->IsAbove(rhs))	     flags.SetKnown0(FLAGS_MASK_CForZF);
    }				     
  if (lhs->IsSignedRange())	     
    {				     
      if (lhs->IsLess(rhs))	     flags.SetKnown1(FLAGS_MASK_SFneqOF);
      if (lhs->IsNotLess(rhs))	     flags.SetKnown0(FLAGS_MASK_SFneqOF);
      if (lhs->IsNotGreater(rhs))    flags.SetKnown1(FLAGS_MASK_ZForSFneqOF);
      if (lhs->IsGreater(rhs))	     flags.SetKnown0(FLAGS_MASK_ZForSFneqOF);
    }
  
  // TODO: calc AF and OF flag if possible
  
  return SetFlagsForRangeSZP(flags,
			     new VariableNRange<unsignedT>(lhs->m_min - rhs->m_value,
							   lhs->m_max - rhs->m_value));
} 

template<class unsignedT>
VariableNRange<unsignedT>* InvSub(const VariableNRange<unsignedT>* rhs, 
				  const VariableNConst<unsignedT>* lhs,
				  StatusFlagsAffected& flags)
{
  SetFlagsUnknownAddSubNeg(flags);
  
  if (rhs->IsNormalRange())
    {
      if (rhs->IsAbove(lhs))	     flags.SetKnown1(FLAGS_MASK_CF);
      if (rhs->IsNotAbove(lhs))	     flags.SetKnown0(FLAGS_MASK_CF);
      if (rhs->IsNotBelow(lhs))	     flags.SetKnown1(FLAGS_MASK_CForZF);
      if (rhs->IsBelow(lhs))	     flags.SetKnown0(FLAGS_MASK_CForZF);
    }				     
  if (rhs->IsSignedRange())	     
    {				     
      if (rhs->IsGreater(lhs))	     flags.SetKnown1(FLAGS_MASK_SFneqOF);
      if (rhs->IsNotGreater(lhs))    flags.SetKnown0(FLAGS_MASK_SFneqOF);
      if (rhs->IsNotLess(lhs))	     flags.SetKnown1(FLAGS_MASK_ZForSFneqOF);
      if (rhs->IsLess(lhs))	     flags.SetKnown0(FLAGS_MASK_ZForSFneqOF);
    }
  
  // TODO: calc AF and OF flag if possible
  
  return SetFlagsForRangeSZP(flags,
			     new VariableNRange<unsignedT>(lhs->m_value - rhs->m_max,
							   lhs->m_value - rhs->m_min));
} 

//////////////////////////////////////////////////////////////////////
// RANGE_RANGE

template<class unsignedT>
VariableNRange<unsignedT>* Add(const VariableNRange<unsignedT>* lhs,
			       const VariableNRange<unsignedT>* rhs, 
			       StatusFlagsAffected& flags)
{
  SetFlagsUnknownAddSubNeg(flags);
  
  if (lhs->IsInside (lhs->m_max + (rhs->m_max - rhs->m_min + 1)))
    return NULL; // spans more than one range (i.e. could be anything)
  
  if (lhs->IsNormalRange() &&
      rhs->IsNegNormalRange())
    {
      if (lhs->IsBelowNeg(rhs))	     flags.SetKnown1(FLAGS_MASK_CF);
      if (lhs->IsNotBelowNeg(rhs))   flags.SetKnown0(FLAGS_MASK_CF);
      if (lhs->IsNotAboveNeg(rhs))   flags.SetKnown1(FLAGS_MASK_CForZF);
      if (lhs->IsAboveNeg(rhs))	     flags.SetKnown0(FLAGS_MASK_CForZF);
    }				     
  if (lhs->IsSignedRange() &&	     
      rhs->IsNegSignedRange())	     
    {				     
      if (lhs->IsLessNeg(rhs))	     flags.SetKnown1(FLAGS_MASK_SFneqOF);
      if (lhs->IsNotLessNeg(rhs))    flags.SetKnown0(FLAGS_MASK_SFneqOF);
      if (lhs->IsNotGreaterNeg(rhs)) flags.SetKnown1(FLAGS_MASK_ZForSFneqOF);
      if (lhs->IsGreaterNeg(rhs))    flags.SetKnown0(FLAGS_MASK_ZForSFneqOF);
    }
  
  // TODO: calc AF and OF flag if possible
  
  return SetFlagsForRangeSZP(flags,
			     new VariableNRange<unsignedT>(lhs->m_min + rhs->m_min,
							   lhs->m_max + rhs->m_max));
  
  
  /*
        unsignedT min = lhs->m_min + rhs->min;
        unsignedT max = lhs->m_max + rhs->max;
        
        
        if ((lhs->m_max - lhs->m_min) >> 1) + ((rhs->m_max - rhs->m_min) >> 1)

  return new VariableNRange<unsignedT>(lhs->m_value + rhs->m_min,
                                                                           lhs->m_value + rhs->m_max);
                                                                           
  */
} 

template<class unsignedT>
VariableNRange<unsignedT>* Sub(const VariableNRange<unsignedT>* lhs,
			       const VariableNRange<unsignedT>* rhs, 
			       StatusFlagsAffected& flags)
{
  SetFlagsUnknownAddSubNeg(flags);

  if (lhs->IsInside (lhs->m_max + (rhs->m_max - rhs->m_min + 1)))
    return NULL; // spans more than one range (i.e. could be anything)
  
  if (lhs->IsNormalRange() &&
      rhs->IsNormalRange())
    {
      if (lhs->IsBelow(rhs))	     flags.SetKnown1(FLAGS_MASK_CF);
      if (lhs->IsNotBelow(rhs))	     flags.SetKnown0(FLAGS_MASK_CF);
      if (lhs->IsNotAbove(rhs))	     flags.SetKnown1(FLAGS_MASK_CForZF);
      if (lhs->IsAbove(rhs))	     flags.SetKnown0(FLAGS_MASK_CForZF);
    }				     
  if (lhs->IsSignedRange() &&	     
      rhs->IsSignedRange())	     
    {				     
      if (lhs->IsLess(rhs))	     flags.SetKnown1(FLAGS_MASK_SFneqOF);
      if (lhs->IsNotLess(rhs))	     flags.SetKnown0(FLAGS_MASK_SFneqOF);
      if (lhs->IsNotGreater(rhs))    flags.SetKnown1(FLAGS_MASK_ZForSFneqOF);
      if (lhs->IsGreater(rhs))	     flags.SetKnown0(FLAGS_MASK_ZForSFneqOF);
    }
  
  // TODO: calc AF and OF flag if possible
  
  return SetFlagsForRangeSZP(flags,
			     new VariableNRange<unsignedT>(lhs->m_min - rhs->m_max,
							   lhs->m_max - rhs->m_min));
  
  /*
    return new VariableNRange<unsignedT>(lhs->m_value - rhs->m_max,
                                                                           lhs->m_value - rhs->m_min);
  */
} 

template<class unsignedT>
VariableNRange<unsignedT>* Inc(const VariableNRange<unsignedT>* lhs, 
			       StatusFlagsAffected& flags)
{
  SetFlagsUnknownIncDec(flags);
  
  if (lhs->IsSignedRange())
    {
      if (lhs->IsLessNeg0())	     flags.SetKnown1(FLAGS_MASK_SFneqOF);
      if (lhs->IsNotLessNeg0())	     flags.SetKnown0(FLAGS_MASK_SFneqOF);
      if (lhs->IsNotGreaterNeg0())   flags.SetKnown1(FLAGS_MASK_ZForSFneqOF);
      if (lhs->IsGreaterNeg0())	     flags.SetKnown0(FLAGS_MASK_ZForSFneqOF);
    }
  
  // TODO: calc AF and OF flag if possible
  
  return SetFlagsForRangeSZP(flags,
			     new VariableNRange<unsignedT>(lhs->m_min + 1,
							   lhs->m_max + 1));
} 


template<class unsignedT>
VariableNRange<unsignedT>* Dec(const VariableNRange<unsignedT>* lhs, 
			       StatusFlagsAffected& flags)
{
  SetFlagsUnknownIncDec(flags);
  
  if (lhs->IsSignedRange())
    {
      if (lhs->IsLess0())	     flags.SetKnown1(FLAGS_MASK_SFneqOF);
      if (lhs->IsNotLess0())	     flags.SetKnown0(FLAGS_MASK_SFneqOF);
      if (lhs->IsNotGreater0())	     flags.SetKnown1(FLAGS_MASK_ZForSFneqOF);
      if (lhs->IsGreater0())	     flags.SetKnown0(FLAGS_MASK_ZForSFneqOF);
    }
  
  // TODO: calc AF and OF flag if possible
  
  return SetFlagsForRangeSZP(flags,
			     new VariableNRange<unsignedT>(lhs->m_min - 1,
							   lhs->m_max - 1));
} 

template<class unsignedT>
VariableNRange<unsignedT>* Neg(const VariableNRange<unsignedT>* lhs, 
			       StatusFlagsAffected& flags)
{
  SetFlagsUnknownAddSubNeg(flags);
  
  if (lhs->IsSignedRange())
    {
      if (lhs->IsGreater0())	    flags.SetKnown1(FLAGS_MASK_SFneqOF);
      if (lhs->IsNotGreater0())	    flags.SetKnown0(FLAGS_MASK_SFneqOF);
      if (lhs->IsNotLess0())	    flags.SetKnown1(FLAGS_MASK_ZForSFneqOF);
      if (lhs->IsLess0())	    flags.SetKnown0(FLAGS_MASK_ZForSFneqOF);
    }
  
  if (lhs->m_min == 0 && lhs->m_max == 0)
    flags.SetKnown0(FLAGS_MASK_CF);
  else if (!lhs->IsInside(0))
    flags.SetKnown1(FLAGS_MASK_CF);
  
  // TODO: calc AF and OF flag if possible
  
  return SetFlagsForRangeSZP(flags,
			     new VariableNRange<unsignedT>(-lhs->m_max,
							   -lhs->m_min));
  /*
2    return new VariableNRange<unsignedT>(lhs->m_value - rhs->m_max,
    lhs->m_value - rhs->m_min);
  */
} 


/*
  #define V_AR_INT_CONST_RANGE_OP(name) \
  template<class unsignedT> \
  VariableNRange<unsignedT>* name(const VariableNConst<unsignedT>* lhs, \
  const VariableNRange<unsignedT>* rhs) \
  { \
  return new VariableNRange<unsignedT>(do_##name(lhs->m_value,rhs->m_min), \
  do_##name(lhs->m_value,rhs->m_max)); \
  } 
  
  V_AR_INT_CONST_RANGE_OP(Add)		// careful if using this macro for other operator
  V_AR_INT_CONST_RANGE_OP(Sub)		// careful if using this macro for other operator
  V_AR_INT_CONST_RANGE_OP(InvSub)	// careful if using this macro for other operator
*/
/*	
	value_utype m_min; // inclusive
	value_utype m_max; // inclusive
	
	// if m_min > m_max, we mean the range that wraps (which is useful for signed numbers)
	// m_min may not be same as m_max, which would either mean only one, or entire range
	// and they are better encoded as a const or an entire range
*/

//////////////////////////////////////////////////////////////////////
// CONST_PATTERN

// template<class unsignedT>
// VariableNPattern<unsignedT>* And_NOT_IN_USE(const VariableNConst<unsignedT>* lhs,
// 				 const VariableNPattern<unsignedT>* rhs, 
// 				 StatusFlagsAffected& flags)
// {
//   return new VariableNPattern<unsignedT>((( rhs->m_known) |  // all known by both
// 					  (~lhs->m_value) | // all made zero by lhs
// 					  (rhs->m_known & ~rhs->m_value)), // all made zero by rhs
// 					 lhs->m_value & rhs->m_value);
// } 
// 
// template<class unsignedT>
// VariableNPattern<unsignedT>* Or_NOT_IN_USE(const VariableNConst<unsignedT>* lhs,
// 				const VariableNPattern<unsignedT>* rhs, 
// 				StatusFlagsAffected& flags)
// {
//   return new VariableNPattern<unsignedT>(((rhs->m_known) | // all known by both
// 					  (lhs->m_value) | // all made one by lhs
// 					  (rhs->m_known & rhs->m_value)), // all made one by rhs
// 					 lhs->m_value | rhs->m_value);
// } 
// 
// template<class unsignedT>
// VariableNPattern<unsignedT>* Xor_NOT_IN_USE(const VariableNConst<unsignedT>* lhs,
// 				 const VariableNPattern<unsignedT>* rhs, 
// 				 StatusFlagsAffected& flags)
// {
//   return new VariableNPattern<unsignedT>(rhs->m_known, // all known by both
// 					 (lhs->m_value ^ rhs->m_value) & (rhs->m_known));
// } 

//////////////////////////////////////////////////////////////////////
// PATTERN_PATTERN

#define RESULT_0    (0x00)
#define RESULT_1    (0x04)
#define RESULT_U    (0x08) // unknown
#define CARRY_0     (0x00)
#define CARRY_1     (0x01)
#define CARRY_U     (0x02) // unknown

#define CARRY_MASK  (0x03)

#define BORROW_0    CARRY_0
#define BORROW_1    CARRY_1
#define BORROW_U    CARRY_U

// left three numbers are the combinations available (sorted (no duplicates))
// the two numbers are r/c, the resultant bit and carry
// adding is symmetric in the two operands)

// 0 0 0   0/0
// 0 0 1   1/0
// 0 0 ?   ?/0   (0/0,1/0)
// 0 1 1   0/1
// 0 1 ?   ?/?   (1/0,0/1)
// 0 ? ?   ?/?   (0/0,1/0,0/1)
// 1 1 1   1/1
// 1 1 ?   ?/1   (0/1,1/1)
// 1 ? ?   ?/?   (1/0,0,1,1/1)
// ? ? ?   ?/?   (0/0,1/0,0/1,1/1)

#define CR_000   (RESULT_0 | CARRY_0)
#define CR_001   (RESULT_1 | CARRY_0)
#define CR_00U   (RESULT_U | CARRY_0)
#define CR_011   (RESULT_0 | CARRY_1)
#define CR_01U   (RESULT_U | CARRY_U)
#define CR_0UU   (RESULT_U | CARRY_U)
#define CR_111   (RESULT_1 | CARRY_1)
#define CR_11U   (RESULT_U | CARRY_1)
#define CR_1UU   (RESULT_U | CARRY_U)
#define CR_UUU   (RESULT_U | CARRY_U)

// the first number is the first operand, 
// the second is the second operand (or perhaps borrow), 
// and the third is borrow (or perhaps second operand)
// subtraction for each bit is symmetric in second operand and borrow, 
// but not in first operand and the other two (scnd and borrow)

#define BR_000   (RESULT_0 | CARRY_0)
#define BR_001   (RESULT_1 | CARRY_1)
#define BR_00U   (RESULT_U | CARRY_U)
#define BR_011   (RESULT_0 | CARRY_1)
#define BR_01U   (RESULT_U | CARRY_U)
#define BR_0UU   (RESULT_U | CARRY_1)

#define BR_100   (RESULT_1 | CARRY_0)
#define BR_101   (RESULT_0 | CARRY_0)
#define BR_10U   (RESULT_U | CARRY_0)
#define BR_111   (RESULT_1 | CARRY_1)
#define BR_11U   (RESULT_U | CARRY_U)
#define BR_1UU   (RESULT_U | CARRY_U)

#define BR_U00   (RESULT_U | CARRY_0)
#define BR_U01   (RESULT_U | CARRY_U)
#define BR_U0U   (RESULT_U | CARRY_U)
#define BR_U11   (RESULT_U | CARRY_1)
#define BR_U1U   (RESULT_U | CARRY_U)
#define BR_UUU   (RESULT_U | CARRY_U)

#define BR_010 BR_U01
#define BR_0U0 BR_U0U
#define BR_0U1 BR_U1U

#define BR_110 BR_U01
#define BR_1U0 BR_U0U
#define BR_1U1 BR_U1U

#define BR_U10 BR_U01
#define BR_UU0 BR_U0U
#define BR_UU1 BR_U1U

byte bAddBitsCarryResultCarry[] =
{
  // CR_000, CR_001, CR_00U, CR_010, CR_011, CR_01U, CR_0U0, CR_0U1, CR_0UU,
  // CR_100, CR_101, CR_10U, CR_110, CR_111, CR_11U, CR_1U0, CR_1U1, CR_1UU,
  // CR_U00, CR_U01, CR_U0U, CR_U10, CR_U11, CR_U1U, CR_UU0, CR_UU1, CR_UUU,
  
  CR_000, CR_001, CR_00U, CR_001, CR_011, CR_01U, CR_00U, CR_01U, CR_0UU,
  CR_001, CR_011, CR_01U, CR_011, CR_111, CR_11U, CR_01U, CR_11U, CR_1UU,
  CR_00U, CR_01U, CR_0UU, CR_01U, CR_11U, CR_1UU, CR_0UU, CR_1UU, CR_UUU,
};

byte bSubBitsBorrowResultBorrow[] =
{
  BR_000, BR_001, BR_00U, BR_010, BR_011, BR_01U, BR_0U0, BR_0U1, BR_0UU,
  BR_100, BR_101, BR_10U, BR_110, BR_111, BR_11U, BR_1U0, BR_1U1, BR_1UU,
  BR_U00, BR_U01, BR_U0U, BR_U10, BR_U11, BR_U1U, BR_UU0, BR_UU1, BR_UUU,
};

/*
dword GetResultBitAndCarry(dword bit1,dword bit2,dword carry)
{
  // perhaps we should have used an table instead??
  // faster!, and probably smaller

  

  bit1  = bit1  ? 0x01010101 : 0;
  bit2  = bit2  ? 0x00110011 : 0;
  carry = carry ? 0x00001111 : 0;

  dword result = bit1 + bit2 + carry;

  dword bitknow   = (result & 0x11111111) != (~result & 0x11111111); 
  dword bit       =  result & 0x11111111;
  dword carryknow = (result & 0x22222222) != (~result & 0x22222222); 
  dword carry     =  result & 0x22222222;
}
*/


template<class unsignedT>
VariableNPattern<unsignedT>* SetFlagsForPatternSZP(StatusFlagsAffected& flags,
						   VariableNPattern<unsignedT>* pattern)
{
  if (pattern->m_known & MOST_SIGNIFICANT_ONE(unsignedT))
    flags.SetKnown(FLAGS_MASK_SF,!!(pattern->m_value & MOST_SIGNIFICANT_ONE(unsignedT)));
  
  if (pattern->m_value)
    flags.SetKnown0(FLAGS_MASK_ZF); // someone nonzero
  else if (pattern->m_known == ALL_ONES(unsignedT))
    flags.SetKnown1(FLAGS_MASK_ZF); // no unknowns (and all zero)
  
  if (((byte) pattern->m_known) == 0xff)
    flags.SetKnown(FLAGS_MASK_PF,parity((byte) pattern->m_value));
  
  return pattern;
}

// adding two patterns, it's hard to know what will come out,
// because of carry from each bit position. We may however know
// this: As far as the lowest bits are known, it can be calculated
// what happen to them, and how carries occur.

// and then for each bit pair this will be the result, and carry to
// the next bit (r/c) we have info from three places: operand a,
// operand b, and the carry from last
//
// 0 0 0
// 0 0 1
// 0 0 ?
// 0 1 0
// 0 1 1
// 0 1 ?
// 0 ? 0
// 0 ? 1
// 0 ? ?
// 1 0 0
// 1 0 1
// 1 0 ?
// 1 1 0
// 1 1 1
// 1 1 ?
// 1 ? 0
// 1 ? 1
// 1 ? ?
// ? 0 0
// ? 0 1
// ? 0 ?
// ? 1 0
// ? 1 1
// ? 1 ?
// ? ? 0
// ? ? 1
// ? ? ?

// left three numbers are the combinations available (sorted (no
// duplicates)) the two numbers are r/c, the resultant bit and carry

// 0 0 0   0/0
// 0 0 1   1/0
// 0 0 ?   ?/0   (0/0,1/0)
// 0 1 1   0/1
// 0 1 ?   ?/?   (1/0,0/1)
// 0 ? ?   ?/?   (0/0,1/0,0/1)
// 1 1 1   1/1
// 1 1 ?   ?/1   (0/1,1/1)
// 1 ? ?   ?/?   (1/0,0,1,1/1)
// ? ? ?   ?/?   (0/0,1/0,0/1,1/1)

// what we do is that we add the three bits together (at eight
// different places, and if the operand was zero, then we use zero,
// and if the operand was 1 then we use one, and if the operand was ?,
// then we use patterns of zero and one (in the different places), to
// simulate both possibilities



template<class unsignedT>
VariableNPattern<unsignedT>* AddSubPattern(const VariableNPattern<unsignedT>* lhs,
					   const VariableNPattern<unsignedT>* rhs,
					   StatusFlagsAffected& flags,
					   byte *bBitsResultCarryBorrow)
{
  SetFlagsUnknownAddSubNeg(flags);
  
  assert (!((~lhs->m_known) & lhs->m_value)); // so value is 0 where known is  
  assert (!((~rhs->m_known) & rhs->m_value)); // so value is 0 where known is
  
  unsignedT l_val     =  lhs->m_value;
  unsignedT l_notknow = ~lhs->m_known;
  unsignedT r_val     =  rhs->m_value;
  unsignedT r_notknow = ~rhs->m_known;
  
  byte carry = 0; // known and zero (0 for 0, 1 for 1, 2 for unknown)
  
  unsignedT value    = 0;
  unsignedT notknown = 0;
  
  for (unsignedT insert_bit = 1; insert_bit; insert_bit <<= 1)
    {
      byte bit1 = (l_val & 1) + 2 * (l_notknow & 1);
      byte bit2 = (r_val & 1) + 2 * (r_notknow & 1);
    
      byte bResultAndCarry = 
	bBitsResultCarryBorrow[bit1 * 9 + bit2 * 3 + carry];
      
      value    = (bResultAndCarry & RESULT_1) ? insert_bit : 0;
      notknown = (bResultAndCarry & RESULT_U) ? insert_bit : 0;
      
      carry    = bResultAndCarry & CARRY_MASK;
      
      l_val     >>= 1;
      l_notknow >>= 1;
      r_val     >>= 1;
      r_notknow >>= 1;
      
      if (insert_bit == 0x80) // carry sets auxilary flag
	{
	  if (!(carry & CARRY_U))
	    flags.SetKnown(FLAGS_MASK_AF,!!(carry & CARRY_1));
	}
      else if (insert_bit == (MOST_SIGNIFICANT_ONE(unsignedT) >> 1))
	{
	  if (!(carry & CARRY_U)) // carry sets overflow flag
	    flags.SetKnown(FLAGS_MASK_OF,!!(carry & CARRY_1));
	}
      
    }
  
  // carry     (      from last bit)
  // auxillary (carry from 4th bit)
  // overflow  (      from signed result, last-1 bit)
  // zero
  // parity
  // sign

  if (!(carry & CARRY_U))
    flags.SetKnown(FLAGS_MASK_CF,!!(carry & CARRY_1));

  VariableNPattern<unsignedT>* pRet;

  if (notknown != ALL_ONES(unsignedT))
    {
      pRet = new VariableNPattern<unsignedT>(~notknown,value);
      SetFlagsForPatternSZP(flags,pRet);
    }

  flags.CombinedFromCZSO();
  
  return pRet;
}

template<class unsignedT>
VariableNPattern<unsignedT>* IncDecPattern(const VariableNPattern<unsignedT>* lhs,
					   StatusFlagsAffected& flags,
					   byte *bBitsResultCarryBorrow)
{
  SetFlagsUnknownIncDec(flags);
  
  assert (!((~lhs->m_known) & lhs->m_value)); // so value is 0 where known is  
  
  unsignedT l_val     =  lhs->m_value;
  unsignedT l_notknow = ~lhs->m_known;
  unsignedT r_val     =  1;
  // unsignedT r_notknow =  0;

  byte carry = 0; // known and zero (0 for 0, 1 for 1, 2 for unknown)

  unsignedT value    = 0;
  unsignedT notknown = 0;

  for (unsignedT insert_bit = 1; insert_bit; insert_bit <<= 1)
  {
    byte bit1 = (l_val & 1) + 2 * (l_notknow & 1);
    byte bit2 = (r_val & 1) /* + 2 * (r_notknow & 1) */;
    
    byte bResultAndCarry = bBitsResultCarryBorrow[bit1 * 9 + bit2 * 3 + carry];
    
    value    = (bResultAndCarry & RESULT_1) ? insert_bit : 0;
    notknown = (bResultAndCarry & RESULT_U) ? insert_bit : 0;
    
    carry    = bResultAndCarry & CARRY_MASK;
    
    l_val     >>= 1;
    l_notknow >>= 1;
    r_val     >>= 1;
    /* r_notknow >>= 1; */
    
    if (insert_bit == 0x80) // carry sets auxilary flag
      {
	if (!(carry & CARRY_U))
	  flags.SetKnown(FLAGS_MASK_AF,!!(carry & CARRY_1));
      }
    else if (insert_bit == (MOST_SIGNIFICANT_ONE(unsignedT) >> 1))
      {
	if (!(carry & CARRY_U)) // carry sets overflow flag
	  flags.SetKnown(FLAGS_MASK_OF,!!(carry & CARRY_1));
      }
    
  }
  
  VariableNPattern<unsignedT>* pRet;
  
  if (notknown != ALL_ONES(unsignedT))
    {
      pRet = new VariableNPattern<unsignedT>(~notknown,value);
      SetFlagsForPatternSZP(flags,pRet);
    }

  flags.CombinedFromZSO();
  
  return pRet;
}

template<class unsignedT>
VariableNPattern<unsignedT>* Add(const VariableNPattern<unsignedT>* lhs,
				 const VariableNPattern<unsignedT>* rhs,
				 StatusFlagsAffected& flags)
{
  return AddSubPattern(lhs,rhs,flags,bAddBitsCarryResultCarry);
} 

template<class unsignedT>
VariableNPattern<unsignedT>* Sub(const VariableNPattern<unsignedT>* lhs,
				 const VariableNPattern<unsignedT>* rhs,
				 StatusFlagsAffected& flags)
{
  return AddSubPattern(lhs,rhs,flags,bSubBitsBorrowResultBorrow);
} 

template<class unsignedT>
VariableNPattern<unsignedT>* InvSub(const VariableNPattern<unsignedT>* lhs,
				 const VariableNPattern<unsignedT>* rhs,
				 StatusFlagsAffected& flags)
{
  return AddSubPattern(rhs,lhs,flags,bSubBitsBorrowResultBorrow);
} 

template<class unsignedT>
VariableNPattern<unsignedT>* Inc(const VariableNPattern<unsignedT>* lhs,
				 StatusFlagsAffected& flags)
{
  return IncDecPattern(lhs,flags,bAddBitsCarryResultCarry);
} 

template<class unsignedT>
VariableNPattern<unsignedT>* Dec(const VariableNPattern<unsignedT>* lhs,
				 StatusFlagsAffected& flags)
{
  return IncDecPattern(lhs,flags,bSubBitsBorrowResultBorrow);
} 


template<class unsignedT>
VariableNPattern<unsignedT>* SetFlagsForPatternAndOrXor(StatusFlagsAffected& flags,
							VariableNPattern<unsignedT>* pattern)
{
  SetFlagsUnknownAndOrXor(flags);
  
  if (pattern->m_known == ALL_ONES(unsignedT))
    {
      bool bZero = !pattern->m_value;
      flags.SetKnown(FLAGS_MASK_ZF | FLAGS_MASK_CForZF,bZero);
      if (bZero)
	flags.SetKnown1(FLAGS_MASK_ZForSFneqOF);
    }
  else if (pattern->m_value) // any known 1 ?, then cant be 0
    flags.SetKnown0(FLAGS_MASK_ZF | FLAGS_MASK_CForZF);
  if (pattern->m_known & MOST_SIGNIFICANT_ONE(unsignedT))
    {
      bool bSign = !!(pattern->m_value & MOST_SIGNIFICANT_ONE(unsignedT));
      flags.SetKnown(FLAGS_MASK_SF | FLAGS_MASK_SFneqOF,bSign);
      if (bSign)
	flags.SetKnown1(FLAGS_MASK_ZForSFneqOF);
    }
  if ((pattern->m_known & ~MOST_SIGNIFICANT_ONE(unsignedT)) &&
      !pattern->m_value) // either 0000 or 1000, so either ZF or SF, and OF = 0
    flags.SetKnown1(FLAGS_MASK_ZForSFneqOF);
  
  if (!flags.IsKnown(FLAGS_MASK_ZForSFneqOF) &&
		flags.AreKnown(FLAGS_MASK_ZF | FLAGS_MASK_SF))
    {
      assert (!flags.GetFlagsOr(FLAGS_MASK_ZF | FLAGS_MASK_SF));
      flags.SetKnown0(FLAGS_MASK_ZForSFneqOF);
    }
  
  if (((byte) pattern->m_known) == 0xff)
    flags.SetKnown(FLAGS_MASK_PF,parity((byte) pattern->m_value));
  
  return pattern;
}

template<class unsignedT>
VariableNPattern<unsignedT>* And(const VariableNPattern<unsignedT>* lhs,
				 const VariableNPattern<unsignedT>* rhs, 
				 StatusFlagsAffected& flags)
{
  return SetFlagsForPatternAndOrXor(flags,
				    new VariableNPattern<unsignedT>(((lhs->m_known & rhs->m_known) |  // all known by both
								     (lhs->m_known & ~lhs->m_value) | // all made zero by lhs
                                          (rhs->m_known & ~rhs->m_value)), // all made zero by rhs
								    lhs->m_value & rhs->m_value));
} 

template<class unsignedT>
VariableNPattern<unsignedT>* Or(const VariableNPattern<unsignedT>* lhs,
				const VariableNPattern<unsignedT>* rhs, 
				StatusFlagsAffected& flags)
{
  return SetFlagsForPatternAndOrXor(flags,
				    new VariableNPattern<unsignedT>(((lhs->m_known & rhs->m_known) | // all known by both
								     (lhs->m_known & lhs->m_value) | // all made one by lhs
								     (rhs->m_known & rhs->m_value)), // all made one by rhs
								    lhs->m_value | rhs->m_value));
} 

template<class unsignedT>
VariableNPattern<unsignedT>* Xor(const VariableNPattern<unsignedT>* lhs,
				 const VariableNPattern<unsignedT>* rhs, 
				 StatusFlagsAffected& flags)
{
  unsignedT known = lhs->m_known & rhs->m_known;
  return SetFlagsForPatternAndOrXor(flags,
				    new VariableNPattern<unsignedT>(known, // all known by both
								    (lhs->m_value ^ rhs->m_value) & (known)));
} 

template<class unsignedT>
VariableNPattern<unsignedT>* Not(const VariableNPattern<unsignedT>* lhs, 
				 StatusFlagsAffected& /* flags */)
{
  // flags not affected
  return new VariableNPattern<unsignedT>(lhs->m_known,~(lhs->m_value));
} 

// binary operator, but one of the operands may be unknown
template<class unsignedT>
VariableNPattern<unsignedT>* And(const VariableNPattern<unsignedT>* lrhs/*,ANY_PARAMETER*/,
				 StatusFlagsAffected& flags)
{

  return SetFlagsForPatternAndOrXor(flags,
				    new VariableNPattern<unsignedT>(lrhs->m_known & ~lrhs->m_value, // all that are zero will be zero (when anded with 0 or 1)
								    0));
} 

// binary operator, but one of the operands may be unknown
template<class unsignedT>
VariableNPattern<unsignedT>* Or(const VariableNPattern<unsignedT>* lrhs/*,ANY_PARAMETER*/,
				StatusFlagsAffected& flags)
{
  assert ((lrhs->m_value | lrhs->m_known) == lrhs->m_known); // if value is 1 for a bit, then it must be known
  
  
  return SetFlagsForPatternAndOrXor(flags,
				    new VariableNPattern<unsignedT>(lrhs->m_value, // all that are one will be one (when ored with 0 or 1)
								    lrhs->m_value));
} 


/*
#define V_AR_INT_CONST_PATTERN_OP(name) \
template<class unsignedT> \
VariableNRange<unsignedT>* name(const VariableNConst<unsignedT>* lhs, \
			        const VariableNPattern<unsignedT>* rhs) \
{ \
  return new VariableNPattern<unsignedT>(do_##name(lhs->m_value,rhs->m_min), \
                                                                           do_##name(lhs->m_value,rhs->m_max)); \
} 

V_AR_INT_CONST_PATTERN_OP(And)	// careful if using this macro for other operator
V_AR_INT_CONST_PATTERN_OP(Or)	// careful if using this macro for other operator
V_AR_INT_CONST_PATTERN_OP(Xor)	// careful if using this macro for other operator
*/
/*
 	value_utype m_known; // bit pattern of known bits
        value_utype m_value; // the values of the known bits

        // known could for ex be 0xf0 and value be 0x00, in which case we know
        // that the variable is 0x0? (the four high bits known to be zero)
*/

#include "gen/VariableOp_OtherOpFcns.h" // dep_chk_unknown  // autogenerated from this file

//////////////////////////////////////////////////////////////////////

#define GENERATE_VAR_INT_BINARY_NOT_ORDERED_OP(name)     // rettype=VariableN<unsignedT>* argtype=const VariableN<unsignedT>* extraargs=StatusFlagsAffected& flags funcnameprefix=VariableOpCalcResult invops=1 
#define GENERATE_VAR_INT_BINARY_ORDERED_OP(name,invname) // rettype=VariableN<unsignedT>* argtype=const VariableN<unsignedT>* extraargs=StatusFlagsAffected& flags funcnameprefix=VariableOpCalcResult invops=1 
#define GENERATE_VAR_INT_UNARY_OP(name)                  // rettype=VariableN<unsignedT>* argtype=const VariableN<unsignedT>* extraargs=StatusFlagsAffected& flags funcnameprefix=VariableOpCalcResult 

GENERATE_VAR_INT_BINARY_NOT_ORDERED_OP(Add)        // defret=SetFlagsUnknownAddSubNeg(flags); return NULL; 
GENERATE_VAR_INT_BINARY_ORDERED_OP    (Sub,InvSub) // defret=SetFlagsUnknownAddSubNeg(flags); return NULL; 
GENERATE_VAR_INT_BINARY_NOT_ORDERED_OP(And)        // defret=SetFlagsUnknownAndOrXor(flags); return NULL; 
GENERATE_VAR_INT_BINARY_NOT_ORDERED_OP(Or)         // defret=SetFlagsUnknownAndOrXor(flags); return NULL;
GENERATE_VAR_INT_BINARY_NOT_ORDERED_OP(Xor)        // defret=SetFlagsUnknownAndOrXor(flags); return NULL;
GENERATE_VAR_INT_UNARY_OP(Neg)                     // defret=SetFlagsUnknownAddSubNeg(flags); return NULL;
GENERATE_VAR_INT_UNARY_OP(Not)                     // defret=return NULL;
GENERATE_VAR_INT_UNARY_OP(Inc)                     // defret=SetFlagsUnknownIncDec(flags); return NULL;
GENERATE_VAR_INT_UNARY_OP(Dec)                     // defret=SetFlagsUnknownIncDec(flags); return NULL;

#define GENERATE_VAR_INT_BINARY_NOT_ORDERED_OP_CONSTRAINED(name)     // rettype=word argtype=const VariableN<unsignedT>* extraargs=const StatusFlagsKnown& flags_known,VariableN<unsignedT>** out_lhs,VariableN<unsignedT>** out_rhs,VariableN<unsignedT>** out_result,StatusFlagsAffected& flags funcnameprefix=VariableOpCalcResult invops=1 
#define GENERATE_VAR_INT_BINARY_ORDERED_OP_CONSTRAINED(name,invname) // rettype=word argtype=const VariableN<unsignedT>* extraargs=const StatusFlagsKnown& flags_known,VariableN<unsignedT>** out_lhs,VariableN<unsignedT>** out_rhs,VariableN<unsignedT>** out_result,StatusFlagsAffected& flags funcnameprefix=VariableOpCalcResult invops=1 
#define GENERATE_VAR_INT_UNARY_OP_CONSTRAINED(name)                  // rettype=word argtype=const VariableN<unsignedT>* extraargs=const StatusFlagsKnown& flags_known,VariableN<unsignedT>** out_lhs,VariableN<unsignedT>** out_result,StatusFlagsAffected& flags funcnameprefix=VariableOpCalcResult

GENERATE_VAR_INT_BINARY_NOT_ORDERED_OP_CONSTRAINED(AddConstrained)        // defret=SetFlagsUnknownAddSubNeg(flags); return CONSTRAINED_NO_CONSTRAINTS; 
GENERATE_VAR_INT_BINARY_ORDERED_OP_CONSTRAINED    (SubConstrained,InvSubConstrained) // defret=SetFlagsUnknownAddSubNeg(flags); return CONSTRAINED_NO_CONSTRAINTS; 
GENERATE_VAR_INT_BINARY_NOT_ORDERED_OP_CONSTRAINED(AndConstrained)        // defret=SetFlagsUnknownAndOrXor(flags); return CONSTRAINED_NO_CONSTRAINTS; 
GENERATE_VAR_INT_BINARY_NOT_ORDERED_OP_CONSTRAINED(OrConstrained)         // defret=SetFlagsUnknownAndOrXor(flags); return CONSTRAINED_NO_CONSTRAINTS;
GENERATE_VAR_INT_BINARY_NOT_ORDERED_OP_CONSTRAINED(XorConstrained)        // defret=SetFlagsUnknownAndOrXor(flags); return CONSTRAINED_NO_CONSTRAINTS;
GENERATE_VAR_INT_UNARY_OP_CONSTRAINED(NegConstrained)                     // defret=SetFlagsUnknownAddSubNeg(flags); return CONSTRAINED_NO_CONSTRAINTS;
GENERATE_VAR_INT_UNARY_OP_CONSTRAINED(NotConstrained)                     // defret=return CONSTRAINED_NO_CONSTRAINTS;
GENERATE_VAR_INT_UNARY_OP_CONSTRAINED(IncConstrained)                     // defret=SetFlagsUnknownIncDec(flags); return CONSTRAINED_NO_CONSTRAINTS;
GENERATE_VAR_INT_UNARY_OP_CONSTRAINED(DecConstrained)                     // defret=SetFlagsUnknownIncDec(flags); return CONSTRAINED_NO_CONSTRAINTS;

//////////////////////////////////////////////////////////////////////

#include "gen/VariableOp_TypedCalls.h" // autogenerated from this file

//////////////////////////////////////////////////////////////////////

void test_var_ops()
{
  {
    Variable8* lhs = new Variable8Const(2);
    Variable8* rhs = new Variable8Const(3);
    
    StatusFlagsAffected flags;
    
    Variable8* result;
    result = VariableOpCalcResultAdd(lhs,rhs,flags);
    result = VariableOpCalcResultSub(lhs,rhs,flags);
    result = VariableOpCalcResultAnd(lhs,rhs,flags);
    result = VariableOpCalcResultOr (lhs,rhs,flags);
    result = VariableOpCalcResultXor(lhs,rhs,flags);
    
    result = VariableOpCalcResultNot(lhs,flags);
    result = VariableOpCalcResultNeg(lhs,flags);
  }
  /*
        {
                Variable8Consts* lhs2 = new Variable8Consts(4);
                Variable8Consts* rhs2 = new Variable8Consts(5);

                // Variable8* result = add(lhs,rhs);
        }
  */
/*
	VariableNBinaryOp<byte> op;

	op.DoSomething();
*/
}

//////////////////////////////////////////////////////////////////////

/* COPY_CALC_RESULT_FUNC @replace=({'Add'=>'Sub'},{'Add'=>'And'},{'Add'=>'Or'},{'Add'=>'Xor'}); */
template<class unsignedT> 
typename VariableNBinaryOpAdd<unsignedT>::var_type* 
VariableNBinaryOpAdd<unsignedT>::CalcResult (StatusFlagsAffected& flags) 
{ 
  return VariableOpCalcResultAdd(m_pSrc1,m_pSrc2,flags); 
} 

template<class unsignedT> 
word
VariableNBinaryOpAdd<unsignedT>::CalcResultConstrained (const StatusFlagsKnown& flags_known,  
							VariableN<unsignedT>** out_lhs,  
							VariableN<unsignedT>** out_rhs,  
							VariableN<unsignedT>** out_result,  
							StatusFlagsAffected& flags) 
{ 
  return VariableOpCalcResultAddConstrained(m_pSrc1,m_pSrc2,flags_known,out_lhs,out_rhs,out_result,flags); 
}
/* END_COPY_CALC_RESULT_FUNC */

/* COPY_CALC_RESULT_FUNC @replace=({'Neg'=>'Not'},{'Neg'=>'Inc'},{'Neg'=>'Dec'}); */
template<class unsignedT> 
typename VariableNUnaryOpNeg<unsignedT>::var_type* 
VariableNUnaryOpNeg<unsignedT>::CalcResult (StatusFlagsAffected& flags) 
{ 
  return VariableOpCalcResultNeg(m_pSrc,flags); 
}

template<class unsignedT> 
word
VariableNUnaryOpNeg<unsignedT>::CalcResultConstrained (const StatusFlagsKnown& flags_known,  
						       VariableN<unsignedT>** out_lhs,  
						       VariableN<unsignedT>** out_result,  
						       StatusFlagsAffected& flags) 
{ 
  return VariableOpCalcResultNegConstrained(m_pSrc,flags_known,out_lhs,out_result,flags); 
}
/* END_COPY_CALC_RESULT_FUNC */

#include "gen/VariableOp_OtherCalcResultFcns.h" // dep_chk_unknown  // autogenerated from this file

//////////////////////////////////////////////////////////////////////

template<class unsignedT>
VariableN<unsignedT>* CalcOpResultType (short nInstruction,VariableN<unsignedT>* pSrc1,VariableN<unsignedT>* pSrc2,
					StatusFlagsAffected& flags)
{
  VariableNBinaryOp<unsignedT>* pVarOp = NULL;
  
  switch (nInstruction)
    {
    case instrADD:
      pVarOp = new VariableNBinaryOpAdd<unsignedT> (pSrc1,pSrc2);
      break;
    case instrADC:
      pVarOp = new VariableNBinaryOpAdd<unsignedT> (pSrc1,pSrc2);
      break;
    case instrSUB:
    case instrCMP:
      pVarOp = new VariableNBinaryOpSub<unsignedT> (pSrc1,pSrc2);
      break;
    case instrSBB:
      pVarOp = new VariableNBinaryOpSub<unsignedT> (pSrc1,pSrc2);
      break;
    case instrOR:
      pVarOp = new VariableNBinaryOpOr<unsignedT> (pSrc1,pSrc2);
      break;
    case instrAND:
    case instrTEST:
      pVarOp = new VariableNBinaryOpAnd<unsignedT> (pSrc1,pSrc2);
      break;
    case instrXOR:
      pVarOp = new VariableNBinaryOpXor<unsignedT> (pSrc1,pSrc2);
      break;
#ifndef NDEBUG
    default:
      assert (false);
#endif
    }
  
  VariableN<unsignedT>* pResult = pVarOp->CalcResult(flags);
  
  if (!pResult)
    {
      /*
	delete pVarOp;
	return NULL;
      */
      
      // instead of deleting and returning a NULL variable,
      // we return an unknown value, so we can still trace how this happened
      
      pResult = new VariableNUnknown<unsignedT>;
    }
  
  pVarOp->AddDepend(pResult);
  pResult->SetSource(pVarOp);
  
  CalcOpVarContent(pResult,pVarOp);

  return pResult;

	/*
#define instrADD			(instrgINT_ARITH | 0x0000) 
#define instrOR				(instrgINT_ARITH | 0x0001) 
#define instrADC			(instrgINT_ARITH | 0x0002) 
#define instrSBB			(instrgINT_ARITH | 0x0003) 
#define instrAND			(instrgINT_ARITH | 0x0004) 
#define instrSUB			(instrgINT_ARITH | 0x0005) 
#define instrXOR			(instrgINT_ARITH | 0x0006) 
#define instrCMP			(instrgINT_ARITH | 0x0007) 

#define instrTEST			(instrgINT_ARITH | 0x0007) 

  
        VariableNBinaryOp<byte> op;
        op.DoSomething();

        VariableNBinaryOp<word> op2;
        op2.DoSomething();
*/
}

//////////////////////////////////////////////////////////////////////

template<class unsignedT>
VariableN<unsignedT>* CalcOpResultType (short nInstruction,
					VariableN<unsignedT>* pSrc1,
					StatusFlagsAffected& flags)
{
  VariableNUnaryOp<unsignedT>* pVarOp = NULL;
  
  switch (nInstruction)
    {
    case instrNOT:
      pVarOp = new VariableNUnaryOpNot<unsignedT> (pSrc1);
      break;
    case instrNEG:
      pVarOp = new VariableNUnaryOpNeg<unsignedT> (pSrc1);
      break;
    case instrINC:
      pVarOp = new VariableNUnaryOpInc<unsignedT> (pSrc1);
      break;
    case instrDEC:
      pVarOp = new VariableNUnaryOpDec<unsignedT> (pSrc1);
      break;
#ifndef NDEBUG
    default:
      assert (false);
#endif
    }
  
  VariableN<unsignedT>* pResult = pVarOp->CalcResult(flags);
  
  if (!pResult)
    {
      /*
	delete pVarOp;
	return NULL;
      */
      
      // instead of deleting and returning a NULL variable,
      // we return an unknown value, so we can still trace how this happened
      
      pResult = new VariableNUnknown<unsignedT>;
    }
  
  pVarOp->AddDepend(pResult);
  pResult->SetSource(pVarOp);
  
  CalcOpVarContent(pResult,pVarOp);

  return pResult;
}

//////////////////////////////////////////////////////////////////////
/*
template<class unsignedT>
void VariableNBinaryOp<unsignedT>::DoSomething ()
{
	printf ("DoSomething: %d\n",sizeof (unsignedT));

}
*/
//////////////////////////////////////////////////////////////////////

Variable* CalcOpResult (int nSize,
			short nInstruction,
			Variable* pSrc1,Variable* pSrc2,
			StatusFlagsAffected& flags)
{
  if (nSize == 1)
    {
      ASSERT_IS_TYPE_OR_NULL(Variable8*,pSrc1);
      ASSERT_IS_TYPE_OR_NULL(Variable8*,pSrc2);
      return CalcOpResultType(nInstruction,
			      (Variable8*) pSrc1,(Variable8*) pSrc2,flags);
    }
  else if (nSize == 2)
    {
      ASSERT_IS_TYPE_OR_NULL(Variable16*,pSrc1);
      ASSERT_IS_TYPE_OR_NULL(Variable16*,pSrc2);
      return CalcOpResultType(nInstruction,
			      (Variable16*) pSrc1,(Variable16*) pSrc2,flags);
    }
  else
    {
      assert (nSize == 4);
      ASSERT_IS_TYPE_OR_NULL(Variable32*,pSrc1);
      ASSERT_IS_TYPE_OR_NULL(Variable32*,pSrc2);
      return CalcOpResultType(nInstruction,
			      (Variable32*) pSrc1,(Variable32*) pSrc2,flags);
    }
}


Variable* CalcOpResult (int nSize,
			short nInstruction,
			Variable* pSrc1,
			StatusFlagsAffected& flags)
{
  if (nSize == 1)
    {
      ASSERT_IS_TYPE_OR_NULL(Variable8*,pSrc1);
      return CalcOpResultType(nInstruction,(Variable8*) pSrc1,flags);
    }
  else if (nSize == 2)
    {
      ASSERT_IS_TYPE_OR_NULL(Variable16*,pSrc1);
      return CalcOpResultType(nInstruction,(Variable16*) pSrc1,flags);
    }
  else
    {
      assert (nSize == 4);
      ASSERT_IS_TYPE_OR_NULL(Variable32*,pSrc1);
      return CalcOpResultType(nInstruction,(Variable32*) pSrc1,flags);
    }
}

Variable* CalcOpResultZeroIdenticalArgs (int nSize,
					 short nInstruction,
					 StatusFlagsAffected& flags)
{
  // return const(0)  for sub/cmp reg,same_reg  or xor reg,same_reg
  
  // set flags accordingly
  
  assert (nInstruction == instrSUB ||
	  nInstruction == instrCMP ||
	  nInstruction == instrXOR);
  
  if (nInstruction == instrXOR)
    {
      SetFlagsUnknownAndOrXor(flags);
      
      flags.SetKnown1(FLAGS_MASK_ZF | FLAGS_MASK_PF |
		      FLAGS_MASK_CForZF | FLAGS_MASK_ZForSFneqOF);
      flags.SetKnown0(FLAGS_MASK_SF |
		      FLAGS_MASK_SFneqOF);
    }
  else
    {
      assert (nInstruction == instrSUB ||
	      nInstruction == instrCMP);
      
      SetFlagsUnknownAddSubNeg(flags);
      
      flags.SetKnown1(FLAGS_MASK_ZF | FLAGS_MASK_PF |
		      FLAGS_MASK_CForZF | FLAGS_MASK_ZForSFneqOF);
      flags.SetKnown0(FLAGS_MASK_CF | FLAGS_MASK_AF | 
		      FLAGS_MASK_OF | FLAGS_MASK_SF |
		      FLAGS_MASK_SFneqOF);
    }
  
  return VariableNZero(nSize);
}

template<class unsignedT>
VariableN<unsignedT>* CalcResultOrUnknown (VariableNOp<unsignedT>* pVarOp)
{
  StatusFlagsAffected flags_dummy;
  
  VariableN<unsignedT>* pResult = pVarOp->CalcResult(flags_dummy);
  
  if (!pResult)
    pResult = new VariableNUnknown<unsignedT>;
  
  pVarOp->AddDepend(pResult);
  pResult->SetSource(pVarOp);

  CalcOpVarContent(pResult,pVarOp);
  
  return pResult;
}

Variable16* CalcMemAddr16 (Variable16* pVarBase,
			   Variable16* pVarIndex,
			   const OperandMemory* pOpMem)
{
  assert (!pVarBase  || pOpMem->m_bBaseReg);
  assert (!pVarIndex || pOpMem->m_bIndexReg);

  // TODO: do we have to do anything about the segment register????,
  // probably, and definately when in 16-bit mode...
  
  // add the stuff together
  
  if (!pOpMem->m_bBaseReg && !pOpMem->m_bIndexReg)
    return new Variable16Const((word) pOpMem->m_dwDisp);

  if (pOpMem->m_bBaseReg && pOpMem->m_bIndexReg)
    {
      pVarBase = 
	CalcResultOrUnknown(new VariableNBinaryOpAdd<word> (pVarBase,
							    pVarIndex));
    }
  else if (pOpMem->m_bIndexReg)
    pVarBase = pVarIndex;

  if (pOpMem->m_dwDisp)
    {
      // add the displacement
      
      pVarBase = 
	CalcResultOrUnknown(new VariableNBinaryOpAdd<word> (pVarBase,
							    new Variable16Const((word) pOpMem->m_dwDisp)));
    }
  
  return pVarBase;
}

Variable32* CalcMemAddr32 (Variable32* pVarBase,
			   Variable32* pVarIndex,
			   const  OperandMemory* pOpMem)
{
  assert (!pVarBase  || pOpMem->m_bBaseReg);
  assert (!pVarIndex || pOpMem->m_bIndexReg);

  if (pOpMem->m_bScale)
    {
      // index must be scaled, perform that
      
      // TODO: pVarIndex shift by pOpMem->m_bScale....
      
      pVarIndex = NULL;
    }
  
  // TODO: do we have to do anything about the segment register????,
  // probably, and definately when in 16-bit mode...

  // add the stuff together

  if (!pOpMem->m_bBaseReg && !pOpMem->m_bIndexReg)
    return new Variable32Const((dword) pOpMem->m_dwDisp);

  if (pOpMem->m_bBaseReg && pOpMem->m_bIndexReg)
    {
      pVarBase = 
	CalcResultOrUnknown(new VariableNBinaryOpAdd<dword> (pVarBase,
							     pVarIndex));
    }
  else if (pOpMem->m_bIndexReg)
    pVarBase = pVarIndex;
  
  if (pOpMem->m_dwDisp)
    {
      // add the displacement
      
      pVarBase = 
	CalcResultOrUnknown(new VariableNBinaryOpAdd<dword> (pVarBase,
							     new Variable32Const((dword) pOpMem->m_dwDisp)));
    }
  
  return pVarBase;
}

Variable* CalcMemAddr (Variable* pVarBase,
		       Variable* pVarIndex,
		       const OperandMemory* pOpMem)
{
  if (pOpMem->IsAddrSize32())
    {
      ASSERT_IS_TYPE_OR_NULL(Variable32*,pVarBase);
      ASSERT_IS_TYPE_OR_NULL(Variable32*,pVarIndex);
      return CalcMemAddr32((Variable32*) pVarBase,
			   (Variable32*) pVarIndex,pOpMem);
    }
  else
    {
      ASSERT_IS_TYPE_OR_NULL(Variable16*,pVarBase);
      ASSERT_IS_TYPE_OR_NULL(Variable16*,pVarIndex);
      return CalcMemAddr16((Variable16*) pVarBase,
			   (Variable16*) pVarIndex,pOpMem);
    }
}

Variable* CalcAddToVar(Variable* pVar,dword dwAdd,int nSize)
{
  if (nSize == 2)
    {
      ASSERT_IS_TYPE_OR_NULL(Variable16*,pVar);
      return CalcResultOrUnknown(new VariableNBinaryOpAdd<word> ((Variable16*) pVar,new Variable16Const((word) dwAdd)));
    }
  else
    {
      assert (nSize == 4);
      ASSERT_IS_TYPE_OR_NULL(Variable32*,pVar);
      return CalcResultOrUnknown(new VariableNBinaryOpAdd<dword> ((Variable32*) pVar,new Variable32Const((dword) dwAdd)));
    }
}













