/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// VariableOp.h: interface for the VariableOp class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(__VARIABLE_OP_H__INCLUDED_)
#define __VARIABLE_OP_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "VariableN.h" 
struct StatusFlagsAffected; 
struct StatusFlagsKnown;

class OperandMemory;

//////////////////////////////////////////////////////////////////////

// possible return values for a Constrain operation (i.e. one where
// part of the result is known (flags)

#define CONSTRAINED_UNREACHABLE        0    // arguments cannot give result
#define CONSTRAINED_NO_CONSTRAINTS     0x01 // no parameter constrained by result, reachable 

#define CONSTRAINED_LHS_NOT_SAME       0x02 // lhs    constrained by known result
#define CONSTRAINED_RHS_NOT_SAME       0x04 // rhs    constrained by known result
#define CONSTRAINED_RESULT_NOT_SAME    0x08 // result constrained by known result

#define CONSTRAINED_LHS_RESULT_NOT_SAME (CONSTRAINED_LHS_NOT_SAME | CONSTRAINED_RESULT_NOT_SAME)

// we only return replacement variables (lhs,rhs,result) if needed
// (i.e. constrained relative to behaviour without result known). But
// we do not return replacements if the entire result was unreachable,
// only return parameters masked by CONSTRAINED_any_NOT_SAME are set

//////////////////////////////////////////////////////////////////////

class VariableOp
{
 public:
  VariableOp();
  virtual ~VariableOp();
  
};

//////////////////////////////////////////////////////////////////////

template<class unsignedT>
class VariableNOp : public VariableNBase<unsignedT>, public VariableOp
{
 public:
  VARIABLE_N_BASE_TYPEDEF_VALUE_TYPES;
  VARIABLE_N_TYPEDEF_VAR_TYPES(VariableN  ,var_type);
  VARIABLE_N_TYPEDEF_VAR_TYPES(VariableNOp,var_type_op);
  
 public:
  typedef VariableN<unsignedT> var_type;
  /*
    public:
    VariableNOp();
    virtual ~VariableNOp();
  */
  
 public:
  virtual var_type* CalcResult (StatusFlagsAffected& flags) = 0;
};

template<class unsignedT>
class VariableNUnaryOp : public VariableNOp<unsignedT>
{
 public:
  VARIABLE_N_BASE_TYPEDEF_VALUE_TYPES;
  VARIABLE_N_TYPEDEF_VAR_TYPES(VariableN       ,var_type);
  VARIABLE_N_TYPEDEF_VAR_TYPES(VariableNUnaryOp,var_type_unary_op);
  
 public:
  VariableNUnaryOp(var_type* pSrc)
    {
      m_pSrc = pSrc;
      if (pSrc)
	pSrc->AddDepend(this);
    }
  
 public:
  var_type* m_pSrc;
  
 public:
  //virtual var_type* CalcResult (StatusFlagsAffected& flags) = 0;
  virtual word CalcResultConstrained (const StatusFlagsKnown& flags_known,  
				      VariableN<unsignedT>** out_lhs,  
				      VariableN<unsignedT>** out_result,  
				      StatusFlagsAffected& flags) = 0;
};

template<class unsignedT>
class VariableNBinaryOp : public VariableNOp<unsignedT>
{
 public:
  VARIABLE_N_BASE_TYPEDEF_VALUE_TYPES;
  VARIABLE_N_TYPEDEF_VAR_TYPES(VariableN        ,var_type);
  VARIABLE_N_TYPEDEF_VAR_TYPES(VariableNBinaryOp,var_type_binary_op);
  
 public:
  VariableNBinaryOp(var_type* pSrc1,var_type* pSrc2)
    {
      m_pSrc1 = pSrc1;
      m_pSrc2 = pSrc2;

      if (pSrc1)
	pSrc1->AddDepend(this);
      if (pSrc2)
	pSrc2->AddDepend(this);
    }
  
 public:
  var_type* m_pSrc1;
  var_type* m_pSrc2;
  
 public:
  //virtual var_type* CalcResult (StatusFlagsAffected& flags) = 0;
  virtual word CalcResultConstrained (const StatusFlagsKnown& flags_known,  
				      VariableN<unsignedT>** out_lhs,  
				      VariableN<unsignedT>** out_rhs,  
				      VariableN<unsignedT>** out_result,  
				      StatusFlagsAffected& flags) = 0;
};

#define VARIABLE_N_BINARY_OP_CLASS(opname) \
template<class unsignedT> \
class VariableNBinaryOp##opname : public VariableNBinaryOp<unsignedT> \
{ \
 public: \
  VARIABLE_N_BASE_TYPEDEF_VALUE_TYPES; \
  VARIABLE_N_TYPEDEF_VAR_TYPES(VariableN        ,var_type); \
  VARIABLE_N_TYPEDEF_VAR_TYPES(VariableNBinaryOp,var_type_binary_op); \
 public: \
  VariableNBinaryOp##opname(var_type* pSrc1,var_type* pSrc2) \
    : VariableNBinaryOp<unsignedT>(pSrc1,pSrc2) {  }; \
 \
 public: \
  virtual var_type* CalcResult (StatusFlagsAffected& flags); \
  virtual word CalcResultConstrained (const StatusFlagsKnown& flags_known, \
				      VariableN<unsignedT>** out_lhs,   \
				      VariableN<unsignedT>** out_rhs,   \
				      VariableN<unsignedT>** out_result,   \
				      StatusFlagsAffected& flags); \
 \
};

VARIABLE_N_BINARY_OP_CLASS(Add)
VARIABLE_N_BINARY_OP_CLASS(Sub)
VARIABLE_N_BINARY_OP_CLASS(Or)
VARIABLE_N_BINARY_OP_CLASS(And)
VARIABLE_N_BINARY_OP_CLASS(Xor)

#define VARIABLE_N_UNARY_OP_CLASS(opname) \
template<class unsignedT> \
class VariableNUnaryOp##opname : public VariableNUnaryOp<unsignedT> \
{ \
 public: \
  VARIABLE_N_BASE_TYPEDEF_VALUE_TYPES; \
  VARIABLE_N_TYPEDEF_VAR_TYPES(VariableN       ,var_type); \
  VARIABLE_N_TYPEDEF_VAR_TYPES(VariableNUnaryOp,var_type_unary_op); \
 public: \
  VariableNUnaryOp##opname(var_type* pSrc1) \
    : VariableNUnaryOp<unsignedT>(pSrc1) {  }; \
 \
 public: \
  virtual var_type* CalcResult (StatusFlagsAffected& flags); \
  virtual word CalcResultConstrained (const StatusFlagsKnown& flags_known,   \
					    VariableN<unsignedT>** out_lhs,   \
					    VariableN<unsignedT>** out_result,   \
					    StatusFlagsAffected& flags);  \
 \
};
     
VARIABLE_N_UNARY_OP_CLASS(Not)
VARIABLE_N_UNARY_OP_CLASS(Neg)
VARIABLE_N_UNARY_OP_CLASS(Inc)
VARIABLE_N_UNARY_OP_CLASS(Dec)








/*
template<class unsignedT>
class VariableNTreatment : public VariableTreatment
{


public:
	virtual Variable* CalcOpResult (short m_nInstruction,Variable* pDest,Variable* pSrc) const
	{
		printf ("Instr: %5d:  %08x %08x\n",m_nInstruction,pDest,pSrc);
		
		return NULL;
	}



};
*/

Variable* CalcOpResult (int nSize,
			short nInstruction,
			Variable* pSrc1,Variable* pSrc2,
			StatusFlagsAffected& flags);

Variable* CalcOpResult (int nSize,
			short nInstruction,
			Variable* pSrc1,
			StatusFlagsAffected& flags);

Variable* CalcOpResultZeroIdenticalArgs (int nSize,
					 short nInstruction,
					 StatusFlagsAffected& flags);

Variable* CalcMemAddr (Variable* pVarBase,
		       Variable* pVarIndex,
		       const OperandMemory* pOpMem);

Variable* CalcAddToVar(Variable* pVar,dword dwAdd,int nSize);

#endif // !defined(__VARIABLE_OP_H__INCLUDED_)
