/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// VariableOp_CalcResult.cpp: 
//
//////////////////////////////////////////////////////////////////////

#include "VariableOp_CalcResult.h" 

//////////////////////////////////////////////////////////////////////

// This is not intended to be good assembly language (I'm not used to it)
// It's just suppose to work
// But it was simpler writing this than writing C logic to figure out the flags
// without actually calling the instruction, besides, this doesn't jump!
//
// Plan:  Take incoming values from pointers 'lhs' and 'rhs'
//        Put them in eax/ax/al and abx/bx/bl
//        Make the operation
//        Push resulting flags16 onto stack
//        Place result of operation (eax/ax/al) into value pointed to by 'result'
//        Pop flags16 to value pointed to by 'flags'
//
// Using: EAX and EBX, restored to original values before return
//        And some stack

//////////////////////////////////////////////////////////////////////

#ifdef _MSC_VER // masm version (unfortunately overgrown)

#define BINARY_CALC_RESULT_WITH_FLAGS_TYPE(opname,type,reg_a,reg_b) \
void CalcResultWithFlags_##opname##_##type(const type* lhs,const type* rhs,type* result,word* flags) \
{ \
  __asm push eax                  /* save */ \
  __asm push ebx                   \
  __asm mov ebx,dword ptr[lhs]    /* get operands */ \
  __asm mov reg_a,type ptr [ebx]   \
  __asm mov ebx,dword ptr[rhs]     \
  __asm mov reg_b,type ptr [ebx]   \
  __asm opname reg_a,reg_b        /* perform operation */ \
  __asm pushf                      \
  __asm mov ebx,dword ptr[result] /* store result & flags */ \
  __asm mov type ptr [ebx],reg_a   \
  __asm mov ebx,dword ptr[flags]   \
  __asm pop word ptr [ebx]         \
  __asm pop ebx                   /* restore */ \
  __asm pop eax                    \
} \
template<> const calc_result_with_flags_type_info<type>::BinaryCalcResultWithFlagsFnc calc_result_with_flags_type_info<type>::CalcResultWithFlags_##opname = CalcResultWithFlags_##opname##_##type;

#define UNARY_CALC_RESULT_WITH_FLAGS_TYPE(opname,type,reg_a,reg_b) \
void CalcResultWithFlags_##opname##_##type(const type* lhs,type* result,word* flags) \
{ \
  __asm push eax                  /* save */ \
  __asm push ebx                   \
  __asm mov ebx,dword ptr[lhs]    /* get operands */ \
  __asm mov reg_a,type ptr [ebx]   \
  __asm opname reg_a              /* perform operation */ \
  __asm pushf                      \
  __asm mov ebx,dword ptr[result] /* store result & flags */ \
  __asm mov type ptr [ebx],reg_a   \
  __asm mov ebx,dword ptr[flags]   \
  __asm pop word ptr [ebx]         \
  __asm pop ebx                   /* restore */ \
  __asm pop eax                    \
} \
template<> const calc_result_with_flags_type_info<type>::UnaryCalcResultWithFlagsFnc calc_result_with_flags_type_info<type>::CalcResultWithFlags_##opname = CalcResultWithFlags_##opname##_##type;

#define BINARY_CALC_RESULT_WITH_FLAGS(opname) \
BINARY_CALC_RESULT_WITH_FLAGS_TYPE(opname,byte,al,bl) \
BINARY_CALC_RESULT_WITH_FLAGS_TYPE(opname,word,ax,bx) \
BINARY_CALC_RESULT_WITH_FLAGS_TYPE(opname,dword,eax,ebx)

#define UNARY_CALC_RESULT_WITH_FLAGS(opname) \
UNARY_CALC_RESULT_WITH_FLAGS_TYPE(opname,byte,al,bl) \
UNARY_CALC_RESULT_WITH_FLAGS_TYPE(opname,word,ax,bx) \
UNARY_CALC_RESULT_WITH_FLAGS_TYPE(opname,dword,eax,ebx)

//////////////////////////////////////////////////////////////////////

#elif __GNUC__ // gas version

#define BINARY_CALC_RESULT_WITH_FLAGS_TYPE(opname,type) \
void CalcResultWithFlags_##opname##_##type(const type* lhs,const type* rhs,type* result,word* flags) \
{ \
  __asm__ ("mov %2,%1\n\t" \
           #opname " %3,%1\n\t" \
           "pushfw\n\t" \
           "popw %0\n\t" \
           : "=q"(*flags),"=q"(*result) \
           : "q"(*lhs),"q"(*rhs) ); \
} \
template<> const calc_result_with_flags_type_info<type>::BinaryCalcResultWithFlagsFnc calc_result_with_flags_type_info<type>::CalcResultWithFlags_##opname = CalcResultWithFlags_##opname##_##type;

#define UNARY_CALC_RESULT_WITH_FLAGS_TYPE(opname,type) \
void CalcResultWithFlags_##opname##_##type(const type* lhs,type* result,word* flags) \
{ \
  __asm__ ("mov %2,%1\n\t" \
           #opname " %1\n\t" \
           "pushfw\n\t" \
           "popw %0\n\t" \
           : "=q"(*flags),"=q"(*result) \
           : "q"(*lhs) ); \
} \
template<> const calc_result_with_flags_type_info<type>::UnaryCalcResultWithFlagsFnc calc_result_with_flags_type_info<type>::CalcResultWithFlags_##opname = CalcResultWithFlags_##opname##_##type;

#define BINARY_CALC_RESULT_WITH_FLAGS(opname) \
BINARY_CALC_RESULT_WITH_FLAGS_TYPE(opname,byte) \
BINARY_CALC_RESULT_WITH_FLAGS_TYPE(opname,word) \
BINARY_CALC_RESULT_WITH_FLAGS_TYPE(opname,dword)

#define UNARY_CALC_RESULT_WITH_FLAGS(opname) \
UNARY_CALC_RESULT_WITH_FLAGS_TYPE(opname,byte) \
UNARY_CALC_RESULT_WITH_FLAGS_TYPE(opname,word) \
UNARY_CALC_RESULT_WITH_FLAGS_TYPE(opname,dword)

#endif /* neither _MSC_VER nor __GNUC__ */

//////////////////////////////////////////////////////////////////////

BINARY_CALC_RESULT_WITH_FLAGS(add)
BINARY_CALC_RESULT_WITH_FLAGS(sub)
BINARY_CALC_RESULT_WITH_FLAGS(and)
BINARY_CALC_RESULT_WITH_FLAGS(or)
BINARY_CALC_RESULT_WITH_FLAGS(xor)

UNARY_CALC_RESULT_WITH_FLAGS(neg)
UNARY_CALC_RESULT_WITH_FLAGS(inc)
UNARY_CALC_RESULT_WITH_FLAGS(dec)

