/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// VariableSimplify.cpp: implementation of the VariableOp class.
//
//////////////////////////////////////////////////////////////////////

#include "VariableSimplify.h" 

#include "VariableNConst.h" 
#include "VariableNConsts.h"
#include "VariableNRange.h" 
#include "VariableNPattern.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

template<class unsignedT>
VariableN<unsignedT>* VariableNConsts<unsignedT>::SimplifyVariable()
{
  assert (m_vectValues.size());
  if (m_vectValues.size () == 1)
    return new VariableNConst<unsignedT>(m_vectValues[0]);
  if (m_vectValues.back () - m_vectValues.front () == 
      (unsignedT) (m_vectValues.size() - 1))
    return new VariableNRange<unsignedT>(m_vectValues.front(),
					 m_vectValues.back());
  // it's perhaps an wrapped range
  uint i = 0;
  while (m_vectValues[i+1] == m_vectValues[i]+1)
    i++;
  if (i < m_vectValues.size() - 1 &&
      m_vectValues[i+1] == 
      ALL_ONES(unsignedT) - (m_vectValues.size() - (i+1)) + 1)
    {
      // it's an wrapped range
      return new VariableNRange<unsignedT>(m_vectValues[i+1],m_vectValues[i]);
    }
  return NULL;
}

template<class unsignedT>
VariableN<unsignedT>* VariableNRange<unsignedT>::SimplifyVariable()
{
  if (m_min == m_max + 1)
    return new VariableNUnknown<unsignedT>();
  if (m_min == m_max)
    return new VariableNConst<unsignedT>(m_min);
  return NULL;
}

template<class unsignedT>
VariableN<unsignedT>* VariableNPattern<unsignedT>::SimplifyVariable()
{
  if (m_known == 0)
    return new VariableNUnknown<unsignedT>();
  if (m_known == ALL_ONES(unsignedT))
    return new VariableNConst<unsignedT>(m_value);
  return NULL;
}

// force instantiation
#define VARIABLE_N_SIMPLIFIERS(type) \
template VariableN<type>* VariableNConsts<type>::SimplifyVariable(); \
template VariableN<type>* VariableNRange<type>::SimplifyVariable(); \
template VariableN<type>* VariableNPattern<type>::SimplifyVariable(); 

VARIABLE_N_SIMPLIFIERS(byte)
VARIABLE_N_SIMPLIFIERS(word)
VARIABLE_N_SIMPLIFIERS(dword)
  
#ifdef _MSC_VER
  // because the compiler wont instanciate on above expressions
  void dummy_function_for_instanciation()
{
#define VARIABLE_N_SIMPLIFIERS_DUMMY_CALL(classname,type) \
{ classname<type> dummy(0); /*dummy.SimplifyVariable();*/ }
  
#define VARIABLE_N_SIMPLIFIERS_DUMMY_CALLS(type) \
VARIABLE_N_SIMPLIFIERS_DUMMY_CALL(VariableNConsts,type) \
VARIABLE_N_SIMPLIFIERS_DUMMY_CALL(VariableNRange,type) \
VARIABLE_N_SIMPLIFIERS_DUMMY_CALL(VariableNPattern,type)

VARIABLE_N_SIMPLIFIERS_DUMMY_CALLS(byte)
VARIABLE_N_SIMPLIFIERS_DUMMY_CALLS(word)
VARIABLE_N_SIMPLIFIERS_DUMMY_CALLS(dword)
}
#endif // _MSC_VER > 1000

