/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// VerifyEncoderDecoder.h: 
//
//////////////////////////////////////////////////////////////////////

#if !defined(__VERIFY_ENCODER_DECODER_H__INCLUDED_)
#define __VERIFY_ENCODER_DECODER_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "InstructionDecoder.h" // dep_chk_needed     
#include "InstructionEncoder.h" // dep_chk_needed     

#include "DataPtrSrc.h" // dep_chk_needed     

void VerifyEncoderDecoder();

void VerifySameInstruction(const Instruction* pInstr1,
			   const Instruction* pInstr2);

class TestCoders
{
 public:
  virtual ~TestCoders() { }
  
 public:
  InstructionDecoder m_decoder;
  InstructionEncoder m_encoder;
  
 public:
  void Test();
  
 public:
  virtual void DoTest() = 0;
  
 protected:
  void VerifyDecodeIntoSame (const Instruction* pInstr,const byte* cCode,const byte bLength);
  
};

//////////////////////////////////////////////////////////////////////

class Rnd
{
  // used to generate pseudo-random numbers so that we can test
  // different arguments. Base seed on instruction number, so it does
  // not change when inserting more instructions. This way the tested
  // arguments will only change when varying the argument generators
  
  // there is no need for the generator to be truly ranodm, but it
  // would be nice if it can generate all possible numbers
  // (i.e. complete)
  
 public:
  Rnd(dword state) { assert (state); m_state = state; }
  
 public:
  long m_state;
  
 protected:
  long r();
  
 public:
  byte rbyte() { return (byte) (r() >> 23); }
  word rword() { return (word) (r() >> 15); };
  dword rdword() { return ((dword) r() ^ (((dword) r()) << 1)); } // because we only get 31 bits
  
  byte rbyte(byte max,byte maxlog2) { return (byte) rword (max,maxlog2); }
  word rword(word max,byte maxlog2);
  
  byte rbyte(byte max) { return (byte) rword (max); } // max must be power of 2
  word rword(word max); // max must be power of 2
  
  byte rbyte_mask(byte mask);
  
};

//////////////////////////////////////////////////////////////////////

#define ARG_INFO_IMM_8				0x00000001
#define ARG_INFO_IMM_16				0x00000002
#define ARG_INFO_IMM_32				0x00000004
#define ARG_INFO_IMM_16_32			(ARG_INFO_IMM_16 | ARG_INFO_IMM_32)
#define ARG_INFO_IMM_HI16_16_32		0x00000008
#define ARG_INFO_REG_STD_8			0x00000010
#define ARG_INFO_REG_STD_16			0x00000020
#define ARG_INFO_REG_STD_32			0x00000040
#define ARG_INFO_REG_STD_16_32		(ARG_INFO_REG_STD_16 | ARG_INFO_REG_STD_32)
#define ARG_INFO_REG_SEG			0x00000100
#define ARG_INFO_REG_DBG_CTRL_TST	0x00000200
#define ARG_INFO_REG_ST				0x00000400
#define ARG_INFO_REG_MM				0x00000800
#define ARG_INFO_REG_XMM			0x00001000
#define ARG_INFO_MEM_8				0x00010000
#define ARG_INFO_MEM_16				0x00020000
#define ARG_INFO_MEM_32				0x00040000
#define ARG_INFO_MEM_16_32			(ARG_INFO_MEM_16 | ARG_INFO_MEM_32)
#define ARG_INFO_MEM_64				0x00080000
#define ARG_INFO_MEM_128			0x00100000
#define ARG_INFO_MEM_OTHER			0x00200000
#define ARG_INFO_MEM_SI_DI			0x00400000
#define ARG_INFO_NO_OPERAND			0x80000000

struct InstrArgOperand
{
 public:
  InstrArgOperand(Operand* pOperand,bool bWrongOp)
  {
    m_pOperand = pOperand;
    m_bWrongOp = bWrongOp;
  }
  
 public:
  Operand*	m_pOperand;
  bool		m_bWrongOp; // operand that should not be encodable for instruction
};

typedef vector<InstrArgOperand> VectorInstrArgOperand;


class InstrArgInfo
{
 public:
  InstrArgInfo()
    {
      dwDestArg = dwSrcArg = dwSrc2Arg = 0;
      wMemOtherSize[0] = wMemOtherSize[1] = 0;
    }
  
 public:
  void GenerateOperands(VectorInstrArgOperand& vectOperandsDest,
			VectorInstrArgOperand& vectOperandsSrc,
			VectorInstrArgOperand& vectOperandsSrc2,
			dword dwRndSeed);
  
 protected:
  void GenerateOperands(VectorInstrArgOperand& vectOperands,
			Rnd& rnd,dword dwArg);
  void GenerateWrongOperands(VectorInstrArgOperand& vectOperands,
			     Rnd& rnd,dword dwArg);
  
 protected:
  void GenerateMemoryOps(VectorInstrArgOperand& vectOperands,
			 Rnd& rnd,bool bWrongOp,
			 byte nNum,word nOpSize,bool bSIDI = false);
  void GenerateRegisterOps(VectorInstrArgOperand& vectOperands,
			   Rnd& rnd,bool bWrongOp,
			   byte bRegType,byte nNum,byte bMask = 0xff);
  
 protected:
  void PushOperand(VectorInstrArgOperand& vectOperands,
		   Operand* pOperand,bool bWrongOp);
  
 public:
  void AddMemOtherSize (word wSize);
  
 public:
  dword dwDestArg;
  dword dwSrcArg;
  dword dwSrc2Arg;
  
  word wMemOtherSize[2];
};

#include <map>

using namespace std;

typedef map<word,InstrArgInfo> MapInstrArgInfo;

class TestEncoderDecoderInstrs : public TestCoders
{
 public:
  virtual ~TestEncoderDecoderInstrs() { }

 public:
  MapInstrArgInfo mapInstrArgInfo;
  
 public:
  virtual void DoTest();
  
 protected:
  void AddArgTypes(const OpcodeInfo& info);
  void AddArgTypes(const OpcodeInfo& info,word nInstr);
};

//////////////////////////////////////////////////////////////////////

class TestDecoderEncoder : public TestCoders
{
 public:
  TestDecoderEncoder();
  virtual ~TestDecoderEncoder() { }
  
 public:
  
  byte m_cCode[MAX_INSTR_CODE_LENGTH];
  DataPtrSrc m_code;
  
 public:
  // used when looping:
  
  int m_nByte;
  int m_nTwoByte;
  
  bool m_bAddrSize32; // prefix included
  
  void TestDispImm(byte* bCode,const byte** pEndDecode,
		   bool bOutputInstrCode = false);
  void TestByte(byte* bCode);
  void TestMMXSSEPrefix(byte* bCode);
  void TestAddrSizePrefix(byte* bCode);
  void TestOpSizePrefix(byte* bCode);
  void TestSegmentPrefix(byte* bCode);
  void TestLockRepPrefix(byte* bCode);
  
  bool TestDecodeEncode(const byte** pEndDecode,bool bOutputInstrCode = false);
  
 public:
  virtual void DoTest();
  

};

#endif // !defined(__VERIFY_ENCODER_DECODER_H__INCLUDED_)
