/*  exmhcg - executable program tuner  
 *  Copyright (C) 2003  Hkan Johansson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// WriteRelocDest.cpp: implementation of the WriteRelocDest class.
//
//////////////////////////////////////////////////////////////////////

#include "WriteRelocDest.h"

#include "RelocatableDest.h"
#include "DataDestRelocPtr.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

WriteRelocDest::WriteRelocDest()
{
  m_pRelocDest = NULL;
  m_pDest = NULL;
}

WriteRelocDest::~WriteRelocDest()
{
  delete m_pRelocDest;
  delete m_pDest;
}

void WriteRelocDest::AllocMemory(dword dwSize)
{
  assert (!m_pRelocDest);
  assert (!m_pDest);

  m_pRelocDest = new RelocatableDest();

  if (dwSize)
    m_pRelocDest->AllocDataBuffer(dwSize); 
  
  m_pDest = new DataDestRelocPtr(m_pRelocDest);
}

void WriteRelocDest::AssureDestMemory(dword dwFreeSize)
{
  // if we do not have the size requested, allocate twice the size
  // requested extra
  
  dword dwOffset = m_pDest ? m_pDest->GetOffset() : 0;
  
  if (!m_pRelocDest)
    m_pRelocDest = new RelocatableDest();
  
  if (m_pRelocDest->m_dwSize - dwOffset < dwFreeSize)
    {
      // so we have too little memory
      
      m_pRelocDest->AllocDataBuffer(dwOffset + dwFreeSize * 2); 
    }
  else
    return; // nothing to do (enough memory)
  
  delete m_pDest;
  m_pDest = new DataDestRelocPtr(m_pRelocDest);
  m_pDest->SetOffset(dwOffset);  
}

RelocatableDest* WriteRelocDest::ReleaseWrite()
{
  RelocatableDest* pReturn = m_pRelocDest; 
  m_pRelocDest = NULL; // because we are no longer responsible for deletion
  return pReturn;
}

void WriteRelocDest::TrimMemoryUsage()
{
  // although we return our contents inside an RelocatableDest, we
  // cannot be fixed up if any of the sections that point to us are
  // relocated (relative to the image base)
  
  // trim memory usage down to the needed one
  
  dword dwOffset = m_pDest ? m_pDest->GetOffset() : 0;
  
  if (!dwOffset)
    {
      delete m_pRelocDest; m_pRelocDest = NULL;
      delete m_pDest;      m_pDest = NULL;       
      return; // no relocations
    }
  
  assert (m_pRelocDest);
  
  m_pRelocDest->AllocDataBuffer(dwOffset);
}

RelocatableDest* WriteRelocDest::EndWrite()    // finalize
{
  TrimMemoryUsage();
  return ReleaseWrite();
}

